Package: tidylearn
Title: A Unified Tidy Interface to R's Machine Learning Ecosystem
Version: 0.5.0
Authors@R: 
    person("Cesaire", "Tobias", email = "cesaire@sheetsolved.com", role = c("aut", "cre"))
Description: Provides a unified tidyverse-compatible interface to R's machine
    learning ecosystem - from data ingestion to model publishing. The tl_read()
    family reads data from files ('CSV', 'Excel', 'Parquet', 'JSON'), databases
    ('SQLite', 'PostgreSQL', 'MySQL', 'BigQuery'), and cloud sources ('S3',
    'GitHub', 'Kaggle'). The tl_model() function wraps established
    implementations from 'glmnet', 'randomForest', 'xgboost', 'e1071', 'rpart',
    'gbm', 'nnet', 'cluster', 'dbscan', and others with consistent function
    signatures and tidy tibble output. Results flow into unified 'ggplot2'-based
    visualization and optional formatted 'gt' tables via the tl_table() family.
    The underlying algorithms are unchanged; 'tidylearn' simply makes them
    easier to use together. Access raw model objects via the $fit slot for a
    supervised method, or $fit$model for an unsupervised one.
    Methods include random forests Breiman (2001) <doi:10.1023/A:1010933404324>,
    LASSO regression Tibshirani (1996) <doi:10.1111/j.2517-6161.1996.tb02080.x>,
    elastic net Zou and Hastie (2005) <doi:10.1111/j.1467-9868.2005.00503.x>,
    support vector machines Cortes and Vapnik (1995) <doi:10.1007/BF00994018>,
    and gradient boosting Friedman (2001) <doi:10.1214/aos/1013203451>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 3.6.0)
Imports: dplyr (>= 1.0.0), ggplot2 (>= 3.3.0), tibble (>= 3.0.0), tidyr
        (>= 1.0.0), purrr (>= 0.3.0), rlang (>= 0.4.0), magrittr,
        parallel, stats, e1071, gbm, glmnet, nnet, randomForest, rpart,
        rsample, ROCR, yardstick, cluster (>= 2.1.0), dbscan (>=
        1.1.0), MASS, smacof (>= 2.1.0)
Suggests: arules, arulesViz, bigrquery, car, DBI, DiagrammeR, DT,
        GGally, ggforce, gridExtra, gt, httr2 (>= 1.0.0), jsonlite,
        keras, knitr, lmtest, moments, nanoparquet, NeuralNetTools,
        paws.storage, readr, RhpcBLASctl, readxl, RMariaDB, rmarkdown,
        RPostgres, rpart.plot, RSQLite, shiny, shinydashboard,
        tensorflow, testthat (>= 3.1.7), withr, xgboost
Config/testthat/edition: 3
URL: https://tidylearn.sheetsolved.com,
        https://github.com/ces0491/tidylearn
BugReports: https://github.com/ces0491/tidylearn/issues
VignetteBuilder: knitr
Collate: 'tidylearn-package.R' 'utils.R' 'read.R' 'read-backends.R'
        'core.R' 'preprocessing.R' 'supervised-classification.R'
        'supervised-regression.R' 'supervised-regularization.R'
        'supervised-trees.R' 'supervised-svm.R'
        'supervised-neural-networks.R' 'supervised-deep-learning.R'
        'supervised-xgboost.R' 'unsupervised-distance.R'
        'unsupervised-pca.R' 'unsupervised-mds.R'
        'unsupervised-clustering.R' 'unsupervised-hclust.R'
        'unsupervised-dbscan.R' 'unsupervised-market-basket.R'
        'unsupervised-validation.R' 'integration.R' 'pipeline.R'
        'model-selection.R' 'compute-detection.R' 'compute-advisor.R'
        'compute-routing.R' 'cloud-consent.R' 'cloud-endpoint.R'
        'cloud-cost.R' 'cloud-serialize.R' 'tuning.R' 'interactions.R'
        'diagnostics.R' 'metrics.R' 'visualization.R' 'tables.R'
        'workflows.R'
NeedsCompilation: no
Packaged: 2026-09-02 05:57:17 UTC; CesaireTobias
Author: Cesaire Tobias [aut, cre]
Maintainer: Cesaire Tobias <cesaire@sheetsolved.com>
Repository: CRAN
Date/Publication: 2026-09-02 06:20:02 UTC
Built: R 4.5.2; ; 2026-09-02 08:16:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,tidylearn_eda)
S3method(plot,tidylearn_model)
S3method(predict,tidylearn_model)
S3method(predict,tidylearn_stratified)
S3method(predict,tidylearn_transfer)
S3method(print,tidy_apriori)
S3method(print,tidy_dbscan)
S3method(print,tidy_gap)
S3method(print,tidy_hclust)
S3method(print,tidy_kmeans)
S3method(print,tidy_mds)
S3method(print,tidy_pam)
S3method(print,tidy_pca)
S3method(print,tidy_silhouette)
S3method(print,tidylearn_automl)
S3method(print,tidylearn_compute_advice)
S3method(print,tidylearn_data)
S3method(print,tidylearn_eda)
S3method(print,tidylearn_gpu_check)
S3method(print,tidylearn_model)
S3method(print,tidylearn_pipeline)
S3method(summary,tidylearn_model)
S3method(summary,tidylearn_pipeline)
S3method(tl_compute_advisor,character)
S3method(tl_compute_advisor,default)
S3method(tl_compute_advisor,tidylearn_supervised)
export("%>%")
export(augment_dbscan)
export(augment_hclust)
export(augment_kmeans)
export(augment_pam)
export(augment_pca)
export(calc_validation_metrics)
export(calc_wss)
export(compare_clusterings)
export(compare_distances)
export(create_cluster_dashboard)
export(explore_dbscan_params)
export(filter_rules_by_item)
export(find_related_items)
export(get_pca_loadings)
export(get_pca_variance)
export(inspect_rules)
export(optimal_clusters)
export(optimal_hclust_k)
export(plot_cluster_comparison)
export(plot_cluster_sizes)
export(plot_clusters)
export(plot_dendrogram)
export(plot_distance_heatmap)
export(plot_elbow)
export(plot_gap_stat)
export(plot_knn_dist)
export(plot_mds)
export(plot_silhouette)
export(plot_variance_explained)
export(recommend_products)
export(standardize_data)
export(suggest_eps)
export(summarize_rules)
export(tidy_apriori)
export(tidy_clara)
export(tidy_cutree)
export(tidy_dbscan)
export(tidy_dendrogram)
export(tidy_dist)
export(tidy_gap_stat)
export(tidy_gower)
export(tidy_hclust)
export(tidy_kmeans)
export(tidy_knn_dist)
export(tidy_mds)
export(tidy_mds_classical)
export(tidy_mds_kruskal)
export(tidy_mds_sammon)
export(tidy_mds_smacof)
export(tidy_pam)
export(tidy_pca)
export(tidy_pca_biplot)
export(tidy_pca_screeplot)
export(tidy_rules)
export(tidy_silhouette)
export(tidy_silhouette_analysis)
export(tl_add_cluster_features)
export(tl_anomaly_aware)
export(tl_auto_interactions)
export(tl_auto_ml)
export(tl_calc_classification_metrics)
export(tl_check_assumptions)
export(tl_check_gpu)
export(tl_cloud_allow_host)
export(tl_cloud_allowed_hosts)
export(tl_cloud_consent)
export(tl_cloud_jobs)
export(tl_compare_cv)
export(tl_compare_pipeline_models)
export(tl_compute_advisor)
export(tl_cv)
export(tl_dashboard)
export(tl_default_param_grid)
export(tl_detect_outliers)
export(tl_diagnostic_dashboard)
export(tl_evaluate)
export(tl_explore)
export(tl_get_best_model)
export(tl_influence_measures)
export(tl_interaction_effects)
export(tl_load_pipeline)
export(tl_model)
export(tl_pipeline)
export(tl_plot_cv_comparison)
export(tl_plot_cv_results)
export(tl_plot_deep_architecture)
export(tl_plot_deep_history)
export(tl_plot_gain)
export(tl_plot_importance_comparison)
export(tl_plot_importance_regularized)
export(tl_plot_influence)
export(tl_plot_interaction)
export(tl_plot_intervals)
export(tl_plot_lift)
export(tl_plot_model_comparison)
export(tl_plot_nn_architecture)
export(tl_plot_nn_tuning)
export(tl_plot_partial_dependence)
export(tl_plot_regularization_cv)
export(tl_plot_regularization_path)
export(tl_plot_svm_boundary)
export(tl_plot_svm_tuning)
export(tl_plot_tree)
export(tl_plot_tuning_results)
export(tl_plot_xgboost_importance)
export(tl_plot_xgboost_shap_dependence)
export(tl_plot_xgboost_shap_summary)
export(tl_plot_xgboost_tree)
export(tl_predict_pipeline)
export(tl_prepare_data)
export(tl_read)
export(tl_read_bigquery)
export(tl_read_csv)
export(tl_read_db)
export(tl_read_dir)
export(tl_read_excel)
export(tl_read_github)
export(tl_read_json)
export(tl_read_kaggle)
export(tl_read_mysql)
export(tl_read_parquet)
export(tl_read_postgres)
export(tl_read_rdata)
export(tl_read_rds)
export(tl_read_s3)
export(tl_read_sqlite)
export(tl_read_tsv)
export(tl_read_zip)
export(tl_reduce_dimensions)
export(tl_run_pipeline)
export(tl_save_pipeline)
export(tl_semisupervised)
export(tl_split)
export(tl_step_selection)
export(tl_stratified_models)
export(tl_table)
export(tl_table_clusters)
export(tl_table_coefficients)
export(tl_table_comparison)
export(tl_table_confusion)
export(tl_table_importance)
export(tl_table_loadings)
export(tl_table_metrics)
export(tl_table_variance)
export(tl_test_interactions)
export(tl_test_model_difference)
export(tl_transfer_learning)
export(tl_tune_deep)
export(tl_tune_grid)
export(tl_tune_nn)
export(tl_tune_random)
export(tl_tune_xgboost)
export(tl_version)
export(tl_xgboost_shap)
export(visualize_rules)
importFrom(ROCR,performance)
importFrom(ROCR,prediction)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,bind_cols)
importFrom(dplyr,desc)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,summarize)
importFrom(dplyr,tibble)
importFrom(e1071,svm)
importFrom(e1071,tune)
importFrom(gbm,gbm)
importFrom(gbm,predict.gbm)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_flip)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_bar)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_contour)
importFrom(ggplot2,geom_density)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_jitter)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_ribbon)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_smooth)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_tile)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_color_gradient)
importFrom(ggplot2,scale_fill_gradient)
importFrom(ggplot2,scale_fill_gradient2)
importFrom(ggplot2,scale_x_log10)
importFrom(ggplot2,stat_qq)
importFrom(ggplot2,stat_qq_line)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,theme_void)
importFrom(glmnet,cv.glmnet)
importFrom(glmnet,glmnet)
importFrom(glmnet,predict.glmnet)
importFrom(magrittr,"%>%")
importFrom(nnet,nnet)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_dbl)
importFrom(purrr,map_lgl)
importFrom(randomForest,importance)
importFrom(randomForest,randomForest)
importFrom(rlang,.data)
importFrom(rlang,.env)
importFrom(rpart,rpart)
importFrom(rpart,rpart.control)
importFrom(stats,AIC)
importFrom(stats,BIC)
importFrom(stats,anova)
importFrom(stats,aov)
importFrom(stats,as.formula)
importFrom(stats,bartlett.test)
importFrom(stats,binomial)
importFrom(stats,coef)
importFrom(stats,cooks.distance)
importFrom(stats,cor)
importFrom(stats,dfbetas)
importFrom(stats,dffits)
importFrom(stats,fitted)
importFrom(stats,formula)
importFrom(stats,glm)
importFrom(stats,hatvalues)
importFrom(stats,influence.measures)
importFrom(stats,kruskal.test)
importFrom(stats,lm)
importFrom(stats,lm.influence)
importFrom(stats,median)
importFrom(stats,model.matrix)
importFrom(stats,p.adjust)
importFrom(stats,poly)
importFrom(stats,predict)
importFrom(stats,qqnorm)
importFrom(stats,reorder)
importFrom(stats,residuals)
importFrom(stats,rstandard)
importFrom(stats,rstudent)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,shapiro.test)
importFrom(stats,step)
importFrom(stats,t.test)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(stats,var)
importFrom(stats,wilcox.test)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tidyr,crossing)
importFrom(tidyr,nest)
importFrom(tidyr,unnest)
importFrom(utils,combn)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,packageVersion)
importFrom(yardstick,accuracy)
importFrom(yardstick,f_meas)
importFrom(yardstick,mae)
importFrom(yardstick,mape)
importFrom(yardstick,pr_auc)
importFrom(yardstick,precision)
importFrom(yardstick,recall)
importFrom(yardstick,rmse)
importFrom(yardstick,roc_auc)
importFrom(yardstick,rsq)

# tidylearn <img src="man/figures/logo.png" align="right" height="139" alt="tidylearn logo" />

Machine Learning for Tidynauts

[![CRAN status](https://www.r-pkg.org/badges/version/tidylearn)](https://cran.r-project.org/package=tidylearn)
[![R-CMD-check](https://github.com/ces0491/tidylearn/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ces0491/tidylearn/actions/workflows/R-CMD-check.yaml)
[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/licenses/MIT)

## Overview

`tidylearn` provides a **unified tidyverse-compatible interface** to R's machine
learning ecosystem. It wraps proven packages like glmnet, randomForest,
xgboost, e1071, cluster, and dbscan - you get the reliability of established
implementations with the convenience of a consistent, tidy API.

**What tidylearn does:**

- Reads data from diverse sources (`tl_read()`) — CSV, Excel, Parquet,
  JSON, databases, S3, Kaggle, and more
- Provides one consistent interface (`tl_model()`) to 20 ML algorithms
- Returns tidy tibbles instead of varied output formats
- Offers unified ggplot2-based visualization and formatted `gt` tables
- Enables pipe-friendly workflows with `%>%`
- Orchestrates complex workflows combining multiple techniques

**What tidylearn is NOT:**

- A reimplementation of ML algorithms (uses established packages under the hood)
- A replacement for the underlying packages (you can access the raw model via
  `model$fit`)

## Why tidylearn?

Each ML package in R has its own API, output format, and conventions. tidylearn
provides a translation layer so you can:

| Without tidylearn                     | With tidylearn           |
| ------------------------------------- | -----------------------  |
| Learn different APIs for each package | One API for everything   |
| Write custom code to extract results  | Consistent tibble output |
| Create different plots for each model | Unified visualization    |
| Manage package-specific quirks        | Focus on your analysis   |

The underlying algorithms are unchanged - tidylearn simply makes them easier to
use together.

## Installation

```r
# Install from CRAN
install.packages("tidylearn")

# Or install development version from GitHub
# devtools::install_github("ces0491/tidylearn")
```

## Quick Start

### Data Ingestion

`tl_read()` auto-detects the format and returns a tidy `tidylearn_data` object:

```r
library(tidylearn)

# Single files — format auto-detected from extension
data <- tl_read("sales.csv")
data <- tl_read("results.xlsx", sheet = "Q1")
data <- tl_read("experiment.parquet")

# Databases
data <- tl_read_sqlite("warehouse.sqlite", "SELECT * FROM sales")
data <- tl_read_postgres("localhost", query = "SELECT * FROM customers",
                         dbname = "analytics", user = "me")

# Cloud and API sources
data <- tl_read_s3("s3://my-bucket/data.csv")
data <- tl_read_kaggle("zillow/zecon", file = "Zip_time_series.csv")

# Multi-file reading
data <- tl_read(c("jan.csv", "feb.csv", "mar.csv"))
data <- tl_read_dir("data/monthly/", format = "csv")
data <- tl_read_zip("download.zip")
```

### Unified Interface

A single `tl_model()` function dispatches to the appropriate underlying package:

```r
library(tidylearn)

# Classification -> uses randomForest::randomForest()
model <- tl_model(iris, Species ~ ., method = "forest")

# Regression -> uses stats::lm()
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

# Regularization -> uses glmnet::glmnet()
model <- tl_model(mtcars, mpg ~ ., method = "lasso")

# Clustering -> uses stats::kmeans()
model <- tl_model(iris[,1:4], method = "kmeans", k = 3)

# PCA -> uses stats::prcomp()
model <- tl_model(iris[,1:4], method = "pca")
```

### Tidy Output

All results come back as tibbles, ready for dplyr and ggplot2:

```r
# Predictions come back as a tibble with a .pred column
predictions <- predict(model, new_data = test_data)

# Metrics as tibbles - pick the metrics you want
metrics <- tl_evaluate(model, test_data)
metrics <- tl_evaluate(model, test_data, metrics = c("rmse", "rsq"))

# Easy to pipe
model %>%
  predict(new_data = test_data) %>%
  bind_cols(test_data) %>%
  ggplot(aes(x = mpg, y = .pred)) +
  geom_point() +
  geom_abline(slope = 1, intercept = 0)
```

For classification, what `.pred` holds depends on `type`: `"class"` gives
labels, `"prob"` gives one column per class. The default `"response"` varies
by method - probabilities for logistic regression, labels for trees and
forests - so pass `type` explicitly, or let `tl_evaluate()` handle it.

### Access the Underlying Model

You always have access to the raw model from the underlying package:

```r
model <- tl_model(iris, Species ~ ., method = "forest")

# Access the randomForest object directly
model$fit  # This is the randomForest::randomForest() result

# Use package-specific functions if needed
randomForest::varImpPlot(model$fit)
```

## Wrapped Packages

tidylearn provides a unified interface to these established R packages:

### Supervised Learning

| Method | Underlying Package | Function Called |
| -------- | ------------------- | ----------------- |
| `"linear"` | stats | `lm()` |
| `"polynomial"` | stats | `lm()` with `poly()` |
| `"logistic"` | stats | `glm(..., family = binomial)` |
| `"ridge"`, `"lasso"`, `"elastic_net"` | glmnet | `glmnet()` |
| `"tree"` | rpart | `rpart()` |
| `"forest"` | randomForest | `randomForest()` |
| `"boost"` | gbm | `gbm()` |
| `"xgboost"` | xgboost | `xgb.train()` |
| `"svm"` | e1071 | `svm()` |
| `"nn"` | nnet | `nnet()` |
| `"deep"` | keras | `keras_model_sequential()` |

### Unsupervised Learning

| Method | Underlying Package | Function Called |
| ------ | ------------------ | --------------- |
| `"pca"` | stats | `prcomp()` |
| `"mds"` | stats, MASS, smacof | `cmdscale()`, `isoMDS()`, etc. |
| `"kmeans"` | stats | `kmeans()` |
| `"pam"` | cluster | `pam()` |
| `"clara"` | cluster | `clara()` |
| `"hclust"` | stats | `hclust()` |
| `"dbscan"` | dbscan | `dbscan()` |

## Integration Workflows

Beyond wrapping individual packages, tidylearn provides orchestration functions
that combine multiple techniques:

### Dimensionality Reduction + Supervised Learning

```r
# Reduce dimensions before classification
reduced <- tl_reduce_dimensions(iris, response = "Species",
                                method = "pca", n_components = 3)
model <- tl_model(reduced$data, Species ~ ., method = "forest")
```

### Cluster-Based Feature Engineering

```r
# Add cluster membership as a feature
enriched <- tl_add_cluster_features(data, response = "target",
                                    method = "kmeans", k = 3)
model <- tl_model(enriched, target ~ ., method = "forest")
```

### Semi-Supervised Learning

```r
# Use clustering to propagate labels to unlabeled data
model <- tl_semisupervised(data, target ~ .,
                          labeled_indices = labeled_idx,
                          cluster_method = "kmeans")
```

### AutoML

```r
# Automatically try multiple approaches
result <- tl_auto_ml(data, target ~ .,
                    time_budget = 300)
result$leaderboard
```

## Compute Backends

Most methods run on the CPU and need no thought. For the two with an
upstream GPU path (`"xgboost"` and `"deep"`), `tl_model()` takes a
`compute` argument:

```r
# Check what this machine can actually do
tl_check_gpu()

# Route a fit to the local GPU (falls back to CPU with a warning if
# no CUDA-capable backend is detected)
model <- tl_model(data, y ~ ., method = "xgboost", compute = "gpu")

# Let tidylearn decide per call
model <- tl_model(data, y ~ ., method = "xgboost", compute = "auto")
```

`tl_compute_advisor()` estimates runtime, peak memory and cost across
local CPU, local GPU and cloud tiers before you commit to a long fit:

```r
tl_compute_advisor("xgboost", data, y ~ ., hyperparams = list(nrounds = 5000))
```

Estimates are order-of-magnitude, not quotes. The cloud tier is reported
for planning only - it is not yet executable.

## Unified Visualization

Consistent ggplot2-based plotting regardless of model type:

```r
# Generic plot method works for all model types
plot(forest_model)       # Automatic visualization based on model type
plot(linear_model)       # Diagnostic plots for regression
plot(pca_model)          # Variance explained for PCA
plot(kmeans_model)       # Cluster scatter plot
plot(hclust_model)       # Dendrogram

# The lower-level helpers take data frames rather than models
plot_clusters(cluster_data, cluster_col = "cluster")
plot_variance_explained(pca_model$fit$variance_explained)

# Interactive dashboard for detailed exploration
tl_dashboard(model, test_data)
```

## Formatted Tables

The `tl_table()` family produces polished `gt` tables for reporting:

```r
# Auto-selects the best table type
tl_table(model)

# Specific table types
tl_table_metrics(model, new_data = test_data)
tl_table_coefficients(model)
tl_table_confusion(model, new_data = test_data)
tl_table_importance(model)

# Compare models side-by-side
tl_table_comparison(model1, model2, model3,
                    new_data = test_data,
                    names = c("Linear", "Forest", "XGBoost"))
```

## Philosophy

tidylearn is built on these principles:

1. **Transparency**: The underlying packages do the real work. tidylearn makes
   them easier to use together without hiding what's happening.

2. **Consistency**: One interface, tidy output, unified visualization - across
   all methods.

3. **Accessibility**: Focus on your analysis, not on learning different package
   APIs.

4. **Interoperability**: Results work seamlessly with dplyr, ggplot2, and the
   broader tidyverse.

## Documentation

```r
# View package help
?tidylearn

# Explore main functions
?tl_read
?tl_model
?tl_evaluate
?tl_table
?tl_auto_ml
```

### Vignettes

- **Getting Started** — Overview of the tidylearn workflow
- **Data Ingestion** — Reading from files, databases, and cloud sources
- **Supervised Learning** — Classification and regression
- **Unsupervised Learning** — PCA, clustering, and MDS
- **Reporting** — Plots and formatted tables
- **Integration Workflows** — Combining multiple techniques
- **AutoML** — Automated machine learning

## Contributing

Contributions are welcome. Before opening a PR, please read
[CONTRIBUTING.md](https://github.com/ces0491/tidylearn/blob/main/CONTRIBUTING.md).

## License

MIT License - see [LICENSE](https://github.com/ces0491/tidylearn/blob/main/LICENSE) for details.

## Author

Cesaire Tobias (<cesaire@sheetsolved.com>)

## Acknowledgments

tidylearn is a wrapper that builds upon the excellent work of many R package
authors. The actual algorithms are implemented in:

- **stats** (base R): lm, glm, prcomp, kmeans, hclust, cmdscale
- **glmnet**: Ridge, LASSO, and elastic net regularization
- **randomForest**: Random forest implementation
- **xgboost**: Gradient boosting
- **gbm**: Gradient boosting machines
- **e1071**: Support vector machines
- **nnet**: Neural networks
- **rpart**: Decision trees
- **cluster**: PAM, CLARA clustering
- **dbscan**: Density-based clustering
- **MASS**: Sammon mapping, isoMDS
- **smacof**: SMACOF MDS algorithm
- **keras/tensorflow**: Deep learning (optional)

Thank you to all the package maintainers whose work makes tidylearn possible.

---

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)
library(ggplot2)

## ----eval=FALSE---------------------------------------------------------------
# # Run AutoML on iris dataset
# result <- tl_auto_ml(iris, Species ~ .,
#                      task = "classification",
#                      time_budget = 60)
# 
# # View best model
# print(result$best_model)

## ----eval=FALSE---------------------------------------------------------------
# # View all models tried
# names(result$models)

## ----eval=FALSE---------------------------------------------------------------
# # View leaderboard
# result$leaderboard

## ----eval=FALSE---------------------------------------------------------------
# # Run AutoML on regression problem
# result_reg <- tl_auto_ml(mtcars, mpg ~ .,
#                          task = "regression",
#                          time_budget = 60)
# 
# # Best model
# print(result_reg$best_model)

## ----eval=FALSE---------------------------------------------------------------
# # Quick sanity check -- 2 fast models, no CV, done in ~1s
# quick <- tl_auto_ml(iris, Species ~ .,
#                     time_budget = 10,
#                     use_reduction = FALSE,
#                     use_clustering = FALSE)
# quick$leaderboard
# #> baseline_tree, baseline_logistic
# 
# # Development iteration -- baselines + forest, some CV
# medium <- tl_auto_ml(iris, Species ~ .,
#                      time_budget = 60,
#                      cv_folds = 3)
# medium$leaderboard
# #> 5--7 models depending on data size
# 
# # Thorough search -- all phases, full CV
# thorough <- tl_auto_ml(iris, Species ~ .,
#                        time_budget = 300,
#                        cv_folds = 5)
# thorough$leaderboard
# #> 9--11 models with cross-validated scores

## ----eval=FALSE---------------------------------------------------------------
# # Factor/character response -> classification
# result_class <- tl_auto_ml(iris, Species ~ ., task = "auto")
# 
# # Numeric response -> regression
# result_reg <- tl_auto_ml(mtcars, mpg ~ ., task = "auto")

## ----eval=FALSE---------------------------------------------------------------
# # Disable dimensionality reduction
# no_reduction <- tl_auto_ml(iris, Species ~ .,
#                            use_reduction = FALSE,
#                            time_budget = 60)
# 
# # Disable cluster features
# no_clustering <- tl_auto_ml(iris, Species ~ .,
#                             use_clustering = FALSE,
#                             time_budget = 60)
# 
# # Baseline models only
# baseline_only <- tl_auto_ml(iris, Species ~ .,
#                             use_reduction = FALSE,
#                             use_clustering = FALSE,
#                             time_budget = 30)

## ----eval=FALSE---------------------------------------------------------------
# # Adjust cross-validation folds
# result_cv <- tl_auto_ml(iris, Species ~ .,
#                         cv_folds = 10,
#                         time_budget = 120)
# 
# # Fewer folds for faster evaluation
# result_fast <- tl_auto_ml(iris, Species ~ .,
#                           cv_folds = 3,
#                           time_budget = 60)

## ----eval=FALSE---------------------------------------------------------------
# result <- tl_auto_ml(iris, Species ~ ., time_budget = 60)
# 
# # Best performing model
# best_model <- result$best_model
# 
# # All models trained
# all_models <- result$models
# 
# # Specific model
# baseline_logistic <- result$models$baseline_logistic
# pca_forest <- result$models$pca_forest

## ----eval=FALSE---------------------------------------------------------------
# # View performance comparison
# leaderboard <- result$leaderboard
# 
# # Sort by score (higher is better for accuracy, lower for RMSE)
# leaderboard <- leaderboard %>%
#   arrange(desc(score))
# 
# print(leaderboard)

## ----eval=FALSE---------------------------------------------------------------
# # Use best model for predictions
# predictions <- predict(result$best_model, new_data = new_data)
# 
# # Or use a specific model
# predictions_pca <- predict(result$models$pca_forest, new_data = new_data)

## ----eval=FALSE---------------------------------------------------------------
# # Split data for evaluation
# split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)
# 
# # Run AutoML on training data
# automl_iris <- tl_auto_ml(split$train, Species ~ .,
#                           time_budget = 90,
#                           cv_folds = 5)
# 
# # Evaluate on test set
# test_preds <- predict(automl_iris$best_model, new_data = split$test)
# test_accuracy <- mean(test_preds$.pred == split$test$Species)
# 
# cat("AutoML Test Accuracy:", round(test_accuracy * 100, 1), "%\n")

## ----eval=FALSE---------------------------------------------------------------
# # Compare models
# for (model_name in names(automl_iris$models)) {
#   model <- automl_iris$models[[model_name]]
#   preds <- predict(model, new_data = split$test)
#   acc <- mean(preds$.pred == split$test$Species)
#   cat(model_name, ":", round(acc * 100, 1), "%\n")
# }

## ----eval=FALSE---------------------------------------------------------------
# # Split mtcars data
# split_mtcars <- tl_split(mtcars, prop = 0.7, seed = 42)
# 
# # Run AutoML
# automl_mpg <- tl_auto_ml(split_mtcars$train, mpg ~ .,
#                          task = "regression",
#                          time_budget = 90)
# 
# # Evaluate
# test_preds_mpg <- predict(automl_mpg$best_model, new_data = split_mtcars$test)
# rmse <- sqrt(mean((test_preds_mpg$.pred - split_mtcars$test$mpg)^2))
# 
# cat("AutoML Test RMSE:", round(rmse, 2), "\n")

## ----eval=FALSE---------------------------------------------------------------
# # Preprocess data first
# processed <- tl_prepare_data(
#   split$train,
#   Species ~ .,
#   scale_method = "standardize",
#   remove_correlated = TRUE
# )
# 
# # Run AutoML on preprocessed data
# automl_processed <- tl_auto_ml(processed$data, Species ~ .,
#                                time_budget = 60)
# 
# # Note: Need to apply same preprocessing to test data
# test_processed <- tl_prepare_data(
#   split$test,
#   Species ~ .,
#   scale_method = "standardize"
# )
# 
# test_preds_proc <- predict(
#   automl_processed$best_model,
#   new_data = test_processed$data
# )

## ----eval=FALSE---------------------------------------------------------------
# # Manual approach: choose one model
# manual_model <- tl_model(split$train, Species ~ ., method = "forest")
# manual_preds <- predict(manual_model, new_data = split$test)
# manual_acc <- mean(manual_preds$.pred == split$test$Species)
# 
# # AutoML approach
# automl_model <- tl_auto_ml(split$train, Species ~ ., time_budget = 60)
# automl_preds <- predict(automl_model$best_model, new_data = split$test)
# automl_acc <- mean(automl_preds$.pred == split$test$Species)
# 
# cat("Manual Selection:", round(manual_acc * 100, 1), "%\n")
# cat("AutoML:", round(automl_acc * 100, 1), "%\n")

## ----eval=FALSE---------------------------------------------------------------
# # First pass: quick exploration
# quick_automl <- tl_auto_ml(split$train, Species ~ .,
#                            time_budget = 30,
#                            use_reduction = TRUE,
#                            use_clustering = FALSE)
# 
# # Analyze what worked — best model name is in the leaderboard
# best_name <- quick_automl$leaderboard$model[1]
# best_method <- quick_automl$best_model$spec$method
# cat("Best model:", best_name, "(method:", best_method, ")\n")
# 
# # Second pass: if a PCA variant won, invest more in reduction
# if (grepl("^pca_", best_name)) {
#   refined_automl <- tl_auto_ml(split$train, Species ~ .,
#                                time_budget = 60,
#                                use_reduction = TRUE,
#                                use_clustering = TRUE)
# }

## ----eval=FALSE---------------------------------------------------------------
# # Get top 3 models
# top_models <- automl_iris$leaderboard %>%
#   arrange(desc(score)) %>%
#   head(3)
# 
# # Make predictions with each
# ensemble_preds <- list()
# for (i in seq_len(nrow(top_models))) {
#   model_name <- top_models$model[i]
#   model <- automl_iris$models[[model_name]]
#   ensemble_preds[[i]] <- predict(model, new_data = split$test)$.pred
# }
# 
# # Majority vote for classification
# final_pred <- apply(do.call(cbind, ensemble_preds), 1, function(x) {
#   names(which.max(table(x)))
# })
# 
# ensemble_acc <- mean(final_pred == split$test$Species)
# cat("Ensemble Accuracy:", round(ensemble_acc * 100, 1), "%\n")

## ----eval=FALSE---------------------------------------------------------------
# # AutoML automatically uses accuracy for classification
# result_class <- tl_auto_ml(iris, Species ~ .,
#                            metric = "accuracy",
#                            time_budget = 60)

## ----eval=FALSE---------------------------------------------------------------
# # AutoML automatically uses RMSE for regression
# result_reg <- tl_auto_ml(mtcars, mpg ~ .,
#                          metric = "rmse",
#                          time_budget = 60)

## ----eval=FALSE---------------------------------------------------------------
# # 1. Reduce CV folds (biggest impact)
# fast_result <- tl_auto_ml(data, formula,
#                           cv_folds = 2,
#                           time_budget = 30)
# 
# # 2. Disable slow phases
# baseline_result <- tl_auto_ml(data, formula,
#                               use_reduction = FALSE,
#                               use_clustering = FALSE,
#                               time_budget = 30)
# 
# # 3. Use a budget under 30s to skip forest/SVM/XGBoost entirely
# quick_result <- tl_auto_ml(data, formula, time_budget = 10)

## ----eval=FALSE---------------------------------------------------------------
# result <- tl_auto_ml(data, formula,
#                      metric = "accuracy",  # or "rmse" for regression
#                      time_budget = 60)

## ----eval=FALSE---------------------------------------------------------------
# # Increase time budget to unlock all phases
# thorough_result <- tl_auto_ml(data, formula, time_budget = 300)
# 
# # Ensure feature engineering is enabled
# full_result <- tl_auto_ml(data, formula,
#                           use_reduction = TRUE,
#                           use_clustering = TRUE,
#                           time_budget = 300)

## ----eval=FALSE---------------------------------------------------------------
# # Complete AutoML workflow
# workflow_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)
# 
# automl_result <- tl_auto_ml(
#   data = workflow_split$train,
#   formula = Species ~ .,
#   task = "auto",
#   use_reduction = TRUE,
#   use_clustering = TRUE,
#   time_budget = 120,
#   cv_folds = 5
# )
# 
# # Evaluate best model
# final_preds <- predict(automl_result$best_model, new_data = workflow_split$test)
# final_accuracy <- mean(final_preds$.pred == workflow_split$test$Species)
# 
# cat("Final AutoML Accuracy:", round(final_accuracy * 100, 1), "%\n")
# cat("Best approach:", automl_result$best_model$spec$method, "\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)
library(ggplot2)

## -----------------------------------------------------------------------------
# Perform PCA on iris data (excluding species)
model_pca <- tl_model(iris[, 1:4], method = "pca")
print(model_pca)

## -----------------------------------------------------------------------------
# Extract variance explained
variance_explained <- model_pca$fit$variance_explained
print(variance_explained)

## -----------------------------------------------------------------------------
# Cumulative variance explained
cumsum(variance_explained$prop_variance)

## -----------------------------------------------------------------------------
# Transform data to principal components
pca_scores <- predict(model_pca)
head(pca_scores)

## -----------------------------------------------------------------------------
# Visualize first two components
pca_plot_data <- pca_scores %>%
  mutate(Species = iris$Species)

ggplot(pca_plot_data, aes(x = PC1, y = PC2, color = Species)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(
    title = "PCA of Iris Dataset",
    x = paste0(
      "PC1 (", round(variance_explained$prop_variance[1] * 100, 1), "%)"
    ),
    y = paste0(
      "PC2 (", round(variance_explained$prop_variance[2] * 100, 1), "%)"
    )
  ) +
  theme_minimal()

## -----------------------------------------------------------------------------
# Examine loadings (variable contributions)
loadings <- model_pca$fit$loadings
print(loadings)

## -----------------------------------------------------------------------------
# Perform MDS
model_mds <- tl_model(iris[, 1:4], method = "mds", k = 2)
print(model_mds)

## -----------------------------------------------------------------------------
# Extract MDS coordinates
mds_points <- predict(model_mds)
head(mds_points)

## -----------------------------------------------------------------------------
# Visualize MDS
mds_plot_data <- mds_points %>%
  mutate(Species = iris$Species)

ggplot(mds_plot_data, aes(x = Dim1, y = Dim2, color = Species)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(title = "MDS of Iris Dataset") +
  theme_minimal()

## -----------------------------------------------------------------------------
# Perform k-means with k=3
model_kmeans <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
print(model_kmeans)

## -----------------------------------------------------------------------------
# Extract cluster assignments
clusters <- model_kmeans$fit$clusters
head(clusters)

## -----------------------------------------------------------------------------
# Compare clusters with actual species
table(Cluster = clusters$cluster, Species = iris$Species)

## -----------------------------------------------------------------------------
# Visualize clusters using PCA
cluster_viz <- pca_scores %>%
  mutate(
    Cluster = as.factor(clusters$cluster),
    Species = iris$Species
  )

ggplot(cluster_viz, aes(x = PC1, y = PC2, color = Cluster, shape = Species)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(title = "K-means Clusters vs True Species") +
  theme_minimal()

## -----------------------------------------------------------------------------
# Access cluster centers
centers <- model_kmeans$fit$centers
print(centers)

## ----eval=FALSE---------------------------------------------------------------
# # Perform PAM clustering
# model_pam <- tl_model(iris[, 1:4], method = "pam", k = 3)
# print(model_pam)
# 
# # Extract clusters
# clusters_pam <- model_pam$fit$clusters
# table(Cluster = clusters_pam$cluster, Species = iris$Species)

## -----------------------------------------------------------------------------
# Perform hierarchical clustering
model_hclust <- tl_model(iris[, 1:4], method = "hclust")
print(model_hclust)

## -----------------------------------------------------------------------------
# Plot dendrogram
plot(model_hclust$fit$model,
     labels = FALSE,
     main = "Hierarchical Clustering of Iris")

## -----------------------------------------------------------------------------
# Cut tree to get clusters
k <- 3
clusters_hc <- cutree(model_hclust$fit$model, k = k)
table(Cluster = clusters_hc, Species = iris$Species)

## -----------------------------------------------------------------------------
# Visualize hierarchical clusters
hc_viz <- pca_scores %>%
  mutate(
    Cluster = as.factor(clusters_hc),
    Species = iris$Species
  )

ggplot(hc_viz, aes(x = PC1, y = PC2, color = Cluster)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(title = "Hierarchical Clustering Results") +
  theme_minimal()

## ----eval=FALSE---------------------------------------------------------------
# # Perform DBSCAN
# model_dbscan <- tl_model(iris[, 1:4], method = "dbscan", eps = 0.5, minPts = 5)
# print(model_dbscan)
# 
# # Extract clusters (0 = noise/outliers)
# clusters_dbscan <- model_dbscan$fit$clusters
# table(clusters_dbscan$cluster)
# 
# # Compare with species
# table(Cluster = clusters_dbscan$cluster, Species = iris$Species)

## ----eval=FALSE---------------------------------------------------------------
# # Create larger dataset
# large_data <- iris[rep(seq_len(nrow(iris)), 10), 1:4]
# 
# # Perform CLARA
# model_clara <- tl_model(large_data, method = "clara", k = 3, samples = 5)
# print(model_clara)
# 
# # Extract clusters
# clusters_clara <- model_clara$fit$clusters

## -----------------------------------------------------------------------------
# Try different values of k
k_values <- 2:8
within_ss <- numeric(length(k_values))

for (i in seq_along(k_values)) {
  k <- k_values[i]
  model <- tl_model(iris[, 1:4], method = "kmeans", k = k)
  within_ss[i] <- model$fit$model$tot.withinss
}

# Plot elbow curve
elbow_data <- data.frame(k = k_values, within_ss = within_ss)

ggplot(elbow_data, aes(x = k, y = within_ss)) +
  geom_line(linewidth = 1) +
  geom_point(size = 3) +
  labs(
    title = "Elbow Method for Optimal k",
    x = "Number of Clusters (k)",
    y = "Total Within-Cluster Sum of Squares"
  ) +
  theme_minimal()

## -----------------------------------------------------------------------------
# Train clustering model
model_train <- tl_model(iris[1:100, 1:4], method = "kmeans", k = 3)

# Predict cluster assignments for new data
new_data <- iris[101:150, 1:4]
new_clusters <- predict(model_train, new_data = new_data)

head(new_clusters)

## -----------------------------------------------------------------------------
# Train PCA model
pca_train <- tl_model(iris[1:100, 1:4], method = "pca")

# Transform new data
new_pca <- predict(pca_train, new_data = new_data)
head(new_pca)

## -----------------------------------------------------------------------------
# Reduce dimensions with PCA
pca_model <- tl_model(iris[, 1:4], method = "pca")
pca_data <- predict(pca_model)

# Select first 2 components
pca_reduced <- pca_data %>% select(PC1, PC2)

# Cluster in reduced space
kmeans_pca <- tl_model(pca_reduced, method = "kmeans", k = 3)
clusters_pca <- kmeans_pca$fit$clusters

# Visualize
viz_combined <- pca_data %>%
  mutate(
    Cluster = as.factor(clusters_pca$cluster),
    Species = iris$Species
  )

ggplot(viz_combined, aes(x = PC1, y = PC2, color = Cluster, shape = Species)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(title = "Clustering in PCA Space") +
  theme_minimal()

## -----------------------------------------------------------------------------
# Simulate customer data
set.seed(42)
customers <- data.frame(
  age = rnorm(200, 40, 15),
  income = rnorm(200, 50000, 20000),
  spending_score = rnorm(200, 50, 25)
)

# Standardize features
customers_scaled <- scale(customers) %>% as.data.frame()

# Cluster customers
customer_segments <- tl_model(customers_scaled, method = "kmeans", k = 4)
customers$segment <- customer_segments$fit$clusters$cluster

# Visualize segments
ggplot(customers,
       aes(x = income, y = spending_score,
           color = as.factor(segment))) +
  geom_point(size = 3, alpha = 0.7) +
  labs(
    title = "Customer Segmentation",
    color = "Segment"
  ) +
  theme_minimal()

## -----------------------------------------------------------------------------
# Use PCA for feature extraction
pca_features <- tl_model(mtcars, method = "pca")

# Keep components explaining 90% of variance
var_exp <- pca_features$fit$variance_explained
cumulative_var <- cumsum(var_exp$prop_variance)
n_components <- which(cumulative_var >= 0.90)[1]

cat("Components needed for 90% variance:", n_components, "\n")
cat("Original features:", ncol(mtcars), "\n")
dim_reduction <- round(
  (1 - n_components / ncol(mtcars)) * 100, 1
)
cat("Dimension reduction:", dim_reduction, "%\n")

## -----------------------------------------------------------------------------
# Complete unsupervised workflow
workflow_data <- iris[, 1:4]

# 1. Reduce dimensions
pca_final <- tl_model(workflow_data, method = "pca")

# 2. Cluster in reduced space
pca_coords <- predict(pca_final) %>% select(PC1, PC2)
clusters_final <- tl_model(pca_coords, method = "kmeans", k = 3)

# 3. Visualize
final_viz <- pca_coords %>%
  mutate(
    Cluster = as.factor(clusters_final$fit$clusters$cluster),
    Species = iris$Species
  )

ggplot(final_viz, aes(x = PC1, y = PC2, color = Cluster)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(title = "Complete Unsupervised Workflow") +
  theme_minimal()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  message = FALSE,
  warning = FALSE
)

# Each backend demo is guarded on its reader package, which lives in
# Suggests -- the vignette has to build without them
has_sqlite <- requireNamespace("DBI", quietly = TRUE) &&
  requireNamespace("RSQLite", quietly = TRUE)

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)

## ----dispatcher-concept, eval = FALSE-----------------------------------------
# # Format is auto-detected from the file extension
# data <- tl_read("sales.csv")
# data <- tl_read("results.xlsx", sheet = "Q1")
# data <- tl_read("experiment.parquet")
# data <- tl_read("config.json")
# data <- tl_read("model_data.rds")
# 
# # Override format detection when the extension is ambiguous
# data <- tl_read("export.txt", format = "tsv")

## ----print-demo---------------------------------------------------------------
tmp <- tempfile(fileext = ".csv")
write.csv(mtcars, tmp, row.names = FALSE)

data <- tl_read(tmp, .quiet = TRUE)
data

## ----cleanup-1, include = FALSE-----------------------------------------------
unlink(tmp)

## ----csv-demo-----------------------------------------------------------------
# Create example files
tmp_csv <- tempfile(fileext = ".csv")
tmp_tsv <- tempfile(fileext = ".tsv")
write.csv(iris, tmp_csv, row.names = FALSE)
write.table(iris, tmp_tsv, sep = "\t", row.names = FALSE)

csv_data <- tl_read_csv(tmp_csv)
tsv_data <- tl_read_tsv(tmp_tsv)
nrow(csv_data)

## ----cleanup-csv, include = FALSE---------------------------------------------
unlink(c(tmp_csv, tmp_tsv))

## ----excel-demo, eval = requireNamespace("readxl", quietly = TRUE)------------
library(readxl)

path <- readxl_example("datasets.xlsx")
excel_data <- tl_read_excel(path, sheet = "mtcars")
head(excel_data, 3)

## ----parquet-demo, eval = requireNamespace("nanoparquet", quietly = TRUE)-----
library(nanoparquet)

tmp_pq <- tempfile(fileext = ".parquet")
write_parquet(iris, tmp_pq)

pq_data <- tl_read_parquet(tmp_pq)
nrow(pq_data)

## ----cleanup-pq, include = FALSE, eval = requireNamespace("nanoparquet", quietly = TRUE)----
unlink(tmp_pq)

## ----json-demo, eval = requireNamespace("jsonlite", quietly = TRUE)-----------
library(jsonlite)

tmp_json <- tempfile(fileext = ".json")
write_json(mtcars[1:5, ], tmp_json)

json_data <- tl_read_json(tmp_json)
json_data

## ----cleanup-json, include = FALSE, eval = requireNamespace("jsonlite", quietly = TRUE)----
unlink(tmp_json)

## ----rds-demo-----------------------------------------------------------------
tmp_rds <- tempfile(fileext = ".rds")
saveRDS(iris, tmp_rds)

rds_data <- tl_read_rds(tmp_rds)
nrow(rds_data)

## ----rdata-demo---------------------------------------------------------------
tmp_rdata <- tempfile(fileext = ".rdata")
my_data <- mtcars
save(my_data, file = tmp_rdata)

# Name is auto-detected when there is a single data frame
rdata_data <- tl_read_rdata(tmp_rdata)
nrow(rdata_data)

## ----cleanup-rds, include = FALSE---------------------------------------------
unlink(c(tmp_rds, tmp_rdata))

## ----sqlite-demo, eval = has_sqlite-------------------------------------------
library(DBI)
library(RSQLite)

# Create an example database
tmp_db <- tempfile(fileext = ".sqlite")
conn <- dbConnect(SQLite(), tmp_db)
dbWriteTable(conn, "iris_tbl", iris)
dbDisconnect(conn)

# Read with tl_read_sqlite
db_data <- tl_read_sqlite(
  tmp_db,
  "SELECT * FROM iris_tbl WHERE Species = 'setosa'"
)
nrow(db_data)

## ----cleanup-sqlite, include = FALSE, eval = has_sqlite-----------------------
unlink(tmp_db)

## ----db-demo, eval = has_sqlite-----------------------------------------------
conn <- dbConnect(SQLite(), ":memory:")
dbWriteTable(conn, "mtcars_tbl", mtcars)

sql <- "SELECT mpg, wt, hp FROM mtcars_tbl WHERE mpg > 20"
db_result <- tl_read_db(conn, sql)
db_result

dbDisconnect(conn)

## ----remote-db, eval = FALSE--------------------------------------------------
# # PostgreSQL
# pg_data <- tl_read_postgres(
#   dsn = "localhost",
#   query = "SELECT * FROM sales WHERE year = 2025",
#   dbname = "analytics",
#   user = "myuser",
#   password = "mypass"
# )
# 
# # MySQL / MariaDB # nolint: commented_code_linter.
# mysql_data <- tl_read_mysql(
#   dsn = "mysql://user:pass@host:3306/mydb",
#   query = "SELECT * FROM customers LIMIT 1000"
# )
# 
# # BigQuery
# bq_data <- tl_read_bigquery(
#   project = "my-gcp-project",
#   query = "SELECT * FROM `dataset.table` LIMIT 1000"
# )

## ----s3, eval = FALSE---------------------------------------------------------
# data <- tl_read_s3("s3://my-bucket/data/sales_2025.csv")
# data <- tl_read_s3("s3://my-bucket/data/results.parquet", region = "eu-west-1")

## ----github, eval = FALSE-----------------------------------------------------
# # Read a CSV from a public GitHub repository
# data <- tl_read_github("tidyverse/dplyr",
#   path = "data-raw/starwars.csv", ref = "main"
# )

## ----kaggle, eval = FALSE-----------------------------------------------------
# data <- tl_read_kaggle("zillow/zecon", file = "Zip_time_series.csv")
# data <- tl_read_kaggle("titanic", file = "train.csv", type = "competition")

## ----multi-path---------------------------------------------------------------
dir <- tempdir()
write.csv(iris[1:50, ], file.path(dir, "batch1.csv"), row.names = FALSE)
write.csv(iris[51:100, ], file.path(dir, "batch2.csv"), row.names = FALSE)

paths <- file.path(dir, c("batch1.csv", "batch2.csv"))
combined <- tl_read(paths, .quiet = TRUE)
table(combined$source_file)

## ----cleanup-multi, include = FALSE-------------------------------------------
unlink(file.path(dir, c("batch1.csv", "batch2.csv")))

## ----dir-demo-----------------------------------------------------------------
dir <- tempfile(pattern = "tl_vignette_")
dir.create(dir)
write.csv(iris[1:50, ], file.path(dir, "jan.csv"), row.names = FALSE)
write.csv(iris[51:100, ], file.path(dir, "feb.csv"), row.names = FALSE)
write.csv(iris[101:150, ], file.path(dir, "mar.csv"), row.names = FALSE)

# Read all CSVs from the directory
all_data <- tl_read_dir(dir, format = "csv", .quiet = TRUE)
nrow(all_data)
table(all_data$source_file)

## ----dir-pattern--------------------------------------------------------------
# Filter with a regex pattern
subset <- tl_read_dir(dir, pattern = "^(jan|feb)", .quiet = TRUE)
nrow(subset)

## ----cleanup-dir, include = FALSE---------------------------------------------
unlink(dir, recursive = TRUE)

## ----dir-dispatch, eval = FALSE-----------------------------------------------
# data <- tl_read("data/monthly_exports/")

## ----zip-demo-----------------------------------------------------------------
# Create an example zip
dir <- tempfile(pattern = "tl_zip_src_")
dir.create(dir)
write.csv(iris, file.path(dir, "iris.csv"), row.names = FALSE)
zip_path <- tempfile(fileext = ".zip")
# -j stores the file without its directory path, so no setwd() is needed
utils::zip(zip_path, file.path(dir, "iris.csv"), flags = "-j9X")

zip_data <- tl_read_zip(zip_path, .quiet = TRUE)
nrow(zip_data)
attr(zip_data, "tl_format")

## ----cleanup-zip, include = FALSE---------------------------------------------
unlink(c(dir, zip_path), recursive = TRUE)

## ----zip-dispatch, eval = FALSE-----------------------------------------------
# data <- tl_read("download.zip")
# data <- tl_read("download.zip", file = "train.csv")

## ----class-demo---------------------------------------------------------------
tmp <- tempfile(fileext = ".csv")
write.csv(mtcars, tmp, row.names = FALSE)
data <- tl_read(tmp, .quiet = TRUE)

# Check metadata
attr(data, "tl_format")

# Works with dplyr
data %>%
  filter(mpg > 20) %>%
  select(mpg, wt, hp) %>%
  head(3)

## ----cleanup-class, include = FALSE-------------------------------------------
unlink(tmp)

## ----pipeline-----------------------------------------------------------------
# 1. Ingest
tmp <- tempfile(fileext = ".csv")
write.csv(iris, tmp, row.names = FALSE)
data <- tl_read(tmp, .quiet = TRUE)

# CSV files lose factor information, so convert character columns as needed
data <- data %>% mutate(Species = as.factor(Species))

# 2. Split
split <- tl_split(data, prop = 0.7, stratify = "Species", seed = 42)

# 3. Model
model <- tl_model(split$train, Species ~ ., method = "forest")

# 4. Evaluate
eval_result <- tl_evaluate(model, new_data = split$test)
eval_result

## ----cleanup-pipeline, include = FALSE----------------------------------------
unlink(tmp)

# tidylearn: Unified Workflow Examples
#
# Seven end-to-end workflows combining supervised and unsupervised
# learning. Run the whole file with:
#
#   source(system.file("examples", "unified_workflow.R",
#                      package = "tidylearn"))
#
# Exercised by tests/testthat/test-examples.R, so it stays runnable.

library(tidylearn)

# ====================================================
# Example 1: Complete Integrative Workflow
# ====================================================

cat("Example 1: Complete Integrative Workflow\n")
cat("=========================================\n\n")

# Load data
data(iris)

# Step 1: Exploratory analysis
cat("Step 1: Exploratory Data Analysis\n")
eda <- tl_explore(iris, response = "Species")
print(eda)

# Step 2: Dimensionality reduction
cat("\nStep 2: Dimensionality Reduction\n")
reduced <- tl_reduce_dimensions(
  iris, response = "Species",
  method = "pca", n_components = 3
)
n_reduced <- sum(grepl("^PC", names(reduced$data)))
cat("Reduced from", ncol(iris) - 1, "to",
    n_reduced, "features\n")

# Step 3: Add cluster features
cat("\nStep 3: Feature Engineering with Clustering\n")
data_enriched <- tl_add_cluster_features(
  reduced$data,
  response = "Species",
  method = "kmeans", k = 3
)
cat("Added cluster feature\n")

# Step 4: Train supervised model
cat("\nStep 4: Supervised Learning\n")
model <- tl_model(data_enriched, Species ~ .,
                  method = "forest")
print(model)


# ====================================================
# Example 2: Semi-Supervised Learning
# ====================================================

cat("\n\nExample 2: Semi-Supervised Learning\n")
cat("====================================\n\n")

# Simulate scenario with limited labels
set.seed(42)
labeled_indices <- sample(nrow(iris), size = 15)

cat(
  "Training with only", length(labeled_indices),
  "labeled observations out of", nrow(iris), "\n"
)

# iris has three species, so the supervised method has to be
# multiclass-capable -- logistic regression is binary only.
semi_model <- tl_semisupervised(
  iris, Species ~ .,
  labeled_indices = labeled_indices,
  cluster_method = "kmeans",
  supervised_method = "forest"
)

cat("Propagated labels to", nrow(iris) - length(labeled_indices),
    "unlabelled observations\n")


# ====================================================
# Example 3: Auto ML
# ====================================================

cat("\n\nExample 3: Automated Machine Learning\n")
cat("======================================\n\n")

auto_result <- tl_auto_ml(
  iris, Species ~ .,
  use_reduction = TRUE,
  use_clustering = TRUE,
  time_budget = 120
)

cat("\nLeaderboard:\n")
print(auto_result$leaderboard)

cat("\nBest model:",
    auto_result$leaderboard$model[1], "\n")


# ====================================================
# Example 4: Anomaly-Aware Modeling
# ====================================================

cat("\n\nExample 4: Anomaly-Aware Modeling\n")
cat("==================================\n\n")

# Add some synthetic outliers
iris_with_outliers <- iris
iris_with_outliers[1:5, 1:4] <-
  iris_with_outliers[1:5, 1:4] * 3

cat("Added 5 outliers to the dataset\n")

anomaly_model <- tl_anomaly_aware(
  iris_with_outliers,
  Species ~ .,
  response = "Species",
  anomaly_method = "dbscan",
  action = "flag",
  supervised_method = "forest",
  eps = 0.5, minPts = 5
)

cat("Detected",
    anomaly_model$anomaly_info$n_anomalies,
    "anomalies\n")


# ====================================================
# Example 5: Stratified Models
# ====================================================

cat("\n\nExample 5: Stratified Models\n")
cat("============================\n\n")

data(mtcars)

stratified <- tl_stratified_models(
  mtcars, mpg ~ .,
  cluster_method = "kmeans", k = 3,
  supervised_method = "linear"
)

cat(
  "Created",
  length(stratified$supervised_models),
  "cluster-specific models\n"
)

# Make predictions
predictions <- predict(stratified, mtcars)
cat("Predictions completed with cluster assignments\n")


# ====================================================
# Example 6: Transfer Learning
# ====================================================

cat("\n\nExample 6: Transfer Learning\n")
cat("============================\n\n")

transfer_model <- tl_transfer_learning(
  iris, Species ~ .,
  pretrain_method = "pca",
  supervised_method = "forest",
  n_components = 3
)

cat("Transfer learning model built on",
    transfer_model$spec$method, "over 3 principal components\n")


# ====================================================
# Example 7: Preprocessing a messy frame
# ====================================================

cat("\n\nExample 7: Preprocessing a messy frame\n")
cat("======================================\n\n")

# Simulate data with issues
set.seed(7)
messy_data <- iris
messy_data[sample(nrow(messy_data), 10),
           sample(4, 1)] <- NA
messy_data$redundant_col <-
  messy_data$Sepal.Length +
  rnorm(nrow(messy_data), 0, 0.01)

cat("Original data:", ncol(messy_data) - 1,
    "features with missing values\n")

processed <- tl_prepare_data(
  messy_data, Species ~ .,
  impute_method = "mean",
  scale_method = "standardize",
  remove_correlated = TRUE,
  correlation_cutoff = 0.95
)

cat(
  "Processed data:", ncol(processed$data) - 1,
  "features, no missing values\n"
)


cat("\n\n========================================\n")
cat("All examples completed successfully!\n")
cat("========================================\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----eval = FALSE-------------------------------------------------------------
# # From CRAN
# install.packages("tidylearn")
# 
# # Development version
# # devtools::install_github("ces0491/tidylearn") # nolint

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)

## -----------------------------------------------------------------------------
# versicolor and virginica overlap, so this is a real classification
# problem -- setosa is linearly separable from the other two, which makes
# logistic regression fail to converge
iris_binary <- iris %>%
  filter(Species %in% c("versicolor", "virginica")) %>%
  mutate(Species = droplevels(Species))

model_logistic <- tl_model(iris_binary, Species ~ ., method = "logistic")
print(model_logistic)

## -----------------------------------------------------------------------------
# Predicted class labels
predictions <- predict(model_logistic, type = "class")
head(predictions)

## -----------------------------------------------------------------------------
# Class probabilities
head(predict(model_logistic, type = "prob"))

## -----------------------------------------------------------------------------
tl_evaluate(model_logistic, metrics = c("accuracy", "f1"))

## -----------------------------------------------------------------------------
# Regression with linear model
model_linear <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
print(model_linear)

## -----------------------------------------------------------------------------
# Predictions
predictions_reg <- predict(model_linear)
head(predictions_reg)

## -----------------------------------------------------------------------------
# Principal Component Analysis
model_pca <- tl_model(iris[, 1:4], method = "pca")
print(model_pca)

## -----------------------------------------------------------------------------
# Transform data
transformed <- predict(model_pca)
head(transformed)

## -----------------------------------------------------------------------------
# K-means clustering
model_kmeans <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
print(model_kmeans)

## -----------------------------------------------------------------------------
# Get cluster assignments
clusters <- model_kmeans$fit$clusters
head(clusters)

## -----------------------------------------------------------------------------
# Compare with actual species
table(clusters$cluster, iris$Species)

## -----------------------------------------------------------------------------
# Prepare data with multiple preprocessing steps
processed <- tl_prepare_data(
  iris,
  Species ~ .,
  impute_method = "mean",
  scale_method = "standardize",
  encode_categorical = FALSE
)

## -----------------------------------------------------------------------------
# Check preprocessing steps applied
names(processed$preprocessing_steps)

## -----------------------------------------------------------------------------
# Use processed data for modeling
model_processed <- tl_model(processed$data, Species ~ ., method = "forest")

## -----------------------------------------------------------------------------
# Simple random split
split <- tl_split(iris, prop = 0.7, seed = 123)

# Train model (three species, so a multiclass-capable method)
model_train <- tl_model(split$train, Species ~ ., method = "forest")

# Test predictions
predictions_test <- predict(model_train, new_data = split$test)
head(predictions_test)

## -----------------------------------------------------------------------------
# Stratified split (maintains class proportions)
split_strat <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

# Check proportions are maintained
prop.table(table(split_strat$train$Species))
prop.table(table(split_strat$test$Species))
prop.table(table(iris$Species))

## -----------------------------------------------------------------------------
# Example: Access the raw randomForest object
model_forest <- tl_model(iris, Species ~ ., method = "forest")
class(model_forest$fit)  # This is the randomForest object

# Use package-specific functions if needed
# randomForest::varImpPlot(model_forest$fit) # nolint

## -----------------------------------------------------------------------------
model_pca <- tl_model(iris, ~ ., method = "pca")
names(model_pca$fit)
class(model_pca$fit$model)  # This is the prcomp object

## -----------------------------------------------------------------------------
# Quick example combining everything
data_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Random forests are scale-invariant, so no scaling is needed here. When a
# method does need scaled inputs, the same transformation has to be applied
# to the test set -- see the Supervised Learning vignette.
model_final <- tl_model(data_split$train, Species ~ ., method = "forest")
test_preds <- predict(model_final, new_data = data_split$test)

accuracy <- mean(test_preds$.pred == data_split$test$Species)
cat("Test accuracy:", round(accuracy * 100, 1), "%\n")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)

## -----------------------------------------------------------------------------
# Create binary classification dataset. setosa is linearly separable from
# the other two species, and logistic regression has no finite maximum
# likelihood estimate on separable data -- glm() fits, warns that the
# algorithm did not converge, and returns coefficients that diverged.
# versicolor and virginica overlap, so this is a real classification
# problem. Even here, a 70% split of 100 rows is separable at some seeds;
# this one is not.
iris_binary <- iris %>%
  filter(Species %in% c("versicolor", "virginica")) %>%
  mutate(Species = droplevels(Species))

# Split data
split <- tl_split(iris_binary, prop = 0.7, stratify = "Species", seed = 42)

## -----------------------------------------------------------------------------
# Train logistic regression
model_logistic <- tl_model(split$train, Species ~ ., method = "logistic")
print(model_logistic)

## -----------------------------------------------------------------------------
# Predicted labels
preds_logistic <- predict(model_logistic, new_data = split$test, type = "class")
head(preds_logistic)

## -----------------------------------------------------------------------------
# Class probabilities
head(predict(model_logistic, new_data = split$test, type = "prob"))

## -----------------------------------------------------------------------------
tl_evaluate(
  model_logistic,
  new_data = split$test,
  metrics = c("accuracy", "precision", "recall", "f1", "auc")
)

## -----------------------------------------------------------------------------
# Train decision tree
model_tree <- tl_model(split$train, Species ~ ., method = "tree")
print(model_tree)

# Predictions
preds_tree <- predict(model_tree, new_data = split$test)

## -----------------------------------------------------------------------------
# Split full iris dataset
split_multi <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

## -----------------------------------------------------------------------------
# Train random forest
model_forest <- tl_model(split_multi$train, Species ~ ., method = "forest")
print(model_forest)

## -----------------------------------------------------------------------------
# Predictions
preds_forest <- predict(model_forest, new_data = split_multi$test)
head(preds_forest)

## -----------------------------------------------------------------------------
# Accuracy on test set
mean(preds_forest$.pred == split_multi$test$Species)

## -----------------------------------------------------------------------------
# Train SVM
model_svm <- tl_model(split_multi$train, Species ~ ., method = "svm")
print(model_svm)

# Predictions
preds_svm <- predict(model_svm, new_data = split_multi$test)

## -----------------------------------------------------------------------------
# Split mtcars data
split_reg <- tl_split(mtcars, prop = 0.7, seed = 123)

# Train linear model
model_lm <- tl_model(split_reg$train, mpg ~ wt + hp + disp, method = "linear")
print(model_lm)

## -----------------------------------------------------------------------------
# Predictions
preds_lm <- predict(model_lm, new_data = split_reg$test)
head(preds_lm)

## -----------------------------------------------------------------------------
# Calculate RMSE
rmse <- sqrt(mean((preds_lm$.pred - split_reg$test$mpg)^2))
cat("RMSE:", round(rmse, 2), "\n")

## -----------------------------------------------------------------------------
# Polynomial regression for non-linear relationships
model_poly <- tl_model(
  split_reg$train, mpg ~ wt,
  method = "polynomial", degree = 2
)
print(model_poly)

## -----------------------------------------------------------------------------
# Predictions
preds_poly <- predict(model_poly, new_data = split_reg$test)

# RMSE
rmse_poly <- sqrt(mean((preds_poly$.pred - split_reg$test$mpg)^2))
cat("Polynomial RMSE:", round(rmse_poly, 2), "\n")

## -----------------------------------------------------------------------------
# Train random forest for regression
model_rf_reg <- tl_model(split_reg$train, mpg ~ ., method = "forest")
print(model_rf_reg)

## -----------------------------------------------------------------------------
# Predictions
preds_rf <- predict(model_rf_reg, new_data = split_reg$test)

# RMSE
rmse_rf <- sqrt(mean((preds_rf$.pred - split_reg$test$mpg)^2))
cat("Random Forest RMSE:", round(rmse_rf, 2), "\n")

## -----------------------------------------------------------------------------
# Ridge regression (L2 regularization)
model_ridge <- tl_model(split_reg$train, mpg ~ ., method = "ridge")
print(model_ridge)

# Predictions
preds_ridge <- predict(model_ridge, new_data = split_reg$test)

## -----------------------------------------------------------------------------
# LASSO (L1 regularization) - performs feature selection
model_lasso <- tl_model(split_reg$train, mpg ~ ., method = "lasso")
print(model_lasso)

# Predictions
preds_lasso <- predict(model_lasso, new_data = split_reg$test)

## -----------------------------------------------------------------------------
# Elastic Net - combines L1 and L2 regularization
model_enet <- tl_model(
  split_reg$train, mpg ~ .,
  method = "elastic_net", alpha = 0.5
)
print(model_enet)

# Predictions
preds_enet <- predict(model_enet, new_data = split_reg$test)

## -----------------------------------------------------------------------------
# Compare multiple models
models <- list(
  linear = tl_model(split_reg$train, mpg ~ ., method = "linear"),
  tree = tl_model(split_reg$train, mpg ~ ., method = "tree"),
  forest = tl_model(split_reg$train, mpg ~ ., method = "forest")
)

## -----------------------------------------------------------------------------
# Calculate RMSE for each model
results <- data.frame(
  Model = character(),
  RMSE = numeric(),
  stringsAsFactors = FALSE
)

for (model_name in names(models)) {
  preds <- predict(models[[model_name]], new_data = split_reg$test)
  rmse <- sqrt(mean((preds$.pred - split_reg$test$mpg)^2))

  results <- rbind(results, data.frame(
    Model = model_name,
    RMSE = rmse
  ))
}

results <- results %>% arrange(RMSE)
print(results)

## -----------------------------------------------------------------------------
# Preprocess data
processed <- tl_prepare_data(
  split_reg$train,
  mpg ~ .,
  scale_method = "standardize",
  remove_correlated = TRUE,
  correlation_cutoff = 0.9
)

## -----------------------------------------------------------------------------
# Train on preprocessed data
model_processed <- tl_model(processed$data, mpg ~ ., method = "linear")
print(model_processed)

## -----------------------------------------------------------------------------
# Interaction terms
model_interact <- tl_model(split_reg$train, mpg ~ wt * hp, method = "linear")

# Polynomial terms using I()
model_poly_manual <- tl_model(
  split_reg$train, mpg ~ wt + I(wt^2), method = "linear"
)

# Subset of predictors
model_subset <- tl_model(
  split_reg$train, mpg ~ wt + hp + disp, method = "linear"
)

## -----------------------------------------------------------------------------
# Create dataset with categorical variables
mtcars_cat <- mtcars %>%
  mutate(
    cyl = as.factor(cyl),
    gear = as.factor(gear),
    am = as.factor(am)
  )

split_cat <- tl_split(mtcars_cat, prop = 0.7, seed = 123)

# Model with categorical predictors
model_cat <- tl_model(split_cat$train, mpg ~ ., method = "forest")
print(model_cat)

## -----------------------------------------------------------------------------
# Create data with missing values. Seeded so the vignette renders the
# same output on every build.
set.seed(123)
mtcars_missing <- mtcars
mtcars_missing[sample(seq_len(nrow(mtcars_missing)), 5), "hp"] <- NA
mtcars_missing[sample(seq_len(nrow(mtcars_missing)), 3), "wt"] <- NA

# Preprocess to handle missing values
processed_missing <- tl_prepare_data(
  mtcars_missing,
  mpg ~ .,
  impute_method = "mean",
  scale_method = "standardize"
)

# Train model
model_imputed <- tl_model(processed_missing$data, mpg ~ ., method = "linear")

## -----------------------------------------------------------------------------
# Complete workflow example
final_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

final_prep <- tl_prepare_data(
  final_split$train, Species ~ .,
  scale_method = "standardize"
)
final_model <- tl_model(final_prep$data, Species ~ ., method = "forest")

## -----------------------------------------------------------------------------
scaling <- final_prep$preprocessing_steps$scaling$scaling_params

final_test <- final_split$test
for (col in names(scaling)) {
  final_test[[col]] <-
    (final_test[[col]] - scaling[[col]]$mean) / scaling[[col]]$sd
}

final_preds <- predict(final_model, new_data = final_test)
accuracy <- mean(final_preds$.pred == final_split$test$Species)
cat("Test Accuracy:", round(accuracy * 100, 1), "%\n")

## ----include = FALSE----------------------------------------------------------
# gt is in Suggests and is the subject of this vignette, so skip the
# whole thing rather than fail the build when it is absent
has_gt <- requireNamespace("gt", quietly = TRUE)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  message = FALSE,
  warning = FALSE,
  eval = has_gt
)

## ----echo = FALSE, results = "asis", eval = TRUE------------------------------
if (!has_gt) {
  cat(
    "> **Note:** the gt package is not installed, so the examples below",
    "are shown without output.\n"
  )
}

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)
library(ggplot2)
library(gt)

## ----plot-regression----------------------------------------------------------
model_reg <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

# Actual vs predicted — one call
plot(model_reg, type = "actual_predicted")

## ----plot-classification------------------------------------------------------
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)
model_clf <- tl_model(split$train, Species ~ ., method = "forest")

plot(model_clf, type = "confusion")

## ----plot-pca-----------------------------------------------------------------
pca <- tidy_pca(USArrests, scale = TRUE)

tidy_pca_screeplot(pca)
tidy_pca_biplot(pca, label_obs = TRUE)

## ----plot-lasso---------------------------------------------------------------
model_lasso <- tl_model(mtcars, mpg ~ ., method = "lasso")

tl_plot_regularization_path(model_lasso)
tl_plot_regularization_cv(model_lasso)

## ----table-auto, eval = FALSE-------------------------------------------------
# tl_table(model)                       # auto-selects the best table type
# tl_table(model, type = "coefficients") # specific type

## ----table-metrics------------------------------------------------------------
tl_table_metrics(model_reg)

## ----table-coef---------------------------------------------------------------
tl_table_coefficients(model_reg)

## ----table-coef-lasso---------------------------------------------------------
tl_table_coefficients(model_lasso)

## ----table-confusion----------------------------------------------------------
tl_table_confusion(model_clf, new_data = split$test)

## ----table-importance---------------------------------------------------------
tl_table_importance(model_clf)

## ----table-variance-----------------------------------------------------------
pca_model <- tl_model(USArrests, method = "pca")
tl_table_variance(pca_model)

## ----table-loadings-----------------------------------------------------------
tl_table_loadings(pca_model)

## ----table-clusters-----------------------------------------------------------
km <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
tl_table_clusters(km)

## ----table-comparison---------------------------------------------------------
m1 <- tl_model(split$train, Species ~ ., method = "svm")
m2 <- tl_model(split$train, Species ~ ., method = "forest")
m3 <- tl_model(split$train, Species ~ ., method = "tree")

tl_table_comparison(
  m1, m2, m3,
  new_data = split$test,
  names = c("SVM", "Random Forest", "Decision Tree")
)

## ----plotly, eval = FALSE-----------------------------------------------------
# library(plotly)
# 
# ggplotly(plot(model_reg, type = "actual_predicted"))
# ggplotly(tidy_pca_biplot(pca, label_obs = TRUE))
# ggplotly(tl_plot_regularization_path(model_lasso))

## ----workflow-----------------------------------------------------------------
# Fit
model <- tl_model(split$train, Species ~ ., method = "forest")

# Evaluate
tl_table_metrics(model, new_data = split$test)

# Visualise
plot(model, type = "confusion")

# Drill into feature importance
tl_table_importance(model, top_n = 4)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----setup--------------------------------------------------------------------
library(tidylearn)
library(dplyr)
library(ggplot2)

## -----------------------------------------------------------------------------
# Reduce dimensions before classification
reduced <- tl_reduce_dimensions(iris,
                                response = "Species",
                                method = "pca",
                                n_components = 3)

# Inspect reduced data
head(reduced$data)

## -----------------------------------------------------------------------------
# Train classifier on the reduced features. iris has three species, so this
# needs a multiclass-capable method -- logistic regression is binary only.
model_reduced <- tl_model(reduced$data, Species ~ ., method = "forest")
print(model_reduced)

## -----------------------------------------------------------------------------
# Training-set accuracy. Use tl_evaluate() rather than comparing $.pred by
# hand: what $.pred holds depends on the method and prediction type.
tl_evaluate(model_reduced)

## -----------------------------------------------------------------------------
# Split data for fair comparison
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

# Model with original features
model_original <- tl_model(split$train, Species ~ ., method = "forest")
eval_original <- tl_evaluate(model_original, new_data = split$test)

# Model with PCA features
reduced_train <- tl_reduce_dimensions(split$train,
                                      response = "Species",
                                      method = "pca",
                                      n_components = 3)

model_pca <- tl_model(reduced_train$data, Species ~ ., method = "forest")

# The test set must be projected through the PCA fitted on the training
# data -- refitting PCA on the test set would leak information
test_transformed <- predict(
  reduced_train$reduction_model,
  new_data = split$test %>% select(-Species)
)
test_transformed$Species <- split$test$Species

eval_pca <- tl_evaluate(model_pca, new_data = test_transformed)

# Compare results
acc <- function(x) round(x$value[x$metric == "accuracy"] * 100, 1)
n_original <- ncol(split$train) - 1
n_reduced <- sum(grepl("^PC", names(reduced_train$data)))
cat("Original features:", n_original, "->", acc(eval_original), "%\n")
cat("PCA features:", n_reduced, "->", acc(eval_pca), "%\n")
cat("Feature reduction:",
    round((1 - n_reduced / n_original) * 100, 1), "%\n")

## -----------------------------------------------------------------------------
# Add cluster features
data_clustered <- tl_add_cluster_features(iris,
                                          response = "Species",
                                          method = "kmeans",
                                          k = 3)

# Check new features
names(data_clustered)

## -----------------------------------------------------------------------------
# Train model with cluster features
model_cluster <- tl_model(data_clustered, Species ~ ., method = "forest")
print(model_cluster)

## -----------------------------------------------------------------------------
# Compare models with and without cluster features
split_comp <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Without cluster features
model_no_cluster <- tl_model(split_comp$train, Species ~ ., method = "forest")
preds_no_cluster <- predict(model_no_cluster, new_data = split_comp$test)
acc_no_cluster <- mean(preds_no_cluster$.pred == split_comp$test$Species)

# With cluster features
train_clustered <- tl_add_cluster_features(split_comp$train,
                                           response = "Species",
                                           method = "kmeans",
                                           k = 3)
model_with_cluster <- tl_model(train_clustered, Species ~ ., method = "forest")

# Need to get cluster model for test data
cluster_model <- attr(train_clustered, "cluster_model")
test_clusters <- predict(cluster_model, new_data = split_comp$test[, -5])
test_clustered <- split_comp$test
test_clustered$cluster_kmeans <- as.factor(test_clusters$cluster)

preds_with_cluster <- predict(model_with_cluster, new_data = test_clustered)
acc_with_cluster <- mean(preds_with_cluster$.pred == split_comp$test$Species)

cat("Without cluster features:", round(acc_no_cluster * 100, 1), "%\n")
cat("With cluster features:", round(acc_with_cluster * 100, 1), "%\n")

## -----------------------------------------------------------------------------
# Use only 10% of labels
set.seed(123)
labeled_indices <- sample(nrow(iris), size = 15)  # 15 of 150 labelled

# supervised_method defaults to "tree". Named here because a forest is
# the better fit for propagated labels: the propagation step introduces
# noise, and averaging over trees absorbs more of it than one tree does.
model_semi <- tl_semisupervised(iris, Species ~ .,
                                labeled_indices = labeled_indices,
                                cluster_method = "kmeans",
                                supervised_method = "forest")

print(model_semi)

## -----------------------------------------------------------------------------
# Check how labels were propagated
label_mapping <- model_semi$semisupervised_info$label_mapping
print(label_mapping)

## -----------------------------------------------------------------------------
# Evaluate against the true labels
preds_semi <- predict(model_semi, new_data = iris, type = "class")
accuracy_semi <- mean(preds_semi$.pred == iris$Species)

cat("Accuracy with only", length(labeled_indices), "labels:",
    round(accuracy_semi * 100, 1), "%\n")
labeled_pct <- round(
  length(labeled_indices) / nrow(iris) * 100, 1
)
cat("Proportion of data labeled:", labeled_pct, "%\n")

## -----------------------------------------------------------------------------
# Fully supervised with same amount of data
labeled_data <- iris[labeled_indices, ]
model_full <- tl_model(labeled_data, Species ~ ., method = "forest")
preds_full <- predict(model_full, new_data = iris, type = "class")
accuracy_full <- mean(preds_full$.pred == iris$Species)

cat("Fully supervised (15 samples):", round(accuracy_full * 100, 1), "%\n")
cat("Semi-supervised (15 labels + propagation):",
    round(accuracy_semi * 100, 1), "%\n")

## ----eval=FALSE---------------------------------------------------------------
# # Flag anomalies as a feature
# model_anomaly_flag <- tl_anomaly_aware(iris, Species ~ .,
#                                        response = "Species",
#                                        anomaly_method = "dbscan",
#                                        action = "flag",
#                                        supervised_method = "forest")
# 
# # Check anomaly info
# cat("Anomalies detected:", model_anomaly_flag$anomaly_info$n_anomalies, "\n")

## ----eval=FALSE---------------------------------------------------------------
# # Remove anomalies before training
# model_anomaly_remove <- tl_anomaly_aware(iris, Species ~ .,
#                                          response = "Species",
#                                          anomaly_method = "dbscan",
#                                          action = "remove",
#                                          supervised_method = "forest")
# 
# cat("Anomalies removed:", model_anomaly_remove$anomalies_removed, "\n")

## -----------------------------------------------------------------------------
# Train separate models for different clusters
stratified_models <- tl_stratified_models(mtcars, mpg ~ .,
                                          cluster_method = "kmeans",
                                          k = 3,
                                          supervised_method = "linear")

# Check structure
names(stratified_models)
length(stratified_models$supervised_models)

## -----------------------------------------------------------------------------
# Predictions using stratified models
preds_stratified <- predict(stratified_models)
head(preds_stratified)

## -----------------------------------------------------------------------------
# Calculate RMSE
rmse_stratified <- sqrt(mean((preds_stratified$.pred - mtcars$mpg)^2))
cat("Stratified Model RMSE:", round(rmse_stratified, 2), "\n")

# Compare with single model
model_single <- tl_model(mtcars, mpg ~ ., method = "linear")
preds_single <- predict(model_single)
rmse_single <- sqrt(mean((preds_single$.pred - mtcars$mpg)^2))
cat("Single Model RMSE:", round(rmse_single, 2), "\n")

## -----------------------------------------------------------------------------
# Step 1: Split data
workflow_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Step 2: Reduce dimensions
workflow_reduced <- tl_reduce_dimensions(workflow_split$train,
                                         response = "Species",
                                         method = "pca",
                                         n_components = 3)

# Step 3: Add cluster features to reduced data
workflow_clustered <- tl_add_cluster_features(workflow_reduced$data,
                                              response = "Species",
                                              method = "kmeans",
                                              k = 3)

# Step 4: Train final model
workflow_model <- tl_model(workflow_clustered, Species ~ ., method = "forest")

print(workflow_model)

## -----------------------------------------------------------------------------
# Transform test data through same pipeline
# 1. Apply PCA transformation
test_pca <- predict(workflow_reduced$reduction_model,
                    new_data = workflow_split$test[, -5])
test_pca$Species <- workflow_split$test$Species

# 2. Get cluster assignments. The cluster model was fitted on the PC
# columns; predict() matches new_data to those columns by name and
# errors on a mismatch rather than assigning against the wrong ones.
cluster_model_wf <- attr(workflow_clustered, "cluster_model")
test_clusters_wf <- predict(cluster_model_wf, new_data = test_pca)
test_pca$cluster_kmeans <- as.factor(test_clusters_wf$cluster)

# 3. Predict
workflow_preds <- predict(workflow_model, new_data = test_pca)
workflow_accuracy <- mean(workflow_preds$.pred == workflow_split$test$Species)

cat("Complete Workflow Accuracy:", round(workflow_accuracy * 100, 1), "%\n")

## -----------------------------------------------------------------------------
# Simulate credit data
set.seed(42)
n <- 500
credit_data <- data.frame(
  age = rnorm(n, 40, 12),
  income = rnorm(n, 50000, 20000),
  debt_ratio = runif(n, 0, 0.5),
  credit_score = rnorm(n, 700, 100),
  years_employed = rpois(n, 5)
)

# Create target variable (default risk)
credit_data$default <- factor(
  ifelse(
    credit_data$debt_ratio > 0.4 & credit_data$credit_score < 650,
    "Yes", "No"
  )
)

# Split data
credit_split <- tl_split(
  credit_data, prop = 0.7, stratify = "default", seed = 123
)

## -----------------------------------------------------------------------------
# Strategy 1: Add customer segments as features
credit_clustered <- tl_add_cluster_features(credit_split$train,
                                            response = "default",
                                            method = "kmeans",
                                            k = 4)

model_credit <- tl_model(credit_clustered, default ~ ., method = "forest")

# Transform test data
cluster_model_credit <- attr(credit_clustered, "cluster_model")
test_clusters_credit <- predict(cluster_model_credit,
                                new_data = credit_split$test[, -6])
test_credit <- credit_split$test
test_credit$cluster_kmeans <- as.factor(test_clusters_credit$cluster)

preds_credit <- predict(model_credit, new_data = test_credit)
accuracy_credit <- mean(preds_credit$.pred == credit_split$test$default)

cat("Credit Risk Model Accuracy:", round(accuracy_credit * 100, 1), "%\n")

#' @title Data Reading Functions for tidylearn
#' @name tidylearn-read
#' @description Functions for reading data from diverse sources into tidy
#'   \code{tidylearn_data} objects. The main dispatcher
#'   \code{tl_read()} auto-detects the format from the file
#'   extension and routes to the appropriate reader.
#'   All readers return a \code{tidylearn_data} object,
#'   which is a tibble subclass carrying metadata about
#'   the data source.
#'
#' @details
#' Supported file formats:
#' \itemize{
#'   \item \strong{CSV}: \code{.csv} files via \pkg{readr}
#'     (with base R fallback)
#'   \item \strong{TSV}: \code{.tsv} files via \pkg{readr}
#'     (with base R fallback)
#'   \item \strong{Excel}: \code{.xls}, \code{.xlsx},
#'     \code{.xlsm} files via \pkg{readxl}
#'   \item \strong{Parquet}: \code{.parquet} files via \pkg{nanoparquet}
#'   \item \strong{JSON}: \code{.json} files via \pkg{jsonlite}
#'   \item \strong{RDS}: \code{.rds} files via base \code{readRDS()}
#'   \item \strong{RData}: \code{.rdata}, \code{.rda}
#'     files via base \code{load()}
#' }
#'
#' Supported databases (via \pkg{DBI}):
#' \itemize{
#'   \item \strong{SQLite}: \code{.sqlite}, \code{.db} files via \pkg{RSQLite}
#'   \item \strong{PostgreSQL}: via \pkg{RPostgres}
#'   \item \strong{MySQL/MariaDB}: via \pkg{RMariaDB}
#'   \item \strong{BigQuery}: via \pkg{bigrquery}
#' }
#'
#' Supported cloud/API sources:
#' \itemize{
#'   \item \strong{S3}: \code{s3://} URIs via \pkg{paws.storage}
#'   \item \strong{GitHub}: raw file download from repositories
#'   \item \strong{Kaggle}: dataset download via Kaggle CLI
#' }
#'
#' Multi-file reading:
#' \itemize{
#'   \item \strong{Multiple paths}: pass a character vector to \code{tl_read()}
#'   \item \strong{Directories}: \code{tl_read_dir()} scans for data files with
#'     optional pattern/format filtering and recursive scanning
#'   \item \strong{Zip archives}: \code{tl_read_zip()} extracts and reads from
#'     \code{.zip} files
#' }
#' When combining multiple files, a \code{source_file} column is added to
#' identify the origin of each row.
NULL

# ---- tidylearn_data class ----

#' Create a tidylearn_data object
#'
#' Constructor for the \code{tidylearn_data} class, a
#' tibble subclass that carries metadata about the data
#' source.
#'
#' @param data A data frame or tibble.
#' @param source Character string describing the data source (e.g., file path).
#' @param format Character string indicating the format (e.g., "csv", "excel").
#' @param timestamp POSIXct timestamp of when the data
#'   was read. Defaults to current time.
#'
#' @return A \code{tidylearn_data} object (tibble
#'   subclass with source metadata).
#' @keywords internal
#' @noRd
new_tidylearn_data <- function(data, source, format, timestamp = Sys.time()) {
  data <- tibble::as_tibble(data)
  structure(
    data,
    class = c("tidylearn_data", class(data)),
    tl_source = source,
    tl_format = format,
    tl_timestamp = timestamp
  )
}

#' Print a tidylearn_data object
#'
#' @param x A \code{tidylearn_data} object.
#' @param ... Additional arguments passed to the tibble print method.
#' @return The input object \code{x}, returned invisibly.
#' @examples
#' \donttest{
#' f <- tempfile(fileext = ".csv")
#' write.csv(iris, f, row.names = FALSE)
#' d <- tl_read(f)
#' print(d)
#' }
#' @export
print.tidylearn_data <- function(x, ...) {
  cat("-- tidylearn data ---------\n")
  cat("Source:", attr(x, "tl_source"), "\n")
  cat("Format:", attr(x, "tl_format"), "\n")
  cat("Read at:", format(attr(x, "tl_timestamp")), "\n\n")
  NextMethod()
}

# ---- Format detection ----

#' Detect data format from source string
#'
#' Infers the data format from a file extension, URL pattern, or connection
#' string. Used internally by \code{tl_read()} when
#' \code{format} is not specified.
#'
#' @param source Character string: a file path, URL, or connection string.
#'
#' @return A character string indicating the detected format.
#' @keywords internal
#' @noRd
tl_detect_format <- function(source) {
  # Check URL/protocol patterns first (these take priority over extensions)
  if (grepl("^https?://", source)) {
    github_pat <- "github\\.com|raw\\.githubusercontent\\.com"
    if (grepl(github_pat, source)) return("github")
    if (grepl("kaggle\\.com", source)) return("kaggle")
  }
  if (grepl("^s3://", source)) return("s3")
  if (grepl("^postgres(ql)?://", source)) return("postgres")
  if (grepl("^mysql://", source)) return("mysql")


  # Fall back to file extension matching
  # Note: .txt defaults to CSV; use format = "tsv" to override
  ext <- tolower(tools::file_ext(source))

  ext_map <- c(
    csv = "csv", tsv = "tsv", txt = "csv",
    xls = "excel", xlsx = "excel", xlsm = "excel",
    rds = "rds", rdata = "rdata", rda = "rdata",
    parquet = "parquet",
    json = "json", ndjson = "json",
    sqlite = "sqlite", db = "sqlite"
  )

  if (ext %in% names(ext_map)) return(ext_map[[ext]])

  stop("Cannot detect format from: '", source,
       "'. Please specify the 'format' argument.",
       call. = FALSE)
}

# ---- Main dispatcher ----

#' Read data from diverse sources
#'
#' Auto-detects the data format from the file extension or source pattern and
#' dispatches to the appropriate reader. All readers
#' return a \code{tidylearn_data} object, which is a
#' tibble subclass carrying metadata about the data
#' source.
#'
#' When \code{source} is a character vector of multiple paths, each file is read
#' and row-bound into a single result with a \code{source_file} column. When
#' \code{source} is a directory path, it is equivalent to calling
#' \code{tl_read_dir()}. When \code{source} is a \code{.zip} file, it is
#' equivalent to calling \code{tl_read_zip()}.
#'
#' @param source A file path, URL, connection string, directory path, or a
#'   character vector of multiple file paths.
#' @param ... Additional arguments passed to the format-specific reader.
#' @param format Optional explicit format override.
#'   One of \code{"csv"}, \code{"tsv"},
#'   \code{"excel"}, \code{"parquet"}, \code{"json"},
#'   \code{"rds"}, \code{"rdata"},
#'   \code{"sqlite"}, \code{"postgres"}, \code{"mysql"}, \code{"bigquery"},
#'   \code{"s3"}, \code{"github"}, \code{"kaggle"}. When \code{NULL} (default),
#'   the format is auto-detected from the file extension
#'   or source pattern. Note: \code{.txt} files default
#'   to CSV; use \code{format = "tsv"} to override.
#' @param .quiet Logical. If \code{TRUE}, suppresses
#'   informational messages. Default is \code{FALSE}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # Read a single CSV file
#' # data <- tl_read("path/to/data.csv")
#'
#' # Read multiple files and row-bind
#' # data <- tl_read(c("jan.csv", "feb.csv", "mar.csv"))
#'
#' # Read all CSVs from a directory
#' # data <- tl_read("data/")
#'
#' # Read from a zip archive
#' # data <- tl_read("data.zip")
#'
#' # Explicit format override
#' # data <- tl_read("path/to/data.txt", format = "tsv")
#' }
#'
#' @export
tl_read <- function(source, ..., format = NULL, .quiet = FALSE) {
  if (!is.character(source) || length(source) == 0) {
    stop("'source' must be a character string or vector of paths",
         call. = FALSE)
  }

  # Multi-path: read each and row-bind
  if (length(source) > 1) {
    return(tl_read_multi(source, ..., format = format, .quiet = .quiet))
  }

  # Directory: delegate to tl_read_dir
  if (dir.exists(source)) {
    return(tl_read_dir(source, ..., format = format, .quiet = .quiet))
  }

  # Zip file: delegate to tl_read_zip
  if (tolower(tools::file_ext(source)) == "zip") {
    return(tl_read_zip(source, ..., format = format, .quiet = .quiet))
  }

  if (is.null(format)) {
    format <- tl_detect_format(source)
  }

  supported_formats <- c(
    "csv", "tsv", "excel", "parquet", "json", "rds", "rdata",
    "sqlite", "postgres", "mysql", "bigquery",
    "s3", "github", "kaggle"
  )

  if (!format %in% supported_formats) {
    stop("Unsupported format: '", format, "'.",
         "\nSupported formats: ", paste(supported_formats, collapse = ", "),
         call. = FALSE)
  }

  if (!.quiet) message("Reading ", format, " data from: ", source)

  result <- switch(format,
    "csv"      = tl_read_csv(source, ...),
    "tsv"      = tl_read_tsv(source, ...),
    "excel"    = tl_read_excel(source, ...),
    "parquet"  = tl_read_parquet(source, ...),
    "json"     = tl_read_json(source, ...),
    "rds"      = tl_read_rds(source),
    "rdata"    = tl_read_rdata(source, ...),
    "sqlite"   = tl_read_sqlite(source, ...),
    "postgres" = tl_read_postgres(source, ...),
    "mysql"    = tl_read_mysql(source, ...),
    "bigquery" = tl_read_bigquery(source, ...),
    "s3"       = tl_read_s3(source, ...),
    "github"   = tl_read_github(source, ...),
    "kaggle"   = tl_read_kaggle(source, ...)
  )

  if (!.quiet) {
    message("Returned: ", nrow(result), " rows x ", ncol(result), " columns")
  }

  result
}

# ---- Multi-file reading ----

#' Read multiple files and row-bind
#'
#' Internal helper that reads multiple file paths and combines them into a
#' single \code{tidylearn_data} object with a \code{source_file} column
#' identifying the origin of each row.
#'
#' @param paths Character vector of file paths.
#' @param ... Additional arguments passed to each reader.
#' @param format Optional format override applied to all files.
#' @param .quiet Suppress messages.
#'
#' @return A \code{tidylearn_data} object with a \code{source_file} column.
#' @keywords internal
#' @noRd
tl_read_multi <- function(paths, ..., format = NULL, .quiet = FALSE) {
  if (!.quiet) {
    message("Reading ", length(paths), " files...")
  }

  results <- lapply(paths, function(p) {
    df <- tl_read(p, ..., format = format, .quiet = TRUE)
    col <- "source_file"
    if (col %in% names(df)) {
      col <- "tl_source_file"
      warning(
        "Column 'source_file' already exists in data. ",
        "Using 'tl_source_file' instead.",
        call. = FALSE
      )
    }
    df[[col]] <- basename(p)
    df
  })

  combined <- dplyr::bind_rows(results)

  if (!.quiet) {
    message("Combined: ", nrow(combined), " rows x ", ncol(combined),
            " columns from ", length(paths), " files")
  }

  new_tidylearn_data(
    combined,
    source = paste0(length(paths), " files"),
    format = if (!is.null(format)) format else "multi"
  )
}

#' Read all matching files from a directory
#'
#' Scans a directory for files matching a pattern or format, reads each one,
#' and row-binds them into a single \code{tidylearn_data} object with a
#' \code{source_file} column identifying the origin of each row.
#'
#' @param path Path to a directory.
#' @param pattern Optional regex pattern to filter file names (e.g.,
#'   \code{"sales_.*\\\\.csv$"}). If \code{NULL}, files are filtered by
#'   \code{format} instead.
#' @param format File format to read. If \code{NULL} and \code{pattern} is
#'   \code{NULL}, all recognized data files are read. If specified, only files
#'   with matching extensions are read.
#' @param recursive Logical. Should subdirectories be scanned? Default is
#'   \code{FALSE}.
#' @param .quiet Suppress messages. Default is \code{FALSE}.
#' @param ... Additional arguments passed to the format-specific reader.
#'
#' @return A \code{tidylearn_data} object with an additional
#'   \code{source_file} column identifying the origin of each row.
#'
#' @examples
#' \donttest{
#' # Read all CSVs from a directory
#' # data <- tl_read_dir("data/", format = "csv")
#'
#' # Read with pattern matching
#' # data <- tl_read_dir("data/", pattern = "^sales_.*\\.csv$")
#'
#' # Read all recognized data files recursively
#' # data <- tl_read_dir("data/", recursive = TRUE)
#' }
#'
#' @export
tl_read_dir <- function(path, pattern = NULL, format = NULL,
                        recursive = FALSE, .quiet = FALSE, ...) {
  if (!dir.exists(path)) {
    stop("Directory not found: '", path, "'", call. = FALSE)
  }

  # Build file list
  if (!is.null(pattern)) {
    files <- list.files(path, pattern = pattern, full.names = TRUE,
                        recursive = recursive)
  } else if (!is.null(format)) {
    ext_patterns <- list(
      csv = "\\.csv$", tsv = "\\.tsv$",
      excel = "\\.(xlsx?|xlsm)$",
      parquet = "\\.parquet$", json = "\\.json$",
      rds = "\\.rds$", rdata = "\\.(rdata|rda)$"
    )
    pat <- ext_patterns[[format]]
    if (is.null(pat)) {
      stop("Cannot scan directory for format '", format, "'. ",
           "Use 'pattern' argument instead.",
           call. = FALSE)
    }
    files <- list.files(path, pattern = pat, full.names = TRUE,
                        recursive = recursive, ignore.case = TRUE)
  } else {
    # All recognized data file extensions
    pat <- "\\.(csv|tsv|xlsx?|xlsm|parquet|json|rds|rdata|rda)$"
    files <- list.files(path, pattern = pat, full.names = TRUE,
                        recursive = recursive, ignore.case = TRUE)
  }

  if (length(files) == 0) {
    stop("No data files found in '", path, "'.",
         if (!is.null(pattern)) paste0("\nPattern: ", pattern),
         if (!is.null(format)) paste0("\nFormat: ", format),
         call. = FALSE)
  }

  if (!.quiet) {
    message("Found ", length(files), " file(s) in ", path)
  }

  tl_read_multi(files, ..., format = format, .quiet = .quiet)
}

#' Read data from a zip archive
#'
#' Extracts a zip archive to a temporary directory and reads the contents.
#' If the archive contains a single data file, it is read directly. If
#' multiple data files are found, they are row-bound with a \code{source_file}
#' column. Use the \code{file} argument to select a specific file from
#' the archive.
#'
#' @param path Path to a zip file.
#' @param file Optional name of a specific file within the archive to read.
#'   Supports partial matching.
#' @param format Optional format override for the file(s) inside the archive.
#' @param .quiet Suppress messages. Default is \code{FALSE}.
#' @param ... Additional arguments passed to the format-specific reader.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}. The archive is extracted to a temporary directory
#'   that is cleaned up automatically. If multiple data files are found,
#'   a \code{source_file} column identifies the origin of each row.
#'
#' @examples
#' \donttest{
#' # Read from a zip archive
#' # data <- tl_read_zip("data.zip")
#'
#' # Read a specific file from the archive
#' # data <- tl_read_zip("data.zip", file = "train.csv")
#' }
#'
#' @export
tl_read_zip <- function(path, file = NULL, format = NULL,
                        .quiet = FALSE, ...) {
  tl_validate_file_path(path)

  # Extract to temp directory
  dest <- tempfile(pattern = "tl_zip_")
  dir.create(dest)
  on.exit(unlink(dest, recursive = TRUE), add = TRUE)

  utils::unzip(path, exdir = dest)

  if (!is.null(file)) {
    # Find the specific file (supports partial matching)
    all_files <- list.files(dest, recursive = TRUE, full.names = TRUE)
    matches <- all_files[grepl(file, basename(all_files), fixed = TRUE)]

    if (length(matches) == 0) {
      available <- basename(all_files)
      stop("File '", file, "' not found in archive.",
           "\nAvailable files: ", paste(available, collapse = ", "),
           call. = FALSE)
    }

    target <- matches[1]
    if (length(matches) > 1 && !.quiet) {
      message("Multiple matches for '", file, "'. Reading: ",
              basename(target))
    }

    result <- tl_read(target, ..., format = format, .quiet = .quiet)
    attr(result, "tl_source") <- paste0(path, "//", basename(target))
    attr(result, "tl_format") <- paste0("zip+", attr(result, "tl_format"))
    return(result)
  }

  # No specific file — read all data files
  data_pattern <- "\\.(csv|tsv|xlsx?|xlsm|parquet|json|rds|rdata|rda)$"
  data_files <- list.files(dest, pattern = data_pattern,
                           full.names = TRUE, recursive = TRUE,
                           ignore.case = TRUE)

  if (length(data_files) == 0) {
    all_files <- list.files(dest, recursive = TRUE)
    stop("No recognized data files found in archive.",
         "\nFiles in archive: ", paste(all_files, collapse = ", "),
         call. = FALSE)
  }

  if (length(data_files) == 1) {
    result <- tl_read(data_files[1], ..., format = format, .quiet = .quiet)
    attr(result, "tl_source") <- paste0(path, "//", basename(data_files[1]))
    attr(result, "tl_format") <- paste0("zip+", attr(result, "tl_format"))
    return(result)
  }

  # Multiple files — row-bind
  if (!.quiet) {
    message("Found ", length(data_files), " data file(s) in ", basename(path))
  }

  result <- tl_read_multi(data_files, ..., format = format, .quiet = .quiet)
  attr(result, "tl_source") <- path
  attr(result, "tl_format") <- "zip+multi"
  result
}

# ---- File format readers ----

#' Read a CSV file
#'
#' Reads a CSV file into a \code{tidylearn_data} object. Uses \pkg{readr} when
#' available for faster parsing, with a base R fallback.
#'
#' @param path Path to a CSV file.
#' @param ... Additional arguments passed to \code{readr::read_csv()} or
#'   \code{utils::read.csv()}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_csv("path/to/data.csv")
#' }
#'
#' @export
tl_read_csv <- function(path, ...) {
  tl_validate_file_path(path)

  if (requireNamespace("readr", quietly = TRUE)) {
    data <- readr::read_csv(path, show_col_types = FALSE, ...)
  } else {
    message("Install 'readr' for faster CSV reading. Using base R.")
    data <- utils::read.csv(path, stringsAsFactors = FALSE, ...)
    data <- tibble::as_tibble(data)
  }

  new_tidylearn_data(data, source = path, format = "csv")
}

#' Read a TSV file
#'
#' Reads a tab-separated file into a
#' \code{tidylearn_data} object. Uses \pkg{readr} when
#' available for faster parsing, with a base R fallback.
#'
#' @param path Path to a TSV file.
#' @param ... Additional arguments passed to \code{readr::read_tsv()} or
#'   \code{utils::read.delim()}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_tsv("path/to/data.tsv")
#' }
#'
#' @export
tl_read_tsv <- function(path, ...) {
  tl_validate_file_path(path)

  if (requireNamespace("readr", quietly = TRUE)) {
    data <- readr::read_tsv(path, show_col_types = FALSE, ...)
  } else {
    message("Install 'readr' for faster TSV reading. Using base R.")
    data <- utils::read.delim(path, stringsAsFactors = FALSE, ...)
    data <- tibble::as_tibble(data)
  }

  new_tidylearn_data(data, source = path, format = "tsv")
}

#' Read an Excel file
#'
#' Reads an Excel file (\code{.xls}, \code{.xlsx}, or \code{.xlsm}) into a
#' \code{tidylearn_data} object. Requires the \pkg{readxl} package.
#'
#' @param path Path to an Excel file.
#' @param sheet Sheet to read. Either a string (the name of a sheet) or an
#'   integer (the position of the sheet). Defaults to the first sheet.
#' @param ... Additional arguments passed to \code{readxl::read_excel()}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_excel("path/to/data.xlsx")
#' # data <- tl_read_excel("path/to/data.xlsx", sheet = "Sheet2")
#' }
#'
#' @export
tl_read_excel <- function(path, sheet = 1, ...) {
  tl_validate_file_path(path)
  tl_check_packages("readxl")

  data <- readxl::read_excel(path, sheet = sheet, ...)

  new_tidylearn_data(data, source = path, format = "excel")
}

#' Read an RDS file
#'
#' Reads an RDS file into a \code{tidylearn_data} object. Uses base R
#' \code{readRDS()} — no additional packages required.
#'
#' @param path Path to an RDS file.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_rds("path/to/data.rds")
#' }
#'
#' @export
tl_read_rds <- function(path) {
  tl_validate_file_path(path)

  data <- readRDS(path)

  if (!is.data.frame(data)) {
    stop("The RDS file does not contain a data frame. ",
         "tl_read_rds() expects tabular data.",
         call. = FALSE)
  }

  new_tidylearn_data(data, source = path, format = "rds")
}

#' Read an RData file
#'
#' Reads an RData (\code{.rdata} or \code{.rda}) file
#' into a \code{tidylearn_data} object. Since RData files
#' can contain multiple objects, use the \code{name}
#' argument to specify which object to extract.
#' If \code{name} is \code{NULL} and
#' the file contains exactly one data frame, it is returned automatically.
#'
#' @param path Path to an RData file.
#' @param name Optional name of the object to extract from the RData file. If
#'   \code{NULL} (default), the function returns the first data frame found, or
#'   errors if there are multiple data frames.
#' @param ... Currently unused.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_rdata("path/to/data.rdata")
#' # data <- tl_read_rdata("path/to/data.rdata", name = "my_data")
#' }
#'
#' @export
tl_read_rdata <- function(path, name = NULL, ...) {
  tl_validate_file_path(path)

  env <- new.env(parent = emptyenv())
  load(path, envir = env)

  objects <- ls(envir = env)

  if (length(objects) == 0) {
    stop("The RData file is empty.", call. = FALSE)
  }

  if (!is.null(name)) {
    if (!name %in% objects) {
      stop("Object '", name, "' not found in the RData file.",
           "\nAvailable objects: ", paste(objects, collapse = ", "),
           call. = FALSE)
    }
    data <- get(name, envir = env)
  } else {
    # Find data frames
    df_objects <- objects[vapply(objects, function(nm) {
      is.data.frame(get(nm, envir = env))
    }, logical(1))]

    if (length(df_objects) == 0) {
      stop("No data frames found in the RData file.",
           "\nAvailable objects: ", paste(objects, collapse = ", "),
           call. = FALSE)
    }

    if (length(df_objects) > 1) {
      stop("Multiple data frames found in the RData file: ",
           paste(df_objects, collapse = ", "),
           "\nPlease specify which one to load with the 'name' argument.",
           call. = FALSE)
    }

    data <- get(df_objects[[1]], envir = env)
  }

  if (!is.data.frame(data)) {
    stop("The selected object is not a data frame. ",
         "tl_read_rdata() expects tabular data.",
         call. = FALSE)
  }

  new_tidylearn_data(data, source = path, format = "rdata")
}

#' Read a Parquet file
#'
#' Reads a Parquet file into a \code{tidylearn_data} object. Requires the
#' \pkg{nanoparquet} package.
#'
#' @param path Path to a Parquet file.
#' @param ... Additional arguments passed to \code{nanoparquet::read_parquet()}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_parquet("path/to/data.parquet")
#' }
#'
#' @export
tl_read_parquet <- function(path, ...) {
  tl_validate_file_path(path)
  tl_check_packages("nanoparquet")

  data <- nanoparquet::read_parquet(path, ...)

  new_tidylearn_data(data, source = path, format = "parquet")
}

#' Read a JSON file
#'
#' Reads a JSON file into a \code{tidylearn_data} object. Expects the JSON to
#' represent tabular data (array of objects or similar). Requires the
#' \pkg{jsonlite} package.
#'
#' @param path Path to a JSON file.
#' @param flatten Logical. Automatically flatten nested data frames? Default is
#'   \code{TRUE}.
#' @param ... Additional arguments passed to \code{jsonlite::fromJSON()}.
#'
#' @return A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
#'   with attributes \code{tl_source}, \code{tl_format}, and
#'   \code{tl_timestamp}.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_json("path/to/data.json")
#' }
#'
#' @export
tl_read_json <- function(path, flatten = TRUE, ...) {
  tl_validate_file_path(path)
  tl_check_packages("jsonlite")

  data <- jsonlite::fromJSON(path, flatten = flatten, ...)

  if (!is.data.frame(data)) {
    stop("The JSON file does not contain tabular data. ",
         "tl_read_json() expects an array of objects.",
         call. = FALSE)
  }

  new_tidylearn_data(data, source = path, format = "json")
}

# ---- Format detection ----

test_that("tl_detect_format identifies file extensions correctly", {
  expect_equal(tl_detect_format("data.csv"), "csv")
  expect_equal(tl_detect_format("data.tsv"), "tsv")
  expect_equal(tl_detect_format("data.txt"), "csv")
  expect_equal(tl_detect_format("data.xls"), "excel")
  expect_equal(tl_detect_format("data.xlsx"), "excel")
  expect_equal(tl_detect_format("data.xlsm"), "excel")
  expect_equal(tl_detect_format("data.rds"), "rds")
  expect_equal(tl_detect_format("data.rdata"), "rdata")
  expect_equal(tl_detect_format("data.rda"), "rdata")
  expect_equal(tl_detect_format("data.parquet"), "parquet")
  expect_equal(tl_detect_format("data.json"), "json")
  expect_equal(tl_detect_format("data.sqlite"), "sqlite")
  expect_equal(tl_detect_format("data.db"), "sqlite")
})

test_that("tl_detect_format is case-insensitive", {
  expect_equal(tl_detect_format("data.CSV"), "csv")
  expect_equal(tl_detect_format("data.XLSX"), "excel")
  expect_equal(tl_detect_format("data.RDS"), "rds")
})

test_that("tl_detect_format identifies URL patterns", {
  expect_equal(
    tl_detect_format("https://github.com/user/repo/blob/main/data.csv"),
    "github"
  )
  url <- "https://raw.githubusercontent.com/user/repo/main/data.csv"
  expect_equal(tl_detect_format(url), "github")
  expect_equal(
    tl_detect_format("https://www.kaggle.com/datasets/user/dataset"),
    "kaggle"
  )
  expect_equal(tl_detect_format("s3://my-bucket/data.csv"), "s3")
  expect_equal(tl_detect_format("postgres://localhost/mydb"), "postgres")
  expect_equal(tl_detect_format("postgresql://localhost/mydb"), "postgres")
  expect_equal(tl_detect_format("mysql://localhost/mydb"), "mysql")
})

test_that("tl_detect_format errors on unrecognizable sources", {
  expect_error(tl_detect_format("unknown_thing"), "Cannot detect format")
  expect_error(tl_detect_format("no_extension"), "Cannot detect format")
})

# ---- tidylearn_data class ----

test_that("new_tidylearn_data creates correct class", {
  data <- new_tidylearn_data(iris, source = "test.csv", format = "csv")
  expect_s3_class(data, "tidylearn_data")
  expect_s3_class(data, "tbl_df")
  expect_equal(attr(data, "tl_source"), "test.csv")
  expect_equal(attr(data, "tl_format"), "csv")
  expect_true(inherits(attr(data, "tl_timestamp"), "POSIXct"))
})

test_that("tidylearn_data works with dplyr verbs", {
  data <- new_tidylearn_data(iris, source = "test.csv", format = "csv")
  filtered <- dplyr::filter(data, Species == "setosa")
  expect_equal(nrow(filtered), 50)
  selected <- dplyr::select(data, Sepal.Length, Species)
  expect_equal(ncol(selected), 2)
})

test_that("print.tidylearn_data shows metadata", {
  data <- new_tidylearn_data(iris, source = "test.csv", format = "csv")
  output <- capture.output(print(data))
  expect_true(any(grepl("tidylearn data", output)))
  expect_true(any(grepl("Source:", output)))
  expect_true(any(grepl("Format:", output)))
  expect_true(any(grepl("Read at:", output)))
})

# ---- tl_read dispatcher ----

test_that("tl_read errors on non-character source", {
  expect_error(tl_read(42), "must be a character string")
  expect_error(tl_read(NULL), "must be a character string")
})

test_that("tl_read errors on missing file", {
  expect_error(tl_read("nonexistent.csv"), "File not found")
})

test_that("tl_read errors on unsupported format", {
  expect_error(tl_read("file.csv", format = "avro"), "Unsupported format")
})

test_that("tl_read auto-detects CSV and reads correctly", {
  tmp <- tempfile(fileext = ".csv")
  on.exit(unlink(tmp), add = TRUE)
  write.csv(iris, tmp, row.names = FALSE)

  result <- tl_read(tmp, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(ncol(result), 5)
})

test_that("tl_read auto-detects RDS and reads correctly", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  saveRDS(mtcars, tmp)

  result <- tl_read(tmp, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

test_that("tl_read respects explicit format override", {
  tmp <- tempfile(fileext = ".txt")
  on.exit(unlink(tmp), add = TRUE)
  write.table(iris, tmp, sep = "\t", row.names = FALSE)

  result <- tl_read(tmp, format = "tsv", .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(attr(result, "tl_format"), "tsv")
})

test_that("tl_read messages can be suppressed", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  saveRDS(iris, tmp)

  expect_silent(tl_read(tmp, .quiet = TRUE))
})

# ---- tl_read_csv ----

test_that("tl_read_csv reads CSV files", {
  tmp <- tempfile(fileext = ".csv")
  on.exit(unlink(tmp), add = TRUE)
  write.csv(iris, tmp, row.names = FALSE)

  result <- tl_read_csv(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(ncol(result), 5)
  expect_equal(attr(result, "tl_format"), "csv")
})

test_that("tl_read_csv errors on missing file", {
  expect_error(tl_read_csv("nonexistent.csv"), "File not found")
})

# ---- tl_read_tsv ----

test_that("tl_read_tsv reads TSV files", {
  tmp <- tempfile(fileext = ".tsv")
  on.exit(unlink(tmp), add = TRUE)
  write.table(iris, tmp, sep = "\t", row.names = FALSE)

  result <- tl_read_tsv(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(attr(result, "tl_format"), "tsv")
})

# ---- tl_read_excel ----

test_that("tl_read_excel requires readxl package", {
  skip_if_not_installed("readxl")

  path <- readxl::readxl_example("datasets.xlsx")
  result <- tl_read_excel(path, sheet = "mtcars")
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
  expect_equal(attr(result, "tl_format"), "excel")
})

test_that("tl_read_excel errors on missing file", {
  skip_if_not_installed("readxl")
  expect_error(tl_read_excel("nonexistent.xlsx"), "File not found")
})

# ---- tl_read_rds ----

test_that("tl_read_rds reads RDS files into tidylearn_data", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  saveRDS(mtcars, tmp)

  result <- tl_read_rds(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
  expect_equal(attr(result, "tl_format"), "rds")
})

test_that("tl_read_rds coerces data.frame to tibble", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  saveRDS(as.data.frame(iris), tmp)

  result <- tl_read_rds(tmp)
  expect_s3_class(result, "tbl_df")
})

test_that("tl_read_rds errors on non-data-frame content", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp), add = TRUE)
  saveRDS(list(a = 1, b = 2), tmp)

  expect_error(tl_read_rds(tmp), "does not contain a data frame")
})

# ---- tl_read_rdata ----

test_that("tl_read_rdata reads single-object RData files", {
  tmp <- tempfile(fileext = ".rdata")
  on.exit(unlink(tmp), add = TRUE)
  my_data <- iris
  save(my_data, file = tmp)

  result <- tl_read_rdata(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(attr(result, "tl_format"), "rdata")
})

test_that("tl_read_rdata extracts named object", {
  tmp <- tempfile(fileext = ".rdata")
  on.exit(unlink(tmp), add = TRUE)
  first_df <- iris
  second_df <- mtcars
  save(first_df, second_df, file = tmp)

  result <- tl_read_rdata(tmp, name = "second_df")
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

test_that("tl_read_rdata errors on multiple data frames without name", {
  tmp <- tempfile(fileext = ".rdata")
  on.exit(unlink(tmp), add = TRUE)
  first_df <- iris
  second_df <- mtcars
  save(first_df, second_df, file = tmp)

  expect_error(tl_read_rdata(tmp), "Multiple data frames found")
})

test_that("tl_read_rdata errors on missing named object", {
  tmp <- tempfile(fileext = ".rdata")
  on.exit(unlink(tmp), add = TRUE)
  my_data <- iris
  save(my_data, file = tmp)

  expect_error(tl_read_rdata(tmp, name = "nonexistent"), "not found")
})

test_that("tl_read_rdata errors on non-data-frame named object", {
  tmp <- tempfile(fileext = ".rdata")
  on.exit(unlink(tmp), add = TRUE)
  my_list <- list(a = 1, b = 2)
  save(my_list, file = tmp)

  expect_error(tl_read_rdata(tmp, name = "my_list"), "not a data frame")
})

# ---- Output consistency ----

test_that("all readers produce tidylearn_data output", {
  # CSV
  tmp_csv <- tempfile(fileext = ".csv")
  write.csv(iris, tmp_csv, row.names = FALSE)

  # TSV
  tmp_tsv <- tempfile(fileext = ".tsv")
  write.table(iris, tmp_tsv, sep = "\t", row.names = FALSE)

  # RDS
  tmp_rds <- tempfile(fileext = ".rds")
  saveRDS(iris, tmp_rds)

  # RData
  tmp_rdata <- tempfile(fileext = ".rdata")
  my_data <- iris
  save(my_data, file = tmp_rdata)

  on.exit(unlink(c(tmp_csv, tmp_tsv, tmp_rds, tmp_rdata)), add = TRUE)

  results <- list(
    tl_read_csv(tmp_csv),
    tl_read_tsv(tmp_tsv),
    tl_read_rds(tmp_rds),
    tl_read_rdata(tmp_rdata)
  )

  for (result in results) {
    expect_s3_class(result, "tidylearn_data")
    expect_s3_class(result, "tbl_df")
    expect_equal(nrow(result), 150)
    expect_false(is.null(attr(result, "tl_source")))
    expect_false(is.null(attr(result, "tl_format")))
    expect_false(is.null(attr(result, "tl_timestamp")))
  }
})

# ---- tl_read_parquet ----

test_that("tl_read_parquet reads parquet files", {
  skip_if_not_installed("nanoparquet")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  nanoparquet::write_parquet(iris, tmp)

  result <- tl_read_parquet(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(ncol(result), 5)
  expect_equal(attr(result, "tl_format"), "parquet")
})

test_that("tl_read auto-detects parquet format", {
  skip_if_not_installed("nanoparquet")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  nanoparquet::write_parquet(mtcars, tmp)

  result <- tl_read(tmp, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

test_that("tl_read_parquet errors on missing file", {
  skip_if_not_installed("nanoparquet")
  expect_error(tl_read_parquet("nonexistent.parquet"), "File not found")
})

# ---- tl_read_json ----

test_that("tl_read_json reads JSON files", {
  skip_if_not_installed("jsonlite")

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  jsonlite::write_json(iris, tmp)

  result <- tl_read_json(tmp)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(attr(result, "tl_format"), "json")
})

test_that("tl_read_json errors on non-tabular JSON", {
  skip_if_not_installed("jsonlite")

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  writeLines('{"key": "value"}', tmp)

  expect_error(tl_read_json(tmp), "does not contain tabular data")
})

test_that("tl_read auto-detects JSON format", {
  skip_if_not_installed("jsonlite")

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp), add = TRUE)
  jsonlite::write_json(mtcars, tmp)

  result <- tl_read(tmp, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

# ---- tl_read_sqlite ----

test_that("tl_read_sqlite queries SQLite databases", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  tmp_db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp_db), add = TRUE)

  conn <- DBI::dbConnect(RSQLite::SQLite(), tmp_db)
  DBI::dbWriteTable(conn, "iris_tbl", iris)
  DBI::dbDisconnect(conn)

  result <- tl_read_sqlite(tmp_db, "SELECT * FROM iris_tbl")
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_equal(attr(result, "tl_format"), "sqlite")
})

test_that("tl_read_sqlite errors without query", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  tmp_db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp_db), add = TRUE)
  conn <- DBI::dbConnect(RSQLite::SQLite(), tmp_db)
  DBI::dbWriteTable(conn, "t", iris)
  DBI::dbDisconnect(conn)

  expect_error(tl_read_sqlite(tmp_db), "query.*required")
})

test_that("tl_read auto-detects sqlite format", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  tmp_db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp_db), add = TRUE)

  conn <- DBI::dbConnect(RSQLite::SQLite(), tmp_db)
  DBI::dbWriteTable(conn, "mtcars_tbl", mtcars)
  DBI::dbDisconnect(conn)

  result <- tl_read(tmp_db, query = "SELECT * FROM mtcars_tbl", .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

test_that("tl_read_sqlite warns on empty result", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  tmp_db <- tempfile(fileext = ".sqlite")
  on.exit(unlink(tmp_db), add = TRUE)

  conn <- DBI::dbConnect(RSQLite::SQLite(), tmp_db)
  DBI::dbWriteTable(conn, "t", iris)
  DBI::dbDisconnect(conn)

  expect_warning(
    tl_read_sqlite(tmp_db, "SELECT * FROM t WHERE 1 = 0"),
    "0 rows"
  )
})

# ---- tl_read_db ----

test_that("tl_read_db reads from live DBI connection", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(conn), add = TRUE)
  DBI::dbWriteTable(conn, "test_tbl", mtcars)

  result <- tl_read_db(conn, "SELECT * FROM test_tbl")
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
  expect_equal(attr(result, "tl_format"), "database")
})

test_that("tl_read_db errors on non-DBI connection", {
  skip_if_not_installed("DBI")
  expect_error(tl_read_db("not_a_connection", "SELECT 1"), "DBI connection")
})

test_that("tl_read_db errors on empty query", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RSQLite")

  conn <- DBI::dbConnect(RSQLite::SQLite(), ":memory:")
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  expect_error(tl_read_db(conn, ""), "non-empty SQL")
})

# ---- tl_read_postgres (error path only) ----

test_that("tl_read_postgres requires query", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RPostgres")

  expect_error(tl_read_postgres("localhost"), "query.*required")
})

# ---- tl_read_mysql (error path only) ----

test_that("tl_read_mysql requires query", {
  skip_if_not_installed("DBI")
  skip_if_not_installed("RMariaDB")

  expect_error(tl_read_mysql("localhost"), "query.*required")
})

# ---- tl_read_bigquery (error path only) ----

test_that("tl_read_bigquery requires query", {
  skip_if_not_installed("bigrquery")

  expect_error(tl_read_bigquery("my-project"), "query.*required")
})

# ---- tl_read_github ----

test_that("tl_read_github requires path for owner/repo format", {
  expect_error(
    tl_read_github("user/repo"),
    "path.*required"
  )
})

# ---- tl_read_kaggle (error path only) ----

test_that("tl_read_kaggle errors when kaggle CLI not installed", {
  skip_on_cran()
  skip_if(nzchar(Sys.which("kaggle")), "kaggle CLI is installed")
  expect_error(tl_read_kaggle("user/dataset"), "Kaggle CLI not found")
})

# ---- tl_read_s3 (error path only) ----

test_that("tl_read_s3 errors on invalid URI", {
  skip_if_not_installed("paws.storage")
  expect_error(tl_read_s3("s3://bucket-only"), "Invalid S3 URI")
})

# ---- Internal helpers ----

test_that("tl_parse_db_url parses connection strings", {
  result <- tl_parse_db_url("mysql://user:pass@localhost:3306/mydb")
  expect_equal(result$user, "user")
  expect_equal(result$password, "pass")
  expect_equal(result$host, "localhost")
  expect_equal(result$port, 3306L)
  expect_equal(result$dbname, "mydb")
})

test_that("tl_parse_db_url handles minimal URLs", {
  result <- tl_parse_db_url("postgres://localhost/mydb")
  expect_null(result$user)
  expect_null(result$password)
  expect_equal(result$host, "localhost")
  expect_null(result$port)
  expect_equal(result$dbname, "mydb")
})

test_that("tl_parse_kaggle_url extracts dataset slug", {
  slug <- tl_parse_kaggle_url(
    "https://www.kaggle.com/datasets/zillow/zecon"
  )
  expect_equal(slug, "zillow/zecon")
})

test_that("tl_parse_kaggle_url extracts competition name", {
  slug <- tl_parse_kaggle_url(
    "https://www.kaggle.com/competitions/titanic"
  )
  expect_equal(slug, "titanic")
})

# ---- Multi-path reading ----

test_that("tl_read accepts multiple paths and row-binds", {
  tmp1 <- tempfile(fileext = ".csv")
  tmp2 <- tempfile(fileext = ".csv")
  on.exit(unlink(c(tmp1, tmp2)), add = TRUE)

  write.csv(iris[1:50, ], tmp1, row.names = FALSE)
  write.csv(iris[51:100, ], tmp2, row.names = FALSE)

  result <- tl_read(c(tmp1, tmp2), .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 100)
  expect_true("source_file" %in% names(result))
  expect_equal(length(unique(result$source_file)), 2)
})

test_that("tl_read multi-path works with explicit format", {
  tmp1 <- tempfile(fileext = ".txt")
  tmp2 <- tempfile(fileext = ".txt")
  on.exit(unlink(c(tmp1, tmp2)), add = TRUE)

  write.table(iris[1:50, ], tmp1, sep = "\t", row.names = FALSE)
  write.table(iris[51:100, ], tmp2, sep = "\t", row.names = FALSE)

  result <- tl_read(c(tmp1, tmp2), format = "tsv", .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 100)
})

# ---- tl_read_dir ----

test_that("tl_read_dir reads all CSVs from a directory", {
  dir <- tempfile(pattern = "tl_test_dir_")
  dir.create(dir)
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  write.csv(iris[1:50, ], file.path(dir, "part1.csv"), row.names = FALSE)
  write.csv(iris[51:100, ], file.path(dir, "part2.csv"), row.names = FALSE)

  result <- tl_read_dir(dir, format = "csv", .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 100)
  expect_true("source_file" %in% names(result))
  expect_equal(sort(unique(result$source_file)), c("part1.csv", "part2.csv"))
})

test_that("tl_read_dir works with pattern filter", {
  dir <- tempfile(pattern = "tl_test_dir_")
  dir.create(dir)
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  write.csv(iris[1:50, ], file.path(dir, "sales_jan.csv"), row.names = FALSE)
  write.csv(iris[51:100, ], file.path(dir, "sales_feb.csv"), row.names = FALSE)
  write.csv(iris[101:150, ], file.path(dir, "other.csv"), row.names = FALSE)

  result <- tl_read_dir(dir, pattern = "^sales_", .quiet = TRUE)
  expect_equal(nrow(result), 100)
  expect_equal(length(unique(result$source_file)), 2)
})

test_that("tl_read_dir scans recursively when asked", {
  dir <- tempfile(pattern = "tl_test_dir_")
  dir.create(dir)
  subdir <- file.path(dir, "sub")
  dir.create(subdir)
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  write.csv(iris[1:50, ], file.path(dir, "a.csv"), row.names = FALSE)
  write.csv(iris[51:100, ], file.path(subdir, "b.csv"), row.names = FALSE)

  # Non-recursive misses subdir file
  result_flat <- tl_read_dir(dir, format = "csv", .quiet = TRUE)
  expect_equal(nrow(result_flat), 50)

  # Recursive finds both
  result_deep <- tl_read_dir(dir, format = "csv", recursive = TRUE,
                             .quiet = TRUE)
  expect_equal(nrow(result_deep), 100)
})

test_that("tl_read_dir errors on empty directory", {
  dir <- tempfile(pattern = "tl_test_dir_")
  dir.create(dir)
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  expect_error(tl_read_dir(dir, .quiet = TRUE), "No data files found")
})

test_that("tl_read_dir errors on non-existent directory", {
  expect_error(tl_read_dir("/fake/dir"), "Directory not found")
})

test_that("tl_read dispatches to tl_read_dir for directories", {
  dir <- tempfile(pattern = "tl_test_dir_")
  dir.create(dir)
  on.exit(unlink(dir, recursive = TRUE), add = TRUE)

  write.csv(iris, file.path(dir, "data.csv"), row.names = FALSE)

  result <- tl_read(dir, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
})

# ---- tl_read_zip ----

test_that("tl_read_zip reads single file from archive", {
  dir <- tempfile(pattern = "tl_zip_src_")
  dir.create(dir)
  zip_path <- tempfile(fileext = ".zip")
  on.exit(unlink(c(dir, zip_path), recursive = TRUE), add = TRUE)

  csv_path <- file.path(dir, "data.csv")
  write.csv(iris, csv_path, row.names = FALSE)

  # Create zip - use full path for the file inside
  old_wd <- getwd()
  setwd(dir)
  utils::zip(zip_path, "data.csv")
  setwd(old_wd)

  result <- tl_read_zip(zip_path, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
  expect_true(grepl("zip\\+", attr(result, "tl_format")))
})

test_that("tl_read_zip reads specific file from archive", {
  dir <- tempfile(pattern = "tl_zip_src_")
  dir.create(dir)
  zip_path <- tempfile(fileext = ".zip")
  on.exit(unlink(c(dir, zip_path), recursive = TRUE), add = TRUE)

  write.csv(iris, file.path(dir, "iris.csv"), row.names = FALSE)
  write.csv(mtcars, file.path(dir, "mtcars.csv"), row.names = FALSE)

  old_wd <- getwd()
  setwd(dir)
  utils::zip(zip_path, c("iris.csv", "mtcars.csv"))
  setwd(old_wd)

  result <- tl_read_zip(zip_path, file = "mtcars.csv", .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 32)
})

test_that("tl_read_zip row-binds multiple files", {
  dir <- tempfile(pattern = "tl_zip_src_")
  dir.create(dir)
  zip_path <- tempfile(fileext = ".zip")
  on.exit(unlink(c(dir, zip_path), recursive = TRUE), add = TRUE)

  write.csv(iris[1:50, ], file.path(dir, "part1.csv"), row.names = FALSE)
  write.csv(iris[51:100, ], file.path(dir, "part2.csv"), row.names = FALSE)

  old_wd <- getwd()
  setwd(dir)
  utils::zip(zip_path, c("part1.csv", "part2.csv"))
  setwd(old_wd)

  result <- tl_read_zip(zip_path, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 100)
  expect_true("source_file" %in% names(result))
})

test_that("tl_read_zip errors on missing file in archive", {
  dir <- tempfile(pattern = "tl_zip_src_")
  dir.create(dir)
  zip_path <- tempfile(fileext = ".zip")
  on.exit(unlink(c(dir, zip_path), recursive = TRUE), add = TRUE)

  write.csv(iris, file.path(dir, "data.csv"), row.names = FALSE)

  old_wd <- getwd()
  setwd(dir)
  utils::zip(zip_path, "data.csv")
  setwd(old_wd)

  expect_error(
    tl_read_zip(zip_path, file = "nonexistent.csv", .quiet = TRUE),
    "not found in archive"
  )
})

test_that("tl_read dispatches to tl_read_zip for .zip files", {
  dir <- tempfile(pattern = "tl_zip_src_")
  dir.create(dir)
  zip_path <- tempfile(fileext = ".zip")
  on.exit(unlink(c(dir, zip_path), recursive = TRUE), add = TRUE)

  write.csv(iris, file.path(dir, "data.csv"), row.names = FALSE)

  old_wd <- getwd()
  setwd(dir)
  utils::zip(zip_path, "data.csv")
  setwd(old_wd)

  result <- tl_read(zip_path, .quiet = TRUE)
  expect_s3_class(result, "tidylearn_data")
  expect_equal(nrow(result), 150)
})

#' Tidy K-Means Clustering
#'
#' Performs k-means clustering with tidy output
#'
#' @param data A data frame or tibble
#' @param k Number of clusters
#' @param cols Columns to include (tidy select).
#'   If NULL, uses all numeric columns.
#' @param nstart Number of random starts (default: 25)
#' @param iter_max Maximum iterations (default: 100)
#' @param algorithm K-means algorithm: "Hartigan-Wong"
#'   (default), "Lloyd", "Forgy", "MacQueen"
#'
#' @return A list of class "tidy_kmeans" containing:
#' \itemize{
#'   \item clusters: tibble with observation IDs and cluster assignments
#'   \item centers: tibble of cluster centers
#'   \item metrics: tibble with clustering quality metrics
#'   \item model: original kmeans object
#' }
#'
#' @examples
#' # Basic k-means
#' km_result <- tidy_kmeans(iris, k = 3)
#'
#' @export
tidy_kmeans <- function(data, k, cols = NULL, nstart = 25, iter_max = 100,
                        algorithm = "Hartigan-Wong") {

  # Select columns
  if (!is.null(cols)) {
    cols_enquo <- rlang::enquo(cols)
    data_selected <- data %>% dplyr::select(!!cols_enquo)
  } else {
    data_selected <- data %>% dplyr::select(where(is.numeric))
  }

  # Perform k-means
  km_model <- stats::kmeans(data_selected, centers = k, nstart = nstart,
                            iter.max = iter_max, algorithm = algorithm)

  # Create clusters tibble
  clusters_tbl <- tibble::tibble(
    .obs_id = rownames(data) %||% as.character(seq_len(nrow(data))),
    cluster = as.integer(km_model$cluster)
  )

  # Create centers tibble
  centers_tbl <- tibble::as_tibble(km_model$centers) %>%
    dplyr::mutate(cluster = seq_len(k), .before = 1)

  # Create metrics tibble
  metrics_tbl <- tibble::tibble(
    k = k,
    tot_withinss = km_model$tot.withinss,
    betweenss = km_model$betweenss,
    tot_ss = km_model$totss,
    iter = km_model$iter,
    converged = km_model$ifault == 0
  )

  # Return tidy object
  result <- list(
    clusters = clusters_tbl,
    centers = centers_tbl,
    metrics = metrics_tbl,
    sizes = km_model$size,
    model = km_model
  )

  class(result) <- c("tidy_kmeans", "list")
  result
}


#' Augment Data with K-Means Cluster Assignments
#'
#' @param kmeans_obj A tidy_kmeans object
#' @param data Original data frame
#'
#' @return A tibble containing the original \code{data} with an additional
#'   \code{cluster} factor column indicating cluster assignments.
#'
#' @examples
#' \donttest{
#' km <- tidy_kmeans(iris[, 1:4], k = 3)
#' augmented <- augment_kmeans(km, iris)
#' }
#'
#' @export
augment_kmeans <- function(kmeans_obj, data) {

  if (!inherits(kmeans_obj, "tidy_kmeans")) {
    stop("kmeans_obj must be a tidy_kmeans object")
  }

  data %>%
    dplyr::bind_cols(
      tibble::tibble(cluster = as.factor(kmeans_obj$model$cluster))
    )
}


#' Tidy PAM (Partitioning Around Medoids)
#'
#' Performs PAM clustering with tidy output
#'
#' @param data A data frame, tibble, or dist object
#' @param k Number of clusters
#' @param metric Distance metric (default: "euclidean").
#'   Use "gower" for mixed data types.
#' @param cols Columns to include (tidy select). If NULL, uses all columns.
#'
#' @return A list of class "tidy_pam" containing:
#' \itemize{
#'   \item clusters: tibble with observation IDs and cluster assignments
#'   \item medoids: tibble of medoid indices and values
#'   \item silhouette: average silhouette width
#'   \item model: original pam object
#' }
#'
#' @examples
#' # PAM with Euclidean distance
#' pam_result <- tidy_pam(iris, k = 3)
#'
#' # PAM with Gower distance for mixed data
#' pam_result <- tidy_pam(mtcars, k = 3, metric = "gower")
#'
#' @export
tidy_pam <- function(data, k, metric = "euclidean", cols = NULL) {

  # Handle dist object
  if (inherits(data, "dist")) {
    dist_mat <- data
    data_orig <- NULL
  } else {
    # Select columns
    if (!is.null(cols)) {
      cols_enquo <- rlang::enquo(cols)
      data_selected <- data %>% dplyr::select(!!cols_enquo)
    } else {
      data_selected <- data
    }

    # Compute distance
    if (metric == "gower") {
      dist_mat <- tidy_gower(data_selected)
    } else {
      dist_mat <- tidy_dist(data_selected, method = metric)
    }

    data_orig <- data_selected
  }

  # Perform PAM
  pam_model <- cluster::pam(dist_mat, k = k, diss = TRUE)

  # Create clusters tibble
  clusters_tbl <- tibble::tibble(
    .obs_id = names(pam_model$clustering) %||%
      as.character(seq_along(pam_model$clustering)),
    cluster = as.integer(pam_model$clustering)
  )

  # Create medoids tibble
  if (!is.null(data_orig)) {
    medoid_data <- data_orig[pam_model$medoids, , drop = FALSE]
    medoids_tbl <- tibble::as_tibble(medoid_data) %>%
      dplyr::mutate(
        cluster = seq_len(k),
        medoid_index = pam_model$medoids,
        .before = 1
      )
  } else {
    medoids_tbl <- tibble::tibble(
      cluster = seq_len(k),
      medoid_index = pam_model$medoids
    )
  }

  # Return tidy object
  result <- list(
    clusters = clusters_tbl,
    medoids = medoids_tbl,
    silhouette_avg = pam_model$silinfo$avg.width,
    silhouette_data = pam_model$silinfo,
    model = pam_model
  )

  class(result) <- c("tidy_pam", "list")
  result
}


#' Augment Data with PAM Cluster Assignments
#'
#' @param pam_obj A tidy_pam object
#' @param data Original data frame
#'
#' @return A tibble containing the original \code{data} with an additional
#'   \code{cluster} factor column indicating cluster assignments.
#'
#' @examples
#' \donttest{
#' pm <- tidy_pam(iris[, 1:4], k = 3)
#' augmented <- augment_pam(pm, iris)
#' }
#'
#' @export
augment_pam <- function(pam_obj, data) {

  if (!inherits(pam_obj, "tidy_pam")) {
    stop("pam_obj must be a tidy_pam object")
  }

  data %>%
    dplyr::bind_cols(
      tibble::tibble(cluster = as.factor(pam_obj$model$clustering))
    )
}


#' Tidy CLARA (Clustering Large Applications)
#'
#' Performs CLARA clustering (scalable version of PAM)
#'
#' @param data A data frame or tibble
#' @param k Number of clusters
#' @param metric Distance metric (default: "euclidean")
#' @param samples Number of samples to draw (default: 50)
#' @param sampsize Sample size (default: min(n, 40 + 2*k))
#'
#' @return A list of class \code{"tidy_clara"} containing:
#' \itemize{
#'   \item clusters: tibble with observation IDs and cluster assignments
#'   \item medoids: tibble of medoid values
#'   \item silhouette_avg: average silhouette width
#'   \item model: original \code{\link[cluster]{clara}} object
#' }
#'
#' @examples
#' \donttest{
#' # CLARA for large datasets
#' large_data <- iris[rep(1:nrow(iris), 10), 1:4]
#' clara_result <- tidy_clara(large_data, k = 3, samples = 50)
#' print(clara_result)
#' }
#'
#' @export
tidy_clara <- function(data, k, metric = "euclidean",
                       samples = 50, sampsize = NULL) {

  # Select numeric columns if data frame
  if (!inherits(data, "dist")) {
    data_numeric <- data %>% dplyr::select(where(is.numeric))
  } else {
    data_numeric <- data
  }

  # Set default sampsize if not provided
  if (is.null(sampsize)) {
    n <- nrow(data_numeric)
    sampsize <- min(n, 40 + 2 * k)
  }

  # Perform CLARA
  clara_model <- cluster::clara(
    data_numeric,
    k = k,
    metric = metric,
    samples = samples,
    sampsize = sampsize
  )

  # Create clusters tibble
  clusters_tbl <- tibble::tibble(
    .obs_id = names(clara_model$clustering) %||%
      as.character(seq_along(clara_model$clustering)),
    cluster = as.integer(clara_model$clustering)
  )

  # Create medoids tibble
  medoids_tbl <- tibble::as_tibble(clara_model$medoids) %>%
    dplyr::mutate(cluster = seq_len(k), .before = 1)

  # Return tidy object
  result <- list(
    clusters = clusters_tbl,
    medoids = medoids_tbl,
    silhouette_avg = clara_model$silinfo$avg.width,
    model = clara_model
  )

  class(result) <- c("tidy_clara", "list")
  result
}


#' Calculate Within-Cluster Sum of Squares for Different k
#'
#' Used for elbow method to determine optimal k
#'
#' @param data A data frame or tibble
#' @param max_k Maximum number of clusters to test (default: 10)
#' @param nstart Number of random starts for each k (default: 25)
#'
#' @return A tibble with columns \code{k} (number of clusters) and
#'   \code{tot_withinss} (total within-cluster sum of squares).
#'
#' @examples
#' \donttest{
#' wss <- calc_wss(iris[, 1:4], max_k = 6)
#' plot(wss$k, wss$tot_withinss, type = "b")
#' }
#'
#' @export
calc_wss <- function(data, max_k = 10, nstart = 25) {

  data_numeric <- data %>% dplyr::select(where(is.numeric))

  wss_values <- purrr::map_dbl(1:max_k, function(k) {
    if (k == 1) {
      # For k=1, total SS is the total variance
      sum(apply(data_numeric, 2, var) * (nrow(data_numeric) - 1))
    } else {
      km <- stats::kmeans(data_numeric, centers = k, nstart = nstart)
      km$tot.withinss
    }
  })

  tibble::tibble(
    k = 1:max_k,
    tot_withinss = wss_values
  )
}


#' Find Optimal Number of Clusters
#'
#' Use multiple methods to suggest optimal k
#'
#' @param data A data frame or tibble
#' @param max_k Maximum k to test (default: 10)
#' @param methods Vector of methods: "silhouette", "gap", "wss" (default: all)
#'
#' @return A list of class \code{"optimal_k_results"} containing one or more of:
#' \itemize{
#'   \item wss: tibble from \code{\link{calc_wss}} (if "wss" method used)
#'   \item silhouette: tibble from \code{\link{tidy_silhouette_analysis}}
#'     (if "silhouette" method used)
#'   \item gap: a \code{tidy_gap} object from \code{\link{tidy_gap_stat}}
#'     (if "gap" method used)
#' }
#'
#' @examples
#' \donttest{
#' opt <- optimal_clusters(iris[, 1:4], max_k = 6, methods = "wss")
#' }
#'
#' @export
optimal_clusters <- function(
    data, max_k = 10,
    methods = c("silhouette", "gap", "wss")) {

  results <- list()

  if ("wss" %in% methods) {
    results$wss <- calc_wss(data, max_k = max_k)
  }

  if ("silhouette" %in% methods) {
    sil_results <- tidy_silhouette_analysis(data, max_k = max_k)
    results$silhouette <- sil_results
  }

  if ("gap" %in% methods) {
    gap_results <- tidy_gap_stat(data, max_k = max_k)
    results$gap <- gap_results
  }

  class(results) <- c("optimal_k_results", "list")
  results
}


#' Print Method for tidy_kmeans
#'
#' @param x A tidy_kmeans object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' km <- tidy_kmeans(iris[, 1:4], k = 3)
#' print(km)
#' }
#'
#' @export
print.tidy_kmeans <- function(x, ...) {
  cat("Tidy K-Means Clustering\n")
  cat("=======================\n\n")
  cat("Number of clusters:", x$metrics$k, "\n")
  cat("Cluster sizes:", paste(x$sizes, collapse = ", "), "\n")
  cat("Total within-cluster SS:", round(x$metrics$tot_withinss, 2), "\n")
  cat("Between-cluster SS:", round(x$metrics$betweenss, 2), "\n")
  cat("Iterations:", x$metrics$iter, "\n")
  cat("Converged:", x$metrics$converged, "\n\n")

  cat("Cluster Centers:\n")
  print(x$centers)

  invisible(x)
}


#' Print Method for tidy_pam
#'
#' @param x A tidy_pam object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' pm <- tidy_pam(iris[, 1:4], k = 3)
#' print(pm)
#' }
#'
#' @export
print.tidy_pam <- function(x, ...) {
  cat("Tidy PAM Clustering\n")
  cat("===================\n\n")
  cat("Number of clusters:", nrow(x$medoids), "\n")
  cat("Average silhouette width:", round(x$silhouette_avg, 4), "\n\n")

  cat("Medoids:\n")
  print(x$medoids)

  cat("\nCluster sizes:\n")
  print(table(x$clusters$cluster))

  invisible(x)
}

#' Fit K-means for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_kmeans <- function(data, formula = NULL, k = 3, ...) {
  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_km <- data[, vars, drop = FALSE]
  } else {
    data_for_km <- data %>% dplyr::select(where(is.numeric))
  }

  # Fit k-means using tidy_kmeans
  km_result <- tidy_kmeans(data_for_km, k = k, ...)

  # Return in expected format
  list(
    clusters = km_result$clusters,
    centers = km_result$centers,
    metrics = km_result$metrics,
    model = km_result$model
  )
}

#' Fit PAM for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_pam <- function(data, formula = NULL, k = 3, ...) {
  tl_check_packages("cluster")

  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_pam <- data[, vars, drop = FALSE]
  } else {
    data_for_pam <- data
  }

  # Fit PAM using tidy_pam
  pam_result <- tidy_pam(data_for_pam, k = k, ...)

  # Return in expected format
  list(
    clusters = pam_result$clusters,
    medoids = pam_result$medoids,
    silhouette_avg = pam_result$silhouette_avg,
    model = pam_result$model
  )
}

#' Fit CLARA for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_clara <- function(data, formula = NULL, k = 3, ...) {
  tl_check_packages("cluster")

  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_clara <- data[, vars, drop = FALSE]
  } else {
    data_for_clara <- data
  }

  # Fit CLARA using tidy_clara
  clara_result <- tidy_clara(data_for_clara, k = k, ...)

  # Return in expected format
  list(
    clusters = clara_result$clusters,
    medoids = clara_result$medoids,
    silhouette_avg = clara_result$silhouette_avg,
    model = clara_result$model
  )
}

#' Tidy Principal Component Analysis
#'
#' Performs PCA on a dataset using tidyverse principles. Returns a tidy list
#' containing scores, loadings, variance explained, and the original model.
#'
#' @param data A data frame or tibble
#' @param cols Columns to include in PCA (tidy select
#'   syntax). If NULL, uses all numeric columns.
#' @param scale Logical; should variables be scaled to
#'   unit variance? Default TRUE.
#' @param center Logical; should variables be centered? Default TRUE.
#' @param method Character; "prcomp" (default, recommended) or "princomp"
#'
#' @return A list of class "tidy_pca" containing:
#' \itemize{
#'   \item scores: tibble of PC scores with observation identifiers
#'   \item loadings: tibble of variable loadings in long format
#'   \item variance: tibble of variance explained by each PC
#'   \item model: the original prcomp/princomp object
#'   \item settings: list of scale, center, method used
#' }
#'
#' @examples
#' # Basic PCA
#' pca_result <- tidy_pca(USArrests)
#'
#'
#' # Access components
#' pca_result$scores
#' pca_result$loadings
#' pca_result$variance
#'
#' @export
tidy_pca <- function(data, cols = NULL, scale = TRUE,
                     center = TRUE, method = "prcomp") {

  # Convert to data frame if needed
  data <- as.data.frame(data)

  # Select columns
  if (!is.null(cols)) {
    cols_enquo <- rlang::enquo(cols)
    data_selected <- data %>% dplyr::select(!!cols_enquo)
  } else {
    # Select only numeric columns
    data_selected <- data %>% dplyr::select(where(is.numeric))
  }

  # Preserve row names/identifiers
  if (!is.null(rownames(data))) {
    obs_id <- rownames(data)
  } else {
    obs_id <- paste0("obs_", seq_len(nrow(data)))
  }

  # Perform PCA
  if (method == "prcomp") {
    pca_model <- stats::prcomp(data_selected, scale. = scale, center = center)
    scores_matrix <- pca_model$x
    loadings_matrix <- pca_model$rotation
    sdev <- pca_model$sdev
  } else if (method == "princomp") {
    pca_model <- stats::princomp(data_selected, cor = scale, scores = TRUE)
    scores_matrix <- pca_model$scores
    loadings_matrix <- pca_model$loadings
    sdev <- pca_model$sdev
  } else {
    stop("method must be 'prcomp' or 'princomp'")
  }

  # Create tidy scores tibble
  scores_tbl <- tibble::as_tibble(scores_matrix) %>%
    dplyr::mutate(.obs_id = obs_id, .before = 1)

  # Create tidy loadings tibble (long format)
  loadings_tbl <- tibble::as_tibble(loadings_matrix, rownames = "variable") %>%
    tidyr::pivot_longer(
      cols = -variable,
      names_to = "component",
      values_to = "loading"
    )

  # Calculate variance explained
  variance <- sdev^2
  prop_var <- variance / sum(variance)
  cumsum_var <- cumsum(prop_var)

  variance_tbl <- tibble::tibble(
    component = colnames(scores_matrix),
    sdev = sdev,
    variance = variance,
    prop_variance = prop_var,
    cum_variance = cumsum_var
  )

  # Return tidy PCA object
  result <- list(
    scores = scores_tbl,
    loadings = loadings_tbl,
    variance = variance_tbl,
    model = pca_model,
    settings = list(scale = scale, center = center, method = method)
  )

  class(result) <- c("tidy_pca", "list")
  result
}


#' Get PCA Loadings in Wide Format
#'
#' @param pca_obj A tidy_pca object
#' @param n_components Number of components to include (default: all)
#'
#' @return A tibble with one row per variable and one column per principal
#'   component, containing the loading values.
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' get_pca_loadings(pca, n_components = 2)
#' }
#'
#' @export
get_pca_loadings <- function(pca_obj, n_components = NULL) {
  if (!inherits(pca_obj, "tidy_pca")) {
    stop("pca_obj must be a tidy_pca object")
  }

  loadings <- pca_obj$loadings

  if (!is.null(n_components)) {
    components_to_keep <- unique(loadings$component)[1:n_components]
    loadings <- loadings %>%
      dplyr::filter(component %in% components_to_keep)
  }

  loadings %>%
    tidyr::pivot_wider(
      names_from = component,
      values_from = loading
    )
}


#' Get Variance Explained Summary
#'
#' @param pca_obj A tidy_pca object
#'
#' @return A tibble with columns \code{component}, \code{sdev},
#'   \code{variance}, \code{prop_variance}, and \code{cum_variance}.
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' get_pca_variance(pca)
#' }
#'
#' @export
get_pca_variance <- function(pca_obj) {
  if (!inherits(pca_obj, "tidy_pca")) {
    stop("pca_obj must be a tidy_pca object")
  }

  pca_obj$variance
}


#' Augment Original Data with PCA Scores
#'
#' Add PC scores to the original dataset
#'
#' @param pca_obj A tidy_pca object
#' @param data Original data frame
#' @param n_components Number of PCs to add (default: all)
#'
#' @return A tibble containing the original \code{data} with additional columns
#'   for each principal component score (named \code{PC1}, \code{PC2}, etc.).
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' augmented <- augment_pca(pca, USArrests, n_components = 2)
#' }
#'
#' @export
augment_pca <- function(pca_obj, data, n_components = NULL) {
  if (!inherits(pca_obj, "tidy_pca")) {
    stop("pca_obj must be a tidy_pca object")
  }

  scores <- pca_obj$scores %>% dplyr::select(-.obs_id)

  if (!is.null(n_components)) {
    scores <- scores %>% dplyr::select(1:n_components)
  }

  dplyr::bind_cols(data, scores)
}


#' Create PCA Scree Plot
#'
#' Visualize variance explained by each principal component
#'
#' @param pca_obj A tidy_pca object
#' @param type Character; "variance" or "proportion" (default)
#' @param add_line Logical; add horizontal line at
#'   eigenvalue = 1? (for Kaiser criterion)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' tidy_pca_screeplot(pca)
#' }
#'
#' @export
tidy_pca_screeplot <- function(pca_obj, type = "proportion", add_line = TRUE) {
  if (!inherits(pca_obj, "tidy_pca")) {
    stop("pca_obj must be a tidy_pca object")
  }

  var_data <- pca_obj$variance

  if (type == "proportion") {
    p <- ggplot2::ggplot(
      var_data,
      ggplot2::aes(x = seq_along(component), y = prop_variance)
    ) +
      ggplot2::geom_line(color = "steelblue", linewidth = 1) +
      ggplot2::geom_point(color = "steelblue", size = 3) +
      ggplot2::labs(
        title = "Scree Plot - Proportion of Variance Explained",
        x = "Principal Component",
        y = "Proportion of Variance"
      )
  } else {
    p <- ggplot2::ggplot(
      var_data,
      ggplot2::aes(x = seq_along(component), y = variance)
    ) +
      ggplot2::geom_line(color = "steelblue", linewidth = 1) +
      ggplot2::geom_point(color = "steelblue", size = 3) +
      ggplot2::labs(
        title = "Scree Plot - Variance (Eigenvalues)",
        x = "Principal Component",
        y = "Variance (Eigenvalue)"
      )

    if (add_line && pca_obj$settings$scale) {
      p <- p + ggplot2::geom_hline(
        yintercept = 1,
        linetype = "dashed",
        color = "red"
      )
    }
  }

  p + ggplot2::theme_minimal()
}


#' Create PCA Biplot
#'
#' Visualize both observations and variables in PC space
#'
#' @param pca_obj A tidy_pca object
#' @param pc_x Principal component for x-axis (default: 1)
#' @param pc_y Principal component for y-axis (default: 2)
#' @param color_by Optional column name to color points by
#' @param arrow_scale Scaling factor for variable arrows (default: 1)
#' @param label_obs Logical; label observations? (default: FALSE)
#' @param label_vars Logical; label variables? (default: TRUE)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' tidy_pca_biplot(pca)
#' }
#'
#' @export
tidy_pca_biplot <- function(pca_obj, pc_x = 1, pc_y = 2,
                            color_by = NULL,
                            arrow_scale = 1,
                            label_obs = FALSE,
                            label_vars = TRUE) {
  if (!inherits(pca_obj, "tidy_pca")) {
    stop("pca_obj must be a tidy_pca object")
  }

  # Get scores
  scores <- pca_obj$scores
  pc_x_name <- paste0("PC", pc_x)
  pc_y_name <- paste0("PC", pc_y)

  if (!pc_x_name %in% names(scores) || !pc_y_name %in% names(scores)) {
    stop("Requested PCs not available")
  }

  # Get loadings for these PCs
  loadings_wide <- pca_obj$loadings %>%
    dplyr::filter(component %in% c(pc_x_name, pc_y_name)) %>%
    tidyr::pivot_wider(names_from = component, values_from = loading)

  # Scale factor for arrows
  score_range <- max(abs(c(scores[[pc_x_name]], scores[[pc_y_name]])))
  loading_range <- max(
    abs(c(loadings_wide[[pc_x_name]], loadings_wide[[pc_y_name]]))
  )
  arrow_scale_factor <- (score_range / loading_range) * 0.8 * arrow_scale

  loadings_wide <- loadings_wide %>%
    dplyr::mutate(
      x_end = .data[[pc_x_name]] * arrow_scale_factor,
      y_end = .data[[pc_y_name]] * arrow_scale_factor
    )

  # Get variance explained
  var_exp <- pca_obj$variance
  var_x <- var_exp$prop_variance[pc_x] * 100
  var_y <- var_exp$prop_variance[pc_y] * 100

  # Create base plot
  p <- ggplot2::ggplot()

  # Add observation points
  if (!is.null(color_by)) {
    p <- p + ggplot2::geom_point(
      data = scores,
      ggplot2::aes(
        x = .data[[pc_x_name]],
        y = .data[[pc_y_name]],
        color = .data[[color_by]]
      ),
      alpha = 0.7,
      size = 2
    )
  } else {
    p <- p + ggplot2::geom_point(
      data = scores,
      ggplot2::aes(x = .data[[pc_x_name]], y = .data[[pc_y_name]]),
      alpha = 0.7,
      size = 2,
      color = "steelblue"
    )
  }

  # Add observation labels if requested
  if (label_obs) {
    p <- p + ggplot2::geom_text(
      data = scores,
      ggplot2::aes(
        x = .data[[pc_x_name]],
        y = .data[[pc_y_name]],
        label = .obs_id
      ),
      size = 2.5,
      vjust = -0.5
    )
  }

  # Add variable arrows
  p <- p + ggplot2::geom_segment(
    data = loadings_wide,
    ggplot2::aes(x = 0, y = 0, xend = x_end, yend = y_end),
    arrow = ggplot2::arrow(length = ggplot2::unit(0.2, "cm")),
    color = "red", size = 0.8, alpha = 0.7
  )

  # Add variable labels if requested
  if (label_vars) {
    p <- p + ggplot2::geom_text(
      data = loadings_wide,
      ggplot2::aes(x = x_end, y = y_end, label = variable),
      color = "red", size = 3, fontface = "bold", vjust = -0.5
    )
  }

  # Add labels and theme
  p <- p +
    ggplot2::labs(
      title = "PCA Biplot",
      x = sprintf("%s (%.1f%% variance)", pc_x_name, var_x),
      y = sprintf("%s (%.1f%% variance)", pc_y_name, var_y)
    ) +
    ggplot2::theme_minimal() +
    ggplot2::coord_equal()

  p
}


#' Print Method for tidy_pca
#'
#' @param x A tidy_pca object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' pca <- tidy_pca(USArrests)
#' print(pca)
#' }
#'
#' @export
print.tidy_pca <- function(x, ...) {
  cat("Tidy PCA Analysis\n")
  cat("=================\n\n")
  cat("Number of observations:", nrow(x$scores), "\n")
  cat("Number of variables:", length(unique(x$loadings$variable)), "\n")
  cat("Number of components:", nrow(x$variance), "\n")
  cat(
    "Settings: scale =", x$settings$scale,
    ", center =", x$settings$center, "\n\n"
  )

  cat("Variance Explained:\n")
  print(x$variance, n = 5)

  cat("\nAccess components with:\n")
  cat("  $scores    - PC scores for each observation\n")
  cat("  $loadings  - Variable loadings on each PC\n")
  cat("  $variance  - Variance explained by each PC\n")
  cat("  $model     - Original PCA model object\n")

  invisible(x)
}

#' Fit PCA for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_pca <- function(data, formula = NULL, scale = TRUE, center = TRUE, ...) {
  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_pca <- data[, vars, drop = FALSE]
  } else {
    data_for_pca <- data %>% dplyr::select(where(is.numeric))
  }

  # Fit PCA using tidy_pca
  pca_result <- tidy_pca(data_for_pca, scale = scale, center = center, ...)

  # Return in expected format
  list(
    scores = pca_result$scores,
    loadings = get_pca_loadings(pca_result),
    variance_explained = pca_result$variance,
    model = pca_result$model,
    settings = pca_result$settings
  )
}

#' Tidy Multidimensional Scaling
#'
#' Unified interface for MDS methods with tidy output
#'
#' @param data A data frame, tibble, or distance matrix
#' @param method Character; "classical" (default),
#'   "metric", "nonmetric", "sammon", or "kruskal"
#' @param ndim Number of dimensions for output (default: 2)
#' @param distance Character; distance metric if data is
#'   not already a dist object (default: "euclidean")
#' @param ... Additional arguments passed to specific MDS functions
#'
#' @return A list of class "tidy_mds" containing:
#' \itemize{
#'   \item config: tibble of MDS configuration (coordinates)
#'   \item stress: goodness-of-fit measure (if applicable)
#'   \item method: character string of method used
#'   \item model: original model object
#' }
#'
#' @examples
#' # Classical MDS
#' mds_result <- tidy_mds(eurodist, method = "classical")
#' print(mds_result)
#'
#' @export
tidy_mds <- function(data, method = "classical",
                     ndim = 2, distance = "euclidean",
                     ...) {

  # Convert to distance matrix if needed
  if (inherits(data, "dist")) {
    dist_mat <- data
  } else {
    data_matrix <- as.matrix(data %>% dplyr::select(where(is.numeric)))
    dist_mat <- stats::dist(data_matrix, method = distance)
  }

  # Call appropriate MDS method
  result <- switch(method,
    classical = tidy_mds_classical(dist_mat, ndim = ndim, ...),
    metric = tidy_mds_smacof(dist_mat, ndim = ndim, type = "ratio", ...),
    nonmetric = tidy_mds_smacof(dist_mat, ndim = ndim, type = "ordinal", ...),
    sammon = tidy_mds_sammon(dist_mat, ndim = ndim, ...),
    kruskal = tidy_mds_kruskal(dist_mat, ndim = ndim, ...),
    stop("method must be one of: classical, metric, nonmetric, sammon, kruskal")
  )

  result
}


#' Classical (Metric) MDS
#'
#' Performs classical multidimensional scaling using cmdscale()
#'
#' @param dist_mat A distance matrix (dist object)
#' @param ndim Number of dimensions (default: 2)
#' @param add_rownames Preserve row names from distance matrix (default: TRUE)
#'
#' @return A list of class \code{"tidy_mds"} containing:
#' \itemize{
#'   \item config: tibble of MDS coordinates
#'   \item stress: \code{NA} (not applicable for classical MDS)
#'   \item gof: goodness-of-fit (proportion of variance retained)
#'   \item eigenvalues: numeric vector of eigenvalues
#'   \item method: \code{"Classical MDS"}
#'   \item model: the \code{\link[stats]{cmdscale}} result
#' }
#'
#' @examples
#' \donttest{
#' d <- dist(USArrests)
#' mds <- tidy_mds_classical(d)
#' print(mds)
#' }
#'
#' @export
tidy_mds_classical <- function(dist_mat, ndim = 2, add_rownames = TRUE) {

  if (!inherits(dist_mat, "dist")) {
    stop("dist_mat must be a dist object")
  }

  # Perform classical MDS
  mds_result <- stats::cmdscale(dist_mat, k = ndim, eig = TRUE)

  # Extract configuration
  config_matrix <- mds_result$points

  # Create tibble
  colnames(config_matrix) <- paste0("Dim", 1:ndim)

  if (add_rownames && !is.null(attr(dist_mat, "Labels"))) {
    config_tbl <- tibble::as_tibble(config_matrix, .name_repair = "minimal") %>%
      dplyr::mutate(.obs_id = attr(dist_mat, "Labels"), .before = 1)
  } else {
    config_tbl <- tibble::as_tibble(config_matrix)
  }

  # Calculate GOF (goodness of fit)
  eigenvalues <- mds_result$eig
  total_var <- sum(abs(eigenvalues))
  retained_var <- sum(eigenvalues[1:ndim])
  gof <- retained_var / total_var

  result <- list(
    config = config_tbl,
    stress = NA,  # Classical MDS doesn't have stress
    gof = gof,
    eigenvalues = eigenvalues,
    method = "Classical MDS",
    model = mds_result
  )

  class(result) <- c("tidy_mds", "list")
  result
}


#' SMACOF MDS (Metric or Non-metric)
#'
#' Performs MDS using SMACOF algorithm from the smacof package
#'
#' @param dist_mat A distance matrix (dist object)
#' @param ndim Number of dimensions (default: 2)
#' @param type Character; "ratio" for metric, "ordinal"
#'   for non-metric (default: "ratio")
#' @param ... Additional arguments passed to smacof::mds()
#'
#' @return A list of class \code{"tidy_mds"} containing:
#' \itemize{
#'   \item config: tibble of MDS coordinates
#'   \item stress: stress value from the SMACOF algorithm
#'   \item method: character string describing the MDS type
#'   \item model: the \code{\link[smacof]{mds}} result
#' }
#'
#' @examples
#' \donttest{
#' d <- dist(USArrests)
#' mds <- tidy_mds_smacof(d, type = "ratio")
#' }
#'
#' @export
tidy_mds_smacof <- function(dist_mat, ndim = 2, type = "ratio", ...) {

  if (!inherits(dist_mat, "dist")) {
    stop("dist_mat must be a dist object")
  }

  # Perform SMACOF
  mds_result <- smacof::mds(dist_mat, ndim = ndim, type = type, ...)

  # Extract configuration
  config_matrix <- mds_result$conf
  colnames(config_matrix) <- paste0("Dim", 1:ndim)

  if (!is.null(attr(dist_mat, "Labels"))) {
    config_tbl <- tibble::as_tibble(config_matrix) %>%
      dplyr::mutate(.obs_id = attr(dist_mat, "Labels"), .before = 1)
  } else {
    config_tbl <- tibble::as_tibble(config_matrix)
  }

  method_name <- ifelse(type == "ordinal", "Non-metric SMACOF", "Metric SMACOF")

  result <- list(
    config = config_tbl,
    stress = mds_result$stress,
    method = method_name,
    model = mds_result
  )

  class(result) <- c("tidy_mds", "list")
  result
}


#' Sammon Mapping
#'
#' Performs Sammon's non-linear mapping
#'
#' @param dist_mat A distance matrix (dist object)
#' @param ndim Number of dimensions (default: 2)
#' @param ... Additional arguments passed to MASS::sammon()
#'
#' @return A list of class \code{"tidy_mds"} containing:
#' \itemize{
#'   \item config: tibble of MDS coordinates
#'   \item stress: Sammon stress value
#'   \item method: \code{"Sammon Mapping"}
#'   \item model: the \code{\link[MASS]{sammon}} result
#' }
#'
#' @examples
#' \donttest{
#' d <- dist(USArrests)
#' mds <- tidy_mds_sammon(d)
#' }
#'
#' @export
tidy_mds_sammon <- function(dist_mat, ndim = 2, ...) {

  if (!inherits(dist_mat, "dist")) {
    stop("dist_mat must be a dist object")
  }

  # Perform Sammon mapping
  mds_result <- MASS::sammon(dist_mat, k = ndim, trace = FALSE, ...)

  # Extract configuration
  config_matrix <- mds_result$points
  colnames(config_matrix) <- paste0("Dim", 1:ndim)

  if (!is.null(attr(dist_mat, "Labels"))) {
    config_tbl <- tibble::as_tibble(config_matrix) %>%
      dplyr::mutate(.obs_id = attr(dist_mat, "Labels"), .before = 1)
  } else {
    config_tbl <- tibble::as_tibble(config_matrix)
  }

  result <- list(
    config = config_tbl,
    stress = mds_result$stress,
    method = "Sammon Mapping",
    model = mds_result
  )

  class(result) <- c("tidy_mds", "list")
  result
}


#' Kruskal's Non-metric MDS
#'
#' Performs Kruskal's isoMDS
#'
#' @param dist_mat A distance matrix (dist object)
#' @param ndim Number of dimensions (default: 2)
#' @param ... Additional arguments passed to MASS::isoMDS()
#'
#' @return A list of class \code{"tidy_mds"} containing:
#' \itemize{
#'   \item config: tibble of MDS coordinates
#'   \item stress: Kruskal stress value
#'   \item method: \code{"Kruskal's isoMDS"}
#'   \item model: the \code{\link[MASS]{isoMDS}} result
#' }
#'
#' @examples
#' \donttest{
#' d <- dist(USArrests)
#' mds <- tidy_mds_kruskal(d)
#' }
#'
#' @export
tidy_mds_kruskal <- function(dist_mat, ndim = 2, ...) {

  if (!inherits(dist_mat, "dist")) {
    stop("dist_mat must be a dist object")
  }

  # Perform isoMDS
  mds_result <- MASS::isoMDS(dist_mat, k = ndim, trace = FALSE, ...)

  # Extract configuration
  config_matrix <- mds_result$points
  colnames(config_matrix) <- paste0("Dim", 1:ndim)

  if (!is.null(attr(dist_mat, "Labels"))) {
    config_tbl <- tibble::as_tibble(config_matrix) %>%
      dplyr::mutate(.obs_id = attr(dist_mat, "Labels"), .before = 1)
  } else {
    config_tbl <- tibble::as_tibble(config_matrix)
  }

  result <- list(
    config = config_tbl,
    stress = mds_result$stress,
    method = "Kruskal's isoMDS",
    model = mds_result
  )

  class(result) <- c("tidy_mds", "list")
  result
}


#' Plot MDS Configuration
#'
#' Visualize MDS results
#'
#' @param mds_obj A tidy_mds object
#' @param color_by Optional variable to color points by
#' @param label_points Logical; add point labels? (default: TRUE)
#' @param dim_x Which dimension for x-axis (default: 1)
#' @param dim_y Which dimension for y-axis (default: 2)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' mds <- tidy_mds(USArrests, method = "classical")
#' plot_mds(mds)
#' }
#'
#' @export
plot_mds <- function(mds_obj, color_by = NULL, label_points = TRUE,
                     dim_x = 1, dim_y = 2) {

  if (!inherits(mds_obj, "tidy_mds")) {
    stop("mds_obj must be a tidy_mds object")
  }

  config <- mds_obj$config
  dim_x_name <- paste0("Dim", dim_x)
  dim_y_name <- paste0("Dim", dim_y)

  # Base plot
  p <- ggplot2::ggplot(
    config,
    ggplot2::aes(x = .data[[dim_x_name]], y = .data[[dim_y_name]])
  )

  # Add points
  if (!is.null(color_by)) {
    p <- p + ggplot2::geom_point(
      ggplot2::aes(color = .data[[color_by]]),
      size = 3, alpha = 0.7
    )
  } else {
    p <- p + ggplot2::geom_point(size = 3, alpha = 0.7, color = "steelblue")
  }

  # Add labels
  if (label_points && ".obs_id" %in% names(config)) {
    p <- p + ggplot2::geom_text(
      ggplot2::aes(label = .obs_id),
      vjust = -0.7, size = 3
    )
  }

  # Add title with stress if available
  title <- mds_obj$method
  if (!is.na(mds_obj$stress)) {
    title <- paste0(title, sprintf(" (Stress = %.4f)", mds_obj$stress))
  } else if (!is.null(mds_obj$gof)) {
    title <- paste0(title, sprintf(" (GOF = %.2f%%)", mds_obj$gof * 100))
  }

  p <- p +
    ggplot2::labs(
      title = title,
      x = paste("Dimension", dim_x),
      y = paste("Dimension", dim_y)
    ) +
    ggplot2::theme_minimal() +
    ggplot2::coord_equal()

  p
}


#' Print Method for tidy_mds
#'
#' @param x A tidy_mds object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' mds <- tidy_mds(USArrests, method = "classical")
#' print(mds)
#' }
#'
#' @export
print.tidy_mds <- function(x, ...) {
  cat("Tidy MDS Analysis\n")
  cat("=================\n\n")
  cat("Method:", x$method, "\n")
  cat("Dimensions:", ncol(x$config) - 1, "\n")
  cat("Observations:", nrow(x$config), "\n")

  if (!is.na(x$stress)) {
    cat("Stress:", sprintf("%.4f", x$stress), "\n")
    cat("\nStress interpretation:\n")
    cat("  < 0.05: Excellent\n")
    cat("  < 0.10: Good\n")
    cat("  < 0.20: Acceptable\n")
    cat("  > 0.20: Poor\n")
  }

  if (!is.null(x$gof)) {
    cat("Goodness of Fit:", sprintf("%.2f%%", x$gof * 100), "\n")
  }

  cat("\nConfiguration (first 6 obs):\n")
  print(head(x$config))

  invisible(x)
}


#' Fit MDS for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_mds <- function(data, formula = NULL, k = 2, method = "classical", ...) {
  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_mds <- data[, vars, drop = FALSE]
  } else {
    data_for_mds <- data %>% dplyr::select(where(is.numeric))
  }

  # Fit MDS using tidy_mds
  mds_result <- tidy_mds(data_for_mds, method = method, ndim = k, ...)

  # Return in expected format
  list(
    points = mds_result$config,
    stress = mds_result$stress,
    gof = mds_result$gof,
    method = mds_result$method,
    model = mds_result$model
  )
}

#' @title Support Vector Machines for tidylearn
#' @name tidylearn-svm
#' @description SVM functionality for classification and
#'   regression
#' @importFrom e1071 svm tune
#' @importFrom stats predict
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate
NULL

#' Fit a support vector machine model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param kernel Kernel function
#'   ("linear", "polynomial", "radial", "sigmoid")
#' @param cost Cost parameter (default: 1)
#' @param gamma Gamma parameter for kernels
#'   (default: 1/ncol(data))
#' @param degree Degree for polynomial kernel (default: 3)
#' @param tune Logical indicating whether to tune
#'   hyperparameters (default: FALSE)
#' @param tune_folds Number of folds for cross-validation
#'   during tuning (default: 5)
#' @param ... Additional arguments to pass to svm()
#' @return A fitted SVM model
#' @keywords internal
tl_fit_svm <- function(data, formula,
                       is_classification = FALSE,
                       kernel = "radial",
                       cost = 1, gamma = NULL,
                       degree = 3, tune = FALSE,
                       tune_folds = 5, ...) {
  # Check if e1071 is installed
  tl_check_packages("e1071")


  # Set default gamma if not provided (use predictor count, not total columns)
  if (is.null(gamma)) {
    n_predictors <- ncol(data) - 1L
    gamma <- 1 / max(n_predictors, 1L)
  }

  # Determine SVM type based on problem type
  if (is_classification) {
    # Get response variable
    response_var <- all.vars(formula)[1]
    y <- data[[response_var]]

    # Check if binary or multiclass
    if (is.factor(y) && length(levels(y)) == 2) {
      # Binary classification
      svm_type <- "C-classification"
    } else {
      # Multiclass classification
      svm_type <- "C-classification"
    }
  } else {
    # Regression
    svm_type <- "eps-regression"
  }

  if (tune) {
    # Tune hyperparameters using cross-validation
    tune_ranges <- list(
      cost = c(0.1, 1, 10, 100)
    )
    if (kernel != "linear") {
      tune_ranges$gamma <- c(0.001, 0.01, 0.1, 1)
    }
    if (kernel == "polynomial") {
      tune_ranges$degree <- c(2, 3, 4)
    }

    tune_result <- e1071::tune(
      svm,
      train.x = formula,
      data = data,
      type = svm_type,
      kernel = kernel,
      ranges = tune_ranges,
      tunecontrol = e1071::tune.control(
        cross = tune_folds
      )
    )

    # Extract best parameters
    best_params <- tune_result$best.parameters
    cost <- best_params$cost
    if (kernel != "linear") {
      gamma <- best_params$gamma
    }
    if (kernel == "polynomial") {
      degree <- best_params$degree
    }

    # Store tuning results for later reference
    tuning_results <- tune_result
  } else {
    tuning_results <- NULL
  }

  # Fit the SVM model
  svm_model <- e1071::svm(
    formula = formula,
    data = data,
    type = svm_type,
    kernel = kernel,
    cost = cost,
    gamma = gamma,
    degree = degree,
    probability = is_classification,
    ...
  )

  # Store tuning results if available
  if (!is.null(tuning_results)) {
    attr(svm_model, "tuning_results") <- tuning_results
  }

  svm_model
}

#' Predict using a support vector machine model
#'
#' @param model A tidylearn SVM model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "prob" (for classification)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_svm <- function(model, new_data,
                           type = "response", ...) {
  # Get the SVM model
  fit <- model$fit
  is_classification <- model$spec$is_classification

  if (is_classification) {
    if (type == "prob") {
      # Check if probability model was enabled
      if (!fit$probability) {
        stop(
          "Probability estimates not available. ",
          "Refit the model with probability = TRUE.",
          call. = FALSE
        )
      }

      # Get class probabilities
      probs <- attr(
        predict(
          fit, newdata = new_data,
          probability = TRUE, ...
        ),
        "probabilities"
      )

      # Convert to tibble with appropriate column names
      class_levels <- colnames(probs)
      prob_df <- as.data.frame(probs)
      names(prob_df) <- class_levels

      tibble::as_tibble(prob_df)
    } else if (type == "class" || type == "response") {
      # Get predicted classes
      preds <- predict(fit, newdata = new_data, ...)
      preds
    } else {
      stop(
        "Invalid prediction type for SVM ",
        "classification. Use 'prob', 'class', ",
        "or 'response'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    preds <- predict(fit, newdata = new_data, ...)
    preds
  }
}

#' Plot SVM decision boundary
#'
#' @param model A tidylearn SVM model object
#' @param x_var Name of the x-axis variable
#' @param y_var Name of the y-axis variable
#' @param grid_size Number of points in each dimension
#'   for the grid (default: 100)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_point geom_contour
#'   scale_fill_gradient2 labs theme_minimal
#' @examples
#' \donttest{
#' if (requireNamespace("e1071", quietly = TRUE)) {
#'   model <- tl_model(iris, Species ~ ., method = "svm")
#'   tl_plot_svm_boundary(model,
#'     x_var = "Sepal.Length", y_var = "Sepal.Width")
#' }
#' }
#' @export
tl_plot_svm_boundary <- function(model,
                                 x_var = NULL,
                                 y_var = NULL,
                                 grid_size = 100,
                                 ...) {
  if (model$spec$method != "svm") {
    stop(
      "Decision boundary plot is only available ",
      "for SVM models",
      call. = FALSE
    )
  }

  if (!model$spec$is_classification) {
    stop(
      "Decision boundary plot is only available ",
      "for classification models",
      call. = FALSE
    )
  }

  # Get original data
  data <- model$data
  formula <- model$spec$formula
  response_var <- all.vars(formula)[1]

  # If x_var and y_var are not specified, use first two predictors
  if (is.null(x_var) || is.null(y_var)) {
    # Use data columns instead of all.vars() -- all.vars(y ~ .) returns only "y"
    predictor_vars <- setdiff(names(data), response_var)
    # Keep only numeric predictors for the boundary grid
    predictor_vars <- predictor_vars[
      vapply(data[predictor_vars], is.numeric, logical(1))
    ]
    if (length(predictor_vars) < 2) {
      stop(
        "At least two numeric predictor variables are ",
        "required for decision boundary plot",
        call. = FALSE
      )
    }
    x_var <- predictor_vars[1]
    y_var <- predictor_vars[2]
  }

  # Check if variables exist
  if (!x_var %in% names(data) ||
        !y_var %in% names(data)) {
    stop(
      "Variables not found in the model data",
      call. = FALSE
    )
  }

  # Create grid for prediction
  x_range <- range(data[[x_var]], na.rm = TRUE)
  y_range <- range(data[[y_var]], na.rm = TRUE)

  x_grid <- seq(
    x_range[1], x_range[2], length.out = grid_size
  )
  y_grid <- seq(
    y_range[1], y_range[2], length.out = grid_size
  )

  grid_data <- expand.grid(x = x_grid, y = y_grid)
  names(grid_data) <- c(x_var, y_var)

  # Add other predictors with mean values
  other_vars <- setdiff(
    names(data), c(response_var, x_var, y_var)
  )
  for (var in other_vars) {
    if (is.factor(data[[var]]) ||
          is.character(data[[var]])) {
      # For categorical variables, use most frequent value
      most_freq <- names(
        sort(table(data[[var]]), decreasing = TRUE)[1]
      )
      grid_data[[var]] <- most_freq
    } else {
      # For continuous variables, use mean
      grid_data[[var]] <- mean(
        data[[var]], na.rm = TRUE
      )
    }
  }

  # Make predictions on the grid -- always produce predicted class labels
  preds <- predict(model$fit, newdata = grid_data)
  grid_data$pred_class <- as.character(preds)

  # For binary classification, also get numeric probability for contour line
  has_numeric_pred <- FALSE
  if (model$fit$type == "C-classification" && model$fit$probability) {
    probs <- attr(
      predict(model$fit, newdata = grid_data, probability = TRUE),
      "probabilities"
    )
    if (!is.null(probs) && ncol(probs) == 2) {
      grid_data$pred_prob <- probs[, 2]
      has_numeric_pred <- TRUE
    }
  }

  # Create plot using geom_raster for decision regions
  p <- ggplot2::ggplot() +
    ggplot2::geom_raster(
      data = grid_data,
      ggplot2::aes(
        x = .data[[x_var]],
        y = .data[[y_var]],
        fill = .data[["pred_class"]]
      ),
      alpha = 0.3
    )

  # Add decision boundary contour line for binary classification
  if (has_numeric_pred) {
    p <- p + ggplot2::geom_contour(
      data = grid_data,
      ggplot2::aes(
        x = .data[[x_var]],
        y = .data[[y_var]],
        z = .data[["pred_prob"]]
      ),
      breaks = 0.5,
      color = "black",
      linewidth = 1
    )
  }

  # Add original data points
  p <- p +
    ggplot2::geom_point(
      data = data,
      ggplot2::aes(
        x = .data[[x_var]],
        y = .data[[y_var]],
        color = .data[[response_var]]
      ),
      size = 3,
      alpha = 0.7
    ) +
    ggplot2::labs(
      title = "SVM Decision Boundary",
      x = x_var,
      y = y_var,
      color = response_var,
      fill = "Predicted Class"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Plot SVM tuning results
#'
#' @param model A tidylearn SVM model object
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_tile
#'   scale_fill_gradient2 labs theme_minimal
#' @examples
#' \donttest{
#' if (requireNamespace("e1071", quietly = TRUE)) {
#'   model <- tl_model(iris, Species ~ ., method = "svm",
#'     kernel = "linear", tune = TRUE, tune_folds = 2)
#'   tl_plot_svm_tuning(model)
#' }
#' }
#' @export
tl_plot_svm_tuning <- function(model, ...) {
  if (model$spec$method != "svm") {
    stop(
      "Tuning plot is only available for SVM models",
      call. = FALSE
    )
  }

  # Check if tuning results are available
  tuning_results <- attr(model$fit, "tuning_results")
  if (is.null(tuning_results)) {
    stop(
      "No tuning results available. ",
      "Fit the model with tune = TRUE.",
      call. = FALSE
    )
  }

  # Extract performance data
  perf_data <- tuning_results$performances

  # Create appropriate plot based on parameters tuned
  if ("gamma" %in% names(perf_data) &&
        "cost" %in% names(perf_data)) {
    # Plot gamma vs cost
    p <- ggplot2::ggplot(
      perf_data,
      ggplot2::aes(
        x = gamma, y = cost, fill = error
      )
    ) +
      ggplot2::geom_tile() +
      ggplot2::scale_x_log10() +
      ggplot2::scale_y_log10() +
      ggplot2::scale_fill_gradient2(
        low = "blue", high = "red",
        mid = "white",
        midpoint = mean(perf_data$error)
      ) +
      ggplot2::labs(
        title = "SVM Parameter Tuning",
        subtitle = paste0(
          "Best parameters: gamma = ",
          tuning_results$best.parameters$gamma,
          ", cost = ",
          tuning_results$best.parameters$cost
        ),
        x = "Gamma (log scale)",
        y = "Cost (log scale)",
        fill = "Error"
      ) +
      ggplot2::theme_minimal()
  } else if ("cost" %in% names(perf_data)) {
    # Plot cost only
    p <- ggplot2::ggplot(
      perf_data,
      ggplot2::aes(x = cost, y = error)
    ) +
      ggplot2::geom_line() +
      ggplot2::geom_point() +
      ggplot2::scale_x_log10() +
      ggplot2::labs(
        title = "SVM Parameter Tuning",
        subtitle = paste0(
          "Best parameter: cost = ",
          tuning_results$best.parameters$cost
        ),
        x = "Cost (log scale)",
        y = "Error"
      ) +
      ggplot2::theme_minimal()
  } else {
    # Generic plot of all parameters
    p <- ggplot2::ggplot(
      perf_data,
      ggplot2::aes(
        x = seq_len(nrow(perf_data)), y = error
      )
    ) +
      ggplot2::geom_line() +
      ggplot2::geom_point() +
      ggplot2::labs(
        title = "SVM Parameter Tuning",
        subtitle = paste0(
          "Best parameters: ",
          paste(
            names(tuning_results$best.parameters),
            tuning_results$best.parameters,
            sep = " = ", collapse = ", "
          )
        ),
        x = "Parameter Combination",
        y = "Error"
      ) +
      ggplot2::theme_minimal()
  }

  p
}

#' @title Tree-based Methods for tidylearn
#' @name tidylearn-trees
#' @description Decision trees, random forests, and boosting functionality
#' @importFrom rpart rpart rpart.control
#' @importFrom stats predict
#' @importFrom randomForest randomForest importance
#' @importFrom gbm gbm predict.gbm
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate arrange desc
NULL

#' Fit a decision tree model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param cp Complexity parameter (default: 0.01)
#' @param minsplit Minimum number of observations in a node for a split
#' @param maxdepth Maximum depth of the tree
#' @param ... Additional arguments to pass to rpart()
#' @return A fitted decision tree model
#' @keywords internal
tl_fit_tree <- function(data, formula, is_classification = FALSE,
                        cp = 0.01, minsplit = 20, maxdepth = 30, ...) {
  # Check if rpart is installed
  tl_check_packages("rpart")

  # Determine method based on problem type
  method <- if (is_classification) "class" else "anova"

  # Fit the tree model
  tree_model <- rpart::rpart(
    formula = formula,
    data = data,
    method = method,
    control = rpart::rpart.control(
      cp = cp,
      minsplit = minsplit,
      maxdepth = maxdepth,
      ...
    )
  )

  tree_model
}

#' Predict using a decision tree model
#'
#' @param model A tidylearn tree model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "prob" or "class" (for classification)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_tree <- function(model, new_data, type = "response", ...) {
  # Get the tree model
  fit <- model$fit
  is_classification <- model$spec$is_classification

  if (is_classification) {
    if (type == "prob") {
      # Get class probabilities
      probs <- predict(fit, newdata = new_data, type = "prob")

      # Convert to tibble with appropriate column names
      class_levels <- colnames(probs)
      prob_df <- as.data.frame(probs)
      names(prob_df) <- class_levels

      tibble::as_tibble(prob_df)
    } else if (type == "class") {
      # Get predicted classes
      preds <- predict(fit, newdata = new_data, type = "class")
      preds
    } else if (type == "response") {
      # Get predicted classes (same as "class" for classification)
      preds <- predict(fit, newdata = new_data, type = "class")
      preds
    } else {
      stop(
        "Invalid prediction type for classification trees. ",
        "Use 'prob', 'class', or 'response'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    preds <- predict(fit, newdata = new_data)
    preds
  }
}

#' Fit a random forest model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param ntree Number of trees to grow (default: 500)
#' @param mtry Number of variables randomly sampled at each split
#' @param importance Whether to compute variable importance (default: TRUE)
#' @param ... Additional arguments to pass to randomForest()
#' @return A fitted random forest model
#' @keywords internal
tl_fit_forest <- function(data, formula, is_classification = FALSE,
                          ntree = 500, mtry = NULL, importance = TRUE, ...) {
  # Check if randomForest is installed
  tl_check_packages("randomForest")

  # Set default mtry if not provided
  if (is.null(mtry)) {
    if (is_classification) {
      # For classification, default is sqrt(p)
      mtry <- floor(sqrt(ncol(data) - 1))
    } else {
      # For regression, default is p/3
      mtry <- max(floor((ncol(data) - 1) / 3), 1)
    }
  }

  # Fit the random forest model
  forest_model <- randomForest::randomForest(
    formula = formula,
    data = data,
    ntree = ntree,
    mtry = mtry,
    importance = importance,
    ...
  )

  forest_model
}

#' Predict using a random forest model
#'
#' @param model A tidylearn forest model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response"
#'   (default), "prob" (for classification)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_forest <- function(model, new_data, type = "response", ...) {
  # Get the random forest model
  fit <- model$fit
  is_classification <- model$spec$is_classification

  if (is_classification) {
    if (type == "prob") {
      # Get class probabilities
      probs <- predict(fit, newdata = new_data, type = "prob")

      # Convert to tibble with appropriate column names
      class_levels <- colnames(probs)
      prob_df <- as.data.frame(probs)
      names(prob_df) <- class_levels

      tibble::as_tibble(prob_df)
    } else if (type == "class" || type == "response") {
      # Get predicted classes
      preds <- predict(fit, newdata = new_data, type = "response")
      preds
    } else {
      stop(
        "Invalid prediction type for random forests. ",
        "Use 'prob', 'class', or 'response'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    preds <- predict(fit, newdata = new_data)
    preds
  }
}

#' Fit a gradient boosting model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param n.trees Number of trees (default: 100)
#' @param interaction.depth Depth of interactions
#'   (default: 3)
#' @param shrinkage Learning rate (default: 0.1)
#' @param n.minobsinnode Minimum number of observations
#'   in terminal nodes (default: 10)
#' @param cv.folds Number of cross-validation folds
#'   (default: 0, no CV)
#' @param ... Additional arguments to pass to gbm()
#' @return A fitted gradient boosting model
#' @keywords internal
tl_fit_boost <- function(
    data, formula,
    is_classification = FALSE,
    n.trees = 100,
    interaction.depth = 3,
    shrinkage = 0.1,
    n.minobsinnode = 10,
    cv.folds = 0, ...) {
  # Check if gbm is installed
  tl_check_packages("gbm")

  # Determine distribution based on problem type
  if (is_classification) {
    # Get response variable
    response_var <- all.vars(formula)[1]
    y <- data[[response_var]]

    # Check if binary or multiclass
    if (is.factor(y) && length(levels(y)) == 2) {
      # Binary classification
      distribution <- "bernoulli"
    } else {
      # Multiclass classification
      distribution <- "multinomial"
    }
  } else {
    # Regression
    distribution <- "gaussian"
  }

  # Fit the boosting model
  boost_model <- gbm::gbm(
    formula = formula,
    data = data,
    distribution = distribution,
    n.trees = n.trees,
    interaction.depth = interaction.depth,
    shrinkage = shrinkage,
    n.minobsinnode = n.minobsinnode,
    cv.folds = cv.folds,
    verbose = FALSE,
    ...
  )

  boost_model
}

#' Predict using a gradient boosting model
#'
#' @param model A tidylearn boost model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response"
#'   (default), "prob" (for classification)
#' @param n.trees Number of trees to use for prediction
#'   (if NULL, uses optimal number)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_boost <- function(
    model, new_data,
    type = "response",
    n.trees = NULL, ...) {
  # Get the boosting model
  fit <- model$fit
  is_classification <- model$spec$is_classification

  # Determine the number of trees to use
  if (is.null(n.trees)) {
    if (fit$cv.folds > 0) {
      # Use the optimal number of trees from CV
      n.trees <- gbm::gbm.perf(
        fit, method = "cv", plot.it = FALSE
      )
    } else {
      # Use all trees
      n.trees <- fit$n.trees
    }
  }

  if (is_classification) {
    # Check distribution
    if (fit$distribution$name == "bernoulli") {
      # Binary classification
      if (type == "prob") {
        # Get probabilities on the scale of the response
        probs <- gbm::predict.gbm(
          fit, newdata = new_data,
          n.trees = n.trees,
          type = "response", ...
        )

        # Get class levels from training data
        response_var <- all.vars(model$spec$formula)[1]
        class_levels <- levels(
          factor(model$data[[response_var]])
        )

        # Create a data frame with probabilities
        prob_df <- tibble::tibble(
          !!class_levels[1] := 1 - probs,
          !!class_levels[2] := probs
        )

        prob_df
      } else if (type == "class" || type == "response") {
        # Get probabilities
        probs <- gbm::predict.gbm(
          fit, newdata = new_data,
          n.trees = n.trees,
          type = "response", ...
        )

        # Get class levels from training data
        response_var <- all.vars(model$spec$formula)[1]
        class_levels <- levels(
          factor(model$data[[response_var]])
        )

        # Convert to classes
        pred_classes <- ifelse(
          probs > 0.5,
          class_levels[2],
          class_levels[1]
        )
        pred_classes <- factor(
          pred_classes, levels = class_levels
        )

        pred_classes
      } else {
        stop(
          "Invalid prediction type for boosting. ",
          "Use 'prob', 'class', or 'response'.",
          call. = FALSE
        )
      }
    } else if (fit$distribution$name == "multinomial") {
      # Multiclass classification
      if (type == "prob") {
        # Get class probabilities
        probs <- gbm::predict.gbm(
          fit, newdata = new_data,
          n.trees = n.trees,
          type = "response", ...
        )

        # Reshape to data frame
        if (is.matrix(probs)) {
          # Get class levels from column names
          class_levels <- colnames(probs)
          prob_df <- as.data.frame(probs)
          names(prob_df) <- class_levels
        } else {
          # For single prediction
          response_var <- all.vars(
            model$spec$formula
          )[1]
          class_levels <- levels(
            factor(model$data[[response_var]])
          )
          prob_df <- as.data.frame(
            matrix(probs, nrow = 1)
          )
          names(prob_df) <- class_levels
        }

        tibble::as_tibble(prob_df)
      } else if (type == "class" || type == "response") {
        # Get class probabilities
        probs <- gbm::predict.gbm(
          fit, newdata = new_data,
          n.trees = n.trees,
          type = "response", ...
        )

        # Convert to classes
        if (is.matrix(probs)) {
          # Find class with highest probability
          class_idx <- apply(probs, 1, which.max)
          class_levels <- colnames(probs)
          pred_classes <- factor(
            class_levels[class_idx],
            levels = class_levels
          )
        } else {
          # For single prediction
          response_var <- all.vars(
            model$spec$formula
          )[1]
          class_levels <- levels(
            factor(model$data[[response_var]])
          )
          class_idx <- which.max(probs)
          pred_classes <- factor(
            class_levels[class_idx],
            levels = class_levels
          )
        }

        pred_classes
      } else {
        stop(
          "Invalid prediction type for boosting. ",
          "Use 'prob', 'class', or 'response'.",
          call. = FALSE
        )
      }
    }
  } else {
    # Regression predictions
    preds <- gbm::predict.gbm(
      fit, newdata = new_data,
      n.trees = n.trees,
      type = "response", ...
    )
    preds
  }
}

#' Plot variable importance for tree-based models
#'
#' @param model A tidylearn tree-based model object
#' @param top_n Number of top features to display (default: 20)
#' @param ... Additional arguments
#' @return A ggplot object
#' @importFrom ggplot2 ggplot aes geom_col coord_flip labs theme_minimal
#' @keywords internal
tl_plot_importance <- function(model, top_n = 20, ...) {
  # Get the model
  fit <- model$fit
  method <- model$spec$method

  if (method == "tree") {
    # Decision tree importance
    # Get variable importance from rpart
    imp <- fit$variable.importance

    # Create a data frame for plotting
    importance_df <- tibble::tibble(
      feature = names(imp),
      importance = as.vector(imp)
    )
  } else if (method == "forest") {
    # Random forest importance
    # Get variable importance from randomForest
    imp <- randomForest::importance(fit)

    # Create a data frame for plotting
    if (model$spec$is_classification) {
      # For classification, use mean decrease in accuracy
      importance_df <- tibble::tibble(
        feature = rownames(imp),
        importance = imp[, "MeanDecreaseAccuracy"]
      )
    } else {
      # For regression, use % increase in MSE
      importance_df <- tibble::tibble(
        feature = rownames(imp),
        importance = imp[, "%IncMSE"]
      )
    }
  } else if (method == "boost") {
    # Gradient boosting importance
    # Get relative influence from gbm
    imp <- summary(fit, plotit = FALSE)

    # Create a data frame for plotting
    importance_df <- tibble::tibble(
      feature = imp$var,
      importance = imp$rel.inf
    )
  } else {
    stop(
      "Variable importance plot not implemented for method: ",
      method,
      call. = FALSE
    )
  }

  # Filter and sort
  importance_df <- importance_df %>%
    dplyr::arrange(dplyr::desc(.data$importance)) %>%
    dplyr::slice_head(n = top_n)

  # Create the plot
  p <- ggplot2::ggplot(
    importance_df,
    ggplot2::aes(
      x = stats::reorder(feature, importance),
      y = importance
    )
  ) +
    ggplot2::geom_col(fill = "steelblue") +
    ggplot2::coord_flip() +
    ggplot2::labs(
      title = "Variable Importance",
      x = NULL,
      y = "Importance"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Plot a decision tree
#'
#' @param model A tidylearn tree model object
#' @param ... Additional arguments to pass to rpart.plot()
#' @return The return value of \code{\link[rpart.plot]{rpart.plot}}, called
#'   for its side effect of drawing the tree.
#' @examples
#' \donttest{
#' model <- tl_model(iris, Species ~ ., method = "tree")
#' tl_plot_tree(model)
#' }
#' @export
tl_plot_tree <- function(model, ...) {
  # Check if rpart.plot is installed
  tl_check_packages("rpart.plot")

  if (model$spec$method != "tree") {
    stop("Tree plot is only available for decision tree models", call. = FALSE)
  }

  # Plot the tree
  rpart.plot::rpart.plot(model$fit, ...)
}

#' Plot partial dependence for tree-based models
#'
#' @param model A tidylearn tree-based model object
#' @param var Variable name to plot
#' @param n.pts Number of points for continuous
#'   variables (default: 20)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_line geom_point
#' @importFrom ggplot2 labs theme_minimal
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ ., method = "forest")
#' tl_plot_partial_dependence(model, var = "wt")
#' }
#' @export
tl_plot_partial_dependence <- function(model, var, n.pts = 20, ...) {
  if (!model$spec$method %in% c("tree", "forest", "boost")) {
    stop(
      "Partial dependence plots are currently ",
      "only implemented for tree-based models",
      call. = FALSE
    )
  }

  # Get the data
  data <- model$data

  # Check if variable exists
  if (!var %in% names(data)) {
    stop(
      "Variable '", var,
      "' not found in the model data",
      call. = FALSE
    )
  }

  # Get variable values
  var_values <- data[[var]]

  # Create grid of values for the variable
  if (is.factor(var_values) || is.character(var_values)) {
    # For categorical variables, use unique values
    grid_values <- unique(var_values)
  } else {
    # For continuous variables, create a sequence
    grid_values <- seq(
      min(var_values, na.rm = TRUE),
      max(var_values, na.rm = TRUE),
      length.out = n.pts
    )
  }

  # Create prediction grid
  pred_data <- purrr::map_dfr(grid_values, function(val) {
    # Make a copy of the original data
    new_data <- data

    # Set the variable to the current value
    new_data[[var]] <- val

    # Make predictions
    if (model$spec$is_classification) {
      # For classification, get probabilities
      probs <- predict(model, new_data, type = "prob")

      # Calculate mean probability for each class
      class_means <- purrr::map_dbl(probs, mean)

      # Create result row
      result <- tibble::tibble(
        var_value = val,
        y = class_means[2]  # For binary classification, use the positive class
      )

      # Add class column for multiclass
      if (length(class_means) > 2) {
        result$class <- names(class_means)[which.max(class_means)]
      }
    } else {
      # For regression, get mean prediction
      preds <- predict(model, new_data, type = "response")
      mean_pred <- mean(preds)

      # Create result row
      result <- tibble::tibble(
        var_value = val,
        y = mean_pred
      )
    }

    result
  })

  # Determine y-axis label
  y_lab <- if (model$spec$is_classification) {
    "Mean Probability"
  } else {
    "Mean Prediction"
  }

  # Create the plot
  if (is.factor(var_values) || is.character(var_values)) {
    # Bar plot for categorical variables
    p <- ggplot2::ggplot(
      pred_data,
      ggplot2::aes(x = var_value, y = y)
    ) +
      ggplot2::geom_col(fill = "steelblue") +
      ggplot2::labs(
        title = paste0(
          "Partial Dependence Plot for ", var
        ),
        x = var,
        y = y_lab
      ) +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        axis.text.x = ggplot2::element_text(
          angle = 45, hjust = 1
        )
      )
  } else {
    # Line plot for continuous variables
    p <- ggplot2::ggplot(
      pred_data,
      ggplot2::aes(x = var_value, y = y)
    ) +
      ggplot2::geom_line(color = "steelblue") +
      ggplot2::geom_point(color = "steelblue") +
      ggplot2::labs(
        title = paste0(
          "Partial Dependence Plot for ", var
        ),
        x = var,
        y = y_lab
      ) +
      ggplot2::theme_minimal()
  }

  p
}

#' Tidy DBSCAN Clustering
#'
#' Performs density-based clustering with tidy output
#'
#' @param data A data frame, tibble, or distance matrix
#' @param eps Neighborhood radius (epsilon)
#' @param minPts Minimum number of points to form a dense region (default: 5)
#' @param cols Columns to include (tidy select).
#'   If NULL, uses all numeric columns.
#' @param distance Distance metric if data is not a
#'   dist object (default: "euclidean")
#'
#' @return A list of class "tidy_dbscan" containing:
#' \itemize{
#'   \item clusters: tibble with observation IDs and
#'     cluster assignments (0 = noise)
#'   \item core_points: logical vector indicating core points
#'   \item n_clusters: number of clusters (excluding noise)
#'   \item n_noise: number of noise points
#'   \item model: original dbscan object
#' }
#'
#' @examples
#' # Basic DBSCAN
#' db_result <- tidy_dbscan(iris, eps = 0.5, minPts = 5)
#'
#' # With suggested eps from k-NN distance plot
#' eps_suggestion <- suggest_eps(iris, minPts = 5)
#' db_result <- tidy_dbscan(iris, eps = eps_suggestion$eps, minPts = 5)
#'
#' @export
tidy_dbscan <- function(data, eps, minPts = 5,
                        cols = NULL,
                        distance = "euclidean") {

  # Handle distance matrix
  if (inherits(data, "dist")) {
    data_matrix <- as.matrix(data)
  } else {
    # Select columns
    if (!is.null(cols)) {
      cols_enquo <- rlang::enquo(cols)
      data_selected <- data %>% dplyr::select(!!cols_enquo)
    } else {
      data_selected <- data %>% dplyr::select(where(is.numeric))
    }

    data_matrix <- as.matrix(data_selected)
  }

  # Perform DBSCAN
  db_model <- dbscan::dbscan(data_matrix, eps = eps, minPts = minPts)

  # Count clusters and noise
  n_clusters <- max(db_model$cluster)
  n_noise <- sum(db_model$cluster == 0)

  # Create clusters tibble
  if (!is.null(rownames(data))) {
    obs_ids <- rownames(data)
  } else {
    obs_ids <- paste0("obs_", seq_len(nrow(data_matrix)))
  }

  clusters_tbl <- tibble::tibble(
    .obs_id = obs_ids,
    cluster = as.integer(db_model$cluster),
    is_noise = cluster == 0,
    is_core = seq_along(cluster) %in% attr(db_model, "core")
  )

  # Create summary statistics
  cluster_summary <- clusters_tbl %>%
    dplyr::filter(!is_noise) %>%
    dplyr::group_by(cluster) %>%
    dplyr::summarise(
      size = dplyr::n(),
      n_core = sum(is_core),
      .groups = "drop"
    )

  # Return tidy object
  result <- list(
    clusters = clusters_tbl,
    summary = cluster_summary,
    n_clusters = n_clusters,
    n_noise = n_noise,
    eps = eps,
    minPts = minPts,
    model = db_model
  )

  class(result) <- c("tidy_dbscan", "list")
  result
}


#' Compute k-NN Distances
#'
#' Calculate distances to k-th nearest neighbor for each point
#'
#' @param data A data frame or matrix
#' @param k Number of nearest neighbors (default: 4)
#' @param cols Columns to include (tidy select).
#'   If NULL, uses all numeric columns.
#'
#' @return A tibble with columns \code{.obs_id} (observation identifier),
#'   \code{knn_dist} (distance to k-th nearest neighbor), and \code{rank}
#'   (rank of the k-NN distance).
#'
#' @examples
#' \donttest{
#' knn <- tidy_knn_dist(iris[, 1:4], k = 5)
#' }
#'
#' @export
tidy_knn_dist <- function(data, k = 4, cols = NULL) {

  # Select columns
  if (!is.null(cols)) {
    cols_enquo <- rlang::enquo(cols)
    data_selected <- data %>% dplyr::select(!!cols_enquo)
  } else {
    data_selected <- data %>% dplyr::select(where(is.numeric))
  }

  data_matrix <- as.matrix(data_selected)

  # Compute k-NN distances
  knn_distances <- dbscan::kNNdist(data_matrix, k = k)

  # Create tibble
  tibble::tibble(
    .obs_id = rownames(data) %||% paste0("obs_", seq_len(nrow(data))),
    knn_dist = as.numeric(knn_distances),
    rank = rank(knn_distances)
  )
}


#' Suggest eps Parameter for DBSCAN
#'
#' Use k-NN distance plot to suggest eps value
#'
#' @param data A data frame or matrix
#' @param minPts Minimum points parameter (used as k for k-NN)
#' @param method Method to suggest eps: "knee" (default), "percentile"
#' @param percentile If method="percentile", which
#'   percentile to use (default: 0.95)
#'
#' @return A list containing:
#' \itemize{
#'   \item eps: suggested epsilon value
#'   \item knn_distances: full tibble of k-NN distances
#'   \item method: method used
#' }
#'
#' @examples
#' eps_info <- suggest_eps(iris, minPts = 5)
#' eps_info$eps
#'
#' @export
suggest_eps <- function(data, minPts = 5,
                        method = "percentile",
                        percentile = 0.95) {

  # Compute k-NN distances
  knn_data <- tidy_knn_dist(data, k = minPts)

  # Suggest eps based on method
  if (method == "percentile") {
    eps_suggested <- stats::quantile(knn_data$knn_dist, percentile)

  } else if (method == "knee") {
    # Find knee/elbow in sorted k-NN distances
    sorted_dist <- sort(knn_data$knn_dist)

    # Calculate differences
    diffs <- diff(sorted_dist)

    # Find maximum jump
    max_jump_idx <- which.max(diffs)
    eps_suggested <- sorted_dist[max_jump_idx]

  } else {
    stop("method must be 'percentile' or 'knee'")
  }

  list(
    eps = as.numeric(eps_suggested),
    knn_distances = knn_data,
    method = method
  )
}


#' Plot k-NN Distance Plot
#'
#' Visualize k-NN distances to help choose eps
#'
#' @param data A data frame or tidy_knn_dist result
#' @param k If data is a data frame, k for k-NN (default: 4)
#' @param add_suggestion Add suggested eps line? (default: TRUE)
#' @param percentile Percentile for suggestion (default: 0.95)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' plot_knn_dist(iris[, 1:4], k = 5)
#' }
#'
#' @export
plot_knn_dist <- function(data, k = 4,
                          add_suggestion = TRUE,
                          percentile = 0.95) {

  # Get k-NN distances if needed
  if (inherits(data, "tbl_df") && "knn_dist" %in% names(data)) {
    knn_data <- data
  } else {
    knn_data <- tidy_knn_dist(data, k = k)
  }

  # Sort by distance
  knn_data <- knn_data %>% dplyr::arrange(knn_dist)

  # Create plot
  p <- ggplot2::ggplot(
    knn_data,
    ggplot2::aes(x = seq_along(knn_dist), y = knn_dist)
  ) +
    ggplot2::geom_line(color = "steelblue", linewidth = 1) +
    ggplot2::labs(
      title = paste0("k-NN Distance Plot (k = ", k, ")"),
      subtitle = "Look for 'elbow' or 'knee' to determine eps",
      x = "Points (sorted by distance)",
      y = paste0(k, "-NN Distance")
    ) +
    ggplot2::theme_minimal()

  # Add suggestion line
  if (add_suggestion) {
    eps_line <- stats::quantile(knn_data$knn_dist, percentile)
    p <- p +
      ggplot2::geom_hline(
        yintercept = eps_line,
        linetype = "dashed", color = "red"
      ) +
      ggplot2::annotate(
        "text",
        x = nrow(knn_data) * 0.7,
        y = eps_line * 1.1,
        label = sprintf(
          "Suggested eps = %.3f\n(%d%% percentile)",
          eps_line, percentile * 100
        ),
        color = "red"
      )
  }

  p
}


#' Augment Data with DBSCAN Cluster Assignments
#'
#' @param dbscan_obj A tidy_dbscan object
#' @param data Original data frame
#'
#' @return A tibble containing the original \code{data} with additional columns
#'   \code{cluster} (factor), \code{is_noise} (logical), and \code{is_core}
#'   (logical).
#'
#' @examples
#' \donttest{
#' db <- tidy_dbscan(iris[, 1:4], eps = 0.5, minPts = 5)
#' augmented <- augment_dbscan(db, iris)
#' }
#'
#' @export
augment_dbscan <- function(dbscan_obj, data) {

  if (!inherits(dbscan_obj, "tidy_dbscan")) {
    stop("dbscan_obj must be a tidy_dbscan object")
  }

  data %>%
    dplyr::bind_cols(
      tibble::tibble(
        cluster = as.factor(dbscan_obj$model$cluster),
        is_noise = dbscan_obj$clusters$is_noise,
        is_core = dbscan_obj$clusters$is_core
      )
    )
}


#' Explore DBSCAN Parameters
#'
#' Test multiple eps and minPts combinations
#'
#' @param data A data frame or matrix
#' @param eps_values Vector of eps values to test
#' @param minPts_values Vector of minPts values to test
#'
#' @return A tibble with columns \code{eps}, \code{minPts}, \code{n_clusters},
#'   \code{n_noise}, and \code{prop_noise} for each parameter combination.
#'
#' @examples
#' \donttest{
#' params <- explore_dbscan_params(iris[, 1:4],
#'   eps_values = c(0.3, 0.5, 0.8), minPts_values = c(3, 5))
#' }
#'
#' @export
explore_dbscan_params <- function(data, eps_values, minPts_values) {  # nolint

  data_numeric <- data %>% dplyr::select(where(is.numeric))

  # Create parameter grid
  param_grid <- expand.grid(
    eps = eps_values,
    minPts = minPts_values,
    stringsAsFactors = FALSE
  )

  # Test each combination
  results <- purrr::map2_dfr(
    param_grid$eps,
    param_grid$minPts,
    function(e, m) {
      db <- tidy_dbscan(
        data_numeric, eps = e, minPts = m
      )
      tibble::tibble(
        eps = e,
        minPts = m,
        n_clusters = db$n_clusters,
        n_noise = db$n_noise,
        prop_noise = db$n_noise / nrow(data_numeric)
      )
    }
  )

  results
}


#' Print Method for tidy_dbscan
#'
#' @param x A tidy_dbscan object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' db <- tidy_dbscan(iris[, 1:4], eps = 0.5, minPts = 5)
#' print(db)
#' }
#'
#' @export
print.tidy_dbscan <- function(x, ...) {
  cat("Tidy DBSCAN Clustering\n")
  cat("======================\n\n")
  cat("Parameters:\n")
  cat("  eps (neighborhood radius):", x$eps, "\n")
  cat("  minPts (minimum points): ", x$minPts, "\n\n")

  cat("Results:\n")
  cat("  Number of clusters:", x$n_clusters, "\n")
  cat("  Number of noise points:", x$n_noise, "\n")
  noise_pct <- (x$n_noise / nrow(x$clusters)) * 100
  cat("  Proportion noise:",
      sprintf("%.1f%%", noise_pct), "\n\n")

  if (nrow(x$summary) > 0) {
    cat("Cluster Summary:\n")
    print(x$summary)
  }

  cat("\nUse augment_dbscan() to add cluster assignments to your data\n")

  invisible(x)
}


#' Fit DBSCAN for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_dbscan <- function(data, formula = NULL, eps = 0.5, minPts = 5, ...) {
  tl_check_packages("dbscan")

  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_db <- data[, vars, drop = FALSE]
  } else {
    data_for_db <- data %>% dplyr::select(where(is.numeric))
  }

  # Fit DBSCAN using tidy_dbscan
  db_result <- tidy_dbscan(data_for_db, eps = eps, minPts = minPts, ...)

  # Return in expected format
  list(
    clusters = db_result$clusters,
    summary = db_result$summary,
    n_clusters = db_result$n_clusters,
    n_noise = db_result$n_noise,
    model = db_result$model
  )
}

#' @title Model Selection Functions for tidylearn
#' @name tidylearn-model-selection
#' @description Functions for stepwise model selection,
#'   cross-validation, and hyperparameter tuning
#' @importFrom stats AIC BIC step
#' @importFrom dplyr filter select mutate
NULL

#' Perform stepwise selection on a linear model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the initial model
#' @param direction Direction of stepwise selection:
#'   "forward", "backward", or "both"
#' @param criterion Criterion for selection: "AIC" or "BIC"
#' @param trace Logical; whether to print progress
#' @param steps Maximum number of steps to take
#' @param ... Additional arguments to pass to step()
#' @return A \code{tidylearn_model} object of class
#'   \code{tidylearn_linear} wrapping the selected \code{\link[stats]{lm}}
#'   model. Access the underlying model via \code{$fit} and the selected
#'   formula via \code{$spec$formula}.
#' @examples
#' \donttest{
#' model <- tl_step_selection(mtcars, mpg ~ ., direction = "backward")
#' summary(model)
#' }
#' @export
tl_step_selection <- function(data, formula, direction = "backward",
                              criterion = "AIC",
                              trace = FALSE,
                              steps = 1000, ...) {
  # Input validation
  direction <- match.arg(direction, c("forward", "backward", "both"))
  criterion <- match.arg(criterion, c("AIC", "BIC"))

  # Create full model
  full_model <- lm(formula, data = data)

  # Create null model (intercept only) for forward selection
  null_formula <- as.formula(paste(all.vars(formula)[1], "~ 1"))
  null_model <- lm(null_formula, data = data)

  # Set penalty parameter k based on criterion
  k <- if (criterion == "AIC") 2 else log(nrow(data))

  # Determine start and scope based on direction
  if (direction == "forward") {
    start_model <- null_model
    scope <- list(lower = null_formula, upper = formula)
  } else if (direction == "backward") {
    start_model <- full_model
    scope <- formula
  } else {  # "both"
    start_model <- null_model
    scope <- list(lower = null_formula, upper = formula)
  }

  # Run stepwise selection
  selected_model <- stats::step(
    start_model,
    scope = scope,
    direction = direction,
    k = k,
    trace = trace,
    steps = steps,
    ...
  )

  # Create tidylearn model wrapper
  is_classification <- is.factor(data[[all.vars(formula)[1]]]) ||
    is.character(data[[all.vars(formula)[1]]])

  # Return selected model as a tidylearn model
  model <- structure(
    list(
      spec = list(
        paradigm = "supervised",
        formula = formula(selected_model),
        method = "linear",
        is_classification = is_classification,
        response_var = all.vars(formula)[1],
        selection = list(
          criterion = criterion,
          direction = direction
        )
      ),
      fit = selected_model,
      data = data
    ),
    class = c("tidylearn_linear", "tidylearn_supervised", "tidylearn_model")
  )

  model
}

#' Compare models using cross-validation
#'
#' @param data A data frame containing the training data
#' @param models A list of tidylearn model objects
#' @param folds Number of cross-validation folds
#' @param metrics Character vector of metrics to compute
#' @param ... Additional arguments
#' @return A list with two elements:
#'   \describe{
#'     \item{\code{$fold_metrics}}{A data frame with columns
#'       \code{metric}, \code{value}, \code{fold}, and \code{model}
#'       containing per-fold results for every model.}
#'     \item{\code{$summary}}{A data frame with columns \code{model},
#'       \code{metric}, \code{mean_value}, \code{sd_value},
#'       \code{min_value}, and \code{max_value} summarizing
#'       cross-validation performance.}
#'   }
#' @examples
#' \donttest{
#' m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
#' m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
#' cv$summary
#' }
#' @export
tl_compare_cv <- function(data, models, folds = 5, metrics = NULL, ...) {
  # Check if all models are tidylearn models
  is_tl_model <- sapply(models, inherits, "tidylearn_model")
  if (!all(is_tl_model)) {
    stop("All models must be tidylearn model objects", call. = FALSE)
  }

  # Get model names if not provided
  model_names <- names(models)
  if (is.null(model_names)) {
    model_names <- paste0("Model_", seq_along(models))
  }

  # Check if all models are of the same type (classification or regression)
  model_types <- sapply(models, function(model) model$spec$is_classification)
  if (length(unique(model_types)) > 1) {
    stop(
      "All models must be of the same type ",
      "(classification or regression)",
      call. = FALSE
    )
  }

  is_classification <- model_types[1]

  # Default metrics based on problem type
  if (is.null(metrics)) {
    if (is_classification) {
      metrics <- c("accuracy", "precision", "recall", "f1", "auc")
    } else {
      metrics <- c("rmse", "mae", "rsq", "mape")
    }
  }

  # Create cross-validation splits
  cv_splits <- rsample::vfold_cv(data, v = folds)

  # For each model, perform cross-validation
  cv_results <- lapply(seq_along(models), function(i) {
    model <- models[[i]]
    model_name <- model_names[i]

    # For each fold, train model and evaluate
    fold_results <- purrr::map_dfr(seq_len(folds), function(j) {
      # Get training and testing data for this fold
      train_data <- rsample::analysis(cv_splits$splits[[j]])
      test_data <- rsample::assessment(cv_splits$splits[[j]])

      # Extract formula and method from the model
      formula <- model$spec$formula
      method <- model$spec$method

      # Train model on this fold
      fold_model <- tl_model(
        train_data, formula = formula,
        method = method, ...
      )

      # Evaluate model on test data
      fold_metrics <- tl_evaluate(fold_model, test_data, metrics = metrics)

      # Add fold number and model name
      fold_metrics$fold <- j
      fold_metrics$model <- model_name

      fold_metrics
    })

    fold_results
  })

  # Combine results from all models
  all_cv_results <- do.call(rbind, cv_results)

  # Calculate summary statistics for each model
  summary_results <- all_cv_results %>%
    dplyr::group_by(.data$model, .data$metric) %>%
    dplyr::summarize(
      mean_value = mean(.data$value, na.rm = TRUE),
      sd_value = sd(.data$value, na.rm = TRUE),
      min_value = min(.data$value, na.rm = TRUE),
      max_value = max(.data$value, na.rm = TRUE),
      .groups = "drop"
    )

  # Return both detailed and summary results
  list(
    fold_metrics = all_cv_results,
    summary = summary_results
  )
}

#' Plot comparison of cross-validation results
#'
#' @param cv_results Results from tl_compare_cv function
#' @param metrics Character vector of metrics to plot
#'   (if NULL, plots all metrics)
#' @return A \code{\link[ggplot2]{ggplot}} object showing boxplots of
#'   cross-validation metric distributions for each model.
#' @importFrom ggplot2 ggplot aes geom_boxplot facet_wrap labs theme_minimal
#' @examples
#' \donttest{
#' m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
#' m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
#' tl_plot_cv_comparison(cv)
#' }
#' @export
tl_plot_cv_comparison <- function(cv_results, metrics = NULL) {
  # Extract fold metrics
  fold_metrics <- cv_results$fold_metrics

  # Filter metrics if specified
  if (!is.null(metrics)) {
    fold_metrics <- fold_metrics %>%
      dplyr::filter(.data$metric %in% metrics)
  }

  # Create the plot
  p <- ggplot2::ggplot(
    fold_metrics,
    ggplot2::aes(x = .data$model, y = .data$value, fill = .data$model)
  ) +
    ggplot2::geom_boxplot() +
    ggplot2::facet_wrap(~ .data$metric, scales = "free_y") +
    ggplot2::labs(
      title = "Cross-Validation Results Comparison",
      x = "Model",
      y = "Metric Value",
      fill = "Model"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))

  p
}

#' Perform statistical comparison of models using cross-validation
#'
#' @param cv_results Results from tl_compare_cv function
#' @param baseline_model Name of the model to use as baseline for comparison
#' @param test Type of statistical test: "t.test" or "wilcox"
#' @param metric Name of the metric to compare
#' @return A data frame with columns \code{metric}, \code{model},
#'   \code{baseline}, \code{mean_diff}, \code{p_value}, and
#'   \code{p_adj} (Holm-adjusted p-value) containing pairwise
#'   statistical comparisons against the baseline model.
#' @importFrom stats t.test wilcox.test p.adjust
#' @examples
#' \donttest{
#' m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
#' m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
#' tl_test_model_difference(cv, baseline_model = "simple", metric = "rmse")
#' }
#' @export
tl_test_model_difference <- function(
    cv_results,
    baseline_model = NULL,
    test = "t.test",
    metric = NULL) {
  # Input validation
  test <- match.arg(test, c("t.test", "wilcox"))

  # Extract fold metrics
  fold_metrics <- cv_results$fold_metrics

  # Get unique models and metrics
  models <- unique(fold_metrics$model)

  if (is.null(baseline_model)) {
    baseline_model <- models[1]
  } else if (!baseline_model %in% models) {
    stop(
      "Baseline model not found in CV results",
      call. = FALSE
    )
  }

  if (is.null(metric)) {
    metrics <- unique(fold_metrics$metric)
  } else {
    if (!metric %in% unique(fold_metrics$metric)) {
      stop(
        "Metric not found in CV results",
        call. = FALSE
      )
    }
    metrics <- metric
  }

  # Perform statistical tests
  results <- lapply(metrics, function(m) {
    # Filter data for current metric
    metric_data <- fold_metrics %>%
      dplyr::filter(.data$metric == m)

    # Get baseline model data
    baseline_data <- metric_data %>%
      dplyr::filter(.data$model == baseline_model) %>%
      dplyr::pull(.data$value)

    # Compare each model to baseline
    other_models <- setdiff(models, baseline_model)
    model_comparisons <- lapply(
      other_models,
      function(model_name) {
        # Get current model data
        model_data <- metric_data %>%
          dplyr::filter(
            .data$model == model_name
          ) %>%
          dplyr::pull(.data$value)

        # Perform statistical test
        if (test == "t.test") {
          test_result <- stats::t.test(
            model_data, baseline_data,
            paired = TRUE
          )
        } else {
          test_result <- stats::wilcox.test(
            model_data, baseline_data,
            paired = TRUE
          )
        }

        # Calculate mean difference
        mean_diff <-
          mean(model_data, na.rm = TRUE) -
          mean(baseline_data, na.rm = TRUE)

        # Return results
        data.frame(
          metric = m,
          model = model_name,
          baseline = baseline_model,
          mean_diff = mean_diff,
          p_value = test_result$p.value
        )
      }
    )

    # Combine results for all models
    model_results <- do.call(
      rbind, model_comparisons
    )

    # Adjust p-values for multiple comparisons
    if (length(models) > 2) {
      model_results$p_adj <- stats::p.adjust(
        model_results$p_value, method = "holm"
      )
    } else {
      model_results$p_adj <- model_results$p_value
    }

    model_results
  })

  # Combine results for all metrics
  all_results <- do.call(rbind, results)

  all_results
}

#' @title Metrics Functionality for tidylearn
#' @name tidylearn-metrics
#' @description Functions for calculating model evaluation metrics
#' @importFrom yardstick accuracy precision recall f_meas
#'   rmse rsq mae mape roc_auc pr_auc
#' @importFrom ROCR prediction performance
#' @importFrom dplyr tibble mutate
NULL

#' Calculate classification metrics
#'
#' @param actuals Actual values (ground truth)
#' @param predicted Predicted class values
#' @param predicted_probs Predicted probabilities (for metrics like AUC)
#' @param metrics Character vector of metrics to compute
#' @param thresholds Optional vector of thresholds to evaluate
#'   for threshold-dependent metrics
#' @param ... Additional arguments
#' @return A \link[tibble]{tibble} with columns \code{metric} (character)
#'   and \code{value} (numeric) containing the requested classification
#'   metrics. When \code{thresholds} are supplied, additional rows are
#'   appended with threshold-specific metric names.
#' @examples
#' \donttest{
#' model <- tl_model(iris, Species ~ ., method = "forest")
#' preds <- predict(model)
#' tl_calc_classification_metrics(iris$Species, preds$.pred)
#' }
#' @export
tl_calc_classification_metrics <- function(
    actuals, predicted,
    predicted_probs = NULL,
    metrics = c("accuracy", "precision",
                "recall", "f1", "auc"),
    thresholds = NULL, ...) {
  # Ensure actuals is a factor
  if (!is.factor(actuals)) {
    actuals <- as.factor(actuals)
  }

  # Ensure predicted is a factor with the same levels
  if (!is.factor(predicted)) {
    predicted <- factor(predicted, levels = levels(actuals))
  }

  # Create a results data frame
  results <- tibble::tibble(metric = character(), value = numeric())

  # Calculate basic classification metrics
  if ("accuracy" %in% metrics) {
    acc <- yardstick::accuracy_vec(actuals, predicted)
    results <- results %>% dplyr::add_row(metric = "accuracy", value = acc)
  }

  if ("precision" %in% metrics) {
    prec <- yardstick::precision_vec(actuals, predicted)
    results <- results %>% dplyr::add_row(metric = "precision", value = prec)
  }

  if ("recall" %in% metrics || "sensitivity" %in% metrics) {
    rec <- yardstick::recall_vec(actuals, predicted)
    results <- results %>% dplyr::add_row(metric = "recall", value = rec)
    if ("sensitivity" %in% metrics) {
      results <- results %>% dplyr::add_row(metric = "sensitivity", value = rec)
    }
  }

  if ("specificity" %in% metrics) {
    spec <- yardstick::specificity_vec(actuals, predicted)
    results <- results %>% dplyr::add_row(metric = "specificity", value = spec)
  }

  if ("f1" %in% metrics) {
    f1 <- yardstick::f_meas_vec(actuals, predicted, beta = 1)
    results <- results %>% dplyr::add_row(metric = "f1", value = f1)
  }

  # Calculate threshold-dependent metrics if probabilities are provided
  has_threshold_metrics <- !is.null(predicted_probs) &&
    ("auc" %in% metrics || "pr_auc" %in% metrics ||
       !is.null(thresholds))
  if (has_threshold_metrics) {

    # For binary classification
    if (ncol(predicted_probs) == 2) {
      # Use the probability of the positive class
      pos_class <- levels(actuals)[2]
      probs <- predicted_probs[[pos_class]]

      # AUC ROC
      if ("auc" %in% metrics) {
        binary_actuals <- as.integer(actuals == pos_class)
        pred_obj <- ROCR::prediction(probs, binary_actuals)
        auc <- unlist(ROCR::performance(pred_obj, "auc")@y.values)
        results <- results %>% dplyr::add_row(metric = "auc", value = auc)
      }

      # PR AUC
      if ("pr_auc" %in% metrics) {
        binary_actuals <- as.integer(actuals == pos_class)
        pred_obj <- ROCR::prediction(probs, binary_actuals)
        perf <- ROCR::performance(pred_obj, "prec", "rec")
        pr_auc <- tl_calculate_pr_auc(perf)
        results <- results %>% dplyr::add_row(metric = "pr_auc", value = pr_auc)
      }

      # Evaluate metrics at different thresholds
      if (!is.null(thresholds)) {
        threshold_metrics <- tl_evaluate_thresholds(
          actuals = actuals,
          probs = probs,
          thresholds = thresholds,
          pos_class = pos_class
        )
        results <- dplyr::bind_rows(results, threshold_metrics)
      }
    } else {
      # Multiclass AUC (one-vs-rest)
      if ("auc" %in% metrics) {
        # Calculate one-vs-rest AUC for each class
        class_aucs <- purrr::map_dbl(
          names(predicted_probs),
          function(class_name) {
            binary_actuals <- as.integer(
              actuals == class_name
            )
            pred_obj <- ROCR::prediction(
              predicted_probs[[class_name]],
              binary_actuals
            )
            unlist(
              ROCR::performance(pred_obj, "auc")@y.values
            )
          }
        )

        # Average AUC across classes
        macro_auc <- mean(class_aucs)
        results <- results %>% dplyr::add_row(metric = "auc", value = macro_auc)

        # Add individual class AUCs
        for (i in seq_along(names(predicted_probs))) {
          class_name <- names(predicted_probs)[i]
          results <- results %>%
            dplyr::add_row(
              metric = paste0("auc_", class_name),
              value = class_aucs[i]
            )
        }
      }
    }
  }

  results
}

#' Calculate the area under the precision-recall curve
#'
#' @param perf A ROCR performance object
#' @return The area under the PR curve
#' @keywords internal
tl_calculate_pr_auc <- function(perf) {
  precision <- perf@y.values[[1]]
  recall <- perf@x.values[[1]]

  # Remove NA/NaN values
  valid <- !is.na(precision) & !is.na(recall)
  precision <- precision[valid]
  recall <- recall[valid]

  # Sort by recall
  ord <- order(recall)
  recall <- recall[ord]
  precision <- precision[ord]

  # Calculate AUC using trapezoidal rule
  auc <- 0
  for (i in 2:length(recall)) {
    width <- recall[i] - recall[i - 1]
    height <- (precision[i] + precision[i - 1]) / 2
    auc <- auc + width * height
  }

  auc
}

#' Evaluate metrics at different thresholds
#'
#' @param actuals Actual values (ground truth)
#' @param probs Predicted probabilities
#' @param thresholds Vector of thresholds to evaluate
#' @param pos_class The positive class
#' @return A tibble of metrics at different thresholds
#' @keywords internal
tl_evaluate_thresholds <- function(actuals, probs, thresholds, pos_class) {
  # No need to convert actuals to binary here,
  # we need the factor for the metrics

  threshold_results <- purrr::map_dfr(thresholds, function(threshold) {
    # Make predictions at this threshold
    pred_vals <- ifelse(
      probs >= threshold, pos_class,
      levels(actuals)[1]
    )
    pred_class <- factor(
      pred_vals, levels = levels(actuals)
    )

    # Calculate metrics
    acc <- yardstick::accuracy_vec(actuals, pred_class)
    prec <- yardstick::precision_vec(actuals, pred_class)
    rec <- yardstick::recall_vec(actuals, pred_class)
    f1 <- yardstick::f_meas_vec(actuals, pred_class, beta = 1)

    # Calculate F2 and F0.5 scores
    f2 <- yardstick::f_meas_vec(actuals, pred_class, beta = 2)
    f0_5 <- yardstick::f_meas_vec(actuals, pred_class, beta = 0.5)

    # Return results for this threshold
    tibble::tibble(
      threshold = threshold,
      metric = c(
        paste0("accuracy_t", threshold),
        paste0("precision_t", threshold),
        paste0("recall_t", threshold),
        paste0("f1_t", threshold),
        paste0("f2_t", threshold),
        paste0("f0.5_t", threshold)
      ),
      value = c(acc, prec, rec, f1, f2, f0_5)
    )
  })

  threshold_results
}


#' Evaluate a tidylearn model
#' @param object A tidylearn model object
#' @param new_data Optional new data for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A \link[tibble]{tibble} with columns \code{metric} (character)
#'   and \code{value} (numeric). For regression models, includes
#'   \code{rmse}, \code{mae}, and \code{rsq}. For classification models,
#'   includes \code{accuracy}.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_evaluate(model)
#' }
#' @export
tl_evaluate <- function(object, new_data = NULL, ...) {
  if (is.null(new_data)) {
    new_data <- object$data
  }

  # Get predictions
  preds <- predict(object, new_data = new_data, ...)

  # Get actual values
  if (inherits(object, "tidylearn_supervised")) {
    response_var <- object$spec$response_var
    actuals <- new_data[[response_var]]

    # Calculate appropriate metrics
    if (object$spec$is_classification) {
      # Classification metrics
      predicted_classes <- preds$.pred
      acc <- mean(predicted_classes == actuals, na.rm = TRUE)

      tibble::tibble(
        metric = "accuracy",
        value = acc
      )
    } else {
      # Regression metrics
      predicted_values <- preds$.pred
      rmse <- sqrt(mean((predicted_values - actuals)^2, na.rm = TRUE))
      mae <- mean(abs(predicted_values - actuals), na.rm = TRUE)
      rsq <- cor(predicted_values, actuals, use = "complete.obs")^2

      tibble::tibble(
        metric = c("rmse", "mae", "rsq"),
        value = c(rmse, mae, rsq)
      )
    }
  } else {
    # Unsupervised evaluation (placeholder)
    tibble::tibble(
      metric = "completed",
      value = 1
    )
  }
}

#' Cross-validation for tidylearn models
#' @param data Data frame
#' @param formula Model formula
#' @param method Modeling method
#' @param folds Number of cross-validation folds
#' @param ... Additional arguments
#' @return A list with two elements:
#'   \describe{
#'     \item{\code{$folds}}{A list of per-fold evaluation
#'       \link[tibble]{tibble}s, each with \code{metric} and
#'       \code{value} columns.}
#'     \item{\code{$summary}}{A \link[tibble]{tibble} with columns
#'       \code{metric}, \code{mean}, and \code{sd} summarizing
#'       performance across folds.}
#'   }
#' @examples
#' \donttest{
#' cv <- tl_cv(mtcars, mpg ~ wt + hp, method = "linear", folds = 3)
#' cv$summary
#' }
#' @export
tl_cv <- function(data, formula, method, folds = 5, ...) {
  n <- nrow(data)
  fold_size <- floor(n / folds)
  indices <- sample(1:n)

  cv_results <- list()

  for (i in 1:folds) {
    # Create fold indices
    test_indices <- indices[
      ((i - 1) * fold_size + 1):min(i * fold_size, n)
    ]
    train_indices <- setdiff(1:n, test_indices)

    # Split data
    train_data <- data[train_indices, ]
    test_data <- data[test_indices, ]

    # Train model
    model <- tl_model(train_data, formula, method = method, ...)

    # Evaluate
    eval_result <- tl_evaluate(model, new_data = test_data)

    cv_results[[i]] <- eval_result
  }

  # Combine results
  all_results <- dplyr::bind_rows(cv_results)

  # Calculate mean and sd for each metric
  summary_results <- all_results %>%
    dplyr::group_by(metric) %>%
    dplyr::summarize(
      mean = mean(value, na.rm = TRUE),
      sd = stats::sd(value, na.rm = TRUE),
      .groups = "drop"
    )

  list(
    folds = cv_results,
    summary = summary_results
  )
}

#' Tidy Distance Matrix Computation
#'
#' Compute distance matrices with tidy output
#'
#' @param data A data frame or tibble
#' @param method Character; distance method
#'   (default: "euclidean"). Options: "euclidean",
#'   "manhattan", "maximum", "gower"
#' @param cols Columns to include (tidy select).
#'   If NULL, uses all numeric columns.
#' @param ... Additional arguments passed to distance functions
#'
#' @return A \code{\link[stats]{dist}} object containing the computed
#'   distance matrix.
#'
#' @examples
#' \donttest{
#' d <- tidy_dist(iris[, 1:4], method = "euclidean")
#' }
#'
#' @export
tidy_dist <- function(data, method = "euclidean", cols = NULL, ...) {

  # Select columns
  if (!is.null(cols)) {
    cols_enquo <- rlang::enquo(cols)
    data_selected <- data %>% dplyr::select(!!cols_enquo)
  } else {
    data_selected <- data
  }

  # Compute distance based on method
  if (method == "gower") {
    dist_mat <- tidy_gower(data_selected, ...)
  } else {
    # Convert to matrix for standard methods
    data_matrix <- as.matrix(data_selected %>% dplyr::select(where(is.numeric)))
    dist_mat <- stats::dist(data_matrix, method = method)
  }

  dist_mat
}


#' Gower Distance Calculation
#'
#' Computes Gower distance for mixed data types (numeric, factor, ordered)
#'
#' @param data A data frame or tibble
#' @param weights Optional named vector of variable
#'   weights (default: equal weights)
#'
#' @return A \code{\link[stats]{dist}} object containing Gower distances, with
#'   the \code{method} attribute set to \code{"gower"}.
#'
#' @details
#' Gower distance handles mixed data types:
#' - Numeric: range-normalized Manhattan distance
#' - Factor/Character: 0 if same, 1 if different
#' - Ordered: treated as numeric ranks
#'
#' Formula: d_ij = sum(w_k * d_ijk) / sum(w_k)
#' where d_ijk is the dissimilarity for variable k between obs i and j
#'
#' @examples
#' # Create example data with mixed types
#' car_data <- data.frame(
#'   horsepower = c(130, 250, 180),
#'   weight = c(1200, 1650, 1420),
#'   color = factor(c("red", "black", "blue"))
#' )
#'
#' # Compute Gower distance
#' gower_dist <- tidy_gower(car_data)
#'
#' @export
tidy_gower <- function(data, weights = NULL) {

  # Convert to data frame
  data <- as.data.frame(data)

  n <- nrow(data)
  p <- ncol(data)

  # Set up weights
  if (is.null(weights)) {
    weights <- rep(1, p)
    names(weights) <- colnames(data)
  }

  # Pre-computation pass
  #
  # 1. Extract each column to a plain vector in a list.
  #    data[i, k] inside a loop dispatches to [.data.frame on every call:
  #    S3 method lookup + argument matching + potential intermediate allocation.
  #    col_vecs[[k]][i] is a hash-table lookup then a C-level pointer offset —
  #    benchmarks show 10-100x faster for scalar access.
  #
  # 2. Store each column's type as a string so we avoid repeated is.numeric() /
  #    is.ordered() / is.factor() S3 predicate calls inside the hot triple loop.
  #
  # 3. Ranges and rank vectors stay precomputed
  col_vecs   <- as.list(data)           # plain-vector views, no copy
  col_ranges <- vector("numeric",   p)
  col_ranks  <- vector("list",      p)
  col_type   <- vector("character", p)  # numeric, ordered, or categorical

  for (k in seq_len(p)) {
    v <- col_vecs[[k]]
    if (is.ordered(v)) {
      r              <- as.numeric(v)
      col_ranks[[k]] <- r
      col_ranges[k]  <- max(r, na.rm = TRUE) - min(r, na.rm = TRUE)
      col_type[k]    <- "ordered"
    } else if (is.numeric(v)) {
      col_ranges[k]  <- max(v, na.rm = TRUE) - min(v, na.rm = TRUE)
      col_type[k]    <- "numeric"
    } else {
      col_type[k]    <- "categorical"
    }
  }

  # Initialize distance matrix
  dist_matrix <- matrix(0, nrow = n, ncol = n)

  # Compute pairwise distances. seq_len() yields an empty sequence when
  # n < 2, so single-row input returns an empty dist instead of erroring.
  for (i in seq_len(n - 1)) {
    for (j in (i + 1):n) {

      total_dist <- 0
      valid_vars <- 0

      # Process each variable
      for (k in seq_len(p)) {

        # Plain vector indexing — avoids [.data.frame overhead on every access
        xi <- col_vecs[[k]][i]
        xj <- col_vecs[[k]][j]

        if (is.na(xi) || is.na(xj)) next

        valid_vars <- valid_vars + weights[k]

        # Column type already resolved — no S3 predicate calls in the hot path
        if (col_type[k] == "numeric") {
          d_k <- if (col_ranges[k] > 0) abs(xi - xj) / col_ranges[k] else 0

        } else if (col_type[k] == "ordered") {
          ri <- col_ranks[[k]][i]
          rj <- col_ranks[[k]][j]
          d_k <- if (col_ranges[k] > 0) abs(ri - rj) / col_ranges[k] else 0

        } else {
          d_k <- if (xi == xj) 0 else 1
        }

        total_dist <- total_dist + weights[k] * d_k
      }

      # Average over valid variables
      if (valid_vars > 0) {
        dist_matrix[i, j] <- total_dist / valid_vars
        dist_matrix[j, i] <- dist_matrix[i, j]  # Symmetric
      }
    }
  }

  # Convert to dist object
  dist_obj <- stats::as.dist(dist_matrix)

  # Preserve row names if available
  if (!is.null(rownames(data))) {
    attr(dist_obj, "Labels") <- rownames(data)
  }

  attr(dist_obj, "method") <- "gower"

  dist_obj
}


#' Standardize Data
#'
#' Center and/or scale numeric variables
#'
#' @param data A data frame or tibble
#' @param center Logical; center variables? (default: TRUE)
#' @param scale Logical; scale variables to unit variance? (default: TRUE)
#'
#' @return A tibble with numeric variables centered and/or scaled as specified;
#'   non-numeric columns are returned unchanged.
#'
#' @examples
#' \donttest{
#' std <- standardize_data(iris[, 1:4])
#' }
#'
#' @export
standardize_data <- function(data, center = TRUE, scale = TRUE) {

  data_std <- data %>%
    dplyr::mutate(
      dplyr::across(
        where(is.numeric),
        ~ if (center && scale) {
          as.numeric(base::scale(.x, center = TRUE, scale = TRUE))
        } else if (center) {
          .x - mean(.x, na.rm = TRUE)
        } else if (scale) {
          .x / stats::sd(.x, na.rm = TRUE)
        } else {
          .x
        }
      )
    )

  data_std
}


#' Compare Distance Methods
#'
#' Compute distances using multiple methods for comparison
#'
#' @param data A data frame or tibble
#' @param methods Character vector of methods to compare
#'
#' @return A named list of \code{\link[stats]{dist}} objects, one per method.
#'
#' @examples
#' \donttest{
#' dists <- compare_distances(
#'   iris[, 1:4], methods = c("euclidean", "manhattan")
#' )
#' }
#'
#' @export
compare_distances <- function(
    data,
    methods = c("euclidean", "manhattan", "maximum")) {

  data_numeric <- data %>% dplyr::select(where(is.numeric))

  dist_list <- purrr::map(methods, function(method) {
    tidy_dist(data = data_numeric, method = method)
  })
  names(dist_list) <- methods

  dist_list
}

test_that("PCA models work", {
  model <- tl_model(iris[, 1:4], method = "pca")

  expect_s3_class(model, "tidylearn_pca")
  expect_equal(model$spec$paradigm, "unsupervised")

  # Check PCA components
  expect_true("scores" %in% names(model$fit))
  expect_true("loadings" %in% names(model$fit))
  expect_true("variance_explained" %in% names(model$fit))

  # Transform data
  transformed <- predict(model)
  expect_s3_class(transformed, "tbl_df")
  expect_true(
    all(grepl("PC", names(transformed)) |
          names(transformed) == ".obs_id")
  )
})

test_that("K-means clustering works", {
  model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)

  expect_s3_class(model, "tidylearn_kmeans")

  # Check cluster assignments
  expect_true("clusters" %in% names(model$fit))
  clusters <- model$fit$clusters
  expect_equal(nrow(clusters), nrow(iris))
  expect_true("cluster" %in% names(clusters))

  # Clusters should be 1 to k
  expect_true(all(clusters$cluster %in% 1:3))
})

test_that("PAM (K-medoids) clustering works", {
  skip_if_not_installed("cluster")

  model <- tl_model(iris[, 1:4], method = "pam", k = 3)

  expect_s3_class(model, "tidylearn_pam")

  # Check cluster assignments
  clusters <- model$fit$clusters
  expect_equal(nrow(clusters), nrow(iris))
  expect_true(all(clusters$cluster %in% 1:3))
})

test_that("CLARA clustering works", {
  skip_if_not_installed("cluster")

  # Create larger dataset for CLARA
  large_data <- iris[rep(seq_len(nrow(iris)), 10), 1:4]

  model <- tl_model(large_data, method = "clara", k = 3, samples = 5)

  expect_s3_class(model, "tidylearn_clara")

  # Check cluster assignments
  clusters <- model$fit$clusters
  expect_equal(nrow(clusters), nrow(large_data))
})

test_that("Hierarchical clustering works", {
  model <- tl_model(iris[, 1:4], method = "hclust")

  expect_s3_class(model, "tidylearn_hclust")

  # Check dendrogram exists
  expect_true("model" %in% names(model$fit))
  expect_s3_class(model$fit$model, "hclust")
})

test_that("DBSCAN clustering works", {
  skip_if_not_installed("dbscan")

  model <- tl_model(iris[, 1:4], method = "dbscan", eps = 0.5, minPts = 5)

  expect_s3_class(model, "tidylearn_dbscan")

  # Check cluster assignments (including noise points as 0)
  clusters <- model$fit$clusters
  expect_equal(nrow(clusters), nrow(iris))
  expect_true("cluster" %in% names(clusters))
})

test_that("MDS works", {
  model <- tl_model(iris[, 1:4], method = "mds", k = 2)

  expect_s3_class(model, "tidylearn_mds")

  # Check MDS points
  expect_true("points" %in% names(model$fit))
  points <- model$fit$points
  expect_equal(nrow(points), nrow(iris))
})

test_that("clustering models predict on new data", {
  # Train clustering model
  model <- tl_model(iris[1:100, 1:4], method = "kmeans", k = 3)

  # Predict on new data
  new_data <- iris[101:150, 1:4]
  predictions <- predict(model, new_data = new_data)

  expect_equal(nrow(predictions), nrow(new_data))
  expect_true("cluster" %in% names(predictions))
})

test_that("PCA retains specified number of components", {
  model <- tl_model(iris[, 1:4], method = "pca")

  # Default should retain all components
  transformed <- predict(model)
  pc_cols <- sum(grepl("^PC", names(transformed)))
  expect_equal(pc_cols, 4)
})

test_that("unsupervised methods handle different data sizes", {
  # Small dataset
  small_data <- iris[1:20, 1:4]
  model_small <- tl_model(small_data, method = "kmeans", k = 2)
  expect_s3_class(model_small, "tidylearn_kmeans")

  # Large dataset
  large_data <- iris[rep(seq_len(nrow(iris)), 5), 1:4]
  model_large <- tl_model(large_data, method = "kmeans", k = 3)
  expect_s3_class(model_large, "tidylearn_kmeans")
})

test_that("clustering validates k parameter", {
  # k should be reasonable - expect an error for invalid k
  expect_error(
    tl_model(iris[, 1:4], method = "kmeans", k = nrow(iris) + 1)
  )

  # Valid k should work
  expect_s3_class(
    tl_model(iris[, 1:4], method = "kmeans", k = 3),
    "tidylearn_kmeans"
  )
})

test_that("tidy_gower returns a dist object with correct metadata", {
  d <- tidy_gower(iris[1:10, 1:4])

  expect_s3_class(d, "dist")
  expect_equal(attr(d, "Size"), 10L)
  expect_equal(attr(d, "method"), "gower")
})

test_that("tidy_gower handles single-row input without erroring", {
  d <- tidy_gower(iris[1, 1:4])

  expect_s3_class(d, "dist")
  expect_equal(length(d), 0L)          # no pairs to compute
  expect_equal(attr(d, "method"), "gower")
})

test_that("tidy_gower distances are non-negative and symmetric", {
  d <- tidy_gower(iris[1:20, 1:4])
  m <- as.matrix(d)

  expect_true(all(m >= 0))
  expect_equal(m, t(m))               # symmetric
  expect_true(all(diag(m) == 0))       # self-distance is 0
})

test_that("tidy_gower gives distance 0 for identical rows", {
  dup <- iris[c(1, 1, 2), 1:4]
  d <- as.matrix(tidy_gower(dup))

  expect_equal(d[1, 2], 0)   # row 1 vs its duplicate
  expect_gt(d[1, 3], 0)      # genuinely different rows are non-zero
})

test_that("tidy_gower numeric-only distances are correct", {
  # Two observations: x in {0, 1}.  range = 1, so d = |0-1|/1 = 1.
  df <- data.frame(x = c(0, 1))
  d  <- as.matrix(tidy_gower(df))

  expect_equal(d[1, 2], 1)

  # Three observations: x in {0, 0.5, 1}.
  # d(1,2) = 0.5, d(1,3) = 1, d(2,3) = 0.5
  df3 <- data.frame(x = c(0, 0.5, 1))
  d3  <- as.matrix(tidy_gower(df3))

  expect_equal(d3[1, 2], 0.5)
  expect_equal(d3[1, 3], 1.0)
  expect_equal(d3[2, 3], 0.5)
})

test_that("tidy_gower categorical distances are correct", {
  # Same category → 0; different category → 1
  df <- data.frame(color = factor(c("red", "red", "blue")))
  d  <- as.matrix(tidy_gower(df))

  expect_equal(d[1, 2], 0)   # red vs red
  expect_equal(d[1, 3], 1)   # red vs blue
  expect_equal(d[2, 3], 1)   # red vs blue
})

test_that("tidy_gower ordered distances are correct", {
  # Ranks: low=1, medium=2, high=3; range = 2
  # d(low, medium) = 1/2, d(low, high) = 2/2 = 1, d(medium, high) = 1/2
  df <- data.frame(
    rating = ordered(c("low", "medium", "high"),
                     levels = c("low", "medium", "high"))
  )
  d <- as.matrix(tidy_gower(df))

  expect_equal(d[1, 2], 0.5)
  expect_equal(d[1, 3], 1.0)
  expect_equal(d[2, 3], 0.5)
})

test_that("tidy_gower handles mixed numeric and categorical columns", {
  # x in {0, 1, 0.5}, color in {red, red, blue}; equal weights, so each
  # pairwise distance is the average of the two d_k.
  # d(1,2): d_x = 1, d_color = 0  → 0.5
  # d(1,3): d_x = 0.5, d_color = 1  → 0.75
  # d(2,3): d_x = 0.5, d_color = 1  → 0.75
  df <- data.frame(
    x     = c(0, 1, 0.5),
    color = factor(c("red", "red", "blue"))
  )
  d <- as.matrix(tidy_gower(df))

  expect_equal(d[1, 2], 0.50)
  expect_equal(d[1, 3], 0.75)
  expect_equal(d[2, 3], 0.75)
})

test_that("tidy_gower skips NA values when computing distances", {
  # x: {0, 1}, y: {1, NA}.  For pair (1,2) only x is valid.
  # d = |0-1| / (1-0) = 1
  df <- data.frame(x = c(0, 1), y = c(1, NA_real_))
  d  <- as.matrix(tidy_gower(df))

  expect_equal(d[1, 2], 1)   # y skipped; only x contributes
})

test_that("tidy_gower respects custom weights", {
  # Upweight x so the numeric dimension dominates
  df <- data.frame(
    x     = c(0, 1),
    color = factor(c("red", "blue"))
  )
  d_equal    <- as.matrix(tidy_gower(df))
  d_weighted <- as.matrix(tidy_gower(df, weights = c(x = 10, color = 1)))

  # Need ≥3 rows so the column range is wider than the pair (1,2) difference.
  # With only 2 rows, range == difference, so d_x = 1 always — weights can't
  # move a result that is already at its maximum.
  #
  # x: range = 1, d(1,2)_x = |0 - 0.1| / 1 = 0.1
  # color: d(1,2)_color = 1  (red vs blue)
  # equal weights: (0.1 + 1) / 2 = 0.55
  # heavy color (1, 100): (1*0.1 + 100*1) / 101 ≈ 0.991
  df2 <- data.frame(
    x     = c(0, 0.1, 1),
    color = factor(c("red", "blue", "red"))
  )
  d_eq2 <- as.matrix(tidy_gower(df2))
  d_wt2 <- as.matrix(tidy_gower(df2, weights = c(x = 1, color = 100)))

  expect_gt(d_wt2[1, 2], d_eq2[1, 2])  # heavy color weight → larger distance
})

test_that("tidy_gower constant column adds to denominator not numerator", {
  # A constant column (range = 0) has d_k = 0 for every pair, so it adds
  # nothing to the numerator.  However, both observations are non-NA, so the
  # column IS counted in valid_vars (the denominator).  The net effect is that
  # the overall distance is exactly halved compared to y alone.
  #
  # pair (1,2): d_x = 0, d_y = |0-1|/1 = 1  → (0 + 1) / 2 = 0.5
  # y alone:    d_y = 1/1 = 1
  df   <- data.frame(x = c(5, 5, 5), y = c(0, 1, 0.5))
  d    <- as.matrix(tidy_gower(df))
  d_y  <- as.matrix(tidy_gower(data.frame(y = c(0, 1, 0.5))))

  expect_equal(d, d_y / 2)
})


test_that("tidy_dist dispatches to tidy_gower for method = 'gower'", {
  df <- data.frame(
    x     = c(1, 2, 3),
    color = factor(c("a", "b", "a"))
  )
  d1 <- tidy_dist(df, method = "gower")
  d2 <- tidy_gower(df)

  expect_equal(as.vector(d1), as.vector(d2))
  expect_equal(attr(d1, "method"), "gower")
})

test_that("unsupervised methods work with formula", {
  # PCA with formula
  model <- tl_model(
    iris,
    ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
    method = "pca"
  )

  expect_s3_class(model, "tidylearn_pca")

  # Clustering with formula
  model2 <- tl_model(iris, ~ Sepal.Length + Sepal.Width,
                     method = "kmeans", k = 3)
  expect_s3_class(model2, "tidylearn_kmeans")
})

#' Tidy Silhouette Analysis
#'
#' Compute silhouette statistics for cluster validation
#'
#' @param clusters Vector of cluster assignments
#' @param dist_mat Distance matrix (dist object)
#'
#' @return A list of class "tidy_silhouette" containing:
#' \itemize{
#'   \item silhouette_data: tibble with silhouette values for each observation
#'   \item avg_width: average silhouette width
#'   \item cluster_avg: average silhouette width by cluster
#' }
#'
#' @examples
#' \donttest{
#' km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
#' d <- dist(iris[, 1:4])
#' sil <- tidy_silhouette(km$cluster, d)
#' }
#'
#' @export
tidy_silhouette <- function(clusters, dist_mat) {

  if (!inherits(dist_mat, "dist")) {
    stop("dist_mat must be a dist object")
  }

  # Compute silhouette
  sil <- cluster::silhouette(clusters, dist_mat)

  # Create silhouette tibble
  sil_tbl <- tibble::as_tibble(sil[, 1:3]) %>%
    dplyr::rename(
      cluster = cluster,
      neighbor = neighbor,
      sil_width = sil_width
    ) %>%
    dplyr::mutate(.id = rownames(sil) %||% seq_len(nrow(sil)), .before = 1)

  # Calculate average by cluster
  cluster_avg <- sil_tbl %>%
    dplyr::group_by(cluster) %>%
    dplyr::summarise(
      n = dplyr::n(),
      avg_sil_width = mean(sil_width),
      .groups = "drop"
    )

  # Overall average
  avg_width <- mean(sil_tbl$sil_width)

  result <- list(
    silhouette_data = sil_tbl,
    avg_width = avg_width,
    cluster_avg = cluster_avg
  )

  class(result) <- c("tidy_silhouette", "list")
  result
}


#' Silhouette Analysis Across Multiple k Values
#'
#' @param data A data frame or tibble
#' @param max_k Maximum number of clusters to test (default: 10)
#' @param method Clustering method: "kmeans" (default) or "hclust"
#' @param nstart If kmeans, number of random starts (default: 25)
#' @param dist_method Distance metric (default: "euclidean")
#' @param linkage_method If hclust, linkage method (default: "average")
#'
#' @return A tibble with columns \code{k} and \code{avg_sil_width}. The
#'   \code{"optimal_k"} attribute contains the k with the highest average
#'   silhouette width.
#'
#' @examples
#' \donttest{
#' sil_analysis <- tidy_silhouette_analysis(iris[, 1:4], max_k = 6)
#' }
#'
#' @export
tidy_silhouette_analysis <- function(data, max_k = 10, method = "kmeans",
                                     nstart = 25, dist_method = "euclidean",
                                     linkage_method = "average") {

  data_numeric <- data %>% dplyr::select(where(is.numeric))
  dist_mat <- stats::dist(data_numeric, method = dist_method)

  # Compute silhouette for k = 2 to max_k
  sil_results <- purrr::map_dfr(2:max_k, function(k) {

    if (method == "kmeans") {
      km <- stats::kmeans(data_numeric, centers = k, nstart = nstart)
      clusters <- km$cluster
    } else if (method == "hclust") {
      hc <- stats::hclust(dist_mat, method = linkage_method)
      clusters <- stats::cutree(hc, k = k)
    } else {
      stop("method must be 'kmeans' or 'hclust'")
    }

    sil <- cluster::silhouette(clusters, dist_mat)

    tibble::tibble(
      k = k,
      avg_sil_width = mean(sil[, 3])
    )
  })

  # Add optimal k
  optimal_k <- sil_results$k[which.max(sil_results$avg_sil_width)]

  attr(sil_results, "optimal_k") <- optimal_k
  attr(sil_results, "method") <- method

  sil_results
}


#' Plot Silhouette Analysis
#'
#' @param sil_obj A tidy_silhouette object or tibble
#'   from tidy_silhouette_analysis
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
#' d <- dist(iris[, 1:4])
#' sil <- tidy_silhouette(km$cluster, d)
#' plot_silhouette(sil)
#' }
#'
#' @export
plot_silhouette <- function(sil_obj) {

  if (inherits(sil_obj, "tidy_silhouette")) {
    # Individual silhouette plot
    sil_data <- sil_obj$silhouette_data

    p <- ggplot2::ggplot(
      sil_data,
      ggplot2::aes(
        x = .id, y = sil_width,
        fill = as.factor(cluster)
      )
    ) +
      ggplot2::geom_col() +
      ggplot2::geom_hline(
        yintercept = sil_obj$avg_width,
        linetype = "dashed", color = "red"
      ) +
      ggplot2::facet_wrap(~cluster, scales = "free_x") +
      ggplot2::labs(
        title = "Silhouette Plot",
        subtitle = sprintf("Average silhouette width: %.3f", sil_obj$avg_width),
        x = "Observation",
        y = "Silhouette Width",
        fill = "Cluster"
      ) +
      ggplot2::theme_minimal() +
      ggplot2::theme(axis.text.x = ggplot2::element_blank())

  } else if (is.data.frame(sil_obj) && "avg_sil_width" %in% names(sil_obj)) {
    # Silhouette across multiple k values
    optimal_k <- attr(sil_obj, "optimal_k")

    p <- ggplot2::ggplot(sil_obj, ggplot2::aes(x = k, y = avg_sil_width)) +
      ggplot2::geom_line(color = "steelblue", linewidth = 1) +
      ggplot2::geom_point(color = "steelblue", size = 3) +
      ggplot2::geom_point(
        data = sil_obj %>% dplyr::filter(k == optimal_k),
        color = "red", size = 5
      ) +
      ggplot2::labs(
        title = "Average Silhouette Width vs Number of Clusters",
        subtitle = sprintf("Optimal k = %d (red point)", optimal_k),
        x = "Number of Clusters (k)",
        y = "Average Silhouette Width"
      ) +
      ggplot2::theme_minimal()

  } else {
    stop(
      "sil_obj must be a tidy_silhouette object ",
      "or silhouette analysis tibble"
    )
  }

  p
}


#' Tidy Gap Statistic
#'
#' Compute gap statistic for determining optimal number of clusters
#'
#' @param data A data frame or tibble
#' @param FUN_cluster Clustering function (default: uses kmeans internally)
#' @param max_k Maximum number of clusters (default: 10)
#' @param B Number of bootstrap samples (default: 50)
#' @param nstart If using kmeans, number of random starts (default: 25)
#'
#' @return A list of class \code{"tidy_gap"} containing:
#' \itemize{
#'   \item gap_data: tibble with gap statistics for each k
#'   \item k_firstSEmax: optimal k via firstSEmax method (most conservative)
#'   \item k_globalmax: optimal k via globalmax method
#'   \item k_firstmax: optimal k via firstmax method
#'   \item recommended_k: recommended k (uses firstSEmax)
#'   \item model: the \code{\link[cluster]{clusGap}} result
#' }
#'
#' @examples
#' \donttest{
#' gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
#' gap$recommended_k
#' }
#'
#' @export
tidy_gap_stat <- function(data, FUN_cluster = NULL,  # nolint
                          max_k = 10, B = 50,  # nolint
                          nstart = 25) {

  data_numeric <- data %>% dplyr::select(where(is.numeric))

  # Use cluster::clusGap
  if (is.null(FUN_cluster)) {
    gap_result <- cluster::clusGap(
      data_numeric,
      FUN = stats::kmeans,
      nstart = nstart,
      K.max = max_k,
      B = B
    )
  } else {
    gap_result <- cluster::clusGap(
      data_numeric,
      FUN = FUN_cluster,
      K.max = max_k,
      B = B
    )
  }

  # Extract results as tibble
  gap_tbl <- tibble::as_tibble(gap_result$Tab) %>%
    dplyr::mutate(k = 1:max_k, .before = 1)

  # Determine optimal k using different methods
  k_firstSEmax <- cluster::maxSE(gap_result$Tab[, "gap"],  # nolint
                                 gap_result$Tab[, "SE.sim"],
                                 method = "firstSEmax")

  k_globalmax <- cluster::maxSE(gap_result$Tab[, "gap"],
                                gap_result$Tab[, "SE.sim"],
                                method = "globalmax")

  k_firstmax <- which.max(gap_result$Tab[, "gap"])

  result <- list(
    gap_data = gap_tbl,
    k_firstSEmax = k_firstSEmax,
    k_globalmax = k_globalmax,
    k_firstmax = k_firstmax,
    recommended_k = k_firstSEmax,  # Most conservative
    model = gap_result
  )

  class(result) <- c("tidy_gap", "list")
  result
}


#' Plot Gap Statistic
#'
#' @param gap_obj A tidy_gap object
#' @param show_methods Logical; show all three k
#'   selection methods? (default: FALSE)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#'
#' @examples
#' \donttest{
#' gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
#' plot_gap_stat(gap)
#' }
#'
#' @export
plot_gap_stat <- function(gap_obj, show_methods = FALSE) {

  if (!inherits(gap_obj, "tidy_gap")) {
    stop("gap_obj must be a tidy_gap object")
  }

  gap_data <- gap_obj$gap_data

  p <- ggplot2::ggplot(gap_data, ggplot2::aes(x = k, y = gap)) +
    ggplot2::geom_line(color = "steelblue", linewidth = 1) +
    ggplot2::geom_point(color = "steelblue", size = 3) +
    ggplot2::geom_errorbar(
      ggplot2::aes(ymin = gap - SE.sim, ymax = gap + SE.sim),
      width = 0.2, alpha = 0.5
    ) +
    ggplot2::labs(
      title = "Gap Statistic",
      subtitle = sprintf(
        "Recommended k = %d (firstSEmax method)",
        gap_obj$recommended_k
      ),
      x = "Number of Clusters (k)",
      y = "Gap Statistic"
    ) +
    ggplot2::theme_minimal()

  # Add vertical lines for different methods
  gap_y <- max(gap_data$gap) * 0.95
  if (show_methods) {
    p <- p +
      ggplot2::geom_vline(
        xintercept = gap_obj$k_firstSEmax,
        color = "red", linetype = "dashed"
      ) +
      ggplot2::geom_vline(
        xintercept = gap_obj$k_globalmax,
        color = "purple", linetype = "dashed"
      ) +
      ggplot2::geom_vline(
        xintercept = gap_obj$k_firstmax,
        color = "green", linetype = "dashed"
      ) +
      ggplot2::annotate(
        "text",
        x = gap_obj$k_firstSEmax, y = gap_y,
        label = "firstSEmax", color = "red",
        angle = 90, vjust = -0.5, size = 3
      ) +
      ggplot2::annotate(
        "text",
        x = gap_obj$k_globalmax, y = gap_y,
        label = "globalmax", color = "purple",
        angle = 90, vjust = -0.5, size = 3
      ) +
      ggplot2::annotate(
        "text",
        x = gap_obj$k_firstmax, y = gap_y,
        label = "firstmax", color = "green",
        angle = 90, vjust = -0.5, size = 3
      )
  } else {
    p <- p +
      ggplot2::geom_vline(
        xintercept = gap_obj$recommended_k,
        color = "red", linetype = "dashed"
      )
  }

  p
}


#' Calculate Cluster Validation Metrics
#'
#' Comprehensive validation metrics for a clustering result
#'
#' @param clusters Vector of cluster assignments
#' @param data Original data frame (for WSS calculation)
#' @param dist_mat Distance matrix (for silhouette)
#'
#' @return A single-row tibble with columns \code{k}, \code{min_size},
#'   \code{max_size}, \code{avg_size}, and optionally \code{avg_silhouette},
#'   \code{min_silhouette} (if \code{dist_mat} provided), and \code{total_wss}
#'   (if \code{data} provided).
#'
#' @examples
#' \donttest{
#' km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
#' d <- dist(iris[, 1:4])
#' metrics <- calc_validation_metrics(km$cluster, iris[, 1:4], d)
#' }
#'
#' @export
calc_validation_metrics <- function(clusters, data = NULL, dist_mat = NULL) {

  metrics <- list()

  # Number of clusters
  k <- length(unique(clusters[clusters != 0]))  # Exclude noise (0) if present
  metrics$k <- k

  # Cluster sizes
  cluster_sizes <- table(clusters)
  metrics$min_size <- min(cluster_sizes)
  metrics$max_size <- max(cluster_sizes)
  metrics$avg_size <- mean(cluster_sizes)

  # Silhouette if distance matrix provided
  if (!is.null(dist_mat)) {
    sil <- cluster::silhouette(clusters, dist_mat)
    metrics$avg_silhouette <- mean(sil[, 3])
    metrics$min_silhouette <- min(sil[, 3])
  }

  # WSS if data provided
  if (!is.null(data)) {
    data_numeric <- data %>% dplyr::select(where(is.numeric))

    # Total within-cluster sum of squares
    wss <- sum(sapply(unique(clusters), function(cl) {
      cluster_data <- data_numeric[clusters == cl, , drop = FALSE]
      if (nrow(cluster_data) > 1) {
        center <- colMeans(cluster_data)
        sum((t(cluster_data) - center)^2)
      } else {
        0
      }
    }))

    metrics$total_wss <- wss
  }

  tibble::as_tibble(metrics)
}


#' Compare Multiple Clustering Results
#'
#' @param cluster_list Named list of cluster assignment vectors
#' @param data Original data
#' @param dist_mat Distance matrix
#'
#' @return A tibble with one row per clustering method and columns for each
#'   validation metric (see \code{\link{calc_validation_metrics}}), plus a
#'   \code{method} column identifying the clustering.
#'
#' @examples
#' \donttest{
#' km3 <- kmeans(iris[, 1:4], 3, nstart = 25)$cluster
#' km4 <- kmeans(iris[, 1:4], 4, nstart = 25)$cluster
#' compare_clusterings(list(k3 = km3, k4 = km4), iris[, 1:4])
#' }
#'
#' @export
compare_clusterings <- function(cluster_list, data, dist_mat = NULL) {

  if (is.null(dist_mat)) {
    data_numeric <- data %>% dplyr::select(where(is.numeric))
    dist_mat <- stats::dist(data_numeric)
  }

  comparison <- purrr::map_dfr(names(cluster_list), function(name) {
    clusters <- cluster_list[[name]]
    metrics <- calc_validation_metrics(clusters, data, dist_mat)
    metrics %>% dplyr::mutate(method = name, .before = 1)
  })

  comparison
}


#' Print Method for tidy_silhouette
#'
#' @param x A tidy_silhouette object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
#' d <- dist(iris[, 1:4])
#' sil <- tidy_silhouette(km$cluster, d)
#' print(sil)
#' }
#'
#' @export
print.tidy_silhouette <- function(x, ...) {
  cat("Tidy Silhouette Analysis\n")
  cat("========================\n\n")
  cat("Average silhouette width:", round(x$avg_width, 4), "\n\n")

  cat("Interpretation:\n")
  cat("  > 0.70: Strong structure\n")
  cat("  > 0.50: Reasonable structure\n")
  cat("  > 0.25: Weak structure\n")
  cat("  < 0.25: No substantial structure\n\n")

  cat("By Cluster:\n")
  print(x$cluster_avg)

  invisible(x)
}


#' Print Method for tidy_gap
#'
#' @param x A tidy_gap object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
#' print(gap)
#' }
#'
#' @export
print.tidy_gap <- function(x, ...) {
  cat("Tidy Gap Statistic\n")
  cat("==================\n\n")
  cat("Recommended k:", x$recommended_k, "(firstSEmax method)\n\n")

  cat("Alternative methods:\n")
  cat("  firstSEmax: k =", x$k_firstSEmax, "(most conservative)\n")
  cat("  globalmax:  k =", x$k_globalmax, "(middle ground)\n")
  cat("  firstmax:   k =", x$k_firstmax, "(most liberal)\n\n")

  cat("Gap Statistics (first 10):\n")
  print(head(x$gap_data, 10))

  invisible(x)
}

#' @title Regularization Functions for tidylearn
#' @name tidylearn-regularization
#' @description Ridge, Lasso, and Elastic Net regularization
#'   functionality
#' @importFrom glmnet glmnet cv.glmnet predict.glmnet
#' @importFrom stats model.matrix as.formula
#' @importFrom tibble tibble
NULL

#' Fit a Ridge regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param alpha Mixing parameter (0 for Ridge, 1 for Lasso,
#'   between 0-1 for Elastic Net)
#' @param lambda Regularization parameter (if NULL, uses
#'   cross-validation to select)
#' @param cv_folds Number of folds for cross-validation
#'   (default: 5)
#' @param ... Additional arguments to pass to glmnet()
#' @return A fitted Ridge regression model
#' @keywords internal
tl_fit_ridge <- function(data, formula,
                         is_classification = FALSE,
                         alpha = 0, lambda = NULL,
                         cv_folds = 5, ...) {
  tl_fit_regularized(
    data, formula, is_classification,
    alpha, lambda, cv_folds, ...
  )
}

#' Fit a Lasso regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param alpha Mixing parameter (0 for Ridge, 1 for Lasso,
#'   between 0-1 for Elastic Net)
#' @param lambda Regularization parameter (if NULL, uses
#'   cross-validation to select)
#' @param cv_folds Number of folds for cross-validation
#'   (default: 5)
#' @param ... Additional arguments to pass to glmnet()
#' @return A fitted Lasso regression model
#' @keywords internal
tl_fit_lasso <- function(data, formula,
                         is_classification = FALSE,
                         alpha = 1, lambda = NULL,
                         cv_folds = 5, ...) {
  tl_fit_regularized(
    data, formula, is_classification,
    alpha, lambda, cv_folds, ...
  )
}

#' Fit an Elastic Net regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param alpha Mixing parameter (default: 0.5 for
#'   Elastic Net)
#' @param lambda Regularization parameter (if NULL, uses
#'   cross-validation to select)
#' @param cv_folds Number of folds for cross-validation
#'   (default: 5)
#' @param ... Additional arguments to pass to glmnet()
#' @return A fitted Elastic Net regression model
#' @keywords internal
tl_fit_elastic_net <- function(data, formula,
                               is_classification = FALSE,
                               alpha = 0.5,
                               lambda = NULL,
                               cv_folds = 5, ...) {
  tl_fit_regularized(
    data, formula, is_classification,
    alpha, lambda, cv_folds, ...
  )
}

#' Fit a regularized regression model
#'
#' Fits Ridge, Lasso, or Elastic Net regularization.
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param alpha Mixing parameter (0 for Ridge, 1 for Lasso,
#'   between 0-1 for Elastic Net)
#' @param lambda Regularization parameter (if NULL, uses
#'   cross-validation to select)
#' @param cv_folds Number of folds for cross-validation
#'   (default: 5)
#' @param ... Additional arguments to pass to glmnet()
#' @return A fitted regularized regression model
#' @keywords internal
tl_fit_regularized <- function(data, formula,
                               is_classification = FALSE,
                               alpha = 0, lambda = NULL,
                               cv_folds = 5, ...) {
  # Check if glmnet is installed
  tl_check_packages("glmnet")

  # Parse the formula
  response_var <- all.vars(formula)[1]

  # Extract response (y) and predictors matrices
  y <- data[[response_var]]

  # Create model matrix for predictors
  # Remove intercept as glmnet adds it by default
  x_mat <- stats::model.matrix(
    formula, data = data
  )[, -1, drop = FALSE]

  # Determine the appropriate family based on problem type
  if (is_classification) {
    if (!is.factor(y)) {
      y <- factor(y)
    }

    if (length(levels(y)) == 2) {
      # Binary classification
      family <- "binomial"
    } else {
      # Multiclass classification
      family <- "multinomial"
    }
  } else {
    # Regression
    family <- "gaussian"
  }

  # Fit the model
  if (is.null(lambda)) {
    # Use cross-validation to select optimal lambda
    cv_model <- glmnet::cv.glmnet(
      x = x_mat,
      y = y,
      alpha = alpha,
      family = family,
      nfolds = cv_folds,
      ...
    )

    # Extract optimal lambda
    lambda_min <- cv_model$lambda.min
    lambda_1se <- cv_model$lambda.1se

    # Fit final model with optimal lambda
    # Use lambda.1se for better generalization
    model <- glmnet::glmnet(
      x = x_mat,
      y = y,
      alpha = alpha,
      family = family,
      lambda = cv_model$lambda,
      ...
    )

    # Store cross-validation results
    attr(model, "cv_results") <- cv_model
    attr(model, "lambda_min") <- lambda_min
    attr(model, "lambda_1se") <- lambda_1se
  } else {
    # Fit model with user-specified lambda
    model <- glmnet::glmnet(
      x = x_mat,
      y = y,
      alpha = alpha,
      family = family,
      lambda = lambda,
      ...
    )

    # Store lambda
    attr(model, "lambda_min") <- lambda
    attr(model, "lambda_1se") <- lambda
  }

  # Store formula, variables, and alpha
  attr(model, "formula") <- formula
  attr(model, "response_var") <- response_var
  attr(model, "alpha") <- alpha
  attr(model, "is_classification") <- is_classification

  model
}

#' Predict using a regularized regression model
#'
#' @param model A tidylearn regularized model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "class" or "prob" (for classification)
#' @param lambda Which lambda to use for prediction
#'   ("1se" or "min", default: "1se")
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_regularized <- function(model, new_data,
                                   type = "response",
                                   lambda = "1se",
                                   ...) {
  fit <- model$fit
  is_classification <- model$spec$is_classification
  formula <- model$spec$formula
  response_var <- model$spec$response_var

  # Extract lambda value to use
  if (lambda == "1se") {
    lambda_val <- attr(fit, "lambda_1se")
  } else if (lambda == "min") {
    lambda_val <- attr(fit, "lambda_min")
  } else if (is.numeric(lambda)) {
    lambda_val <- lambda
  } else {
    stop(
      "Invalid lambda specification. ",
      "Use '1se', 'min', or a numeric value.",
      call. = FALSE
    )
  }

  # Create model matrix for new data (excluding intercept)
  x_new <- stats::model.matrix(
    formula, data = new_data
  )[, -1, drop = FALSE]

  # Make predictions
  if (is_classification) {
    # Classification predictions
    if (type == "response" || type == "link") {
      # Linear predictor
      preds <- glmnet::predict.glmnet(
        fit,
        newx = x_new,
        s = lambda_val,
        type = "link",
        ...
      )
      as.vector(preds)
    } else if (type == "class") {
      # Predicted classes
      if ("family" %in% names(fit) &&
            fit$family == "binomial") {
        # Binary classification
        probs <- glmnet::predict.glmnet(
          fit,
          newx = x_new,
          s = lambda_val,
          type = "response",
          ...
        )

        # Get class levels from training data
        class_levels <- levels(
          factor(model$data[[response_var]])
        )

        # Classify based on probability > 0.5
        pred_classes <- ifelse(
          as.vector(probs) > 0.5,
          class_levels[2],
          class_levels[1]
        )
        pred_classes <- factor(
          pred_classes, levels = class_levels
        )

        pred_classes
      } else {
        # Multiclass classification
        preds <- glmnet::predict.glmnet(
          fit,
          newx = x_new,
          s = lambda_val,
          type = "class",
          ...
        )
        as.vector(preds)
      }
    } else if (type == "prob") {
      # Predicted probabilities
      if ("family" %in% names(fit) &&
            fit$family == "binomial") {
        # Binary classification
        pos_probs <- as.vector(
          glmnet::predict.glmnet(
            fit,
            newx = x_new,
            s = lambda_val,
            type = "response",
            ...
          )
        )
        neg_probs <- 1 - pos_probs

        # Get class levels from training data
        class_levels <- levels(
          factor(model$data[[response_var]])
        )

        # Create a data frame with probs for each class
        prob_df <- tibble::tibble(
          !!class_levels[1] := neg_probs,
          !!class_levels[2] := pos_probs
        )

        prob_df
      } else {
        # Multiclass classification
        probs <- glmnet::predict.glmnet(
          fit,
          newx = x_new,
          s = lambda_val,
          type = "response",
          ...
        )

        # Reshape to data frame
        prob_df <- as.data.frame(probs)
        names(prob_df) <- levels(
          factor(model$data[[response_var]])
        )

        tibble::as_tibble(prob_df)
      }
    } else {
      stop(
        "Invalid prediction type for classification. ",
        "Use 'response', 'class', or 'prob'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    preds <- glmnet::predict.glmnet(
      fit,
      newx = x_new,
      s = lambda_val,
      type = "response",
      ...
    )
    as.vector(preds)
  }
}

#' Predict using a Ridge regression model
#'
#' @param model A tidylearn Ridge model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_ridge <- function(model, new_data,
                             type = "response",
                             ...) {
  tl_predict_regularized(model, new_data, type, ...)
}

#' Predict using a Lasso regression model
#'
#' @param model A tidylearn Lasso model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_lasso <- function(model, new_data,
                             type = "response",
                             ...) {
  tl_predict_regularized(model, new_data, type, ...)
}

#' Predict using an Elastic Net regression model
#'
#' @param model A tidylearn Elastic Net model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_elastic_net <- function(model, new_data,
                                   type = "response",
                                   ...) {
  tl_predict_regularized(model, new_data, type, ...)
}

#' Plot regularization path for a regularized model
#'
#' @param model A tidylearn regularized model object
#' @param label_n Number of top features to label
#'   (default: 5)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ ., method = "lasso")
#' tl_plot_regularization_path(model)
#' }
#' @importFrom ggplot2 ggplot aes geom_line
#'   scale_x_log10 labs theme_minimal
#' @export
tl_plot_regularization_path <- function(model,
                                        label_n = 5,
                                        ...) {
  # Extract the glmnet model
  fit <- model$fit

  # Get coefficient matrix
  coef_matrix <- as.matrix(coef(fit))

  # Exclude intercept
  coef_matrix <- coef_matrix[-1, , drop = FALSE]

  # Create a data frame for plotting
  coef_df <- tibble::tibble(
    lambda = rep(
      fit$lambda, each = nrow(coef_matrix)
    ),
    feature = rep(
      rownames(coef_matrix),
      times = length(fit$lambda)
    ),
    coefficient = as.vector(coef_matrix)
  )

  # Identify the top features (by max absolute coef)
  top_features <- coef_df %>%
    dplyr::group_by(.data$feature) %>%
    dplyr::summarize(
      max_abs_coef = max(abs(.data$coefficient)),
      .groups = "drop"
    ) %>%
    dplyr::arrange(dplyr::desc(.data$max_abs_coef)) %>%
    dplyr::slice_head(n = label_n) %>%
    dplyr::pull(.data$feature)

  # Mark top features for labeling
  coef_df <- coef_df %>%
    dplyr::mutate(
      is_top = .data$feature %in% top_features
    )

  # Get optimal lambda values
  lambda_min <- attr(fit, "lambda_min")
  lambda_1se <- attr(fit, "lambda_1se")

  # Create the plot
  p <- ggplot2::ggplot(
    coef_df,
    ggplot2::aes(
      x = lambda,
      y = coefficient,
      group = feature,
      color = is_top
    )
  ) +
    ggplot2::geom_line(
      ggplot2::aes(alpha = is_top, size = is_top)
    ) +
    ggplot2::geom_vline(
      xintercept = lambda_min,
      linetype = "dashed",
      color = "blue"
    ) +
    ggplot2::geom_vline(
      xintercept = lambda_1se,
      linetype = "dashed",
      color = "red"
    ) +
    ggplot2::scale_x_log10() +
    ggplot2::scale_alpha_manual(
      values = c(0.3, 1)
    ) +
    ggplot2::scale_size_manual(
      values = c(0.5, 1.2)
    ) +
    ggplot2::scale_color_manual(
      values = c("gray", "steelblue")
    ) +
    ggplot2::labs(
      title = "Regularization Path",
      subtitle = paste0(
        "Blue: lambda.min, Red: lambda.1se"
      ),
      x = "Lambda (log scale)",
      y = "Coefficients"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(legend.position = "none")

  # Add feature labels at the right end of the plot
  if (label_n > 0) {
    # Get the rightmost lambda value
    rightmost_lambda <- min(fit$lambda)

    # Get coefficients at rightmost lambda for top feats
    label_data <- coef_df %>%
      dplyr::filter(
        .data$is_top,
        .data$lambda == rightmost_lambda
      )

    # Add labels
    p <- p + ggplot2::geom_text(
      data = label_data,
      ggplot2::aes(
        label = feature,
        x = lambda * 1.1,
        y = coefficient
      ),
      hjust = 0, vjust = 0.5, size = 3
    )
  }

  p
}

#' Plot cross-validation results for a regularized model
#'
#' Shows the cross-validation error as a function of
#' lambda for ridge, lasso, or elastic net models fitted
#' with cv.glmnet.
#'
#' @param model A tidylearn regularized model object
#'   (ridge, lasso, or elastic_net)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ ., method = "ridge")
#' tl_plot_regularization_cv(model)
#' }
#' @importFrom ggplot2 ggplot aes geom_point geom_line
#'   geom_ribbon scale_x_log10 labs theme_minimal
#' @export
tl_plot_regularization_cv <- function(model, ...) {
  # Extract the glmnet model
  fit <- model$fit

  # Check if cross-validation results are available
  cv_results <- attr(fit, "cv_results")
  if (is.null(cv_results)) {
    stop(
      "Cross-validation results not available. ",
      "Model must be fitted with lambda = NULL.",
      call. = FALSE
    )
  }

  # Get lambda values and mean cross-validated error
  lambda <- cv_results$lambda
  cvm <- cv_results$cvm
  cvsd <- cv_results$cvsd

  # Create a data frame for plotting
  cv_df <- tibble::tibble(
    lambda = lambda,
    error = cvm,
    error_upper = cvm + cvsd,
    error_lower = cvm - cvsd
  )

  # Get optimal lambda values
  lambda_min <- cv_results$lambda.min
  lambda_1se <- cv_results$lambda.1se

  # Determine y-axis label based on model type
  if (model$spec$is_classification) {
    y_label <- "Binomial Deviance"
  } else {
    y_label <- "Mean Squared Error"
  }

  # Create the plot
  p <- ggplot2::ggplot(
    cv_df,
    ggplot2::aes(x = lambda, y = error)
  ) +
    ggplot2::geom_ribbon(
      ggplot2::aes(
        ymin = error_lower,
        ymax = error_upper
      ),
      alpha = 0.2
    ) +
    ggplot2::geom_point() +
    ggplot2::geom_line() +
    ggplot2::geom_vline(
      xintercept = lambda_min,
      linetype = "dashed",
      color = "blue"
    ) +
    ggplot2::geom_vline(
      xintercept = lambda_1se,
      linetype = "dashed",
      color = "red"
    ) +
    ggplot2::scale_x_log10() +
    ggplot2::labs(
      title = "Cross-Validation Results",
      subtitle = paste0(
        "Blue: lambda.min, Red: lambda.1se"
      ),
      x = "Lambda (log scale)",
      y = y_label
    ) +
    ggplot2::theme_minimal()

  p
}

#' Plot variable importance for a regularized model
#'
#' @param model A tidylearn regularized model object
#' @param lambda Which lambda to use
#'   ("1se" or "min", default: "1se")
#' @param top_n Number of top features to display
#'   (default: 20)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ ., method = "lasso")
#' tl_plot_importance_regularized(model)
#' }
#' @importFrom ggplot2 ggplot aes geom_col coord_flip
#'   labs theme_minimal
#' @export
tl_plot_importance_regularized <- function(model,
                                           lambda = "1se",
                                           top_n = 20,
                                           ...) {
  # Extract the glmnet model
  fit <- model$fit

  # Extract lambda value to use
  if (lambda == "1se") {
    lambda_val <- attr(fit, "lambda_1se")
  } else if (lambda == "min") {
    lambda_val <- attr(fit, "lambda_min")
  } else if (is.numeric(lambda)) {
    lambda_val <- lambda
  } else {
    stop(
      "Invalid lambda specification. ",
      "Use '1se', 'min', or a numeric value.",
      call. = FALSE
    )
  }

  # Get coefficients at selected lambda
  coefs <- as.matrix(coef(fit, s = lambda_val))

  # Exclude intercept
  coefs <- coefs[-1, , drop = FALSE]

  # Create a data frame for plotting
  importance_df <- tibble::tibble(
    feature = rownames(coefs),
    importance = abs(as.vector(coefs))
  ) %>%
    dplyr::arrange(dplyr::desc(.data$importance)) %>%
    dplyr::filter(.data$importance > 0) %>%
    dplyr::slice_head(n = top_n)

  # Create the plot
  p <- ggplot2::ggplot(
    importance_df,
    ggplot2::aes(
      x = stats::reorder(feature, importance),
      y = importance
    )
  ) +
    ggplot2::geom_col(fill = "steelblue") +
    ggplot2::coord_flip() +
    ggplot2::labs(
      title = "Feature Importance",
      subtitle = paste0(
        "Based on coefficient magnitudes at ",
        "lambda = ", round(lambda_val, 5)
      ),
      x = NULL,
      y = "Absolute Coefficient Value"
    ) +
    ggplot2::theme_minimal()

  p
}


#' Predict using a glmnet model
#' @keywords internal
#' @noRd
tl_predict_glmnet <- function(model, new_data,
                              type = "response",
                              ...) {
  fit <- model$fit
  is_classification <- model$spec$is_classification

  # Create model matrix for new data
  formula <- model$spec$formula
  response_var <- all.vars(formula)[1]

  # Get predictor variables from original formula
  all_vars <- all.vars(formula)
  predictor_vars <- all_vars[-1]  # Remove response

  # Create formula manually without using update()
  if (length(predictor_vars) == 0 ||
        (length(predictor_vars) == 1 &&
           predictor_vars[1] == ".")) {
    # Use all predictors except response
    predictor_names <- setdiff(
      names(new_data), response_var
    )
    predictor_formula <- as.formula(paste(
      "~",
      paste(predictor_names, collapse = " + "),
      "- 1"
    ))
  } else {
    predictor_formula <- as.formula(paste(
      "~",
      paste(predictor_vars, collapse = " + "),
      "- 1"
    ))
  }

  # Create model matrix
  x_new <- stats::model.matrix(
    predictor_formula, data = new_data
  )

  # Get optimal lambda value from model
  lambda_val <- attr(fit, "lambda_min")
  if (is.null(lambda_val)) {
    # If no cv was used, use the first lambda
    lambda_val <- fit$lambda[1]
  }

  # Make predictions
  if (is_classification) {
    preds <- predict(
      fit,
      newx = x_new,
      type = "class",
      s = lambda_val
    )
    as.vector(preds)
  } else {
    preds <- predict(
      fit,
      newx = x_new,
      type = "response",
      s = lambda_val
    )
    as.vector(preds)
  }
}

#' @title Advanced Diagnostics Functions for tidylearn
#' @name tidylearn-diagnostics
#' @description Functions for advanced model diagnostics,
#'   assumption checking, and outlier detection
#' @importFrom stats influence.measures cooks.distance hatvalues dffits dfbetas
#' @importFrom stats lm.influence rstudent rstandard
#' @importFrom stats shapiro.test bartlett.test kruskal.test
#' @importFrom dplyr %>% filter select mutate arrange
#' @importFrom ggplot2 ggplot aes geom_point geom_text labs theme_minimal
NULL

#' Calculate influence measures for a linear model
#'
#' @param model A tidylearn model object
#' @param threshold_cook Cook's distance threshold (default: 4/n)
#' @param threshold_leverage Leverage threshold (default: 2*(p+1)/n)
#' @param threshold_dffits DFFITS threshold (default: 2*sqrt((p+1)/n))
#' @return A data frame with one row per observation containing influence
#'   measures: \code{cooks_distance}, \code{leverage}, \code{dffits},
#'   \code{std_residual}, \code{stud_residual}, boolean flags for each
#'   threshold (\code{is_cook_influential}, \code{is_leverage_influential},
#'   \code{is_dffits_influential}, \code{is_outlier}), per-coefficient
#'   \code{dfbetas_*} columns, and an overall \code{is_influential} flag.
#'   Threshold values are stored as attributes.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_influence_measures(model)
#' }
#' @export
tl_influence_measures <- function(model, threshold_cook = NULL,
                                  threshold_leverage = NULL,
                                  threshold_dffits = NULL) {
  # Check if model is supported
  supported_methods <- c(
    "linear", "logistic", "polynomial", "ridge", "lasso", "elastic_net"
  )
  if (!inherits(model, "tidylearn_model") ||
        !model$spec$method %in% supported_methods) {
    stop(
      "Influence measures are only available for linear-based models",
      call. = FALSE
    )
  }

  # Extract fitted model
  fit <- model$fit

  # Data dimensions
  n <- nrow(model$data)
  p <- length(coef(fit)) - 1  # Number of predictors (excluding intercept)

  # Set default thresholds if not provided
  if (is.null(threshold_cook)) threshold_cook <- 4 / n
  if (is.null(threshold_leverage)) {
    threshold_leverage <- 2 * (p + 1) / n
  }
  if (is.null(threshold_dffits)) {
    threshold_dffits <- 2 * sqrt((p + 1) / n)
  }

  # Calculate influence measures
  cooks_d <- cooks.distance(fit)
  leverage <- hatvalues(fit)
  dffits_val <- dffits(fit)
  dfbetas_val <- dfbetas(fit)

  # Get standardized residuals
  std_resid <- rstandard(fit)
  stud_resid <- rstudent(fit)

  # Create data frame
  influence_df <- data.frame(
    observation = 1:n,
    cooks_distance = cooks_d,
    leverage = leverage,
    dffits = dffits_val,
    std_residual = std_resid,
    stud_residual = stud_resid
  )

  # Add flags for influential observations
  influence_df$is_cook_influential <-
    influence_df$cooks_distance > threshold_cook
  influence_df$is_leverage_influential <-
    influence_df$leverage > threshold_leverage
  influence_df$is_dffits_influential <-
    abs(influence_df$dffits) > threshold_dffits
  influence_df$is_outlier <- abs(influence_df$std_residual) > 3

  # Add dfbetas as separate columns
  coef_names <- names(coef(fit))
  for (i in seq_along(coef_names)) {
    col_name <- paste0("dfbetas_", gsub("[^[:alnum:]]", "_", coef_names[i]))
    influence_df[[col_name]] <- dfbetas_val[, i]
  }

  # Add summary column for overall influence
  influence_df$is_influential <- influence_df$is_cook_influential |
    influence_df$is_leverage_influential |
    influence_df$is_dffits_influential |
    influence_df$is_outlier

  # Add threshold values as attributes
  attr(influence_df, "threshold_cook") <- threshold_cook
  attr(influence_df, "threshold_leverage") <- threshold_leverage
  attr(influence_df, "threshold_dffits") <- threshold_dffits

  influence_df
}

#' Plot influence diagnostics
#'
#' @param model A tidylearn model object
#' @param plot_type Type of influence plot: "cook", "leverage", "index"
#' @param threshold_cook Cook's distance threshold (default: 4/n)
#' @param threshold_leverage Leverage threshold (default: 2*(p+1)/n)
#' @param threshold_dffits DFFITS threshold (default: 2*sqrt((p+1)/n))
#' @param n_labels Number of points to label (default: 3)
#' @param label_size Text size for labels (default: 3)
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_plot_influence(model, plot_type = "cook")
#' }
#' @export
tl_plot_influence <- function(model,
                              plot_type = "cook",
                              threshold_cook = NULL,
                              threshold_leverage = NULL,
                              threshold_dffits = NULL,
                              n_labels = 3,
                              label_size = 3) {
  # Get influence measures
  influence_df <- tl_influence_measures(
    model,
    threshold_cook = threshold_cook,
    threshold_leverage = threshold_leverage,
    threshold_dffits = threshold_dffits
  )

  # Get thresholds from attributes
  threshold_cook <- attr(influence_df, "threshold_cook")
  threshold_leverage <- attr(influence_df, "threshold_leverage")
  threshold_dffits <- attr(influence_df, "threshold_dffits")

  # Create plot based on type
  if (plot_type == "cook") {
    # Cook's distance plot
    # Identify top points to label
    n_to_label <- min(n_labels, nrow(influence_df))
    top_idx <- order(
      influence_df$cooks_distance,
      decreasing = TRUE
    )[1:n_to_label]
    influence_df$label <- ifelse(
      influence_df$observation %in% influence_df$observation[top_idx],
      as.character(influence_df$observation),
      ""
    )

    subtitle_text <- paste(
      "Threshold:",
      round(threshold_cook, 4),
      "- Points above are influential"
    )
    p <- ggplot2::ggplot(
      influence_df,
      ggplot2::aes(x = observation, y = cooks_distance)
    ) +
      ggplot2::geom_point(
        ggplot2::aes(color = is_cook_influential),
        size = 3,
        alpha = 0.7
      ) +
      ggplot2::geom_hline(
        yintercept = threshold_cook,
        linetype = "dashed",
        color = "red"
      ) +
      ggplot2::geom_text(
        ggplot2::aes(label = label),
        hjust = -0.3,
        vjust = 0.5,
        size = label_size
      ) +
      ggplot2::scale_color_manual(values = c("blue", "red")) +
      ggplot2::labs(
        title = "Cook's Distance Plot",
        subtitle = subtitle_text,
        x = "Observation Index",
        y = "Cook's Distance",
        color = "Influential"
      ) +
      ggplot2::theme_minimal()

  } else if (plot_type == "leverage") {
    # Leverage-Residual plot (Bubble plot with Cook's distance)
    # Identify top points to label
    n_to_label <- min(n_labels, nrow(influence_df))
    top_idx <- order(
      influence_df$cooks_distance,
      decreasing = TRUE
    )[1:n_to_label]
    influence_df$label <- ifelse(
      influence_df$observation %in% influence_df$observation[top_idx],
      as.character(influence_df$observation),
      ""
    )

    p <- ggplot2::ggplot(
      influence_df,
      ggplot2::aes(
        x = leverage,
        y = std_residual,
        size = cooks_distance,
        color = is_influential
      )
    ) +
      ggplot2::geom_point(alpha = 0.7) +
      ggplot2::geom_hline(
        yintercept = c(-3, 0, 3),
        linetype = "dashed",
        color = c("red", "black", "red")
      ) +
      ggplot2::geom_vline(
        xintercept = threshold_leverage,
        linetype = "dashed",
        color = "red"
      ) +
      ggplot2::geom_text(
        ggplot2::aes(label = label),
        hjust = -0.3,
        vjust = 0.5,
        size = label_size
      ) +
      ggplot2::scale_color_manual(values = c("blue", "red")) +
      ggplot2::labs(
        title = "Leverage-Residual Plot",
        subtitle = paste(
          "Leverage threshold:", round(threshold_leverage, 4),
          "- Residual threshold: +/-3"
        ),
        x = "Leverage (Hat Values)",
        y = "Standardized Residuals",
        size = "Cook's Distance",
        color = "Influential"
      ) +
      ggplot2::theme_minimal()

  } else if (plot_type == "index") {
    # Index plot of residuals
    # Identify top points to label
    n_to_label <- min(n_labels, nrow(influence_df))
    top_idx <- order(
      abs(influence_df$std_residual),
      decreasing = TRUE
    )[1:n_to_label]
    influence_df$label <- ifelse(
      influence_df$observation %in% influence_df$observation[top_idx],
      as.character(influence_df$observation),
      ""
    )

    p <- ggplot2::ggplot(
      influence_df,
      ggplot2::aes(x = observation, y = std_residual)
    ) +
      ggplot2::geom_point(
        ggplot2::aes(color = is_outlier),
        size = 3,
        alpha = 0.7
      ) +
      ggplot2::geom_hline(
        yintercept = c(-3, 0, 3),
        linetype = "dashed",
        color = c("red", "black", "red")
      ) +
      ggplot2::geom_text(
        ggplot2::aes(label = label),
        hjust = -0.3,
        vjust = 0.5,
        size = label_size
      ) +
      ggplot2::scale_color_manual(values = c("blue", "red")) +
      ggplot2::labs(
        title = "Index Plot of Standardized Residuals",
        subtitle = "Points outside +/-3 are considered outliers",
        x = "Observation Index",
        y = "Standardized Residuals",
        color = "Outlier"
      ) +
      ggplot2::theme_minimal()
  } else {
    stop(
      "Invalid plot_type. Use 'cook', 'leverage', or 'index'.",
      call. = FALSE
    )
  }

  p
}

#' Check model assumptions
#'
#' @param model A tidylearn model object
#' @param test Logical; whether to perform statistical tests
#' @param verbose Logical; whether to print test results and explanations
#' @return A named list with one element per assumption checked
#'   (\code{linearity}, \code{independence}, \code{homoscedasticity},
#'   \code{normality}, \code{multicollinearity}, \code{outliers}), each
#'   containing \code{assumption} (character label), \code{check} (logical
#'   or \code{NULL}), \code{details} (character), and
#'   \code{recommendation} (character). An additional \code{overall} element
#'   summarises the number of assumptions checked, violated, and satisfied.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_check_assumptions(model)
#' }
#' @export
tl_check_assumptions <- function(model, test = TRUE, verbose = TRUE) {
  # Check if model is supported
  supported <- c(
    "linear", "logistic", "polynomial", "ridge", "lasso", "elastic_net"
  )
  if (!inherits(model, "tidylearn_model") ||
        !model$spec$method %in% supported) {
    stop(
      "Assumption checking is only available for linear-based models",
      call. = FALSE
    )
  }

  # Extract fitted model and data
  fit <- model$fit
  data <- model$data

  # Get residuals
  residuals <- residuals(fit)
  fitted_values <- fitted(fit)

  # Initialize results list
  assumptions <- list()

  # 1. Linearity
  # Check correlation between fitted values and residuals
  linearity_cor <- cor(fitted_values, residuals)
  linearity_details <- paste(
    "Correlation between fitted values and residuals:",
    round(linearity_cor, 4)
  )
  assumptions$linearity <- list(
    assumption = "Linearity",
    check = abs(linearity_cor) < 0.1,
    details = linearity_details,
    recommendation = if (abs(linearity_cor) >= 0.1) {
      "Consider non-linear transformations or polynomial terms"
    } else {
      "Linearity assumption appears satisfied"
    }
  )

  # 2. Independence
  # Durbin-Watson test for autocorrelation
  if (test && requireNamespace("car", quietly = TRUE)) {
    dw_test <- car::durbinWatsonTest(fit)
    dw_statistic <- dw_test$dw
    dw_recommendation <- if (dw_statistic < 1.5 || dw_statistic > 2.5) {
      paste(
        "Possible autocorrelation in residuals.",
        "Check for time-series structure or clustering."
      )
    } else {
      "Independence assumption appears satisfied"
    }
    assumptions$independence <- list(
      assumption = "Independence",
      check = dw_statistic >= 1.5 && dw_statistic <= 2.5,
      details = paste("Durbin-Watson statistic:", round(dw_statistic, 4)),
      recommendation = dw_recommendation
    )
  } else {
    assumptions$independence <- list(
      assumption = "Independence",
      check = NULL,
      details = "No statistical test performed",
      recommendation = paste(
        "Ensure observations are independent.",
        "Consider the data collection process."
      )
    )
  }

  # 3. Homoscedasticity (Equal Variance)
  # Breusch-Pagan test
  if (test && requireNamespace("lmtest", quietly = TRUE)) {
    bp_test <- lmtest::bptest(fit)
    bp_p_value <- bp_test$p.value
    bp_recommendation <- if (bp_p_value < 0.05) {
      paste(
        "Heteroscedasticity detected.",
        "Consider variance stabilizing transformations or robust SEs."
      )
    } else {
      "Homoscedasticity assumption appears satisfied"
    }
    assumptions$homoscedasticity <- list(
      assumption = "Homoscedasticity",
      check = bp_p_value >= 0.05,
      details = paste("Breusch-Pagan test p-value:", round(bp_p_value, 4)),
      recommendation = bp_recommendation
    )
  } else {
    # Simpler check: correlation between abs(residuals) and fitted values
    het_cor <- cor(abs(residuals), fitted_values)
    het_details <- paste(
      "Correlation between |residuals| and fitted values:",
      round(het_cor, 4)
    )
    het_recommendation <- if (abs(het_cor) >= 0.2) {
      paste(
        "Possible heteroscedasticity.",
        "Consider variance stabilizing transformations or robust SEs."
      )
    } else {
      "Homoscedasticity assumption appears satisfied"
    }
    assumptions$homoscedasticity <- list(
      assumption = "Homoscedasticity",
      check = abs(het_cor) < 0.2,
      details = het_details,
      recommendation = het_recommendation
    )
  }

  # 4. Normality of Residuals
  # Shapiro-Wilk test
  # Shapiro-Wilk limited to 5000 observations
  if (test && length(residuals) <= 5000) {
    sw_test <- shapiro.test(residuals)
    sw_p_value <- sw_test$p.value
    norm_recommendation <- if (sw_p_value < 0.05) {
      paste(
        "Residuals may not be normally distributed.",
        "Consider transformations or robust regression."
      )
    } else {
      "Normality assumption appears satisfied"
    }
    assumptions$normality <- list(
      assumption = "Normality of Residuals",
      check = sw_p_value >= 0.05,
      details = paste("Shapiro-Wilk test p-value:", round(sw_p_value, 4)),
      recommendation = norm_recommendation
    )
  } else {
    # Check skewness and kurtosis
    if (requireNamespace("moments", quietly = TRUE)) {
      skew <- moments::skewness(residuals)
      kurt <- moments::kurtosis(residuals)
      norm_check <- abs(skew) < 0.5 && abs(kurt - 3) < 1
      norm_details <- paste(
        "Skewness:", round(skew, 4),
        "Kurtosis:", round(kurt, 4)
      )
      norm_recommendation <- if (!norm_check) {
        paste(
          "Residuals may not be normally distributed.",
          "Consider transformations or robust regression."
        )
      } else {
        "Normality assumption appears satisfied"
      }
      assumptions$normality <- list(
        assumption = "Normality of Residuals",
        check = norm_check,
        details = norm_details,
        recommendation = norm_recommendation
      )
    } else {
      # Simplified check with quantiles
      q_norm <- qqnorm(residuals, plot = FALSE)
      cor_norm <- cor(q_norm$x, q_norm$y)
      qq_recommendation <- if (cor_norm < 0.98) {
        paste(
          "Residuals may not be normally distributed.",
          "Consider transformations or robust regression."
        )
      } else {
        "Normality assumption appears satisfied"
      }
      assumptions$normality <- list(
        assumption = "Normality of Residuals",
        check = cor_norm >= 0.98,
        details = paste("QQ-plot correlation:", round(cor_norm, 4)),
        recommendation = qq_recommendation
      )
    }
  }

  # 5. Multicollinearity
  if (requireNamespace("car", quietly = TRUE)) {
    # Attempt to calculate VIF
    tryCatch({
      vif_values <- car::vif(fit)
      max_vif <- max(vif_values)
      vif_recommendation <- if (max_vif >= 5) {
        paste(
          "Multicollinearity detected.",
          "Consider removing or combining highly correlated predictors."
        )
      } else {
        "No serious multicollinearity detected"
      }
      assumptions$multicollinearity <- list(
        assumption = "No Multicollinearity",
        check = max_vif < 5,
        details = paste("Maximum VIF:", round(max_vif, 4)),
        recommendation = vif_recommendation
      )
    }, error = function(e) {
      # If VIF calculation fails, use correlation matrix
      if (verbose) {
        message("VIF calculation failed. Checking correlations instead.")
      }

      # Create correlation matrix of predictors
      formula <- model$spec$formula
      predictor_vars <- all.vars(formula)[-1]  # Remove response variable

      if (length(predictor_vars) > 1) {
        pred_data <- stats::model.matrix(formula, data)[, -1]
        cor_matrix <- cor(pred_data)
        max_cor <- max(abs(cor_matrix[upper.tri(cor_matrix)]))

        cor_details <- paste(
          "Maximum correlation between predictors:",
          round(max_cor, 4)
        )
        cor_recommendation <- if (max_cor >= 0.7) {
          paste(
            "Potential multicollinearity.",
            "Consider removing or combining correlated predictors."
          )
        } else {
          "No serious multicollinearity detected"
        }
        assumptions$multicollinearity <- list(
          assumption = "No Multicollinearity",
          check = max_cor < 0.7,
          details = cor_details,
          recommendation = cor_recommendation
        )
      } else {
        assumptions$multicollinearity <- list(
          assumption = "No Multicollinearity",
          check = TRUE,
          details = "Model has only one predictor",
          recommendation = "Not applicable with a single predictor"
        )
      }
    })
  } else {
    # If car package not available, use correlation matrix
    formula <- model$spec$formula
    predictor_vars <- all.vars(formula)[-1]

    if (length(predictor_vars) > 1) {
      pred_data <- stats::model.matrix(formula, data)[, -1]
      cor_matrix <- cor(pred_data)
      max_cor <- max(abs(cor_matrix[upper.tri(cor_matrix)]))

      cor_details2 <- paste(
        "Maximum correlation between predictors:",
        round(max_cor, 4)
      )
      cor_recommendation2 <- if (max_cor >= 0.7) {
        paste(
          "Potential multicollinearity.",
          "Consider removing or combining correlated predictors."
        )
      } else {
        "No serious multicollinearity detected"
      }
      assumptions$multicollinearity <- list(
        assumption = "No Multicollinearity",
        check = max_cor < 0.7,
        details = cor_details2,
        recommendation = cor_recommendation2
      )
    } else {
      assumptions$multicollinearity <- list(
        assumption = "No Multicollinearity",
        check = TRUE,
        details = "Model has only one predictor",
        recommendation = "Not applicable with a single predictor"
      )
    }
  }

  # 6. Outliers and Influential Points
  influence_df <- tl_influence_measures(model)
  n_influential <- sum(influence_df$is_influential)

  outlier_recommendation <- if (n_influential > 0) {
    influential_obs <- influence_df$observation[influence_df$is_influential]
    obs_to_show <- influential_obs[1:min(5, n_influential)]
    paste(
      "Consider inspecting observations:",
      paste(obs_to_show, collapse = ", "),
      "... (and potentially others)"
    )
  } else {
    "No influential outliers detected"
  }
  assumptions$outliers <- list(
    assumption = "No Influential Outliers",
    check = n_influential == 0,
    details = paste(n_influential, "influential observations detected"),
    recommendation = outlier_recommendation
  )

  # Print summary if verbose
  if (verbose) {
    message("Model Assumptions Check Summary:")
    message("--------------------------------")
    for (name in names(assumptions)) {
      check <- assumptions[[name]]
      check_status <- if (is.null(check$check)) {
        "UNKNOWN"
      } else if (check$check) {
        "SATISFIED"
      } else {
        "VIOLATED"
      }

      message(
        check$assumption, ": ", check_status, "\n",
        "  Details: ", check$details, "\n",
        "  Recommendation: ", check$recommendation, "\n"
      )
    }
  }

  # Add overall assessment
  checks <- Filter(
    Negate(is.null),
    lapply(assumptions, function(x) x$check)
  )
  checks <- unlist(checks)

  if (length(checks) > 0) {
    overall_status <- if (all(checks)) {
      "All checked assumptions appear to be satisfied."
    } else {
      n_violated <- sum(!checks)
      paste(
        n_violated,
        "assumption(s) appear to be violated. See details."
      )
    }
  } else {
    overall_status <- "Could not perform complete assumption checks."
  }

  assumptions$overall <- list(
    status = overall_status,
    n_checked = length(checks),
    n_violated = sum(!checks),
    n_satisfied = sum(checks)
  )

  assumptions
}

#' Create a comprehensive diagnostic dashboard
#'
#' @param model A tidylearn model object
#' @param include_influence Logical; whether to include influence diagnostics
#' @param include_assumptions Logical; whether to include assumption checks
#' @param include_performance Logical; whether to include performance metrics
#' @param arrange_plots Layout arrangement (e.g., "grid", "row", "column")
#' @return A \code{\link[gridExtra]{grid.arrange}} object (a
#'   \code{\link[grid]{grob}}) containing the arranged diagnostic plots.
#' @examples
#' \donttest{
#' if (requireNamespace("gridExtra")) {
#'   model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#'   tl_diagnostic_dashboard(model)
#' }
#' }
#' @export
tl_diagnostic_dashboard <- function(model, include_influence = TRUE,
                                    include_assumptions = TRUE,
                                    include_performance = TRUE,
                                    arrange_plots = "grid") {

  # Check package dependencies
  if (!requireNamespace("gridExtra", quietly = TRUE)) {
    stop(
      "Package 'gridExtra' is required for creating dashboards",
      call. = FALSE
    )
  }

  # Initialize plots list
  plots <- list()

  # Add basic diagnostic plots
  plots$residuals_vs_fitted <- tl_plot_residuals(model, type = "fitted")
  plots$residual_hist <- tl_plot_residuals(model, type = "histogram")
  plots$qq_plot <- ggplot2::ggplot(
    data.frame(residuals = rstandard(model$fit)),
    ggplot2::aes(sample = residuals)
  ) +
    ggplot2::stat_qq() +
    ggplot2::stat_qq_line(color = "red") +
    ggplot2::labs(
      title = "Normal Q-Q Plot",
      x = "Theoretical Quantiles",
      y = "Sample Quantiles"
    ) +
    ggplot2::theme_minimal()

  # Add scale-location plot
  diag_plots <- tl_plot_diagnostics(model, which = 3)
  plots$scale_location <- diag_plots[["scale_location"]]

  # Add influence plots if requested
  if (include_influence) {
    plots$cook_distance <- tl_plot_influence(model, plot_type = "cook")
    plots$leverage_plot <- tl_plot_influence(model, plot_type = "leverage")
  }

  # Add assumption check if requested
  if (include_assumptions) {
    assumptions <- tl_check_assumptions(model, verbose = FALSE)

    # Create plot with assumption check results
    assumption_results <- data.frame(
      Assumption = sapply(assumptions[1:5], function(x) x$assumption),
      Status = sapply(assumptions[1:5], function(x) {
        if (is.null(x$check)) return("Unknown")
        ifelse(x$check, "Satisfied", "Violated")
      }),
      Details = sapply(assumptions[1:5], function(x) x$details)
    )

    # Create a textual summary plot
    status_colors <- c(
      "Satisfied" = "green",
      "Violated" = "red",
      "Unknown" = "gray"
    )
    plots$assumptions <- ggplot2::ggplot(
      assumption_results,
      ggplot2::aes(x = 1, y = Assumption, fill = Status)
    ) +
      ggplot2::geom_tile() +
      ggplot2::geom_text(
        ggplot2::aes(label = Details),
        hjust = 0,
        x = 1.05
      ) +
      ggplot2::scale_fill_manual(values = status_colors) +
      ggplot2::labs(title = "Assumption Check Results") +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        axis.title = ggplot2::element_blank(),
        axis.text.x = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_blank(),
        panel.grid = ggplot2::element_blank()
      ) +
      ggplot2::coord_cartesian(xlim = c(0.5, 2.5))
  }

  # Add performance metrics if requested
  if (include_performance) {
    # Calculate model metrics
    metrics <- tl_evaluate(model)

    # Create performance summary plot
    plots$performance <- ggplot2::ggplot(
      metrics,
      ggplot2::aes(x = metric, y = value)
    ) +
      ggplot2::geom_col(fill = "steelblue") +
      ggplot2::geom_text(
        ggplot2::aes(label = round(value, 3)),
        vjust = -0.5
      ) +
      ggplot2::labs(title = "Model Performance Metrics") +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        axis.title = ggplot2::element_blank(),
        axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)
      )
  }

  # Arrange plots based on specified layout
  if (arrange_plots == "grid") {
    # Determine grid dimensions
    n_plots <- length(plots)
    n_cols <- min(3, n_plots)
    n_rows <- ceiling(n_plots / n_cols)

    # Arrange in grid
    combined_plot <- gridExtra::grid.arrange(
      grobs = plots,
      ncol = n_cols,
      nrow = n_rows
    )
  } else if (arrange_plots == "row") {
    # Arrange in a single row
    combined_plot <- gridExtra::grid.arrange(
      grobs = plots,
      ncol = length(plots)
    )
  } else if (arrange_plots == "column") {
    # Arrange in a single column
    combined_plot <- gridExtra::grid.arrange(
      grobs = plots,
      nrow = length(plots)
    )
  } else {
    stop(
      "Invalid arrange_plots value. Use 'grid', 'row', or 'column'.",
      call. = FALSE
    )
  }

  combined_plot
}

#' Detect outliers in the data
#'
#' @param data A data frame containing the data
#' @param variables Character vector of variables to check for outliers
#' @param method Method for outlier detection: "boxplot", "z-score", "cook",
#'   "iqr", "mahalanobis"
#' @param threshold Threshold for outlier detection
#' @param plot Logical; whether to create a plot of outliers
#' @return A list with outlier detection results:
#'   \describe{
#'     \item{method}{The detection method used (character).}
#'     \item{method_name}{Human-readable method name (character).}
#'     \item{threshold}{The threshold value used (numeric).}
#'     \item{threshold_label}{Formatted threshold description (character).}
#'     \item{outlier_flags}{A logical matrix (observations x variables).}
#'     \item{any_outlier}{Logical vector indicating if each observation is an
#'       outlier in any variable.}
#'     \item{outlier_counts}{List with \code{total}, \code{by_variable}, and
#'       \code{by_observation} counts.}
#'     \item{outlier_indices}{Integer vector of outlier row indices.}
#'     \item{plot}{A \code{\link[ggplot2]{ggplot}} object, or \code{NULL} if
#'       \code{plot = FALSE}.}
#'   }
#' @examples
#' \donttest{
#' tl_detect_outliers(mtcars, variables = c("mpg", "wt"), method = "iqr")
#' }
#' @export
tl_detect_outliers <- function(data, variables = NULL, method = "iqr",
                               threshold = NULL, plot = TRUE) {
  # Handle variables selection
  if (is.null(variables)) {
    # Select only numeric variables
    variables <- names(data)[sapply(data, is.numeric)]
    if (length(variables) == 0) {
      stop("No numeric variables found in the data", call. = FALSE)
    }
  } else {
    # Check if specified variables exist and are numeric
    for (var in variables) {
      if (!var %in% names(data)) {
        stop("Variable not found in data: ", var, call. = FALSE)
      }
      if (!is.numeric(data[[var]])) {
        stop("Variable is not numeric: ", var, call. = FALSE)
      }
    }
  }

  # Extract data for selected variables
  var_data <- data[, variables, drop = FALSE]

  # Set default threshold based on method
  if (is.null(threshold)) {
    threshold <- switch(
      method,
      "boxplot" = 1.5, # IQR multiplier
      "z-score" = 3, # Standard deviations
      "cook" = 4 / nrow(data),
      "iqr" = 1.5,
      "mahalanobis" = 0.975,
      2 # Default multiplier
    )
  }

  # Detect outliers based on method
  if (method == "boxplot" || method == "iqr") {
    # IQR method for each variable
    outlier_flags <- sapply(variables, function(var) {
      q1 <- stats::quantile(var_data[[var]], 0.25, na.rm = TRUE)
      q3 <- stats::quantile(var_data[[var]], 0.75, na.rm = TRUE)
      iqr <- q3 - q1
      lower_bound <- q1 - threshold * iqr
      upper_bound <- q3 + threshold * iqr
      var_data[[var]] < lower_bound | var_data[[var]] > upper_bound
    })

    method_name <- "Interquartile Range (IQR)"
    threshold_label <- paste0("IQR multiplier: ", threshold)

  } else if (method == "z-score") {
    # Z-score method for each variable
    outlier_flags <- sapply(variables, function(var) {
      z_scores <- abs(scale(var_data[[var]]))
      z_scores > threshold
    })

    method_name <- "Z-Score"
    threshold_label <- paste0("Standard deviations: ", threshold)

  } else if (method == "cook") {
    # Cook's distance method (requires fitting a model)
    # First need to create a model formula
    if (length(variables) < 2) {
      stop(
        "Cook's distance method requires at least 2 variables",
        call. = FALSE
      )
    }

    # Use first variable as response
    formula <- stats::as.formula(
      paste(variables[1], "~", paste(variables[-1], collapse = " + "))
    )

    # Fit linear model
    model <- stats::lm(formula, data = data)

    # Calculate Cook's distance
    cooks_d <- stats::cooks.distance(model)

    # Create flags (only available for the whole observation, not by variable)
    outlier_flags <- matrix(
      cooks_d > threshold,
      nrow = nrow(data),
      ncol = length(variables)
    )
    colnames(outlier_flags) <- variables

    method_name <- "Cook's Distance"
    threshold_label <- paste0("Threshold: ", threshold)

  } else if (method == "mahalanobis") {
    # Mahalanobis distance for multivariate outlier detection
    if (length(variables) < 2) {
      stop(
        "Mahalanobis distance requires at least 2 variables",
        call. = FALSE
      )
    }

    # Calculate center (means) and covariance matrix
    center <- colMeans(var_data, na.rm = TRUE)
    cov_matrix <- stats::cov(var_data, use = "pairwise.complete.obs")

    # Calculate Mahalanobis distances
    mahal_dist <- stats::mahalanobis(var_data, center, cov_matrix)

    # Convert threshold to chi-square quantile
    chi2_quantile <- stats::qchisq(threshold, df = length(variables))

    # Flag outliers (same for all variables)
    outlier_flags <- matrix(
      mahal_dist > chi2_quantile,
      nrow = nrow(data),
      ncol = length(variables)
    )
    colnames(outlier_flags) <- variables

    method_name <- "Mahalanobis Distance"
    threshold_label <- paste0(
      "Chi-square quantile (p = ", threshold,
      ", df = ", length(variables), "): ", round(chi2_quantile, 2)
    )

  } else {
    stop(
      "Invalid method. Use 'boxplot', 'z-score',",
      " 'cook', 'iqr', or 'mahalanobis'.",
      call. = FALSE
    )
  }

  # Combine flags across variables
  any_outlier <- apply(outlier_flags, 1, any)

  # Count outliers
  outlier_counts <- list(
    total = sum(any_outlier),
    by_variable = colSums(outlier_flags, na.rm = TRUE),
    by_observation = rowSums(outlier_flags, na.rm = TRUE)
  )

  # Create plot if requested
  if (plot) {
    if (method == "mahalanobis" && length(variables) >= 2) {
      # For Mahalanobis, create scatter plot matrix with outliers highlighted
      # Create a data frame with outlier flags
      plot_data <- var_data
      plot_data$is_outlier <- any_outlier

      # Create pairs plot
      if (requireNamespace("GGally", quietly = TRUE)) {
        outlier_plot <- GGally::ggpairs(
          plot_data,
          columns = seq_along(variables),
          aes(color = is_outlier),
          progress = FALSE
        ) +
          ggplot2::scale_color_manual(values = c("blue", "red")) +
          ggplot2::labs(
            title = paste("Outlier Detection using", method_name),
            subtitle = threshold_label
          )
      } else {
        # Fallback to basic scatter plot of first two variables
        outlier_plot <- ggplot2::ggplot(
          plot_data,
          ggplot2::aes(
            x = .data[[variables[1]]],
            y = .data[[variables[2]]],
            color = is_outlier
          )
        ) +
          ggplot2::geom_point() +
          ggplot2::scale_color_manual(values = c("blue", "red")) +
          ggplot2::labs(
            title = paste("Outlier Detection using", method_name),
            subtitle = threshold_label,
            x = variables[1],
            y = variables[2],
            color = "Outlier"
          ) +
          ggplot2::theme_minimal()
      }
    } else if (method == "cook") {
      # For Cook's distance, create index plot
      plot_data <- data.frame(
        observation = seq_along(cooks_d),
        cooks_distance = cooks_d,
        is_outlier = cooks_d > threshold
      )

      outlier_plot <- ggplot2::ggplot(
        plot_data,
        ggplot2::aes(
          x = observation,
          y = cooks_distance,
          color = is_outlier
        )
      ) +
        ggplot2::geom_point() +
        ggplot2::geom_hline(
          yintercept = threshold,
          linetype = "dashed",
          color = "red"
        ) +
        ggplot2::scale_color_manual(values = c("blue", "red")) +
        ggplot2::labs(
          title = paste("Outlier Detection using", method_name),
          subtitle = threshold_label,
          x = "Observation Index",
          y = "Cook's Distance",
          color = "Outlier"
        ) +
        ggplot2::theme_minimal()
    } else {
      # For other methods, create boxplots with outliers highlighted
      # Prepare data for plotting
      plot_data <- tidyr::pivot_longer(
        var_data,
        cols = dplyr::everything(),
        names_to = "variable",
        values_to = "value"
      )

      # Add outlier flag
      plot_data$is_outlier <- FALSE
      for (i in seq_len(nrow(plot_data))) {
        var_idx <- match(plot_data$variable[i], variables)
        obs_idx <- (i - 1) %% nrow(data) + 1
        plot_data$is_outlier[i] <- outlier_flags[obs_idx, var_idx]
      }

      outlier_plot <- ggplot2::ggplot(
        plot_data,
        ggplot2::aes(
          x = variable,
          y = value,
          fill = is_outlier
        )
      ) +
        ggplot2::geom_boxplot(outlier.shape = NA) +
        ggplot2::geom_jitter(
          ggplot2::aes(color = is_outlier),
          width = 0.2,
          alpha = 0.7
        ) +
        ggplot2::scale_fill_manual(values = c("lightblue", "lightpink")) +
        ggplot2::scale_color_manual(values = c("blue", "red")) +
        ggplot2::labs(
          title = paste("Outlier Detection using", method_name),
          subtitle = threshold_label,
          x = "Variable",
          y = "Value",
          fill = "Outlier",
          color = "Outlier"
        ) +
        ggplot2::theme_minimal() +
        ggplot2::theme(
          axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)
        )
    }
  } else {
    outlier_plot <- NULL
  }

  # Create output list
  result <- list(
    method = method,
    method_name = method_name,
    threshold = threshold,
    threshold_label = threshold_label,
    outlier_flags = outlier_flags,
    any_outlier = any_outlier,
    outlier_counts = outlier_counts,
    outlier_indices = which(any_outlier),
    plot = outlier_plot
  )

  result
}

# ---- Diagnostics functions ----

# -- Influence measures --

test_that("tl_influence_measures returns data frame for linear model", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  inf <- tl_influence_measures(model)

  expect_s3_class(inf, "data.frame")
  expect_equal(nrow(inf), nrow(mtcars))
  expect_true("cooks_distance" %in% names(inf))
  expect_true("leverage" %in% names(inf))
  expect_true("dffits" %in% names(inf))
  expect_true("is_influential" %in% names(inf))
})

test_that("tl_influence_measures respects custom thresholds", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

  # Very strict threshold should flag more observations
  strict <- tl_influence_measures(model, threshold_cook = 0.001)
  # Very lenient threshold should flag fewer
  lenient <- tl_influence_measures(model, threshold_cook = 10)

  expect_gte(
    sum(strict$is_cook_influential),
    sum(lenient$is_cook_influential)
  )
})

test_that("tl_influence_measures errors for unsupported methods", {
  skip_if_not_installed("randomForest")

  model <- tl_model(iris, Species ~ ., method = "forest")

  expect_error(tl_influence_measures(model), "linear-based")
})

test_that("tl_influence_measures includes dfbetas columns", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  inf <- tl_influence_measures(model)

  # Should have dfbetas for intercept, wt, hp
  dfbetas_cols <- grep("^dfbetas_", names(inf), value = TRUE)
  expect_true(length(dfbetas_cols) >= 3)
})

test_that("tl_influence_measures stores thresholds as attributes", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  inf <- tl_influence_measures(model)

  expect_true(!is.null(attr(inf, "threshold_cook")))
  expect_true(!is.null(attr(inf, "threshold_leverage")))
  expect_true(!is.null(attr(inf, "threshold_dffits")))
})

# -- Influence plotting --

test_that("tl_plot_influence returns ggplot for cook type", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  p <- tl_plot_influence(model, plot_type = "cook")

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_influence returns ggplot for leverage type", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  p <- tl_plot_influence(model, plot_type = "leverage")

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_influence returns ggplot for index type", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  p <- tl_plot_influence(model, plot_type = "index")

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_influence errors for invalid plot type", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

  expect_error(tl_plot_influence(model, plot_type = "invalid"), "Invalid")
})

# -- Assumption checking --

test_that("tl_check_assumptions returns list for linear model", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  assumptions <- tl_check_assumptions(model, test = FALSE, verbose = FALSE)

  expect_type(assumptions, "list")
  expect_true("linearity" %in% names(assumptions))
  expect_true("normality" %in% names(assumptions))
  expect_true("homoscedasticity" %in% names(assumptions))
  expect_true("multicollinearity" %in% names(assumptions))
  expect_true("outliers" %in% names(assumptions))
  expect_true("overall" %in% names(assumptions))
})

test_that("tl_check_assumptions each check has standard structure", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  assumptions <- tl_check_assumptions(model, test = FALSE, verbose = FALSE)

  # Each assumption (except overall) should have assumption, check,
  # details, recommendation
  for (name in setdiff(names(assumptions), "overall")) {
    check <- assumptions[[name]]
    expect_true("assumption" %in% names(check),
                label = paste(name, "has assumption field"))
    expect_true("details" %in% names(check),
                label = paste(name, "has details field"))
    expect_true("recommendation" %in% names(check),
                label = paste(name, "has recommendation field"))
  }
})

test_that("tl_check_assumptions overall has correct counts", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  assumptions <- tl_check_assumptions(model, test = FALSE, verbose = FALSE)

  overall <- assumptions$overall
  expect_true("n_checked" %in% names(overall))
  expect_true("n_violated" %in% names(overall))
  expect_true("n_satisfied" %in% names(overall))
  expect_equal(overall$n_checked, overall$n_violated + overall$n_satisfied)
})

test_that("tl_check_assumptions errors for unsupported methods", {
  skip_if_not_installed("randomForest")

  model <- tl_model(iris, Species ~ ., method = "forest")

  expect_error(tl_check_assumptions(model), "linear-based")
})

test_that("tl_check_assumptions works with statistical tests when available", {
  skip_if_not_installed("car")
  skip_if_not_installed("lmtest")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  assumptions <- tl_check_assumptions(model, test = TRUE, verbose = FALSE)

  # Should include Durbin-Watson and Breusch-Pagan results
  expect_true(grepl("Durbin-Watson", assumptions$independence$details))
  expect_true(grepl("Breusch-Pagan", assumptions$homoscedasticity$details))
})

test_that("tl_check_assumptions works for polynomial model", {
  model <- tl_model(mtcars, mpg ~ wt, method = "polynomial", degree = 2)
  assumptions <- tl_check_assumptions(model, test = FALSE, verbose = FALSE)

  expect_type(assumptions, "list")
  expect_true("linearity" %in% names(assumptions))
})

# -- Outlier detection --

test_that("tl_detect_outliers works with IQR method", {
  result <- tl_detect_outliers(iris, variables = c("Sepal.Length"),
                               method = "iqr", plot = FALSE)

  expect_type(result, "list")
  expect_true("outlier_flags" %in% names(result))
  expect_true("outlier_counts" %in% names(result))
  expect_true("outlier_indices" %in% names(result))
  expect_equal(result$method, "iqr")
})

test_that("tl_detect_outliers works with z-score method", {
  result <- tl_detect_outliers(iris, variables = c("Sepal.Length"),
                               method = "z-score", plot = FALSE)

  expect_type(result, "list")
  expect_equal(result$method, "z-score")
})

test_that("tl_detect_outliers works with cook method", {
  result <- tl_detect_outliers(
    mtcars,
    variables = c("mpg", "wt", "hp"),
    method = "cook", plot = FALSE
  )

  expect_type(result, "list")
  expect_equal(result$method, "cook")
})

test_that("tl_detect_outliers works with mahalanobis method", {
  result <- tl_detect_outliers(
    iris,
    variables = c("Sepal.Length", "Sepal.Width"),
    method = "mahalanobis", plot = FALSE
  )

  expect_type(result, "list")
  expect_equal(result$method, "mahalanobis")
})

test_that("tl_detect_outliers auto-selects numeric variables", {
  result <- tl_detect_outliers(iris, method = "iqr", plot = FALSE)

  expect_type(result, "list")
  # Should have used all 4 numeric columns
  expect_equal(ncol(result$outlier_flags), 4)
})

test_that("tl_detect_outliers errors for non-numeric variable", {
  expect_error(
    tl_detect_outliers(iris, variables = "Species", plot = FALSE),
    "not numeric"
  )
})

test_that("tl_detect_outliers errors for nonexistent variable", {
  expect_error(
    tl_detect_outliers(iris, variables = "nonexistent", plot = FALSE),
    "not found"
  )
})

test_that("tl_detect_outliers respects threshold parameter", {
  strict <- tl_detect_outliers(iris, method = "iqr",
                               threshold = 0.5, plot = FALSE)
  lenient <- tl_detect_outliers(iris, method = "iqr",
                                threshold = 3.0, plot = FALSE)

  # Stricter threshold should find more outliers
  expect_gte(strict$outlier_counts$total, lenient$outlier_counts$total)
})

test_that("tl_detect_outliers creates plot when requested", {
  result <- tl_detect_outliers(iris, variables = c("Sepal.Length"),
                               method = "iqr", plot = TRUE)

  expect_true(!is.null(result$plot))
  expect_s3_class(result$plot, "ggplot")
})

test_that("tl_detect_outliers returns NULL plot when plot = FALSE", {
  result <- tl_detect_outliers(iris, method = "iqr", plot = FALSE)

  expect_null(result$plot)
})

test_that("tl_detect_outliers errors for invalid method", {
  expect_error(
    tl_detect_outliers(iris, method = "invalid", plot = FALSE),
    "Invalid method"
  )
})

test_that("tl_detect_outliers cook method requires 2+ variables", {
  expect_error(
    tl_detect_outliers(iris, variables = "Sepal.Length",
                       method = "cook", plot = FALSE),
    "at least 2"
  )
})

test_that("tl_detect_outliers mahalanobis requires 2+ variables", {
  expect_error(
    tl_detect_outliers(iris, variables = "Sepal.Length",
                       method = "mahalanobis", plot = FALSE),
    "at least 2"
  )
})

# -- Diagnostic dashboard --

test_that("tl_diagnostic_dashboard errors without gridExtra", {
  skip_if(requireNamespace("gridExtra", quietly = TRUE),
          "gridExtra is installed, cannot test missing-package path")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  expect_error(tl_diagnostic_dashboard(model), "gridExtra")
})

# -- Classification auto-detection fix --

test_that("numeric response with few unique values is treated as regression", {
  # Create data with numeric response having <= 10 unique values
  set.seed(42)
  data <- data.frame(
    y = sample(1:5, 50, replace = TRUE),
    x1 = rnorm(50),
    x2 = rnorm(50)
  )

  # Should be treated as regression (not classification) since y is numeric
  expect_message(
    model <- tl_model(data, y ~ x1 + x2, method = "linear"),
    "unique numeric values"
  )
  expect_false(model$spec$is_classification)
})

test_that("factor response is treated as classification regardless of levels", {
  data <- data.frame(
    y = factor(rep(c("A", "B"), 25)),
    x1 = rnorm(50)
  )

  model <- tl_model(data, y ~ x1, method = "logistic")
  expect_true(model$spec$is_classification)
})

#' Utility functions for tidylearn
#' @keywords internal
#' @importFrom stats aov coef cor fitted median qqnorm
#'   reorder residuals runif sd setNames terms update var
#' @importFrom utils combn getFromNamespace head packageVersion
#' @noRd

# Suppress R CMD check notes about global variables from tidyverse NSE
utils::globalVariables(c(
  ".", ".id", ".obs_id", ".row_id", ":=",
  "Actual", "Assumption", "Details", "Freq",
  "Predicted", "SE.sim", "Status",
  "abs_shap_value", "actual", "all_of",
  "avg_sil_width", "cluster", "cluster_label",
  "coefficient", "component",
  "conf_lower", "conf_upper", "confidence",
  "cooks_distance", "cost", "cum_variance",
  "decay", "decile", "distance",
  "epoch", "error", "error_lower", "error_upper",
  "feature", "feature_value", "fold",
  "fpr", "frac_pos", "gap",
  "id1", "id2", "interaction_value", "is_best",
  "is_cook_influential", "is_core",
  "is_influential", "is_noise", "is_outlier",
  "is_top", "k", "knn_dist", "label", "lambda",
  "leverage", "lhs", "lift", "loading",
  "mean_pred_prob", "mean_value", "metric",
  "model", "n", "neighbor", "obs_id",
  "observation", "pc_num", "percentage",
  "pred", "pred_lower", "pred_upper",
  "predicted", "prop_variance",
  "residuals", "rhs", "score", "shap_value",
  "sil_width", "size", "sqrt_abs_residuals",
  "std_residual", "support", "tl_plot_model",
  "tl_plot_unsupervised",
  "tl_prediction_intervals", "tot_withinss",
  "tpr", "value", "var_value", "variable",
  "variance", "where", "x", "x_end",
  "y", "y_end", "abs_estimate", "estimate",
  "p_value", "significant", "std_error", "term"
))


# Null-coalescing operator
`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

#' Extract numeric columns from data
#' @keywords internal
#' @noRd
get_numeric_cols <- function(data, cols = NULL) {
  if (!is.null(cols)) {
    cols_enquo <- rlang::enquo(cols)
    data %>% dplyr::select(!!cols_enquo)
  } else {
    data %>% dplyr::select(where(is.numeric))
  }
}

#' Extract response variable from formula
#' @keywords internal
#' @noRd
extract_response <- function(formula, data) {
  if (is.null(formula)) {
    return(NULL)
  }

  vars <- all.vars(formula)
  if (length(vars) == 0) {
    return(NULL)
  }

  response_var <- vars[1]
  if (response_var %in% names(data)) {
    return(data[[response_var]])
  }
  NULL
}

#' Create observation IDs
#' @keywords internal
#' @noRd
create_obs_ids <- function(data) {
  has_names <- !is.null(rownames(data)) &&
    !all(rownames(data) == as.character(
      seq_len(nrow(data))
    ))
  if (has_names) {
    return(rownames(data))
  }
  paste0("obs_", seq_len(nrow(data)))
}

#' Validate data for modeling
#' @keywords internal
#' @noRd
validate_data <- function(data, allow_missing = FALSE) {
  if (!is.data.frame(data)) {
    stop("data must be a data frame or tibble",
         call. = FALSE)
  }

  if (nrow(data) == 0) {
    stop("data has no rows", call. = FALSE)
  }

  if (!allow_missing && any(is.na(data))) {
    warning(
      "Missing values detected in data. ",
      "Consider imputation or removing missing values."
    )
  }

  invisible(TRUE)
}

#' Safe extraction of formula variables
#' @keywords internal
#' @noRd
get_formula_vars <- function(formula, data) {
  if (is.null(formula)) {
    return(names(data)[sapply(data, is.numeric)])
  }

  # Check if it's a one-sided formula (unsupervised)
  if (length(formula) == 2) {
    # One-sided: ~ vars
    rhs <- formula[[2]]
    if (rhs == ".") {
      names(data)[sapply(data, is.numeric)]
    } else {
      all.vars(formula)
    }
  } else {
    # Two-sided: response ~ predictors
    vars <- all.vars(formula)
    vars[-1]  # Exclude response
  }
}

#' Validate that a file path exists
#' @keywords internal
#' @noRd
tl_validate_file_path <- function(path) {
  if (!is.character(path) || length(path) != 1) {
    stop("'path' must be a single character string",
         call. = FALSE)
  }
  if (!file.exists(path)) {
    stop("File not found: '", path, "'",
         call. = FALSE)
  }
  invisible(TRUE)
}

#' Check if required packages are installed
#' @keywords internal
#' @noRd
tl_check_packages <- function(...) {
  packages <- c(...)

  for (pkg in packages) {
    if (!requireNamespace(pkg, quietly = TRUE)) {
      stop("Package '", pkg, "' is required but ",
           "not installed. ",
           "Please install it with: install.packages('",
           pkg, "')",
           call. = FALSE)
    }
  }

  invisible(TRUE)
}

#' Integration Functions: Combining Supervised and Unsupervised Learning
#'
#' These functions demonstrate the power of tidylearn's unified approach by
#' seamlessly integrating supervised and unsupervised learning techniques.

#' Feature Engineering via Dimensionality Reduction
#'
#' Use PCA, MDS, or other dimensionality reduction as a preprocessing step
#' for supervised learning. This can improve model performance
#' and interpretability.
#'
#' @param data A data frame
#' @param response Response variable name (will be preserved)
#' @param method Dimensionality reduction method: "pca", "mds"
#' @param n_components Number of components to retain
#' @param ... Additional arguments for the dimensionality reduction method
#' @return A list with components:
#'   \describe{
#'     \item{data}{The transformed data frame with reduced-dimension columns
#'       and the response variable (if provided).}
#'     \item{reduction_model}{The fitted tidylearn dimensionality reduction
#'       model.}
#'     \item{original_data}{The original input data frame.}
#'     \item{response}{The response variable name, or \code{NULL}.}
#'   }
#' @export
#' @examples
#' \donttest{
#' # Reduce dimensions before classification
#' reduced <- tl_reduce_dimensions(
#'   iris, response = "Species",
#'   method = "pca", n_components = 3
#' )
#' model <- tl_model(reduced$data, Species ~ ., method = "tree")
#' }
tl_reduce_dimensions <- function(data,
                                 response = NULL,
                                 method = "pca",
                                 n_components = NULL,
                                 ...) {
  # Separate response if provided
  if (!is.null(response)) {
    if (!response %in% names(data)) {
      stop(
        "Response variable '", response,
        "' not found in data", call. = FALSE
      )
    }
    response_data <- data[[response]]
    predictor_data <- data %>% dplyr::select(-dplyr::all_of(response))
  } else {
    response_data <- NULL
    predictor_data <- data
  }

  # Apply dimensionality reduction
  reduction_model <- tl_model(predictor_data, method = method, ...)

  # Transform data
  if (method == "pca") {
    transformed <- reduction_model$fit$scores

    # Select components
    if (!is.null(n_components)) {
      pc_cols <- paste0("PC", seq_len(n_components))
      transformed <- transformed %>%
        dplyr::select(dplyr::all_of(pc_cols))
    }

    # Add response back
    if (!is.null(response)) {
      transformed[[response]] <- response_data
    }

  } else if (method == "mds") {
    transformed <- reduction_model$fit$points

    # Select dimensions
    if (!is.null(n_components)) {
      dim_cols <- paste0("Dim", seq_len(n_components))
      transformed <- transformed %>%
        dplyr::select(dplyr::all_of(dim_cols))
    }

    # Add response back
    if (!is.null(response)) {
      transformed[[response]] <- response_data
    }
  }

  # Drop the .obs_id row identifier -- it is internal bookkeeping, not a
  # feature. Leaving it in place lets it reach downstream supervised models
  # as a high-cardinality predictor, which makes tree-based fits intractable.
  if (".obs_id" %in% names(transformed)) {
    transformed <- transformed[, names(transformed) != ".obs_id",
                               drop = FALSE]
  }

  list(
    data = transformed,
    reduction_model = reduction_model,
    original_data = data,
    response = response
  )
}

#' Cluster-Based Features
#'
#' Add cluster assignments as features for supervised learning.
#' This semi-supervised approach can capture non-linear patterns.
#'
#' @param data A data frame
#' @param response Response variable name (will be excluded from clustering)
#' @param method Clustering method: "kmeans", "pam", "hclust", "dbscan"
#' @param ... Additional arguments for clustering
#' @return The original data frame with an additional factor column named
#'   \code{cluster_<method>} containing cluster assignments. The fitted
#'   cluster model is stored as an attribute \code{"cluster_model"}.
#' @export
#' @examples
#' \donttest{
#' # Add cluster features before supervised learning
#' data_with_clusters <- tl_add_cluster_features(iris, response = "Species",
#'                                                 method = "kmeans", k = 3)
#' model <- tl_model(data_with_clusters, Species ~ ., method = "forest")
#' }
tl_add_cluster_features <- function(data,
                                    response = NULL,
                                    method = "kmeans",
                                    ...) {
  # Separate response if provided
  if (!is.null(response)) {
    if (!response %in% names(data)) {
      stop(
        "Response variable '", response,
        "' not found in data", call. = FALSE
      )
    }
    predictor_data <- data %>% dplyr::select(-dplyr::all_of(response))
  } else {
    predictor_data <- data
  }

  # Perform clustering
  cluster_model <- tl_model(predictor_data, method = method, ...)

  # Extract cluster assignments
  if (method %in% c("kmeans", "pam", "clara")) {
    clusters <- cluster_model$fit$clusters$cluster
  } else if (method == "hclust") {
    # For hclust, we need k to cut the tree
    if (!"k" %in% names(list(...))) {
      warning("k not specified for hclust, using k=3")
      k <- 3
    } else {
      k <- list(...)$k
    }
    clusters <- stats::cutree(cluster_model$fit$model, k = k)
  } else if (method == "dbscan") {
    clusters <- cluster_model$fit$clusters$cluster
  }

  # Add to original data
  data_augmented <- data %>%
    dplyr::mutate(
      !!paste0("cluster_", method) := as.factor(clusters)
    )

  attr(data_augmented, "cluster_model") <- cluster_model
  data_augmented
}

#' Semi-Supervised Learning via Clustering
#'
#' Train a supervised model with limited labels by first clustering the data
#' and propagating labels within clusters.
#'
#' @param data A data frame
#' @param formula Model formula
#' @param labeled_indices Indices of labeled observations
#' @param cluster_method Clustering method for label propagation
#' @param supervised_method Supervised learning method for final model
#' @param ... Additional arguments
#' @return A tidylearn model object with additional class
#'   \code{"tidylearn_semisupervised"}, trained on pseudo-labeled data. The
#'   model includes a \code{semisupervised_info} element with
#'   \code{labeled_indices}, \code{cluster_model}, and
#'   \code{label_mapping}.
#' @export
#' @examples
#' \donttest{
#' # Use only 10% of labels
#' labeled_idx <- sample(nrow(iris), size = 15)
#' model <- tl_semisupervised(iris, Species ~ ., labeled_indices = labeled_idx,
#'   cluster_method = "kmeans",
#'   supervised_method = "tree"
#' )
#' }
tl_semisupervised <- function(data, formula, labeled_indices,
                              cluster_method = "kmeans",
                              supervised_method = "logistic", ...) {
  # Extract response variable
  response_var <- all.vars(formula)[1]

  # Create training data with only labeled observations
  labeled_data <- data[labeled_indices, ]

  # Cluster the full dataset (excluding response)
  predictor_data <- data %>% dplyr::select(-dplyr::all_of(response_var))

  # Determine k from labeled data
  k <- length(unique(labeled_data[[response_var]]))

  cluster_model <- tl_model(predictor_data, method = cluster_method, k = k, ...)

  # Get cluster assignments
  clusters <- cluster_model$fit$clusters$cluster

  # Propagate labels within clusters
  cluster_labels <- tibble::tibble(
    obs_id = seq_len(nrow(data)),
    cluster = clusters,
    label = data[[response_var]]
  )

  # For each cluster, find the most common label from labeled data
  label_mapping <- cluster_labels %>%
    dplyr::filter(obs_id %in% labeled_indices) %>%
    dplyr::group_by(cluster) %>%
    dplyr::summarize(
      cluster_label = names(which.max(table(label))),
      .groups = "drop"
    )

  # Assign pseudo-labels to unlabeled data
  pseudo_labeled <- cluster_labels %>%
    dplyr::left_join(label_mapping, by = "cluster") %>%
    dplyr::mutate(
      final_label = dplyr::if_else(
        obs_id %in% labeled_indices,
        as.character(label), cluster_label
      )
    )

  # Create pseudo-labeled dataset
  data_pseudo <- data
  data_pseudo[[response_var]] <- as.factor(pseudo_labeled$final_label)

  # Train supervised model on pseudo-labeled data
  model <- tl_model(data_pseudo, formula, method = supervised_method, ...)

  # Add metadata
  model$semisupervised_info <- list(
    labeled_indices = labeled_indices,
    cluster_model = cluster_model,
    label_mapping = label_mapping
  )

  class(model) <- c("tidylearn_semisupervised", class(model))
  model
}

#' Anomaly-Aware Supervised Learning
#'
#' Detect outliers using DBSCAN or other methods, then optionally
#' remove them or down-weight them before supervised learning.
#'
#' @param data A data frame
#' @param formula Model formula
#' @param response Response variable name
#' @param anomaly_method Method for anomaly detection:
#'   "dbscan", "isolation_forest"
#' @param action Action to take: "remove", "flag", "downweight"
#' @param supervised_method Supervised learning method
#' @param ... Additional arguments
#' @return A tidylearn model object with additional class
#'   \code{"tidylearn_anomaly_aware"}. The model includes an
#'   \code{anomaly_info} element with \code{anomaly_model},
#'   \code{is_anomaly} (logical vector), \code{n_anomalies}, and
#'   \code{action}.
#' @export
#' @examples
#' \donttest{
#' model <- tl_anomaly_aware(iris, Species ~ ., response = "Species",
#'                            anomaly_method = "dbscan", action = "flag")
#' }
tl_anomaly_aware <- function(data, formula, response,
                             anomaly_method = "dbscan",
                             action = "flag",
                             supervised_method = "logistic",
                             ...) {
  # Separate predictors for anomaly detection
  predictor_data <- data %>% dplyr::select(-dplyr::all_of(response))

  # Detect anomalies
  if (anomaly_method == "dbscan") {
    anomaly_model <- tl_model(predictor_data, method = "dbscan", ...)
    is_anomaly <- anomaly_model$fit$clusters$cluster == 0  # DBSCAN noise points
  } else {
    stop("Unsupported anomaly method: ", anomaly_method, call. = FALSE)
  }

  # Take action based on anomalies
  if (action == "remove") {
    data_clean <- data[!is_anomaly, ]
    model <- tl_model(data_clean, formula, method = supervised_method)
    model$anomalies_removed <- sum(is_anomaly)
  } else if (action == "flag") {
    data_flagged <- data %>%
      dplyr::mutate(is_anomaly = is_anomaly)
    # Include anomaly flag as a feature - manually construct formula
    response_var <- all.vars(formula)[1]
    all_vars <- all.vars(formula)
    predictor_vars <- all_vars[-1]

    if (length(predictor_vars) == 1 && predictor_vars[1] == ".") {
      # Formula was response ~ .
      formula_updated <- as.formula(paste(response_var, "~ . + is_anomaly"))
    } else {
      # Explicit predictors
      preds_str <- paste(c(predictor_vars, "is_anomaly"), collapse = " + ")
      formula_updated <- as.formula(paste(response_var, "~", preds_str))
    }
    model <- tl_model(data_flagged, formula_updated, method = supervised_method)
  } else if (action == "downweight") {
    # Create weights (anomalies get lower weight)
    weights <- ifelse(is_anomaly, 0.1, 1.0)
    model <- tl_model(
      data, formula,
      method = supervised_method, weights = weights
    )
  }

  # Add anomaly detection info
  model$anomaly_info <- list(
    anomaly_model = anomaly_model,
    is_anomaly = is_anomaly,
    n_anomalies = sum(is_anomaly),
    action = action
  )

  class(model) <- c("tidylearn_anomaly_aware", class(model))
  model
}

#' Stratified Features via Clustering
#'
#' Create cluster-specific supervised models for heterogeneous data
#'
#' @param data A data frame
#' @param formula Model formula
#' @param cluster_method Clustering method
#' @param k Number of clusters
#' @param supervised_method Supervised learning method
#' @param ... Additional arguments
#' @return A list with class \code{"tidylearn_stratified"} containing:
#'   \describe{
#'     \item{cluster_model}{The fitted clustering model.}
#'     \item{supervised_models}{Named list of tidylearn models, one per
#'       cluster.}
#'     \item{formula}{The model formula.}
#'     \item{data}{The original training data.}
#'   }
#' @export
#' @examples
#' \donttest{
#' models <- tl_stratified_models(mtcars, mpg ~ ., cluster_method = "kmeans",
#'                                 k = 3, supervised_method = "linear")
#' }
tl_stratified_models <- function(data, formula, cluster_method = "kmeans",
                                 k = 3, supervised_method = "linear", ...) {
  # Extract response variable
  response_var <- all.vars(formula)[1]

  # Cluster the predictors
  predictor_data <- data %>% dplyr::select(-dplyr::all_of(response_var))
  cluster_model <- tl_model(predictor_data, method = cluster_method, k = k, ...)

  # Get cluster assignments
  clusters <- cluster_model$fit$clusters$cluster

  # Train a model for each cluster
  cluster_models <- list()
  for (i in seq_len(k)) {
    cluster_data <- data[clusters == i, ]
    if (nrow(cluster_data) > 0) {
      cluster_models[[paste0("cluster_", i)]] <- tl_model(
        cluster_data, formula, method = supervised_method, ...
      )
    }
  }

  # Return stratified model object
  structure(
    list(
      cluster_model = cluster_model,
      supervised_models = cluster_models,
      formula = formula,
      data = data
    ),
    class = c("tidylearn_stratified", "list")
  )
}

#' Predict from stratified models
#' @param object A tidylearn_stratified model object
#' @param new_data New data for predictions
#' @param ... Additional arguments
#' @return A \link[tibble]{tibble} with a \code{.pred} column containing
#'   predictions and a \code{.cluster} column with cluster assignments.
#' @examples
#' \donttest{
#' models <- tl_stratified_models(mtcars, mpg ~ .,
#'   cluster_method = "kmeans", k = 2, supervised_method = "linear")
#' preds <- predict(models)
#' }
#' @export
predict.tidylearn_stratified <- function(object, new_data = NULL, ...) {
  if (is.null(new_data)) {
    new_data <- object$data
  }

  # Get response variable
  response_var <- all.vars(object$formula)[1]

  # Assign new data to clusters
  predictor_data <- new_data %>% dplyr::select(-dplyr::all_of(response_var))
  new_clusters <- predict(object$cluster_model, new_data = predictor_data)

  # Predict from appropriate cluster model
  predictions <- vector("list", nrow(new_data))
  for (i in seq_len(nrow(new_data))) {
    cluster_id <- new_clusters$cluster[i]
    model_name <- paste0("cluster_", cluster_id)

    if (model_name %in% names(object$supervised_models)) {
      pred <- predict(
        object$supervised_models[[model_name]],
        new_data = new_data[i, , drop = FALSE], ...
      )
      predictions[[i]] <- pred$.pred[1]
    } else {
      predictions[[i]] <- NA
    }
  }

  tibble::tibble(
    .pred = unlist(predictions),
    .cluster = new_clusters$cluster
  )
}

test_that("tl_reduce_dimensions works with PCA", {
  result <- tl_reduce_dimensions(iris, response = "Species",
                                 method = "pca", n_components = 3)

  expect_type(result, "list")
  expect_true("data" %in% names(result))
  expect_true("reduction_model" %in% names(result))

  # Check transformed data has PC columns
  expect_true(any(grepl("PC", names(result$data))))

  # Response should be preserved
  expect_true("Species" %in% names(result$data))
  expect_equal(result$data$Species, iris$Species)

  # Should have requested number of components
  pc_cols <- sum(grepl("^PC\\d+$", names(result$data)))
  expect_equal(pc_cols, 3)
})

test_that("tl_reduce_dimensions works without response", {
  result <- tl_reduce_dimensions(iris[, 1:4], method = "pca", n_components = 2)

  expect_type(result, "list")
  expect_true("data" %in% names(result))

  # Should have PC columns
  pc_cols <- sum(grepl("^PC", names(result$data)))
  expect_gte(pc_cols, 2)
})

test_that("tl_add_cluster_features adds cluster columns", {
  data_with_clusters <- tl_add_cluster_features(iris, response = "Species",
                                                method = "kmeans", k = 3)

  # Should have cluster column
  expect_true(any(grepl("cluster_", names(data_with_clusters))))

  # Original columns should be preserved
  expect_true(all(names(iris) %in% names(data_with_clusters)))

  # Cluster column should be a factor
  cluster_col <- grep("cluster_", names(data_with_clusters), value = TRUE)
  expect_s3_class(data_with_clusters[[cluster_col]], "factor")
})

test_that("tl_add_cluster_features works with different clustering methods", {
  # K-means
  data_kmeans <- tl_add_cluster_features(iris, response = "Species",
                                         method = "kmeans", k = 3)
  expect_true("cluster_kmeans" %in% names(data_kmeans))

  # PAM
  skip_if_not_installed("cluster")
  data_pam <- tl_add_cluster_features(iris, response = "Species",
                                      method = "pam", k = 3)
  expect_true("cluster_pam" %in% names(data_pam))
})

test_that("tl_semisupervised performs label propagation", {
  # Use only 10% of labels
  set.seed(123)
  labeled_idx <- sample(nrow(iris), size = 15)

  model <- tl_semisupervised(iris, Species ~ .,
                             labeled_indices = labeled_idx,
                             cluster_method = "kmeans",
                             supervised_method = "logistic")

  expect_s3_class(model, "tidylearn_semisupervised")
  expect_s3_class(model, "tidylearn_supervised")

  # Should have semisupervised info
  expect_true("semisupervised_info" %in% names(model))
  expect_equal(model$semisupervised_info$labeled_indices, labeled_idx)

  # Can predict
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(iris))
})

test_that("tl_anomaly_aware detects and handles outliers", {
  skip_if_not_installed("dbscan")

  # Flag anomalies
  model_flag <- tl_anomaly_aware(iris, Species ~ .,
                                 response = "Species",
                                 anomaly_method = "dbscan",
                                 action = "flag",
                                 supervised_method = "logistic")

  expect_s3_class(model_flag, "tidylearn_anomaly_aware")
  expect_true("anomaly_info" %in% names(model_flag))
  expect_equal(model_flag$anomaly_info$action, "flag")

  # Remove anomalies
  model_remove <- tl_anomaly_aware(iris, Species ~ .,
                                   response = "Species",
                                   anomaly_method = "dbscan",
                                   action = "remove",
                                   supervised_method = "logistic")

  expect_s3_class(model_remove, "tidylearn_anomaly_aware")
  expect_true("anomalies_removed" %in% names(model_remove))
})

test_that("tl_stratified_models creates cluster-specific models", {
  models <- tl_stratified_models(mtcars, mpg ~ .,
                                 cluster_method = "kmeans",
                                 k = 3,
                                 supervised_method = "linear")

  expect_s3_class(models, "tidylearn_stratified")
  expect_true("cluster_model" %in% names(models))
  expect_true("supervised_models" %in% names(models))

  # Should have one model per cluster
  expect_gte(length(models$supervised_models), 1)
  expect_lte(length(models$supervised_models), 3)
})

test_that("predict.tidylearn_stratified assigns to clusters and predicts", {
  models <- tl_stratified_models(mtcars, mpg ~ .,
                                 cluster_method = "kmeans",
                                 k = 2,
                                 supervised_method = "linear")

  # Predict on training data
  preds <- predict(models)
  expect_equal(nrow(preds), nrow(mtcars))
  expect_true(".pred" %in% names(preds))
  expect_true(".cluster" %in% names(preds))

  # Predict on new data
  preds_new <- predict(models, new_data = mtcars[1:10, ])
  expect_equal(nrow(preds_new), 10)
})

test_that("integration functions validate inputs", {
  # Invalid response variable
  expect_error(
    tl_reduce_dimensions(iris, response = "InvalidColumn", method = "pca"),
    "Response variable.*not found"
  )

  expect_error(
    tl_add_cluster_features(iris, response = "InvalidColumn",
                            method = "kmeans", k = 3),
    "Response variable.*not found"
  )
})

test_that("reduced data can be used for supervised learning", {
  # Reduce dimensions
  reduced <- tl_reduce_dimensions(iris, response = "Species",
                                  method = "pca", n_components = 3)

  # Train model on reduced data
  model <- tl_model(reduced$data, Species ~ ., method = "logistic")

  expect_s3_class(model, "tidylearn_logistic")

  # Can predict
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(iris))
})

test_that("cluster features improve model", {
  # This is more of an integration test to ensure the workflow works
  data_clustered <- tl_add_cluster_features(iris,
                                            response = "Species",
                                            method = "kmeans", k = 3)

  # Train model with cluster features
  model <- tl_model(data_clustered, Species ~ ., method = "logistic")

  expect_s3_class(model, "tidylearn_logistic")

  # Can predict
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(data_clustered))
})

#' @title Visualization Functions for tidylearn
#' @name tidylearn-visualization
#' @description General visualization functions for tidylearn models
#' @importFrom ggplot2 ggplot aes geom_line geom_point geom_bar geom_boxplot
#' @importFrom ggplot2 geom_histogram geom_density
#' @importFrom ggplot2 geom_jitter scale_color_gradient
#' @importFrom ggplot2 labs theme_minimal
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate filter group_by summarize arrange
NULL

#' Plot feature importance across multiple models
#'
#' @param ... tidylearn model objects to compare
#' @param top_n Number of top features to display (default: 10)
#' @param names Optional character vector of model names
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' m1 <- tl_model(iris, Species ~ ., method = "forest")
#' m2 <- tl_model(iris, Species ~ ., method = "boost")
#' tl_plot_importance_comparison(m1, m2, names = c("Forest", "Boost"))
#' }
#' @export
tl_plot_importance_comparison <- function(..., top_n = 10, names = NULL) {
  # Get models
  models <- list(...)

  # Get model names if not provided
  if (is.null(names)) {
    names <- paste0("Model ", seq_along(models))
  } else if (length(names) != length(models)) {
    stop("Length of 'names' must match the number of models", call. = FALSE)
  }

  # Extract importance for each model
  all_importance <- purrr::map2_dfr(models, names, function(model, name) {
    # Check model type
    if (model$spec$method %in% c("tree", "forest", "boost")) {
      # Tree-based models
      imp_data <- tl_extract_importance(model)

      # Add model name
      imp_data$model <- name

      imp_data
    } else if (model$spec$method %in% c("ridge", "lasso", "elastic_net")) {
      # Regularized regression
      imp_data <- tl_get_importance_regularized(model)

      # Add model name
      imp_data$model <- name

      imp_data
    } else {
      warning(
        "Importance extraction not implemented for model type: ",
        model$spec$method,
        call. = FALSE
      )
      NULL
    }
  })


  # If no importances could be extracted, return NULL
  if (is.null(all_importance) || nrow(all_importance) == 0) {
    NULL
  }

  # Find top features across all models
  top_features <- all_importance %>%
    dplyr::group_by(.data[["feature"]]) %>%
    dplyr::summarize(
      avg_importance = mean(.data[["importance"]]),
      .groups = "drop"
    ) %>%
    dplyr::arrange(dplyr::desc(.data[["avg_importance"]])) %>%
    dplyr::slice_head(n = top_n) %>%
    dplyr::pull(.data[["feature"]])

  # Filter to only top features
  plot_data <- all_importance %>%
    dplyr::filter(.data[["feature"]] %in% top_features)

  # Create the plot
  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(
      x = stats::reorder(feature, importance),
      y = importance,
      fill = model
    )
  ) +
    ggplot2::geom_col(position = "dodge") +
    ggplot2::coord_flip() +
    ggplot2::labs(
      title = "Feature Importance Comparison",
      x = NULL,
      y = "Importance",
      fill = "Model"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Extract importance from a tree-based model
#'
#' @param model A tidylearn model object
#' @return A data frame with feature importance values
#' @keywords internal
tl_extract_importance <- function(model) {
  # Get the model
  fit <- model$fit
  method <- model$spec$method

  if (method == "tree") {
    # Decision tree importance
    # Get variable importance from rpart
    imp <- fit$variable.importance

    # Create a data frame for plotting
    importance_df <- tibble::tibble(
      feature = names(imp),
      importance = as.vector(imp)
    )
  } else if (method == "forest") {
    # Random forest importance
    # Get variable importance from randomForest
    imp <- randomForest::importance(fit)

    # Create a data frame for plotting
    if (model$spec$is_classification) {
      # For classification, use mean decrease in accuracy
      importance_df <- tibble::tibble(
        feature = rownames(imp),
        importance = imp[, "MeanDecreaseAccuracy"]
      )
    } else {
      # For regression, use % increase in MSE
      importance_df <- tibble::tibble(
        feature = rownames(imp),
        importance = imp[, "%IncMSE"]
      )
    }
  } else if (method == "boost") {
    # Gradient boosting importance
    # Get relative influence from gbm
    imp <- summary(fit, plotit = FALSE)

    # Create a data frame for plotting
    importance_df <- tibble::tibble(
      feature = imp$var,
      importance = imp$rel.inf
    )
  } else {
    stop(
      "Variable importance extraction not implemented for method: ",
      method,
      call. = FALSE
    )
  }

  # Normalize importance to 0-100 scale
  importance_df <- importance_df %>%
    dplyr::mutate(
      importance = 100 * .data[["importance"]] / max(.data[["importance"]])
    )

  importance_df
}

#' Extract importance from a regularized regression model
#'
#' @param model A tidylearn regularized model object
#' @param lambda Which lambda to use ("1se" or "min", default: "1se")
#' @return A data frame with feature importance values
#' @keywords internal
tl_get_importance_regularized <- function(model, lambda = "1se") {
  # Extract the glmnet model
  fit <- model$fit

  # Extract lambda value to use
  if (lambda == "1se") {
    lambda_val <- attr(fit, "lambda_1se")
  } else if (lambda == "min") {
    lambda_val <- attr(fit, "lambda_min")
  } else if (is.numeric(lambda)) {
    lambda_val <- lambda
  } else {
    stop(
      "Invalid lambda specification. Use '1se', 'min', or a numeric value.",
      call. = FALSE
    )
  }

  # Get coefficients at selected lambda
  coefs <- as.matrix(coef(fit, s = lambda_val))

  # Exclude intercept
  coefs <- coefs[-1, , drop = FALSE]

  # Create a data frame for plotting
  importance_df <- tibble::tibble(
    feature = rownames(coefs),
    importance = abs(as.vector(coefs))
  ) %>%
    dplyr::filter(.data[["importance"]] > 0)

  # Normalize importance to 0-100 scale
  importance_df <- importance_df %>%
    dplyr::mutate(
      importance = 100 * .data[["importance"]] / max(.data[["importance"]])
    )

  importance_df
}

#' Plot model comparison
#'
#' @param ... tidylearn model objects to compare
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param metrics Character vector of metrics to compute
#' @param names Optional character vector of model names
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' m1 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "lasso")
#' tl_plot_model_comparison(m1, m2, names = c("Linear", "Lasso"))
#' }
#' @export
tl_plot_model_comparison <- function(
    ...,
    new_data = NULL,
    metrics = NULL,
    names = NULL) {
  # Get models
  models <- list(...)

  # Get model names if not provided
  if (is.null(names)) {
    names <- purrr::map_chr(models, function(model) {
      task <- ifelse(
        model$spec$is_classification,
        "classification",
        "regression"
      )
      paste0(model$spec$method, " (", task, ")")
    })
  } else if (length(names) != length(models)) {
    stop("Length of 'names' must match the number of models", call. = FALSE)
  }

  # Check if all models are of the same type (classification or regression)
  is_classifications <- purrr::map_lgl(
    models,
    function(model) model$spec$is_classification
  )
  if (length(unique(is_classifications)) > 1) {
    stop(
      "All models must be of the same type (classification or regression)",
      call. = FALSE
    )
  }

  is_classification <- is_classifications[1]

  # Use first model's training data if new_data not provided
  if (is.null(new_data)) {
    new_data <- models[[1]]$data
    message(
      "Evaluating on training data. ",
      "For model validation, provide separate test data."
    )
  }

  # Default metrics based on problem type
  if (is.null(metrics)) {
    if (is_classification) {
      metrics <- c("accuracy", "precision", "recall", "f1", "auc")
    } else {
      metrics <- c("rmse", "mae", "rsq", "mape")
    }
  }

  # Evaluate each model
  model_results <- purrr::map2_dfr(models, names, function(model, name) {
    # Evaluate model
    eval_results <- tl_evaluate(model, new_data, metrics)

    # Add model name
    eval_results$model <- name

    eval_results
  })

  # Create the plot
  p <- ggplot2::ggplot(
    model_results,
    ggplot2::aes(x = model, y = value, fill = metric)
  ) +
    ggplot2::geom_col(position = "dodge") +
    ggplot2::facet_wrap(~ metric, scales = "free_y") +
    ggplot2::labs(
      title = "Model Comparison",
      x = NULL,
      y = "Metric Value",
      fill = "Metric"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)
    )

  p
}

#' Plot cross-validation results
#'
#' @param cv_results Cross-validation results from tl_cv function
#' @param metrics Character vector of metrics to plot
#'   (if NULL, plots all metrics)
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @export
tl_plot_cv_results <- function(cv_results, metrics = NULL) {
  # Extract fold metrics
  fold_metrics <- cv_results$fold_metrics

  # Filter metrics if specified
  if (!is.null(metrics)) {
    fold_metrics <- fold_metrics %>%
      dplyr::filter(.data[["metric"]] %in% metrics)
  }

  # Create the plot
  p <- ggplot2::ggplot(
    fold_metrics,
    ggplot2::aes(
      x = factor(fold),
      y = value,
      group = metric,
      color = metric
    )
  ) +
    ggplot2::geom_line() +
    ggplot2::geom_point() +
    ggplot2::facet_wrap(~ metric, scales = "free_y") +
    ggplot2::geom_hline(
      data = cv_results$summary,
      ggplot2::aes(yintercept = mean_value, color = metric),
      linetype = "dashed"
    ) +
    ggplot2::labs(
      title = "Cross-Validation Results",
      subtitle = "Dashed lines represent mean values across folds",
      x = "Fold",
      y = "Metric Value",
      color = "Metric"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Create interactive visualization dashboard for a model
#'
#' @param model A tidylearn model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A \code{\link[shiny]{shinyApp}} object.
#' @examples
#' \donttest{
#' if (requireNamespace("shiny")) {
#'   model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#'   app <- tl_dashboard(model)
#' }
#' }
#' @export
tl_dashboard <- function(model, new_data = NULL, ...) {
  # Check if required packages are installed
  tl_check_packages(c("shiny", "shinydashboard", "DT"))

  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Define UI
  ui <- shinydashboard::dashboardPage(
    shinydashboard::dashboardHeader(title = "tidylearn Model Dashboard"),

    shinydashboard::dashboardSidebar(
      shinydashboard::sidebarMenu(
        shinydashboard::menuItem(
          "Overview", tabName = "overview",
          icon = shiny::icon("dashboard")
        ),
        shinydashboard::menuItem(
          "Performance", tabName = "performance",
          icon = shiny::icon("chart-line")
        ),
        shinydashboard::menuItem(
          "Predictions", tabName = "predictions",
          icon = shiny::icon("table")
        ),
        shinydashboard::menuItem(
          "Diagnostics", tabName = "diagnostics",
          icon = shiny::icon("chart-area")
        )
      )
    ),

    shinydashboard::dashboardBody(
      shinydashboard::tabItems(
        # Overview tab
        shinydashboard::tabItem(
          tabName = "overview",
          shiny::fluidRow(
            shinydashboard::box(
              title = "Model Summary",
              width = 12,
              shiny::verbatimTextOutput("model_summary")
            )
          ),
          shiny::fluidRow(
            shinydashboard::box(
              title = "Feature Importance",
              width = 12,
              shiny::plotOutput("importance_plot")
            )
          )
        ),

        # Performance tab
        shinydashboard::tabItem(
          tabName = "performance",
          shiny::fluidRow(
            shinydashboard::box(
              title = "Performance Metrics",
              width = 12,
              DT::DTOutput("metrics_table")
            )
          ),
          shiny::conditionalPanel(
            condition = "output.is_classification == true",
            shiny::fluidRow(
              shinydashboard::box(
                title = "ROC Curve",
                width = 6,
                shiny::plotOutput("roc_plot")
              ),
              shinydashboard::box(
                title = "Confusion Matrix",
                width = 6,
                shiny::plotOutput("confusion_plot")
              )
            )
          ),
          shiny::conditionalPanel(
            condition = "output.is_classification == false",
            shiny::fluidRow(
              shinydashboard::box(
                title = "Actual vs Predicted",
                width = 6,
                shiny::plotOutput("actual_predicted_plot")
              ),
              shinydashboard::box(
                title = "Residuals",
                width = 6,
                shiny::plotOutput("residuals_plot")
              )
            )
          )
        ),

        # Predictions tab
        shinydashboard::tabItem(
          tabName = "predictions",
          shiny::fluidRow(
            shinydashboard::box(
              title = "Predictions",
              width = 12,
              DT::DTOutput("predictions_table")
            )
          )
        ),

        # Diagnostics tab
        shinydashboard::tabItem(
          tabName = "diagnostics",
          shiny::conditionalPanel(
            condition = "output.is_classification == false",
            shiny::fluidRow(
              shinydashboard::box(
                title = "Diagnostic Plots",
                width = 12,
                shiny::plotOutput("diagnostics_plot")
              )
            )
          ),
          shiny::conditionalPanel(
            condition = "output.is_classification == true",
            shiny::fluidRow(
              shinydashboard::box(
                title = "Calibration Plot",
                width = 6,
                shiny::plotOutput("calibration_plot")
              ),
              shinydashboard::box(
                title = "Precision-Recall Curve",
                width = 6,
                shiny::plotOutput("pr_curve_plot")
              )
            )
          )
        )
      )
    )
  )

  # Define server logic
  server <- function(input, output, session) {
    # Flag for classification or regression
    output$is_classification <- shiny::reactive({
      model$spec$is_classification
    })
    shiny::outputOptions(output, "is_classification", suspendWhenHidden = FALSE)

    # Model summary
    output$model_summary <- shiny::renderPrint({
      summary(model)
    })

    # Performance metrics
    output$metrics_table <- DT::renderDT({
      metrics <- tl_evaluate(model, new_data)
      DT::datatable(metrics,
                    options = list(pageLength = 10),
                    rownames = FALSE)
    })

    # Feature importance
    output$importance_plot <- shiny::renderPlot({
      tree_or_reg <- c(
        "tree", "forest", "boost",
        "ridge", "lasso", "elastic_net"
      )
      if (model$spec$method %in% tree_or_reg) {
        tl_plot_importance(model)
      } else {
        shiny::validate(
          shiny::need(
            FALSE,
            "Feature importance not available for this model type"
          )
        )
      }
    })

    # Predictions
    output$predictions_table <- DT::renderDT({
      # Get actual values
      actuals <- new_data[[model$spec$response_var]]

      if (model$spec$is_classification) {
        # Classification
        pred_class <- predict(model, new_data, type = "class")$.pred
        pred_prob <- predict(model, new_data, type = "prob")

        # Combine into a data frame
        results <- cbind(
          data.frame(actual = actuals, predicted = pred_class),
          pred_prob
        )
      } else {
        # Regression
        predictions <- predict(model, new_data)$.pred

        # Combine into a data frame
        results <- data.frame(
          actual = actuals,
          predicted = predictions,
          residual = actuals - predictions
        )
      }

      DT::datatable(results,
                    options = list(pageLength = 10),
                    rownames = FALSE)
    })

    # ROC plot (for classification)
    output$roc_plot <- shiny::renderPlot({
      if (model$spec$is_classification) {
        tl_plot_roc(model, new_data)
      }
    })

    # Confusion matrix (for classification)
    output$confusion_plot <- shiny::renderPlot({
      if (model$spec$is_classification) {
        tl_plot_confusion(model, new_data)
      }
    })

    # Actual vs predicted plot (for regression)
    output$actual_predicted_plot <- shiny::renderPlot({
      if (!model$spec$is_classification) {
        tl_plot_actual_predicted(model, new_data)
      }
    })

    # Residuals plot (for regression)
    output$residuals_plot <- shiny::renderPlot({
      if (!model$spec$is_classification) {
        tl_plot_residuals(model, new_data)
      }
    })

    # Diagnostics plots (for regression)
    output$diagnostics_plot <- shiny::renderPlot({
      if (!model$spec$is_classification) {
        tl_plot_diagnostics(model)
      }
    })

    # Calibration plot (for classification)
    output$calibration_plot <- shiny::renderPlot({
      if (model$spec$is_classification) {
        tl_plot_calibration(model, new_data)
      }
    })

    # Precision-Recall curve (for classification)
    output$pr_curve_plot <- shiny::renderPlot({
      if (model$spec$is_classification) {
        tl_plot_precision_recall(model, new_data)
      }
    })
  }

  # Return the Shiny app
  shiny::shinyApp(ui, server)
}

#' Plot lift chart for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param bins Number of bins for grouping predictions
#'   (default: 10)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_line geom_point
#' @importFrom ggplot2 geom_hline labs theme_minimal
#' @examples
#' \donttest{
#' iris_bin <- iris[iris$Species != "setosa", ]
#' iris_bin$Species <- factor(iris_bin$Species)
#' model <- tl_model(iris_bin, Species ~ ., method = "logistic")
#' tl_plot_lift(model)
#' }
#' @export
tl_plot_lift <- function(model, new_data = NULL, bins = 10, ...) {
  if (!model$spec$is_classification) {
    stop(
      "Lift chart is only available for classification models",
      call. = FALSE
    )
  }

  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # For binary classification
  if (length(levels(actuals)) == 2) {
    # Get probabilities
    probs <- predict(model, new_data, type = "prob")
    pos_class <- levels(actuals)[2]
    pos_probs <- probs[[pos_class]]

    # Convert actuals to binary (0/1)
    binary_actuals <- as.integer(actuals == pos_class)

    # Order by probability
    ordered_data <- tibble::tibble(
      prob = pos_probs,
      actual = binary_actuals
    ) %>%
      dplyr::arrange(dplyr::desc(.data[["prob"]]))

    # Calculate cumulative metrics
    decile_size <- ceiling(nrow(ordered_data) / bins)

    # Calculate lift by decile
    lift_data <- tibble::tibble(
      decile = integer(),
      cumulative_responders = integer(),
      cumulative_total = integer(),
      cumulative_response_rate = numeric(),
      baseline_rate = numeric(),
      lift = numeric()
    )

    baseline_rate <- mean(binary_actuals)
    cumulative_responders <- 0
    cumulative_total <- 0

    for (i in 1:bins) {
      # Get current decile indices
      start_idx <- (i - 1) * decile_size + 1
      end_idx <- min(i * decile_size, nrow(ordered_data))

      # Update cumulative counts
      current_responders <- sum(ordered_data$actual[start_idx:end_idx])
      current_total <- end_idx - start_idx + 1

      cumulative_responders <- cumulative_responders + current_responders
      cumulative_total <- cumulative_total + current_total

      # Calculate metrics
      cumulative_response_rate <- cumulative_responders / cumulative_total
      lift <- cumulative_response_rate / baseline_rate

      # Add to results
      lift_data <- lift_data %>%
        dplyr::add_row(
          decile = i,
          cumulative_responders = cumulative_responders,
          cumulative_total = cumulative_total,
          cumulative_response_rate = cumulative_response_rate,
          baseline_rate = baseline_rate,
          lift = lift
        )
    }

    # Create the plot
    p <- ggplot2::ggplot(lift_data, ggplot2::aes(x = decile, y = lift)) +
      ggplot2::geom_line(color = "blue") +
      ggplot2::geom_point(color = "blue", size = 3) +
      ggplot2::geom_hline(yintercept = 1, linetype = "dashed", color = "red") +
      ggplot2::labs(
        title = "Lift Chart",
        subtitle = "Cumulative lift by decile",
        x = "Decile (sorted by predicted probability)",
        y = "Cumulative Lift"
      ) +
      ggplot2::scale_x_continuous(breaks = 1:bins) +
      ggplot2::theme_minimal()

    p
  } else {
    stop(
      "Lift chart is currently only implemented for binary classification",
      call. = FALSE
    )
  }
}

#' Plot gain chart for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param bins Number of bins for grouping predictions
#'   (default: 10)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_line geom_point
#' @importFrom ggplot2 geom_abline labs theme_minimal
#' @examples
#' \donttest{
#' iris_bin <- iris[iris$Species != "setosa", ]
#' iris_bin$Species <- factor(iris_bin$Species)
#' model <- tl_model(iris_bin, Species ~ ., method = "logistic")
#' tl_plot_gain(model)
#' }
#' @export
tl_plot_gain <- function(model, new_data = NULL, bins = 10, ...) {
  if (!model$spec$is_classification) {
    stop(
      "Gain chart is only available for classification models",
      call. = FALSE
    )
  }

  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # For binary classification
  if (length(levels(actuals)) == 2) {
    # Get probabilities
    probs <- predict(model, new_data, type = "prob")
    pos_class <- levels(actuals)[2]
    pos_probs <- probs[[pos_class]]

    # Convert actuals to binary (0/1)
    binary_actuals <- as.integer(actuals == pos_class)

    # Order by probability
    ordered_data <- tibble::tibble(
      prob = pos_probs,
      actual = binary_actuals
    ) %>%
      dplyr::arrange(dplyr::desc(.data[["prob"]]))

    # Calculate cumulative metrics
    decile_size <- ceiling(nrow(ordered_data) / bins)
    total_responders <- sum(binary_actuals)

    # Calculate gain by decile
    gain_data <- tibble::tibble(
      decile = integer(),
      cumulative_pct_population = numeric(),
      cumulative_pct_responders = numeric()
    )

    cumulative_responders <- 0

    for (i in 1:bins) {
      # Get current decile indices
      start_idx <- (i - 1) * decile_size + 1
      end_idx <- min(i * decile_size, nrow(ordered_data))

      # Update cumulative counts
      current_responders <- sum(ordered_data$actual[start_idx:end_idx])
      cumulative_responders <- cumulative_responders + current_responders

      # Calculate metrics
      cumulative_pct_population <- end_idx / nrow(ordered_data) * 100
      cumulative_pct_responders <-
        cumulative_responders / total_responders * 100

      # Add to results
      gain_data <- gain_data %>%
        dplyr::add_row(
          decile = i,
          cumulative_pct_population = cumulative_pct_population,
          cumulative_pct_responders = cumulative_pct_responders
        )
    }

    # Add origin point
    gain_data <- dplyr::bind_rows(
      tibble::tibble(
        decile = 0,
        cumulative_pct_population = 0,
        cumulative_pct_responders = 0
      ),
      gain_data
    )

    # Create the plot
    p <- ggplot2::ggplot(
      gain_data,
      ggplot2::aes(
        x = cumulative_pct_population,
        y = cumulative_pct_responders
      )
    ) +
      ggplot2::geom_line(color = "blue", linewidth = 1) +
      ggplot2::geom_point(color = "blue", size = 3) +
      ggplot2::geom_abline(
        intercept = 0,
        slope = 1,
        linetype = "dashed",
        color = "red"
      ) +
      ggplot2::labs(
        title = "Cumulative Gain Chart",
        subtitle = "Cumulative % of responders by % of population",
        x = "Cumulative % of Population",
        y = "Cumulative % of Responders"
      ) +
      ggplot2::coord_fixed() +
      ggplot2::scale_x_continuous(breaks = seq(0, 100, by = 10)) +
      ggplot2::scale_y_continuous(breaks = seq(0, 100, by = 10)) +
      ggplot2::theme_minimal()

    p
  } else {
    stop(
      "Gain chart is currently only implemented for binary classification",
      call. = FALSE
    )
  }
}
#' Plot Clusters in 2D Space
#'
#' Visualize clustering results using first two dimensions
#' or specified dimensions
#'
#' @param data A data frame with cluster assignments
#' @param cluster_col Name of cluster column (default: "cluster")
#' @param x_col X-axis variable (if NULL, uses first numeric column)
#' @param y_col Y-axis variable (if NULL, uses second numeric column)
#' @param centers Optional data frame of cluster centers
#' @param title Plot title
#' @param color_noise_black If TRUE, color noise points (cluster 0) black
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' km <- tidy_kmeans(iris[, 1:4], k = 3)
#' clustered <- augment_kmeans(km, iris[, 1:4])
#' plot_clusters(clustered)
#' }
#' @export
plot_clusters <- function(data,
                          cluster_col = "cluster",
                          x_col = NULL,
                          y_col = NULL,
                          centers = NULL,
                          title = "Cluster Plot",
                          color_noise_black = TRUE) {

  # Find numeric columns if not specified
  numeric_cols <- names(data)[sapply(data, is.numeric)]

  if (is.null(x_col)) {
    x_col <- numeric_cols[1]
  }

  if (is.null(y_col)) {
    y_col <- if (length(numeric_cols) > 1) {
      numeric_cols[2]
    } else {
      numeric_cols[1]
    }
  }

  # Ensure cluster column is factor
  plot_data <- data %>%
    dplyr::mutate(!!cluster_col := as.factor(!!rlang::sym(cluster_col)))

  # Create base plot
  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(
      x = !!rlang::sym(x_col),
      y = !!rlang::sym(y_col),
      color = !!rlang::sym(cluster_col)
    )
  ) +
    ggplot2::geom_point(size = 2.5, alpha = 0.7) +
    ggplot2::labs(
      title = title,
      x = x_col,
      y = y_col,
      color = "Cluster"
    ) +
    ggplot2::theme_minimal()

  # Color noise points black if requested
  if (color_noise_black && "0" %in% unique(plot_data[[cluster_col]])) {
    n_clusters <- length(unique(plot_data[[cluster_col]])) - 1
    other_clusters <- setdiff(unique(plot_data[[cluster_col]]), "0")
    cluster_colors <- setNames(grDevices::rainbow(n_clusters), other_clusters)
    p <- p + ggplot2::scale_color_manual(
      values = c("0" = "black", cluster_colors)
    )
  }

  # Add centers if provided
  if (!is.null(centers)) {
    p <- p + ggplot2::geom_point(
      data = centers,
      ggplot2::aes(x = !!rlang::sym(x_col), y = !!rlang::sym(y_col)),
      color = "black", size = 5, shape = 4, stroke = 2,
      inherit.aes = FALSE
    )
  }

  p
}


#' Create Elbow Plot for K-Means
#'
#' Plot total within-cluster sum of squares vs number of clusters
#'
#' @param wss_data A tibble with columns k and tot_withinss (from calc_wss)
#' @param add_line Add vertical line at suggested optimal k? (default: FALSE)
#' @param suggested_k If add_line=TRUE, which k to highlight
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' wss <- data.frame(k = 2:6, tot_withinss = c(150, 90, 60, 50, 45))
#' plot_elbow(wss)
#' }
#' @export
plot_elbow <- function(wss_data, add_line = FALSE, suggested_k = NULL) {

  p <- ggplot2::ggplot(wss_data, ggplot2::aes(x = k, y = tot_withinss)) +
    ggplot2::geom_line(color = "steelblue", linewidth = 1) +
    ggplot2::geom_point(color = "steelblue", size = 3) +
    ggplot2::labs(
      title = "Elbow Method - Total Within-Cluster Sum of Squares",
      subtitle = "Look for 'elbow' in the curve",
      x = "Number of Clusters (k)",
      y = "Total Within-Cluster SS"
    ) +
    ggplot2::theme_minimal()

  if (add_line && !is.null(suggested_k)) {
    p <- p +
      ggplot2::geom_vline(
        xintercept = suggested_k,
        linetype = "dashed",
        color = "red"
      ) +
      ggplot2::annotate(
        "text",
        x = suggested_k,
        y = max(wss_data$tot_withinss) * 0.9,
        label = sprintf("k = %d", suggested_k),
        color = "red",
        hjust = -0.2
      )
  }

  p
}


#' Create Cluster Comparison Plot
#'
#' Compare multiple clustering results side-by-side
#'
#' @param data Data frame with multiple cluster columns
#' @param cluster_cols Vector of cluster column names
#' @param x_col X-axis variable
#' @param y_col Y-axis variable
#'
#' @return The return value of \code{\link[gridExtra]{grid.arrange}}, a
#'   \code{\link[gtable]{gtable}} drawn as a side effect.
#' @examples
#' \donttest{
#' df <- iris[, 1:4]
#' df$km3 <- kmeans(df, 3)$cluster
#' df$km4 <- kmeans(df, 4)$cluster
#' plot_cluster_comparison(df, c("km3", "km4"), "Sepal.Length", "Sepal.Width")
#' }
#' @export
plot_cluster_comparison <- function(data, cluster_cols, x_col, y_col) {

  plots <- purrr::map(cluster_cols, function(col) {
    plot_clusters(data, cluster_col = col, x_col = x_col, y_col = y_col,
                  title = paste("Clusters:", col))
  })

  # Combine plots
  gridExtra::grid.arrange(
    grobs = plots,
    ncol = ceiling(sqrt(length(plots)))
  )
}


#' Plot Cluster Size Distribution
#'
#' Create bar plot of cluster sizes
#'
#' @param clusters Vector of cluster assignments
#' @param title Plot title (default: "Cluster Size Distribution")
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' clusters <- kmeans(iris[, 1:4], 3)$cluster
#' plot_cluster_sizes(clusters)
#' }
#' @export
plot_cluster_sizes <- function(clusters, title = "Cluster Size Distribution") {

  cluster_counts <- tibble::tibble(cluster = as.factor(clusters)) %>%
    dplyr::count(cluster)

  ggplot2::ggplot(
    cluster_counts,
    ggplot2::aes(x = cluster, y = n, fill = cluster)
  ) +
    ggplot2::geom_col() +
    ggplot2::geom_text(ggplot2::aes(label = n), vjust = -0.5) +
    ggplot2::labs(
      title = title,
      x = "Cluster",
      y = "Number of Observations"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(legend.position = "none")
}


#' Plot Variance Explained (PCA)
#'
#' Create combined scree plot showing individual and cumulative variance
#'
#' @param variance_tbl Variance tibble from tidy_pca
#' @param threshold Horizontal line for variance threshold
#'   (default: 0.8 for 80%)
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_model(iris[, 1:4], method = "pca")
#' plot_variance_explained(model$fit$variance_explained)
#' }
#' @export
plot_variance_explained <- function(variance_tbl, threshold = 0.8) {

  # Prepare data for plotting
  plot_data <- variance_tbl %>%
    dplyr::mutate(pc_num = seq_len(dplyr::n()))

  # Create dual-axis plot
  subtitle_text <- sprintf(
    "Red line: cumulative variance | Green line: %.0f%% threshold",
    threshold * 100
  )
  p1 <- ggplot2::ggplot(plot_data, ggplot2::aes(x = pc_num)) +
    ggplot2::geom_col(
      ggplot2::aes(y = prop_variance),
      fill = "steelblue",
      alpha = 0.7
    ) +
    ggplot2::geom_line(
      ggplot2::aes(y = cum_variance),
      color = "red",
      linewidth = 1
    ) +
    ggplot2::geom_point(
      ggplot2::aes(y = cum_variance),
      color = "red",
      size = 2
    ) +
    ggplot2::geom_hline(
      yintercept = threshold,
      linetype = "dashed",
      color = "darkgreen"
    ) +
    ggplot2::labs(
      title = "Variance Explained by Principal Components",
      subtitle = subtitle_text,
      x = "Principal Component",
      y = "Proportion of Variance Explained"
    ) +
    ggplot2::scale_y_continuous(
      labels = function(x) paste0(round(x * 100, 1), "%")
    ) +
    ggplot2::theme_minimal()

  p1
}


#' Plot Dendrogram with Cluster Highlights
#'
#' Enhanced dendrogram with colored cluster rectangles
#'
#' @param hclust_obj Hierarchical clustering object (hclust or tidy_hclust)
#' @param k Number of clusters to highlight
#' @param title Plot title
#'
#' @return Invisibly returns the \code{\link[stats]{hclust}} object. The
#'   dendrogram is drawn as a side effect.
#' @examples
#' \donttest{
#' hc <- hclust(dist(iris[, 1:4]))
#' plot_dendrogram(hc, k = 3)
#' }
#' @export
plot_dendrogram <- function(hclust_obj,
                            k = NULL,
                            title = "Hierarchical Clustering Dendrogram") {

  if (inherits(hclust_obj, "tidy_hclust")) {
    hc <- hclust_obj$model
  } else {
    hc <- hclust_obj
  }

  plot(hc, main = title, xlab = "", ylab = "Height", sub = "", cex = 0.7)

  if (!is.null(k)) {
    stats::rect.hclust(hc, k = k, border = 2:(k + 1))
  }

  invisible(hc)
}


#' Create Summary Dashboard
#'
#' Generate a multi-panel summary of clustering results
#'
#' @param data Data frame with cluster assignments
#' @param cluster_col Cluster column name
#' @param validation_metrics Optional tibble of validation metrics
#'
#' @return Invisibly returns a list of \code{\link[ggplot2]{ggplot}} objects.
#'   The combined plot grid is drawn as a side effect via
#'   \code{\link[gridExtra]{grid.arrange}}.
#' @examples
#' \donttest{
#' df <- iris[, 1:4]
#' df$cluster <- kmeans(df, 3)$cluster
#' create_cluster_dashboard(df)
#' }
#' @export
create_cluster_dashboard <- function(data,
                                     cluster_col = "cluster",
                                     validation_metrics = NULL) {

  plots <- list()

  # 1. Cluster scatter plot (first two numeric columns)
  numeric_cols <- names(data)[sapply(data, is.numeric)]
  if (length(numeric_cols) >= 2) {
    plots[[1]] <- plot_clusters(
      data, cluster_col = cluster_col,
      x_col = numeric_cols[1],
      y_col = numeric_cols[2]
    )
  }

  # 2. Cluster sizes
  plots[[2]] <- plot_cluster_sizes(data[[cluster_col]])

  # 3. If validation metrics provided, create metrics plot
  if (!is.null(validation_metrics)) {
    # Create a text plot with metrics
    fmt <- paste0(
      "Validation Metrics\n\n",
      "Number of Clusters: %d\n",
      "Avg Silhouette: %.3f\n",
      "Min Size: %d\nMax Size: %d"
    )
    metrics_text <- sprintf(
      fmt,
      validation_metrics$k,
      validation_metrics$avg_silhouette %||% NA,
      validation_metrics$min_size,
      validation_metrics$max_size
    )

    plots[[3]] <- ggplot2::ggplot() +
      ggplot2::annotate(
        "text", x = 0.5, y = 0.5,
        label = metrics_text, size = 5
      ) +
      ggplot2::theme_void()
  }

  # Combine plots
  if (length(plots) > 0) {
    gridExtra::grid.arrange(grobs = plots, ncol = 2)
  }

  invisible(plots)
}


#' Create Distance Heatmap
#'
#' Visualize distance matrix as heatmap
#'
#' @param dist_mat Distance matrix (dist object)
#' @param cluster_order Optional vector to reorder observations by cluster
#' @param title Plot title
#'
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' d <- dist(iris[1:20, 1:4])
#' plot_distance_heatmap(d)
#' }
#' @export
plot_distance_heatmap <- function(dist_mat,
                                  cluster_order = NULL,
                                  title = "Distance Heatmap") {

  # Convert to matrix
  dist_matrix <- as.matrix(dist_mat)

  # Reorder if cluster order provided
  if (!is.null(cluster_order)) {
    order_idx <- order(cluster_order)
    dist_matrix <- dist_matrix[order_idx, order_idx]
  }

  # Convert to long format
  dist_long <- dist_matrix %>%
    tibble::as_tibble(rownames = "id1") %>%
    tidyr::pivot_longer(-id1, names_to = "id2", values_to = "distance")

  # Create heatmap
  ggplot2::ggplot(dist_long, ggplot2::aes(x = id1, y = id2, fill = distance)) +
    ggplot2::geom_tile() +
    ggplot2::scale_fill_gradient(low = "white", high = "red") +
    ggplot2::labs(title = title, x = "", y = "", fill = "Distance") +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 90, hjust = 1, size = 6),
      axis.text.y = ggplot2::element_text(size = 6)
    )
}

# ---- Visualization functions ----

# -- Supervised visualization helpers --

test_that("tl_plot_actual_predicted returns ggplot for regression", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  p <- tl_plot_actual_predicted(model)

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_residuals returns ggplot for regression", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  p <- tl_plot_residuals(model, type = "fitted")

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_diagnostics returns a list of plots", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  plots <- tl_plot_diagnostics(model, which = 1:2)

  expect_type(plots, "list")
  expect_length(plots, 2)
})

test_that("tl_plot_confusion returns ggplot for binary classification", {
  data <- iris[iris$Species != "setosa", ]
  data$Species <- droplevels(data$Species)
  model <- tl_model(data, Species ~ ., method = "logistic")
  p <- tl_plot_confusion(model)

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_roc returns ggplot for binary classification", {
  # Create binary classification dataset
  data <- iris[iris$Species != "setosa", ]
  data$Species <- droplevels(data$Species)
  model <- tl_model(data, Species ~ ., method = "logistic")
  p <- tl_plot_roc(model)

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_importance works for tree-based models", {
  skip_if_not_installed("randomForest")

  model <- tl_model(mtcars, mpg ~ wt + hp + cyl, method = "forest")
  p <- tl_plot_importance(model)

  expect_s3_class(p, "ggplot")
})

# -- plot.tidylearn_model dispatch --

test_that("plot.tidylearn_model dispatches correctly for regression", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

  # Default type = "auto" should give actual_predicted for regression

  p <- plot(model)
  expect_s3_class(p, "ggplot")

  # Explicit type
  p2 <- plot(model, type = "residuals")
  expect_s3_class(p2, "ggplot")
})

test_that("plot.tidylearn_model dispatches correctly for classification", {
  data <- iris[iris$Species != "setosa", ]
  data$Species <- droplevels(data$Species)
  model <- tl_model(data, Species ~ ., method = "logistic")

  # Default type = "auto" should give confusion for classification
  p <- plot(model)
  expect_s3_class(p, "ggplot")
})

# -- Lift and gain charts --

test_that("tl_plot_lift works for binary classification", {
  data <- iris[iris$Species != "setosa", ]
  data$Species <- droplevels(data$Species)
  model <- tl_model(data, Species ~ ., method = "logistic")
  p <- tl_plot_lift(model, bins = 5)

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_gain works for binary classification", {
  data <- iris[iris$Species != "setosa", ]
  data$Species <- droplevels(data$Species)
  model <- tl_model(data, Species ~ ., method = "logistic")
  p <- tl_plot_gain(model, bins = 5)

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_lift errors for regression models", {
  model <- tl_model(mtcars, mpg ~ wt, method = "linear")

  expect_error(tl_plot_lift(model), "classification")
})

test_that("tl_plot_gain errors for regression models", {
  model <- tl_model(mtcars, mpg ~ wt, method = "linear")

  expect_error(tl_plot_gain(model), "classification")
})

# -- Model comparison --

test_that("tl_plot_model_comparison returns ggplot", {
  model1 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  model2 <- tl_model(mtcars, mpg ~ wt + hp, method = "polynomial", degree = 2)

  p <- tl_plot_model_comparison(model1, model2, names = c("Linear", "Poly"))

  expect_s3_class(p, "ggplot")
})

test_that("tl_plot_model_comparison errors for mixed model types", {
  model_reg <- tl_model(mtcars, mpg ~ wt, method = "linear")
  model_cls <- tl_model(iris, Species ~ ., method = "tree")

  expect_error(
    tl_plot_model_comparison(model_reg, model_cls),
    "same type"
  )
})

# -- Importance comparison --

test_that("tl_plot_importance_comparison works for tree-based models", {
  skip_if_not_installed("randomForest")

  model1 <- tl_model(mtcars, mpg ~ wt + hp + cyl, method = "forest")
  model2 <- tl_model(mtcars, mpg ~ wt + hp + cyl, method = "tree")
  p <- tl_plot_importance_comparison(model1, model2,
                                     names = c("Forest", "Tree"))

  expect_s3_class(p, "ggplot")
})

# -- CV results plotting --

test_that("tl_plot_cv_results returns ggplot", {
  cv_res <- tl_cv(mtcars, mpg ~ wt + hp, method = "linear", folds = 3)

  # tl_plot_cv_results expects fold_metrics and summary in a specific format

  # Build compatible structure
  fold_metrics <- do.call(rbind, lapply(seq_along(cv_res$folds), function(i) {
    df <- cv_res$folds[[i]]
    df$fold <- i
    df
  }))

  cv_input <- list(
    fold_metrics = fold_metrics,
    summary = dplyr::rename(cv_res$summary, mean_value = mean)
  )

  p <- tl_plot_cv_results(cv_input)
  expect_s3_class(p, "ggplot")
})

# -- Unsupervised visualization --

test_that("plot_clusters returns ggplot", {
  km <- tidy_kmeans(iris[, 1:4], k = 3)
  clustered_data <- augment_kmeans(km, iris[, 1:4])
  p <- plot_clusters(clustered_data)

  expect_s3_class(p, "ggplot")
})

test_that("plot_cluster_sizes returns ggplot", {
  clusters <- sample(1:3, 50, replace = TRUE)
  p <- plot_cluster_sizes(clusters)

  expect_s3_class(p, "ggplot")
})

test_that("plot_elbow returns ggplot", {
  wss <- calc_wss(iris[, 1:4], max_k = 5)
  p <- plot_elbow(wss)

  expect_s3_class(p, "ggplot")
})

test_that("plot_variance_explained returns ggplot", {
  pca_obj <- tidy_pca(iris[, 1:4])
  variance_tbl <- get_pca_variance(pca_obj)
  p <- plot_variance_explained(variance_tbl)

  expect_s3_class(p, "ggplot")
})

test_that("plot_dendrogram works", {
  hc <- tidy_hclust(iris[1:20, 1:4])
  # plot_dendrogram uses base graphics; just test it doesn't error
  expect_invisible(plot_dendrogram(hc, k = 3))
})

test_that("plot_distance_heatmap returns ggplot", {
  d <- dist(iris[1:15, 1:4])
  p <- plot_distance_heatmap(d)

  expect_s3_class(p, "ggplot")
})

# -- Dashboard (Shiny) --

test_that("tl_dashboard errors without shiny installed", {
  skip_if(requireNamespace("shiny", quietly = TRUE) &&
            requireNamespace("shinydashboard", quietly = TRUE) &&
            requireNamespace("DT", quietly = TRUE),
          "Shiny stack is installed, cannot test missing-package path")

  model <- tl_model(mtcars, mpg ~ wt, method = "linear")
  expect_error(tl_dashboard(model))
})

test_that("tl_dashboard returns shiny.appobj when packages available", {
  skip_if_not_installed("shiny")
  skip_if_not_installed("shinydashboard")
  skip_if_not_installed("DT")

  model <- tl_model(mtcars, mpg ~ wt, method = "linear")
  app <- tl_dashboard(model)

  expect_s3_class(app, "shiny.appobj")
})

#' @title High-Level Workflows for Common Machine Learning Patterns
#' @name tidylearn-workflows
#' @description Functions providing end-to-end workflows that showcase
#'   tidylearn's ability to seamlessly combine multiple learning paradigms
NULL

#' Auto ML: Automated Machine Learning Workflow
#'
#' Automatically explores multiple modeling approaches including
#' dimensionality reduction, clustering, and various supervised methods.
#' Returns the best performing model based on cross-validation.
#'
#' @param data A data frame
#' @param formula Model formula (for supervised learning)
#' @param task Task type: "classification", "regression", or "auto" (default)
#' @param use_reduction Whether to try dimensionality reduction (default: TRUE)
#' @param use_clustering Whether to add cluster features (default: TRUE)
#' @param time_budget Time budget in seconds (default: 300). Controls which
#'   models are attempted and whether cross-validation is used for evaluation.
#'   The budget is checked **between** model fits, not during them -- once a
#'   model starts training it runs to completion because R cannot safely
#'   interrupt C-level code (e.g. randomForest, xgboost, e1071).
#'
#'   How the budget shapes the workflow:
#'   \itemize{
#'     \item **Under 30s**: Only fast models are attempted
#'       (tree, logistic/linear).
#'       Cross-validation is skipped; models are ranked on training-set metrics
#'       only. Expect 2 models in the leaderboard. Use this for quick sanity
#'       checks or interactive exploration.
#'     \item **30--120s**: All baseline models are attempted including random
#'       forest. Cross-validation runs when enough time remains after each
#'       model fit; otherwise training metrics are used. Advanced models
#'       (SVM, XGBoost / ridge, lasso) are attempted if 40\% of the budget
#'       remains after baselines. Dimensionality reduction and clustering
#'       pipelines run if enabled and 10\% of the budget remains.
#'     \item **120s+ (recommended)**: The full pipeline runs -- all baselines,
#'       advanced models, PCA-augmented variants, and cluster-augmented
#'       variants, each with cross-validation. Expect 9--11 models in the
#'       leaderboard.
#'   }
#'
#'   Because individual model fits (especially forest, SVM, XGBoost with CV)
#'   can take 5--30s each depending on data size, the actual wall-clock time
#'   may modestly exceed the budget by the duration of the last model that
#'   was started before the budget expired.
#' @param cv_folds Number of cross-validation folds (default: 5). Reducing
#'   this (e.g. to 2 or 3) is an effective way to stay closer to the time
#'   budget since CV is typically the most expensive step.
#' @param metric Evaluation metric (default: auto-selected based on task).
#'   For classification: "accuracy"; for regression: "rmse".
#' @return A list with class \code{"tidylearn_automl"} containing:
#'   \describe{
#'     \item{best_model}{The best tidylearn model object}
#'     \item{models}{Named list of all successfully trained models}
#'     \item{leaderboard}{Tibble ranking models by the chosen metric}
#'     \item{task}{Detected or specified task type}
#'     \item{metric}{Metric used for ranking}
#'     \item{runtime}{Total elapsed time as a difftime object}
#'   }
#' @export
#' @examples
#' \donttest{
#' # Quick run with fast models only (< 30s budget skips forest/SVM/XGBoost)
#' result <- tl_auto_ml(iris, Species ~ .,
#'   time_budget = 10,
#'   use_reduction = FALSE,
#'   use_clustering = FALSE,
#'   cv_folds = 2)
#' result$leaderboard
#' }
tl_auto_ml <- function(data, formula, task = "auto",
                       use_reduction = TRUE, use_clustering = TRUE,
                       time_budget = 300, cv_folds = 5, metric = NULL) {
  start_time <- Sys.time()

  # Helper: remaining seconds in the budget
  time_remaining <- function() {
    as.numeric(difftime(Sys.time(), start_time, units = "secs"))
  }
  budget_left <- function() time_budget - time_remaining()

  # Helper: train a model with error handling.
  # Note: R cannot safely interrupt C-level code (randomForest, xgboost),
  # so we control budget by skipping models rather than killing them.
  safe_train <- function(expr_fn, label) {
    if (budget_left() <= 0) {
      message("    ", label, ": skipped (time budget exhausted)")
      return(NULL)
    }
    tryCatch(
      expr_fn(),
      error = function(e) {
        message("    ", label, ": failed - ", e$message)
        NULL
      }
    )
  }

  # Determine task type
  if (task == "auto") {
    response_var <- all.vars(formula)[1]
    y <- data[[response_var]]
    task <- if (is.factor(y) || is.character(y)) {
      "classification"
    } else {
      "regression"
    }
  }

  # Set default metric
  if (is.null(metric)) {
    metric <- if (task == "classification") "accuracy" else "rmse"
  }

  message("Starting Auto ML with task: ", task)
  message("Time budget: ", time_budget, " seconds")

  # Prepare candidate models
  models <- list()
  results <- list()

  # 1. Baseline models (ordered fast to slow)
  # Slow methods (forest, svm, xgboost) involve C code that cannot be

  # interrupted, so we only attempt them when the budget is generous.
  message("\n[1/4] Training baseline models...")
  fast_methods <- if (task == "classification") {
    c("tree", "logistic")
  } else {
    c("tree", "linear")
  }
  slow_methods <- "forest"  # C-level, ~5s+ per fit

  baseline_methods <- if (time_budget >= 30) {
    c(fast_methods, slow_methods)
  } else {
    fast_methods
  }

  for (method in baseline_methods) {
    if (budget_left() <= 0) break

    model_name <- paste0("baseline_", method)
    message("  Training: ", model_name)

    model <- safe_train(function() {
      tl_model(data, formula, method = method)
    }, model_name)
    if (is.null(model)) next

    # CV only when enough budget remains
    eval_result <- if (budget_left() > time_budget * 0.3) {
      safe_train(function() {
        tl_cv(data, formula, method = method, folds = cv_folds)
      }, paste0(model_name, " CV"))
    } else {
      NULL
    }
    if (is.null(eval_result)) eval_result <- tl_evaluate(model)

    models[[model_name]] <- model
    results[[model_name]] <- eval_result
  }

  # 2. Models with dimensionality reduction
  if (use_reduction && budget_left() > max(5, time_budget * 0.1)) {
    message("\n[2/4] Training models with dimensionality reduction...")

    tryCatch({
      response_var <- all.vars(formula)[1]
      n_predictors <- ncol(data) - 1
      n_components <- min(5, ceiling(n_predictors / 2))

      reduced <- tl_reduce_dimensions(
        data, response = response_var,
        method = "pca",
        n_components = n_components
      )

      # Update formula for reduced data -- exclude .obs_id
      pred_vars <- names(reduced$data)[
        !names(reduced$data) %in% c(".obs_id", response_var)
      ]
      formula_reduced <- as.formula(paste(
        response_var, "~",
        paste(pred_vars, collapse = " + ")
      ))

      # Drop .obs_id from reduced data
      reduced_data <- reduced$data
      if (".obs_id" %in% names(reduced_data)) {
        reduced_data <- reduced_data[, names(reduced_data) != ".obs_id",
                                     drop = FALSE]
      }

      for (method in baseline_methods) {
        if (budget_left() < max(2, time_budget * 0.05)) break

        model_name <- paste0("pca_", method)
        message("  Training: ", model_name)

        result <- safe_train(function() {
          model <- tl_model(reduced_data, formula_reduced, method = method)
          model$reduction_info <- list(
            reduction_model = reduced$reduction_model,
            n_components = n_components
          )
          eval_result <- tl_evaluate(model)
          list(model = model, eval_result = eval_result)
        }, model_name)

        if (!is.null(result)) {
          models[[model_name]] <- result$model
          results[[model_name]] <- result$eval_result
        }
      }
    }, error = function(e) {
      message("  Dimensionality reduction failed: ", e$message)
    })
  }

  # 3. Models with cluster features
  if (use_clustering && budget_left() > max(5, time_budget * 0.1)) {
    message("\n[3/4] Training models with cluster features...")

    tryCatch({
      response_var <- all.vars(formula)[1]
      k <- if (task == "classification") {
        length(unique(data[[response_var]]))
      } else {
        3
      }

      data_clustered <- tl_add_cluster_features(
        data, response = response_var,
        method = "kmeans", k = k
      )

      for (method in baseline_methods) {
        if (budget_left() < max(2, time_budget * 0.05)) break

        model_name <- paste0("clustered_", method)
        message("  Training: ", model_name)

        result <- safe_train(function() {
          model <- tl_model(data_clustered, formula, method = method)
          eval_result <- tl_evaluate(model)
          list(model = model, eval_result = eval_result)
        }, model_name)

        if (!is.null(result)) {
          models[[model_name]] <- result$model
          results[[model_name]] <- result$eval_result
        }
      }
    }, error = function(e) {
      message("  Cluster feature engineering failed: ", e$message)
    })
  }

  # 4. Advanced models if time allows (these are C-heavy and slow)
  message("\n[4/4] Training advanced models...")
  if (budget_left() > time_budget * 0.4 && time_budget >= 30) {
    advanced_methods <- if (task == "classification") {
      c("svm", "xgboost")
    } else {
      c("ridge", "lasso")
    }

    for (method in advanced_methods) {
      if (budget_left() <= 0) break

      model_name <- paste0("advanced_", method)
      message("  Training: ", model_name)

      model <- safe_train(function() {
        tl_model(data, formula, method = method)
      }, model_name)
      if (is.null(model)) next

      eval_result <- if (budget_left() > time_budget * 0.3) {
        safe_train(function() {
          tl_cv(data, formula, method = method, folds = cv_folds)
        }, paste0(model_name, " CV"))
      } else {
        NULL
      }
      if (is.null(eval_result)) eval_result <- tl_evaluate(model)

      models[[model_name]] <- model
      results[[model_name]] <- eval_result
    }
  }

  # Create leaderboard
  message("\n[*] Creating leaderboard...")
  leaderboard <- create_leaderboard(results, metric, task)

  # Get best model
  if (nrow(leaderboard) == 0 || length(models) == 0) {
    warning("No models were successfully trained within the time budget.",
            call. = FALSE)
    best_model_name <- NA_character_
    best_model <- NULL
  } else {
    best_model_name <- leaderboard$model[1]
    best_model <- models[[best_model_name]]
  }

  total_time <- difftime(Sys.time(), start_time, units = "secs")
  message("\nAuto ML complete in ", round(total_time, 2), " seconds")
  message("Best model: ", best_model_name)

  structure(
    list(
      best_model = best_model,
      models = models,              # Add for test compatibility
      all_models = models,          # Keep for backward compatibility
      results = results,            # Add for test compatibility
      leaderboard = leaderboard,
      task = task,
      metric = metric,
      runtime = total_time
    ),
    class = c("tidylearn_automl", "list")
  )
}

#' Create leaderboard from results
#' @keywords internal
#' @noRd
create_leaderboard <- function(results, metric, task) {
  if (length(results) == 0) {
    return(tibble::tibble(model = character(0), score = numeric(0)))
  }

  scores <- vapply(results, function(r) {
    if (is.list(r) && metric %in% names(r)) {
      as.numeric(r[[metric]])
    } else if (is.list(r) && "metrics" %in% names(r)) {
      val <- r$metrics[[metric]]
      if (is.null(val)) NA_real_ else as.numeric(val)
    } else {
      NA_real_
    }
  }, numeric(1))

  leaderboard <- tibble::tibble(
    model = names(scores),
    score = scores
  )

  # Sort: ascending for error metrics, descending for accuracy metrics
  if (metric %in% c("rmse", "mae", "mse")) {
    leaderboard <- leaderboard %>% dplyr::arrange(score)
  } else {
    leaderboard <- leaderboard %>% dplyr::arrange(dplyr::desc(score))
  }

  leaderboard
}

#' Print auto ML results
#' @param x A tidylearn_automl object
#' @param ... Additional arguments (ignored)
#' @return The input object \code{x}, returned invisibly.
#' @export
print.tidylearn_automl <- function(x, ...) {
  cat("tidylearn Auto ML Results\n")
  cat("=========================\n")
  cat("Task:", x$task, "\n")
  cat("Metric:", x$metric, "\n")
  cat("Runtime:", round(x$runtime, 2), "seconds\n")
  cat("Models trained:", length(x$all_models), "\n\n")

  cat("Leaderboard:\n")
  print(x$leaderboard, n = 10)

  cat("\nBest model:", x$leaderboard$model[1], "\n")
  cat("Best score:", x$leaderboard$score[1], "\n")

  invisible(x)
}

#' Exploratory Data Analysis Workflow
#'
#' Comprehensive EDA combining unsupervised learning techniques
#' to understand data structure before modeling
#'
#' @param data A data frame
#' @param response Optional response variable for colored visualizations
#' @param max_components Maximum PCA components to compute (default: 5)
#' @param k_range Range of k values for clustering (default: 2:6)
#' @return A list with class \code{"tidylearn_eda"} containing:
#'   \describe{
#'     \item{data}{The original data frame.}
#'     \item{response}{The response variable name, or \code{NULL}.}
#'     \item{pca}{The fitted PCA model.}
#'     \item{optimal_k}{List with optimal cluster count results.}
#'     \item{kmeans}{The fitted k-means model.}
#'     \item{hclust}{The fitted hierarchical clustering model.}
#'     \item{summary}{List with \code{n_obs}, \code{n_vars},
#'       \code{n_components}, and \code{best_k}.}
#'   }
#' @export
#' @examples
#' \donttest{
#' eda <- tl_explore(iris, response = "Species")
#' plot(eda)
#' }
tl_explore <- function(data, response = NULL,
                       max_components = 5,
                       k_range = 2:6) {
  message("Running Exploratory Data Analysis...")

  # 1. Dimensionality Reduction
  message("[1/4] PCA analysis...")
  predictor_data <- if (!is.null(response)) {
    data %>% dplyr::select(-dplyr::all_of(response))
  } else {
    data
  }

  pca_result <- tl_model(predictor_data, method = "pca")

  # 2. Optimal clustering
  message("[2/4] Finding optimal clusters...")
  optimal_k <- tl_optimal_clusters(predictor_data, k_range = k_range)

  # 3. Cluster analysis with optimal k
  message("[3/4] Clustering analysis...")
  k_best <- optimal_k$best_k
  kmeans_result <- tl_model(predictor_data, method = "kmeans", k = k_best)
  hclust_result <- tl_model(predictor_data, method = "hclust")

  # 4. Distance analysis
  message("[4/4] Distance analysis...")

  message("EDA complete!")

  structure(
    list(
      data = data,
      response = response,
      pca = pca_result,
      optimal_k = optimal_k,
      kmeans = kmeans_result,
      hclust = hclust_result,
      summary = list(
        n_obs = nrow(data),
        n_vars = ncol(predictor_data),
        n_components = max_components,
        best_k = k_best
      )
    ),
    class = c("tidylearn_eda", "list")
  )
}

#' Print EDA results
#' @param x A tidylearn_eda object
#' @param ... Additional arguments (ignored)
#' @return The input object \code{x}, returned invisibly.
#' @examples
#' \donttest{
#' eda <- tl_explore(iris, response = "Species")
#' print(eda)
#' }
#' @export
print.tidylearn_eda <- function(x, ...) {
  cat("tidylearn Exploratory Data Analysis\n")
  cat("===================================\n")
  cat("Observations:", x$summary$n_obs, "\n")
  cat("Variables:", x$summary$n_vars, "\n")
  cat("Optimal clusters:", x$summary$best_k, "\n\n")

  cat("PCA Variance Explained (first 5 components):\n")
  print(head(x$pca$fit$variance, 5))

  cat("\nCluster sizes (k =", x$summary$best_k, "):\n")
  print(table(x$kmeans$fit$clusters$cluster))

  invisible(x)
}

#' Plot EDA results
#' @param x A tidylearn_eda object
#' @param ... Additional arguments (ignored)
#' @return The input object \code{x}, returned invisibly. Called for its
#'   side effect of plotting a PCA scatter plot coloured by cluster.
#' @examples
#' \donttest{
#' eda <- tl_explore(iris, response = "Species")
#' plot(eda)
#' }
#' @export
plot.tidylearn_eda <- function(x, ...) {
  # Get PCA scores for visualization
  pca_scores <- x$pca$fit$scores
  clusters <- x$kmeans$fit$clusters$cluster

  # Create plot data
  plot_data <- data.frame(
    PC1 = pca_scores$PC1,
    PC2 = pca_scores$PC2,
    Cluster = as.factor(clusters)
  )

  # Add response if available
  if (!is.null(x$response) && x$response %in% names(x$data)) {
    plot_data$Response <- x$data[[x$response]]
  }

  # Create the plot (use .data$ to avoid R CMD check NOTEs)
  p <- ggplot2::ggplot(plot_data, ggplot2::aes(
    x = .data[["PC1"]],
    y = .data[["PC2"]],
    color = .data[["Cluster"]]
  )) +
    ggplot2::geom_point(size = 2, alpha = 0.7) +
    ggplot2::labs(
      title = "EDA: PCA with K-means Clusters",
      subtitle = paste("k =", x$summary$best_k, "clusters"),
      x = "Principal Component 1",
      y = "Principal Component 2"
    ) +
    ggplot2::theme_minimal()

  print(p)
  invisible(x)
}

#' Find optimal number of clusters
#' @keywords internal
#' @noRd
tl_optimal_clusters <- function(data, k_range = 2:6, method = "silhouette") {
  scores <- numeric(length(k_range))

  for (i in seq_along(k_range)) {
    k <- k_range[i]
    km <- tl_model(data, method = "kmeans", k = k)

    # Compute silhouette score
    if (requireNamespace("cluster", quietly = TRUE)) {
      dist_mat <- stats::dist(
        dplyr::select(data, where(is.numeric))
      )
      sil <- cluster::silhouette(
        km$fit$clusters$cluster, dist_mat
      )
      scores[i] <- mean(sil[, 3])
    } else {
      # Fallback to within-cluster sum of squares
      scores[i] <- -km$fit$metrics$tot_withinss
    }
  }

  best_idx <- which.max(scores)

  list(
    k_values = k_range,
    scores = scores,
    best_k = k_range[best_idx],
    best_score = scores[best_idx]
  )
}

#' Transfer Learning Workflow
#'
#' Use unsupervised pre-training (e.g., autoencoder
#' features) before supervised learning
#'
#' @param data Training data
#' @param formula Model formula
#' @param pretrain_method Pre-training method: "pca", "autoencoder"
#' @param supervised_method Supervised learning method
#' @param ... Additional arguments
#' @return A list with class \code{"tidylearn_transfer"} containing:
#'   \describe{
#'     \item{pretrain_model}{The fitted dimensionality reduction model.}
#'     \item{supervised_model}{The fitted supervised tidylearn model.}
#'     \item{formula}{The model formula.}
#'     \item{method}{The supervised learning method used.}
#'   }
#' @export
#' @examples
#' \donttest{
#' model <- tl_transfer_learning(iris, Species ~ .,
#'   pretrain_method = "pca", supervised_method = "logistic")
#' }
tl_transfer_learning <- function(data, formula, pretrain_method = "pca",
                                 supervised_method = "logistic", ...) {
  message("Transfer Learning Workflow")
  message("==========================")

  response_var <- all.vars(formula)[1]

  # Phase 1: Unsupervised pre-training
  message("[Phase 1] Unsupervised pre-training with ", pretrain_method, "...")
  pretrain_model <- tl_reduce_dimensions(
    data, response = response_var,
    method = pretrain_method, ...
  )

  # Phase 2: Supervised learning on transformed features
  message("[Phase 2] Supervised learning with ", supervised_method, "...")

  # Remove .obs_id from PCA/MDS output (row identifier, not a feature)

  supervised_data <- pretrain_model$data
  if (".obs_id" %in% names(supervised_data)) {
    supervised_data <- supervised_data[, names(supervised_data) != ".obs_id",
                                       drop = FALSE]
  }

  supervised_model <- tl_model(
    supervised_data, formula,
    method = supervised_method
  )

  # Combine models
  structure(
    list(
      pretrain_model = pretrain_model$reduction_model,
      supervised_model = supervised_model,
      formula = formula,
      method = supervised_method
    ),
    class = c("tidylearn_transfer", "list")
  )
}

#' Predict with transfer learning model
#' @param object A tidylearn_transfer model object
#' @param new_data New data for predictions
#' @param ... Additional arguments
#' @return A \link[tibble]{tibble} with a \code{.pred} column containing
#'   predictions.
#' @examples
#' \donttest{
#' model <- tl_transfer_learning(iris, Species ~ .,
#'   pretrain_method = "pca", supervised_method = "logistic")
#' preds <- predict(model, iris[1:5, ])
#' }
#' @export
predict.tidylearn_transfer <- function(object, new_data, ...) {
  # Transform new data using pre-trained model
  transformed <- predict(object$pretrain_model, new_data = new_data)

  # Remove .obs_id from transformed data (row identifier, not a feature)
  if (".obs_id" %in% names(transformed)) {
    transformed <- transformed[, names(transformed) != ".obs_id", drop = FALSE]
  }

  # Predict using supervised model
  predict(object$supervised_model, new_data = transformed, ...)
}

test_that("tl_auto_ml runs with basic settings", {
  skip_on_cran()
  skip_if_not_installed("randomForest")

  # Run with short time budget for testing
  result <- tl_auto_ml(iris, Species ~ .,
                       use_reduction = TRUE,
                       use_clustering = TRUE,
                       time_budget = 10)

  expect_type(result, "list")
  expect_true("models" %in% names(result))
  expect_true("results" %in% names(result))
  expect_true("best_model" %in% names(result))

  # Should have trained at least one model
  expect_gte(length(result$models), 1)
})

test_that("tl_auto_ml detects task type automatically", {
  skip_on_cran()

  # Classification task
  result_class <- tl_auto_ml(iris, Species ~ .,
                             task = "auto",
                             time_budget = 5)

  expect_type(result_class, "list")

  # Regression task
  result_reg <- tl_auto_ml(mtcars, mpg ~ wt + hp,
                           task = "auto",
                           time_budget = 5)

  expect_type(result_reg, "list")
})

test_that("tl_auto_ml respects time budget", {
  skip_on_cran()

  # Very short time budget
  start_time <- Sys.time()
  result <- tl_auto_ml(iris, Species ~ .,
                       time_budget = 5,
                       use_reduction = FALSE,
                       use_clustering = FALSE)
  end_time <- Sys.time()

  # Should complete within reasonable time (with some buffer)
  elapsed <- as.numeric(difftime(end_time, start_time, units = "secs"))
  expect_lt(elapsed, 20)  # Should finish well before 20 seconds
})

test_that("tl_auto_ml can disable reduction and clustering", {
  skip_on_cran()

  result <- tl_auto_ml(iris, Species ~ .,
                       use_reduction = FALSE,
                       use_clustering = FALSE,
                       time_budget = 10)

  expect_type(result, "list")

  # Should still have baseline models
  expect_gte(length(result$models), 1)

  # Model names should not include "pca_" or "clustered_"
  model_names <- names(result$models)
  expect_false(any(grepl("^pca_", model_names)))
  expect_false(any(grepl("^clustered_", model_names)))
})

test_that("tl_auto_ml handles small datasets", {
  skip_on_cran()

  # Small dataset
  small_data <- iris[1:30, ]

  result <- tl_auto_ml(small_data, Species ~ .,
                       time_budget = 5,
                       cv_folds = 3)

  expect_type(result, "list")
  expect_gte(length(result$models), 1)
})

test_that("tl_auto_ml returns best model", {
  skip_on_cran()

  result <- tl_auto_ml(iris, Species ~ .,
                       time_budget = 10)

  expect_true("best_model" %in% names(result))
  expect_s3_class(result$best_model, "tidylearn_model")

  # Best model should be one of the trained models
  expect_true(!is.null(result$best_model))
})

test_that("tl_auto_ml works with regression tasks", {
  skip_on_cran()

  result <- tl_auto_ml(mtcars, mpg ~ .,
                       task = "regression",
                       time_budget = 10)

  expect_type(result, "list")
  expect_gte(length(result$models), 1)
})

test_that("tl_auto_ml handles errors gracefully", {
  skip_on_cran()

  # Should not crash even if some models fail
  result <- tl_auto_ml(iris, Species ~ .,
                       time_budget = 5)

  expect_type(result, "list")
  # Should have trained at least one successful model
  expect_gte(length(result$models), 1)
})

#' Data Preprocessing for tidylearn
#'
#' Unified preprocessing functions that work with both
#' supervised and unsupervised workflows

#' Prepare Data for Machine Learning
#'
#' Comprehensive preprocessing pipeline including imputation, scaling,
#' encoding, and feature engineering
#'
#' @param data A data frame
#' @param formula Optional formula (for supervised learning)
#' @param impute_method Method for missing value imputation:
#'   "mean", "median", "mode", "knn"
#' @param scale_method Scaling method: "standardize",
#'   "normalize", "robust", "none"
#' @param encode_categorical Whether to encode categorical
#'   variables (default: TRUE)
#' @param remove_zero_variance Remove zero-variance features (default: TRUE)
#' @param remove_correlated Remove highly correlated features (default: FALSE)
#' @param correlation_cutoff Correlation threshold for removal (default: 0.95)
#' @return A list with components:
#'   \describe{
#'     \item{\code{data}}{The processed data frame.}
#'     \item{\code{original_data}}{The original unprocessed data frame.}
#'     \item{\code{preprocessing_steps}}{A list of metadata for each
#'       preprocessing step applied (imputation values, encoding maps,
#'       scaling parameters, etc.).}
#'     \item{\code{formula}}{The formula passed in (or \code{NULL}).}
#'   }
#' @export
#' @examples
#' \donttest{
#' processed <- tl_prepare_data(iris, Species ~ ., scale_method = "standardize")
#' model <- tl_model(processed$data, Species ~ ., method = "tree")
#' }
tl_prepare_data <- function(data, formula = NULL,
                            impute_method = "mean",
                            scale_method = "standardize",
                            encode_categorical = TRUE,
                            remove_zero_variance = TRUE,
                            remove_correlated = FALSE,
                            correlation_cutoff = 0.95) {

  processed_data <- data
  preprocessing_steps <- list()

  # Extract response if formula provided
  response_var <- NULL
  if (!is.null(formula)) {
    response_var <- all.vars(formula)[1]
  }

  # Separate predictors and response
  if (!is.null(response_var)) {
    response_data <- processed_data[[response_var]]
    predictor_data <- processed_data %>%
      dplyr::select(-dplyr::all_of(response_var))
  } else {
    response_data <- NULL
    predictor_data <- processed_data
  }

  # 1. Handle missing values
  if (any(is.na(predictor_data))) {
    message("Imputing missing values using method: ", impute_method)
    imputation_info <- impute_missing(predictor_data, method = impute_method)
    predictor_data <- imputation_info$data
    preprocessing_steps$imputation <- imputation_info
  }

  # 2. Encode categorical variables
  if (encode_categorical) {
    cat_vars <- names(predictor_data)[
      sapply(
        predictor_data,
        function(x) is.factor(x) || is.character(x)
      )
    ]

    if (length(cat_vars) > 0) {
      message("Encoding ", length(cat_vars), " categorical variables")
      encoding_info <- encode_categoricals(predictor_data, cat_vars)
      predictor_data <- encoding_info$data
      preprocessing_steps$encoding <- encoding_info
    }
  }

  # 3. Remove zero variance features
  if (remove_zero_variance) {
    zero_var_cols <- find_zero_variance(predictor_data)
    if (length(zero_var_cols) > 0) {
      message("Removing ", length(zero_var_cols), " zero-variance features")
      predictor_data <- predictor_data %>%
        dplyr::select(-dplyr::all_of(zero_var_cols))
      preprocessing_steps$zero_variance <- zero_var_cols
    }
  }

  # 4. Remove highly correlated features
  if (remove_correlated) {
    numeric_data <- predictor_data %>% dplyr::select(where(is.numeric))
    if (ncol(numeric_data) > 1) {
      cor_matrix <- stats::cor(numeric_data, use = "pairwise.complete.obs")
      high_cor <- find_high_correlation(cor_matrix, cutoff = correlation_cutoff)

      if (length(high_cor) > 0) {
        message("Removing ", length(high_cor), " highly correlated features")
        predictor_data <- predictor_data %>%
          dplyr::select(-dplyr::all_of(high_cor))
        preprocessing_steps$high_correlation <- high_cor
      }
    }
  }

  # 5. Scale numeric features
  if (scale_method != "none") {
    numeric_cols <- names(predictor_data)[sapply(predictor_data, is.numeric)]

    if (length(numeric_cols) > 0) {
      message("Scaling numeric features using method: ", scale_method)
      scaling_info <- scale_features(
        predictor_data, numeric_cols,
        method = scale_method
      )
      predictor_data <- scaling_info$data
      preprocessing_steps$scaling <- scaling_info
    }
  }

  # Recombine with response
  if (!is.null(response_var)) {
    processed_data <- predictor_data %>%
      dplyr::mutate(!!response_var := response_data)
  } else {
    processed_data <- predictor_data
  }

  list(
    data = processed_data,
    original_data = data,
    preprocessing_steps = preprocessing_steps,
    formula = formula
  )
}

#' Impute missing values
#' @keywords internal
#' @noRd
impute_missing <- function(data, method = "mean") {
  imputed_data <- data
  imputation_values <- list()

  numeric_cols <- names(data)[sapply(data, is.numeric)]

  for (col in numeric_cols) {
    if (any(is.na(data[[col]]))) {
      if (method == "mean") {
        impute_val <- mean(data[[col]], na.rm = TRUE)
      } else if (method == "median") {
        impute_val <- stats::median(data[[col]], na.rm = TRUE)
      } else {
        impute_val <- mean(data[[col]], na.rm = TRUE)
      }

      imputed_data[[col]][is.na(imputed_data[[col]])] <- impute_val
      imputation_values[[col]] <- impute_val
    }
  }

  list(
    data = imputed_data,
    method = method,
    imputation_values = imputation_values
  )
}

#' Encode categorical variables
#' @keywords internal
#' @noRd
encode_categoricals <- function(data, cat_vars) {
  encoded_data <- data
  encoding_map <- list()

  for (var in cat_vars) {
    if (is.character(data[[var]])) {
      encoded_data[[var]] <- as.factor(data[[var]])
    }

    # One-hot encode if more than 2 levels
    if (nlevels(encoded_data[[var]]) > 2) {
      # Create dummy variables
      dummies <- stats::model.matrix(
        ~ . - 1,
        data = data.frame(x = encoded_data[[var]])
      )
      colnames(dummies) <- paste0(var, "_", gsub("^x", "", colnames(dummies)))

      # Remove original column and add dummies
      encoded_data <- encoded_data %>%
        dplyr::select(-dplyr::all_of(var)) %>%
        dplyr::bind_cols(as.data.frame(dummies))

      encoding_map[[var]] <- colnames(dummies)
    }
  }

  list(
    data = encoded_data,
    encoding_map = encoding_map
  )
}

#' Find zero variance columns
#' @keywords internal
#' @noRd
find_zero_variance <- function(data) {
  numeric_data <- data %>% dplyr::select(where(is.numeric))

  zero_var <- sapply(numeric_data, function(x) {
    stats::var(x, na.rm = TRUE) == 0 || all(is.na(x))
  })

  names(zero_var)[zero_var]
}

#' Find highly correlated features
#' @keywords internal
#' @noRd
find_high_correlation <- function(cor_matrix, cutoff = 0.95) {
  cor_matrix[lower.tri(cor_matrix, diag = TRUE)] <- 0

  high_cor_pairs <- which(abs(cor_matrix) > cutoff, arr.ind = TRUE)

  if (nrow(high_cor_pairs) == 0) {
    return(character(0))
  }

  # For each pair, remove the one with higher average correlation
  to_remove <- character()
  for (i in seq_len(nrow(high_cor_pairs))) {
    row_idx <- high_cor_pairs[i, 1]
    col_idx <- high_cor_pairs[i, 2]

    row_name <- rownames(cor_matrix)[row_idx]
    col_name <- colnames(cor_matrix)[col_idx]

    # Calculate average correlation
    row_avg_cor <- mean(abs(cor_matrix[row_idx, ]), na.rm = TRUE)
    col_avg_cor <- mean(abs(cor_matrix[, col_idx]), na.rm = TRUE)

    if (row_avg_cor > col_avg_cor) {
      to_remove <- c(to_remove, row_name)
    } else {
      to_remove <- c(to_remove, col_name)
    }
  }

  unique(to_remove)
}

#' Scale numeric features
#' @keywords internal
#' @noRd
scale_features <- function(data, numeric_cols, method = "standardize") {
  scaled_data <- data
  scaling_params <- list()

  for (col in numeric_cols) {
    if (method == "standardize") {
      # Z-score standardization
      mean_val <- mean(data[[col]], na.rm = TRUE)
      sd_val <- stats::sd(data[[col]], na.rm = TRUE)

      if (sd_val > 0) {
        scaled_data[[col]] <- (data[[col]] - mean_val) / sd_val
        scaling_params[[col]] <- list(mean = mean_val, sd = sd_val)
      }

    } else if (method == "normalize") {
      # Min-max normalization
      min_val <- min(data[[col]], na.rm = TRUE)
      max_val <- max(data[[col]], na.rm = TRUE)

      if (max_val > min_val) {
        scaled_data[[col]] <- (data[[col]] - min_val) / (max_val - min_val)
        scaling_params[[col]] <- list(min = min_val, max = max_val)
      }

    } else if (method == "robust") {
      # Robust scaling using median and IQR
      median_val <- stats::median(data[[col]], na.rm = TRUE)
      q1 <- stats::quantile(data[[col]], 0.25, na.rm = TRUE)
      q3 <- stats::quantile(data[[col]], 0.75, na.rm = TRUE)
      iqr_val <- q3 - q1

      if (iqr_val > 0) {
        scaled_data[[col]] <- (data[[col]] - median_val) / iqr_val
        scaling_params[[col]] <- list(median = median_val, iqr = iqr_val)
      }
    }
  }

  list(
    data = scaled_data,
    method = method,
    scaling_params = scaling_params
  )
}

#' Split data into train and test sets
#'
#' @param data A data frame
#' @param prop Proportion for training set (default: 0.8)
#' @param stratify Column name for stratified splitting
#' @param seed Random seed for reproducibility
#' @return A list with two elements:
#'   \describe{
#'     \item{\code{$train}}{A data frame containing the training subset.}
#'     \item{\code{$test}}{A data frame containing the test subset.}
#'   }
#' @export
#' @examples
#' \donttest{
#' split_data <- tl_split(iris, prop = 0.7, stratify = "Species")
#' train <- split_data$train
#' test <- split_data$test
#' }
tl_split <- function(data, prop = 0.8, stratify = NULL, seed = NULL) {
  if (!is.null(seed)) {
    set.seed(seed)
  }

  n <- nrow(data)

  if (!is.null(stratify)) {
    if (!stratify %in% names(data)) {
      stop("Stratify variable not found in data", call. = FALSE)
    }

    # Stratified sampling
    groups <- split(seq_len(n), data[[stratify]])
    train_indices <- unlist(lapply(groups, function(idx) {
      sample(idx, size = floor(length(idx) * prop))
    }))

  } else {
    # Simple random sampling
    train_indices <- sample(seq_len(n), size = floor(n * prop))
  }

  list(
    train = data[train_indices, ],
    test = data[-train_indices, ]
  )
}

test_that("tl_prepare_data handles missing values", {
  # Create data with missing values
  data_missing <- iris
  data_missing[1:5, "Sepal.Length"] <- NA
  data_missing[10:15, "Petal.Width"] <- NA

  # Prepare data with imputation
  result <- tl_prepare_data(data_missing, Species ~ .,
                            impute_method = "mean",
                            scale_method = "none",
                            encode_categorical = FALSE)

  # Check that NAs are imputed
  expect_false(any(is.na(result$data)))
  expect_true("imputation" %in% names(result$preprocessing_steps))
})

test_that("tl_prepare_data scales features correctly", {
  # Standardization
  result_std <- tl_prepare_data(iris, Species ~ .,
                                impute_method = "mean",
                                scale_method = "standardize",
                                encode_categorical = FALSE)

  numeric_cols <- sapply(result_std$data, is.numeric)
  numeric_data <- result_std$data[, numeric_cols]

  # Check means are close to 0 and sds close to 1 (excluding response)
  means <- colMeans(numeric_data[, names(numeric_data) != "Species"])
  expect_true(all(abs(means) < 1e-10))

  # Normalization
  result_norm <- tl_prepare_data(iris, Species ~ .,
                                 impute_method = "mean",
                                 scale_method = "normalize",
                                 encode_categorical = FALSE)

  numeric_data_norm <- result_norm$data[, numeric_cols]
  # Check values are in [0, 1]
  expect_true(
    all(numeric_data_norm >= 0 & numeric_data_norm <= 1, na.rm = TRUE)
  )
})

test_that("tl_prepare_data encodes categorical variables", {
  # Create data with categorical variable
  test_data <- data.frame(
    x1 = rnorm(100),
    x2 = rnorm(100),
    cat_var = factor(rep(c("A", "B", "C"), length.out = 100)),
    y = rnorm(100)
  )

  result <- tl_prepare_data(test_data, y ~ .,
                            encode_categorical = TRUE,
                            scale_method = "none")

  # Original categorical variable should be replaced with dummies
  expect_false("cat_var" %in% names(result$data))
  expect_true(any(grepl("cat_var_", names(result$data))))
})

test_that("tl_prepare_data removes zero variance features", {
  # Create data with zero variance column
  test_data <- iris
  test_data$zero_var <- 1

  result <- tl_prepare_data(test_data, Species ~ .,
                            remove_zero_variance = TRUE,
                            scale_method = "none",
                            encode_categorical = FALSE)

  # Zero variance column should be removed
  expect_false("zero_var" %in% names(result$data))
  expect_true("zero_variance" %in% names(result$preprocessing_steps))
})

test_that("tl_prepare_data removes highly correlated features", {
  # Create data with highly correlated columns
  test_data <- iris
  test_data$Sepal.Length.Copy <-
    test_data$Sepal.Length + rnorm(nrow(iris), 0, 0.01)

  result <- tl_prepare_data(test_data, Species ~ .,
                            remove_correlated = TRUE,
                            correlation_cutoff = 0.95,
                            scale_method = "none",
                            encode_categorical = FALSE)

  # One of the correlated columns should be removed
  has_original <- "Sepal.Length" %in% names(result$data)
  has_copy <- "Sepal.Length.Copy" %in% names(result$data)

  expect_true(xor(has_original, has_copy))
})

test_that("tl_split creates train/test splits correctly", {
  # Simple split
  split <- tl_split(iris, prop = 0.7, seed = 123)

  expect_type(split, "list")
  expect_equal(names(split), c("train", "test"))
  expect_equal(nrow(split$train), 105)
  expect_equal(nrow(split$test), 45)
  expect_equal(nrow(split$train) + nrow(split$test), nrow(iris))

  # Check no overlap
  train_idx <- as.numeric(rownames(split$train))
  test_idx <- as.numeric(rownames(split$test))
  expect_equal(length(intersect(train_idx, test_idx)), 0)
})

test_that("tl_split supports stratified splitting", {
  # Stratified split
  split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

  # Check proportions are maintained
  train_props <- prop.table(table(split$train$Species))
  test_props <- prop.table(table(split$test$Species))
  original_props <- prop.table(table(iris$Species))

  # Proportions should be similar (within 5%)
  expect_true(all(abs(train_props - original_props) < 0.05))
  expect_true(all(abs(test_props - original_props) < 0.05))
})

test_that("tl_split validates inputs", {
  expect_error(
    tl_split(iris, prop = 0.7, stratify = "NonexistentColumn"),
    "Stratify variable not found"
  )
})

test_that("tl_prepare_data preserves response variable", {
  result <- tl_prepare_data(iris, Species ~ .,
                            scale_method = "standardize",
                            encode_categorical = FALSE)

  # Response should be present and unchanged
  expect_true("Species" %in% names(result$data))
  expect_equal(result$data$Species, iris$Species)
})

test_that("linear regression models work", {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

  expect_s3_class(model, "tidylearn_linear")
  expect_false(model$spec$is_classification)

  # Predictions should be numeric
  preds <- predict(model)
  expect_type(preds$.pred, "double")
  expect_equal(nrow(preds), nrow(mtcars))
})

test_that("logistic regression models work for classification", {
  model <- tl_model(iris, Species ~ ., method = "logistic")

  expect_s3_class(model, "tidylearn_logistic")
  expect_true(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(iris))
})

test_that("tree models work for classification", {
  skip_if_not_installed("rpart")

  model <- tl_model(iris, Species ~ Sepal.Length + Sepal.Width, method = "tree")

  expect_s3_class(model, "tidylearn_tree")
  expect_true(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(iris))
})

test_that("tree models work for regression", {
  skip_if_not_installed("rpart")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "tree")

  expect_s3_class(model, "tidylearn_tree")
  expect_false(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_type(preds$.pred, "double")
})

test_that("random forest models work for classification", {
  skip_if_not_installed("randomForest")

  model <- tl_model(iris, Species ~ ., method = "forest")

  expect_s3_class(model, "tidylearn_forest")
  expect_true(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(iris))
})

test_that("random forest models work for regression", {
  skip_if_not_installed("randomForest")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "forest")

  expect_s3_class(model, "tidylearn_forest")
  expect_false(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_type(preds$.pred, "double")
})

test_that("ridge regression works", {
  skip_if_not_installed("glmnet")

  model <- tl_model(mtcars, mpg ~ ., method = "ridge")

  expect_s3_class(model, "tidylearn_ridge")

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(mtcars))
})

test_that("lasso regression works", {
  skip_if_not_installed("glmnet")

  model <- tl_model(mtcars, mpg ~ ., method = "lasso")

  expect_s3_class(model, "tidylearn_lasso")

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(mtcars))
})

test_that("elastic net works", {
  skip_if_not_installed("glmnet")

  model <- tl_model(mtcars, mpg ~ ., method = "elastic_net", alpha = 0.5)

  expect_s3_class(model, "tidylearn_elastic_net")

  # Predictions
  preds <- predict(model)
  expect_equal(nrow(preds), nrow(mtcars))
})

test_that("polynomial regression works", {
  model <- tl_model(mtcars, mpg ~ wt, method = "polynomial", degree = 2)

  expect_s3_class(model, "tidylearn_polynomial")
  expect_false(model$spec$is_classification)

  # Predictions
  preds <- predict(model)
  expect_type(preds$.pred, "double")
})

test_that("supervised models handle new data correctly", {
  # Split data
  split <- tl_split(iris, prop = 0.7, seed = 123)

  # Train on training set
  model <- tl_model(split$train, Species ~ ., method = "logistic")

  # Predict on test set
  preds <- predict(model, new_data = split$test)

  expect_equal(nrow(preds), nrow(split$test))
})

test_that("supervised models work with formula variations", {
  # Formula with interaction
  model1 <- tl_model(mtcars, mpg ~ wt * hp, method = "linear")
  expect_s3_class(model1, "tidylearn_linear")

  # Formula with all variables
  model2 <- tl_model(iris, Species ~ ., method = "logistic")
  expect_s3_class(model2, "tidylearn_logistic")

  # Formula with subset of variables
  model3 <- tl_model(iris, Species ~ Sepal.Length + Petal.Length,
                     method = "logistic")
  expect_s3_class(model3, "tidylearn_logistic")
})

#' @title Hyperparameter Tuning Functions for tidylearn
#' @name tidylearn-tuning
#' @description
#' Functions for automatic hyperparameter tuning and selection
#' @importFrom stats model.matrix as.formula
#' @importFrom dplyr filter select mutate arrange
#' @importFrom tidyr crossing
NULL

#' Tune hyperparameters for a model using grid search
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param method The modeling method to tune
#' @param param_grid A named list of parameter values
#'   to tune
#' @param folds Number of cross-validation folds
#' @param metric Metric to optimize
#' @param maximize Logical; whether to maximize (TRUE)
#'   or minimize (FALSE) the metric
#' @param verbose Logical; whether to print progress
#' @param ... Additional arguments passed to tl_model
#' @return A tidylearn model object fitted with the best hyperparameters.
#'   Tuning results are stored as an attribute \code{"tuning_results"},
#'   a list containing \code{param_grid}, \code{results} (data frame of
#'   all evaluated combinations), \code{best_params}, \code{best_metric},
#'   \code{metric}, and \code{maximize}.
#' @examples
#' \donttest{
#' model <- tl_tune_grid(iris, Species ~ ., method = "tree",
#'   param_grid = list(cp = c(0.01, 0.1), minsplit = c(10, 20)),
#'   folds = 2, verbose = FALSE)
#' }
#' @export
tl_tune_grid <- function(data, formula, method,
                         param_grid, folds = 5,
                         metric = NULL,
                         maximize = NULL,
                         verbose = TRUE, ...) {
  # Input validation
  if (!is.list(param_grid)) {
    stop("param_grid must be a named list", call. = FALSE)
  }

  # Determine if classification or regression
  response_var <- all.vars(formula)[1]
  y <- data[[response_var]]
  is_classification <- is.factor(y) || is.character(y)

  # Default metric based on problem type
  if (is.null(metric)) {
    if (is_classification) {
      metric <- "accuracy"
      if (is.null(maximize)) maximize <- TRUE
    } else {
      metric <- "rmse"
      if (is.null(maximize)) maximize <- FALSE
    }
  }

  # Create parameter grid
  param_df <- do.call(tidyr::crossing, param_grid)

  if (verbose) {
    message(
      "Tuning ", method, " model with ",
      nrow(param_df), " parameter combinations"
    )
    message(
      "Cross-validation with ", folds, " folds"
    )
    message(
      "Optimizing for ", metric, ", ",
      ifelse(maximize, "maximizing", "minimizing")
    )
  }

  # Create cross-validation splits
  cv_splits <- rsample::vfold_cv(data, v = folds)

  # Initialize results storage
  tuning_results <- list()

  # Loop through parameter combinations
  for (i in seq_len(nrow(param_df))) {
    if (verbose) {
      message(
        "Parameter set ", i, " of ",
        nrow(param_df), ": ",
        paste(
          names(param_df), param_df[i, ],
          sep = "=", collapse = ", "
        )
      )
    }

    # Extract parameters for this iteration
    params <- as.list(param_df[i, ])

    # Initialize metrics storage for this parameter set
    fold_metrics <- numeric(folds)

    # Cross-validation loop
    for (j in seq_len(folds)) {
      # Get training and validation data for this fold
      train_fold <- rsample::analysis(
        cv_splits$splits[[j]]
      )
      valid_fold <- rsample::assessment(
        cv_splits$splits[[j]]
      )

      # Fit model with current parameters
      model_args <- c(
        list(
          data = train_fold,
          formula = formula,
          method = method
        ),
        params,
        list(...)
      )

      # Train model
      fold_model <- tryCatch({
        do.call(tl_model, model_args)
      }, error = function(e) {
        warning(
          "Error fitting model with parameters: ",
          paste(
            names(params), params,
            sep = "=", collapse = ", "
          ),
          ". Error: ", e$message
        )
        NULL
      })

      # If model failed, skip this fold
      if (is.null(fold_model)) {
        fold_metrics[j] <- NA
        next
      }

      # Evaluate model
      eval_metrics <- tl_evaluate(
        fold_model, valid_fold, metrics = metric
      )

      # Store metric value
      fold_metrics[j] <- eval_metrics$value[
        eval_metrics$metric == metric
      ]
    }

    # Calculate mean metric across folds
    mean_metric <- mean(fold_metrics, na.rm = TRUE)

    # Store result for this parameter set
    tuning_results[[i]] <- c(
      params,
      list(
        mean_metric = mean_metric,
        fold_metrics = fold_metrics,
        metric_name = metric,
        maximize = maximize
      )
    )

    if (verbose) {
      message(
        "  Mean ", metric, ": ",
        round(mean_metric, 4)
      )
    }
  }

  # Convert results to data frame
  results_df <- do.call(
    rbind,
    lapply(tuning_results, function(x) {
      df <- data.frame(
        mean_metric = x$mean_metric
      )

      # Add parameters
      for (param in names(param_grid)) {
        df[[param]] <- x[[param]]
      }

      df
    })
  )

  # Find best parameter set
  if (maximize) {
    best_idx <- which.max(results_df$mean_metric)
  } else {
    best_idx <- which.min(results_df$mean_metric)
  }

  # Extract best parameters
  best_params <- as.list(
    results_df[best_idx, names(param_grid)]
  )

  if (verbose) {
    message(
      "Best parameters found: ",
      paste(
        names(best_params), best_params,
        sep = "=", collapse = ", "
      )
    )
    message(
      "Best ", metric, ": ",
      round(results_df$mean_metric[best_idx], 4)
    )
  }

  # Fit final model with best parameters
  final_model_args <- c(
    list(
      data = data,
      formula = formula,
      method = method
    ),
    best_params,
    list(...)
  )

  final_model <- do.call(tl_model, final_model_args)

  # Add tuning results to model
  attr(final_model, "tuning_results") <- list(
    param_grid = param_grid,
    results = results_df,
    best_params = best_params,
    best_metric = results_df$mean_metric[best_idx],
    metric = metric,
    maximize = maximize
  )

  final_model
}

#' Tune hyperparameters using random search
#'
#' @param data A data frame containing the training
#'   data
#' @param formula A formula specifying the model
#' @param method The modeling method to tune
#' @param param_space A named list of parameter spaces
#'   to sample from
#' @param n_iter Number of random parameter
#'   combinations to try
#' @param folds Number of cross-validation folds
#' @param metric Metric to optimize
#' @param maximize Logical; whether to maximize (TRUE)
#'   or minimize (FALSE) the metric
#' @param verbose Logical; whether to print progress
#' @param seed Random seed for reproducibility
#' @param ... Additional arguments passed to tl_model
#' @return A tidylearn model object fitted with the best hyperparameters.
#'   Tuning results are stored as an attribute \code{"tuning_results"},
#'   a list containing \code{param_space}, \code{results} (data frame of
#'   all evaluated iterations), \code{best_params}, \code{best_metric},
#'   \code{metric}, and \code{maximize}.
#' @examples
#' \donttest{
#' model <- tl_tune_random(mtcars, mpg ~ ., method = "tree",
#'   param_space = list(cp = c(0.01, 0.1), minsplit = c(10, 20)),
#'   n_iter = 3, folds = 2, verbose = FALSE)
#' }
#' @export
tl_tune_random <- function(data, formula, method,
                           param_space,
                           n_iter = 10,
                           folds = 5,
                           metric = NULL,
                           maximize = NULL,
                           verbose = TRUE,
                           seed = NULL, ...) {
  # Set seed if provided
  if (!is.null(seed)) {
    set.seed(seed)
  }

  # Input validation
  if (!is.list(param_space)) {
    stop(
      "param_space must be a named list",
      call. = FALSE
    )
  }

  # Determine if classification or regression
  response_var <- all.vars(formula)[1]
  y <- data[[response_var]]
  is_classification <- is.factor(y) || is.character(y)

  # Default metric based on problem type
  if (is.null(metric)) {
    if (is_classification) {
      metric <- "accuracy"
      if (is.null(maximize)) maximize <- TRUE
    } else {
      metric <- "rmse"
      if (is.null(maximize)) maximize <- FALSE
    }
  }

  if (verbose) {
    message(
      "Random search for ", method,
      " model with ", n_iter, " iterations"
    )
    message(
      "Cross-validation with ", folds, " folds"
    )
    message(
      "Optimizing for ", metric, ", ",
      ifelse(maximize, "maximizing", "minimizing")
    )
  }

  # Create cross-validation splits
  cv_splits <- rsample::vfold_cv(data, v = folds)

  # Initialize results storage
  tuning_results <- list()

  # Generate random parameter combinations
  param_combinations <- list()
  for (i in seq_len(n_iter)) {
    params <- list()
    for (param_name in names(param_space)) {
      param_def <- param_space[[param_name]]

      if (is.function(param_def)) {
        # Custom sampling function
        params[[param_name]] <- param_def()
      } else if (is.numeric(param_def) &&
                   length(param_def) == 2) {
        # Numeric range: [min, max]
        params[[param_name]] <- runif(
          1, param_def[1], param_def[2]
        )
      } else if (is.numeric(param_def) &&
                   length(param_def) == 3 &&
                   param_def[3] == "log") {
        # Log-uniform range: [min, max, "log"]
        params[[param_name]] <- exp(runif(
          1, log(param_def[1]), log(param_def[2])
        ))
      } else if (is.integer(param_def) ||
                   (is.numeric(param_def) &&
                      all(param_def == floor(param_def)))) {
        # Integer range or discrete values
        if (length(param_def) == 2) {
          # Integer range: [min, max]
          params[[param_name]] <- sample(
            param_def[1]:param_def[2], 1
          )
        } else {
          # Discrete values
          params[[param_name]] <- sample(
            param_def, 1
          )
        }
      } else if (is.character(param_def) ||
                   is.factor(param_def)) {
        # Categorical parameter
        params[[param_name]] <- sample(
          param_def, 1
        )
      } else if (is.logical(param_def)) {
        # Logical parameter
        params[[param_name]] <- sample(
          c(TRUE, FALSE), 1
        )
      } else {
        stop(
          "Unsupported parameter space definition ",
          "for ", param_name, call. = FALSE
        )
      }
    }
    param_combinations[[i]] <- params
  }

  # Loop through parameter combinations
  for (i in seq_len(n_iter)) {
    params <- param_combinations[[i]]

    if (verbose) {
      message(
        "Iteration ", i, " of ", n_iter, ": ",
        paste(
          names(params),
          round(unlist(params), 4),
          sep = "=", collapse = ", "
        )
      )
    }

    # Initialize metrics storage for this parameter set
    fold_metrics <- numeric(folds)

    # Cross-validation loop
    for (j in seq_len(folds)) {
      # Get training and validation data for this fold
      train_fold <- rsample::analysis(
        cv_splits$splits[[j]]
      )
      valid_fold <- rsample::assessment(
        cv_splits$splits[[j]]
      )

      # Fit model with current parameters
      model_args <- c(
        list(
          data = train_fold,
          formula = formula,
          method = method
        ),
        params,
        list(...)
      )

      # Train model
      fold_model <- tryCatch({
        do.call(tl_model, model_args)
      }, error = function(e) {
        warning(
          "Error fitting model with parameters: ",
          paste(
            names(params), params,
            sep = "=", collapse = ", "
          ),
          ". Error: ", e$message
        )
        NULL
      })

      # If model failed, skip this fold
      if (is.null(fold_model)) {
        fold_metrics[j] <- NA
        next
      }

      # Evaluate model
      eval_metrics <- tl_evaluate(
        fold_model, valid_fold, metrics = metric
      )

      # Store metric value
      fold_metrics[j] <- eval_metrics$value[
        eval_metrics$metric == metric
      ]
    }

    # Calculate mean metric across folds
    mean_metric <- mean(fold_metrics, na.rm = TRUE)

    # Store result for this parameter set
    tuning_results[[i]] <- c(
      params,
      list(
        mean_metric = mean_metric,
        fold_metrics = fold_metrics,
        metric_name = metric,
        maximize = maximize
      )
    )

    if (verbose) {
      message(
        "  Mean ", metric, ": ",
        round(mean_metric, 4)
      )
    }
  }

  # Convert results to data frame
  results_df <- do.call(
    rbind,
    lapply(seq_along(tuning_results), function(i) {
      result <- tuning_results[[i]]

      df <- data.frame(
        iteration = i,
        mean_metric = result$mean_metric
      )

      # Add parameters
      for (param in names(param_space)) {
        df[[param]] <- result[[param]]
      }

      df
    })
  )

  # Find best parameter set
  if (maximize) {
    best_idx <- which.max(results_df$mean_metric)
  } else {
    best_idx <- which.min(results_df$mean_metric)
  }

  # Extract best parameters
  best_params <- as.list(
    results_df[best_idx, names(param_space)]
  )

  if (verbose) {
    message(
      "Best parameters found: ",
      paste(
        names(best_params),
        round(unlist(best_params), 4),
        sep = "=", collapse = ", "
      )
    )
    message(
      "Best ", metric, ": ",
      round(results_df$mean_metric[best_idx], 4)
    )
  }

  # Fit final model with best parameters
  final_model_args <- c(
    list(
      data = data,
      formula = formula,
      method = method
    ),
    best_params,
    list(...)
  )

  final_model <- do.call(tl_model, final_model_args)

  # Add tuning results to model
  attr(final_model, "tuning_results") <- list(
    param_space = param_space,
    results = results_df,
    best_params = best_params,
    best_metric = results_df$mean_metric[best_idx],
    metric = metric,
    maximize = maximize
  )

  final_model
}

#' Plot hyperparameter tuning results
#'
#' @param model A tidylearn model object with tuning
#'   results
#' @param top_n Number of top parameter sets to
#'   highlight
#' @param param1 First parameter to plot (for 2D grid
#'   or scatter plots)
#' @param param2 Second parameter to plot (for 2D grid
#'   or scatter plots)
#' @param plot_type Type of plot: "scatter", "grid",
#'   "parallel", "importance"
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' model <- tl_tune_grid(iris, Species ~ ., method = "tree",
#'   param_grid = list(cp = c(0.01, 0.1), minsplit = c(10, 20)),
#'   folds = 2, verbose = FALSE)
#' tl_plot_tuning_results(model)
#' }
#' @importFrom ggplot2 ggplot aes geom_point geom_tile
#'   scale_fill_gradient2 labs theme_minimal
#' @export
tl_plot_tuning_results <- function(model,
                                   top_n = 5,
                                   param1 = NULL,
                                   param2 = NULL,
                                   plot_type =
                                     "scatter") {
  # Check if model has tuning results
  tuning_results <- attr(model, "tuning_results")
  if (is.null(tuning_results)) {
    stop(
      "Model does not have tuning results",
      call. = FALSE
    )
  }

  # Extract results data frame
  results_df <- tuning_results$results

  # Get parameter names
  param_names <- setdiff(
    names(results_df),
    c("iteration", "mean_metric")
  )

  # Default parameters for plotting if not specified
  if (is.null(param1) && length(param_names) > 0) {
    param1 <- param_names[1]
  }

  if (is.null(param2) && length(param_names) > 1) {
    param2 <- param_names[2]
  }

  # Create ranking column
  results_df$rank <- rank(
    if (tuning_results$maximize) {
      -results_df$mean_metric
    } else {
      results_df$mean_metric
    }
  )
  results_df$is_top <- results_df$rank <= top_n

  # Create plot based on plot_type
  if (plot_type == "scatter" &&
        !is.null(param1) && !is.null(param2)) {
    # Scatter plot of two parameters
    p <- ggplot2::ggplot(
      results_df,
      ggplot2::aes(
        x = .data[[param1]],
        y = .data[[param2]],
        color = .data$mean_metric,
        size = .data$is_top,
        shape = .data$is_top
      )
    ) +
      ggplot2::geom_point(alpha = 0.7) +
      ggplot2::scale_color_gradient2(
        low = "blue",
        high = "red",
        mid = "white",
        midpoint = median(
          results_df$mean_metric, na.rm = TRUE
        )
      ) +
      ggplot2::scale_size_manual(
        values = c(3, 5)
      ) +
      ggplot2::scale_shape_manual(
        values = c(16, 18)
      ) +
      ggplot2::labs(
        title = "Hyperparameter Tuning Results",
        subtitle = paste(
          "Optimizing", tuning_results$metric,
          "- top", top_n, "results highlighted"
        ),
        x = param1,
        y = param2,
        color = tuning_results$metric,
        size = "Top Result",
        shape = "Top Result"
      ) +
      ggplot2::theme_minimal()

  } else if (plot_type == "grid" &&
               !is.null(param1) && !is.null(param2)) {
    # Heat map for grid search
    n_unique_p1 <- length(
      unique(results_df[[param1]])
    )
    n_unique_p2 <- length(
      unique(results_df[[param2]])
    )
    if (n_unique_p1 <= 20 && n_unique_p2 <= 20) {
      p <- ggplot2::ggplot(
        results_df,
        ggplot2::aes(
          x = .data[[param1]],
          y = .data[[param2]],
          fill = .data$mean_metric
        )
      ) +
        ggplot2::geom_tile() +
        ggplot2::geom_text(
          ggplot2::aes(
            label = round(.data$mean_metric, 3)
          ),
          color = "black",
          size = 3
        ) +
        ggplot2::scale_fill_gradient2(
          low = "blue",
          high = "red",
          mid = "white",
          midpoint = median(
            results_df$mean_metric, na.rm = TRUE
          )
        ) +
        ggplot2::labs(
          title = "Hyperparameter Tuning Results Grid",
          subtitle = paste(
            "Optimizing", tuning_results$metric
          ),
          x = param1,
          y = param2,
          fill = tuning_results$metric
        ) +
        ggplot2::theme_minimal()
    } else {
      warning(
        "Parameters have too many unique values ",
        "for a grid plot. ",
        "Using scatter plot instead."
      )
      plot_type <- "scatter"
      tl_plot_tuning_results(
        model, top_n, param1, param2, "scatter"
      )
    }

  } else if (plot_type == "parallel") {
    # Parallel coordinates plot
    results_norm <- results_df

    for (param in param_names) {
      if (is.numeric(results_df[[param]])) {
        param_min <- min(
          results_df[[param]], na.rm = TRUE
        )
        param_max <- max(
          results_df[[param]], na.rm = TRUE
        )
        norm_col <- paste0(param, "_norm")
        if (param_max > param_min) {
          results_norm[[norm_col]] <-
            (results_df[[param]] - param_min) /
            (param_max - param_min)
        } else {
          results_norm[[norm_col]] <- 0.5
        }
      } else {
        # For categorical parameters
        norm_col <- paste0(param, "_norm")
        n_levels <- length(
          unique(results_df[[param]])
        )
        results_norm[[norm_col]] <-
          as.numeric(
            factor(results_df[[param]])
          ) / n_levels
      }
    }

    # Prepare data for parallel coordinates
    param_norm_names <- paste0(
      param_names, "_norm"
    )

    # Convert to long format
    plot_data <- tidyr::pivot_longer(
      results_norm,
      cols = all_of(param_norm_names),
      names_to = "parameter",
      values_to = "value"
    )

    # Remove _norm suffix for plotting
    plot_data$parameter <- gsub(
      "_norm$", "", plot_data$parameter
    )

    # Create parallel coordinates plot
    p <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(
        x = .data$parameter,
        y = .data$value,
        group = .data$rank,
        color = .data$mean_metric,
        size = .data$is_top,
        alpha = .data$is_top
      )
    ) +
      ggplot2::geom_line() +
      ggplot2::scale_color_gradient2(
        low = "blue",
        high = "red",
        mid = "white",
        midpoint = median(
          results_df$mean_metric, na.rm = TRUE
        )
      ) +
      ggplot2::scale_size_manual(
        values = c(0.5, 1.5)
      ) +
      ggplot2::scale_alpha_manual(
        values = c(0.3, 1)
      ) +
      ggplot2::labs(
        title = paste(
          "Parallel Coordinates Plot of",
          "Hyperparameter Tuning Results"
        ),
        subtitle = paste(
          "Optimizing", tuning_results$metric,
          "- top", top_n, "results highlighted"
        ),
        x = "Parameter",
        y = "Normalized Value",
        color = tuning_results$metric,
        size = "Top Result",
        alpha = "Top Result"
      ) +
      ggplot2::theme_minimal() +
      ggplot2::theme(
        panel.grid.major.x = ggplot2::element_line(
          color = "gray90"
        ),
        panel.grid.minor = ggplot2::element_blank(),
        axis.text.x = ggplot2::element_text(
          angle = 45, hjust = 1
        )
      )

  } else if (plot_type == "importance") {
    # Parameter importance plot
    param_importance <- lapply(
      param_names,
      function(param) {
        if (is.numeric(results_df[[param]])) {
          # For numeric parameters, use correlation
          cor_val <- cor(
            results_df[[param]],
            results_df$mean_metric,
            use = "pairwise.complete.obs"
          )
          data.frame(
            parameter = param,
            importance = abs(cor_val),
            correlation = cor_val
          )
        } else {
          # For categorical parameters, use ANOVA
          results_df[[param]] <- as.factor(
            results_df[[param]]
          )

          # Run ANOVA
          anova_result <- summary(aov(
            mean_metric ~ .data[[param]],
            data = results_df
          ))

          # Calculate eta squared
          ss_total <- sum(
            anova_result[[1]]$"Sum Sq"
          )
          ss_param <- anova_result[[1]]$"Sum Sq"[1]
          eta_squared <- ss_param / ss_total

          data.frame(
            parameter = param,
            importance = eta_squared,
            correlation = NA
          )
        }
      }
    )

    # Combine results
    importance_df <- do.call(
      rbind, param_importance
    )

    # Sort by importance
    importance_df <- importance_df[
      order(
        importance_df$importance,
        decreasing = TRUE
      ),
    ]

    # Create bar plot
    p <- ggplot2::ggplot(
      importance_df,
      ggplot2::aes(
        x = reorder(
          .data$parameter,
          .data$importance
        ),
        y = .data$importance,
        fill = .data$correlation
      )
    ) +
      ggplot2::geom_col() +
      ggplot2::scale_fill_gradient2(
        low = "blue",
        high = "red",
        mid = "white",
        midpoint = 0,
        na.value = "gray"
      ) +
      ggplot2::coord_flip() +
      ggplot2::labs(
        title = "Hyperparameter Importance",
        subtitle = paste(
          "Correlation with",
          tuning_results$metric
        ),
        x = "Parameter",
        y = "Importance (|Correlation| or Eta Sq)",
        fill = "Correlation\n(numeric only)"
      ) +
      ggplot2::theme_minimal()
  } else {
    stop(
      "Invalid plot_type or insufficient ",
      "parameters for plotting",
      call. = FALSE
    )
  }

  p
}

#' Create pre-defined parameter grids for common models
#'
#' @param method Model method ("tree", "forest",
#'   "boost", "svm", etc.)
#' @param size Grid size: "small", "medium", "large"
#' @param is_classification Whether the task is
#'   classification or regression
#' @return A named list of parameter values suitable for passing to
#'   \code{\link{tl_tune_grid}} or \code{\link{tl_tune_random}}. Each
#'   element is a numeric or character vector of candidate values for
#'   that hyperparameter.
#' @examples
#' \donttest{
#' grid <- tl_default_param_grid("tree", size = "small")
#' grid <- tl_default_param_grid("forest", size = "medium")
#' }
#' @export
tl_default_param_grid <- function(method,
                                  size = "medium",
                                  is_classification =
                                    TRUE) {
  # Input validation
  size <- match.arg(size, c("small", "medium", "large"))

  # Define default grids for different methods
  if (method == "tree") {
    if (size == "small") {
      list(
        cp = c(0.01, 0.1),
        minsplit = c(10, 20)
      )
    } else if (size == "medium") {
      list(
        cp = c(0.001, 0.01, 0.1),
        minsplit = c(5, 10, 20, 30),
        maxdepth = c(10, 20, 30)
      )
    } else { # large
      list(
        cp = c(0.0001, 0.001, 0.01, 0.1),
        minsplit = c(5, 10, 20, 30, 50),
        maxdepth = c(5, 10, 15, 20, 30)
      )
    }
  } else if (method == "forest") {
    if (size == "small") {
      list(
        mtry = c(2, 3),
        ntree = c(100, 500)
      )
    } else if (size == "medium") {
      list(
        mtry = c(2, 3, 4, 5),
        ntree = c(100, 300, 500)
      )
    } else { # large
      list(
        mtry = c(1, 2, 3, 4, 5, 6),
        ntree = c(100, 300, 500, 1000),
        sampsize = c(0.5, 0.632, 0.8, 1.0),
        nodesize = c(1, 3, 5)
      )
    }
  } else if (method == "boost") {
    if (size == "small") {
      list(
        n.trees = c(50, 100),
        interaction.depth = c(2, 3),
        shrinkage = c(0.01, 0.1)
      )
    } else if (size == "medium") {
      list(
        n.trees = c(50, 100, 200),
        interaction.depth = c(1, 2, 3, 4),
        shrinkage = c(0.001, 0.01, 0.1),
        n.minobsinnode = c(5, 10)
      )
    } else { # large
      list(
        n.trees = c(50, 100, 200, 500, 1000),
        interaction.depth = c(1, 2, 3, 4, 5),
        shrinkage = c(0.001, 0.01, 0.05, 0.1),
        n.minobsinnode = c(1, 5, 10, 20),
        bag.fraction = c(0.5, 0.632, 0.8, 1.0)
      )
    }
  } else if (method == "svm") {
    if (size == "small") {
      list(
        kernel = c("linear", "radial"),
        cost = c(0.1, 1, 10)
      )
    } else if (size == "medium") {
      list(
        kernel = c("linear", "polynomial", "radial"),
        cost = c(0.01, 0.1, 1, 10, 100),
        gamma = c(0.01, 0.1, 1)
      )
    } else { # large
      list(
        kernel = c(
          "linear", "polynomial",
          "radial", "sigmoid"
        ),
        cost = c(
          0.001, 0.01, 0.1, 1, 10, 100, 1000
        ),
        gamma = c(0.001, 0.01, 0.1, 1, 10),
        degree = c(2, 3, 4),
        coef0 = c(0, 0.1, 1)
      )
    }
  } else if (method == "nn") {
    if (size == "small") {
      list(
        size = c(3, 5),
        decay = c(0, 0.1)
      )
    } else if (size == "medium") {
      list(
        size = c(2, 5, 10),
        decay = c(0, 0.01, 0.1)
      )
    } else { # large
      list(
        size = c(2, 5, 10, 20, 50),
        decay = c(0, 0.001, 0.01, 0.1, 0.5),
        maxit = c(100, 200, 500)
      )
    }
  } else if (method %in%
               c("ridge", "lasso", "elastic_net")) {
    if (method == "elastic_net") {
      # Alpha values for elastic net
      alpha_values <- if (size == "small") {
        c(0.25, 0.5, 0.75)
      } else if (size == "medium") {
        c(0.1, 0.25, 0.5, 0.75, 0.9)
      } else { # large
        seq(0.1, 0.9, by = 0.1)
      }
    } else {
      # For ridge and lasso, alpha is fixed
      alpha_values <- if (method == "ridge") {
        0
      } else {
        1
      }
    }

    # Lambda values (regularization strength)
    if (size == "small") {
      lambda_values <- c(0.001, 0.01, 0.1, 1)
    } else if (size == "medium") {
      lambda_values <- c(
        0.0001, 0.001, 0.01, 0.1, 1, 10
      )
    } else { # large
      lambda_values <- c(
        0.0001, 0.0005, 0.001, 0.005,
        0.01, 0.05, 0.1, 0.5, 1, 5, 10
      )
    }

    # Combine parameters
    if (method == "elastic_net") {
      list(
        alpha = alpha_values,
        lambda = lambda_values
      )
    } else {
      list(
        lambda = lambda_values
      )
    }
  } else if (method == "logistic") {
    # For logistic regression, tune regularization
    tl_default_param_grid(
      "ridge", size, is_classification
    )
  } else if (method == "deep") {
    if (size == "small") {
      list(
        hidden_layers = list(c(10), c(10, 5)),
        activation = c("relu", "tanh"),
        dropout = c(0, 0.2)
      )
    } else if (size == "medium") {
      list(
        hidden_layers = list(
          c(10), c(20, 10), c(20, 10, 5)
        ),
        activation = c("relu", "tanh", "sigmoid"),
        dropout = c(0, 0.2, 0.5),
        batch_size = c(16, 32, 64)
      )
    } else { # large
      list(
        hidden_layers = list(
          c(10), c(20), c(50),
          c(20, 10), c(50, 25), c(50, 25, 10)
        ),
        activation = c(
          "relu", "tanh", "sigmoid", "elu"
        ),
        dropout = c(0, 0.1, 0.2, 0.3, 0.5),
        batch_size = c(16, 32, 64, 128),
        learning_rate = c(
          0.0001, 0.001, 0.01, 0.1
        )
      )
    }
  } else {
    # Default empty grid for unknown method
    warning(
      "Unknown method: ", method,
      ". Returning empty parameter grid."
    )
    list()
  }
}

# ---- Tuning functions ----

test_that("tl_default_param_grid returns named list for tree", {
  grid <- tl_default_param_grid("tree", size = "small")

  expect_type(grid, "list")
  expect_true("cp" %in% names(grid))
  expect_true("minsplit" %in% names(grid))
})

test_that("tl_default_param_grid returns named list for forest", {
  grid <- tl_default_param_grid("forest", size = "medium")

  expect_type(grid, "list")
  expect_true("mtry" %in% names(grid))
  expect_true("ntree" %in% names(grid))
})

test_that("tl_default_param_grid returns named list for svm", {
  grid <- tl_default_param_grid("svm", size = "small")

  expect_type(grid, "list")
  expect_true("kernel" %in% names(grid))
  expect_true("cost" %in% names(grid))
})

test_that("tl_default_param_grid handles all supported methods", {
  methods <- c("tree", "forest", "boost", "svm", "nn",
               "ridge", "lasso", "elastic_net", "deep")

  for (method in methods) {
    grid <- tl_default_param_grid(method, size = "small")
    expect_type(grid, "list")
    expect_true(length(grid) > 0)
  }
})

test_that("tl_default_param_grid respects size parameter", {
  small <- tl_default_param_grid("forest", size = "small")
  medium <- tl_default_param_grid("forest", size = "medium")
  large <- tl_default_param_grid("forest", size = "large")

  # Larger grids should have more parameter values
  small_combos <- prod(sapply(small, length))
  medium_combos <- prod(sapply(medium, length))
  large_combos <- prod(sapply(large, length))

  expect_true(small_combos <= medium_combos)
  expect_true(medium_combos <= large_combos)
})

test_that("tl_default_param_grid warns for unknown method", {
  expect_warning(
    grid <- tl_default_param_grid("nonexistent"),
    "Unknown method"
  )
  expect_equal(length(grid), 0)
})

test_that("tl_default_param_grid elastic_net includes alpha", {
  grid <- tl_default_param_grid("elastic_net", size = "small")

  expect_true("alpha" %in% names(grid))
  expect_true(all(grid$alpha > 0 & grid$alpha < 1))
})

test_that("tl_default_param_grid ridge has no alpha in grid", {
  grid <- tl_default_param_grid("ridge", size = "medium")

  # Ridge should only have lambda, not alpha
  expect_true("lambda" %in% names(grid))
  expect_false("alpha" %in% names(grid))
})

test_that("tl_tune_grid works with tree method", {
  skip_if_not_installed("rpart")
  skip_if_not_installed("rsample")

  set.seed(42)
  param_grid <- list(cp = c(0.01, 0.1))

  result <- suppressMessages(
    tl_tune_grid(
      iris, Species ~ ., method = "tree",
      param_grid = param_grid, folds = 2,
      verbose = FALSE
    )
  )

  expect_s3_class(result, "tidylearn_model")
  expect_true(!is.null(attr(result, "tuning_results")))

  tuning <- attr(result, "tuning_results")
  expect_true("best_params" %in% names(tuning))
  expect_true("results" %in% names(tuning))
  expect_equal(nrow(tuning$results), 2)  # 2 param combos
})

test_that("tl_tune_grid validates param_grid input", {
  expect_error(
    tl_tune_grid(mtcars, mpg ~ wt, method = "linear",
                 param_grid = "not a list"),
    "param_grid must be a named list"
  )
})

test_that("tl_tune_random works with tree method", {
  skip_if_not_installed("rpart")
  skip_if_not_installed("rsample")

  set.seed(42)
  param_space <- list(cp = c(0.001, 0.1))

  result <- suppressMessages(
    tl_tune_random(
      iris, Species ~ ., method = "tree",
      param_space = param_space, n_iter = 2,
      folds = 2, verbose = FALSE, seed = 42
    )
  )

  expect_s3_class(result, "tidylearn_model")
  expect_true(!is.null(attr(result, "tuning_results")))

  tuning <- attr(result, "tuning_results")
  expect_equal(nrow(tuning$results), 2)  # 2 iterations
})

test_that("tl_tune_random validates param_space input", {
  expect_error(
    tl_tune_random(mtcars, mpg ~ wt, method = "linear",
                   param_space = "not a list"),
    "param_space must be a named list"
  )
})

test_that("tl_plot_tuning_results returns ggplot", {
  skip_if_not_installed("rpart")
  skip_if_not_installed("rsample")

  set.seed(42)
  param_grid <- list(
    cp = c(0.001, 0.01, 0.1),
    minsplit = c(5, 20)
  )

  model <- suppressMessages(
    tl_tune_grid(
      iris, Species ~ ., method = "tree",
      param_grid = param_grid, folds = 2,
      verbose = FALSE
    )
  )

  # Scatter plot
  p <- tl_plot_tuning_results(model, plot_type = "scatter")
  expect_s3_class(p, "ggplot")

  # Grid plot
  p2 <- tl_plot_tuning_results(model, plot_type = "grid")
  expect_s3_class(p2, "ggplot")
})

test_that("tl_plot_tuning_results errors without tuning results", {
  model <- tl_model(mtcars, mpg ~ wt, method = "linear")

  expect_error(
    tl_plot_tuning_results(model),
    "tuning results"
  )
})

test_that("tl_tune_grid handles model fitting failures gracefully", {
  skip_if_not_installed("rsample")

  # Use a dataset where some parameter combos might fail
  param_grid <- list(cp = c(0.01, 0.5))

  # Should complete without error even if some folds perform poorly
  expect_no_error(
    suppressMessages(suppressWarnings(
      tl_tune_grid(
        iris, Species ~ ., method = "tree",
        param_grid = param_grid, folds = 2,
        verbose = FALSE
      )
    ))
  )
})

#' Tidy Hierarchical Clustering
#'
#' Performs hierarchical clustering with tidy output
#'
#' @param data A data frame, tibble, or dist object
#' @param method Agglomeration method: "ward.D2",
#'   "single", "complete", "average" (default),
#'   "mcquitty", "median", "centroid"
#' @param distance Distance metric if data is not a
#'   dist object (default: "euclidean")
#' @param cols Columns to include (tidy select).
#'   If NULL, uses all numeric columns.
#'
#' @return A list of class "tidy_hclust" containing:
#' \itemize{
#'   \item model: hclust object
#'   \item dist: distance matrix used
#'   \item method: linkage method used
#'   \item data: original data (for plotting)
#' }
#'
#' @examples
#' # Basic hierarchical clustering
#' hc_result <- tidy_hclust(USArrests, method = "average")
#'
#' # With specific distance
#' hc_result <- tidy_hclust(mtcars, method = "complete", distance = "manhattan")
#'
#' @export
tidy_hclust <- function(data, method = "average",
                        distance = "euclidean",
                        cols = NULL) {

  # Handle dist object
  if (inherits(data, "dist")) {
    dist_mat <- data
    data_orig <- NULL
  } else {
    # Select columns
    if (!is.null(cols)) {
      cols_enquo <- rlang::enquo(cols)
      data_selected <- data %>% dplyr::select(!!cols_enquo)
    } else {
      data_selected <- data %>% dplyr::select(where(is.numeric))
    }

    # Compute distance
    dist_mat <- tidy_dist(data_selected, method = distance)
    data_orig <- data_selected
  }

  # Perform hierarchical clustering
  hc_model <- stats::hclust(dist_mat, method = method)

  # Create result object
  result <- list(
    model = hc_model,
    dist = dist_mat,
    method = method,
    distance_method = distance,
    data = data_orig
  )

  class(result) <- c("tidy_hclust", "list")
  result
}


#' Cut Hierarchical Clustering Tree
#'
#' Cut dendrogram to obtain cluster assignments
#'
#' @param hclust_obj A tidy_hclust object or hclust object
#' @param k Number of clusters (optional)
#' @param h Height at which to cut (optional)
#'
#' @return A tibble with columns \code{.obs_id} (observation identifier) and
#'   \code{cluster} (integer cluster assignment).
#'
#' @examples
#' \donttest{
#' hc <- tidy_hclust(USArrests, method = "ward.D2")
#' clusters <- tidy_cutree(hc, k = 3)
#' }
#'
#' @export
tidy_cutree <- function(hclust_obj, k = NULL, h = NULL) {

  if (inherits(hclust_obj, "tidy_hclust")) {
    hc_model <- hclust_obj$model
  } else if (inherits(hclust_obj, "hclust")) {
    hc_model <- hclust_obj
  } else {
    stop("hclust_obj must be a tidy_hclust or hclust object")
  }

  if (is.null(k) && is.null(h)) {
    stop("Either k or h must be specified")
  }

  # Cut tree
  if (!is.null(k)) {
    clusters <- stats::cutree(hc_model, k = k)
  } else {
    clusters <- stats::cutree(hc_model, h = h)
  }

  # Create tibble
  tibble::tibble(
    .obs_id = names(clusters) %||% as.character(seq_along(clusters)),
    cluster = as.integer(clusters)
  )
}


#' Augment Data with Hierarchical Cluster Assignments
#'
#' Add cluster assignments to original data
#'
#' @param hclust_obj A tidy_hclust object
#' @param data Original data frame
#' @param k Number of clusters (optional)
#' @param h Height at which to cut (optional)
#'
#' @return A tibble containing the original \code{data} with an additional
#'   \code{cluster} integer column indicating cluster assignments.
#'
#' @examples
#' \donttest{
#' hc <- tidy_hclust(USArrests, method = "ward.D2")
#' augmented <- augment_hclust(hc, USArrests, k = 3)
#' }
#'
#' @export
augment_hclust <- function(hclust_obj, data, k = NULL, h = NULL) {

  if (!inherits(hclust_obj, "tidy_hclust")) {
    stop("hclust_obj must be a tidy_hclust object")
  }

  cluster_assignments <- tidy_cutree(hclust_obj, k = k, h = h)

  # Add clusters to data
  data %>%
    dplyr::mutate(.row_id = dplyr::row_number()) %>%
    dplyr::left_join(
      cluster_assignments %>% dplyr::mutate(.row_id = dplyr::row_number()),
      by = ".row_id"
    ) %>%
    dplyr::select(-.row_id, -.obs_id)
}


#' Plot Dendrogram
#'
#' Create dendrogram visualization
#'
#' @param hclust_obj A tidy_hclust object or hclust object
#' @param k Optional; number of clusters to highlight with rectangles
#' @param hang Fraction of plot height to hang labels (default: 0.01)
#' @param cex Label size (default: 0.7)
#'
#' @return The \code{\link[stats]{hclust}} object, returned invisibly. The
#'   dendrogram is plotted as a side effect.
#'
#' @examples
#' \donttest{
#' hc <- tidy_hclust(USArrests, method = "ward.D2")
#' tidy_dendrogram(hc, k = 3)
#' }
#'
#' @export
tidy_dendrogram <- function(hclust_obj, k = NULL, hang = 0.01, cex = 0.7) {

  if (inherits(hclust_obj, "tidy_hclust")) {
    hc_model <- hclust_obj$model
    method_label <- hclust_obj$method
  } else if (inherits(hclust_obj, "hclust")) {
    hc_model <- hclust_obj
    method_label <- hc_model$method
  } else {
    stop("hclust_obj must be a tidy_hclust or hclust object")
  }

  # Plot dendrogram
  plot(hc_model,
       main = paste0(
         "Hierarchical Clustering Dendrogram\n(",
         method_label, " linkage)"
       ),
       xlab = "",
       ylab = "Height",
       sub = "",
       hang = hang,
       cex = cex)

  # Add rectangles if k specified
  if (!is.null(k)) {
    stats::rect.hclust(hc_model, k = k, border = 2:(k + 1))
  }

  invisible(hc_model)
}


#' Determine Optimal Number of Clusters for Hierarchical Clustering
#'
#' Use silhouette or gap statistic to find optimal k
#'
#' @param hclust_obj A tidy_hclust object
#' @param method Character; "silhouette" (default) or "gap"
#' @param max_k Maximum number of clusters to test (default: 10)
#'
#' @return A list containing:
#' \itemize{
#'   \item optimal_k: the recommended number of clusters
#'   \item method: the evaluation method used
#'   \item values: numeric vector of evaluation scores (for silhouette)
#'   \item k_range: integer vector of k values tested (for silhouette)
#' }
#' If \code{method = "gap"}, returns a \code{tidy_gap} object instead.
#'
#' @examples
#' \donttest{
#' hc <- tidy_hclust(USArrests, method = "ward.D2")
#' opt <- optimal_hclust_k(hc, method = "silhouette", max_k = 6)
#' }
#'
#' @export
optimal_hclust_k <- function(hclust_obj, method = "silhouette", max_k = 10) {

  if (!inherits(hclust_obj, "tidy_hclust")) {
    stop("hclust_obj must be a tidy_hclust object")
  }

  dist_mat <- hclust_obj$dist
  hc_model <- hclust_obj$model

  if (method == "silhouette") {
    # Compute silhouette for k = 2 to max_k
    sil_widths <- purrr::map_dbl(2:max_k, function(k) {
      clusters <- stats::cutree(hc_model, k = k)
      sil <- cluster::silhouette(clusters, dist_mat)
      mean(sil[, 3])
    })

    optimal_k <- which.max(sil_widths) + 1

    result <- list(
      optimal_k = optimal_k,
      method = "silhouette",
      values = sil_widths,
      k_range = 2:max_k
    )

  } else if (method == "gap") {
    # Use gap statistic
    gap_result <- tidy_gap_stat(
      hclust_obj$data,
      FUN_cluster = function(data, k) {
        hc_temp <- stats::hclust(stats::dist(data), method = hclust_obj$method)
        stats::cutree(hc_temp, k = k)
      },
      max_k = max_k
    )

    result <- gap_result

  } else {
    stop("method must be 'silhouette' or 'gap'")
  }

  result
}


#' Print Method for tidy_hclust
#'
#' @param x A tidy_hclust object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' hc <- tidy_hclust(USArrests, method = "ward.D2")
#' print(hc)
#' }
#'
#' @export
print.tidy_hclust <- function(x, ...) {
  cat("Tidy Hierarchical Clustering\n")
  cat("=============================\n\n")
  cat("Linkage method:", x$method, "\n")
  cat("Distance method:", x$distance_method, "\n")
  cat("Number of observations:", length(x$model$order), "\n")
  cat("Number of merges:", nrow(x$model$merge), "\n\n")

  cat("Use tidy_cutree() to cut the tree and obtain cluster assignments\n")
  cat("Use tidy_dendrogram() to visualize the dendrogram\n")

  invisible(x)
}


#' Fit hierarchical clustering for tidylearn models
#' @keywords internal
#' @noRd
tl_fit_hclust <- function(data, formula = NULL,
                          method = "average",
                          distance = "euclidean", ...) {
  # Extract variables to use
  if (!is.null(formula)) {
    vars <- get_formula_vars(formula, data)
    data_for_hc <- data[, vars, drop = FALSE]
  } else {
    data_for_hc <- data %>% dplyr::select(where(is.numeric))
  }

  # Fit hierarchical clustering using tidy_hclust
  hc_result <- tidy_hclust(
    data_for_hc, method = method,
    distance = distance, ...
  )

  # Return in expected format
  list(
    model = hc_result$model,
    dist = hc_result$dist,
    method = hc_result$method
  )
}

#' Tidy Apriori Algorithm
#'
#' Mine association rules using the Apriori algorithm with tidy output
#'
#' @param transactions A transactions object or data frame
#' @param support Minimum support (default: 0.01)
#' @param confidence Minimum confidence (default: 0.5)
#' @param minlen Minimum rule length (default: 2)
#' @param maxlen Maximum rule length (default: 10)
#' @param target Type of association mined: "rules"
#'   (default), "frequent itemsets",
#'   "maximally frequent itemsets"
#'
#' @return A list of class "tidy_rules" containing:
#' \itemize{
#'   \item rules_tbl: tibble of rules with lhs, rhs, and quality measures
#'   \item rules: original rules object
#'   \item parameters: parameters used
#' }
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#' data("Groceries", package = "arules")
#'
#' # Basic apriori
#' rules <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'
#' # Access rules
#' rules$rules_tbl
#' }
#' }
#'
#' @export
tidy_apriori <- function(transactions, support = 0.01, confidence = 0.5,
                         minlen = 2, maxlen = 10, target = "rules") {


  # Check if arules is installed
  tl_check_packages("arules")

  # Set up parameters
  params <- list(
    supp = support,
    conf = confidence,
    minlen = minlen,
    maxlen = maxlen,
    target = target
  )

  # Run Apriori
  rules_obj <- arules::apriori(transactions, parameter = params)

  # Convert to tidy format if rules
  if (target == "rules") {
    rules_tbl <- tidy_rules(rules_obj)
  } else {
    rules_tbl <- NULL
  }

  result <- list(
    rules_tbl = rules_tbl,
    rules = rules_obj,
    parameters = params,
    n_rules = length(rules_obj)
  )

  class(result) <- c("tidy_apriori", "list")
  result
}


#' Convert Association Rules to Tidy Tibble
#'
#' @param rules A rules object from arules
#'
#' @return A tibble with columns \code{rule_id}, \code{lhs}, \code{rhs}, and
#'   quality measures (e.g., \code{support}, \code{confidence}, \code{lift}).
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   rules_obj <- arules::apriori(Groceries,
#'     parameter = list(supp = 0.001, conf = 0.5))
#'   rules_tbl <- tidy_rules(rules_obj)
#' }
#' }
#'
#' @export
tidy_rules <- function(rules) {

  # Check if arules is installed
  tl_check_packages("arules")

  if (length(rules) == 0) {
    return(tibble::tibble())
  }

  # Extract LHS and RHS as character
  lhs <- arules::labels(arules::lhs(rules))
  rhs <- arules::labels(arules::rhs(rules))

  # Get quality measures
  quality_df <- arules::quality(rules)

  # Combine into tibble
  rules_tbl <- tibble::tibble(
    rule_id = seq_along(lhs),
    lhs = lhs,
    rhs = rhs
  ) %>%
    dplyr::bind_cols(tibble::as_tibble(quality_df))

  rules_tbl
}


#' Inspect Association Rules
#'
#' View rules sorted by various quality measures
#'
#' @param rules_obj A tidy_apriori object or rules object
#' @param by Sort by: "support", "confidence", "lift" (default), "count"
#' @param n Number of rules to display (default: 10)
#' @param decreasing Sort in decreasing order? (default: TRUE)
#'
#' @return A tibble of the top \code{n} rules sorted by the specified quality
#'   measure.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   inspect_rules(res, by = "lift", n = 5)
#' }
#' }
#'
#' @export
inspect_rules <- function(rules_obj, by = "lift", n = 10, decreasing = TRUE) {

  # Handle different input types
  if (inherits(rules_obj, "tidy_apriori")) {
    rules_tbl <- rules_obj$rules_tbl
  } else if (inherits(rules_obj, "rules")) {
    rules_tbl <- tidy_rules(rules_obj)
  } else if (is.data.frame(rules_obj)) {
    rules_tbl <- rules_obj
  } else {
    stop("rules_obj must be a tidy_apriori object, rules object, or tibble")
  }

  # Sort and select top n
  if (by %in% names(rules_tbl)) {
    rules_tbl <- rules_tbl %>%
      dplyr::arrange(dplyr::desc(!!rlang::sym(by))) %>%
      dplyr::slice(seq_len(min(n, nrow(rules_tbl))))

    if (!decreasing) {
      rules_tbl <- rules_tbl %>%
        dplyr::arrange(!!rlang::sym(by))
    }
  } else {
    warning("Sorting column not found, returning first n rules")
    rules_tbl <- rules_tbl %>%
      dplyr::slice(seq_len(min(n, nrow(rules_tbl))))
  }

  rules_tbl
}


#' Filter Rules by Item
#'
#' Subset rules containing specific items
#'
#' @param rules_obj A tidy_apriori object or tibble of rules
#' @param item Character; item to filter by
#' @param where Character; "lhs", "rhs", or "both" (default: "both")
#'
#' @return A tibble of rules containing the specified \code{item} in the
#'   requested position.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   filter_rules_by_item(res, "whole milk", where = "rhs")
#' }
#' }
#'
#' @export
filter_rules_by_item <- function(rules_obj, item, where = "both") {

  # Get rules tibble
  if (inherits(rules_obj, "tidy_apriori")) {
    rules_tbl <- rules_obj$rules_tbl
  } else {
    rules_tbl <- rules_obj
  }

  # Filter based on location
  if (where == "lhs") {
    filtered <- rules_tbl %>%
      dplyr::filter(grepl(item, lhs, fixed = TRUE))
  } else if (where == "rhs") {
    filtered <- rules_tbl %>%
      dplyr::filter(grepl(item, rhs, fixed = TRUE))
  } else {
    filtered <- rules_tbl %>%
      dplyr::filter(
        grepl(item, lhs, fixed = TRUE) |
          grepl(item, rhs, fixed = TRUE)
      )
  }

  filtered
}


#' Find Related Items
#'
#' Find items frequently purchased with a given item
#'
#' @param rules_obj A tidy_apriori object
#' @param item Character; item to find associations for
#' @param min_lift Minimum lift threshold (default: 1.5)
#' @param top_n Number of top associations to return (default: 10)
#'
#' @return A tibble of rules involving the specified \code{item}, filtered by
#'   \code{min_lift} and sorted by lift in descending order.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   find_related_items(res, "whole milk", min_lift = 1.5)
#' }
#' }
#'
#' @export
find_related_items <- function(rules_obj, item, min_lift = 1.5, top_n = 10) {

  # Get rules tibble
  if (inherits(rules_obj, "tidy_apriori")) {
    rules_tbl <- rules_obj$rules_tbl
  } else {
    rules_tbl <- rules_obj
  }

  # Filter rules containing the item
  related <- rules_tbl %>%
    dplyr::filter(
      grepl(item, lhs, fixed = TRUE) |
        grepl(item, rhs, fixed = TRUE)
    ) %>%
    dplyr::filter(lift >= min_lift) %>%
    dplyr::arrange(dplyr::desc(lift)) %>%
    dplyr::slice(seq_len(min(top_n, dplyr::n())))

  related
}


#' Summarize Association Rules
#'
#' Get summary statistics about rules
#'
#' @param rules_obj A tidy_apriori object or rules tibble
#'
#' @return A list with \code{n_rules} and summary statistics (\code{min},
#'   \code{max}, \code{mean}, \code{median}) for \code{support},
#'   \code{confidence}, and \code{lift}.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   summarize_rules(res)
#' }
#' }
#'
#' @export
summarize_rules <- function(rules_obj) {

  # Get rules tibble
  if (inherits(rules_obj, "tidy_apriori")) {
    rules_tbl <- rules_obj$rules_tbl
    n_rules <- rules_obj$n_rules
  } else {
    rules_tbl <- rules_obj
    n_rules <- nrow(rules_tbl)
  }

  if (n_rules == 0) {
    return(list(n_rules = 0, message = "No rules found"))
  }

  summary_list <- list(
    n_rules = n_rules,
    support = list(
      min = min(rules_tbl$support),
      max = max(rules_tbl$support),
      mean = mean(rules_tbl$support),
      median = stats::median(rules_tbl$support)
    ),
    confidence = list(
      min = min(rules_tbl$confidence),
      max = max(rules_tbl$confidence),
      mean = mean(rules_tbl$confidence),
      median = stats::median(rules_tbl$confidence)
    ),
    lift = list(
      min = min(rules_tbl$lift),
      max = max(rules_tbl$lift),
      mean = mean(rules_tbl$lift),
      median = stats::median(rules_tbl$lift)
    )
  )

  summary_list
}


#' Visualize Association Rules
#'
#' Create visualizations of association rules
#'
#' @param rules_obj A tidy_apriori object, rules object, or rules tibble
#' @param method Visualization method: "scatter"
#'   (default), "graph", "grouped", "paracoord"
#' @param top_n Number of top rules to visualize (default: 50)
#' @param ... Additional arguments passed to plot() for rules visualization
#'
#' @return A \code{\link[ggplot2]{ggplot}} object when
#'   \code{method = "scatter"}.
#'   For other methods, the plot is produced as a side effect via
#'   \pkg{arulesViz}.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   visualize_rules(res, method = "scatter")
#' }
#' }
#'
#' @export
visualize_rules <- function(rules_obj, method = "scatter", top_n = 50, ...) {

  # Get rules object
  if (inherits(rules_obj, "tidy_apriori")) {
    rules <- rules_obj$rules
  } else if (inherits(rules_obj, "rules")) {
    rules <- rules_obj
  } else if (is.data.frame(rules_obj)) {
    stop(
      "Cannot visualize tibble directly; ",
      "provide tidy_apriori or rules object"
    )
  } else {
    stop("rules_obj must be a tidy_apriori or rules object")
  }

  # Subset to top n rules
  if (length(rules) > top_n) {
    rules <- head(rules, top_n)
  }

  # Create visualization based on method
  if (method == "scatter") {
    # Scatter plot with ggplot2
    rules_tbl <- tidy_rules(rules)

    p <- ggplot2::ggplot(
      rules_tbl,
      ggplot2::aes(
        x = support, y = confidence,
        color = lift, size = lift
      )
    ) +
      ggplot2::geom_point(alpha = 0.6) +
      ggplot2::scale_color_gradient(low = "lightblue", high = "red") +
      ggplot2::labs(
        title = "Association Rules - Support vs Confidence",
        subtitle = sprintf("Top %d rules (colored by lift)", nrow(rules_tbl)),
        x = "Support",
        y = "Confidence"
      ) +
      ggplot2::theme_minimal()

    p

  } else {
    # Use arulesViz for other methods
    # Check if arulesViz is available
    if (!requireNamespace("arulesViz", quietly = TRUE)) {
      stop(
        "Package 'arulesViz' is required for ",
        "this visualization method.",
        call. = FALSE
      )
    }
    plot(rules, method = method, ...)
  }
}


#' Generate Product Recommendations
#'
#' Get product recommendations based on basket contents
#'
#' @param rules_obj A tidy_apriori object
#' @param basket Character vector of items in current basket
#' @param top_n Number of recommendations to return (default: 5)
#' @param min_confidence Minimum confidence threshold (default: 0.5)
#'
#' @return A tibble with columns \code{rhs} (recommended item),
#'   \code{confidence}, \code{lift}, and \code{support}, sorted by lift in
#'   descending order.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   recommend_products(res, basket = c("whole milk", "butter"))
#' }
#' }
#'
#' @export
recommend_products <- function(rules_obj, basket,
                               top_n = 5,
                               min_confidence = 0.5) {

  # Get rules tibble
  if (inherits(rules_obj, "tidy_apriori")) {
    rules_tbl <- rules_obj$rules_tbl
  } else {
    rules_tbl <- rules_obj
  }

  # Find rules where LHS matches basket items
  recommendations <- rules_tbl %>%
    dplyr::filter(confidence >= min_confidence) %>%
    dplyr::filter(sapply(lhs, function(rule_lhs) {
      # Check if all items in LHS are in basket
      rule_items <- strsplit(gsub("[{}]", "", rule_lhs), ",")[[1]]
      rule_items <- trimws(rule_items)
      all(rule_items %in% basket)
    })) %>%
    dplyr::arrange(dplyr::desc(lift)) %>%
    dplyr::select(rhs, confidence, lift, support) %>%
    dplyr::slice(seq_len(min(top_n, dplyr::n())))

  recommendations
}


#' Print Method for tidy_apriori
#'
#' @param x A tidy_apriori object
#' @param ... Additional arguments (ignored)
#'
#' @return The input object \code{x}, returned invisibly.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("arules", quietly = TRUE)) {
#'   data("Groceries", package = "arules")
#'   res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
#'   print(res)
#' }
#' }
#'
#' @export
print.tidy_apriori <- function(x, ...) {
  cat("Tidy Apriori Results\n")
  cat("====================\n\n")
  cat("Parameters:\n")
  cat("  Minimum support:   ", x$parameters$supp, "\n")
  cat("  Minimum confidence:", x$parameters$conf, "\n")
  cat("  Rule length:       ",
      x$parameters$minlen, "-",
      x$parameters$maxlen, "\n\n")

  cat("Results:\n")
  cat("  Number of rules:", x$n_rules, "\n\n")

  if (x$n_rules > 0) {
    summary <- summarize_rules(x)

    cat("Quality Measure Summary:\n")
    cat("  Support:    ",
        sprintf("%.4f - %.4f (mean: %.4f)",
                summary$support$min,
                summary$support$max,
                summary$support$mean), "\n")
    cat("  Confidence: ",
        sprintf("%.4f - %.4f (mean: %.4f)",
                summary$confidence$min,
                summary$confidence$max,
                summary$confidence$mean), "\n")
    cat("  Lift:       ",
        sprintf("%.2f - %.2f (mean: %.2f)",
                summary$lift$min,
                summary$lift$max,
                summary$lift$mean), "\n\n")

    cat("Top 5 rules by lift:\n")
    print(inspect_rules(x, by = "lift", n = 5))
  }

  cat("\nUse inspect_rules() to view more rules\n")
  cat("Use visualize_rules() to create visualizations\n")

  invisible(x)
}

#' @title Neural Networks for tidylearn
#' @name tidylearn-neural-networks
#' @description Neural network functionality for classification and regression
#' @importFrom nnet nnet
#' @importFrom stats predict
#' @importFrom stats model.matrix as.formula
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate
NULL

#' Fit a neural network model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param size Number of units in the hidden layer (default: 5)
#' @param decay Weight decay parameter (default: 0)
#' @param maxit Maximum number of iterations (default: 100)
#' @param trace Logical; whether to print progress (default: FALSE)
#' @param ... Additional arguments to pass to nnet()
#' @return A fitted neural network model
#' @keywords internal
tl_fit_nn <- function(data, formula, is_classification = FALSE,
                      size = 5, decay = 0, maxit = 100, trace = FALSE, ...) {
  # Check if nnet is installed
  tl_check_packages("nnet")

  # Get response variable
  response_var <- all.vars(formula)[1]

  if (is_classification) {
    # For classification, ensure response is a factor
    if (!is.factor(data[[response_var]])) {
      data[[response_var]] <- factor(data[[response_var]])
    }

    # Fit classification neural network
    nn_model <- nnet::nnet(
      formula = formula,
      data = data,
      size = size,
      decay = decay,
      maxit = maxit,
      trace = trace,
      # For classification
      entropy = TRUE,  # Use cross-entropy error function
      ...
    )
  } else {
    # Fit regression neural network
    nn_model <- nnet::nnet(
      formula = formula,
      data = data,
      size = size,
      decay = decay,
      maxit = maxit,
      trace = trace,
      linout = TRUE,  # Linear output for regression
      ...
    )
  }

  nn_model
}

#' Predict using a neural network model
#'
#' @param model A tidylearn neural network model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "prob" (for classification), "class" (for classification)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_nn <- function(model, new_data, type = "response", ...) {
  # Get the neural network model
  fit <- model$fit
  is_classification <- model$spec$is_classification
  response_var <- model$spec$response_var

  if (is_classification) {
    # Get original response levels
    levels <- levels(factor(model$data[[response_var]]))

    if (type == "prob") {
      # Get class probabilities
      probs <- as.data.frame(
        predict(fit, newdata = new_data, type = "raw", ...)
      )

      # For binary classification
      if (length(levels) == 2) {
        # nnet returns probability for the second class only
        if (ncol(probs) == 1) {
          # Create data frame with probabilities for both classes
          prob_df <- tibble::tibble(
            !!levels[1] := 1 - probs[[1]],
            !!levels[2] := probs[[1]]
          )
        } else {
          # Already has both classes
          names(probs) <- levels
          prob_df <- tibble::as_tibble(probs)
        }
      } else {
        # Multiclass - ensure column names match class levels
        names(probs) <- levels
        prob_df <- tibble::as_tibble(probs)
      }

      prob_df
    } else if (type == "class") {
      # Get probabilities first
      probs <- as.data.frame(
        predict(fit, newdata = new_data, type = "raw", ...)
      )

      # For binary classification
      if (length(levels) == 2) {
        if (ncol(probs) == 1) {
          # Classify based on probability > 0.5
          pred_classes <- ifelse(probs[[1]] > 0.5, levels[2], levels[1])
        } else {
          # Find class with highest probability
          pred_classes <- levels[apply(probs, 1, which.max)]
        }
      } else {
        # Multiclass - find class with highest probability
        pred_classes <- levels[apply(probs, 1, which.max)]
      }

      # Convert to factor with original levels
      pred_classes <- factor(pred_classes, levels = levels)

      pred_classes
    } else if (type == "response") {
      # Same as "class" for classification
      tl_predict_nn(model, new_data, type = "class", ...)
    } else {
      stop(
        "Invalid prediction type for neural networks. ",
        "Use 'prob', 'class', or 'response'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    preds <- as.vector(predict(fit, newdata = new_data, type = "raw", ...))
    preds
  }
}

#' Plot neural network architecture
#'
#' @param model A tidylearn neural network model object
#' @param ... Additional arguments
#' @return The return value of \code{\link[NeuralNetTools]{plotnet}}, called for
#'   its side effect of drawing the network diagram, or \code{NULL} if the
#'   \pkg{NeuralNetTools} package is not installed.
#' @examples
#' \donttest{
#' if (requireNamespace("NeuralNetTools", quietly = TRUE)) {
#'   model <- tl_model(iris, Species ~ ., method = "nn", size = 3)
#'   tl_plot_nn_architecture(model)
#' }
#' }
#' @importFrom ggplot2 ggplot aes geom_segment geom_point geom_text theme_void
#' @export
tl_plot_nn_architecture <- function(model, ...) {
  if (model$spec$method != "nn") {
    stop(
      "Neural network architecture plot is only ",
      "available for neural network models",
      call. = FALSE
    )
  }

  # Check if NeuralNetTools is installed
  if (!requireNamespace("NeuralNetTools", quietly = TRUE)) {
    message(
      "Package 'NeuralNetTools' is required for ",
      "neural network visualization. Please install it."
    )
    return(NULL)
  }

  # Plot using NeuralNetTools
  NeuralNetTools::plotnet(model$fit, ...)
}

#' Tune a neural network model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param sizes Vector of hidden layer sizes to try
#' @param decays Vector of weight decay parameters to try
#' @param folds Number of cross-validation folds (default: 5)
#' @param ... Additional arguments to pass to nnet()
#' @return A list with elements \code{model} (the best fitted \code{nnet}
#'   model), \code{best_size} (optimal hidden-layer size), \code{best_decay}
#'   (optimal weight decay), and \code{tuning_results} (a data frame of all
#'   parameter combinations and their cross-validated errors).
#' @export
tl_tune_nn <- function(data, formula, is_classification = FALSE,
                       sizes = c(1, 2, 5, 10), decays = c(0, 0.001, 0.01, 0.1),
                       folds = 5, ...) {
  # Check if nnet is installed
  tl_check_packages("nnet")

  # Create cross-validation splits
  cv_splits <- rsample::vfold_cv(data, v = folds)

  # Initialize results
  tune_results <- expand.grid(
    size = sizes,
    decay = decays,
    error = NA
  )

  # Get response variable
  response_var <- all.vars(formula)[1]

  # Loop through parameter combinations
  for (i in seq_len(nrow(tune_results))) {
    size <- tune_results$size[i]
    decay <- tune_results$decay[i]

    # Cross-validation for this parameter combination
    cv_errors <- numeric(folds)

    for (j in 1:folds) {
      # Get training and testing data for this fold
      train_data <- rsample::analysis(cv_splits$splits[[j]])
      test_data <- rsample::assessment(cv_splits$splits[[j]])

      # Train neural network
      nn <- tl_fit_nn(
        data = train_data,
        formula = formula,
        is_classification = is_classification,
        size = size,
        decay = decay,
        maxit = 100,
        trace = FALSE,
        ...
      )

      # Make predictions
      preds <- predict(nn, newdata = test_data, type = "raw")

      # Calculate error
      if (is_classification) {
        # For classification, use accuracy
        if (is.factor(test_data[[response_var]])) {
          actuals <- test_data[[response_var]]
        } else {
          actuals <- factor(test_data[[response_var]])
        }

        # Convert predictions to class labels
        levels <- levels(actuals)

        if (length(levels) == 2) {
          if (is.vector(preds)) {
            # Binary classification with single output
            pred_classes <- ifelse(preds > 0.5, levels[2], levels[1])
          } else {
            # Binary classification with multiple outputs
            pred_classes <- levels[apply(preds, 1, which.max)]
          }
        } else {
          # Multiclass classification
          pred_classes <- levels[apply(preds, 1, which.max)]
        }

        pred_classes <- factor(pred_classes, levels = levels)

        # Calculate classification error (1 - accuracy)
        cv_errors[j] <- 1 - mean(pred_classes == actuals)
      } else {
        # For regression, use MSE
        actuals <- test_data[[response_var]]
        cv_errors[j] <- mean((preds - actuals)^2)
      }
    }

    # Store mean error for this parameter combination
    tune_results$error[i] <- mean(cv_errors)
  }

  # Find best parameters
  best_idx <- which.min(tune_results$error)
  best_size <- tune_results$size[best_idx]
  best_decay <- tune_results$decay[best_idx]

  # Train final model with best parameters
  best_model <- tl_fit_nn(
    data = data,
    formula = formula,
    is_classification = is_classification,
    size = best_size,
    decay = best_decay,
    maxit = 100,
    ...
  )

  # Return results
  list(
    model = best_model,
    best_size = best_size,
    best_decay = best_decay,
    tuning_results = tune_results
  )
}

#' Plot neural network training history
#'
#' @param model A tidylearn neural network model object
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_line labs theme_minimal
#' @export
tl_plot_nn_tuning <- function(model, ...) {
  if (!is.list(model) || !"tuning_results" %in% names(model)) {
    stop("This function requires the output from tl_tune_nn()", call. = FALSE)
  }

  # Extract tuning results
  tune_results <- model$tuning_results

  # Create data for heatmap
  heatmap_data <- tune_results %>%
    dplyr::mutate(size = factor(.data$size), decay = factor(.data$decay))

  # Create heatmap
  p <- ggplot2::ggplot(
    heatmap_data,
    ggplot2::aes(x = decay, y = size, fill = error)
  ) +
    ggplot2::geom_tile() +
    ggplot2::geom_text(ggplot2::aes(label = round(error, 4)), color = "white") +
    ggplot2::scale_fill_gradient(low = "blue", high = "red") +
    ggplot2::labs(
      title = "Neural Network Parameter Tuning",
      subtitle = paste0(
        "Best parameters: size = ", model$best_size,
        ", decay = ", model$best_decay
      ),
      x = "Weight Decay",
      y = "Hidden Layer Size",
      fill = "Error"
    ) +
    ggplot2::theme_minimal()

  p
}

#' @title Regression Functions for tidylearn
#' @name tidylearn-regression
#' @description Linear and polynomial regression functionality
#' @importFrom stats lm predict poly model.matrix
#' @importFrom tibble tibble
#' @importFrom dplyr bind_cols
NULL

#' Fit a linear regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param ... Additional arguments to pass to lm()
#' @return A fitted linear regression model
#' @keywords internal
tl_fit_linear <- function(data, formula, ...) {
  lm_model <- stats::lm(formula, data = data, ...)
  lm_model
}

#' Predict using a linear regression model
#'
#' @param model A tidylearn linear model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "confidence", "prediction"
#' @param level Confidence level for intervals (default: 0.95)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_linear <- function(model, new_data,
                              type = "response",
                              level = 0.95, ...) {
  if (type == "response") {
    preds <- stats::predict(
      model$fit, newdata = new_data, ...
    )
    preds
  } else if (type == "confidence") {
    pred_obj <- stats::predict(
      model$fit, newdata = new_data,
      interval = "confidence", level = level, ...
    )
    as.data.frame(pred_obj)
  } else if (type == "prediction") {
    pred_obj <- stats::predict(
      model$fit, newdata = new_data,
      interval = "prediction", level = level, ...
    )
    as.data.frame(pred_obj)
  } else {
    stop(
      "Invalid prediction type. ",
      "Use 'response', 'confidence', or 'prediction'.",
      call. = FALSE
    )
  }
}

#' Fit a polynomial regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param degree Degree of the polynomial (default: 2)
#' @param ... Additional arguments to pass to lm()
#' @return A fitted polynomial regression model
#' @keywords internal
tl_fit_polynomial <- function(data, formula, degree = 2, ...) {
  # Parse the formula to get the response and predictor variables
  terms <- stats::terms(formula, data = data)
  response_var <- all.vars(formula)[1]
  predictor_vars <- attr(terms, "term.labels")

  # Create a new formula with polynomial terms
  poly_formula <- paste(response_var, "~")

  for (i in seq_along(predictor_vars)) {
    var <- predictor_vars[i]
    # Check if this is an interaction term or has special characters
    if (grepl(":", var) || grepl("\\*", var) || grepl("^I\\(", var)) {
      # Keep interaction terms as they are
      poly_formula <- paste(
        poly_formula, var,
        ifelse(i < length(predictor_vars), "+", "")
      )
    } else {
      # Create polynomial term
      poly_term <- paste0(
        "poly(", var,
        ", degree = ", degree,
        ", raw = TRUE)"
      )
      poly_formula <- paste(
        poly_formula, poly_term,
        ifelse(i < length(predictor_vars), "+", "")
      )
    }
  }

  # Fit the polynomial model
  poly_model <- stats::lm(
    as.formula(poly_formula), data = data, ...
  )

  # Store original formula and degree for future reference
  attr(poly_model, "original_formula") <- formula
  attr(poly_model, "poly_degree") <- degree

  poly_model
}

#' Predict using a polynomial regression model
#'
#' @param model A tidylearn polynomial model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "confidence", "prediction"
#' @param level Confidence level for intervals (default: 0.95)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_polynomial <- function(model, new_data,
                                  type = "response",
                                  level = 0.95, ...) {
  # Extract polynomial degree
  degree <- attr(model$fit, "poly_degree")
  if (is.null(degree)) degree <- 2  # Default if not stored

  # Call the underlying prediction function
  if (type == "response") {
    preds <- stats::predict(
      model$fit, newdata = new_data, ...
    )
    preds
  } else if (type == "confidence") {
    pred_obj <- stats::predict(
      model$fit, newdata = new_data,
      interval = "confidence", level = level, ...
    )
    as.data.frame(pred_obj)
  } else if (type == "prediction") {
    pred_obj <- stats::predict(
      model$fit, newdata = new_data,
      interval = "prediction", level = level, ...
    )
    as.data.frame(pred_obj)
  } else {
    stop(
      "Invalid prediction type. ",
      "Use 'response', 'confidence', or 'prediction'.",
      call. = FALSE
    )
  }
}

#' Plot diagnostics for a regression model
#'
#' @param model A tidylearn regression model object
#' @param which Which plots to create (1:4)
#' @param ... Additional arguments
#' @return A ggplot object (or list of ggplot objects)
#' @importFrom ggplot2 ggplot aes geom_point geom_smooth
#'   geom_hline geom_text
#' @importFrom ggplot2 labs theme_minimal scale_color_gradient
#'   stat_qq stat_qq_line
#' @keywords internal
tl_plot_diagnostics <- function(model, which = 1:4, ...) {
  # Get residuals and fitted values
  fitted_vals <- fitted(model$fit)
  residuals <- residuals(model$fit)
  std_residuals <- rstandard(model$fit)

  # Create data frame for plotting
  plot_data <- tibble::tibble(
    fitted = fitted_vals,
    residuals = residuals,
    std_residuals = std_residuals,
    abs_residuals = abs(residuals),
    sqrt_abs_residuals = sqrt(abs(residuals)),
    leverage = hatvalues(model$fit),
    cooks_distance = cooks.distance(model$fit)
  )

  plots <- list()

  # Plot 1: Residuals vs Fitted
  if (1 %in% which) {
    p1 <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = fitted, y = residuals)
    ) +
      ggplot2::geom_point(alpha = 0.6) +
      ggplot2::geom_hline(
        yintercept = 0, linetype = "dashed",
        color = "red"
      ) +
      ggplot2::geom_smooth(
        se = FALSE, color = "blue", method = "loess"
      ) +
      ggplot2::labs(
        title = "Residuals vs Fitted",
        x = "Fitted values",
        y = "Residuals"
      ) +
      ggplot2::theme_minimal()

    plots[["residuals_vs_fitted"]] <- p1
  }

  # Plot 2: Normal Q-Q
  if (2 %in% which) {
    p2 <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(sample = std_residuals)
    ) +
      ggplot2::stat_qq() +
      ggplot2::stat_qq_line(color = "red") +
      ggplot2::labs(
        title = "Normal Q-Q",
        x = "Theoretical Quantiles",
        y = "Standardized Residuals"
      ) +
      ggplot2::theme_minimal()

    plots[["qq"]] <- p2
  }

  # Plot 3: Scale-Location
  if (3 %in% which) {
    p3 <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = fitted, y = sqrt_abs_residuals)
    ) +
      ggplot2::geom_point(alpha = 0.6) +
      ggplot2::geom_smooth(
        se = FALSE, color = "blue", method = "loess"
      ) +
      ggplot2::labs(
        title = "Scale-Location",
        x = "Fitted values",
        y = "sqrt(|Standardized residuals|)"
      ) +
      ggplot2::theme_minimal()

    plots[["scale_location"]] <- p3
  }

  # Plot 4: Cook's distance / Leverage
  if (4 %in% which) {
    p4 <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = leverage, y = std_residuals)
    ) +
      ggplot2::geom_point(
        ggplot2::aes(
          size = cooks_distance,
          color = cooks_distance
        ),
        alpha = 0.6
      ) +
      ggplot2::geom_hline(
        yintercept = c(-2, 0, 2),
        linetype = "dashed", color = "red"
      ) +
      ggplot2::scale_color_gradient(
        low = "blue", high = "red"
      ) +
      ggplot2::labs(
        title = "Residuals vs Leverage",
        x = "Leverage",
        y = "Standardized residuals",
        size = "Cook's distance",
        color = "Cook's distance"
      ) +
      ggplot2::theme_minimal()

    plots[["residuals_vs_leverage"]] <- p4
  }

  # Return a single plot or list of plots
  if (length(plots) == 1) {
    plots[[1]]
  } else {
    plots
  }
}

#' Plot actual vs predicted values for a regression model
#'
#' @param model A tidylearn regression model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A ggplot object
#' @importFrom ggplot2 ggplot aes geom_point geom_abline
#'   labs theme_minimal
#' @keywords internal
tl_plot_actual_predicted <- function(model,
                                     new_data = NULL,
                                     ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual and predicted values
  actuals <- new_data[[model$spec$response_var]]
  predictions <- predict(model, new_data)$.pred

  # Create data frame for plotting
  plot_data <- tibble::tibble(
    actual = actuals,
    predicted = predictions
  )

  # Calculate correlation
  corr <- round(cor(actuals, predictions), 3)
  r_squared <- round(cor(actuals, predictions)^2, 3)

  # Create the plot
  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(x = actual, y = predicted)
  ) +
    ggplot2::geom_point(alpha = 0.6) +
    ggplot2::geom_abline(
      intercept = 0, slope = 1,
      color = "red", linetype = "dashed"
    ) +
    ggplot2::labs(
      title = "Actual vs Predicted Values",
      subtitle = paste0(
        "Correlation: ", corr,
        ", R-squared: ", r_squared
      ),
      x = "Actual values",
      y = "Predicted values"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Plot residuals for a regression model
#'
#' @param model A tidylearn regression model object
#' @param type Type of residual plot: "fitted" (default),
#'   "histogram", "predicted"
#' @param ... Additional arguments
#' @return A ggplot object
#' @importFrom ggplot2 ggplot aes geom_point geom_hline
#'   geom_histogram labs theme_minimal
#' @keywords internal
tl_plot_residuals <- function(model, type = "fitted", ...) {
  # Get residuals and fitted values
  fitted_vals <- fitted(model$fit)
  residuals <- residuals(model$fit)

  # Create data frame for plotting
  plot_data <- tibble::tibble(
    fitted = fitted_vals,
    residuals = residuals
  )

  # Create the plot based on type
  if (type == "fitted") {
    p <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = fitted, y = residuals)
    ) +
      ggplot2::geom_point(alpha = 0.6) +
      ggplot2::geom_hline(
        yintercept = 0, linetype = "dashed",
        color = "red"
      ) +
      ggplot2::labs(
        title = "Residuals vs Fitted Values",
        x = "Fitted values",
        y = "Residuals"
      ) +
      ggplot2::theme_minimal()
  } else if (type == "histogram") {
    p <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = residuals)
    ) +
      ggplot2::geom_histogram(
        bins = 30, fill = "steelblue",
        color = "white", alpha = 0.7
      ) +
      ggplot2::labs(
        title = "Histogram of Residuals",
        x = "Residuals",
        y = "Count"
      ) +
      ggplot2::theme_minimal()
  } else if (type == "predicted") {
    # Get predictions
    predictions <- predict(model, model$data)$.pred

    # Add to plot data
    plot_data$predicted <- predictions

    p <- ggplot2::ggplot(
      plot_data,
      ggplot2::aes(x = predicted, y = residuals)
    ) +
      ggplot2::geom_point(alpha = 0.6) +
      ggplot2::geom_hline(
        yintercept = 0, linetype = "dashed",
        color = "red"
      ) +
      ggplot2::labs(
        title = "Residuals vs Predicted Values",
        x = "Predicted values",
        y = "Residuals"
      ) +
      ggplot2::theme_minimal()
  } else {
    stop(
      "Invalid plot type. ",
      "Use 'fitted', 'histogram', or 'predicted'.",
      call. = FALSE
    )
  }

  p
}

#' Create confidence and prediction interval plots
#'
#' @param model A tidylearn regression model object
#' @param new_data Optional data frame for prediction
#'   (if NULL, uses training data)
#' @param level Confidence level (default: 0.95)
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_point geom_ribbon
#'   labs theme_minimal
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt, method = "linear")
#' tl_plot_intervals(model)
#' }
#' @export
tl_plot_intervals <- function(model,
                              new_data = NULL,
                              level = 0.95, ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  if (model$spec$is_classification) {
    stop(
      "Interval plots are only available ",
      "for regression models",
      call. = FALSE
    )
  }

  # Get actual values
  x_var <- all.vars(model$spec$formula)[-1][1]
  y_var <- model$spec$response_var

  # Sort data by x variable for smooth curves
  sorted_data <- new_data[order(new_data[[x_var]]), ]

  # Calculate confidence and prediction intervals from the raw model
  conf_int <- stats::predict(model$fit, newdata = sorted_data,
                             interval = "confidence", level = level)
  pred_int <- stats::predict(model$fit, newdata = sorted_data,
                             interval = "prediction", level = level)

  # Create plot data
  plot_data <- tibble::tibble(
    x = sorted_data[[x_var]],
    y = sorted_data[[y_var]],
    pred = conf_int[, "fit"],
    conf_lower = conf_int[, "lwr"],
    conf_upper = conf_int[, "upr"],
    pred_lower = pred_int[, "lwr"],
    pred_upper = pred_int[, "upr"]
  )

  # Create the plot
  p <- ggplot2::ggplot(
    plot_data, ggplot2::aes(x = x)
  ) +
    # Prediction intervals (wider)
    ggplot2::geom_ribbon(
      ggplot2::aes(
        ymin = pred_lower, ymax = pred_upper
      ),
      fill = "lightblue", alpha = 0.3
    ) +
    # Confidence intervals (narrower)
    ggplot2::geom_ribbon(
      ggplot2::aes(
        ymin = conf_lower, ymax = conf_upper
      ),
      fill = "steelblue", alpha = 0.5
    ) +
    # Fitted line
    ggplot2::geom_line(
      ggplot2::aes(y = pred),
      color = "blue", linewidth = 1
    ) +
    # Actual points
    ggplot2::geom_point(
      ggplot2::aes(y = y), alpha = 0.6
    ) +
    ggplot2::labs(
      title = paste0(
        "Regression with ", level * 100,
        "% Confidence & Prediction Intervals"
      ),
      subtitle = paste0(
        "Dark band: Confidence interval, ",
        "Light band: Prediction interval"
      ),
      x = x_var,
      y = y_var
    ) +
    ggplot2::theme_minimal()

  p
}

#' @title Deep Learning for tidylearn
#' @name tidylearn-deep-learning
#' @description Deep learning functionality using Keras/TensorFlow
#' @importFrom stats model.matrix as.formula
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate
NULL

#' Fit a deep learning model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param hidden_layers Vector of units in each hidden layer
#'   (default: c(32, 16))
#' @param activation Activation function for hidden layers
#'   (default: "relu")
#' @param dropout Dropout rate for regularization (default: 0.2)
#' @param epochs Number of training epochs (default: 30)
#' @param batch_size Batch size for training (default: 32)
#' @param validation_split Proportion of data for validation
#'   (default: 0.2)
#' @param verbose Verbosity mode (0 = silent, 1 = progress bar,
#'   2 = one line per epoch) (default: 0)
#' @param ... Additional arguments
#' @return A fitted deep learning model
#' @keywords internal
tl_fit_deep <- function(data, formula,
                        is_classification = FALSE,
                        hidden_layers = c(32, 16),
                        activation = "relu",
                        dropout = 0.2,
                        epochs = 30, batch_size = 32,
                        validation_split = 0.2,
                        verbose = 0, ...) {
  # Check if keras is installed
  tl_check_packages(c("keras", "tensorflow"))

  # Parse the formula
  response_var <- all.vars(formula)[1]

  # Prepare data
  # Extract response (y)
  y <- data[[response_var]]

  # Create model matrix for predictors, excluding the intercept
  x_mat <- stats::model.matrix(formula, data = data)[, -1, drop = FALSE]

  # Normalize features
  x_means <- colMeans(x_mat)
  x_sds <- apply(x_mat, 2, sd)
  x_scaled <- scale(x_mat, center = x_means, scale = x_sds)

  # Prepare y based on problem type
  if (is_classification) {
    if (!is.factor(y)) {
      y <- factor(y)
    }

    if (length(levels(y)) == 2) {
      # Binary classification
      y_numeric <- as.integer(y) - 1  # Convert to 0/1
      output_units <- 1
      output_activation <- "sigmoid"
      loss <- "binary_crossentropy"
      metrics <- c("accuracy")
    } else {
      # Multiclass classification
      # One-hot encode the response
      y_onehot <- keras::to_categorical(as.integer(y) - 1)
      y_numeric <- y_onehot
      output_units <- length(levels(y))
      output_activation <- "softmax"
      loss <- "categorical_crossentropy"
      metrics <- c("accuracy")
    }
  } else {
    # Regression
    y_numeric <- y
    output_units <- 1
    output_activation <- "linear"
    loss <- "mse"
    metrics <- c("mae")
  }

  # Create sequential model
  model <- keras::keras_model_sequential()

  # Add input layer with appropriate shape
  model %>% keras::layer_dense(
    units = hidden_layers[1],
    activation = activation,
    input_shape = ncol(x_mat)
  )

  # Add dropout for regularization
  if (dropout > 0) {
    model %>% keras::layer_dropout(rate = dropout)
  }

  # Add hidden layers
  for (i in 2:length(hidden_layers)) {
    model %>% keras::layer_dense(
      units = hidden_layers[i],
      activation = activation
    )

    if (dropout > 0) {
      model %>% keras::layer_dropout(rate = dropout)
    }
  }

  # Add output layer
  model %>% keras::layer_dense(
    units = output_units,
    activation = output_activation
  )

  # Compile the model
  model %>% keras::compile(
    optimizer = "adam",
    loss = loss,
    metrics = metrics
  )

  # Fit the model
  history <- model %>% keras::fit(
    x = x_scaled,
    y = y_numeric,
    epochs = epochs,
    batch_size = batch_size,
    validation_split = validation_split,
    verbose = verbose,
    ...
  )

  # Store data for future predictions
  model_data <- list(
    model = model,
    history = history,
    x_means = x_means,
    x_sds = x_sds,
    formula = formula,
    is_classification = is_classification,
    levels = if (is_classification) levels(factor(y)) else NULL
  )

  model_data
}

#' Predict using a deep learning model
#'
#' @param model A tidylearn deep learning model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "prob" (for classification), "class" (for classification)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_deep <- function(model, new_data,
                            type = "response", ...) {
  # Extract the deep learning model and associated data
  fit <- model$fit
  is_classification <- model$spec$is_classification
  formula <- model$spec$formula

  # Create model matrix for new data
  x_new <- stats::model.matrix(
    formula, data = new_data
  )[, -1, drop = FALSE]

  # Scale using the training data parameters
  x_new_scaled <- scale(
    x_new, center = fit$x_means, scale = fit$x_sds
  )

  # Make predictions
  raw_preds <- predict(fit$model, x_new_scaled)

  if (is_classification) {
    if (length(fit$levels) == 2) {
      # Binary classification
      if (type == "prob") {
        # Get probabilities
        prob_df <- tibble::tibble(
          !!fit$levels[1] := 1 - raw_preds[, 1],
          !!fit$levels[2] := raw_preds[, 1]
        )

        prob_df
      } else if (type == "class" || type == "response") {
        # Get classes
        pred_classes <- ifelse(
          raw_preds[, 1] > 0.5,
          fit$levels[2], fit$levels[1]
        )
        pred_classes <- factor(
          pred_classes, levels = fit$levels
        )

        pred_classes
      } else {
        stop(
          "Invalid prediction type for deep learning ",
          "classification. Use 'prob', 'class', ",
          "or 'response'.",
          call. = FALSE
        )
      }
    } else {
      # Multiclass classification
      if (type == "prob") {
        # Convert to data frame with class probabilities
        prob_df <- as.data.frame(raw_preds)
        names(prob_df) <- fit$levels

        tibble::as_tibble(prob_df)
      } else if (type == "class" || type == "response") {
        # Get classes with highest probability
        pred_idx <- apply(raw_preds, 1, which.max)
        pred_classes <- fit$levels[pred_idx]
        pred_classes <- factor(
          pred_classes, levels = fit$levels
        )

        pred_classes
      } else {
        stop(
          "Invalid prediction type for deep learning ",
          "classification. Use 'prob', 'class', ",
          "or 'response'.",
          call. = FALSE
        )
      }
    }
  } else {
    # Regression predictions
    as.vector(raw_preds)
  }
}

#' Plot deep learning model training history
#'
#' @param model A tidylearn deep learning model object
#' @param metrics Which metrics to plot
#'   (default: c("loss", "val_loss"))
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \dontrun{
#' if (requireNamespace("keras", quietly = TRUE)) {
#'   model <- tl_model(iris, Species ~ ., method = "deep", epochs = 5)
#'   tl_plot_deep_history(model)
#' }
#' }
#' @importFrom ggplot2 ggplot aes geom_line labs theme_minimal
#' @export
tl_plot_deep_history <- function(model,
                                 metrics = c("loss",
                                             "val_loss"),
                                 ...) {
  if (model$spec$method != "deep") {
    stop(
      "Training history plot is only available ",
      "for deep learning models",
      call. = FALSE
    )
  }

  # Extract training history
  history <- model$fit$history

  # Convert to data frame
  history_df <- tibble::tibble(
    epoch = seq_along(history$metrics$loss),
    loss = history$metrics$loss
  )

  # Add validation metrics if available
  if (!is.null(history$metrics$val_loss)) {
    history_df$val_loss <- history$metrics$val_loss
  }

  # Add accuracy metrics if available
  if (!is.null(history$metrics$accuracy)) {
    history_df$accuracy <- history$metrics$accuracy
    if (!is.null(history$metrics$val_accuracy)) {
      history_df$val_accuracy <-
        history$metrics$val_accuracy
    }
  }

  # Add MAE metrics if available
  if (!is.null(history$metrics$mae)) {
    history_df$mae <- history$metrics$mae
    if (!is.null(history$metrics$val_mae)) {
      history_df$val_mae <- history$metrics$val_mae
    }
  }

  # Convert to long format for plotting
  history_long <- history_df %>%
    tidyr::pivot_longer(
      cols = -epoch,
      names_to = "metric",
      values_to = "value"
    ) %>%
    dplyr::filter(.data$metric %in% metrics)

  # Create the plot
  p <- ggplot2::ggplot(
    history_long,
    ggplot2::aes(
      x = epoch, y = value, color = metric
    )
  ) +
    ggplot2::geom_line() +
    ggplot2::labs(
      title = "Deep Learning Training History",
      x = "Epoch",
      y = "Value",
      color = "Metric"
    ) +
    ggplot2::theme_minimal()

  p
}

#' Plot deep learning model architecture
#'
#' @param model A tidylearn deep learning model object
#' @param ... Additional arguments
#' @return The return value of \code{keras::plot_model()}, an architecture
#'   diagram of the Keras model.
#' @examples
#' \dontrun{
#' if (requireNamespace("keras", quietly = TRUE)) {
#'   model <- tl_model(iris, Species ~ ., method = "deep", epochs = 5)
#'   tl_plot_deep_architecture(model)
#' }
#' }
#' @export
tl_plot_deep_architecture <- function(model, ...) {
  if (model$spec$method != "deep") {
    stop(
      "Architecture plot is only available ",
      "for deep learning models",
      call. = FALSE
    )
  }

  # Check if keras is installed
  tl_check_packages("keras")

  # Plot model architecture
  getFromNamespace("plot_model", "keras")(
    model$fit$model,
    show_shapes = TRUE,
    show_layer_names = TRUE,
    ...
  )
}

#' Tune a deep learning model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param hidden_layers_options List of vectors defining hidden
#'   layer configurations to try
#' @param learning_rates Learning rates to try
#'   (default: c(0.01, 0.001, 0.0001))
#' @param batch_sizes Batch sizes to try
#'   (default: c(16, 32, 64))
#' @param epochs Number of training epochs (default: 30)
#' @param validation_split Proportion of data for validation
#'   (default: 0.2)
#' @param ... Additional arguments
#' @return A list with elements \code{model} (the best fitted deep learning
#'   model), \code{best_hidden_layers} (optimal layer configuration),
#'   \code{best_learning_rate}, \code{best_batch_size}, and
#'   \code{tuning_results} (a data frame of all hyperparameter combinations
#'   and their validation losses).
#' @examples
#' \dontrun{
#' if (requireNamespace("keras", quietly = TRUE)) {
#'   result <- tl_tune_deep(iris, Species ~ .,
#'     is_classification = TRUE,
#'     hidden_layers_options = list(c(10), c(10, 5)),
#'     learning_rates = c(0.01, 0.001), batch_sizes = c(32),
#'     epochs = 5)
#' }
#' }
#' @export
tl_tune_deep <- function(data, formula,
                         is_classification = FALSE,
                         hidden_layers_options = list(
                           c(32), c(64, 32),
                           c(128, 64, 32)
                         ),
                         learning_rates = c(
                           0.01, 0.001, 0.0001
                         ),
                         batch_sizes = c(16, 32, 64),
                         epochs = 30,
                         validation_split = 0.2, ...) {
  # Check if keras is installed
  tl_check_packages(c("keras", "tensorflow"))

  # Create grid of hyperparameters
  hyperparams <- expand.grid(
    hidden_layers_idx = seq_along(hidden_layers_options),
    learning_rate = learning_rates,
    batch_size = batch_sizes,
    val_loss = NA
  )

  # Train models with different hyperparameters
  for (i in seq_len(nrow(hyperparams))) {
    # Get current hyperparameters
    hl_idx <- hyperparams$hidden_layers_idx[i]
    hidden_layers <- hidden_layers_options[[hl_idx]]
    learning_rate <- hyperparams$learning_rate[i]
    batch_size <- hyperparams$batch_size[i]

    # Fit model with current hyperparameters
    model <- tryCatch({
      tl_fit_deep(
        data = data,
        formula = formula,
        is_classification = is_classification,
        hidden_layers = hidden_layers,
        epochs = epochs,
        batch_size = batch_size,
        validation_split = validation_split,
        optimizer = keras::optimizer_adam(
          learning_rate = learning_rate
        ),
        verbose = 0,
        ...
      )
    }, error = function(e) {
      message(
        "Error fitting model with hyperparameters: ",
        "hidden_layers=",
        paste(hidden_layers, collapse = ","),
        ", learning_rate=", learning_rate,
        ", batch_size=", batch_size
      )
      message("Error message: ", e$message)
      NULL
    })

    # If model was successfully trained, store validation loss
    if (!is.null(model) && !is.null(model$history)) {
      val_losses <- model$history$metrics$val_loss
      hyperparams$val_loss[i] <- min(
        val_losses, na.rm = TRUE
      )
    }
  }

  # Find best hyperparameters (minimizing validation loss)
  best_idx <- which.min(hyperparams$val_loss)
  best_hl_idx <- hyperparams$hidden_layers_idx[best_idx]
  best_hidden_layers <-
    hidden_layers_options[[best_hl_idx]]
  best_learning_rate <-
    hyperparams$learning_rate[best_idx]
  best_batch_size <- hyperparams$batch_size[best_idx]

  # Train final model with best hyperparameters
  best_model <- tl_fit_deep(
    data = data,
    formula = formula,
    is_classification = is_classification,
    hidden_layers = best_hidden_layers,
    epochs = epochs,
    batch_size = best_batch_size,
    validation_split = validation_split,
    optimizer = keras::optimizer_adam(
      learning_rate = best_learning_rate
    ),
    ...
  )

  # Return results
  list(
    model = best_model,
    best_hidden_layers = best_hidden_layers,
    best_learning_rate = best_learning_rate,
    best_batch_size = best_batch_size,
    tuning_results = hyperparams
  )
}

#' @title tidylearn: A Unified Tidy Interface to R's Machine Learning Ecosystem
#' @name tidylearn-core
#' @description Core functionality for tidylearn. This package
#'   provides a unified tidyverse-compatible interface to
#'   established R machine learning packages including glmnet,
#'   randomForest, xgboost, e1071, rpart, gbm, nnet, cluster,
#'   and dbscan. The underlying algorithms are unchanged -
#'   tidylearn wraps them with consistent function signatures,
#'   tidy tibble output, and unified ggplot2-based
#'   visualization. Access raw model objects via model$fit.
#' @importFrom magrittr %>%
#' @importFrom rlang .data .env
#' @importFrom dplyr filter select mutate group_by summarize arrange
#' @importFrom tibble tibble as_tibble
#' @importFrom purrr map map_dbl map_lgl map2
#' @importFrom tidyr nest unnest
#' @importFrom stats predict model.matrix formula as.formula
NULL

#' Pipe operator
#'
#' @name %>%
#' @rdname pipe
#' @param lhs A value or the magrittr placeholder.
#' @param rhs A function call using the magrittr semantics.
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @usage lhs \%>\% rhs
#' @return The result of applying rhs to lhs.
#' @description See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
NULL

#' @export
#' @rdname pipe
`%>%` <- magrittr::`%>%`

#' Create a tidylearn model
#'
#' Unified interface for creating machine learning models
#' by wrapping established R packages. This function dispatches
#' to the appropriate underlying package based on the method.
#'
#' The wrapped packages include: stats (lm, glm, prcomp,
#' kmeans, hclust), glmnet, randomForest, xgboost, gbm,
#' e1071, nnet, rpart, cluster, and dbscan. The underlying
#' algorithms are unchanged - this function provides a
#' consistent interface and returns tidy output.
#'
#' Access the raw model object from the underlying package via \code{model$fit}.
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model. For
#'   unsupervised methods, use \code{~ vars} or NULL.
#' @param method The modeling method. Supervised: "linear"
#'   (stats::lm), "logistic" (stats::glm), "tree" (rpart),
#'   "forest" (randomForest), "boost" (gbm),
#'   "ridge"/"lasso"/"elastic_net" (glmnet), "svm" (e1071),
#'   "nn" (nnet), "deep" (keras), "xgboost" (xgboost).
#'   Unsupervised: "pca" (stats::prcomp),
#'   "mds" (stats/MASS/smacof), "kmeans" (stats::kmeans),
#'   "pam"/"clara" (cluster), "hclust" (stats::hclust),
#'   "dbscan" (dbscan).
#' @param ... Additional arguments passed to the underlying model function
#' @return A \code{tidylearn_model} object (S3) containing the fitted model
#'   (\code{$fit}), model specification (\code{$spec}), and training data
#'   (\code{$data}). The object also inherits from a method-specific class
#'   (e.g., \code{tidylearn_linear}) and a paradigm class
#'   (\code{tidylearn_supervised} or \code{tidylearn_unsupervised}).
#' @export
#' @examples
#' \donttest{
#' # Classification -> wraps randomForest::randomForest()
#' model <- tl_model(iris, Species ~ ., method = "forest")
#' model$fit  # Access the raw randomForest object
#'
#' # Regression -> wraps stats::lm()
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' model$fit  # Access the raw lm object
#'
#' # PCA -> wraps stats::prcomp()
#' model <- tl_model(iris, ~ ., method = "pca")
#' model$fit  # Access the raw prcomp object
#'
#' # Clustering -> wraps stats::kmeans()
#' model <- tl_model(iris, method = "kmeans", k = 3)
#' model$fit  # Access the raw kmeans object
#' }
tl_model <- function(data, formula = NULL, method = "linear", ...) {
  # Validate inputs
  if (!is.data.frame(data)) {
    stop("'data' must be a data frame", call. = FALSE)
  }

  # Define supervised and unsupervised methods
  supervised_methods <- c(
    "linear", "polynomial", "logistic", "tree",
    "forest", "boost", "ridge", "lasso",
    "elastic_net", "svm", "nn", "deep", "xgboost"
  )
  unsupervised_methods <- c(
    "pca", "mds", "kmeans", "pam",
    "clara", "hclust", "dbscan"
  )

  # Determine paradigm
  is_supervised <- method %in% supervised_methods
  is_unsupervised <- method %in% unsupervised_methods

  if (!is_supervised && !is_unsupervised) {
    stop(
      "Unknown method: ", method,
      "\nSupervised methods: ",
      paste(supervised_methods, collapse = ", "),
      "\nUnsupervised methods: ",
      paste(unsupervised_methods, collapse = ", "),
      call. = FALSE
    )
  }

  # Route to appropriate function
  if (is_supervised) {
    tl_model_supervised(data, formula, method, ...)
  } else {
    tl_model_unsupervised(data, formula, method, ...)
  }
}

#' Create a supervised learning model
#'
#' Internal function for creating supervised models
#' @keywords internal
#' @noRd
tl_model_supervised <- function(data, formula, method, ...) {
  if (!inherits(formula, "formula")) {
    formula <- as.formula(formula)
  }

  # Extract response variable
  response_var <- all.vars(formula)[1]

  # Determine if classification or regression
  y <- data[[response_var]]
  is_classification <- is.factor(y) || is.character(y)

  if (!is_classification && is.numeric(y) &&
        length(unique(y)) <= 10) {
    message(
      "Note: Response '", response_var, "' has ",
      length(unique(y)), " unique numeric values. ",
      "Treating as regression. Convert to factor ",
      "for classification."
    )
  }

  # Create model specification
  model_spec <- list(
    paradigm = "supervised",
    formula = formula,
    method = method,
    is_classification = is_classification,
    response_var = response_var
  )

  # Fit the model based on method
  fitted_model <- switch(
    method,
    "linear" = tl_fit_linear(data, formula, ...),
    "polynomial" = tl_fit_polynomial(data, formula, ...),
    "logistic" = tl_fit_logistic(data, formula, ...),
    "tree" = tl_fit_tree(data, formula, is_classification, ...),
    "forest" = tl_fit_forest(data, formula, is_classification, ...),
    "boost" = tl_fit_boost(data, formula, is_classification, ...),
    "ridge" = tl_fit_ridge(data, formula, is_classification, ...),
    "lasso" = tl_fit_lasso(data, formula, is_classification, ...),
    "elastic_net" = tl_fit_elastic_net(data, formula, is_classification, ...),
    "svm" = tl_fit_svm(data, formula, is_classification, ...),
    "nn" = tl_fit_nn(data, formula, is_classification, ...),
    "deep" = tl_fit_deep(data, formula, is_classification, ...),
    "xgboost" = tl_fit_xgboost(data, formula, is_classification, ...),
    stop("Unsupported supervised method: ", method, call. = FALSE)
  )

  # Create and return tidylearn model object
  model <- structure(
    list(
      spec = model_spec,
      fit = fitted_model,
      data = data
    ),
    class = c(
      paste0("tidylearn_", method),
      "tidylearn_supervised", "tidylearn_model"
    )
  )

  model
}

#' Create an unsupervised learning model
#'
#' Internal function for creating unsupervised models
#' @keywords internal
#' @noRd
tl_model_unsupervised <- function(data, formula = NULL, method, ...) {
  # For unsupervised learning, formula can be NULL or ~ vars

  # Create model specification
  model_spec <- list(
    paradigm = "unsupervised",
    formula = formula,
    method = method
  )

  # Fit the model based on method
  fitted_model <- switch(
    method,
    "pca" = tl_fit_pca(data, formula, ...),
    "mds" = tl_fit_mds(data, formula, ...),
    "kmeans" = tl_fit_kmeans(data, formula, ...),
    "pam" = tl_fit_pam(data, formula, ...),
    "clara" = tl_fit_clara(data, formula, ...),
    "hclust" = tl_fit_hclust(data, formula, ...),
    "dbscan" = tl_fit_dbscan(data, formula, ...),
    stop("Unsupported unsupervised method: ", method, call. = FALSE)
  )

  # Create and return tidylearn model object
  model <- structure(
    list(
      spec = model_spec,
      fit = fitted_model,
      data = data
    ),
    class = c(
      paste0("tidylearn_", method),
      "tidylearn_unsupervised", "tidylearn_model"
    )
  )

  model
}

#' Predict using a tidylearn model
#'
#' Unified prediction interface for both supervised and unsupervised models
#'
#' @param object A tidylearn model object
#' @param new_data A data frame containing the new data.
#'   If NULL, uses training data.
#' @param type Type of prediction. For supervised:
#'   "response" (default), "prob", "class". For
#'   unsupervised: "scores", "clusters", "transform"
#'   depending on method.
#' @param ... Additional arguments
#' @return A \link[tibble]{tibble} with a \code{.pred} column containing
#'   predictions. For classification with \code{type = "prob"}, returns
#'   columns for each class probability.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' predict(model)
#' predict(model, new_data = mtcars[1:5, ])
#' }
#' @export
predict.tidylearn_model <- function(object,
                                    new_data = NULL,
                                    type = "response",
                                    ...) {
  if (is.null(new_data)) {
    new_data <- object$data
  }

  # Route to appropriate predict method
  if (inherits(object, "tidylearn_supervised")) {
    predict_supervised(object, new_data, type, ...)
  } else if (inherits(object, "tidylearn_unsupervised")) {
    predict_unsupervised(object, new_data, type, ...)
  } else {
    stop("Unknown model type", call. = FALSE)
  }
}

#' Predict using supervised models
#' @keywords internal
#' @noRd
predict_supervised <- function(object, new_data, type = "response", ...) {
  method <- object$spec$method

  # Route to method-specific prediction
  preds <- switch(
    method,
    "linear" = predict(object$fit, newdata = new_data, ...),
    "polynomial" = predict(object$fit, newdata = new_data, ...),
    "logistic" = tl_predict_logistic(object, new_data, type, ...),
    "tree" = tl_predict_tree(object, new_data, type, ...),
    "forest" = tl_predict_forest(object, new_data, type, ...),
    "boost" = tl_predict_boost(object, new_data, type, ...),
    "ridge" = tl_predict_glmnet(object, new_data, type, ...),
    "lasso" = tl_predict_glmnet(object, new_data, type, ...),
    "elastic_net" = tl_predict_glmnet(object, new_data, type, ...),
    "svm" = tl_predict_svm(object, new_data, type, ...),
    "nn" = tl_predict_nn(object, new_data, type, ...),
    "deep" = tl_predict_deep(object, new_data, type, ...),
    "xgboost" = tl_predict_xgboost(object, new_data, type, ...),
    stop(
      "Unsupported supervised method for prediction: ",
      method, call. = FALSE
    )
  )

  # Ensure tibble output
  if (is.data.frame(preds) || inherits(preds, "tbl")) {
    preds
  } else {
    tibble::tibble(.pred = preds)
  }
}

#' Predict using unsupervised models
#' @keywords internal
#' @noRd
predict_unsupervised <- function(object, new_data, type = "response", ...) {
  method <- object$spec$method

  result <- switch(
    method,
    "pca" = {
      # For PCA, transform the new data
      if (is.null(new_data) || nrow(new_data) == nrow(object$data)) {
        object$fit$scores
      } else {
        # Transform new data using the PCA rotation
        x_mat <- new_data %>%
          dplyr::select(where(is.numeric)) %>%
          as.matrix()
        if (object$fit$settings$center) {
          x_mat <- scale(
            x_mat,
            center = object$fit$model$center,
            scale = FALSE
          )
        }
        if (object$fit$settings$scale) {
          x_mat <- scale(
            x_mat,
            center = FALSE,
            scale = object$fit$model$scale
          )
        }
        scores <- x_mat %*% object$fit$model$rotation
        colnames(scores) <- paste0(
          "PC", seq_len(ncol(scores))
        )
        tibble::as_tibble(scores) %>%
          dplyr::mutate(
            .obs_id = as.character(seq_len(nrow(scores))),
            .before = 1
          )
      }
    },
    "kmeans" = {
      if (is.null(new_data) || nrow(new_data) == nrow(object$data)) {
        object$fit$clusters
      } else {
        # Assign to nearest center
        x_mat <- new_data %>%
          dplyr::select(where(is.numeric)) %>%
          as.matrix()
        centers <- object$fit$model$centers
        dists <- apply(x_mat, 1, function(x) {
          apply(centers, 1, function(c) sum((x - c)^2))
        })
        clusters <- apply(dists, 2, which.min)
        tibble::tibble(cluster = as.integer(clusters))
      }
    },
    "mds" = {
      # MDS doesn't naturally support new data prediction
      object$fit$points
    },
    "hclust" = {
      # Hierarchical clustering doesn't support standard prediction
      warning(
        "Hierarchical clustering does not support ",
        "standard out-of-sample prediction."
      )
      object$fit$clusters
    },
    "dbscan" = {
      # DBSCAN doesn't support standard prediction
      warning("DBSCAN does not support standard out-of-sample prediction.")
      object$fit$clusters
    },
    stop(
      "Unsupported unsupervised method for prediction: ",
      method, call. = FALSE
    )
  )

  result
}


#' Print method for tidylearn models
#' @param x A tidylearn model object
#' @param ... Additional arguments (ignored)
#' @return The input object \code{x}, returned invisibly.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' print(model)
#' }
#' @export
print.tidylearn_model <- function(x, ...) {
  cat("tidylearn Model\n")
  cat("===============\n")
  cat("Paradigm:", x$spec$paradigm, "\n")
  cat("Method:", x$spec$method, "\n")

  if (x$spec$paradigm == "supervised") {
    cat(
      "Task:",
      ifelse(
        x$spec$is_classification,
        "Classification", "Regression"
      ), "\n"
    )
    cat("Formula:", deparse(x$spec$formula), "\n")
  } else {
    cat("Technique:", x$spec$method, "\n")
  }

  cat("\nTraining observations:", nrow(x$data), "\n")
  invisible(x)
}

#' Summary method for tidylearn models
#' @param object A tidylearn model object
#' @param ... Additional arguments (ignored)
#' @return The input \code{object}, returned invisibly. Called for its
#'   side effect of printing model summary and training performance.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' summary(model)
#' }
#' @export
summary.tidylearn_model <- function(object, ...) {
  print(object)

  cat("\n")
  if (inherits(object, "tidylearn_supervised")) {
    # Evaluate on training data
    eval_results <- tl_evaluate(object)
    cat("Training Performance:\n")
    print(eval_results)
  } else {
    # Show unsupervised model details
    cat("Model Components:\n")
    print(names(object$fit))
  }

  invisible(object)
}

#' Plot a supervised tidylearn model
#'
#' Dispatches to the appropriate plotting function based on model type and
#' requested plot type.
#'
#' @param model A tidylearn supervised model object
#' @param type Plot type. For regression: "auto", "actual_predicted",
#'   "residuals", "diagnostics". For classification: "auto", "confusion",
#'   "roc", "precision_recall", "calibration", "lift", "gain".
#'   "importance" is available for tree-based and regularized models.
#' @param ... Additional arguments passed to the underlying plot function
#' @return A ggplot2 object (invisibly for base-graphics plots)
#' @keywords internal
tl_plot_model <- function(model, type = "auto", ...) {
  is_class <- model$spec$is_classification

  if (type == "auto") {
    type <- if (is_class) "confusion" else "actual_predicted"
  }

  switch(
    type,
    # Regression plots
    "actual_predicted" = tl_plot_actual_predicted(model, ...),
    "residuals"        = tl_plot_residuals(model, ...),
    "diagnostics"      = tl_plot_diagnostics(model, ...),
    # Classification plots
    "confusion"        = tl_plot_confusion(model, ...),
    "roc"              = tl_plot_roc(model, ...),
    "precision_recall" = tl_plot_precision_recall(model, ...),
    "calibration"      = tl_plot_calibration(model, ...),
    "lift"             = tl_plot_lift(model, ...),
    "gain"             = tl_plot_gain(model, ...),
    # Shared
    "importance"       = tl_plot_importance(model, ...),
    stop(
      "Unknown plot type '", type, "'. ",
      if (is_class) {
        paste0(
          "Use: 'confusion', 'roc', ",
          "'precision_recall', 'calibration', ",
          "'lift', 'gain', or 'importance'."
        )
      } else {
        paste0(
          "Use: 'actual_predicted', ",
          "'residuals', 'diagnostics', ",
          "or 'importance'."
        )
      },
      call. = FALSE
    )
  )
}

#' Plot an unsupervised tidylearn model
#'
#' Dispatches to the appropriate plotting function based on the unsupervised
#' model method.
#'
#' @param model A tidylearn unsupervised model object
#' @param type Plot type (default: "auto"). Currently unused; reserved for
#'   future sub-type selection.
#' @param ... Additional arguments passed to the underlying plot function
#' @return A ggplot2 object or invisible result
#' @keywords internal
tl_plot_unsupervised <- function(model, type = "auto", ...) {
  method <- model$spec$method

  switch(
    method,
    "pca"    = plot_variance_explained(get_pca_variance(model$fit), ...),
    "kmeans" = ,
    "pam"    = ,
    "clara"  = ,
    "dbscan" = plot_clusters(
      cbind(model$data, cluster = model$fit$cluster), ...
    ),
    "hclust" = plot_dendrogram(model$fit, ...),
    "mds"    = plot_mds(model$fit, ...),
    stop(
      "Plotting not implemented for unsupervised method: ", method,
      call. = FALSE
    )
  )
}

#' Plot method for tidylearn models
#' @param x A tidylearn model object
#' @param type Plot type (default: "auto")
#' @param ... Additional arguments passed to plotting functions
#' @return A \code{\link[ggplot2]{ggplot}} object. The specific plot depends
#'   on the model paradigm and \code{type} argument.
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' plot(model, type = "actual_predicted")
#' }
#' @export
plot.tidylearn_model <- function(x, type = "auto", ...) {
  if (inherits(x, "tidylearn_supervised")) {
    tl_plot_model(x, type, ...)
  } else if (inherits(x, "tidylearn_unsupervised")) {
    tl_plot_unsupervised(x, type, ...)
  }
}

#' Get tidylearn version information
#' @return A package_version object containing the version number
#' @examples
#' tl_version()
#' @export
tl_version <- function() {
  packageVersion("tidylearn")
}

test_that("tl_model creates supervised models correctly", {
  # Test with classification
  model <- tl_model(iris, Species ~ ., method = "logistic")

  expect_s3_class(model, "tidylearn_model")
  expect_s3_class(model, "tidylearn_supervised")
  expect_s3_class(model, "tidylearn_logistic")
  expect_equal(model$spec$paradigm, "supervised")
  expect_equal(model$spec$method, "logistic")
  expect_true(model$spec$is_classification)
  expect_equal(model$spec$response_var, "Species")
})

test_that("tl_model creates unsupervised models correctly", {
  # Test PCA
  model <- tl_model(iris[, 1:4], method = "pca")

  expect_s3_class(model, "tidylearn_model")
  expect_s3_class(model, "tidylearn_unsupervised")
  expect_s3_class(model, "tidylearn_pca")
  expect_equal(model$spec$paradigm, "unsupervised")
  expect_equal(model$spec$method, "pca")
})

test_that("tl_model validates inputs", {
  # Invalid data type
  expect_error(
    tl_model("not_a_dataframe", method = "linear"),
    "data.*must be a data frame"
  )

  # Unknown method
  expect_error(
    tl_model(iris, Species ~ ., method = "unknown_method"),
    "Unknown method"
  )
})

test_that("tl_model determines task type correctly", {
  # Classification with factor
  model_factor <- tl_model(iris, Species ~ ., method = "logistic")
  expect_true(model_factor$spec$is_classification)

  # Regression with numeric
  model_numeric <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  expect_false(model_numeric$spec$is_classification)
})

test_that("predict.tidylearn_model works for supervised models", {
  model <- tl_model(iris, Species ~ ., method = "logistic")

  # Predict on training data
  pred_train <- predict(model)
  expect_s3_class(pred_train, "tbl_df")
  expect_equal(nrow(pred_train), nrow(iris))

  # Predict on new data
  pred_new <- predict(model, new_data = iris[1:10, ])
  expect_equal(nrow(pred_new), 10)
})

test_that("predict.tidylearn_model works for unsupervised models", {
  model <- tl_model(iris[, 1:4], method = "pca")

  # Transform data
  transformed <- predict(model)
  expect_s3_class(transformed, "tbl_df")
})

test_that("print.tidylearn_model displays correctly", {
  model <- tl_model(iris, Species ~ ., method = "forest")

  # Should print without error
  expect_output(print(model), "tidylearn Model")
  expect_output(print(model), "Paradigm: supervised")
  expect_output(print(model), "Method: forest")
  expect_output(print(model), "Task: Classification")
})

test_that("tl_version returns package version", {
  version <- tl_version()
  expect_s3_class(version, "package_version")
})

#' @title Interaction Analysis Functions for tidylearn
#' @name tidylearn-interactions
#' @description Functions for testing, visualizing, and analyzing interactions
#' @importFrom stats lm anova
#' @importFrom dplyr filter select mutate
#' @importFrom ggplot2 ggplot aes geom_line geom_point
#'   facet_wrap labs theme_minimal
NULL

#' Test for significant interactions between variables
#'
#' @param data A data frame containing the data
#' @param formula A formula specifying the base model without interactions
#' @param var1 First variable to test for interactions
#' @param var2 Second variable to test for interactions
#'   (if NULL, tests var1 with all others)
#' @param all_pairs Logical; whether to test all variable pairs
#' @param categorical_only Logical; whether to only test categorical variables
#' @param numeric_only Logical; whether to only test numeric variables
#' @param mixed_only Logical; whether to only test numeric-categorical pairs
#' @param alpha Significance level for interaction tests
#' @return A data frame with one row per tested interaction pair, containing
#'   columns \code{var1}, \code{var2}, \code{p_value}, \code{significant}
#'   (logical), \code{delta_r2} (change in R-squared), and
#'   \code{f_statistic}, sorted by \code{p_value} ascending.
#' @examples
#' \donttest{
#' results <- tl_test_interactions(mtcars, mpg ~ wt + hp + cyl,
#'   var1 = "wt", var2 = "hp")
#' }
#' @export
tl_test_interactions <- function(data, formula, var1 = NULL, var2 = NULL,
                                 all_pairs = FALSE, categorical_only = FALSE,
                                 numeric_only = FALSE, mixed_only = FALSE,
                                 alpha = 0.05) {
  # Extract predictor variables
  predictors <- all.vars(formula)[-1]

  # Categorize variables
  var_types <- sapply(data[predictors], function(x) {
    if (is.factor(x) || is.character(x)) {
      "categorical"
    } else if (is.numeric(x)) {
      "numeric"
    } else {
      "other"
    }
  })

  # Generate pairs to test
  if (all_pairs) {
    # Test all variable pairs
    pairs <- combn(predictors, 2, simplify = FALSE)
  } else if (!is.null(var1) && !is.null(var2)) {
    # Test specific pair
    pairs <- list(c(var1, var2))
  } else if (!is.null(var1)) {
    # Test var1 with all other variables
    pairs <- lapply(setdiff(predictors, var1), function(v) c(var1, v))
  } else {
    stop("Must specify at least var1 or set all_pairs = TRUE", call. = FALSE)
  }

  # Filter pairs based on variable types
  if (categorical_only) {
    pairs <- pairs[sapply(
      pairs, function(p) all(var_types[p] == "categorical")
    )]
  } else if (numeric_only) {
    pairs <- pairs[sapply(pairs, function(p) all(var_types[p] == "numeric"))]
  } else if (mixed_only) {
    pairs <- pairs[sapply(pairs, function(p) {
      var_types[p[1]] != var_types[p[2]] &&
        all(var_types[p] %in% c("categorical", "numeric"))
    })]
  }

  # Test interactions
  results <- lapply(pairs, function(pair) {
    # Build base model
    base_model <- lm(formula, data = data)

    # Build model with interaction
    int_formula <- update(formula, paste0(". ~ . + ", pair[1], ":", pair[2]))
    int_model <- lm(int_formula, data = data)

    # Perform ANOVA to compare models
    models_comparison <- anova(base_model, int_model)

    # Extract p-value for interaction term
    p_value <- models_comparison$`Pr(>F)`[2]

    # Calculate interaction effect size (change in R-squared)
    r2_base <- summary(base_model)$r.squared
    r2_int <- summary(int_model)$r.squared
    delta_r2 <- r2_int - r2_base

    # Create result row
    data.frame(
      var1 = pair[1],
      var2 = pair[2],
      p_value = p_value,
      significant = p_value < alpha,
      delta_r2 = delta_r2,
      f_statistic = models_comparison$F[2]
    )
  })

  # Combine results
  all_results <- do.call(rbind, results)

  # Sort by significance
  all_results <- all_results[order(all_results$p_value), ]

  all_results
}

#' Plot interaction effects
#'
#' @param model A tidylearn model object
#' @param var1 First variable in the interaction
#' @param var2 Second variable in the interaction
#' @param n_points Number of points to use for continuous variables
#' @param fixed_values Named list of values for other variables in the model
#' @param confidence Logical; whether to show confidence intervals
#' @param ... Additional arguments to pass to predict()
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @export
tl_plot_interaction <- function(model, var1, var2,
                                n_points = 100,
                                fixed_values = NULL,
                                confidence = TRUE,
                                ...) {
  # Extract data
  data <- model$data

  # Check if variables exist in the model
  formula <- model$spec$formula
  all_vars <- all.vars(formula)[-1] # Remove response variable
  if (!var1 %in% all_vars || !var2 %in% all_vars) {
    stop("Variables not found in model formula", call. = FALSE)
  }

  # Determine variable types
  var1_type <- if (
    is.factor(data[[var1]]) || is.character(data[[var1]])
  ) "categorical" else "numeric"
  var2_type <- if (
    is.factor(data[[var2]]) || is.character(data[[var2]])
  ) "categorical" else "numeric"

  # Create grid of values
  if (var1_type == "categorical") {
    if (is.factor(data[[var1]])) {
      var1_values <- levels(data[[var1]])
    } else {
      var1_values <- unique(data[[var1]])
    }
  } else {
    var1_values <- seq(min(data[[var1]], na.rm = TRUE),
                       max(data[[var1]], na.rm = TRUE),
                       length.out = n_points)
  }

  if (var2_type == "categorical") {
    if (is.factor(data[[var2]])) {
      var2_values <- levels(data[[var2]])
    } else {
      var2_values <- unique(data[[var2]])
    }
  } else {
    var2_values <- seq(min(data[[var2]], na.rm = TRUE),
                       max(data[[var2]], na.rm = TRUE),
                       length.out = n_points)
  }

  # Create grid of all combinations
  grid <- expand.grid(
    var1 = var1_values,
    var2 = var2_values
  )
  names(grid) <- c(var1, var2)

  # Add fixed values for other variables
  all_other_vars <- setdiff(all_vars, c(var1, var2))

  for (v in all_other_vars) {
    if (!is.null(fixed_values) && v %in% names(fixed_values)) {
      # Use provided value
      grid[[v]] <- fixed_values[[v]]
    } else {
      # Use median/mode as default
      if (is.numeric(data[[v]])) {
        grid[[v]] <- median(data[[v]], na.rm = TRUE)
      } else if (is.factor(data[[v]]) || is.character(data[[v]])) {
        # Use most frequent value
        tab <- table(data[[v]])
        grid[[v]] <- names(tab)[which.max(tab)]

        # Convert to factor if necessary
        if (is.factor(data[[v]])) {
          grid[[v]] <- factor(grid[[v]], levels = levels(data[[v]]))
        }
      } else {
        # Use first value
        grid[[v]] <- data[[v]][1]
      }
    }
  }

  # Make predictions
  predictions <- predict(model, grid, ...)

  # Add predictions to grid -- tidylearn predict returns tibble with .pred
  if (is.data.frame(predictions) && ".pred" %in% names(predictions)) {
    grid$prediction <- predictions$.pred
    y_col <- "prediction"
    lower_col <- NULL
    upper_col <- NULL
  } else if (is.data.frame(predictions)) {
    # Non-tidylearn output (e.g., confidence intervals with fit/lwr/upr)
    grid <- cbind(grid, predictions)
    y_col <- if ("fit" %in% names(predictions)) "fit" else names(predictions)[1]
    lower_col <- if ("lwr" %in% names(predictions)) "lwr" else NULL
    upper_col <- if ("upr" %in% names(predictions)) "upr" else NULL
  } else {
    grid$prediction <- predictions
    y_col <- "prediction"
    lower_col <- NULL
    upper_col <- NULL
  }

  # Create plot
  if (var1_type == "numeric" && var2_type == "categorical") {
    # Line plot with x = var1, color = var2
    p <- ggplot2::ggplot(grid, ggplot2::aes(
      x = .data[[var1]], y = .data[[y_col]],
      color = .data[[var2]]
    )) +
      ggplot2::geom_line() +
      ggplot2::labs(
        title = paste("Interaction between", var1, "and", var2),
        x = var1,
        y = "Predicted Value",
        color = var2
      )

    # Add confidence intervals if available
    if (confidence && !is.null(lower_col) && !is.null(upper_col)) {
      p <- p + ggplot2::geom_ribbon(
        ggplot2::aes(
          ymin = .data[[lower_col]],
          ymax = .data[[upper_col]],
          fill = .data[[var2]]
        ),
        alpha = 0.2,
        linetype = 0
      )
    }
  } else if (var1_type == "categorical" && var2_type == "numeric") {
    # Line plot with x = var2, color = var1
    p <- ggplot2::ggplot(grid, ggplot2::aes(
      x = .data[[var2]], y = .data[[y_col]],
      color = .data[[var1]]
    )) +
      ggplot2::geom_line() +
      ggplot2::labs(
        title = paste("Interaction between", var1, "and", var2),
        x = var2,
        y = "Predicted Value",
        color = var1
      )

    # Add confidence intervals if available
    if (confidence && !is.null(lower_col) && !is.null(upper_col)) {
      p <- p + ggplot2::geom_ribbon(
        ggplot2::aes(
          ymin = .data[[lower_col]],
          ymax = .data[[upper_col]],
          fill = .data[[var1]]
        ),
        alpha = 0.2,
        linetype = 0
      )
    }
  } else if (var1_type == "numeric" && var2_type == "numeric") {
    # Contour plot or heat map
    p <- ggplot2::ggplot(grid, ggplot2::aes(
      x = .data[[var1]], y = .data[[var2]],
      z = .data[[y_col]]
    )) +
      ggplot2::geom_contour_filled() +
      ggplot2::labs(
        title = paste("Interaction between", var1, "and", var2),
        x = var1,
        y = var2,
        fill = "Predicted Value"
      )
  } else {
    # Categorical x Categorical: Faceted bar plot
    p <- ggplot2::ggplot(grid, ggplot2::aes(
      x = .data[[var1]], y = .data[[y_col]],
      fill = .data[[var2]]
    )) +
      ggplot2::geom_col(position = "dodge") +
      ggplot2::labs(
        title = paste("Interaction between", var1, "and", var2),
        x = var1,
        y = "Predicted Value",
        fill = var2
      )
  }

  # Apply minimal theme
  p <- p + ggplot2::theme_minimal()

  p
}

#' Find important interactions automatically
#'
#' @param data A data frame containing the data
#' @param formula A formula specifying the base model without interactions
#' @param top_n Number of top interactions to return
#' @param min_r2_change Minimum change in R-squared to consider
#' @param max_p_value Maximum p-value for significance
#' @param exclude_vars Character vector of variables to exclude
#'   from interaction testing
#' @return A tidylearn model object (class \code{"tidylearn_model"}) fitted
#'   with the top significant interaction terms added to the formula.
#'   The interaction test results and selected interactions are stored as
#'   attributes \code{"interaction_tests"} and
#'   \code{"selected_interactions"}.
#' @examples
#' \donttest{
#' model <- tl_auto_interactions(mtcars, mpg ~ wt + hp + cyl, top_n = 2)
#' }
#' @export
tl_auto_interactions <- function(data, formula, top_n = 3, min_r2_change = 0.01,
                                 max_p_value = 0.05, exclude_vars = NULL) {
  # Extract predictor variables
  predictors <- all.vars(formula)[-1]

  # Remove excluded variables
  if (!is.null(exclude_vars)) {
    predictors <- setdiff(predictors, exclude_vars)
  }

  # Test all interactions
  test_results <- tl_test_interactions(
    data, formula, all_pairs = TRUE,
    alpha = max_p_value
  )

  # Filter significant interactions
  significant <- test_results %>%
    dplyr::filter(.data$p_value < max_p_value, .data$delta_r2 >= min_r2_change)

  # Select top interactions
  if (nrow(significant) > top_n) {
    top_interactions <- significant[1:top_n, ]
  } else {
    top_interactions <- significant
  }

  if (nrow(top_interactions) == 0) {
    message("No significant interactions found based on criteria")
    # Return original model
    return(tl_model(data, formula, method = "linear"))
  }

  # Build formula with interactions
  interaction_terms <- apply(top_interactions, 1, function(row) {
    paste0(row["var1"], ":", row["var2"])
  })

  int_str <- paste(interaction_terms, collapse = " + ")
  new_formula <- update(
    formula, paste0(". ~ . +", int_str)
  )

  # Fit model with interactions
  interaction_model <- tl_model(data, new_formula, method = "linear")

  # Add interaction test results to model
  attr(interaction_model, "interaction_tests") <- test_results
  attr(interaction_model, "selected_interactions") <- top_interactions

  interaction_model
}

#' Calculate partial effects based on a model with interactions
#'
#' @param model A tidylearn model object
#' @param var Variable to calculate effects for
#' @param by_var Variable to calculate effects by (interaction variable)
#' @param at_values Named list of values at which to hold other variables
#' @param intervals Logical; whether to include confidence intervals
#' @return For numeric \code{var}: a list with \code{effects} (data frame of
#'   predicted values across the variable range for each level of
#'   \code{by_var}) and \code{slopes} (data frame with estimated slopes and
#'   standard errors per level). For categorical \code{var}: a data frame of
#'   predicted values at each factor level for each level of \code{by_var}.
#' @export
tl_interaction_effects <- function(model, var, by_var,
                                   at_values = NULL,
                                   intervals = TRUE) {
  # Extract data
  data <- model$data
  formula <- model$spec$formula

  # Check if variables exist in the model
  all_vars <- all.vars(formula)[-1] # Remove response variable
  if (!var %in% all_vars || !by_var %in% all_vars) {
    stop("Variables not found in model formula", call. = FALSE)
  }

  # Check if the interaction term exists in the model
  int_term <- paste0(var, ":", by_var)
  rev_int_term <- paste0(by_var, ":", var)

  has_interaction <- int_term %in% attr(terms(formula), "term.labels") ||
    rev_int_term %in% attr(terms(formula), "term.labels")

  if (!has_interaction) {
    warning(paste("Interaction term", int_term, "not found in model formula"),
            call. = FALSE)
  }

  # Determine by_var values
  if (is.factor(data[[by_var]]) || is.character(data[[by_var]])) {
    if (is.factor(data[[by_var]])) {
      by_values <- levels(data[[by_var]])
    } else {
      by_values <- unique(data[[by_var]])
    }
  } else {
    # For continuous by_var, use quantiles
    by_values <- stats::quantile(
      data[[by_var]], probs = seq(0, 1, 0.25),
      na.rm = TRUE
    )
    names(by_values) <- paste0("Q", seq(0, 100, 25))
  }

  # Set up variable values
  if (is.factor(data[[var]]) || is.character(data[[var]])) {
    if (is.factor(data[[var]])) {
      var_values <- levels(data[[var]])
    } else {
      var_values <- unique(data[[var]])
    }
    var_is_numeric <- FALSE
  } else {
    var_range <- range(data[[var]], na.rm = TRUE)
    var_values <- seq(var_range[1], var_range[2], length.out = 100)
    var_is_numeric <- TRUE
  }

  # Create grid for predictions
  grid_list <- list()

  for (bv in by_values) {
    # Create data frame for this by_value
    grid <- data.frame(var_values)
    names(grid) <- var
    grid[[by_var]] <- bv

    # Add values for other variables
    other_vars <- setdiff(all_vars, c(var, by_var))
    for (v in other_vars) {
      if (!is.null(at_values) && v %in% names(at_values)) {
        grid[[v]] <- at_values[[v]]
      } else {
        # Use median/mode
        if (is.numeric(data[[v]])) {
          grid[[v]] <- median(data[[v]], na.rm = TRUE)
        } else if (is.factor(data[[v]])) {
          grid[[v]] <- levels(data[[v]])[1]
        } else {
          grid[[v]] <- unique(data[[v]])[1]
        }
      }
    }

    # Make predictions -- use raw model for se.fit (tidylearn predict
    # doesn't support it), and extract .pred for the plain case
    if (intervals) {
      raw_preds <- stats::predict(model$fit, newdata = grid, se.fit = TRUE)
      grid$fit <- as.vector(raw_preds$fit)
      grid$se <- as.vector(raw_preds$se.fit)
      grid$lower <- grid$fit - 1.96 * grid$se
      grid$upper <- grid$fit + 1.96 * grid$se
    } else {
      preds <- predict(model, grid)
      grid$fit <- if (is.data.frame(preds)) preds$.pred else preds
    }

    # Add by_value label
    grid$by_value <- bv
    if (!is.null(names(by_values)) && !is.na(match(bv, by_values))) {
      grid$by_label <- names(by_values)[match(bv, by_values)]
    } else {
      grid$by_label <- as.character(bv)
    }

    # Store in list
    grid_list[[length(grid_list) + 1]] <- grid
  }

  # Combine all grids
  final_grid <- do.call(rbind, grid_list)

  # Calculate slopes for numeric variables
  if (var_is_numeric) {
    # Group by by_value
    slopes <- lapply(by_values, function(bv) {
      # Get data for this by_value
      sub_grid <- final_grid[final_grid$by_value == bv, ]

      # If only one value, can't compute slope
      if (nrow(sub_grid) <= 1) {
        data.frame(
          by_value = bv,
          by_label = if (is.null(names(by_values))) {
            as.character(bv)
          } else {
            names(by_values)[match(bv, by_values)]
          },
          slope = NA,
          slope_se = NA
        )
      }

      # Fit linear model to get slope
      slope_formula <- stats::as.formula(paste("fit ~", var))
      slope_model <- lm(slope_formula, data = sub_grid)
      slope_coef <- coef(summary(slope_model))

      data.frame(
        by_value = bv,
        by_label = if (is.null(names(by_values))) {
          as.character(bv)
        } else {
          names(by_values)[match(bv, by_values)]
        },
        slope = slope_coef[2, 1],
        slope_se = slope_coef[2, 2]
      )
    })

    # Combine all slopes
    final_slopes <- do.call(rbind, slopes)

    # Return both grid and slopes
    list(
      effects = final_grid,
      slopes = final_slopes
    )
  } else {
    # For categorical variables, just return the effects
    final_grid
  }
}

#' @title XGBoost Functions for tidylearn
#' @name tidylearn-xgboost
#' @description XGBoost-specific implementation for gradient boosting
#' @importFrom stats model.matrix as.formula
#' @importFrom dplyr filter select mutate
NULL

#' Fit an XGBoost model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param nrounds Number of boosting rounds (default: 100)
#' @param max_depth Maximum depth of trees (default: 6)
#' @param eta Learning rate (default: 0.3)
#' @param subsample Subsample ratio of observations (default: 1)
#' @param colsample_bytree Subsample ratio of columns (default: 1)
#' @param min_child_weight Minimum sum of instance weight
#'   needed in a child (default: 1)
#' @param gamma Minimum loss reduction to make a further partition (default: 0)
#' @param alpha L1 regularization term (default: 0)
#' @param lambda L2 regularization term (default: 1)
#' @param early_stopping_rounds Early stopping rounds (default: NULL)
#' @param nthread Number of threads (default: max available)
#' @param verbose Verbose output (default: 0)
#' @param ... Additional arguments to pass to xgb.train()
#' @return A fitted XGBoost model
#' @keywords internal
tl_fit_xgboost <- function(data, formula, is_classification = FALSE,
                           nrounds = 100, max_depth = 6, eta = 0.3,
                           subsample = 1, colsample_bytree = 1,
                           min_child_weight = 1, gamma = 0,
                           alpha = 0, lambda = 1,
                           early_stopping_rounds = NULL,
                           nthread = NULL, verbose = 0, ...) {
  # Check if xgboost is installed
  tl_check_packages("xgboost")

  # Parse formula
  response_var <- all.vars(formula)[1]

  # Prepare data for XGBoost
  # Create model matrix for predictors (exclude intercept)
  x_mat <- stats::model.matrix(formula, data = data)[, -1, drop = FALSE]

  # Extract response variable
  y <- data[[response_var]]

  # Prepare response variable based on problem type
  if (is_classification) {
    if (!is.factor(y)) {
      y <- factor(y)
    }

    if (length(levels(y)) == 2) {
      # Binary classification: convert to 0/1
      y_numeric <- as.integer(y) - 1
      objective <- "binary:logistic"
      eval_metric <- "logloss"
    } else {
      # Multiclass classification: convert to 0-based index
      y_numeric <- as.integer(y) - 1
      objective <- "multi:softprob"
      eval_metric <- "mlogloss"
      num_class <- length(levels(y))
    }
  } else {
    # Regression
    y_numeric <- y
    objective <- "reg:squarederror"
    eval_metric <- "rmse"
  }

  # Create DMatrix object
  dtrain <- xgboost::xgb.DMatrix(data = as.matrix(x_mat), label = y_numeric)

  # Set parameters
  params <- list(
    objective = objective,
    eval_metric = eval_metric,
    max_depth = max_depth,
    eta = eta,
    subsample = subsample,
    colsample_bytree = colsample_bytree,
    min_child_weight = min_child_weight,
    gamma = gamma,
    alpha = alpha,
    lambda = lambda
  )

  # Add num_class parameter for multiclass classification
  if (is_classification && length(levels(y)) > 2) {
    params$num_class <- num_class
  }

  # Add nthread parameter if specified
  if (!is.null(nthread)) {
    params$nthread <- nthread
  }

  # Train XGBoost model
  xgb_model <- xgboost::xgb.train(
    params = params,
    data = dtrain,
    nrounds = nrounds,
    early_stopping_rounds = early_stopping_rounds,
    verbose = verbose,
    ...
  )

  # Store additional information for later use
  attr(xgb_model, "feature_names") <- colnames(x_mat)
  attr(xgb_model, "response_var") <- response_var
  attr(xgb_model, "is_classification") <- is_classification

  if (is_classification) {
    attr(xgb_model, "response_levels") <- levels(y)
  }

  xgb_model
}

#' Predict using an XGBoost model
#'
#' @param model A tidylearn XGBoost model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "response" (default),
#'   "prob" (for classification), "class" (for classification)
#' @param ntreelimit Limit number of trees used for prediction
#'   (default: NULL, uses all trees)
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_xgboost <- function(model, new_data,
                               type = "response",
                               ntreelimit = NULL, ...) {
  # Extract XGBoost model
  xgb_model <- model$fit

  # Extract metadata
  feature_names <- attr(xgb_model, "feature_names")
  is_classification <- model$spec$is_classification

  # Create model matrix for new data (exclude intercept)
  formula <- model$spec$formula
  x_new <- stats::model.matrix(formula, data = new_data)[, -1, drop = FALSE]

  # Check column names match
  if (!all(colnames(x_new) %in% feature_names)) {
    missing_cols <- setdiff(colnames(x_new), feature_names)
    warning("New data contains columns not in the training data: ",
            paste(missing_cols, collapse = ", "))
  }

  # Create DMatrix for prediction
  x_subset <- x_new[, feature_names, drop = FALSE]
  dtest <- xgboost::xgb.DMatrix(
    data = as.matrix(x_subset)
  )

  # Make predictions
  if (is_classification) {
    if (type == "prob") {
      # Get class probabilities
      response_levels <- attr(xgb_model, "response_levels")
      n_classes <- length(response_levels)

      if (n_classes == 2) {
        # Binary classification
        prob <- predict(
          xgb_model, newdata = dtest,
          ntreelimit = ntreelimit
        )

        # Create data frame with probabilities for both classes
        prob_df <- data.frame(
          prob0 = 1 - prob,
          prob1 = prob
        )
        colnames(prob_df) <- response_levels

        prob_df
      } else {
        # Multiclass classification
        probs <- predict(
          xgb_model, newdata = dtest,
          ntreelimit = ntreelimit,
          reshape = TRUE
        )
        colnames(probs) <- response_levels

        as.data.frame(probs)
      }
    } else if (type == "class" || type == "response") {
      # Get predicted classes
      response_levels <- attr(xgb_model, "response_levels")
      n_classes <- length(response_levels)

      if (n_classes == 2) {
        # Binary classification
        prob <- predict(
          xgb_model, newdata = dtest,
          ntreelimit = ntreelimit
        )
        pred_classes <- ifelse(
          prob > 0.5,
          response_levels[2],
          response_levels[1]
        )
      } else {
        # Multiclass classification
        probs <- predict(
          xgb_model, newdata = dtest,
          ntreelimit = ntreelimit,
          reshape = TRUE
        )
        pred_idx <- max.col(probs)
        pred_classes <- response_levels[pred_idx]
      }

      # Convert to factor with original levels
      factor(pred_classes, levels = response_levels)
    } else {
      stop(
        "Invalid prediction type for XGBoost ",
        "classification. Use 'prob', 'class', ",
        "or 'response'.",
        call. = FALSE
      )
    }
  } else {
    # Regression predictions
    predict(
      xgb_model, newdata = dtest,
      ntreelimit = ntreelimit
    )
  }
}

#' Plot feature importance for an XGBoost model
#'
#' @param model A tidylearn XGBoost model object
#' @param top_n Number of top features to display (default: 10)
#' @param importance_type Type of importance: "gain", "cover", "frequency"
#' @param ... Additional arguments
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
#'   tl_plot_xgboost_importance(model)
#' }
#' }
#' @export
tl_plot_xgboost_importance <- function(model, top_n = 10,
                                       importance_type = "gain",
                                       ...) {
  # Check if model is an XGBoost model
  if (!inherits(model, "tidylearn_model") || model$spec$method != "xgboost") {
    stop("This function requires an XGBoost model", call. = FALSE)
  }

  # Extract XGBoost model
  xgb_model <- model$fit

  # Calculate feature importance
  importance <- xgboost::xgb.importance(
    model = xgb_model,
    feature_names = attr(xgb_model, "feature_names"),
    ...
  )

  # Limit to top features
  if (nrow(importance) > top_n) {
    importance <- importance[1:top_n, ]
  }

  # Create importance plot
  xgboost::xgb.plot.importance(
    importance,
    rel_to_first = TRUE,
    xlab = "Relative importance"
  )
}

#' Plot XGBoost tree visualization
#'
#' @param model A tidylearn XGBoost model object
#' @param tree_index Index of the tree to plot (default: 0, first tree)
#' @param ... Additional arguments
#' @return The return value of \code{\link[xgboost]{xgb.plot.tree}}, a
#'   tree diagram rendered via the \pkg{DiagrammeR} package.
#' @export
tl_plot_xgboost_tree <- function(model, tree_index = 0, ...) {
  # Check if model is an XGBoost model
  if (!inherits(model, "tidylearn_model") || model$spec$method != "xgboost") {
    stop("This function requires an XGBoost model", call. = FALSE)
  }

  # Extract XGBoost model
  xgb_model <- model$fit

  # Plot tree
  xgboost::xgb.plot.tree(model = xgb_model, tree_index = tree_index, ...)
}

#' Tune XGBoost hyperparameters
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param is_classification Logical indicating if this is a
#'   classification problem
#' @param param_grid Named list of parameter values to try
#' @param cv_folds Number of cross-validation folds (default: 5)
#' @param early_stopping_rounds Early stopping rounds (default: 10)
#' @param verbose Logical indicating whether to print progress (default: TRUE)
#' @param ... Additional arguments
#' @return A \code{tidylearn_model} object (the refit on full data using the
#'   best hyperparameters) with an attribute \code{"tuning_results"} containing
#'   a list with elements \code{param_grid}, \code{results} (per-combination CV
#'   output), \code{best_params}, \code{best_iteration}, \code{best_score}, and
#'   \code{minimize}.
#' @export
tl_tune_xgboost <- function(data, formula, is_classification = FALSE,
                            param_grid = NULL, cv_folds = 5,
                            early_stopping_rounds = 10,
                            verbose = TRUE, ...) {
  # Check if xgboost is installed
  tl_check_packages("xgboost")

  # Set default parameter grid if not provided
  if (is.null(param_grid)) {
    param_grid <- list(
      max_depth = c(3, 6, 9),
      eta = c(0.01, 0.1, 0.3),
      subsample = c(0.7, 1.0),
      colsample_bytree = c(0.7, 1.0),
      min_child_weight = c(1, 3, 5),
      gamma = c(0, 0.1, 0.2)
    )
  }

  # Parse formula
  response_var <- all.vars(formula)[1]

  # Prepare data for XGBoost
  x_mat <- stats::model.matrix(formula, data = data)[, -1, drop = FALSE]
  y <- data[[response_var]]

  # Prepare response variable based on problem type
  if (is_classification) {
    if (!is.factor(y)) {
      y <- factor(y)
    }

    if (length(levels(y)) == 2) {
      # Binary classification: convert to 0/1
      y_numeric <- as.integer(y) - 1
      objective <- "binary:logistic"
      eval_metric <- "logloss"
    } else {
      # Multiclass classification: convert to 0-based index
      y_numeric <- as.integer(y) - 1
      objective <- "multi:softprob"
      eval_metric <- "mlogloss"
      num_class <- length(levels(y))
    }
  } else {
    # Regression
    y_numeric <- y
    objective <- "reg:squarederror"
    eval_metric <- "rmse"
  }

  # Create DMatrix object
  dtrain <- xgboost::xgb.DMatrix(data = as.matrix(x_mat), label = y_numeric)

  # Create parameter grid
  param_df <- expand.grid(param_grid)

  if (verbose) {
    message("Tuning XGBoost with ", nrow(param_df), " parameter combinations")
    message("Cross-validation with ", cv_folds, " folds")
  }

  # Initialize storage for results
  results <- list()

  # Loop through parameter combinations
  for (i in seq_len(nrow(param_df))) {
    if (verbose) {
      message(
        "Parameter set ", i, " of ", nrow(param_df),
        ": ",
        paste(
          names(param_df), param_df[i, ],
          sep = "=", collapse = ", "
        )
      )
    }

    # Extract parameters for this iteration
    params <- as.list(param_df[i, ])

    # Set basic parameters
    params$objective <- objective
    params$eval_metric <- eval_metric

    # Add num_class parameter for multiclass classification
    if (is_classification && length(levels(y)) > 2) {
      params$num_class <- num_class
    }

    # Run cross-validation
    cv_result <- xgboost::xgb.cv(
      params = params,
      data = dtrain,
      nrounds = 1000,  # Set high, will be limited by early stopping
      nfold = cv_folds,
      early_stopping_rounds = early_stopping_rounds,
      verbose = ifelse(verbose, 1, 0),
      ...
    )

    # Extract best iteration and performance
    best_iteration <- cv_result$best_iteration
    metric_col <- paste0("test_", eval_metric, "_mean")
    best_score <- cv_result$evaluation_log[
      best_iteration,
    ][[metric_col]]

    # Store results
    results[[i]] <- list(
      params = params,
      best_iteration = best_iteration,
      best_score = best_score,
      cv_result = cv_result
    )

    if (verbose) {
      message("  Best iteration: ", best_iteration,
              ", Best ", eval_metric, ": ", round(best_score, 6))
    }
  }

  # Find best parameter set
  best_scores <- sapply(results, function(x) x$best_score)

  # Determine if we should minimize or maximize the metric
  minimize <- eval_metric %in% c("error", "logloss", "mlogloss", "rmse", "mae")

  if (minimize) {
    best_idx <- which.min(best_scores)
  } else {
    best_idx <- which.max(best_scores)
  }

  # Extract best parameters and iteration
  best_params <- results[[best_idx]]$params
  best_iteration <- results[[best_idx]]$best_iteration

  if (verbose) {
    message(
      "Best parameters found: ",
      paste(
        names(best_params), best_params,
        sep = "=", collapse = ", "
      )
    )
    message("Best iteration: ", best_iteration)
    best_sc <- results[[best_idx]]$best_score
    message(
      "Best ", eval_metric, ": ",
      round(best_sc, 6)
    )
  }

  # Train final model with best parameters
  final_model <- xgboost::xgb.train(
    params = best_params,
    data = dtrain,
    nrounds = best_iteration,
    verbose = 0,
    ...
  )

  # Store additional information for later use
  attr(final_model, "feature_names") <- colnames(x_mat)
  attr(final_model, "response_var") <- response_var
  attr(final_model, "is_classification") <- is_classification

  if (is_classification) {
    attr(final_model, "response_levels") <- levels(y)
  }

  # Create tidylearn model wrapper
  model <- structure(
    list(
      spec = list(
        paradigm = "supervised",
        formula = formula,
        method = "xgboost",
        is_classification = is_classification,
        response_var = response_var
      ),
      fit = final_model,
      data = data
    ),
    class = c("tidylearn_xgboost", "tidylearn_supervised", "tidylearn_model")
  )

  # Add tuning results to model
  attr(model, "tuning_results") <- list(
    param_grid = param_grid,
    results = results,
    best_params = best_params,
    best_iteration = best_iteration,
    best_score = results[[best_idx]]$best_score,
    minimize = minimize
  )

  model
}

#' Generate SHAP values for XGBoost model interpretation
#'
#' @param model A tidylearn XGBoost model object
#' @param data Data for SHAP value calculation
#'   (default: NULL, uses training data)
#' @param n_samples Number of samples to use (default: 100, NULL for all)
#' @param trees_idx Trees to include (default: NULL, uses all trees)
#' @return A data frame with one column of SHAP values per feature, a
#'   \code{BIAS} column, a \code{row_id} column, and the original data columns
#'   appended for reference.
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
#'   shap <- tl_xgboost_shap(model, n_samples = 20)
#' }
#' }
#' @export
tl_xgboost_shap <- function(model, data = NULL,
                            n_samples = 100,
                            trees_idx = NULL) {
  # Check if model is an XGBoost model
  if (!inherits(model, "tidylearn_model") || model$spec$method != "xgboost") {
    stop("This function requires an XGBoost model", call. = FALSE)
  }

  # Check if xgboost is installed
  tl_check_packages("xgboost")

  # Use training data if data is not provided
  if (is.null(data)) {
    data <- model$data
  }

  # Extract model and feature names
  xgb_model <- model$fit
  feature_names <- attr(xgb_model, "feature_names")

  # Prepare data for XGBoost
  formula <- model$spec$formula
  x_mat <- stats::model.matrix(formula, data = data)[, -1, drop = FALSE]

  # Sample data if needed
  if (!is.null(n_samples) && n_samples < nrow(x_mat)) {
    sample_idx <- sample(nrow(x_mat), n_samples)
    x_mat <- x_mat[sample_idx, , drop = FALSE]
    data <- data[sample_idx, ]
  }

  # Create DMatrix object
  dmatrix <- xgboost::xgb.DMatrix(data = as.matrix(x_mat))

  # Calculate SHAP values
  shap_values <- predict(
    xgb_model, dmatrix,
    predcontrib = TRUE,
    approxcontrib = FALSE,
    trees_idx = trees_idx
  )

  # Remove BIAS column (last column) if present
  if (ncol(shap_values) == ncol(x_mat) + 1) {
    bias <- shap_values[, ncol(shap_values)]
    shap_values <- shap_values[, -ncol(shap_values), drop = FALSE]
  } else {
    bias <- rep(0, nrow(shap_values))
  }

  # Convert to data frame and add feature names
  shap_df <- as.data.frame(shap_values)
  colnames(shap_df) <- feature_names

  # Add bias column
  shap_df$BIAS <- bias

  # Add row identifiers
  shap_df$row_id <- seq_len(nrow(shap_df))

  # Add original data for reference
  if (nrow(shap_df) == nrow(data)) {
    # Combine with original data for later reference
    for (col in names(data)) {
      if (!col %in% names(shap_df)) {  # Avoid duplicates
        shap_df[[col]] <- data[[col]]
      }
    }
  }

  shap_df
}

#' Plot SHAP summary for XGBoost model
#'
#' @param model A tidylearn XGBoost model object
#' @param data Data for SHAP value calculation
#'   (default: NULL, uses training data)
#' @param top_n Number of top features to display (default: 10)
#' @param n_samples Number of samples to use
#'   (default: 100, NULL for all)
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_point
#'   scale_color_gradient labs theme_minimal
#' @examples
#' \donttest{
#' if (requireNamespace("xgboost", quietly = TRUE)) {
#'   model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
#'   tl_plot_xgboost_shap_summary(model, n_samples = 20)
#' }
#' }
#' @export
tl_plot_xgboost_shap_summary <- function(model,
                                         data = NULL,
                                         top_n = 10,
                                         n_samples = 100) {
  # Resolve data and handle sampling here so feature values and SHAP values

  # come from exactly the same rows
  use_data <- if (is.null(data)) model$data else data
  if (!is.null(n_samples) && n_samples < nrow(use_data)) {
    sample_idx <- sample(nrow(use_data), n_samples)
    use_data <- use_data[sample_idx, , drop = FALSE]
  }

  # Calculate SHAP values on the (already-sampled) data; pass n_samples = NULL
  shap_df <- tl_xgboost_shap(model, use_data, n_samples = NULL)

  # Convert wide to long format for plotting
  feature_cols <- attr(model$fit, "feature_names")

  # Compute absolute mean SHAP value for each feature
  feature_importance <- sapply(feature_cols, function(feat) {
    mean(abs(shap_df[[feat]]))
  })

  # Sort features by importance
  sorted_features <- names(sort(feature_importance, decreasing = TRUE))

  # Limit to top features
  if (length(sorted_features) > top_n) {
    top_features <- sorted_features[seq_len(top_n)]
  } else {
    top_features <- sorted_features
  }

  # Prepare data for plotting (use_data has the same rows as shap_df)
  plot_data <- NULL
  for (feat in top_features) {
    # Get feature values from the sampled data (same rows as SHAP values)
    feat_vals <- if (feat %in% names(use_data)) use_data[[feat]] else NA
    shap_vals <- shap_df[[feat]]

    # Combine into data frame
    feat_data <- data.frame(
      feature = feat,
      feature_value = feat_vals,
      shap_value = shap_vals,
      abs_shap_value = abs(shap_vals)
    )

    # Add to plot data
    if (is.null(plot_data)) {
      plot_data <- feat_data
    } else {
      plot_data <- rbind(plot_data, feat_data)
    }
  }

  # Create plot
  shap_aes <- ggplot2::aes(
    x = reorder(feature, abs_shap_value),
    y = shap_value
  )
  color_aes <- ggplot2::aes(color = feature_value)
  grad <- ggplot2::scale_color_gradient2(
    low = "blue", mid = "white",
    high = "red", midpoint = 0
  )
  shap_labs <- ggplot2::labs(
    title = "SHAP Feature Importance",
    subtitle = "Features sorted by mean absolute SHAP value",
    x = NULL,
    y = "SHAP Value (impact on prediction)",
    color = "Feature Value"
  )

  if (requireNamespace("ggforce", quietly = TRUE)) {
    # Use violin plots with jittered points
    p <- ggplot2::ggplot(plot_data, shap_aes) +
      ggforce::geom_sina(
        color_aes, size = 2, alpha = 0.7
      ) +
      grad +
      ggplot2::coord_flip() +
      shap_labs +
      ggplot2::theme_minimal()
  } else {
    # Fall back to basic jittered points
    p <- ggplot2::ggplot(plot_data, shap_aes) +
      ggplot2::geom_jitter(
        color_aes, width = 0.2, alpha = 0.7
      ) +
      grad +
      ggplot2::coord_flip() +
      shap_labs +
      ggplot2::theme_minimal()
  }

  p
}

#' Plot SHAP dependence for a specific feature
#'
#' @param model A tidylearn XGBoost model object
#' @param feature Feature name to plot
#' @param interaction_feature Feature to use for coloring (default: NULL)
#' @param data Data for SHAP value calculation
#'   (default: NULL, uses training data)
#' @param n_samples Number of samples to use
#'   (default: 100, NULL for all)
#' @return A \code{\link[ggplot2]{ggplot}} object.
#' @importFrom ggplot2 ggplot aes geom_point
#'   geom_smooth scale_color_gradient labs theme_minimal
#' @export
tl_plot_xgboost_shap_dependence <- function( # nolint: object_length_linter.
    model, feature,
    interaction_feature = NULL,
    data = NULL,
    n_samples = 100) {
  # Check if feature exists
  feature_names <- attr(model$fit, "feature_names")
  if (!feature %in% feature_names) {
    stop(
      "Feature not found in model: ", feature,
      call. = FALSE
    )
  }

  # Check interaction feature if provided
  if (!is.null(interaction_feature) &&
        !interaction_feature %in% feature_names) {
    stop(
      "Interaction feature not found in model: ",
      interaction_feature, call. = FALSE
    )
  }

  # Calculate SHAP values
  shap_df <- tl_xgboost_shap(model, data, n_samples)

  # Use original data if available
  if (is.null(data)) {
    data <- model$data
  }

  # Sample data if needed
  if (!is.null(n_samples) && n_samples < nrow(data)) {
    sample_idx <- sample(nrow(data), n_samples)
    data <- data[sample_idx, ]
  }

  # Get feature values and SHAP values
  feature_values <- data[[feature]]
  shap_values <- shap_df[[feature]]

  # Create plot data
  plot_data <- data.frame(
    feature_value = feature_values,
    shap_value = shap_values
  )

  # Add interaction feature if provided
  if (!is.null(interaction_feature)) {
    plot_data$interaction_value <- data[[interaction_feature]]
  }

  # Create plot
  base_aes <- ggplot2::aes(
    x = feature_value, y = shap_value
  )
  dep_title <- paste(
    "SHAP Dependence Plot for", feature
  )

  if (is.null(interaction_feature)) {
    # Without interaction feature
    p <- ggplot2::ggplot(plot_data, base_aes) +
      ggplot2::geom_point(alpha = 0.7) +
      ggplot2::geom_smooth(
        method = "loess", formula = y ~ x,
        se = TRUE, color = "red"
      ) +
      ggplot2::labs(
        title = dep_title,
        x = feature,
        y = "SHAP Value (impact on prediction)"
      ) +
      ggplot2::theme_minimal()
  } else {
    # With interaction feature
    int_aes <- ggplot2::aes(
      x = feature_value, y = shap_value,
      color = interaction_value
    )
    p <- ggplot2::ggplot(plot_data, int_aes) +
      ggplot2::geom_point(alpha = 0.7) +
      ggplot2::scale_color_gradient2(
        low = "blue", mid = "white",
        high = "red"
      ) +
      ggplot2::labs(
        title = dep_title,
        subtitle = paste(
          "Colored by", interaction_feature
        ),
        x = feature,
        y = "SHAP Value (impact on prediction)",
        color = interaction_feature
      ) +
      ggplot2::theme_minimal()
  }

  p
}

#' @title Table Functions for tidylearn
#' @name tidylearn-tables
#' @description Functions for producing formatted gt tables
#'   from tidylearn models. Provides a parallel interface to
#'   the plot functions: \code{tl_table(model, type)}
#'   dispatches to the appropriate table formatter based on model type.
#'   Requires the gt package (suggested dependency).
NULL

# ── Internal helpers ──────────────────────────────────────────────────────────

#' Apply tidylearn gt theme
#' @param gt_tbl A gt table object
#' @param title Optional table title
#' @param subtitle Optional subtitle
#' @param source_note Optional source note
#' @return A styled gt object
#' @keywords internal
#' @noRd
tl_gt_theme <- function(gt_tbl, title = NULL,
                        subtitle = NULL,
                        source_note = NULL) {
  gt_tbl <- gt_tbl %>%
    gt::tab_options(
      heading.background.color = "#2c3e50",
      heading.title.font.size = gt::px(16),
      heading.subtitle.font.size = gt::px(12),
      column_labels.background.color = "#34495e",
      column_labels.font.weight = "bold",
      row.striping.include_table_body = TRUE,
      row.striping.background_color = "#f8f9fa",
      table.border.top.color = "#2c3e50",
      table.border.bottom.color = "#2c3e50",
      table.font.size = gt::px(13)
    ) %>%
    gt::tab_style(
      style = gt::cell_text(color = "white"),
      locations = gt::cells_column_labels()
    )

  if (!is.null(title)) {
    gt_tbl <- gt_tbl %>% gt::tab_header(title = title, subtitle = subtitle)
  }
  if (!is.null(source_note)) {
    gt_tbl <- gt_tbl %>% gt::tab_source_note(source_note = source_note)
  }

  gt_tbl
}

#' Build model info string for table footnotes
#' @param model A tidylearn model object
#' @return Character string describing the model
#' @keywords internal
#' @noRd
tl_model_info <- function(model) {
  method <- model$spec$method
  if (model$spec$paradigm == "supervised") {
    task <- if (model$spec$is_classification) "classification" else "regression"
    formula_text <- deparse(model$spec$formula)
    paste0("tidylearn | ", method, " (", task, ") | ", formula_text,
           " | n = ", nrow(model$data))
  } else {
    paste0("tidylearn | ", method, " | n = ", nrow(model$data))
  }
}

# ── Main dispatcher ──────────────────────────────────────────────────────────

#' Create formatted tables for tidylearn models
#'
#' Dispatches to the appropriate table function based on model type and
#' requested table type. Requires the gt package.
#'
#' @param model A tidylearn model object
#' @param type Table type (default: "auto"). For supervised models: "metrics",
#'   "coefficients", "confusion", "importance". For unsupervised models:
#'   "variance", "loadings", "clusters". MDS models are not supported.
#' @param ... Additional arguments passed to the underlying table function
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_table(model)
#' tl_table(model, type = "coefficients")
#' }
tl_table <- function(model, type = "auto", ...) {
  tl_check_packages("gt")

  if (!inherits(model, "tidylearn_model")) {
    stop("'model' must be a tidylearn_model object", call. = FALSE)
  }

  if (inherits(model, "tidylearn_supervised")) {
    tl_table_model(model, type, ...)
  } else if (inherits(model, "tidylearn_unsupervised")) {
    tl_table_unsupervised(model, type, ...)
  }
}

#' Supervised table dispatcher
#' @keywords internal
#' @noRd
tl_table_model <- function(model, type = "auto", ...) {
  is_class <- model$spec$is_classification

  if (type == "auto") {
    type <- "metrics"
  }

  switch(
    type,
    "metrics"      = tl_table_metrics(model, ...),
    "coefficients" = tl_table_coefficients(model, ...),
    "confusion"    = tl_table_confusion(model, ...),
    "importance"   = tl_table_importance(model, ...),
    stop(
      "Unknown table type '", type, "'. ",
      if (is_class) {
        "Use: 'metrics', 'coefficients', 'confusion', or 'importance'."
      } else {
        "Use: 'metrics', 'coefficients', or 'importance'."
      },
      call. = FALSE
    )
  )
}

#' Unsupervised table dispatcher
#' @keywords internal
#' @noRd
tl_table_unsupervised <- function(model, type = "auto", ...) {
  method <- model$spec$method

  if (method == "mds") {
    stop("Table output is not supported for MDS models. Use plot() instead.",
         call. = FALSE)
  }

  if (type == "auto") {
    type <- switch(
      method,
      "pca"    = "variance",
      "kmeans" = , "pam" = , "clara" = , "dbscan" = , "hclust" = "clusters",
      stop("No default table type for method '", method, "'.", call. = FALSE)
    )
  }

  switch(
    type,
    "variance" = tl_table_variance(model, ...),
    "loadings" = tl_table_loadings(model, ...),
    "clusters" = tl_table_clusters(model, ...),
    stop(
      "Unknown table type '", type,
      "'. Use: 'variance', 'loadings', or 'clusters'.",
      call. = FALSE
    )
  )
}

# ── Supervised table functions ───────────────────────────────────────────────

#' Formatted evaluation metrics table
#'
#' Produces a styled gt table of model evaluation metrics from
#' \code{\link{tl_evaluate}}.
#'
#' @param model A tidylearn supervised model object
#' @param new_data Optional test data. If NULL, uses training data.
#' @param digits Number of decimal places (default: 4)
#' @param ... Additional arguments passed to \code{tl_evaluate}
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_table_metrics(model)
#' }
tl_table_metrics <- function(model, new_data = NULL, digits = 4, ...) {
  tl_check_packages("gt")

  eval_results <- tl_evaluate(model, new_data = new_data, ...)

  eval_results %>%
    dplyr::mutate(
      metric = gsub("_", " ", .data$metric),
      metric = tools::toTitleCase(.data$metric)
    ) %>%
    gt::gt() %>%
    gt::cols_label(metric = "Metric", value = "Value") %>%
    gt::fmt_number(columns = "value", decimals = digits) %>%
    tl_gt_theme(
      title = "Model Evaluation Metrics",
      source_note = tl_model_info(model)
    )
}

#' Formatted model coefficients table
#'
#' Produces a styled gt table of model coefficients. Supports linear,
#' polynomial, logistic, ridge, lasso, and elastic net models.
#'
#' @param model A tidylearn model object
#' @param lambda For regularised models: "1se" (default) or "min"
#' @param digits Number of decimal places (default: 4)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
#' tl_table_coefficients(model)
#' }
tl_table_coefficients <- function(model, lambda = "1se", digits = 4, ...) {
  tl_check_packages("gt")

  method <- model$spec$method

  if (method %in% c("linear", "polynomial", "logistic")) {
    summ <- summary(model$fit)
    coef_mat <- summ$coefficients
    coef_tbl <- tibble::tibble(
      term = rownames(coef_mat),
      estimate = coef_mat[, 1],
      std_error = coef_mat[, 2],
      statistic = coef_mat[, 3],
      p_value = coef_mat[, 4]
    ) %>%
      dplyr::mutate(significant = ifelse(.data$p_value < 0.05, "*", ""))

    stat_label <- if (method == "logistic") "z value" else "t value"

    gt_tbl <- coef_tbl %>%
      gt::gt() %>%
      gt::cols_label(
        term = "Term", estimate = "Estimate", std_error = "Std. Error",
        statistic = stat_label, p_value = "p", significant = ""
      ) %>%
      gt::fmt_number(
        columns = c("estimate", "std_error", "statistic"),
        decimals = digits
      ) %>%
      gt::fmt_scientific(columns = "p_value", decimals = 2) %>%
      gt::tab_style(
        style = gt::cell_fill(color = "#d4edda"),
        locations = gt::cells_body(rows = coef_tbl$p_value < 0.05)
      ) %>%
      tl_gt_theme(
        title = paste0(tools::toTitleCase(method), " Model Coefficients"),
        source_note = tl_model_info(model)
      )

  } else if (method %in% c("ridge", "lasso", "elastic_net")) {
    fit <- model$fit

    if (lambda == "1se") {
      lambda_val <- attr(fit, "lambda_1se")
    } else if (lambda == "min") {
      lambda_val <- attr(fit, "lambda_min")
    } else {
      lambda_val <- lambda
    }

    coefs <- as.matrix(stats::coef(fit, s = lambda_val))

    coef_tbl <- tibble::tibble(
      term = rownames(coefs),
      estimate = as.vector(coefs)
    ) %>%
      dplyr::mutate(abs_estimate = abs(.data$estimate)) %>%
      dplyr::arrange(dplyr::desc(.data$abs_estimate))

    gt_tbl <- coef_tbl %>%
      gt::gt() %>%
      gt::cols_label(
        term = "Term", estimate = "Coefficient", abs_estimate = "|Coefficient|"
      ) %>%
      gt::fmt_number(
        columns = c("estimate", "abs_estimate"), decimals = digits
      ) %>%
      gt::tab_style(
        style = gt::cell_text(color = "#999999"),
        locations = gt::cells_body(rows = coef_tbl$estimate == 0)
      ) %>%
      tl_gt_theme(
        title = paste0(tools::toTitleCase(method), " Coefficients"),
        subtitle = paste0(
          "lambda = ", signif(lambda_val, 4),
          " (", lambda, ")"
        ),
        source_note = tl_model_info(model)
      )

  } else {
    stop(
      "Coefficient table not available for method '", method,
      "'. Try type = 'importance' instead.",
      call. = FALSE
    )
  }

  gt_tbl
}

#' Formatted confusion matrix table
#'
#' Produces a styled gt confusion matrix with correct predictions highlighted.
#' Only available for classification models.
#'
#' @param model A tidylearn classification model
#' @param new_data Optional test data. If NULL, uses training data.
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(iris, Species ~ ., method = "forest")
#' tl_table_confusion(model)
#' }
tl_table_confusion <- function(model, new_data = NULL, ...) {
  tl_check_packages("gt")

  if (!model$spec$is_classification) {
    stop("Confusion matrix table is only available for classification models",
         call. = FALSE)
  }

  if (is.null(new_data)) new_data <- model$data

  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) actuals <- factor(actuals)
  predicted <- predict(model, new_data, type = "class")$.pred

  cm <- table(Actual = actuals, Predicted = predicted)
  cm_df <- as.data.frame.matrix(cm)
  cm_df$Actual <- rownames(cm_df)
  cm_df <- cm_df %>% dplyr::select("Actual", dplyr::everything())

  gt_tbl <- gt::gt(cm_df, rowname_col = "Actual") %>%
    gt::tab_spanner(label = "Predicted", columns = -1) %>%
    gt::tab_stubhead(label = "Actual") %>%
    tl_gt_theme(
      title = "Confusion Matrix",
      source_note = tl_model_info(model)
    )

  # Highlight diagonal (correct predictions)
  class_levels <- levels(actuals)
  for (cls in class_levels) {
    if (cls %in% colnames(cm_df)) {
      gt_tbl <- gt_tbl %>%
        gt::tab_style(
          style = gt::cell_fill(color = "#d4edda"),
          locations = gt::cells_body(columns = cls, rows = cls)
        )
    }
  }

  gt_tbl
}

#' Formatted feature importance table
#'
#' Produces a styled gt table of feature importance with a colour gradient.
#' Supports tree-based, regularised, and xgboost models.
#'
#' @param model A tidylearn model object
#' @param top_n Maximum number of features to display (default: 20)
#' @param digits Number of decimal places (default: 2)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(iris, Species ~ ., method = "forest")
#' tl_table_importance(model)
#' }
tl_table_importance <- function(model, top_n = 20, digits = 2, ...) {
  tl_check_packages("gt")

  method <- model$spec$method

  if (method %in% c("tree", "forest", "boost")) {
    imp_df <- tl_extract_importance(model)
  } else if (method %in% c("ridge", "lasso", "elastic_net")) {
    imp_df <- tl_get_importance_regularized(model)
  } else {
    stop("Importance table not available for method '", method, "'.",
         call. = FALSE)
  }

  imp_df <- imp_df %>%
    dplyr::arrange(dplyr::desc(.data$importance)) %>%
    dplyr::slice_head(n = top_n)

  imp_df %>%
    gt::gt() %>%
    gt::cols_label(feature = "Feature", importance = "Importance") %>%
    gt::fmt_number(columns = "importance", decimals = digits) %>%
    gt::data_color(
      columns = "importance",
      palette = c("#f8f9fa", "#2c3e50")
    ) %>%
    tl_gt_theme(
      title = "Feature Importance",
      subtitle = paste0("Top ", min(top_n, nrow(imp_df)), " features"),
      source_note = tl_model_info(model)
    )
}

# ── Unsupervised table functions ─────────────────────────────────────────────

#' Formatted PCA variance explained table
#'
#' Produces a styled gt table of variance explained by each principal component,
#' with a colour gradient on cumulative variance.
#'
#' @param model A tidylearn PCA model object
#' @param n_components Maximum number of components to show (default: all)
#' @param digits Number of decimal places (default: 4)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(iris[, 1:4], method = "pca")
#' tl_table_variance(model)
#' }
tl_table_variance <- function(model, n_components = NULL, digits = 4, ...) {
  tl_check_packages("gt")

  if (model$spec$method != "pca") {
    stop("Variance table is only available for PCA models", call. = FALSE)
  }

  var_tbl <- model$fit$variance_explained
  if (!is.null(n_components)) {
    var_tbl <- var_tbl %>% dplyr::slice_head(n = n_components)
  }

  var_tbl %>%
    gt::gt() %>%
    gt::cols_label(
      component = "Component", sdev = "Std. Dev.", variance = "Variance",
      prop_variance = "Proportion", cum_variance = "Cumulative"
    ) %>%
    gt::fmt_number(columns = c("sdev", "variance"), decimals = digits) %>%
    gt::fmt_percent(
      columns = c("prop_variance", "cum_variance"),
      decimals = 1
    ) %>%
    gt::data_color(
      columns = "cum_variance",
      palette = c("#ffffff", "#27ae60")
    ) %>%
    tl_gt_theme(
      title = "PCA Variance Explained",
      source_note = tl_model_info(model)
    )
}

#' Formatted PCA loadings table
#'
#' Produces a styled gt table of variable loadings on each principal component,
#' with a diverging colour scale to highlight strong loadings.
#'
#' @param model A tidylearn PCA model object
#' @param n_components Number of components to show (default: all)
#' @param digits Number of decimal places (default: 3)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(iris[, 1:4], method = "pca")
#' tl_table_loadings(model)
#' }
tl_table_loadings <- function(model, n_components = NULL, digits = 3, ...) {
  tl_check_packages("gt")

  if (model$spec$method != "pca") {
    stop("Loadings table is only available for PCA models", call. = FALSE)
  }

  loadings_wide <- model$fit$loadings
  if (!is.null(n_components)) {
    pc_cols <- paste0("PC", seq_len(n_components))
    loadings_wide <- loadings_wide %>%
      dplyr::select("variable", dplyr::any_of(pc_cols))
  }

  pc_cols <- setdiff(names(loadings_wide), "variable")

  loadings_wide %>%
    gt::gt() %>%
    gt::cols_label(variable = "Variable") %>%
    gt::fmt_number(columns = dplyr::all_of(pc_cols), decimals = digits) %>%
    gt::data_color(
      columns = dplyr::all_of(pc_cols),
      palette = c("#c0392b", "#ffffff", "#2980b9"),
      domain = c(-1, 1)
    ) %>%
    tl_gt_theme(
      title = "PCA Loadings",
      source_note = tl_model_info(model)
    )
}

#' Formatted cluster summary table
#'
#' Produces a styled gt table showing cluster sizes and mean feature values.
#' Supports kmeans, pam, clara, dbscan, and hclust models.
#'
#' @param model A tidylearn clustering model object
#' @param k For hclust models, the number of clusters to cut (default: 3)
#' @param digits Number of decimal places (default: 2)
#' @param ... Additional arguments (currently unused)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
#' tl_table_clusters(model)
#' }
tl_table_clusters <- function(model, k = 3, digits = 2, ...) {
  tl_check_packages("gt")

  method <- model$spec$method

  if (method == "kmeans") {
    centers <- model$fit$centers
    sizes <- model$fit$model$size
    summary_tbl <- centers %>%
      dplyr::mutate(size = sizes, .after = "cluster")
  } else if (method %in% c("pam", "clara")) {
    centers <- model$fit$medoids
    cluster_counts <- model$fit$clusters %>%
      dplyr::count(.data$cluster, name = "size")
    summary_tbl <- centers %>%
      dplyr::left_join(cluster_counts, by = "cluster")
  } else if (method == "hclust") {
    clusters <- stats::cutree(model$fit$model, k = k)
    data_with_clusters <- model$data %>%
      dplyr::select(where(is.numeric)) %>%
      dplyr::mutate(cluster = as.integer(clusters))
    summary_tbl <- data_with_clusters %>%
      dplyr::group_by(.data$cluster) %>%
      dplyr::summarise(
        size = dplyr::n(),
        dplyr::across(where(is.numeric) & !dplyr::any_of("cluster"), mean),
        .groups = "drop"
      )
  } else if (method == "dbscan") {
    cluster_assignments <- model$fit$clusters
    data_with_clusters <- model$data %>%
      dplyr::select(where(is.numeric)) %>%
      dplyr::mutate(cluster = cluster_assignments$cluster)
    summary_tbl <- data_with_clusters %>%
      dplyr::group_by(.data$cluster) %>%
      dplyr::summarise(
        size = dplyr::n(),
        dplyr::across(where(is.numeric) & !dplyr::any_of("cluster"), mean),
        .groups = "drop"
      )
  } else {
    stop("Cluster table not available for method '",
         method, "'.", call. = FALSE)
  }

  numeric_cols <- names(summary_tbl)[
    vapply(summary_tbl, is.numeric, logical(1))
  ]
  numeric_cols <- setdiff(numeric_cols, c("cluster", "size", "medoid_index"))

  summary_tbl %>%
    gt::gt() %>%
    gt::cols_label(cluster = "Cluster", size = "Size") %>%
    gt::fmt_number(columns = dplyr::all_of(numeric_cols), decimals = digits) %>%
    gt::fmt_integer(columns = dplyr::any_of(c("cluster", "size"))) %>%
    tl_gt_theme(
      title = "Cluster Summary",
      subtitle = paste0(
        method, " | ",
        length(unique(summary_tbl$cluster)), " clusters"
      ),
      source_note = tl_model_info(model)
    )
}

# ── Standalone comparison function ───────────────────────────────────────────

#' Compare multiple models in a formatted table
#'
#' Evaluates multiple tidylearn models and presents the results side-by-side
#' in a styled gt table.
#'
#' @param ... tidylearn model objects to compare
#' @param new_data Optional test data for evaluation. If NULL, uses the
#'   training data of the first model.
#' @param names Optional character vector of model names
#' @param digits Number of decimal places (default: 4)
#' @return A \code{\link[gt]{gt}} table object.
#' @export
#' @examples
#' \donttest{
#' m1 <- tl_model(mtcars, mpg ~ ., method = "linear")
#' m2 <- tl_model(mtcars, mpg ~ ., method = "lasso")
#' tl_table_comparison(m1, m2, names = c("Linear", "Lasso"))
#' }
tl_table_comparison <- function(..., new_data = NULL,
                                names = NULL,
                                digits = 4) {
  tl_check_packages("gt")

  models <- list(...)

  if (length(models) < 2) {
    stop("Provide at least 2 models to compare", call. = FALSE)
  }

  if (is.null(names)) {
    names <- vapply(models, function(m) {
      task <- if (m$spec$paradigm == "supervised") {
        if (m$spec$is_classification) "cls" else "reg"
      } else {
        m$spec$method
      }
      paste0(m$spec$method, " (", task, ")")
    }, character(1))
  }

  if (is.null(new_data)) {
    new_data <- models[[1]]$data
  }

  results <- purrr::map2_dfr(models, names, function(model, name) {
    eval_res <- tl_evaluate(model, new_data = new_data)
    eval_res$model <- name
    eval_res
  })

  wide_results <- results %>%
    dplyr::mutate(
      metric = gsub("_", " ", .data$metric),
      metric = tools::toTitleCase(.data$metric)
    ) %>%
    tidyr::pivot_wider(names_from = "model", values_from = "value")

  wide_results %>%
    gt::gt() %>%
    gt::cols_label(metric = "Metric") %>%
    gt::fmt_number(columns = -"metric", decimals = digits) %>%
    tl_gt_theme(
      title = "Model Comparison",
      subtitle = paste0(length(models), " models compared"),
      source_note = paste0("tidylearn | n = ", nrow(new_data))
    )
}

test_that("tl_table dispatches correctly for supervised models", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  tbl <- tl_table(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table dispatches correctly for unsupervised models", {
  skip_if_not_installed("gt")

  model <- tl_model(iris[, 1:4], method = "pca")
  tbl <- tl_table(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table rejects non-tidylearn objects", {
  skip_if_not_installed("gt")

  expect_error(tl_table(lm(mpg ~ wt, data = mtcars)), "tidylearn_model")
})

test_that("tl_table_metrics returns gt_tbl", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  tbl <- tl_table_metrics(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_coefficients works for linear models", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  tbl <- tl_table_coefficients(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_coefficients works for regularised models", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "lasso")
  tbl <- tl_table_coefficients(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_coefficients errors for unsupported methods", {
  skip_if_not_installed("gt")

  model <- tl_model(iris, Species ~ ., method = "forest")
  expect_error(tl_table_coefficients(model), "importance")
})

test_that("tl_table_confusion works for classification", {
  skip_if_not_installed("gt")

  model <- tl_model(iris, Species ~ ., method = "forest")
  tbl <- tl_table_confusion(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_confusion errors for regression", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  expect_error(tl_table_confusion(model), "classification")
})

test_that("tl_table_importance works for tree-based models", {
  skip_if_not_installed("gt")

  model <- tl_model(iris, Species ~ ., method = "forest")
  tbl <- tl_table_importance(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_variance works for PCA", {
  skip_if_not_installed("gt")

  model <- tl_model(iris[, 1:4], method = "pca")
  tbl <- tl_table_variance(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_variance errors for non-PCA models", {
  skip_if_not_installed("gt")

  model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
  expect_error(tl_table_variance(model), "PCA")
})

test_that("tl_table_loadings works for PCA", {
  skip_if_not_installed("gt")

  model <- tl_model(iris[, 1:4], method = "pca")
  tbl <- tl_table_loadings(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_clusters works for kmeans", {
  skip_if_not_installed("gt")

  model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
  tbl <- tl_table_clusters(model)
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table_comparison requires at least 2 models", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  expect_error(tl_table_comparison(model), "at least 2")
})

test_that("tl_table_comparison works with multiple models", {
  skip_if_not_installed("gt")

  m1 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "lasso")
  tbl <- tl_table_comparison(m1, m2, names = c("Linear", "Lasso"))
  expect_s3_class(tbl, "gt_tbl")
})

test_that("tl_table errors for unknown type", {
  skip_if_not_installed("gt")

  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  expect_error(tl_table(model, type = "nonexistent"), "Unknown table type")
})

#' @title Classification Functions for tidylearn
#' @name tidylearn-classification
#' @description Logistic regression and classification metrics functionality
#' @importFrom stats glm predict binomial
#' @importFrom tibble tibble as_tibble
#' @importFrom dplyr mutate
#' @importFrom ROCR prediction performance
#' @importFrom ggplot2 ggplot aes geom_line geom_abline labs theme_minimal
NULL

#' Fit a logistic regression model
#'
#' @param data A data frame containing the training data
#' @param formula A formula specifying the model
#' @param ... Additional arguments to pass to glm()
#' @return A fitted logistic regression model
#' @keywords internal
tl_fit_logistic <- function(data, formula, ...) {
  # Ensure response variable is a factor
  response_var <- all.vars(formula)[1]
  if (!is.factor(data[[response_var]])) {
    warning(
      "Converting response variable to factor ",
      "for logistic regression",
      call. = FALSE
    )
    data[[response_var]] <- factor(data[[response_var]])
  }

  # Fit the logistic regression model
  glm_model <- stats::glm(formula, data = data, family = stats::binomial(), ...)

  glm_model
}

#' Predict using a logistic regression model
#'
#' @param model A tidylearn logistic model object
#' @param new_data A data frame containing the new data
#' @param type Type of prediction: "prob" (default), "class", "response"
#' @param ... Additional arguments
#' @return Predictions
#' @keywords internal
tl_predict_logistic <- function(model, new_data, type = "prob", ...) {
  # Extract necessary information
  response_var <- model$spec$response_var

  # Get original response variable from training data
  original_response <- model$data[[response_var]]
  if (!is.factor(original_response)) {
    original_response <- factor(original_response)
  }
  class_levels <- levels(original_response)

  # Make predictions based on the type
  if (type == "response") {
    # Get the linear predictor
    preds <- stats::predict(
      model$fit, newdata = new_data, type = "response", ...
    )
    preds
  } else if (type == "prob") {
    # Binary classification
    if (length(class_levels) == 2) {
      # Get probabilities for the positive class
      pos_probs <- stats::predict(
        model$fit, newdata = new_data,
        type = "response", ...
      )
      neg_probs <- 1 - pos_probs

      # Create a data frame with probabilities for each class
      prob_df <- tibble::tibble(
        !!class_levels[1] := neg_probs,
        !!class_levels[2] := pos_probs
      )

      prob_df
    } else {
      # Multiclass classification (requires multinomial logistic regression)
      stop(
        "Multiclass logistic regression not ",
        "currently implemented",
        call. = FALSE
      )
    }
  } else if (type == "class") {
    # Binary classification
    if (length(class_levels) == 2) {
      # Get probabilities for the positive class
      pos_probs <- stats::predict(
        model$fit, newdata = new_data,
        type = "response", ...
      )

      # Classify based on probability > 0.5
      pred_classes <- ifelse(pos_probs > 0.5, class_levels[2], class_levels[1])
      pred_classes <- factor(pred_classes, levels = class_levels)

      pred_classes
    } else {
      # Multiclass classification (requires multinomial logistic regression)
      stop(
        "Multiclass logistic regression not ",
        "currently implemented",
        call. = FALSE
      )
    }
  } else {
    stop(
      "Invalid prediction type. ",
      "Use 'prob', 'class', or 'response'.",
      call. = FALSE
    )
  }
}

#' Plot ROC curve for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A ggplot object with ROC curve
#' @importFrom ROCR prediction performance
#' @importFrom ggplot2 ggplot aes geom_line geom_abline labs theme_minimal
#' @keywords internal
tl_plot_roc <- function(model, new_data = NULL, ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # For binary classification
  if (length(levels(actuals)) == 2) {
    # Get probabilities
    probs <- predict(model, new_data, type = "prob")
    pos_class <- levels(actuals)[2]
    pos_probs <- probs[[pos_class]]

    # Convert actuals to binary (0/1)
    binary_actuals <- as.integer(actuals == pos_class)

    # Create ROC curve
    pred_obj <- ROCR::prediction(pos_probs, binary_actuals)
    perf <- ROCR::performance(pred_obj, "tpr", "fpr")

    # Calculate AUC
    auc <- unlist(ROCR::performance(pred_obj, "auc")@y.values)

    # Create data frame for plotting
    roc_data <- tibble::tibble(
      fpr = unlist(perf@x.values),
      tpr = unlist(perf@y.values)
    )

    # Create the plot
    p <- ggplot2::ggplot(roc_data, ggplot2::aes(x = fpr, y = tpr)) +
      ggplot2::geom_line(color = "blue", linewidth = 1) +
      ggplot2::geom_abline(
        intercept = 0, slope = 1,
        linetype = "dashed", color = "gray"
      ) +
      ggplot2::labs(
        title = "ROC Curve",
        subtitle = paste0("AUC = ", round(auc, 3)),
        x = "False Positive Rate",
        y = "True Positive Rate"
      ) +
      ggplot2::coord_fixed() +
      ggplot2::theme_minimal()

    p
  } else {
    # Multiclass ROC (one-vs-rest)
    stop("Multiclass ROC curves not currently implemented", call. = FALSE)
  }
}

#' Plot confusion matrix for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A ggplot object with confusion matrix
#' @importFrom ggplot2 ggplot aes geom_tile geom_text scale_fill_gradient
#' @keywords internal
tl_plot_confusion <- function(model, new_data = NULL, ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # Get predicted classes
  predicted <- predict(model, new_data, type = "class")$.pred

  # Create confusion matrix
  cm <- table(Actual = actuals, Predicted = predicted)

  # Convert to data frame for plotting
  cm_df <- as.data.frame(as.table(cm))

  # Calculate percentages
  cm_df$percentage <- cm_df$Freq / sum(cm_df$Freq) * 100

  # Create the plot
  p <- ggplot2::ggplot(cm_df, ggplot2::aes(x = Predicted, y = Actual)) +
    ggplot2::geom_tile(ggplot2::aes(fill = Freq), color = "white") +
    ggplot2::geom_text(
      ggplot2::aes(
        label = paste0(
          Freq, "\n(", round(percentage, 1), "%)"
        )
      ),
      color = "black", size = 4
    ) +
    ggplot2::scale_fill_gradient(low = "white", high = "steelblue") +
    ggplot2::labs(
      title = "Confusion Matrix",
      x = "Predicted Class",
      y = "Actual Class",
      fill = "Count"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(panel.grid = ggplot2::element_blank())

  p
}

#' Plot precision-recall curve for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param ... Additional arguments
#' @return A ggplot object with precision-recall curve
#' @importFrom ROCR prediction performance
#' @importFrom ggplot2 ggplot aes geom_line labs theme_minimal
#' @keywords internal
tl_plot_precision_recall <- function(model, new_data = NULL, ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # For binary classification
  if (length(levels(actuals)) == 2) {
    # Get probabilities
    probs <- predict(model, new_data, type = "prob")
    pos_class <- levels(actuals)[2]
    pos_probs <- probs[[pos_class]]

    # Convert actuals to binary (0/1)
    binary_actuals <- as.integer(actuals == pos_class)

    # Create precision-recall curve
    pred_obj <- ROCR::prediction(pos_probs, binary_actuals)
    perf <- ROCR::performance(pred_obj, "prec", "rec")

    # Calculate area under PR curve
    pr_auc <- tl_calculate_pr_auc(perf)

    # Create data frame for plotting
    pr_data <- tibble::tibble(
      recall = unlist(perf@x.values),
      precision = unlist(perf@y.values)
    )

    # Remove NA/NaN values
    pr_data <- pr_data[!is.na(pr_data$precision) & !is.na(pr_data$recall), ]

    # Create the plot
    p <- ggplot2::ggplot(pr_data, ggplot2::aes(x = recall, y = precision)) +
      ggplot2::geom_line(color = "blue", linewidth = 1) +
      ggplot2::labs(
        title = "Precision-Recall Curve",
        subtitle = paste0("Area Under PR Curve = ", round(pr_auc, 3)),
        x = "Recall",
        y = "Precision"
      ) +
      ggplot2::ylim(0, 1) +
      ggplot2::xlim(0, 1) +
      ggplot2::theme_minimal()

    p
  } else {
    # Multiclass precision-recall curves (not implemented)
    stop(
      "Multiclass precision-recall curves not ",
      "currently implemented",
      call. = FALSE
    )
  }
}

#' Plot calibration curve for a classification model
#'
#' @param model A tidylearn classification model object
#' @param new_data Optional data frame for evaluation
#'   (if NULL, uses training data)
#' @param bins Number of bins for grouping predictions (default: 10)
#' @param ... Additional arguments
#' @return A ggplot object with calibration curve
#' @importFrom ggplot2 ggplot aes geom_point geom_line
#'   geom_abline labs theme_minimal
#' @keywords internal
tl_plot_calibration <- function(model, new_data = NULL, bins = 10, ...) {
  if (is.null(new_data)) {
    new_data <- model$data
  }

  # Get actual values
  response_var <- model$spec$response_var
  actuals <- new_data[[response_var]]
  if (!is.factor(actuals)) {
    actuals <- factor(actuals)
  }

  # For binary classification
  if (length(levels(actuals)) == 2) {
    # Get probabilities
    probs <- predict(model, new_data, type = "prob")
    pos_class <- levels(actuals)[2]
    pos_probs <- probs[[pos_class]]

    # Convert actuals to binary (0/1)
    binary_actuals <- as.integer(actuals == pos_class)

    # Create bins of predictions
    bin_breaks <- seq(0, 1, length.out = bins + 1)

    # Assign each prediction to a bin
    bins_idx <- cut(
      pos_probs, breaks = bin_breaks,
      labels = FALSE, include.lowest = TRUE
    )

    # Calculate mean predicted probability and fraction of
    # positives for each bin
    calibration_data <- tibble::tibble(
      bin = bins_idx,
      prob = pos_probs,
      actual = binary_actuals
    ) %>%
      dplyr::group_by(.data$bin) %>%
      dplyr::summarize(
        mean_pred_prob = mean(.data$prob),
        frac_pos = mean(.data$actual),
        n = dplyr::n(),
        .groups = "drop"
      )

    # Create the plot
    p <- ggplot2::ggplot(
      calibration_data,
      ggplot2::aes(x = mean_pred_prob, y = frac_pos)
    ) +
      ggplot2::geom_point(ggplot2::aes(size = n), alpha = 0.7) +
      ggplot2::geom_line(color = "blue") +
      ggplot2::geom_abline(
        intercept = 0, slope = 1,
        linetype = "dashed", color = "gray"
      ) +
      ggplot2::labs(
        title = "Calibration Curve",
        subtitle = paste0(
          "Perfectly calibrated predictions ",
          "should lie on the diagonal"
        ),
        x = "Mean Predicted Probability",
        y = "Fraction of Positives",
        size = "Count"
      ) +
      ggplot2::coord_fixed() +
      ggplot2::theme_minimal()

    p
  } else {
    # Multiclass calibration (not implemented)
    stop(
      "Multiclass calibration curves not ",
      "currently implemented",
      call. = FALSE
    )
  }
}

#' @title Data Reading Backends for tidylearn
#' @name tidylearn-read-backends
#' @description Backend readers for databases and cloud/API sources.
#'   All backends are optional dependencies checked at call time via
#'   \code{tl_check_packages()}.
#'
#' @details
#' Database backends (via \pkg{DBI}):
#' \itemize{
#'   \item \strong{SQLite}: via \pkg{RSQLite}
#'   \item \strong{PostgreSQL}: via \pkg{RPostgres}
#'   \item \strong{MySQL/MariaDB}: via \pkg{RMariaDB}
#'   \item \strong{BigQuery}: via \pkg{bigrquery}
#' }
#'
#' Cloud/API backends:
#' \itemize{
#'   \item \strong{S3}: via \pkg{paws.storage}
#'   \item \strong{GitHub}: via base \code{download.file()}
#'   \item \strong{Kaggle}: via Kaggle CLI
#' }
NULL

# ---- Database readers ----

#' Read from a DBI database connection
#'
#' Executes a SQL query against an existing \pkg{DBI} connection and returns
#' the result as a \code{tidylearn_data} object. The connection is not closed
#' by this function — the caller is responsible for managing the connection
#' lifecycle.
#'
#' @param conn A \pkg{DBI} connection object (e.g., from
#'   \code{DBI::dbConnect()}).
#' @param query A SQL query string.
#' @param ... Additional arguments passed to \code{DBI::dbGetQuery()}.
#'
#' @return A \code{tidylearn_data} object containing the query results.
#'
#' @examples
#' \donttest{
#' # conn <- DBI::dbConnect(RSQLite::SQLite(), "my_database.sqlite")
#' # data <- tl_read_db(conn, "SELECT * FROM my_table")
#' # DBI::dbDisconnect(conn)
#' }
#'
#' @export
tl_read_db <- function(conn, query, ...) {
  tl_check_packages("DBI")

  if (!inherits(conn, "DBIConnection")) {
    stop("'conn' must be a DBI connection object. ",
         "Create one with DBI::dbConnect().",
         call. = FALSE)
  }

  if (!is.character(query) || length(query) != 1 || !nzchar(query)) {
    stop("'query' must be a non-empty SQL string.", call. = FALSE)
  }

  data <- DBI::dbGetQuery(conn, query, ...)

  if (nrow(data) == 0) {
    warning("Query returned 0 rows.", call. = FALSE)
  }

  source_desc <- paste0(class(conn)[1], ": ", substr(query, 1, 80))
  new_tidylearn_data(data, source = source_desc, format = "database")
}

#' Read from a SQLite database
#'
#' Opens a SQLite database file, executes a SQL query, and returns the result
#' as a \code{tidylearn_data} object. The connection is automatically closed
#' when done. Requires \pkg{DBI} and \pkg{RSQLite}.
#'
#' @param path Path to a SQLite database file (\code{.sqlite} or \code{.db}).
#' @param query A SQL query string.
#' @param ... Additional arguments passed to \code{DBI::dbGetQuery()}.
#'
#' @return A \code{tidylearn_data} object containing the query results.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_sqlite("my_database.sqlite", "SELECT * FROM my_table")
#' }
#'
#' @export
tl_read_sqlite <- function(path, query, ...) {
  tl_validate_file_path(path)
  tl_check_packages("DBI", "RSQLite")

  if (missing(query) || !is.character(query) || !nzchar(query)) {
    stop("'query' is required. Provide a SQL string, e.g., ",
         "'SELECT * FROM my_table'.",
         call. = FALSE)
  }

  conn <- DBI::dbConnect(RSQLite::SQLite(), path)
  on.exit(DBI::dbDisconnect(conn), add = TRUE)

  data <- DBI::dbGetQuery(conn, query, ...)

  if (nrow(data) == 0) {
    warning("Query returned 0 rows.", call. = FALSE)
  }

  new_tidylearn_data(data, source = path, format = "sqlite")
}

#' Read from a PostgreSQL database
#'
#' Connects to a PostgreSQL database, executes a SQL query, and returns the
#' result as a \code{tidylearn_data} object. Accepts either a connection string
#' or individual connection parameters. Requires \pkg{DBI} and \pkg{RPostgres}.
#'
#' @param dsn A PostgreSQL connection string (e.g.,
#'   \code{"postgres://user:pass@host:port/dbname"}), or the database host if
#'   using named parameters.
#' @param query A SQL query string.
#' @param dbname Database name (if not in \code{dsn}).
#' @param user Username (if not in \code{dsn}).
#' @param password Password (if not in \code{dsn}).
#' @param port Port number. Default is 5432.
#' @param ... Additional arguments passed to \code{DBI::dbConnect()}.
#'
#' @return A \code{tidylearn_data} object containing the query results.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_postgres(
#' #   dsn = "localhost",
#' #   query = "SELECT * FROM my_table",
#' #   dbname = "mydb",
#' #   user = "myuser",
#' #   password = "mypass"
#' # )
#' }
#'
#' @export
tl_read_postgres <- function(dsn, query, dbname = NULL, user = NULL,
                             password = NULL, port = 5432, ...) {
  tl_check_packages("DBI", "RPostgres")

  if (missing(query) || !is.character(query) || !nzchar(query)) {
    stop("'query' is required. Provide a SQL string.",
         call. = FALSE)
  }

  conn <- NULL
  on.exit({
    if (!is.null(conn)) DBI::dbDisconnect(conn)
  }, add = TRUE)

  # Parse connection string if provided
  if (grepl("^postgres(ql)?://", dsn)) {
    conn <- tryCatch(
      DBI::dbConnect(
        RPostgres::Postgres(), dsn = dsn, ...
      ),
      error = function(e) {
        stop(
          "Failed to connect to PostgreSQL: ",
          e$message,
          "\nCheck your connection string.",
          call. = FALSE
        )
      }
    )
  } else {
    conn <- tryCatch(
      DBI::dbConnect(
        RPostgres::Postgres(),
        host = dsn,
        dbname = dbname,
        user = user,
        password = password,
        port = port,
        ...
      ),
      error = function(e) {
        stop(
          "Failed to connect to PostgreSQL: ",
          e$message,
          "\nCheck your connection parameters.",
          call. = FALSE
        )
      }
    )
  }

  data <- DBI::dbGetQuery(conn, query)

  if (nrow(data) == 0) {
    warning("Query returned 0 rows.", call. = FALSE)
  }

  source_desc <- if (grepl("^postgres", dsn)) {
    dsn
  } else {
    paste0("postgres://", dsn)
  }
  new_tidylearn_data(data, source = source_desc, format = "postgres")
}

#' Read from a MySQL/MariaDB database
#'
#' Connects to a MySQL or MariaDB database, executes a SQL query, and returns
#' the result as a \code{tidylearn_data} object. Accepts either a connection
#' string or individual connection parameters. Requires \pkg{DBI} and
#' \pkg{RMariaDB}.
#'
#' @param dsn A MySQL connection string (e.g.,
#'   \code{"mysql://user:pass@host:port/dbname"}), or the database host if
#'   using named parameters.
#' @param query A SQL query string.
#' @param dbname Database name (if not in \code{dsn}).
#' @param user Username (if not in \code{dsn}).
#' @param password Password (if not in \code{dsn}).
#' @param port Port number. Default is 3306.
#' @param ... Additional arguments passed to \code{DBI::dbConnect()}.
#'
#' @return A \code{tidylearn_data} object containing the query results.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_mysql(
#' #   dsn = "localhost",
#' #   query = "SELECT * FROM my_table",
#' #   dbname = "mydb",
#' #   user = "myuser",
#' #   password = "mypass"
#' # )
#' }
#'
#' @export
tl_read_mysql <- function(dsn, query, dbname = NULL, user = NULL,
                          password = NULL, port = 3306, ...) {
  tl_check_packages("DBI", "RMariaDB")

  if (missing(query) || !is.character(query) || !nzchar(query)) {
    stop("'query' is required. Provide a SQL string.",
         call. = FALSE)
  }

  conn <- NULL
  on.exit({
    if (!is.null(conn)) DBI::dbDisconnect(conn)
  }, add = TRUE)

  # Parse connection string or use named params
  if (grepl("^mysql://", dsn)) {
    parsed <- tl_parse_db_url(dsn)
    conn <- tryCatch(
      DBI::dbConnect(
        RMariaDB::MariaDB(),
        host = parsed$host,
        dbname = parsed$dbname,
        user = parsed$user,
        password = parsed$password,
        port = parsed$port %||% 3306L,
        ...
      ),
      error = function(e) {
        stop(
          "Failed to connect to MySQL: ",
          e$message,
          "\nCheck your connection string.",
          call. = FALSE
        )
      }
    )
  } else {
    conn <- tryCatch(
      DBI::dbConnect(
        RMariaDB::MariaDB(),
        host = dsn,
        dbname = dbname,
        user = user,
        password = password,
        port = port,
        ...
      ),
      error = function(e) {
        stop(
          "Failed to connect to MySQL: ",
          e$message,
          "\nCheck your connection parameters.",
          call. = FALSE
        )
      }
    )
  }

  data <- DBI::dbGetQuery(conn, query)

  if (nrow(data) == 0) {
    warning("Query returned 0 rows.", call. = FALSE)
  }

  source_desc <- if (grepl("^mysql", dsn)) dsn else paste0("mysql://", dsn)
  new_tidylearn_data(data, source = source_desc, format = "mysql")
}

#' Read from Google BigQuery
#'
#' Executes a SQL query against Google BigQuery and returns the result as a
#' \code{tidylearn_data} object. Requires the \pkg{bigrquery} package and
#' valid Google Cloud authentication.
#'
#' @param project Google Cloud project ID.
#' @param query A SQL query string (Standard SQL).
#' @param dataset Optional default dataset for unqualified table names.
#' @param ... Additional arguments passed to
#'   \code{bigrquery::bq_project_query()}.
#'
#' @return A \code{tidylearn_data} object containing the query results.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_bigquery(
#' #   project = "my-project",
#' #   query = "SELECT * FROM `my_dataset.my_table` LIMIT 1000"
#' # )
#' }
#'
#' @export
tl_read_bigquery <- function(project, query, dataset = NULL, ...) {
  tl_check_packages("bigrquery")

  if (missing(query) || !is.character(query) || !nzchar(query)) {
    stop("'query' is required. Provide a SQL string.", call. = FALSE)
  }

  # Handle bigquery:// URI format from dispatcher
  if (grepl("^bigquery://", project)) {
    parts <- strsplit(sub("^bigquery://", "", project), "/")[[1]]
    project <- parts[1]
    if (length(parts) > 1 && is.null(dataset)) {
      dataset <- parts[2]
    }
  }

  tb <- tryCatch(
    bigrquery::bq_project_query(project, query, ...),
    error = function(e) {
      stop("BigQuery query failed: ", e$message,
           "\nCheck your project ID, query, and authentication.",
           call. = FALSE)
    }
  )

  data <- bigrquery::bq_table_download(tb)

  if (nrow(data) == 0) {
    warning("Query returned 0 rows.", call. = FALSE)
  }

  source_desc <- paste0("bigquery://", project)
  if (!is.null(dataset)) {
    source_desc <- paste0(source_desc, "/", dataset)
  }
  new_tidylearn_data(data, source = source_desc, format = "bigquery")
}

# ---- Cloud/API readers ----

#' Read from Amazon S3
#'
#' Downloads a file from an S3 bucket and reads it into a \code{tidylearn_data}
#' object. The file format is auto-detected from the key's extension, or can be
#' specified explicitly. Requires the \pkg{paws.storage} package and valid AWS
#' credentials.
#'
#' @param source An S3 URI (e.g., \code{"s3://bucket/path/to/file.csv"}).
#' @param format Optional format override for the downloaded file. If
#'   \code{NULL}, auto-detected from the S3 key extension.
#' @param region AWS region. If \code{NULL}, uses the default from your AWS
#'   configuration.
#' @param ... Additional arguments passed to the format-specific reader.
#'
#' @return A \code{tidylearn_data} object containing the downloaded data.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_s3("s3://my-bucket/data/sales.csv")
#' # data <- tl_read_s3("s3://my-bucket/data/results.parquet")
#' }
#'
#' @export
tl_read_s3 <- function(source, format = NULL, region = NULL, ...) {
  tl_check_packages("paws.storage")

  # Parse s3:// URI
  s3_path <- sub("^s3://", "", source)
  parts <- strsplit(s3_path, "/", fixed = TRUE)[[1]]

  if (length(parts) < 2) {
    stop("Invalid S3 URI: '", source, "'. ",
         "Expected format: s3://bucket/key",
         call. = FALSE)
  }

  bucket <- parts[1]
  key <- paste(parts[-1], collapse = "/")

  # Detect format from key extension
  if (is.null(format)) {
    format <- tryCatch(
      tl_detect_format(key),
      error = function(e) {
        stop("Cannot detect file format from S3 key '", key, "'. ",
             "Specify the 'format' argument.",
             call. = FALSE)
      }
    )
  }

  # Create S3 client
  config <- list()
  if (!is.null(region)) config$region <- region

  s3 <- tryCatch(
    paws.storage::s3(config = config),
    error = function(e) {
      stop("Failed to create S3 client: ", e$message,
           "\nCheck your AWS credentials and configuration.",
           call. = FALSE)
    }
  )

  # Download to temp file
  ext <- tools::file_ext(key)
  tmp <- tempfile(fileext = paste0(".", ext))
  on.exit(unlink(tmp), add = TRUE)

  resp <- tryCatch(
    s3$get_object(Bucket = bucket, Key = key),
    error = function(e) {
      stop("Failed to download s3://", bucket, "/", key, ": ", e$message,
           call. = FALSE)
    }
  )

  writeBin(resp$Body, tmp)

  # Read the downloaded file using the appropriate reader
  result <- switch(format,
    "csv"     = tl_read_csv(tmp, ...),
    "tsv"     = tl_read_tsv(tmp, ...),
    "excel"   = tl_read_excel(tmp, ...),
    "parquet" = tl_read_parquet(tmp, ...),
    "json"    = tl_read_json(tmp, ...),
    "rds"     = tl_read_rds(tmp),
    "rdata"   = tl_read_rdata(tmp, ...),
    stop("Unsupported format '", format, "' for S3 source.", call. = FALSE)
  )

  # Override source to show S3 URI instead of temp path
  attr(result, "tl_source") <- source
  attr(result, "tl_format") <- paste0("s3+", format)
  result
}

#' Read from GitHub
#'
#' Downloads a raw file from a GitHub repository and reads it into a
#' \code{tidylearn_data} object. Accepts either a full GitHub URL or a
#' \code{owner/repo} shorthand with a file path.
#'
#' @param source A GitHub URL or \code{"owner/repo"} string.
#' @param path Path to the file within the repository (required when
#'   \code{source} is \code{"owner/repo"} format).
#' @param ref Branch, tag, or commit SHA. Default is \code{"main"}.
#' @param ... Additional arguments passed to the format-specific reader.
#'
#' @return A \code{tidylearn_data} object containing the downloaded data.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_github("user/repo", path = "data/file.csv")
#' # data <- tl_read_github(
#' #   "https://github.com/user/repo/blob/main/data/file.csv"
#' # )
#' }
#'
#' @export
tl_read_github <- function(source, path = NULL, ref = "main", ...) {
  # Build raw URL
  if (grepl("^https?://github\\.com", source)) {
    # Convert GitHub blob URL to raw URL
    raw_url <- sub("github\\.com", "raw.githubusercontent.com", source)
    raw_url <- sub("/blob/", "/", raw_url)
  } else if (grepl("^https?://raw\\.githubusercontent\\.com", source)) {
    raw_url <- source
  } else {
    # owner/repo shorthand
    if (is.null(path)) {
      stop("'path' is required when 'source' is in 'owner/repo' format.",
           call. = FALSE)
    }
    raw_url <- paste0(
      "https://raw.githubusercontent.com/",
      source, "/", ref, "/", path
    )
  }

  # Detect format from the URL file extension
  format <- tryCatch(
    tl_detect_format(basename(raw_url)),
    error = function(e) {
      stop("Cannot detect file format from GitHub URL. ",
           "The file must have a recognizable extension (csv, json, etc.).",
           call. = FALSE)
    }
  )

  # Download to temp file
  ext <- tools::file_ext(raw_url)
  tmp <- tempfile(fileext = paste0(".", ext))
  on.exit(unlink(tmp), add = TRUE)

  tryCatch(
    utils::download.file(raw_url, tmp, mode = "wb", quiet = TRUE),
    error = function(e) {
      stop("Failed to download from GitHub: ", e$message,
           "\nURL: ", raw_url,
           call. = FALSE)
    }
  )

  # Read the downloaded file
  result <- switch(format,
    "csv"     = tl_read_csv(tmp, ...),
    "tsv"     = tl_read_tsv(tmp, ...),
    "excel"   = tl_read_excel(tmp, ...),
    "parquet" = tl_read_parquet(tmp, ...),
    "json"    = tl_read_json(tmp, ...),
    "rds"     = tl_read_rds(tmp),
    "rdata"   = tl_read_rdata(tmp, ...),
    stop("Unsupported format '", format, "' for GitHub source.", call. = FALSE)
  )

  # Override source to show GitHub URL instead of temp path
  attr(result, "tl_source") <- source
  attr(result, "tl_format") <- paste0("github+", format)
  result
}

#' Read from Kaggle
#'
#' Downloads a dataset file from Kaggle using the Kaggle CLI and reads it into
#' a \code{tidylearn_data} object. Requires the Kaggle CLI to be installed and
#' configured (\code{pip install kaggle}).
#'
#' @param source A Kaggle dataset slug (e.g., \code{"user/dataset-name"}) or a
#'   Kaggle URL.
#' @param file The specific file to read from the dataset. If \code{NULL} and
#'   the dataset contains exactly one file, it is read automatically.
#' @param dest Directory to download files to. Default is a temporary directory.
#' @param type Either \code{"dataset"} (default) or \code{"competition"}.
#' @param ... Additional arguments passed to the format-specific reader.
#'
#' @return A \code{tidylearn_data} object containing the downloaded data.
#'
#' @examples
#' \donttest{
#' # data <- tl_read_kaggle("zillow/zecon", file = "Zip_time_series.csv")
#' # data <- tl_read_kaggle("titanic", file = "train.csv", type = "competition")
#' }
#'
#' @export
tl_read_kaggle <- function(source, file = NULL, dest = tempdir(),
                           type = "dataset", ...) {
  # Check Kaggle CLI is installed
  tl_check_kaggle_cli()

  # Parse Kaggle URL to slug if needed
  if (grepl("kaggle\\.com", source)) {
    source <- tl_parse_kaggle_url(source)
  }

  # Download the dataset
  if (type == "competition") {
    args <- c("competitions", "download", "-c", source, "-p", dest)
    if (!is.null(file)) args <- c(args, "-f", file)
  } else {
    args <- c("datasets", "download", "-d", source, "-p", dest, "--unzip")
    if (!is.null(file)) args <- c(args, "-f", file)
  }

  result <- tryCatch(
    system2("kaggle", args, stdout = TRUE, stderr = TRUE),
    error = function(e) {
      stop("Kaggle CLI failed: ", e$message, call. = FALSE)
    }
  )

  status <- attr(result, "status")
  if (!is.null(status) && status != 0) {
    stop("Kaggle download failed:\n", paste(result, collapse = "\n"),
         call. = FALSE)
  }

  # Find the downloaded file
  if (!is.null(file)) {
    downloaded <- file.path(dest, file)
  } else {
    # Find files in dest that match common data formats
    data_exts <- c("csv", "tsv", "json", "parquet", "xlsx", "xls")
    pattern <- paste0("\\.(", paste(data_exts, collapse = "|"), ")$")
    candidates <- list.files(dest, pattern = pattern, full.names = TRUE)
    candidates <- candidates[order(file.mtime(candidates), decreasing = TRUE)]

    if (length(candidates) == 0) {
      stop("No data files found in downloaded Kaggle dataset. ",
           "Specify the 'file' argument.",
           call. = FALSE)
    }

    if (length(candidates) > 1) {
      message("Multiple files found. Reading: ", basename(candidates[1]))
    }

    downloaded <- candidates[1]
  }

  if (!file.exists(downloaded)) {
    stop("Downloaded file not found: '", downloaded, "'.",
         call. = FALSE)
  }

  # Detect format and read
  format <- tl_detect_format(downloaded)
  result <- switch(format,
    "csv"     = tl_read_csv(downloaded, ...),
    "tsv"     = tl_read_tsv(downloaded, ...),
    "excel"   = tl_read_excel(downloaded, ...),
    "parquet" = tl_read_parquet(downloaded, ...),
    "json"    = tl_read_json(downloaded, ...),
    stop("Unsupported format '", format, "' in Kaggle download.", call. = FALSE)
  )

  # Override source metadata
  attr(result, "tl_source") <- paste0("kaggle://", source)
  attr(result, "tl_format") <- paste0("kaggle+", format)
  result
}

# ---- Internal helpers ----

#' Check that the Kaggle CLI is installed
#' @keywords internal
#' @noRd
tl_check_kaggle_cli <- function() {
  result <- tryCatch(
    system2("kaggle", "--version", stdout = TRUE, stderr = TRUE),
    error = function(e) NULL,
    warning = function(w) NULL
  )

  if (is.null(result)) {
    stop("Kaggle CLI not found. Install it with: pip install kaggle\n",
         "Then configure credentials: ",
         "https://github.com/Kaggle/kaggle-cli#api-credentials",
         call. = FALSE)
  }
  invisible(TRUE)
}

#' Parse a Kaggle URL into a dataset slug
#' @keywords internal
#' @noRd
tl_parse_kaggle_url <- function(url) {
  # https://www.kaggle.com/datasets/user/dataset-name
  # https://www.kaggle.com/competitions/competition-name
  if (grepl("/datasets/", url)) {
    parts <- regmatches(url, regexpr("[^/]+/[^/]+$", url))
  } else if (grepl("/competitions/", url)) {
    parts <- regmatches(url, regexpr("[^/]+$", url))
  } else {
    stop("Cannot parse Kaggle URL: '", url, "'.", call. = FALSE)
  }
  parts
}

#' Parse a database connection URL
#' @keywords internal
#' @noRd
tl_parse_db_url <- function(url) {
  # mysql://user:password@host:port/dbname
  # postgres://user:password@host:port/dbname
  pattern <- paste0(
    "^[a-z]+://(?:([^:@]+)(?::([^@]*))?@)?",
    "([^:/]+)(?::([0-9]+))?(?:/(.+))?$"
  )
  m <- regmatches(url, regexec(pattern, url))[[1]]

  if (length(m) == 0) {
    stop("Cannot parse database URL: '", url, "'.", call. = FALSE)
  }

  list(
    user = if (nzchar(m[2])) m[2] else NULL,
    password = if (nzchar(m[3])) m[3] else NULL,
    host = m[4],
    port = if (nzchar(m[5])) as.integer(m[5]) else NULL,
    dbname = if (nzchar(m[6])) m[6] else NULL
  )
}

#' @title Model Pipeline Functions for tidylearn
#' @name tidylearn-pipeline
#' @description Functions for creating end-to-end model pipelines
#' @importFrom stats formula
#' @importFrom dplyr filter select mutate
NULL

#' Create a modeling pipeline
#'
#' @param data A data frame containing the data
#' @param formula A formula specifying the model
#' @param preprocessing A list of preprocessing steps
#' @param models A list of models to train
#' @param evaluation A list of evaluation criteria
#' @param ... Additional arguments
#' @return A \code{tidylearn_pipeline} object (S3 list) with components
#'   \code{$formula}, \code{$data}, \code{$preprocessing},
#'   \code{$models}, \code{$evaluation}, and \code{$results}
#'   (initially \code{NULL}; populated after \code{\link{tl_run_pipeline}}).
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .,
#'   models = list(tree = list(method = "tree")))
#' print(pipe)
#' }
#' @export
tl_pipeline <- function(data, formula,
                        preprocessing = NULL,
                        models = NULL,
                        evaluation = NULL, ...) {
  # Create default preprocessing if not provided
  if (is.null(preprocessing)) {
    preprocessing <- list(
      impute_missing = TRUE,
      standardize = TRUE,
      dummy_encode = TRUE
    )
  }

  # Create default models if not provided
  if (is.null(models)) {
    # Determine if classification or regression
    response_var <- all.vars(formula)[1]
    y <- data[[response_var]]
    is_classification <- is.factor(y) || is.character(y)

    if (is_classification) {
      models <- list(
        logistic = list(method = "logistic"),
        tree = list(method = "tree"),
        forest = list(method = "forest", ntree = 500)
      )
    } else {
      models <- list(
        linear = list(method = "linear"),
        lasso = list(method = "lasso"),
        forest = list(method = "forest", ntree = 500)
      )
    }
  }

  # Create default evaluation if not provided
  if (is.null(evaluation)) {
    # Determine if classification or regression
    response_var <- all.vars(formula)[1]
    y <- data[[response_var]]
    is_classification <- is.factor(y) || is.character(y)

    if (is_classification) {
      evaluation <- list(
        metrics = c("accuracy", "precision", "recall", "f1", "auc"),
        validation = "cv",
        cv_folds = 5,
        best_metric = "f1"
      )
    } else {
      evaluation <- list(
        metrics = c("rmse", "mae", "rsq", "mape"),
        validation = "cv",
        cv_folds = 5,
        best_metric = "rmse"
      )
    }
  }

  # Create pipeline object
  pipeline <- list(
    formula = formula,
    data = data,
    preprocessing = preprocessing,
    models = models,
    evaluation = evaluation,
    results = NULL
  )

  class(pipeline) <- "tidylearn_pipeline"

  pipeline
}

#' Run a tidylearn pipeline
#'
#' @param pipeline A tidylearn pipeline object
#' @param verbose Logical; whether to print progress
#' @return The input \code{tidylearn_pipeline} object with its
#'   \code{$results} component populated. Results include
#'   \code{$processed_data}, \code{$model_results} (a named list of
#'   per-model fits and metrics), \code{$best_model_name},
#'   \code{$best_model} (the winning \code{tidylearn_model}), and
#'   \code{$metric_values}.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .,
#'   models = list(tree = list(method = "tree")),
#'   evaluation = list(metrics = "accuracy", validation = "cv",
#'     cv_folds = 2, best_metric = "accuracy"))
#' pipe <- tl_run_pipeline(pipe, verbose = FALSE)
#' }
#' @export
tl_run_pipeline <- function(pipeline, verbose = TRUE) {
  # Check if pipeline is valid
  if (!inherits(pipeline, "tidylearn_pipeline")) {
    stop("Input must be a tidylearn pipeline object", call. = FALSE)
  }

  # Extract components
  data <- pipeline$data
  formula <- pipeline$formula
  preprocessing <- pipeline$preprocessing
  models <- pipeline$models
  evaluation <- pipeline$evaluation

  # Apply preprocessing
  if (verbose) {
    message("Applying preprocessing steps...")
  }

  processed_data <- data

  if (preprocessing$impute_missing) {
    if (verbose) {
      message("  - Imputing missing values")
    }

    # Simple imputation for numeric and categorical variables
    for (col in names(processed_data)) {
      if (is.numeric(processed_data[[col]])) {
        # Impute with median for numeric
        na_idx <- is.na(processed_data[[col]])
        if (any(na_idx)) {
          med <- median(processed_data[[col]], na.rm = TRUE)
          processed_data[[col]][na_idx] <- med
        }
      } else if (is.factor(processed_data[[col]]) ||
                   is.character(processed_data[[col]])) {
        # Impute with mode for categorical
        na_idx <- is.na(processed_data[[col]])
        if (any(na_idx)) {
          # Calculate mode
          if (is.factor(processed_data[[col]])) {
            tab <- table(processed_data[[col]])
          } else {
            tab <- table(processed_data[[col]], useNA = "no")
          }
          mode_val <- names(tab)[which.max(tab)]
          processed_data[[col]][na_idx] <- mode_val
        }
      }
    }
  }

  if (preprocessing$standardize) {
    if (verbose) {
      message("  - Standardizing numeric features")
    }

    # Identify numeric columns (excluding response)
    response_var <- all.vars(formula)[1]
    numeric_cols <- sapply(processed_data, is.numeric)
    numeric_cols[response_var] <- FALSE  # Don't standardize response

    # Standardize each numeric column (as.vector avoids matrix-column)
    for (col in names(processed_data)[numeric_cols]) {
      processed_data[[col]] <- as.vector(scale(processed_data[[col]]))
    }
  }

  if (preprocessing$dummy_encode) {
    if (verbose) {
      message("  - Creating dummy variables for categorical features")
    }

    # Let model.matrix handle this during model fitting
    # We don't modify the processed_data here
  }

  # Set up validation strategy
  if (evaluation$validation == "cv") {
    cv_folds <- evaluation$cv_folds

    if (verbose) {
      message("Setting up ", cv_folds, "-fold cross-validation")
    }

    # Create cross-validation splits
    cv_splits <- rsample::vfold_cv(processed_data, v = cv_folds)
  } else if (evaluation$validation == "split") {
    train_prop <- evaluation$train_prop
    if (is.null(train_prop)) train_prop <- 0.8

    if (verbose) {
      message("Setting up train/test split (", train_prop * 100, "% / ",
              (1 - train_prop) * 100, "%)")
    }

    # Create a single train/test split
    train_idx <- sample(
      nrow(processed_data),
      round(train_prop * nrow(processed_data))
    )
    train_data <- processed_data[train_idx, ]
    test_data <- processed_data[-train_idx, ]
  }

  # Train and evaluate models
  model_results <- list()

  for (model_name in names(models)) {
    if (verbose) {
      message("Training model: ", model_name)
    }

    # Extract model configuration
    model_config <- models[[model_name]]
    method <- model_config$method

    # Remove method from config to use rest as parameters
    model_params <- model_config[names(model_config) != "method"]

    if (evaluation$validation == "cv") {
      # Cross-validation approach
      cv_results <- list()

      for (i in 1:cv_folds) {
        if (verbose) {
          message("  - Fold ", i, "/", cv_folds)
        }

        # Get training and testing data for this fold
        train_fold <- rsample::analysis(cv_splits$splits[[i]])
        test_fold <- rsample::assessment(cv_splits$splits[[i]])

        # Fit model on training fold
        model_args <- c(
          list(
            data = train_fold,
            formula = formula,
            method = method
          ),
          model_params
        )

        fold_model <- do.call(tl_model, model_args)

        # Evaluate on test fold
        fold_metrics <- tl_evaluate(
          fold_model, test_fold,
          metrics = evaluation$metrics
        )

        # Store fold results
        cv_results[[i]] <- list(
          model = fold_model,
          metrics = fold_metrics
        )
      }

      # Calculate average metrics across folds
      all_metrics <- do.call(rbind, lapply(cv_results, function(x) x$metrics))

      avg_metrics <- all_metrics %>%
        dplyr::group_by(.data$metric) %>%
        dplyr::summarize(
          mean_value = mean(.data$value, na.rm = TRUE),
          sd_value = sd(.data$value, na.rm = TRUE)
        )

      # Train final model on all data
      final_model_args <- c(
        list(
          data = processed_data,
          formula = formula,
          method = method
        ),
        model_params
      )

      final_model <- do.call(tl_model, final_model_args)

      # Store results
      model_results[[model_name]] <- list(
        model = final_model,
        cv_results = cv_results,
        avg_metrics = avg_metrics
      )

      if (verbose) {
        for (i in seq_len(nrow(avg_metrics))) {
          metric <- avg_metrics$metric[i]
          mean_val <- avg_metrics$mean_value[i]
          sd_val <- avg_metrics$sd_value[i]

          message("    ", metric, ": ", round(mean_val, 4),
                  " (+/-", round(sd_val, 4), ")")
        }
      }
    } else if (evaluation$validation == "split") {
      # Train/test split approach
      # Fit model on training data
      model_args <- c(
        list(
          data = train_data,
          formula = formula,
          method = method
        ),
        model_params
      )

      split_model <- do.call(tl_model, model_args)

      # Evaluate on test data
      test_metrics <- tl_evaluate(
        split_model, test_data,
        metrics = evaluation$metrics
      )

      # Store results
      model_results[[model_name]] <- list(
        model = split_model,
        test_metrics = test_metrics
      )

      if (verbose) {
        for (i in seq_len(nrow(test_metrics))) {
          metric <- test_metrics$metric[i]
          value <- test_metrics$value[i]

          message("    ", metric, ": ", round(value, 4))
        }
      }
    }
  }

  # Select the best model
  best_metric <- evaluation$best_metric

  if (verbose) {
    message("Selecting best model based on ", best_metric)
  }

  # Extract metric values for each model
  metric_values <- sapply(
    names(model_results),
    function(model_name) {
      result <- model_results[[model_name]]

      if (evaluation$validation == "cv") {
        metric_row <- result$avg_metrics$metric ==
          best_metric
        if (any(metric_row)) {
          val <- result$avg_metrics$mean_value[metric_row]
          if (is.nan(val) || is.na(val)) NA_real_ else val
        } else {
          NA_real_
        }
      } else {
        metric_row <- result$test_metrics$metric ==
          best_metric
        if (any(metric_row)) {
          val <- result$test_metrics$value[metric_row]
          if (is.nan(val) || is.na(val)) NA_real_ else val
        } else {
          NA_real_
        }
      }
    }
  )

  # Determine if higher or lower is better for this metric
  metrics_higher_better <- c(
    "accuracy", "precision", "recall",
    "f1", "auc", "rsq"
  )
  is_higher_better <- best_metric %in%
    metrics_higher_better

  # Find best model (use na.rm-safe which.max/which.min)
  valid_values <- !is.na(metric_values)
  if (!any(valid_values)) {
    best_idx <- 1L
    warning(
      "Could not determine best model from metric '",
      best_metric, "' -- all values NA. Using first model.",
      call. = FALSE
    )
  } else if (is_higher_better) {
    best_idx <- which.max(metric_values)
  } else {
    best_idx <- which.min(metric_values)
  }

  best_model_name <- names(model_results)[best_idx]
  best_model <- model_results[[best_model_name]]$model

  if (verbose) {
    message("Best model: ", best_model_name, " with ", best_metric, " = ",
            round(metric_values[best_idx], 4))
  }

  # Update pipeline with results
  pipeline$results <- list(
    processed_data = processed_data,
    model_results = model_results,
    best_model_name = best_model_name,
    best_model = best_model,
    metric_values = metric_values
  )

  pipeline
}

#' Get the best model from a pipeline
#'
#' @param pipeline A tidylearn pipeline object with results
#' @return The best \code{tidylearn_model} object from the pipeline,
#'   selected by the metric specified in \code{evaluation$best_metric}.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .,
#'   models = list(tree = list(method = "tree")),
#'   evaluation = list(metrics = "accuracy", validation = "cv",
#'     cv_folds = 2, best_metric = "accuracy"))
#' pipe <- tl_run_pipeline(pipe, verbose = FALSE)
#' best <- tl_get_best_model(pipe)
#' }
#' @export
tl_get_best_model <- function(pipeline) {
  # Check if pipeline has results
  if (is.null(pipeline$results)) {
    stop(
      "Pipeline has not been run yet. Use tl_run_pipeline() first.",
      call. = FALSE
    )
  }

  pipeline$results$best_model
}

#' Compare models from a pipeline
#'
#' @param pipeline A tidylearn pipeline object with results
#' @param metrics Character vector of metrics to compare
#'   (if NULL, uses all available)
#' @return A \code{\link[ggplot2]{ggplot}} object showing a faceted bar
#'   chart comparing metric values across models, with the best model
#'   highlighted.
#' @importFrom ggplot2 ggplot aes geom_col facet_wrap labs theme_minimal
#' @export
tl_compare_pipeline_models <- function(pipeline, metrics = NULL) {
  # Check if pipeline has results
  if (is.null(pipeline$results)) {
    stop(
      "Pipeline has not been run yet. Use tl_run_pipeline() first.",
      call. = FALSE
    )
  }

  # Extract model results
  model_results <- pipeline$results$model_results

  # Create data frame for plotting
  comparison_data <- NULL

  for (model_name in names(model_results)) {
    result <- model_results[[model_name]]

    if (!is.null(pipeline$evaluation) &&
          pipeline$evaluation$validation == "cv") {
      # Get from average metrics
      model_metrics <- result$avg_metrics

      # Filter metrics if specified
      if (!is.null(metrics)) {
        model_metrics <- model_metrics[model_metrics$metric %in% metrics, ]
      }

      # Add model name
      model_metrics$model <- model_name

      # Rename columns
      names(model_metrics)[names(model_metrics) == "mean_value"] <- "value"

      # Add to comparison data
      if (is.null(comparison_data)) {
        comparison_data <- model_metrics
      } else {
        comparison_data <- rbind(comparison_data, model_metrics)
      }
    } else {
      # Get from test metrics
      model_metrics <- result$test_metrics

      # Filter metrics if specified
      if (!is.null(metrics)) {
        model_metrics <- model_metrics[model_metrics$metric %in% metrics, ]
      }

      # Add model name
      model_metrics$model <- model_name

      # Add to comparison data
      if (is.null(comparison_data)) {
        comparison_data <- model_metrics
      } else {
        comparison_data <- rbind(comparison_data, model_metrics)
      }
    }
  }

  # Add highlight for best model
  comparison_data$is_best <-
    comparison_data$model ==
    pipeline$results$best_model_name

  # Determine which metrics are "higher is better"
  metrics_higher_better <- c(
    "accuracy", "precision", "recall",
    "f1", "auc", "rsq"
  )
  comparison_data$higher_better <-
    comparison_data$metric %in% metrics_higher_better

  # Create comparison plot
  p <- ggplot2::ggplot(
    comparison_data,
    ggplot2::aes(
      x = model,
      y = value,
      fill = is_best
    )
  ) +
    ggplot2::geom_col() +
    ggplot2::facet_wrap(~ metric, scales = "free_y") +
    ggplot2::geom_text(
      ggplot2::aes(label = round(value, 3)),
      vjust = -0.5,
      size = 3
    ) +
    ggplot2::scale_fill_manual(values = c("steelblue", "darkred")) +
    ggplot2::labs(
      title = "Model Comparison",
      subtitle = "Best model highlighted in red",
      x = "Model",
      y = "Metric Value",
      fill = "Best Model"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      axis.text.x = ggplot2::element_text(angle = 45, hjust = 1),
      legend.position = "bottom"
    )

  p
}

#' Make predictions using a pipeline
#'
#' @param pipeline A tidylearn pipeline object with results
#' @param new_data A data frame containing the new data
#' @param type Type of prediction (default: "response")
#' @param model_name Name of model to use (if NULL, uses the best model)
#' @param ... Additional arguments passed to predict
#' @return A \link[tibble]{tibble} with a \code{.pred} column containing
#'   predictions from the selected (or best) pipeline model, after
#'   applying the same preprocessing steps used during training.
#' @export
tl_predict_pipeline <- function(pipeline,
                                new_data,
                                type = "response",
                                model_name = NULL,
                                ...) {
  # Check if pipeline has results
  if (is.null(pipeline$results)) {
    stop(
      "Pipeline has not been run yet. Use tl_run_pipeline() first.",
      call. = FALSE
    )
  }

  # Determine which model to use
  if (is.null(model_name)) {
    # Use best model
    model <- pipeline$results$best_model
  } else {
    # Use specified model
    if (!model_name %in% names(pipeline$results$model_results)) {
      stop("Model not found in pipeline: ", model_name, call. = FALSE)
    }
    model <- pipeline$results$model_results[[model_name]]$model
  }

  # Apply same preprocessing steps to new data
  processed_new_data <- new_data

  if (!is.null(pipeline$preprocessing)) {
    if (pipeline$preprocessing$impute_missing) {
      # Simple imputation for numeric and categorical variables
      for (col in names(processed_new_data)) {
        if (is.numeric(processed_new_data[[col]])) {
          # Impute with median for numeric
          na_idx <- is.na(processed_new_data[[col]])
          if (any(na_idx)) {
            # Use median from original processed data if available
            if (col %in% names(pipeline$results$processed_data)) {
              med <- median(
                pipeline$results$processed_data[[col]],
                na.rm = TRUE
              )
            } else {
              med <- median(processed_new_data[[col]], na.rm = TRUE)
            }
            processed_new_data[[col]][na_idx] <- med
          }
        } else if (is.factor(processed_new_data[[col]]) ||
                     is.character(processed_new_data[[col]])) {
          # Impute with mode for categorical
          na_idx <- is.na(processed_new_data[[col]])
          if (any(na_idx)) {
            # Use mode from original processed data if available
            if (col %in% names(pipeline$results$processed_data)) {
              if (is.factor(pipeline$results$processed_data[[col]])) {
                tab <- table(pipeline$results$processed_data[[col]])
              } else {
                tab <- table(
                  pipeline$results$processed_data[[col]],
                  useNA = "no"
                )
              }
            } else {
              if (is.factor(processed_new_data[[col]])) {
                tab <- table(processed_new_data[[col]])
              } else {
                tab <- table(processed_new_data[[col]], useNA = "no")
              }
            }
            mode_val <- names(tab)[which.max(tab)]
            processed_new_data[[col]][na_idx] <- mode_val
          }
        }
      }
    }

    if (pipeline$preprocessing$standardize) {
      # Identify numeric columns (excluding response)
      response_var <- all.vars(pipeline$formula)[1]
      numeric_cols <- sapply(processed_new_data, is.numeric)
      if (response_var %in% names(numeric_cols)) {
        numeric_cols[response_var] <- FALSE  # Don't standardize response
      }

      # Standardize using means and sds from training data
      for (col in names(processed_new_data)[numeric_cols]) {
        if (col %in% names(pipeline$results$processed_data)) {
          # Use mean and sd from training data
          col_mean <- mean(pipeline$results$processed_data[[col]], na.rm = TRUE)
          col_sd <- sd(pipeline$results$processed_data[[col]], na.rm = TRUE)

          processed_new_data[[col]] <-
            (processed_new_data[[col]] - col_mean) /
            col_sd
        } else {
          # Just standardize with its own mean and sd
          processed_new_data[[col]] <- scale(processed_new_data[[col]])
        }
      }
    }
  }

  # Make predictions
  predict(model, processed_new_data, type = type, ...)
}

#' Save a pipeline to disk
#'
#' @param pipeline A tidylearn pipeline object
#' @param file Path to save the pipeline
#' @return Called for its side effect of saving to disk; returns
#'   \code{NULL} invisibly.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .)
#' tl_save_pipeline(pipe, tempfile(fileext = ".rds"))
#' }
#' @export
tl_save_pipeline <- function(pipeline, file) {
  # Validate input
  if (!inherits(pipeline, "tidylearn_pipeline")) {
    stop("Input must be a tidylearn pipeline object", call. = FALSE)
  }

  # Save as RDS
  saveRDS(pipeline, file = file)

  invisible(NULL)
}

#' Load a pipeline from disk
#'
#' @param file Path to the pipeline file
#' @return A \code{tidylearn_pipeline} object previously saved with
#'   \code{\link{tl_save_pipeline}}.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .)
#' f <- tempfile(fileext = ".rds")
#' tl_save_pipeline(pipe, f)
#' pipe2 <- tl_load_pipeline(f)
#' }
#' @export
tl_load_pipeline <- function(file) {
  # Load RDS
  pipeline <- readRDS(file)

  # Validate
  if (!inherits(pipeline, "tidylearn_pipeline")) {
    stop("Loaded object is not a tidylearn pipeline", call. = FALSE)
  }

  pipeline
}

#' Print a tidylearn pipeline
#'
#' @param x A tidylearn pipeline object
#' @param ... Additional arguments (not used)
#' @return The input pipeline object \code{x}, returned invisibly.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .)
#' print(pipe)
#' }
#' @export
print.tidylearn_pipeline <- function(x, ...) {
  # Extract pipeline components
  formula <- x$formula
  data_dims <- dim(x$data)
  preprocessing <- names(x$preprocessing)[sapply(x$preprocessing, isTRUE)]
  models <- names(x$models)

  # Print pipeline summary
  cat("Tidylearn Pipeline\n")
  cat("=================\n")
  cat("Formula:", deparse(formula), "\n")
  cat("Data:", data_dims[1], "observations,", data_dims[2], "variables\n")
  cat("Preprocessing:", paste(preprocessing, collapse = ", "), "\n")
  cat("Models:", paste(models, collapse = ", "), "\n")

  # Print evaluation if available
  if (!is.null(x$evaluation)) {
    validation <- x$evaluation$validation
    metrics <- paste(x$evaluation$metrics, collapse = ", ")

    cat("Evaluation: ", validation)
    if (validation == "cv") {
      cat(" (", x$evaluation$cv_folds, " folds)", sep = "")
    }
    cat("\n")
    cat("Metrics:", metrics, "\n")
    cat("Best metric:", x$evaluation$best_metric, "\n")
  }

  # Print results if available
  if (!is.null(x$results)) {
    cat("\nResults\n")
    cat("=======\n")
    cat("Best model:", x$results$best_model_name, "\n")

    # Print metric values
    cat("Performance:\n")
    for (i in seq_along(x$results$metric_values)) {
      m_name <- names(x$results$metric_values)[i]
      metric_value <- x$results$metric_values[i]
      is_best <- if (m_name == x$results$best_model_name) " (best)" else ""

      cat(
        "  ", m_name, ": ", x$evaluation$best_metric, " = ",
        round(metric_value, 4), is_best, "\n",
        sep = ""
      )
    }
  }

  invisible(x)
}

#' Summarize a tidylearn pipeline
#'
#' @param object A tidylearn pipeline object
#' @param ... Additional arguments (not used)
#' @return The input pipeline \code{object}, returned invisibly. Called
#'   for its side effect of printing detailed pipeline and model results.
#' @examples
#' \donttest{
#' pipe <- tl_pipeline(iris, Species ~ .)
#' summary(pipe)
#' }
#' @export
summary.tidylearn_pipeline <- function(object, ...) {
  # If no results, just print the pipeline
  if (is.null(object$results)) {
    print(object)
    return(invisible(object))
  }

  # Print pipeline info
  print(object)

  # Print more detailed results
  cat("\nDetailed Results\n")
  cat("===============\n")

  for (model_name in names(object$results$model_results)) {
    cat("\nModel:", model_name, "\n")

    result <- object$results$model_results[[model_name]]

    if (!is.null(object$evaluation) && object$evaluation$validation == "cv") {
      # Print average metrics with standard deviation
      cat("Cross-validation metrics:\n")
      for (i in seq_len(nrow(result$avg_metrics))) {
        metric <- result$avg_metrics$metric[i]
        mean_val <- result$avg_metrics$mean_value[i]
        sd_val <- result$avg_metrics$sd_value[i]

        cat("  ", metric, ": ", round(mean_val, 4),
            " (+/-", round(sd_val, 4), ")", "\n", sep = "")
      }
    } else {
      # Print test metrics
      cat("Test metrics:\n")
      for (i in seq_len(nrow(result$test_metrics))) {
        metric <- result$test_metrics$metric[i]
        value <- result$test_metrics$value[i]

        cat("  ", metric, ": ", round(value, 4), "\n", sep = "")
      }
    }
  }

  # Summary of best model
  cat("\nBest Model Summary\n")
  cat("=================\n")
  print(summary(object$results$best_model))

  invisible(object)
}

# tidylearn cloud compute — threat model

**Status:** DRAFT — describes the commitments tidylearn's cloud-compute
integration will meet when the Modal-backed `compute = "cloud"` path lands.
**Last updated:** 2026-08-06
**Scope:** the cloud-compute paths in tidylearn (`tl_compute_advisor()`'s
cloud-tier estimates, `tl_resolve_compute()`'s `"cloud"` branch, and the
forthcoming `tl_model(..., compute = "cloud")` submission to Modal).

**API names below are provisional.** `confirm_upload`, `dry_run` and
`tl_cloud_consent()` describe the intended design and do not exist yet;
`tl_model()` is the existing entry point they would attach to. The
commitments are binding, the spellings are not.

**Transport.** tidylearn talks to Modal over plain HTTPS from R, using
`httr2` against a Modal Web Function the user has deployed to their own
Modal account. Modal's control plane is gRPC with SDKs for Python, JS and
Go only; there is no R SDK and no general REST API for invoking Functions,
so the Web Function endpoint is the supported language-agnostic surface.
tidylearn does not embed a Python runtime and does not call Modal through
reticulate. The worker container runs R, so fitted models cross the wire as
native R serialisation rather than a per-method export format.

This document is the contract for what cloud-compute in tidylearn will and
will not do. Anyone reviewing the Modal-integration PR should be able to
read this doc and verify, via grep / code review / test runs, that each
mitigation actually holds in the implementation.

The doc is intentionally narrow. It does not cover Modal's own
infrastructure security (see Modal's security docs and SOC 2 report), nor
local OS / R session hardening, nor threats requiring physical access to
the user's machine, nor cryptanalysis of TLS.

---

## 1. Trust boundaries

```text
+-------------------------+                      +-----------------------+
| User's R session        |                      | Modal infrastructure  |
| (trusted)               |                      | (third-party)         |
|                         |                      |                       |
|  +-------------------+  |   HTTPS + proxy      |  +-----------------+  |
|  | tidylearn R code  |--+---token headers----->+->| Web Function    |  |
|  | (trusted)         |  |   to *.modal.run     |  | endpoint        |  |
|  |   via httr2       |<-+---model bytes--------+--+--------+--------+  |
|  +---------+---------+  |                      |           |           |
|            |            |                      |  +--------v--------+  |
|  +---------v---------+  |                      |  | ephemeral R     |  |
|  | local data        |  |                      |  | worker container|  |
|  | (sensitive,       |  |                      |  | (per job)       |  |
|  |  user-owned)      |  |                      |  +-----------------+  |
|  +-------------------+  |                      +-----------------------+
|                         |
|  +-------------------+  |   read at call time, never logged
|  | proxy token       |  |   (env var / user-supplied)
|  | wk-... / ws-...   |  |
|  +-------------------+  |
+-------------------------+

  Deployment (one-time, outside R): user runs `modal deploy` with Modal's
  Python CLI. The account API token stays with the CLI and never enters R.
```

Boundaries:

- **R session ↔ tidylearn code** — trusted in both directions. tidylearn
  runs in the user's R process with the user's privileges.
- **tidylearn code ↔ Modal endpoint** — tidylearn constructs the HTTPS
  request itself with `httr2`. There is no SDK sub-component inside the
  trusted boundary; the transport is tidylearn's own code and is
  therefore fully in scope for this document.
- **tidylearn ↔ Modal infrastructure** — TLS to `*.modal.run`. tidylearn
  does not implement its own cryptography and does not disable
  certificate verification.
- **User's local data → Modal** — sensitive data crossing a third-party
  boundary. This is the most consequential transition and is gated by
  explicit consent (see [Section 4](#4-data-egress-consent-ux)).
- **User's Modal account API token** — never crosses into tidylearn code.
  It is used only by Modal's Python CLI at deploy time, which runs outside
  R (see [T1](#t1-token-handling-in-tidylearn)).
- **User's Modal proxy token** — a distinct, endpoint-scoped credential
  (`wk-…` ID plus `ws-…` secret) that tidylearn *does* handle, because it
  must set the `Modal-Key` / `Modal-Secret` request headers itself. This
  is the significant change from the reticulate-based design and is
  covered by [T1](#t1-token-handling-in-tidylearn).

## 2. Assets

| Asset | Sensitivity | Storage | Lifetime |
| --- | --- | --- | --- |
| Modal account API token | High | `~/.modal.toml` (user OS), used only by the Python CLI at deploy time | Until revoked |
| Modal proxy token (`wk-` / `ws-`) | High — endpoint-scoped | User environment; read into the R process per request, never persisted | Until revoked |
| Training data (rows × cols) | User-determined; may contain PII | In-memory in R, transmitted to Modal during fit | Ephemeral on Modal — destroyed with container |
| Fitted model artifacts | Medium — may memorize training data | Returned to R session as raw bytes; not persisted on Modal. Keras models travel as a separate hdf5 payload, since they cannot cross a process boundary through base R serialisation | Lives in user's R session |
| Hyperparameters / model spec | Low | In-memory in R, transmitted to Modal | Ephemeral |
| Job stdout / logs | Low — should contain no row-level data | Modal log retention (per Modal's policy) | Per Modal |

## 3. Threats and mitigations

### T1: Token handling in tidylearn

**Scenario.** tidylearn causes a Modal credential to leak — by reading the
account token, by printing a token to the console, by including one in an
error message or an `httr2` verbose/debug dump, or by writing one into a
saved object.

**Risk.** High for the account API token: it can spin up arbitrary Modal
jobs at the user's expense and reach anything in their workspace. Lower
but still material for the proxy token, which is scoped to invoking
deployed endpoints — a leak means an attacker can run fits on the user's
account and see their results.

The reticulate design could promise that no token material entered R at
all. The HTTPS design cannot: tidylearn sets the auth headers itself, so
the proxy token is necessarily in the R process. The two credentials are
therefore treated differently.

**Mitigation.**

Account API token — unchanged, still never touched:

- tidylearn R code MUST NOT read `~/.modal.toml`, `~/.config/modal/`,
  `~/.modal/`, or any Modal token environment variable. The account token
  is used only by Modal's Python CLI at deploy time, outside R.
- No `Sys.getenv("MODAL_TOKEN*")` calls in tidylearn source.

Proxy token — held, but contained:

- Sourced from a single documented environment variable pair read at
  request time. tidylearn MUST NOT persist it, cache it in an R object
  that outlives the call, or write it into a fitted model.
- Attached with `httr2::req_headers_redacted()` so the values are masked
  in `req_dry_run()`, `last_request()`, printed request objects, and
  `httr2` verbosity. Plain `req_headers()` MUST NOT be used for the
  `Modal-Key` / `Modal-Secret` headers.
- Never interpolated into `message()`, `warning()`, `stop()`, or
  condition metadata.
- Accepting a token as a function argument is permitted only if it is
  consumed within the call and not stored; the environment variable is
  the documented path.

**Verification.**

- `grep -rn 'modal.toml\|MODAL_TOKEN\|MODAL_AUTH' R/ tests/` returns nothing.
- `grep -rn 'ak-\|wk-\|ws-' R/` returns no token-shaped literals (tests use
  opaque mocks).
- `grep -rn 'req_headers(' R/` shows no Modal auth headers set through the
  non-redacting variant.
- A test asserts that a constructed request's printed form does not
  contain the mock token value.

### T2: Silent data upload without user knowledge

**Scenario.** A user fits a model with `compute = "cloud"` and the
training data is silently uploaded to a third-party service without their
explicit awareness.

**Risk.** High. Even with a valid Modal account, the user may have
data-handling obligations (regulatory, contractual, or organisational)
that forbid silent egress to third parties.

**Mitigation.**

Two-layer consent (see [Section 4](#4-data-egress-consent-ux)):

- **Per-call:** `tl_model(..., compute = "cloud", confirm_upload = TRUE)`
  must be passed for every cloud submission, OR a session-level lock
  must be in place.
- **Session-level:** `tl_cloud_consent()` opts in for the rest of the R
  session and is documented as intended for batch / non-interactive use.
  The session lock is *not* persisted across R restarts.
- Before any data crosses the network, tidylearn prints a one-line
  summary of what is about to be uploaded (rows × cols, estimated MB,
  destination region) so the user can confirm the shape of the egress.
- A dry-run mode (`dry_run = TRUE`) prints the summary and exits without
  uploading.

**Verification.**

- A test that calls `tl_model(..., compute = "cloud")` without consent
  expects an error.
- A test that calls with `confirm_upload = TRUE` and a mocked submission
  layer verifies the upload-summary message fires before the submission
  call.
- `grep -rn 'cloud_submit\|req_perform' R/` flags every call site that
  performs a cloud request, so the consent gate is auditable.

### T3: Data persisted on Modal beyond job lifetime

**Scenario.** Training data lingers on Modal volumes or in Modal's
storage after the fit completes, beyond what the user expects.

**Risk.** Medium. Sensitive data accumulates in a third-party system
without the user actively choosing to store it there.

**Mitigation.**

- The Modal app definition tidylearn ships MUST NOT mount any Modal
  Volume by default. Job containers are ephemeral and destroyed after
  the fit returns.
- Persistent volumes are an explicit opt-in (a separate API, e.g.
  `tl_cloud_cache(...)`, not yet designed) for users who want to cache
  large datasets across fits. The opt-in path is out of scope for the
  first cloud-integration release.
- Returned model artifacts are streamed back to the R session and not
  persisted on Modal.

**Verification.**

- The Modal app definition file in tidylearn is reviewed for any
  `modal.Volume`, `modal.NetworkFileSystem`, or persistent-storage
  references.
- A test runs a cloud fit and then queries the Modal app's storage to
  confirm nothing was persisted (this is a manual integration test, not
  part of CRAN's automated suite).

### T4: Network interception in transit

**Scenario.** An attacker on the network path between the user and
Modal intercepts the upload, captures training data, or tampers with the
returned model.

**Risk.** Low (Modal endpoints are HTTPS-only) but worth naming. Higher
than under the reticulate design, because tidylearn now owns the transport
rather than delegating it.

**Mitigation.**

- All cloud transit uses `httr2` over HTTPS to Modal-hosted endpoints.
- tidylearn MUST NOT disable or relax TLS verification. No
  `req_options(ssl_verifypeer = 0)`, no `ssl_verifyhost` downgrade, no
  custom CA bundle.
- Any endpoint URL is rejected unless its scheme is `https` and its host
  is a Modal host (see [T9](#t9-egress-to-a-non-modal-host)).
- tidylearn MUST NOT call out to non-Modal endpoints during cloud
  submission (no telemetry, no analytics, no third-party storage).

**Verification.**

- The earlier form of this check — "no `httr*` in cloud paths" — no longer
  applies; `httr2` *is* the transport. The check is now about destination
  and TLS posture, not about the presence of an HTTP client.
- `grep -rn 'ssl_verify\|req_options' R/` shows no TLS downgrades.
- `grep -rn 'request(' R/` — every request base URL traces back to the
  validated endpoint from [T9](#t9-egress-to-a-non-modal-host), not to a
  literal or an unchecked user string.
- `grep -rnE 'curl::|download\.file|\burl\(' R/` returns nothing in cloud
  paths; `httr2` is the only client. (The `\b` matters — without it the
  pattern also matches identifiers ending in `url`, such as
  `tl_validate_modal_url(`.)

### T5: Telemetry / phone-home from tidylearn

**Scenario.** tidylearn collects usage data (which methods are run,
dataset sizes, model performance) and transmits it to a third party for
analytics, error reporting, or otherwise.

**Risk.** Medium. Even anonymised telemetry leaks information about the
user's work patterns and may be unacceptable in regulated environments.

**Mitigation.**

- tidylearn ships with NO telemetry. Period. No analytics, no
  crash reporting to a third party, no usage pings, no install-time
  callbacks.
- This is a project-wide policy, not specific to cloud compute, but it
  is restated here because the cloud path is the obvious vector for
  such a regression.

**Verification.**

- `grep -rn 'analytics\|telemetry\|track\|pingback\|sentry\|posthog'
  R/ tests/` returns nothing.
- CRAN review process is part of the defence (CRAN policy forbids
  phoning home without explicit opt-in).

### T6: Sensitive data in logs and messages

**Scenario.** tidylearn or the Modal SDK logs row-level training data,
model coefficients that encode sensitive values, or token-shaped
strings.

**Risk.** Medium. Logs may end up in CI artefacts, terminals, or shared
contexts.

**Mitigation.**

- tidylearn's `message()` / `warning()` / `cat()` output MUST contain
  only metadata: row counts, column counts, method names, timestamps,
  tier labels, cost estimates. Never row values, never coefficient
  values from sensitive fits, never tokens.
- The upload-summary message in [T2](#t2-silent-data-upload-without-user-knowledge)
  is constructed from dimensions, not data.
- Modal's own logs are user-readable and ephemeral; outside tidylearn's
  direct control but documented as such.

**Verification.**

- Code review: any new `cat()` / `message()` / `print()` in cloud paths
  is reviewed for data leakage.
- The existing `tl_compute_advisor()` print method is already
  metadata-only and is the template for cloud-path messages.

### T7: Compromised dependency exfiltrates data

**Scenario.** A dependency in the cloud path is compromised and uses that
path to exfiltrate user data. Under the HTTPS + R-worker design the
dependency surface is `httr2` on the client side, and the R packages
baked into the worker image on the remote side.

**Risk.** Low-to-medium. Hard to fully defend against, but the impact
surface can be limited. Smaller than under the reticulate design, which
put a Python interpreter and the Modal SDK's transitive tree inside the
user's session.

**Mitigation.**

- The worker image pins its **R** dependencies — a fixed `rocker/r-ver`
  base tag plus a package snapshot (a repository snapshot date or a
  committed `renv.lock`). Reproducible image builds, lock file committed
  to the repo.
- Because `method = "deep"` ships in the first cloud release, the worker
  image also carries a Python runtime with TensorFlow and keras, and
  **those are pinned too**. The R `keras` package and the Python keras it
  binds to must be a matched pair; an unpinned image would drift into the
  Keras 2 / Keras 3 split and fail at deserialisation rather than at
  build time. This is the cost of shipping `deep`: the other twelve
  methods need only R and CRAN packages.
- R-side cloud deps live in `Suggests:` so the install footprint is
  minimal. `httr2` is the only addition for the twelve pure-R methods.
  The `deep` path additionally calls `keras`, which tidylearn already
  suggests for local deep fits — it adds no new dependency, but it does
  mean reticulate and a Python runtime are present on the client for
  that one method.
- Direct dep tree is reviewed before each tidylearn release.
- This is an industry-wide problem; tidylearn does not solve it
  unilaterally but commits to good hygiene.

**Verification.**

- The worker image's pinned R dependency manifest is part of the
  cloud-integration PR and is reviewed, as is the pinned Python/TF
  manifest for the `deep` variant.
- The image build is reproducible from the committed manifest.
- `grep -rn 'reticulate::\|library(reticulate)\|require(reticulate)' R/`
  returns nothing — tidylearn never drives reticulate directly, even on
  the `deep` path, where it calls `keras::serialize_model()` and lets
  keras own the Python boundary. (Match on calls, not on the word: the
  source explains in comments why `python.builtin.object` is the right
  class to detect.)
- Cloud paths import only `httr2` and, for `deep`, `keras`.

### T8: Multi-tenant data leakage on Modal

**Scenario.** Another Modal user observes or accesses tidylearn user's
job data due to a Modal-side tenant-isolation failure.

**Risk.** Out of tidylearn's direct control.

**Mitigation.**

- Each tidylearn user deploys the Modal app to their own Modal account.
  tidylearn does NOT operate a shared, multi-tenant Modal service on
  behalf of users.
- Tenant isolation between Modal users is Modal's responsibility (see
  Modal's security docs).
- tidylearn's cloud documentation makes this explicit: cloud compute
  means the user's data on the user's Modal account, billed to the
  user, isolated from other Modal users by Modal's own controls.

**Verification.**

- The tidylearn Modal-deploy helper (`tl_cloud_setup()`) operates
  against the locally-configured Modal account. No tidylearn-operated
  endpoint.

### T9: Egress to a non-Modal host

**Scenario.** Because each user deploys their own Modal app, the endpoint
URL is user-supplied configuration. A typo, a copied-and-edited example, a
poisoned option in a shared `.Rprofile`, or a modified environment variable
sends training data to a host that is not Modal — while every consent gate
in [T2](#t2-silent-data-upload-without-user-knowledge) is satisfied and the
upload summary still reads as expected.

**Risk.** Medium-to-high. This threat did not exist under the reticulate
design, where the SDK resolved the destination from the user's Modal
account and tidylearn never named a host. Owning the transport means owning
the destination.

**Mitigation.**

- The endpoint URL is validated before any request carrying user data.
  It MUST parse, use scheme `https`, and have a host on the allowlist.
  Anything else is a hard error, not a warning.
- The allowlist defaults to Modal's own domains (`*.modal.run`, and
  `modal.com` for API hosts). Modal customers serving Web Functions from
  a custom domain extend it with `tl_cloud_allow_host()`.
- Extension is a **per-session function call**, never an option or an
  environment variable. A shared `.Rprofile` or an inherited environment
  must not be able to add an upload destination without the user having
  written the call. Additions are not persisted and die with the session.
- Added hosts are validated as bare host names. URLs, ports, paths and
  wildcards are refused, as is any single-label name — `"com"` would
  otherwise open an entire TLD and defeat this control completely.
- Host matching is anchored on a leading dot, for default and added
  hosts alike, so a host is matched only by itself and its subdomains.
- The pre-upload summary distinguishes a Modal host from a host added
  this session, so an extended allowlist is visible at the moment
  consent is acted on.
- Validation happens at the single choke point that builds the request,
  so no call site can bypass it.
- The pre-upload summary in [Section 4](#4-data-egress-consent-ux)
  displays the resolved destination host, so a wrong destination is
  visible to the user before they consent.
- Redirects are not followed for requests carrying user data — a 3xx to
  an off-Modal host would otherwise defeat host validation. (Modal's own
  150-second 303 continuation is a different mechanism and is handled by
  the documented poll loop, which re-validates each URL it is given.)

**Verification.**

- A test supplies `http://` , a non-Modal host, and a Modal-lookalike host
  (e.g. `modal.run.evil.test`) and expects an error in each case before any
  request is made.
- Tests confirm that adding a host does not widen anything above it: with
  `fits.example.com` allowed, `example.com`, `evil-fits.example.com` and
  `fits.example.com.evil.test` all remain refused.
- Tests confirm `https` is still required on an added host.
- `grep -rn 'request(' R/` confirms every request is constructed through
  the validating helper.
- Code review confirms redirect-following is disabled on data-carrying
  requests.

### T10: Runaway spend and orphaned jobs

**Scenario.** A user submits a cloud fit and is billed far more than they
expected. The ways this happens are mostly not adversarial:

- The R session goes away — Ctrl-C, a closed IDE, a crash, a closed
  laptop — and the job keeps running on Modal, because the session was
  only polling for a result and was never what kept the work alive.
- The job hangs or diverges and runs to whatever timeout it inherited.
- A retry policy silently multiplies the bill: Modal's timeouts are per
  execution attempt, so `n` retries can cost `n + 1` full timeouts.
- A submission is retried over a flaky network and runs twice.
- The advisor's estimate is optimistic and the real fit costs more.

**Risk.** High, and the consequence is financial rather than
informational. It was previously listed as out of scope on the grounds
that quota exhaustion is a usability concern; that was wrong. A library
that spends a user's money on their behalf owns this.

**Mitigation.**

The controls split into two groups, and the distinction matters more
than any individual control:

*Survives the client dying* — the only real guarantees:

- Every submission sets an explicit timeout, derived from the advisor's
  estimate with headroom and capped well below Modal's 24-hour maximum.
  tidylearn MUST NOT inherit Modal's default timeout, and MUST NOT
  submit a job with no timeout.
- Retries are off by default on the tidylearn worker function. A failed
  fit is the user's decision to repeat, not a silent cost multiplier.
- `tl_cloud_setup()` directs the user to set a spend budget on their
  Modal workspace. That budget is the only true hard cap, and it is not
  tidylearn's to set.

*Best-effort, lost if the session is killed:*

- A fit whose own estimate exceeds the timeout cap is refused before
  submission. Such a job would bill the full timeout and then be killed
  before producing anything.
- The user is shown, and gated on, the **worst-case** cost — timeout
  multiplied by the tier rate — not the expected cost. The estimate is
  order-of-magnitude; the timeout is the actual bound.
- A worst-case cost above `max_cost` is refused. Raising it is explicit.
- The submitted call id is recorded and printed, and in-flight jobs are
  listable, so a job is never invisible.
- Interrupt and error handlers cancel the remote call rather than
  orphaning it, and an explicit cancel is available.
- The submit POST is never automatically retried; only the idempotent
  result poll is.

**Verification.**

- A test asserts that a fit estimated beyond the timeout cap is refused
  rather than submitted.
- A test asserts the gate uses worst case, not the estimate, and that
  the refusal message says so.
- Tests assert the derived timeout has headroom over the estimate, a
  floor, and a cap.
- `grep -rn 'req_retry' R/` shows retry configured only on polling
  requests, never on submission.
- The Modal app definition sets an explicit timeout and `retries = 0`.
- Code review confirms in-flight jobs are registered before the request
  is performed, not after, so a response that never arrives still leaves
  a record.

## 4. Data egress consent UX

The egress contract has two layers:

**Per-call argument.** `tl_model(..., compute = "cloud", confirm_upload =
TRUE)`. Required on each call unless a session lock is in place. Default
is `confirm_upload = FALSE`, so naive calls error rather than silently
upload.

**Session lock.** `tl_cloud_consent()` opts in for the rest of the R
session. Intended for batch / non-interactive use where per-call
arguments would be repetitive. Not persisted across R restarts. Can be
revoked mid-session with `tl_cloud_consent(FALSE)`.

**Pre-upload summary.** Before any data crosses the network, tidylearn
prints:

```text
Uploading to Modal:
  Method:           xgboost
  Destination:      my-workspace--tidylearn-fit.modal.run
  Rows x cols:      1,500,000 x 47
  Estimated MB:     564
  Modal tier:       A10G (24 GB VRAM / 24 GB RAM)
  Estimated:        14 min, $0.43
  Timeout:          42 min (job is killed at this point)
  Most it can bill: $0.77
```

The destination line is part of the mitigation for
[T9](#t9-egress-to-a-non-modal-host): the host the data will actually reach
is shown before consent is acted on, and is marked when it is a host the
user added this session rather than one of Modal's own.

The last two lines are the mitigation for
[T10](#t10-runaway-spend-and-orphaned-jobs). The estimate is
order-of-magnitude and could be wrong; the timeout is what actually
bounds the bill, so the figure the user is asked to accept is the worst
case rather than the expectation.

The summary is generated from metadata only — no row values.

**Dry-run.** `tl_model(..., compute = "cloud", dry_run = TRUE)` prints the
summary and returns without submitting.

**Non-interactive contexts.** Rscript, CI, knit, etc. have no terminal
to prompt. In those contexts the only valid path is the session lock
(`tl_cloud_consent()` called earlier in the script) or per-call
`confirm_upload = TRUE`. tidylearn does NOT use `readline()` or any
interactive prompt that would break batch use.

## 5. Audit checklist (Modal-integration PR)

Before the Modal-integration PR is merged, reviewers must confirm:

- [ ] No `modal.toml` / `MODAL_TOKEN` / `MODAL_AUTH*` references in `R/`
      or `tests/`.
- [ ] No `Sys.getenv("MODAL_TOKEN*")` calls in `R/` (the account token is
      never read by R).
- [ ] No token-shaped string literals (`ak-...`, `wk-...`, `ws-...`)
      anywhere in source or tests (use opaque mocks).
- [ ] Proxy-token headers are set with `req_headers_redacted()`, never
      plain `req_headers()`, and a test asserts the token does not appear
      in a printed request.
- [ ] The proxy token is not persisted, cached beyond the call, or stored
      in a fitted model object.
- [ ] Every cloud submission call site is gated on either
      `confirm_upload = TRUE` or a session lock check.
- [ ] The pre-upload summary message is generated from metadata, not
      from row values, and names the resolved destination host.
- [ ] The Modal app definition includes no `modal.Volume`,
      `modal.NetworkFileSystem`, or persistent-storage references.
- [ ] No telemetry / analytics / pingback calls in cloud paths.
- [ ] The worker image's R dependency manifest is pinned and committed,
      and the build is reproducible from it. For the `deep` variant, the
      Python/TensorFlow/keras stack is pinned as a matched pair with the
      R `keras` package.
- [ ] Every request is built through the URL-validating helper; scheme is
      `https` and host is on the allowlist, enforced as an error.
- [ ] The allowlist is extensible only by an explicit per-session call,
      not by an option or environment variable, and added hosts are
      validated as bare host names with no single-label entries.
- [ ] Redirect-following is disabled on requests carrying user data.
- [ ] No TLS downgrades (`ssl_verifypeer`, `ssl_verifyhost`, custom CA).
- [ ] Every submission sets an explicit timeout; none inherits Modal's
      default or omits one.
- [ ] The Modal app definition sets `retries = 0`, so a hung job cannot
      bill several full timeouts.
- [ ] The submit request is never auto-retried; `req_retry()` appears
      only on result polling.
- [ ] The pre-upload summary states the worst-case cost and the timeout,
      and the budget gate refuses on worst case rather than estimate.
- [ ] A fit whose estimate exceeds the timeout cap is refused before
      submission.
- [ ] In-flight jobs are registered before the request is performed and
      are listable with `tl_cloud_jobs()`.
- [ ] `httr2` is the only HTTP client in cloud paths; no `curl::`,
      `download.file()`, or `url()`.

## 6. Out of scope

The following are intentionally not addressed by this threat model:

- Modal's internal infrastructure security (covered by Modal's own
  security docs and SOC 2 report).
- Local OS / file system security on the user's machine.
- R session-level isolation (e.g., another local user reading R objects
  from `/tmp` or memory).
- Threats requiring physical access to the user's machine.
- Cryptanalysis of TLS.
- Modal's own billing accuracy, and the spend budget the user sets on
  their Modal workspace. tidylearn bounds what it submits (see
  [T10](#t10-runaway-spend-and-orphaned-jobs)) but cannot enforce a cap
  on the account itself.

## 7. Open questions

These are commitments deferred to the Modal-integration PR:

- Exact API shape of `tl_cloud_consent()` (singleton vs. per-method
  granularity).
- How `dry_run = TRUE` interacts with the advisor — should `dry_run`
  also short-circuit the network-time portion of the cloud estimate?
- Behaviour when a user revokes consent mid-session while a fit is
  in-flight (likely: complete the in-flight fit, refuse new ones).
- Poll-loop behaviour on `404` from the result endpoint: Modal expires
  results after 7 days, which is indistinguishable at the HTTP level from
  a call ID that never existed.
- Logging cadence: at what verbosity level (`verbose = 0/1/2`) does
  the upload-summary print, and at what level (if any) is it
  suppressed?

Two questions listed here previously are now settled and described above:
the endpoint is read from the `TIDYLEARN_MODAL_ENDPOINT` environment
variable rather than an R option, and the host allowlist is extensible
via `tl_cloud_allow_host()` under the constraints in
[T9](#t9-egress-to-a-non-modal-host).

## 8. Revision history

- **2026-08-06** — added [T10](#t10-runaway-spend-and-orphaned-jobs),
  runaway spend and orphaned jobs, and removed the out-of-scope line that
  dismissed quota exhaustion as a usability concern. A submitted Modal
  job runs to completion regardless of what the R session does, so a
  closed session leaves a job billing invisibly; the mandatory
  submission timeout, not any client-side handler, is what bounds that.
  Section 4's summary now shows the timeout and the worst-case cost, and
  the gate refuses on worst case rather than on the estimate.
- **2026-08-06** — revised for the HTTPS + R-worker architecture. Modal has
  no R SDK and no general REST API for invoking Functions, so tidylearn
  calls a user-deployed Web Function over HTTPS with `httr2` instead of
  driving the Python SDK through reticulate, and the worker container runs
  R rather than Python. Consequences: T1 rewritten (tidylearn now handles a
  proxy token, so the "no token material in R" commitment is replaced by
  containment rules); T4 rewritten (the ban on `httr*` in cloud paths
  became a destination-and-TLS check); T7 now pins the worker's R
  dependencies rather than Python ones; T9 added for egress to a non-Modal
  host, a threat introduced by owning the transport. Sections 1, 4 and 5
  updated to match.
- **2026-05-22** — initial draft. Written alongside slice 4 of the
  compute-backends initiative (post local-GPU routing + advisor reframe;
  pre Modal-integration code). All mitigations described as commitments
  to be verified when Modal code lands.

---
title: "Supervised Learning with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Supervised Learning with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## Introduction

This vignette demonstrates supervised learning capabilities in tidylearn. All
methods shown here wrap established R packages - the algorithms are unchanged,
tidylearn simply provides a consistent interface and tidy output.

**Wrapped packages include:**

- stats (`lm()`, `glm()`) for linear and logistic regression
- rpart for decision trees
- randomForest for random forests
- gbm and xgboost for gradient boosting
- glmnet for regularization (ridge, lasso, elastic net)
- e1071 for support vector machines
- nnet for neural networks

Access raw model objects via `model$fit` for package-specific functionality.

## Classification

### Binary Classification

Let's create a binary classification problem from the iris dataset:

```{r}
# Create binary classification dataset
iris_binary <- iris %>%
  filter(Species %in% c("setosa", "versicolor")) %>%
  mutate(Species = droplevels(Species))

# Split data
split <- tl_split(iris_binary, prop = 0.7, stratify = "Species", seed = 123)
```

#### Logistic Regression

```{r}
# Train logistic regression
model_logistic <- tl_model(split$train, Species ~ ., method = "logistic")
print(model_logistic)
```

```{r}
# Predictions
preds_logistic <- predict(model_logistic, new_data = split$test)
head(preds_logistic)
```

#### Decision Trees

```{r}
# Train decision tree
model_tree <- tl_model(split$train, Species ~ ., method = "tree")
print(model_tree)

# Predictions
preds_tree <- predict(model_tree, new_data = split$test)
```

### Multi-class Classification

```{r}
# Split full iris dataset
split_multi <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)
```

#### Random Forest

```{r}
# Train random forest
model_forest <- tl_model(split_multi$train, Species ~ ., method = "forest")
print(model_forest)
```

```{r}
# Predictions
preds_forest <- predict(model_forest, new_data = split_multi$test)
head(preds_forest)
```

```{r}
# Accuracy on test set
mean(preds_forest$.pred == split_multi$test$Species)
```

#### Support Vector Machines

```{r, eval=FALSE}
# Train SVM
model_svm <- tl_model(split_multi$train, Species ~ ., method = "svm")
print(model_svm)

# Predictions
preds_svm <- predict(model_svm, new_data = split_multi$test)
```

## Regression

### Linear Regression

```{r}
# Split mtcars data
split_reg <- tl_split(mtcars, prop = 0.7, seed = 123)

# Train linear model
model_lm <- tl_model(split_reg$train, mpg ~ wt + hp + disp, method = "linear")
print(model_lm)
```

```{r}
# Predictions
preds_lm <- predict(model_lm, new_data = split_reg$test)
head(preds_lm)
```

```{r}
# Calculate RMSE
rmse <- sqrt(mean((preds_lm$.pred - split_reg$test$mpg)^2))
cat("RMSE:", round(rmse, 2), "\n")
```

### Polynomial Regression

```{r}
# Polynomial regression for non-linear relationships
model_poly <- tl_model(
  split_reg$train, mpg ~ wt,
  method = "polynomial", degree = 2
)
print(model_poly)
```

```{r}
# Predictions
preds_poly <- predict(model_poly, new_data = split_reg$test)

# RMSE
rmse_poly <- sqrt(mean((preds_poly$.pred - split_reg$test$mpg)^2))
cat("Polynomial RMSE:", round(rmse_poly, 2), "\n")
```

### Random Forest Regression

```{r}
# Train random forest for regression
model_rf_reg <- tl_model(split_reg$train, mpg ~ ., method = "forest")
print(model_rf_reg)
```

```{r}
# Predictions
preds_rf <- predict(model_rf_reg, new_data = split_reg$test)

# RMSE
rmse_rf <- sqrt(mean((preds_rf$.pred - split_reg$test$mpg)^2))
cat("Random Forest RMSE:", round(rmse_rf, 2), "\n")
```

## Regularized Regression

Regularization helps prevent overfitting by adding penalties to model complexity.

### Ridge Regression

```{r, eval=FALSE}
# Ridge regression (L2 regularization)
model_ridge <- tl_model(split_reg$train, mpg ~ ., method = "ridge")
print(model_ridge)

# Predictions
preds_ridge <- predict(model_ridge, new_data = split_reg$test)
```

### LASSO

```{r, eval=FALSE}
# LASSO (L1 regularization) - performs feature selection
model_lasso <- tl_model(split_reg$train, mpg ~ ., method = "lasso")
print(model_lasso)

# Predictions
preds_lasso <- predict(model_lasso, new_data = split_reg$test)
```

### Elastic Net

```{r, eval=FALSE}
# Elastic Net - combines L1 and L2 regularization
model_enet <- tl_model(
  split_reg$train, mpg ~ .,
  method = "elastic_net", alpha = 0.5
)
print(model_enet)

# Predictions
preds_enet <- predict(model_enet, new_data = split_reg$test)
```

## Model Comparison

```{r}
# Compare multiple models
models <- list(
  linear = tl_model(split_reg$train, mpg ~ ., method = "linear"),
  tree = tl_model(split_reg$train, mpg ~ ., method = "tree"),
  forest = tl_model(split_reg$train, mpg ~ ., method = "forest")
)
```

```{r}
# Calculate RMSE for each model
results <- data.frame(
  Model = character(),
  RMSE = numeric(),
  stringsAsFactors = FALSE
)

for (model_name in names(models)) {
  preds <- predict(models[[model_name]], new_data = split_reg$test)
  rmse <- sqrt(mean((preds$.pred - split_reg$test$mpg)^2))

  results <- rbind(results, data.frame(
    Model = model_name,
    RMSE = rmse
  ))
}

results <- results %>% arrange(RMSE)
print(results)
```

## Advanced Features

### Using Preprocessed Data

```{r}
# Preprocess data
processed <- tl_prepare_data(
  split_reg$train,
  mpg ~ .,
  scale_method = "standardize",
  remove_correlated = TRUE,
  correlation_cutoff = 0.9
)
```

```{r}
# Train on preprocessed data
model_processed <- tl_model(processed$data, mpg ~ ., method = "linear")
print(model_processed)
```

### Formula Variations

```{r}
# Interaction terms
model_interact <- tl_model(split_reg$train, mpg ~ wt * hp, method = "linear")

# Polynomial terms using I()
model_poly_manual <- tl_model(
  split_reg$train, mpg ~ wt + I(wt^2), method = "linear"
)

# Subset of predictors
model_subset <- tl_model(
  split_reg$train, mpg ~ wt + hp + disp, method = "linear"
)
```

## Handling Different Data Types

### Categorical Predictors

```{r}
# Create dataset with categorical variables
mtcars_cat <- mtcars %>%
  mutate(
    cyl = as.factor(cyl),
    gear = as.factor(gear),
    am = as.factor(am)
  )

split_cat <- tl_split(mtcars_cat, prop = 0.7, seed = 123)

# Model with categorical predictors
model_cat <- tl_model(split_cat$train, mpg ~ ., method = "forest")
print(model_cat)
```

### Missing Values

```{r}
# Create data with missing values
mtcars_missing <- mtcars
mtcars_missing[sample(seq_len(nrow(mtcars_missing)), 5), "hp"] <- NA
mtcars_missing[sample(seq_len(nrow(mtcars_missing)), 3), "wt"] <- NA

# Preprocess to handle missing values
processed_missing <- tl_prepare_data(
  mtcars_missing,
  mpg ~ .,
  impute_method = "mean",
  scale_method = "standardize"
)

# Train model
model_imputed <- tl_model(processed_missing$data, mpg ~ ., method = "linear")
```

## Best Practices

1. **Always split your data** before training to properly evaluate performance
2. **Use stratified splitting** for classification to maintain class proportions
3. **Preprocess your data** for better model performance
4. **Compare multiple models** to find the best approach
5. **Consider regularization** when dealing with many predictors
6. **Use appropriate metrics** - accuracy for classification, RMSE/MAE for regression

## Summary

tidylearn provides a unified interface for supervised learning:

- **Classification**: Logistic regression, decision trees, random forests, SVM, etc.
- **Regression**: Linear, polynomial, random forests, regularized methods
- **Preprocessing**: Integrated data preparation tools
- **Consistent API**: Same function (`tl_model()`) for all methods
- **Tidy Output**: Easy-to-use predictions and model objects

```{r}
# Complete workflow example
final_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)
final_prep <- tl_prepare_data(
  final_split$train, Species ~ .,
  scale_method = "standardize"
)
final_model <- tl_model(final_prep$data, Species ~ ., method = "forest")
final_preds <- predict(final_model, new_data = final_split$test)

# Evaluate
accuracy <- mean(final_preds$.pred == final_split$test$Species)
cat("Test Accuracy:", round(accuracy * 100, 1), "%\n")
```

---
title: "Tuning and Pipelines"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tuning and Pipelines}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## Overview

`tl_model()` fits one model with the hyperparameters you name. Two families
build on that:

- **Tuning** searches a space of hyperparameters and returns the model fitted
  with the winner. `tl_tune_grid()` walks every combination; `tl_tune_random()`
  samples from ranges.
- **Pipelines** bundle preprocessing, a set of models, and an evaluation
  scheme into one object you can run, save, reload, and predict from.

The two compose: tune to find the settings, then put the winning settings in a
pipeline so the whole recipe is reproducible.

## Grid Search

`tl_tune_grid()` takes a named list of candidate values, one entry per
hyperparameter, and cross-validates every combination.

```{r}
tuned_tree <- tl_tune_grid(
  iris, Species ~ .,
  method = "tree",
  param_grid = list(cp = c(0.001, 0.01, 0.1), minsplit = c(5, 20)),
  folds = 3,
  verbose = FALSE
)
```

What comes back is an ordinary tidylearn model, already fitted with the
winning settings, so everything you would normally do with a model still
works:

```{r}
print(tuned_tree)
```

The search itself is attached as a `"tuning_results"` attribute:

```{r}
tuning <- attr(tuned_tree, "tuning_results")
names(tuning)
```

```{r}
tuning$results
```

```{r}
# The settings that won, and the score they won with
tuning$best_params
tuning$best_metric
```

The winning values reach the underlying fit, not just the report:

```{r}
tuned_tree$fit$control$cp
```

### Choosing the metric

Without a `metric`, tuning uses accuracy for classification and RMSE for
regression. Name one explicitly and the optimisation direction follows from
the metric: `rmse`, `mse`, `mae` and `mape` are minimised, everything else
maximised. Pass `maximize` only to override that.

```{r}
tuned_reg <- tl_tune_grid(
  mtcars, mpg ~ .,
  method = "forest",
  param_grid = list(mtry = c(2, 4), ntree = c(100, 300)),
  folds = 3,
  metric = "rmse",
  verbose = FALSE
)

attr(tuned_reg, "tuning_results")$best_params
```

### Starting from a default grid

`tl_default_param_grid()` supplies a reasonable starting grid per method, at
three sizes. Use it as a first pass, then narrow around whatever won.

```{r}
tl_default_param_grid("tree", size = "small")
```

```{r}
tl_default_param_grid("forest", size = "medium")
```

```{r}
tuned_default <- tl_tune_grid(
  iris, Species ~ .,
  method = "tree",
  param_grid = tl_default_param_grid("tree", size = "small"),
  folds = 3,
  verbose = FALSE
)

attr(tuned_default, "tuning_results")$best_params
```

## Random Search

Grid search cost is the product of the candidate counts, so it grows quickly.
`tl_tune_random()` samples `n_iter` points instead.

How `param_space` describes each parameter decides how it is sampled:

| Specification | Sampled as |
|---|---|
| Two numbers, `c(lo, hi)` | Uniform on the continuous interval |
| Three or more numbers | Drawn from exactly those values |
| A character vector | Drawn from those levels |
| A function of no arguments | Whatever the function returns |

A two-element vector is always continuous, so `minsplit = c(2, 40)` samples
values like 11.34. For a parameter that has to be a whole number, list the
candidates instead:

```{r}
tuned_random <- tl_tune_random(
  iris, Species ~ .,
  method = "tree",
  param_space = list(
    cp = c(0.0001, 0.2),                 # continuous
    minsplit = c(2, 5, 10, 20, 30, 40)   # drawn from these six
  ),
  n_iter = 8,
  folds = 3,
  seed = 42,
  verbose = FALSE
)

attr(tuned_random, "tuning_results")$best_params
```

Pass `seed` whenever you want the search to be reproducible. Without it, two
runs sample different points and can pick different winners.

## Looking at the Search

`tl_plot_tuning_results()` reads the attribute and draws it four ways.

```{r}
tl_plot_tuning_results(tuned_tree, plot_type = "scatter")
```

`"grid"` draws the two-parameter heat map that grid search is built for:

```{r}
tl_plot_tuning_results(tuned_tree, plot_type = "grid")
```

`"parallel"` puts every parameter on its own axis, which scales past two:

```{r}
tl_plot_tuning_results(tuned_tree, plot_type = "parallel")
```

`"importance"` ranks parameters by how much of the score variation each one
explains — a quick read on which knob is worth refining:

```{r}
tl_plot_tuning_results(tuned_tree, plot_type = "importance")
```

Every one of these is a ggplot2 object, so add to it as usual.

## Pipelines

A pipeline records preprocessing, the models to fit, and how to evaluate
them. Building it does no work; `tl_run_pipeline()` does.

```{r}
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

pipe <- tl_pipeline(
  split$train, Species ~ .,
  preprocessing = list(standardize = TRUE, dummy_encode = FALSE),
  models = list(
    tree = list(method = "tree"),
    forest = list(method = "forest", ntree = 300)
  ),
  evaluation = list(
    validation = "cv",
    cv_folds = 3,
    metrics = c("accuracy", "f1"),
    best_metric = "accuracy"
  )
)

print(pipe)
```

Anything you leave out of `preprocessing` or `evaluation` takes its default,
so a partial list is fine. An unrecognised name is an error rather than a step
that quietly does nothing.

```{r, error = TRUE}
tl_pipeline(split$train, Species ~ .,
            preprocessing = list(scale_method = "standardize"))
```

### Running it

```{r}
run <- tl_run_pipeline(pipe, verbose = FALSE)

names(run$models)
```

```{r}
print(run)
```

`tl_get_best_model()` returns the model that won on `best_metric`:

```{r}
best <- tl_get_best_model(run)
best$spec$method
```

### Predicting through the pipeline

This is the reason to use a pipeline rather than a bare model. Predicting on
raw new data replays the preprocessing the pipeline learned during the run,
applying the *training* centre and scale rather than recomputing them from
the new rows.

```{r}
preds <- tl_predict_pipeline(run, new_data = split$test, model_name = "forest")
head(preds)
```

```{r}
mean(preds$.pred == split$test$Species)
```

Omit `model_name` to predict with the best model.

### Saving and reloading

```{r}
path <- tempfile(fileext = ".rds")
tl_save_pipeline(run, path)

reloaded <- tl_load_pipeline(path)
names(reloaded$models)
```

```{r}
# Predictions survive the round trip, preprocessing included
reloaded_preds <- tl_predict_pipeline(
  reloaded, new_data = split$test, model_name = "forest"
)
identical(reloaded_preds$.pred, preds$.pred)
```

```{r, include = FALSE}
unlink(path)
```

## Tuning into a Pipeline

Tuning tells you the settings; the pipeline holds them alongside the
preprocessing that produced them.

```{r}
tuned <- tl_tune_grid(
  split$train, Species ~ .,
  method = "forest",
  param_grid = list(mtry = c(2, 3), ntree = c(100, 300)),
  folds = 3,
  verbose = FALSE
)

best_params <- attr(tuned, "tuning_results")$best_params
best_params
```

```{r}
final <- tl_pipeline(
  split$train, Species ~ .,
  models = list(
    forest = c(list(method = "forest"), best_params)
  ),
  evaluation = list(cv_folds = 3, metrics = "accuracy",
                    best_metric = "accuracy")
)

final_run <- tl_run_pipeline(final, verbose = FALSE)
final_preds <- tl_predict_pipeline(final_run, new_data = split$test)

mean(final_preds$.pred == split$test$Species)
```

## Cost

Tuning multiplies fits. A grid of *g* combinations at *k* folds is *g × k*
fits, plus one more to build the final model. The tree grid at the top of this
vignette is 6 combinations × 3 folds = 18 fits, 19 with the final model, of
a method that takes milliseconds. The same grid on `method = "xgboost"` with
1000 rounds is the same 19 fits of something much slower.

Two levers, in the order worth pulling:

1. **Fewer folds.** Going from 5 to 3 removes 40% of the work and still gives
   out-of-sample estimates.
2. **Random over grid.** `tl_tune_random(n_iter = 10)` costs a fixed 10 points
   regardless of how many parameters you are searching, where a grid over the
   same parameters costs their product.

`tl_compute_advisor()` estimates the cost of a single fit before you multiply
it by the search — worth a look before starting a long grid.

## Where to Go Next

- **AutoML** (`vignette("automl")`) searches across methods rather than within
  one, and manages its own budget.
- **Diagnostics** (`vignette("diagnostics")`) covers what to check once you
  have a fitted model.

---
title: "Reporting with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reporting with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
# gt is in Suggests and is the subject of this vignette, so skip the
# whole thing rather than fail the build when it is absent
has_gt <- requireNamespace("gt", quietly = TRUE)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  message = FALSE,
  warning = FALSE,
  eval = has_gt
)
```

```{r, echo = FALSE, results = "asis", eval = TRUE}
if (!has_gt) {
  cat(
    "> **Note:** the gt package is not installed, so the examples below",
    "are shown without output.\n"
  )
}
```

## Overview

Two families turn a fitted model into something publishable. `plot()` and
the `tl_plot_*()` functions return ggplot2 objects; `tl_table()` and the
`tl_table_*()` functions return `gt` tables. Both dispatch on model type, so
the same call covers a forest and a lasso fit, and both hand back an object
you can keep editing rather than printed output you cannot.

```{r setup}
library(tidylearn)
library(dplyr)
library(ggplot2)
library(gt)
```

## Plots

`plot()` picks the visualisation from the model type, and `type` narrows it
further where a model supports more than one view.

### Regression

```{r plot-regression}
model_reg <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")

# Actual vs predicted — one call
plot(model_reg, type = "actual_predicted")
```

### Classification

```{r plot-classification}
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)
model_clf <- tl_model(split$train, Species ~ ., method = "forest")

plot(model_clf, type = "confusion")
```

### PCA

```{r plot-pca}
pca <- tidy_pca(USArrests, scale = TRUE)

tidy_pca_screeplot(pca)
tidy_pca_biplot(pca, label_obs = TRUE)
```

### Regularisation

```{r plot-lasso}
model_lasso <- tl_model(mtcars, mpg ~ ., method = "lasso")

tl_plot_regularization_path(model_lasso)
tl_plot_regularization_cv(model_lasso)
```

## Tables

`tl_table()` mirrors the plot interface, dispatching on model type and an
optional `type`:

```{r table-auto, eval = FALSE}
tl_table(model)                       # auto-selects the best table type
tl_table(model, type = "coefficients") # specific type
```

### Evaluation Metrics

```{r table-metrics}
tl_table_metrics(model_reg)
```

### Coefficients

For linear and logistic models, the table includes standard errors, test
statistics, and p-values, with significant terms highlighted:

```{r table-coef}
tl_table_coefficients(model_reg)
```

For regularised models, coefficients are sorted by magnitude and zero
coefficients are greyed out:

```{r table-coef-lasso}
tl_table_coefficients(model_lasso)
```

### Confusion Matrix

A formatted confusion matrix with correct predictions highlighted on the
diagonal:

```{r table-confusion}
tl_table_confusion(model_clf, new_data = split$test)
```

### Feature Importance

A ranked importance table with a colour gradient:

```{r table-importance}
tl_table_importance(model_clf)
```

### PCA Variance Explained

Cumulative variance is coloured green to highlight how many components are
needed:

```{r table-variance}
pca_model <- tl_model(USArrests, method = "pca")
tl_table_variance(pca_model)
```

### PCA Loadings

A diverging red–blue colour scale highlights strong positive and negative
loadings:

```{r table-loadings}
tl_table_loadings(pca_model)
```

### Cluster Summary

Cluster sizes and mean feature values:

```{r table-clusters}
km <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
tl_table_clusters(km)
```

### Model Comparison

Compare multiple models side-by-side:

```{r table-comparison}
m1 <- tl_model(split$train, Species ~ ., method = "svm")
m2 <- tl_model(split$train, Species ~ ., method = "forest")
m3 <- tl_model(split$train, Species ~ ., method = "tree")

tl_table_comparison(
  m1, m2, m3,
  new_data = split$test,
  names = c("SVM", "Random Forest", "Decision Tree")
)
```

## Interactive Reporting with plotly

Every plot function returns a ggplot2 object, so `ggplotly()` takes any of
them without special handling:

```{r plotly, eval = FALSE}
library(plotly)

ggplotly(plot(model_reg, type = "actual_predicted"))
ggplotly(tidy_pca_biplot(pca, label_obs = TRUE))
ggplotly(tl_plot_regularization_path(model_lasso))
```

## Putting It Together

Fit, score, look, drill in — the four calls that make up most reporting
sections:

```{r workflow}
# Fit
model <- tl_model(split$train, Species ~ ., method = "forest")

# Evaluate
tl_table_metrics(model, new_data = split$test)

# Visualise
plot(model, type = "confusion")

# Drill into feature importance
tl_table_importance(model, top_n = 4)
```

Swap `method = "forest"` for `method = "tree"` or `method = "svm"` and the
reporting code above works without modification.

---
title: "Diagnostics"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Diagnostics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

# car, lmtest and moments back individual assumption tests; gridExtra
# arranges the dashboard. All are in Suggests.
has_assumption_pkgs <- requireNamespace("car", quietly = TRUE) &&
  requireNamespace("lmtest", quietly = TRUE)
has_gridextra <- requireNamespace("gridExtra", quietly = TRUE)
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## Overview

A model that fits is not the same as a model you should use. These functions
answer four questions:

- **Do the assumptions hold?** `tl_check_assumptions()`
- **Which observations drove the fit?** `tl_influence_measures()`,
  `tl_detect_outliers()`
- **Is the difference between two models real?** `tl_compare_cv()`,
  `tl_test_model_difference()`
- **Are there interactions I have not modelled?** `tl_test_interactions()`,
  `tl_interaction_effects()`

`tl_explore()` runs an unsupervised sweep over a dataset before you model it
at all.

We use a linear model throughout, because that is where assumption checking
has teeth.

```{r}
model <- tl_model(mtcars, mpg ~ wt + hp + disp, method = "linear")
```

## Checking Assumptions

`tl_check_assumptions()` runs six checks and returns a verdict on each.

```{r, eval = has_assumption_pkgs}
assumptions <- tl_check_assumptions(model, verbose = FALSE)
names(assumptions)
```

```{r, eval = has_assumption_pkgs}
assumptions$overall
```

Each entry carries the test that was run, the verdict, and what to do about
it. Nothing here is a pass/fail gate — the recommendation is a prompt, not an
instruction.

```{r, eval = has_assumption_pkgs}
assumptions$normality
```

```{r, eval = has_assumption_pkgs}
assumptions$multicollinearity
```

A compact table of every check:

```{r, eval = has_assumption_pkgs}
checks <- c("linearity", "independence", "homoscedasticity",
            "normality", "multicollinearity", "outliers")

data.frame(
  assumption = vapply(checks, function(x) assumptions[[x]]$assumption,
                      character(1)),
  holds = vapply(checks, function(x) isTRUE(assumptions[[x]]$check),
                 logical(1)),
  detail = vapply(checks, function(x) assumptions[[x]]$details, character(1)),
  row.names = NULL
)
```

`disp` correlating with both `wt` and `hp` is what drives the VIF here, and it
is the kind of thing that is invisible in a coefficient table.

### The dashboard

`tl_diagnostic_dashboard()` draws the standard panels in one grid.

```{r, eval = has_assumption_pkgs && has_gridextra, fig.height = 7}
tl_diagnostic_dashboard(model)
```

Switch off any section you do not want with `include_influence`,
`include_assumptions` or `include_performance`.

## Influence

`tl_influence_measures()` returns one row per observation with Cook's
distance, leverage, DFFITS, standardised and studentised residuals, DFBETAS
per coefficient, and a flag for each.

```{r}
influence <- tl_influence_measures(model)
dim(influence)
```

```{r}
influence %>%
  filter(is_influential) %>%
  select(observation, cooks_distance, leverage, dffits, std_residual)
```

The flags use conventional cutoffs, which you can override with
`threshold_cook`, `threshold_leverage` and `threshold_dffits`.

The DFBETAS columns say *which coefficient* an observation moved, which is
usually the more useful question:

```{r}
influence %>%
  select(observation, starts_with("dfbetas_")) %>%
  arrange(desc(abs(dfbetas_wt))) %>%
  head(4)
```

### Refit without the influential rows

The point of the exercise is to see whether the conclusion survives.

```{r}
keep <- !influence$is_influential
refit <- tl_model(mtcars[keep, ], mpg ~ wt + hp + disp, method = "linear")

data.frame(
  term = names(coef(model$fit)),
  all_rows = round(unname(coef(model$fit)), 4),
  without_influential = round(unname(coef(refit$fit)), 4)
)
```

```{r}
sum(!keep)
```

If dropping a handful of rows moves a coefficient materially, that
coefficient describes those rows rather than the population you sampled.

## Outliers in the Data

`tl_influence_measures()` is about a fitted model. `tl_detect_outliers()`
works on the data itself, before or independently of any fit.

```{r}
outliers <- tl_detect_outliers(
  mtcars,
  variables = c("mpg", "hp", "wt"),
  method = "iqr",
  plot = FALSE
)

outliers$outlier_counts$total
outliers$outlier_counts$by_variable
```

```{r}
mtcars[outliers$outlier_indices, c("mpg", "hp", "wt")]
```

`method` also takes `"zscore"` and `"mahalanobis"`. The first two treat each
variable separately; Mahalanobis distance accounts for the correlation
between them, so it finds points that are unremarkable on every single axis
and unusual in combination.

```{r}
mahal <- tl_detect_outliers(
  mtcars,
  variables = c("mpg", "hp", "wt"),
  method = "mahalanobis",
  plot = FALSE
)

mahal$outlier_indices
```

Set `plot = TRUE` to get a ggplot2 object back in `$plot`.

## Comparing Models

A difference in a single held-out score is not evidence. `tl_compare_cv()`
scores several fitted models over the same folds.

```{r}
simple <- tl_model(mtcars, mpg ~ wt, method = "linear")
full <- tl_model(mtcars, mpg ~ wt + hp + disp, method = "linear")
tree <- tl_model(mtcars, mpg ~ wt + hp + disp, method = "tree")

cv <- tl_compare_cv(
  mtcars,
  models = list(simple = simple, full = full, tree = tree),
  folds = 5,
  metrics = c("rmse", "rsq")
)

names(cv)
```

```{r}
cv$summary
```

Per-fold scores are kept as well, which is what makes a test possible:

```{r}
head(cv$fold_metrics)
```

### Is the difference real?

`tl_test_model_difference()` compares each model against a baseline using the
per-fold scores.

```{r}
tl_test_model_difference(
  cv,
  baseline_model = "simple",
  metric = "rmse",
  test = "t.test"
)
```

With five folds this has very little power, so treat a non-significant result
as "these folds do not separate the models" rather than as evidence they are
equivalent. `test = "wilcox.test"` drops the normality assumption, which
matters more at small fold counts than the loss of power costs you.

## Interactions

`tl_test_interactions()` fits each candidate interaction and reports whether
it earns its degrees of freedom.

```{r}
interactions <- tl_test_interactions(
  mtcars, mpg ~ wt + hp + disp,
  all_pairs = TRUE
)

interactions
```

`delta_r2` is the more useful column: a p-value tells you the term is
detectable, `delta_r2` tells you whether it is worth carrying.

Restrict the search with `numeric_only`, `categorical_only` or `mixed_only`,
or name a single pair with `var1` and `var2`.

### Reading an interaction

Once a term is in the model, `tl_interaction_effects()` says what it does at
different levels of the moderator.

```{r}
model_int <- tl_model(mtcars, mpg ~ wt * hp, method = "linear")
effects <- tl_interaction_effects(model_int, var = "wt", by_var = "hp")

effects$slopes
```

The slope of `mpg` on `wt` weakens as `hp` rises — extra weight costs less
fuel economy in a high-powered car, which already had little to lose.

`slope_se` describes the straight line fitted to the prediction grid rather
than the sampling uncertainty of the marginal effect — for a linear model the
grid is exactly linear, so it is near zero by construction. Use
`summary(model_int$fit)` for inference on the interaction coefficient.

```{r}
summary(model_int$fit)$coefficients
```

`tl_auto_interactions()` does the search and the refit in one step, returning
a model with the surviving interactions already in the formula:

```{r}
auto <- tl_auto_interactions(mtcars, mpg ~ wt + hp + disp)
auto$spec$formula
```

## Exploring Before Modelling

`tl_explore()` runs PCA, picks a cluster count, clusters, and computes a
distance summary in one call. It is a first look at a dataset, not a
diagnostic of a fit.

```{r}
eda <- tl_explore(iris, response = "Species", max_components = 4, k_range = 2:5)
names(eda)
```

```{r}
eda$optimal_k
```

```{r}
get_pca_variance(eda$pca)
```

```{r, fig.height = 6}
plot(eda)
```

## A Checklist

For a linear model, in order:

1. `tl_check_assumptions()` — six checks, with the reason each one failed.
2. `tl_influence_measures()` — refit without the flagged rows and see whether
   the coefficients hold.
3. `tl_test_interactions()` — the effect you assumed was additive may not be.
4. `tl_compare_cv()` then `tl_test_model_difference()` — before preferring
   one model over another.

For tree-based and other non-parametric methods, steps 1 and 3 do not apply;
step 2 is available through `tl_detect_outliers()` on the data, and step 4
works unchanged.

---
title: "Unsupervised Learning with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Unsupervised Learning with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

```{r setup}
library(tidylearn)
library(dplyr)
library(ggplot2)
```

## Two Ways In

tidylearn offers two entry points to the same algorithms, and which one you
want depends on what you are doing.

`tl_model(method = "kmeans")` puts clustering behind the same signature as
every supervised method, which is what you want when the clustering is a step
inside a larger workflow — see `vignette("integration-workflows")`.

The `tidy_*()` family is the fuller interface, and the subject of this
vignette. Each function returns a list of tibbles rather than a fitted object
you have to take apart, and each has a matching `augment_*()` that glues the
result back onto your data.

```{r}
# Same algorithm, two interfaces
model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
km <- tidy_kmeans(iris[, 1:4], k = 3)

names(km)
```

Both wrap `stats::kmeans()`; the algorithms are unchanged. Reach the raw
object through `model$fit$model` or `km$model` — an unsupervised `$fit`
is the list of tidied components, with the wrapped object among them.

**Wrapped packages:**

- stats (`prcomp()`, `kmeans()`, `hclust()`, `cmdscale()`)
- cluster (`pam()`, `clara()`)
- dbscan for density-based clustering
- MASS (`isoMDS()`, `sammon()`)
- smacof for MDS algorithms
- arules for association rules — see `vignette("market-basket")`

## Scale First

Every distance-based method here answers a question about distance, and
distance is measured in whatever units your columns happen to use. A variable
measured in thousands will dominate one measured in tenths, whatever its
actual relevance.

`standardize_data()` centres and scales:

```{r}
iris_scaled <- standardize_data(iris[, 1:4])

sapply(iris_scaled, function(x) round(c(mean = mean(x), sd = sd(x)), 3))
```

iris happens to have four variables on a similar scale, so the examples below
use the raw values and stay comparable to the species labels. On real data,
scale first.

## Principal Component Analysis

```{r}
pca <- tidy_pca(iris[, 1:4], scale = TRUE)
names(pca)
```

Accessors rather than list-digging:

```{r}
get_pca_variance(pca)
```

```{r}
get_pca_loadings(pca, n_components = 2)
```

Two components carry 96% of the variance. `plot_variance_explained()` marks
where a threshold is crossed:

```{r}
plot_variance_explained(get_pca_variance(pca), threshold = 0.9)
```

`tidy_pca_screeplot()` is the conventional scree plot:

```{r}
tidy_pca_screeplot(pca)
```

### Scores back on the data

`augment_pca()` returns the original data with component scores appended, so
the labels you already have stay attached:

```{r}
scored <- augment_pca(pca, iris, n_components = 2)
head(scored, 3)
```

```{r}
ggplot(scored, aes(x = PC1, y = PC2, color = Species)) +
  geom_point(size = 3, alpha = 0.7) +
  labs(
    title = "PCA of Iris",
    x = paste0("PC1 (",
               round(get_pca_variance(pca)$prop_variance[1] * 100, 1), "%)"),
    y = paste0("PC2 (",
               round(get_pca_variance(pca)$prop_variance[2] * 100, 1), "%)")
  ) +
  theme_minimal()
```

`tidy_pca_biplot()` overlays the loadings on the same scatter, which is how
you read what the components mean:

```{r}
tidy_pca_biplot(pca, color_by = iris$Species)
```

## How Many Clusters?

Guessing *k* and checking the answer against labels you happen to have is not
a method. `optimal_clusters()` runs three criteria at once.

```{r}
opt <- optimal_clusters(iris[, 1:4], max_k = 8)
names(opt)
```

```{r}
opt$wss
```

```{r}
opt$silhouette
```

```{r}
c(silhouette = attr(opt$silhouette, "optimal_k"),
  gap = opt$gap$recommended_k)
```

The three criteria disagree, which is normal. The elbow is a judgement call,
silhouette favours well-separated compact clusters, and the gap statistic
compares against a null of no structure. Silhouette says 2 here because
*versicolor* and *virginica* overlap; the botanical answer is 3.

```{r}
plot_elbow(opt$wss, suggested_k = 3)
```

```{r}
plot_gap_stat(opt$gap)
```

`calc_wss()` gives the within-cluster sums of squares on their own if that is
all you need:

```{r}
calc_wss(iris[, 1:4], max_k = 6)
```

## K-means

```{r}
km <- tidy_kmeans(iris[, 1:4], k = 3)
km$centers
```

```{r}
km$clusters
```

`augment_kmeans()` puts the assignment back on the data:

```{r}
iris_clustered <- augment_kmeans(km, iris)
table(Cluster = iris_clustered$cluster, Species = iris_clustered$Species)
```

```{r}
plot_cluster_sizes(km$clusters$cluster)
```

```{r}
plot_clusters(iris_clustered, cluster_col = "cluster",
              x_col = "Petal.Length", y_col = "Petal.Width")
```

### Was it a good clustering?

Silhouette width scores each observation on how much better it fits its own
cluster than the next-nearest one. Values near 1 are comfortable, near 0
borderline, negative means the point is on the wrong side.

```{r}
dist_mat <- tidy_dist(iris[, 1:4])
sil <- tidy_silhouette(km$clusters$cluster, dist_mat)

sil$avg_width
```

```{r}
sil$cluster_avg
```

```{r}
plot_silhouette(sil)
```

Cluster 1 is clean; the other two are the *versicolor*/*virginica* boundary,
and their scores say so without needing the labels.

`calc_validation_metrics()` collects the summary numbers in one row:

```{r}
calc_validation_metrics(km$clusters$cluster, iris[, 1:4], dist_mat)
```

## PAM and CLARA

PAM picks actual observations as cluster centres, which makes it less
sensitive to outliers than k-means and gives you a representative row rather
than an average.

```{r}
pam_result <- tidy_pam(iris[, 1:4], k = 3)
pam_result$medoids
```

```{r}
pam_result$silhouette_avg
```

```{r}
table(Cluster = augment_pam(pam_result, iris)$cluster, Species = iris$Species)
```

CLARA samples rather than computing the full distance matrix, which is what
makes it usable when *n* is large enough that an *n × n* matrix is not:

```{r}
large_data <- iris[rep(seq_len(nrow(iris)), 10), 1:4]
clara_result <- tidy_clara(large_data, k = 3, samples = 5)

table(clara_result$clusters$cluster)
```

## Hierarchical Clustering

`tidy_hclust()` builds the tree; cutting it is a separate decision.

```{r}
hc <- tidy_hclust(iris[, 1:4], method = "average")
plot_dendrogram(hc, k = 3)
```

`optimal_hclust_k()` scores cut heights the way `optimal_clusters()` scores
*k*:

```{r}
optimal_hclust_k(hc, method = "silhouette", max_k = 8)$optimal_k
```

```{r}
cuts <- tidy_cutree(hc, k = 3)
head(cuts, 3)
```

```{r}
hc_data <- augment_hclust(hc, iris, k = 3)
table(Cluster = hc_data$cluster, Species = hc_data$Species)
```

Linkage matters more than most people expect. `"average"`, `"complete"`,
`"single"` and `"ward.D2"` can produce different trees from the same distances:

```{r}
linkages <- c("single", "average", "complete", "ward.D2")

sapply(linkages, function(m) {
  cl <- tidy_cutree(tidy_hclust(iris[, 1:4], method = m), k = 3)$cluster
  max(table(cl))
})
```

Single linkage chains, so it puts almost everything in one cluster. That is a
property of the linkage, not a finding about irises.

## DBSCAN

DBSCAN finds arbitrarily shaped clusters and labels sparse points as noise.
It needs `eps` (the neighbourhood radius) and `minPts`. Rather than guessing,
`suggest_eps()` reads it off the k-nearest-neighbour distance curve.

```{r}
eps_suggestion <- suggest_eps(iris[, 1:4], minPts = 5)
eps_suggestion$eps
```

```{r}
plot_knn_dist(iris[, 1:4], k = 5)
```

The elbow in that curve is where points stop having close neighbours, which
is the radius you want.

```{r}
db <- tidy_dbscan(iris[, 1:4], eps = eps_suggestion$eps, minPts = 5)

c(clusters = db$n_clusters, noise = db$n_noise)
```

```{r}
db_data <- augment_dbscan(db, iris)
table(Cluster = db_data$cluster, Species = db_data$Species)
```

Cluster 0 is noise, not a cluster.

`explore_dbscan_params()` sweeps the two parameters together, which is more
informative than tuning either alone:

```{r}
explore_dbscan_params(
  iris[, 1:4],
  eps_values = c(0.4, 0.6, 0.8, 1.0),
  minPts_values = c(4, 5, 10)
)
```

Read `prop_noise` alongside `n_clusters`: a setting that finds many clusters
by discarding a third of the data has not found structure.

## Multidimensional Scaling

MDS places observations so that their plotted distances reproduce their
distances in the original space. Unlike PCA it can work from any distance
matrix, including non-Euclidean ones.

```{r}
mds <- tidy_mds(iris[, 1:4], method = "classical", ndim = 2)
head(mds$config, 3)
```

```{r}
plot_mds(mds, color_by = iris$Species, label_points = FALSE)
```

`method` also takes `"metric"` and `"nonmetric"` (both via smacof),
`"sammon"` and `"kruskal"`. The last two minimise a stress function by
dividing through the observed distances, so they need every pairwise distance
to be strictly positive. iris contains one duplicated row, which is enough to
stop them:

```{r, error = TRUE}
tidy_mds(iris[, 1:4], method = "sammon", ndim = 2)
```

Drop the duplicates and they run:

```{r}
distinct_iris <- iris[!duplicated(iris[, 1:4]), 1:4]
sammon <- tidy_mds(distinct_iris, method = "sammon", ndim = 2)
sammon$stress
```

Stress is the number to check before reading anything into a non-metric
layout: below about 0.05 the picture is a faithful rendering of the
distances, and above about 0.2 it is decoration.

## Comparing Clusterings

`compare_clusterings()` scores several partitions of the same data side by
side.

```{r}
comparison <- compare_clusterings(
  list(
    kmeans = km$clusters$cluster,
    pam = pam_result$clusters$cluster,
    hclust = cuts$cluster,
    dbscan = db$clusters$cluster
  ),
  iris[, 1:4],
  dist_mat
)

comparison
```

```{r}
plot_cluster_comparison(
  iris[, 1:4] %>%
    mutate(kmeans = km$clusters$cluster, hclust = cuts$cluster),
  cluster_cols = c("kmeans", "hclust"),
  x_col = "Petal.Length",
  y_col = "Petal.Width"
)
```

The distance metric is a choice too. `compare_distances()` computes several
so you can see whether your conclusion depends on it:

```{r}
names(compare_distances(iris[, 1:4]))
```

```{r}
plot_distance_heatmap(dist_mat)
```

Look for block structure on the diagonal — contiguous runs of small
distances are what a cluster looks like here.

## A Worked Sequence

Putting the pieces in the order they belong:

```{r}
data_matrix <- standardize_data(iris[, 1:4])

# 1. How many clusters does the data support?
choice <- optimal_clusters(data_matrix, max_k = 8)
k <- attr(choice$silhouette, "optimal_k")
k
```

```{r}
# 2. Cluster at that k
final_km <- tidy_kmeans(data_matrix, k = k)

# 3. Score the result before believing it
final_sil <- tidy_silhouette(final_km$clusters$cluster, tidy_dist(data_matrix))
final_sil$avg_width
```

```{r}
# 4. Attach the assignment and look at it
final_data <- augment_kmeans(final_km, iris)
table(Cluster = final_data$cluster, Species = final_data$Species)
```

```{r}
plot_clusters(final_data, cluster_col = "cluster",
              x_col = "Petal.Length", y_col = "Petal.Width")
```

Silhouette chose 2, and the table shows what that means: *setosa* separated,
the other two species merged. That is the honest reading of this data at this
metric — the third cluster exists botanically but is not well separated in
these four measurements.

## Function Reference

| Task | Function |
|---|---|
| Prepare | `standardize_data()`, `tidy_dist()`, `compare_distances()` |
| Choose *k* | `optimal_clusters()`, `calc_wss()`, `tidy_gap_stat()`, `optimal_hclust_k()` |
| Reduce | `tidy_pca()`, `tidy_mds()` |
| Cluster | `tidy_kmeans()`, `tidy_pam()`, `tidy_clara()`, `tidy_hclust()`, `tidy_dbscan()` |
| Attach results | `augment_pca()`, `augment_kmeans()`, `augment_pam()`, `augment_hclust()`, `augment_dbscan()` |
| Validate | `tidy_silhouette()`, `calc_validation_metrics()`, `compare_clusterings()` |
| Tune DBSCAN | `suggest_eps()`, `plot_knn_dist()`, `explore_dbscan_params()` |
| Plot | `plot_elbow()`, `plot_silhouette()`, `plot_gap_stat()`, `plot_clusters()`, `plot_cluster_sizes()`, `plot_dendrogram()`, `plot_mds()`, `plot_distance_heatmap()`, `plot_variance_explained()` |
| Accessors | `get_pca_variance()`, `get_pca_loadings()`, `tidy_cutree()` |

Association rules have their own vignette: `vignette("market-basket")`.

---
title: "Data Ingestion with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Data Ingestion with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  message = FALSE,
  warning = FALSE
)

# Each backend demo is guarded on its reader package, which lives in
# Suggests -- the vignette has to build without them
has_sqlite <- requireNamespace("DBI", quietly = TRUE) &&
  requireNamespace("RSQLite", quietly = TRUE)
```

## Overview

Every machine learning workflow starts with data. tidylearn's `tl_read()`
family provides a single consistent interface for loading data from files,
databases, cloud storage, and APIs into tidy tibbles — ready for
`tl_prepare_data()` and `tl_model()`.

All readers return a `tidylearn_data` object, a tibble subclass that carries
metadata about the source, format, and read timestamp.

```{r setup}
library(tidylearn)
library(dplyr)
```

---

## The `tl_read()` Dispatcher

`tl_read()` auto-detects the data format and dispatches to the appropriate
backend — just like `tl_model()` routes to the right algorithm:

```{r dispatcher-concept, eval = FALSE}
# Format is auto-detected from the file extension
data <- tl_read("sales.csv")
data <- tl_read("results.xlsx", sheet = "Q1")
data <- tl_read("experiment.parquet")
data <- tl_read("config.json")
data <- tl_read("model_data.rds")

# Override format detection when the extension is ambiguous
data <- tl_read("export.txt", format = "tsv")
```

The result always prints with a metadata header:

```{r print-demo}
tmp <- tempfile(fileext = ".csv")
write.csv(mtcars, tmp, row.names = FALSE)

data <- tl_read(tmp, .quiet = TRUE)
data
```

```{r cleanup-1, include = FALSE}
unlink(tmp)
```

---

## File Formats

### CSV and TSV

Uses [readr](https://readr.tidyverse.org/) when available for fast,
column-type-aware parsing. Falls back to base R automatically if readr is not
installed.

```{r csv-demo}
# Create example files
tmp_csv <- tempfile(fileext = ".csv")
tmp_tsv <- tempfile(fileext = ".tsv")
write.csv(iris, tmp_csv, row.names = FALSE)
write.table(iris, tmp_tsv, sep = "\t", row.names = FALSE)

csv_data <- tl_read_csv(tmp_csv)
tsv_data <- tl_read_tsv(tmp_tsv)
nrow(csv_data)
```

```{r cleanup-csv, include = FALSE}
unlink(c(tmp_csv, tmp_tsv))
```

### Excel

Reads `.xls`, `.xlsx`, and `.xlsm` files via
[readxl](https://readxl.tidyverse.org/). Select sheets by name or position:

```{r excel-demo, eval = requireNamespace("readxl", quietly = TRUE)}
library(readxl)

path <- readxl_example("datasets.xlsx")
excel_data <- tl_read_excel(path, sheet = "mtcars")
head(excel_data, 3)
```

### Parquet

Lightweight, columnar storage for large datasets. Uses
[nanoparquet](https://cran.r-project.org/package=nanoparquet) — a fast,
dependency-free reader:

```{r parquet-demo, eval = requireNamespace("nanoparquet", quietly = TRUE)}
library(nanoparquet)

tmp_pq <- tempfile(fileext = ".parquet")
write_parquet(iris, tmp_pq)

pq_data <- tl_read_parquet(tmp_pq)
nrow(pq_data)
```

```{r cleanup-pq, include = FALSE, eval = requireNamespace("nanoparquet", quietly = TRUE)}
unlink(tmp_pq)
```

### JSON

Reads tabular JSON (array of objects) via
[jsonlite](https://cran.r-project.org/package=jsonlite). Nested structures
are automatically flattened:

```{r json-demo, eval = requireNamespace("jsonlite", quietly = TRUE)}
library(jsonlite)

tmp_json <- tempfile(fileext = ".json")
write_json(mtcars[1:5, ], tmp_json)

json_data <- tl_read_json(tmp_json)
json_data
```

```{r cleanup-json, include = FALSE, eval = requireNamespace("jsonlite", quietly = TRUE)}
unlink(tmp_json)
```

### RDS and RData

Native R serialisation formats — no extra packages needed:

```{r rds-demo}
tmp_rds <- tempfile(fileext = ".rds")
saveRDS(iris, tmp_rds)

rds_data <- tl_read_rds(tmp_rds)
nrow(rds_data)
```

```{r rdata-demo}
tmp_rdata <- tempfile(fileext = ".rdata")
my_data <- mtcars
save(my_data, file = tmp_rdata)

# Name is auto-detected when there is a single data frame
rdata_data <- tl_read_rdata(tmp_rdata)
nrow(rdata_data)
```

```{r cleanup-rds, include = FALSE}
unlink(c(tmp_rds, tmp_rdata))
```

---

## Databases

All database readers use [DBI](https://dbi.r-dbi.org/) as the interface
layer. Each reader manages its own connection lifecycle — connect, query,
disconnect — so you only need to provide the path or credentials and a SQL
query.

### SQLite

The simplest database backend — no server required:

```{r sqlite-demo, eval = has_sqlite}
library(DBI)
library(RSQLite)

# Create an example database
tmp_db <- tempfile(fileext = ".sqlite")
conn <- dbConnect(SQLite(), tmp_db)
dbWriteTable(conn, "iris_tbl", iris)
dbDisconnect(conn)

# Read with tl_read_sqlite
db_data <- tl_read_sqlite(
  tmp_db,
  "SELECT * FROM iris_tbl WHERE Species = 'setosa'"
)
nrow(db_data)
```

```{r cleanup-sqlite, include = FALSE, eval = has_sqlite}
unlink(tmp_db)
```

### Using a Live Connection

If you already have a DBI connection, use `tl_read_db()` directly — it will
not close your connection:

```{r db-demo, eval = has_sqlite}
conn <- dbConnect(SQLite(), ":memory:")
dbWriteTable(conn, "mtcars_tbl", mtcars)

sql <- "SELECT mpg, wt, hp FROM mtcars_tbl WHERE mpg > 20"
db_result <- tl_read_db(conn, sql)
db_result

dbDisconnect(conn)
```

### PostgreSQL, MySQL, and BigQuery

These require a running database server or cloud service. The API is the
same — provide connection details and a SQL query:

```{r remote-db, eval = FALSE}
# PostgreSQL
pg_data <- tl_read_postgres(
  dsn = "localhost",
  query = "SELECT * FROM sales WHERE year = 2025",
  dbname = "analytics",
  user = "myuser",
  password = "mypass"
)

# MySQL / MariaDB # nolint: commented_code_linter.
mysql_data <- tl_read_mysql(
  dsn = "mysql://user:pass@host:3306/mydb",
  query = "SELECT * FROM customers LIMIT 1000"
)

# BigQuery
bq_data <- tl_read_bigquery(
  project = "my-gcp-project",
  query = "SELECT * FROM `dataset.table` LIMIT 1000"
)
```

---

## Cloud and API Sources

### Amazon S3

Downloads a file from S3 and auto-detects the format from the key's
extension. Requires valid AWS credentials:

```{r s3, eval = FALSE}
data <- tl_read_s3("s3://my-bucket/data/sales_2025.csv")
data <- tl_read_s3("s3://my-bucket/data/results.parquet", region = "eu-west-1")
```

### GitHub

Downloads raw files directly from a repository. Accepts full URLs or
`owner/repo` shorthand:

```{r github, eval = FALSE}
# Read a CSV from a public GitHub repository
data <- tl_read_github("tidyverse/dplyr",
  path = "data-raw/starwars.csv", ref = "main"
)
```

### Kaggle

Downloads datasets via the
[Kaggle CLI](https://github.com/Kaggle/kaggle-cli). Install with
`pip install kaggle` and configure your API credentials:

```{r kaggle, eval = FALSE}
data <- tl_read_kaggle("zillow/zecon", file = "Zip_time_series.csv")
data <- tl_read_kaggle("titanic", file = "train.csv", type = "competition")
```

---

## Multi-File Reading

Real-world data is often split across multiple files. tidylearn handles
three common patterns.

### Multiple Paths

Pass a character vector to `tl_read()` — each file is read and row-bound,
with a `source_file` column tracking origin:

```{r multi-path}
dir <- tempdir()
write.csv(iris[1:50, ], file.path(dir, "batch1.csv"), row.names = FALSE)
write.csv(iris[51:100, ], file.path(dir, "batch2.csv"), row.names = FALSE)

paths <- file.path(dir, c("batch1.csv", "batch2.csv"))
combined <- tl_read(paths, .quiet = TRUE)
table(combined$source_file)
```

```{r cleanup-multi, include = FALSE}
unlink(file.path(dir, c("batch1.csv", "batch2.csv")))
```

### Directory Scanning

Point `tl_read_dir()` at a directory. Filter by format, regex pattern, or
scan recursively:

```{r dir-demo}
dir <- tempfile(pattern = "tl_vignette_")
dir.create(dir)
write.csv(iris[1:50, ], file.path(dir, "jan.csv"), row.names = FALSE)
write.csv(iris[51:100, ], file.path(dir, "feb.csv"), row.names = FALSE)
write.csv(iris[101:150, ], file.path(dir, "mar.csv"), row.names = FALSE)

# Read all CSVs from the directory
all_data <- tl_read_dir(dir, format = "csv", .quiet = TRUE)
nrow(all_data)
table(all_data$source_file)
```

```{r dir-pattern}
# Filter with a regex pattern
subset <- tl_read_dir(dir, pattern = "^(jan|feb)", .quiet = TRUE)
nrow(subset)
```

```{r cleanup-dir, include = FALSE}
unlink(dir, recursive = TRUE)
```

Passing a directory path directly to `tl_read()` works too:

```{r dir-dispatch, eval = FALSE}
data <- tl_read("data/monthly_exports/")
```

### Zip Archives

`tl_read_zip()` extracts the archive, auto-detects the file format, and
reads the contents. Select a specific file or let it discover data files
automatically:

```{r zip-demo}
# Create an example zip
dir <- tempfile(pattern = "tl_zip_src_")
dir.create(dir)
write.csv(iris, file.path(dir, "iris.csv"), row.names = FALSE)
zip_path <- tempfile(fileext = ".zip")
# -j stores the file without its directory path, so no setwd() is needed
utils::zip(zip_path, file.path(dir, "iris.csv"), flags = "-j9X")

zip_data <- tl_read_zip(zip_path, .quiet = TRUE)
nrow(zip_data)
attr(zip_data, "tl_format")
```

```{r cleanup-zip, include = FALSE}
unlink(c(dir, zip_path), recursive = TRUE)
```

Zip files are also auto-detected by `tl_read()`:

```{r zip-dispatch, eval = FALSE}
data <- tl_read("download.zip")
data <- tl_read("download.zip", file = "train.csv")
```

---

## The `tidylearn_data` Class

Every reader returns a `tidylearn_data` object — a tibble subclass with
three metadata attributes:

| Attribute | Description |
|-----------|-------------|
| `tl_source` | File path, URL, or description of the data source |
| `tl_format` | Detected or specified format (e.g., `"csv"`, `"zip+csv"`) |
| `tl_timestamp` | POSIXct timestamp of when the data was read |

Because `tidylearn_data` inherits from `tbl_df`, all dplyr verbs, ggplot2,
and tidylearn functions work transparently:

```{r class-demo}
tmp <- tempfile(fileext = ".csv")
write.csv(mtcars, tmp, row.names = FALSE)
data <- tl_read(tmp, .quiet = TRUE)

# Check metadata
attr(data, "tl_format")

# Works with dplyr
data %>%
  filter(mpg > 20) %>%
  select(mpg, wt, hp) %>%
  head(3)
```

```{r cleanup-class, include = FALSE}
unlink(tmp)
```

---

## Full Pipeline

Combining `tl_read()` with the rest of tidylearn gives you a complete
workflow from raw data to published results:

```{r pipeline}
# 1. Ingest
tmp <- tempfile(fileext = ".csv")
write.csv(iris, tmp, row.names = FALSE)
data <- tl_read(tmp, .quiet = TRUE)

# CSV files lose factor information, so convert character columns as needed
data <- data %>% mutate(Species = as.factor(Species))

# 2. Split
split <- tl_split(data, prop = 0.7, stratify = "Species", seed = 42)

# 3. Model
model <- tl_model(split$train, Species ~ ., method = "forest")

# 4. Evaluate
eval_result <- tl_evaluate(model, new_data = split$test)
eval_result
```

```{r cleanup-pipeline, include = FALSE}
unlink(tmp)
```

---

## Supported Formats Reference

| Format | Function | Backend | Dependency |
|--------|----------|---------|------------|
| CSV | `tl_read_csv()` | readr / base R | Suggests (readr) |
| TSV | `tl_read_tsv()` | readr / base R | Suggests (readr) |
| Excel | `tl_read_excel()` | readxl | Suggests |
| Parquet | `tl_read_parquet()` | nanoparquet | Suggests |
| JSON | `tl_read_json()` | jsonlite | Suggests |
| RDS | `tl_read_rds()` | base R | None |
| RData | `tl_read_rdata()` | base R | None |
| SQLite | `tl_read_sqlite()` | DBI + RSQLite | Suggests |
| PostgreSQL | `tl_read_postgres()` | DBI + RPostgres | Suggests |
| MySQL | `tl_read_mysql()` | DBI + RMariaDB | Suggests |
| BigQuery | `tl_read_bigquery()` | bigrquery | Suggests |
| S3 | `tl_read_s3()` | paws.storage | Suggests |
| GitHub | `tl_read_github()` | base R | None |
| Kaggle | `tl_read_kaggle()` | Kaggle CLI | None (system) |
| Directory | `tl_read_dir()` | (dispatches) | — |
| Zip | `tl_read_zip()` | base R + (dispatches) | — |
| Multi-path | `tl_read(c(...))` | (dispatches) | — |

---
title: "Integration Workflows: Combining Supervised and Unsupervised Learning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Integration Workflows: Combining Supervised and Unsupervised Learning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

```{r setup}
library(tidylearn)
library(dplyr)
library(ggplot2)
```

## Introduction

Five functions put an unsupervised step in front of a supervised one:
reduce the features, add cluster membership as a feature, propagate labels
you do not have, handle outliers explicitly, or fit one model per cluster.

They coordinate the wrapped packages rather than implementing anything new.
Each step is an ordinary `tl_model()` object, so `$fit` still reaches what
the underlying package returned — `$fit$model` for the unsupervised step.

One rule governs all five, and it is the thing that goes wrong most often:
the unsupervised step is fitted on training data and must then be *applied*
to the test set, never refitted on it. Every example below carries the
transformation across explicitly, because refitting leaks the test set into
the model.

## Dimensionality Reduction as Preprocessing

PCA or MDS collapses correlated predictors into fewer components before the
supervised fit. That costs some information and buys a smaller, less
collinear feature space.

### Basic Usage

```{r}
# Reduce dimensions before classification
reduced <- tl_reduce_dimensions(iris,
                                response = "Species",
                                method = "pca",
                                n_components = 3)

# Inspect reduced data
head(reduced$data)
```

```{r}
# Train classifier on the reduced features. iris has three species, so this
# needs a multiclass-capable method -- logistic regression is binary only.
model_reduced <- tl_model(reduced$data, Species ~ ., method = "forest")
print(model_reduced)
```

```{r}
# Training-set accuracy. Use tl_evaluate() rather than comparing $.pred by
# hand: what $.pred holds depends on the method and prediction type.
tl_evaluate(model_reduced)
```

### Comparison: Original vs Reduced Features

```{r}
# Split data for fair comparison
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

# Model with original features
model_original <- tl_model(split$train, Species ~ ., method = "forest")
eval_original <- tl_evaluate(model_original, new_data = split$test)

# Model with PCA features
reduced_train <- tl_reduce_dimensions(split$train,
                                      response = "Species",
                                      method = "pca",
                                      n_components = 3)

model_pca <- tl_model(reduced_train$data, Species ~ ., method = "forest")

# The test set must be projected through the PCA fitted on the training
# data -- refitting PCA on the test set would leak information
test_transformed <- predict(
  reduced_train$reduction_model,
  new_data = split$test %>% select(-Species)
)
test_transformed$Species <- split$test$Species

eval_pca <- tl_evaluate(model_pca, new_data = test_transformed)

# Compare results
acc <- function(x) round(x$value[x$metric == "accuracy"] * 100, 1)
n_original <- ncol(split$train) - 1
n_reduced <- sum(grepl("^PC", names(reduced_train$data)))
cat("Original features:", n_original, "->", acc(eval_original), "%\n")
cat("PCA features:", n_reduced, "->", acc(eval_pca), "%\n")
cat("Feature reduction:",
    round((1 - n_reduced / n_original) * 100, 1), "%\n")
```

## Cluster-Based Feature Engineering

Add cluster assignments as a feature, so a model that cannot express group
structure directly gets a column that encodes it. The cluster model is kept
on the result as a `"cluster_model"` attribute, which is what you need to
assign test rows to the same clusters.

### Adding Cluster Features

```{r}
# Add cluster features
data_clustered <- tl_add_cluster_features(iris,
                                          response = "Species",
                                          method = "kmeans",
                                          k = 3)

# Check new features
names(data_clustered)
```

```{r}
# Train model with cluster features
model_cluster <- tl_model(data_clustered, Species ~ ., method = "forest")
print(model_cluster)
```

### Performance Comparison

```{r}
# Compare models with and without cluster features
split_comp <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Without cluster features
model_no_cluster <- tl_model(split_comp$train, Species ~ ., method = "forest")
preds_no_cluster <- predict(model_no_cluster, new_data = split_comp$test)
acc_no_cluster <- mean(preds_no_cluster$.pred == split_comp$test$Species)

# With cluster features
train_clustered <- tl_add_cluster_features(split_comp$train,
                                           response = "Species",
                                           method = "kmeans",
                                           k = 3)
model_with_cluster <- tl_model(train_clustered, Species ~ ., method = "forest")

# Need to get cluster model for test data
cluster_model <- attr(train_clustered, "cluster_model")
test_clusters <- predict(cluster_model, new_data = split_comp$test[, -5])
test_clustered <- split_comp$test
test_clustered$cluster_kmeans <- as.factor(test_clusters$cluster)

preds_with_cluster <- predict(model_with_cluster, new_data = test_clustered)
acc_with_cluster <- mean(preds_with_cluster$.pred == split_comp$test$Species)

cat("Without cluster features:", round(acc_no_cluster * 100, 1), "%\n")
cat("With cluster features:", round(acc_with_cluster * 100, 1), "%\n")
```

## Semi-Supervised Learning

When labels are expensive and unlabelled rows are plentiful, cluster the
full dataset and give each cluster the majority label of whatever labelled
rows fall in it. The propagated labels are guesses, and the fit is only as
good as the assumption that clusters line up with classes.

### Training with Limited Labels

```{r}
# Use only 10% of labels
set.seed(123)
labeled_indices <- sample(nrow(iris), size = 15)  # 15 of 150 labelled

# supervised_method defaults to "tree". Named here because a forest is
# the better fit for propagated labels: the propagation step introduces
# noise, and averaging over trees absorbs more of it than one tree does.
model_semi <- tl_semisupervised(iris, Species ~ .,
                                labeled_indices = labeled_indices,
                                cluster_method = "kmeans",
                                supervised_method = "forest")

print(model_semi)
```

```{r}
# Check how labels were propagated
label_mapping <- model_semi$semisupervised_info$label_mapping
print(label_mapping)
```

```{r}
# Evaluate against the true labels
preds_semi <- predict(model_semi, new_data = iris, type = "class")
accuracy_semi <- mean(preds_semi$.pred == iris$Species)

cat("Accuracy with only", length(labeled_indices), "labels:",
    round(accuracy_semi * 100, 1), "%\n")
labeled_pct <- round(
  length(labeled_indices) / nrow(iris) * 100, 1
)
cat("Proportion of data labeled:", labeled_pct, "%\n")
```

### Comparison: Semi-Supervised vs Fully Supervised

```{r}
# Fully supervised with same amount of data
labeled_data <- iris[labeled_indices, ]
model_full <- tl_model(labeled_data, Species ~ ., method = "forest")
preds_full <- predict(model_full, new_data = iris, type = "class")
accuracy_full <- mean(preds_full$.pred == iris$Species)

cat("Fully supervised (15 samples):", round(accuracy_full * 100, 1), "%\n")
cat("Semi-supervised (15 labels + propagation):",
    round(accuracy_semi * 100, 1), "%\n")
```

## Anomaly-Aware Modeling

`tl_anomaly_aware()` runs outlier detection first and then does one of two
things with what it finds: `action = "flag"` adds an indicator column and
keeps the rows, `action = "remove"` drops them.

### Flagging Anomalies

```{r, eval=FALSE}
# Flag anomalies as a feature
model_anomaly_flag <- tl_anomaly_aware(iris, Species ~ .,
                                       response = "Species",
                                       anomaly_method = "dbscan",
                                       action = "flag",
                                       supervised_method = "forest")

# Check anomaly info
cat("Anomalies detected:", model_anomaly_flag$anomaly_info$n_anomalies, "\n")
```

### Removing Anomalies

```{r, eval=FALSE}
# Remove anomalies before training
model_anomaly_remove <- tl_anomaly_aware(iris, Species ~ .,
                                         response = "Species",
                                         anomaly_method = "dbscan",
                                         action = "remove",
                                         supervised_method = "forest")

cat("Anomalies removed:", model_anomaly_remove$anomalies_removed, "\n")
```

## Stratified Models

One model per cluster, for data where the relationship differs between
groups. It reads more easily than a single model carrying many interaction
terms, and it needs enough rows in every cluster for each fit to be
estimable.

### Training Stratified Models

```{r}
# Train separate models for different clusters
stratified_models <- tl_stratified_models(mtcars, mpg ~ .,
                                          cluster_method = "kmeans",
                                          k = 3,
                                          supervised_method = "linear")

# Check structure
names(stratified_models)
length(stratified_models$supervised_models)
```

```{r}
# Predictions using stratified models
preds_stratified <- predict(stratified_models)
head(preds_stratified)
```

```{r}
# Calculate RMSE
rmse_stratified <- sqrt(mean((preds_stratified$.pred - mtcars$mpg)^2))
cat("Stratified Model RMSE:", round(rmse_stratified, 2), "\n")

# Compare with single model
model_single <- tl_model(mtcars, mpg ~ ., method = "linear")
preds_single <- predict(model_single)
rmse_single <- sqrt(mean((preds_single$.pred - mtcars$mpg)^2))
cat("Single Model RMSE:", round(rmse_single, 2), "\n")
```

## Complete Integration Workflow

Combining multiple integration techniques:

```{r}
# Step 1: Split data
workflow_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Step 2: Reduce dimensions
workflow_reduced <- tl_reduce_dimensions(workflow_split$train,
                                         response = "Species",
                                         method = "pca",
                                         n_components = 3)

# Step 3: Add cluster features to reduced data
workflow_clustered <- tl_add_cluster_features(workflow_reduced$data,
                                              response = "Species",
                                              method = "kmeans",
                                              k = 3)

# Step 4: Train final model
workflow_model <- tl_model(workflow_clustered, Species ~ ., method = "forest")

print(workflow_model)
```

```{r}
# Transform test data through same pipeline
# 1. Apply PCA transformation
test_pca <- predict(workflow_reduced$reduction_model,
                    new_data = workflow_split$test[, -5])
test_pca$Species <- workflow_split$test$Species

# 2. Get cluster assignments. The cluster model was fitted on the PC
# columns; predict() matches new_data to those columns by name and
# errors on a mismatch rather than assigning against the wrong ones.
cluster_model_wf <- attr(workflow_clustered, "cluster_model")
test_clusters_wf <- predict(cluster_model_wf, new_data = test_pca)
test_pca$cluster_kmeans <- as.factor(test_clusters_wf$cluster)

# 3. Predict
workflow_preds <- predict(workflow_model, new_data = test_pca)
workflow_accuracy <- mean(workflow_preds$.pred == workflow_split$test$Species)

cat("Complete Workflow Accuracy:", round(workflow_accuracy * 100, 1), "%\n")
```

## Practical Example: Credit Risk Assessment

```{r}
# Simulate credit data
set.seed(42)
n <- 500
credit_data <- data.frame(
  age = rnorm(n, 40, 12),
  income = rnorm(n, 50000, 20000),
  debt_ratio = runif(n, 0, 0.5),
  credit_score = rnorm(n, 700, 100),
  years_employed = rpois(n, 5)
)

# Create target variable (default risk)
credit_data$default <- factor(
  ifelse(
    credit_data$debt_ratio > 0.4 & credit_data$credit_score < 650,
    "Yes", "No"
  )
)

# Split data
credit_split <- tl_split(
  credit_data, prop = 0.7, stratify = "default", seed = 123
)
```

```{r}
# Strategy 1: Add customer segments as features
credit_clustered <- tl_add_cluster_features(credit_split$train,
                                            response = "default",
                                            method = "kmeans",
                                            k = 4)

model_credit <- tl_model(credit_clustered, default ~ ., method = "forest")

# Transform test data
cluster_model_credit <- attr(credit_clustered, "cluster_model")
test_clusters_credit <- predict(cluster_model_credit,
                                new_data = credit_split$test[, -6])
test_credit <- credit_split$test
test_credit$cluster_kmeans <- as.factor(test_clusters_credit$cluster)

preds_credit <- predict(model_credit, new_data = test_credit)
accuracy_credit <- mean(preds_credit$.pred == credit_split$test$default)

cat("Credit Risk Model Accuracy:", round(accuracy_credit * 100, 1), "%\n")
```

## What Each Combination Buys You

- **Dimensionality reduction before fitting** trades some accuracy for a
  smaller feature space. Whether the trade is worth making is empirical --
  run the comparison above on your own data.
- **Cluster features** give a model a handle on group structure it cannot
  otherwise express. On data with no group structure they add noise.
- **Semi-supervised learning** is worth reaching for when labels are
  expensive and unlabelled observations are plentiful.
- **Anomaly-aware modelling** decides what happens to outliers explicitly
  rather than leaving it to the loss function.
- **Stratified models** fit one model per cluster, which reads more easily
  than a single model with many interaction terms, and needs enough
  observations in every cluster to be estimable.

## Function Reference

| Function | Puts this in front of the supervised fit |
|---|---|
| `tl_reduce_dimensions()` | PCA or MDS |
| `tl_add_cluster_features()` | Cluster membership as a column |
| `tl_semisupervised()` | Label propagation from clusters |
| `tl_anomaly_aware()` | Outlier detection, flagged or removed |
| `tl_stratified_models()` | One model per cluster |

## Where to Go Next

- `vignette("unsupervised-learning")` — the clustering and ordination steps
  above, driven directly
- `vignette("tuning-and-pipelines")` — `tl_pipeline()` does the
  train-then-replay bookkeeping for you
- `vignette("automl")` — `tl_auto_ml()` applies the PCA and clustering
  variants automatically

---
title: "Market Basket Analysis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Market Basket Analysis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
# arules supplies the algorithm and the example data; arulesViz backs the
# graph and grouped-matrix plots. Both are in Suggests, so the whole
# vignette is conditional on them.
has_arules <- requireNamespace("arules", quietly = TRUE)
has_arulesviz <- requireNamespace("arulesViz", quietly = TRUE)

knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  message = FALSE,
  warning = FALSE,
  eval = has_arules
)
```

```{r, echo = FALSE, results = "asis", eval = TRUE}
if (!has_arules) {
  cat(
    "> **Note:** the arules package is not installed, so the examples",
    "below are shown without output.\n"
  )
}
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## Overview

Association rule mining finds statements of the form *when a basket contains
A, it tends to also contain B*. It is unsupervised: there is no response
variable, and every item is a candidate on both sides.

`tidy_apriori()` wraps the Apriori algorithm from
[arules](https://cran.r-project.org/package=arules) and returns the rules as
a tibble instead of an S4 object you have to `inspect()` to read. The rest of
the family filters, ranks and applies those rules.

We use `Groceries`, a month of point-of-sale data from a grocery outlet:
9,835 transactions over 169 item categories.

```{r}
data("Groceries", package = "arules")
Groceries
```

## Mining Rules

Three numbers describe every rule, and two of them are thresholds you set up
front:

- **Support** — the fraction of all transactions containing the whole rule.
  A support floor is what makes the search tractable, and what stops you
  acting on a pattern that occurred four times.
- **Confidence** — of the baskets containing the left-hand side, the fraction
  that also contain the right-hand side. This is the conditional probability.
- **Lift** — confidence divided by the right-hand side's own frequency. Lift
  of 1 means the two are independent; above 1 means the left-hand side makes
  the right-hand side more likely than chance.

```{r}
rules <- tidy_apriori(
  Groceries,
  support = 0.001,      # at least ~10 of the 9,835 transactions
  confidence = 0.5,     # right-hand side follows at least half the time
  minlen = 2            # rules with something on both sides
)
```

```{r}
print(rules)
```

The tibble is the part you work with:

```{r}
rules$rules_tbl
```

```{r}
names(rules)
```

`$rules` holds the underlying arules object for anything the tidy layer does
not cover, in the same way `$fit` does for `tl_model()`.

### Choosing the thresholds

Support and confidence trade recall against volume, and the trade is steep.

```{r}
grid <- expand.grid(
  support = c(0.001, 0.005, 0.01),
  confidence = c(0.3, 0.5, 0.7)
)

grid$n_rules <- mapply(function(s, c) {
  tidy_apriori(Groceries, support = s, confidence = c)$n_rules
}, grid$support, grid$confidence)

grid
```

Dropping the support floor by a factor of ten multiplies the rule count by
more than a hundred. A high-confidence rule at very low support is usually a
description of a handful of shoppers, not of the shop.

## Reading the Rules

`inspect_rules()` sorts and takes the head, which is what you want almost
every time:

```{r}
inspect_rules(rules, by = "lift", n = 10)
```

`by` also accepts `"support"`, `"confidence"` and `"count"`. Set
`decreasing = FALSE` to look at the other end.

`summarize_rules()` gives the distribution of each quality measure across the
whole rule set:

```{r}
summary_stats <- summarize_rules(rules)
summary_stats$n_rules
```

```{r}
data.frame(
  measure = c("support", "confidence", "lift"),
  min = c(summary_stats$support$min, summary_stats$confidence$min,
          summary_stats$lift$min),
  median = c(summary_stats$support$median, summary_stats$confidence$median,
             summary_stats$lift$median),
  max = c(summary_stats$support$max, summary_stats$confidence$max,
          summary_stats$lift$max)
)
```

Because the result is a tibble, dplyr works directly:

```{r}
rules$rules_tbl %>%
  filter(lift > 5, count >= 15) %>%
  arrange(desc(confidence)) %>%
  select(lhs, rhs, confidence, lift, count)
```

## Working With One Item

`filter_rules_by_item()` narrows to rules mentioning an item. `where` picks
the side: `"lhs"` for what the item leads to, `"rhs"` for what leads to it,
`"both"` for either.

```{r}
# What predicts a purchase of whole milk?
filter_rules_by_item(rules, "whole milk", where = "rhs") %>%
  arrange(desc(lift)) %>%
  select(lhs, confidence, lift, count) %>%
  head(5)
```

```{r}
# And what does a basket containing yoghurt lead to?
filter_rules_by_item(rules, "yogurt", where = "lhs") %>%
  arrange(desc(lift)) %>%
  select(lhs, rhs, confidence, lift) %>%
  head(5)
```

`find_related_items()` is the shortcut for the common question, with a lift
floor built in so that co-occurrence by sheer popularity is excluded:

```{r}
find_related_items(rules, "yogurt", min_lift = 1.5, top_n = 5) %>%
  select(lhs, rhs, confidence, lift)
```

## Recommending From a Basket

`recommend_products()` takes the items currently in a basket and returns what
the rules suggest adding.

```{r}
recommend_products(
  rules,
  basket = c("flour", "baking powder"),
  top_n = 5
)
```

A rule only fires when the basket covers its **entire** left-hand side, so a
basket of one or two very common items often matches nothing above the
confidence floor:

```{r}
recommend_products(rules, basket = c("whole milk", "butter"))
```

That empty result is the honest answer rather than a failure. `whole milk`
appears in about a quarter of all baskets, so very little follows it with 50%
confidence.

`min_confidence` filters the rules you already mined, so raising the ceiling
means re-mining, not re-filtering. `rules` above was mined at
`confidence = 0.5`, and no amount of filtering will produce a rule that was
never generated:

```{r}
broad <- tidy_apriori(
  Groceries,
  support = 0.001, confidence = 0.15, minlen = 2
)

broad$n_rules
```

```{r}
recommend_products(
  broad,
  basket = c("whole milk", "butter"),
  min_confidence = 0.15,
  top_n = 5
)
```

Confidence around 0.2 is weak on its own — one basket in five. Lift near 3.4
is what makes these worth reading: eggs are three times more likely in a
basket that already holds milk and butter than in a basket picked at random.
For common items, mine wide and rank on lift.

## Visualising

`visualize_rules()` returns a ggplot2 object for the scatter method, so it
composes like any other plot in the package.

```{r}
visualize_rules(rules, method = "scatter", top_n = 200)
```

Support on one axis against confidence on the other, coloured by lift, is the
standard first look. Rules sitting well away from the main cloud are the ones
to read.

```{r, eval = has_arules && has_arulesviz}
visualize_rules(rules, method = "graph", top_n = 20)
```

The graph method needs arulesViz and draws items as nodes with rules as
edges, which is the more useful view once you have narrowed to a handful of
rules worth reading.

## From a Data Frame

`Groceries` is already a `transactions` object. Real data usually arrives as
one row per line item, which `arules` coerces:

```{r}
receipts <- data.frame(
  basket_id = c(1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5),
  item = c("bread", "butter", "jam",
           "bread", "butter",
           "bread", "butter", "jam",
           "milk", "bread",
           "bread", "butter", "jam"),
  stringsAsFactors = TRUE
)

baskets <- split(as.character(receipts$item), receipts$basket_id)
transactions <- as(baskets, "transactions")
transactions
```

```{r}
small_rules <- tidy_apriori(
  transactions,
  support = 0.4, confidence = 0.6, minlen = 2
)

small_rules$rules_tbl %>%
  arrange(desc(lift)) %>%
  select(lhs, rhs, support, confidence, lift)
```

With five baskets these numbers mean nothing — the point is the shape of the
input. Anything `arules` accepts as `transactions`, `tidy_apriori()` accepts.

## What the Numbers Do Not Tell You

Three cautions worth carrying:

1. **Lift is symmetric; the rule is not.** *A ⇒ B* and *B ⇒ A* have the same
   lift and usually different confidence. Which direction you act on is a
   decision about the business, not a result from the data.
2. **A rule is not a cause.** Bread and butter co-occur because people buy
   both, not because one drives the other. Moving butter next to bread is a
   testable hypothesis, and the test is an experiment.
3. **Rules describe the period you mined.** A month of grocery data carries
   that month's seasonality. Re-mine rather than reusing a stored rule set
   across a season boundary.

## Function Reference

| Function | Purpose |
|---|---|
| `tidy_apriori()` | Mine rules, return a tibble |
| `tidy_rules()` | Convert an arules rules object to a tibble |
| `inspect_rules()` | Sort by a quality measure and take the head |
| `summarize_rules()` | Distribution of support, confidence and lift |
| `filter_rules_by_item()` | Rules mentioning an item, by side |
| `find_related_items()` | Items associated with one item, above a lift floor |
| `recommend_products()` | Suggestions for a given basket |
| `visualize_rules()` | Scatter, graph and grouped-matrix plots |

---
title: "Getting Started with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Introduction

Every machine learning package in R has its own API and its own output
format. tidylearn puts one signature over 20 of them: `tl_model()` picks the
underlying package from the `method` you name, and whatever comes back is a
tibble or a ggplot2 object.

The algorithms are untouched. glmnet, randomForest, xgboost, e1071, cluster
and dbscan do the fitting, and `model$fit` hands you the object they
returned — `model$fit$model` for an unsupervised method, whose `$fit` is
the list of tidied components — so nothing here closes off the package
underneath.

This vignette covers the shape of a workflow end to end. The articles listed
at the bottom go deeper on each step.

## Installation

```{r, eval = FALSE}
# From CRAN
install.packages("tidylearn")

# Development version
# devtools::install_github("ces0491/tidylearn") # nolint
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## The Unified Interface

The core of tidylearn is the `tl_model()` function, which dispatches to the
appropriate underlying package based on the method you specify. The wrapped
packages include stats, glmnet, randomForest, xgboost, gbm, e1071, nnet, rpart,
cluster, and dbscan.

### Supervised Learning

#### Classification

Logistic regression handles two-class problems, so we take a binary subset
of iris here. For three or more classes use `"tree"`, `"forest"`, `"svm"`
or `"nn"`.

```{r}
# versicolor and virginica overlap, so this is a real classification
# problem -- setosa is linearly separable from the other two, which makes
# logistic regression fail to converge
iris_binary <- iris %>%
  filter(Species %in% c("versicolor", "virginica")) %>%
  mutate(Species = droplevels(Species))

model_logistic <- tl_model(iris_binary, Species ~ ., method = "logistic")
print(model_logistic)
```

Predictions come back as a tibble with a `.pred` column. What `.pred`
contains depends on `type`: `"class"` gives the predicted label, `"prob"`
gives one column per class.

```{r}
# Predicted class labels
predictions <- predict(model_logistic, type = "class")
head(predictions)
```

```{r}
# Class probabilities
head(predict(model_logistic, type = "prob"))
```

Note that the default `type = "response"` means different things across
methods — probabilities for logistic regression, class labels for trees and
forests. Ask for `type = "class"` explicitly when you want labels, or let
`tl_evaluate()` handle it:

```{r}
tl_evaluate(model_logistic, metrics = c("accuracy", "f1"))
```

#### Regression

```{r}
# Regression with linear model
model_linear <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
print(model_linear)
```

```{r}
# Predictions
predictions_reg <- predict(model_linear)
head(predictions_reg)
```

### Unsupervised Learning

#### Dimensionality Reduction

```{r}
# Principal Component Analysis
model_pca <- tl_model(iris[, 1:4], method = "pca")
print(model_pca)
```

```{r}
# Transform data
transformed <- predict(model_pca)
head(transformed)
```

#### Clustering

```{r}
# K-means clustering
model_kmeans <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
print(model_kmeans)
```

```{r}
# Get cluster assignments
clusters <- model_kmeans$fit$clusters
head(clusters)
```

```{r}
# Compare with actual species
table(clusters$cluster, iris$Species)
```

## Data Preprocessing

`tl_prepare_data()` handles imputation, scaling and encoding in one call, and
records what it did so the same transformation can be replayed on new data:

```{r}
# Prepare data with multiple preprocessing steps
processed <- tl_prepare_data(
  iris,
  Species ~ .,
  impute_method = "mean",
  scale_method = "standardize",
  encode_categorical = FALSE
)
```

```{r}
# Check preprocessing steps applied
names(processed$preprocessing_steps)
```

```{r}
# Use processed data for modeling
model_processed <- tl_model(processed$data, Species ~ ., method = "forest")
```

## Train-Test Splitting

```{r}
# Simple random split
split <- tl_split(iris, prop = 0.7, seed = 123)

# Train model (three species, so a multiclass-capable method)
model_train <- tl_model(split$train, Species ~ ., method = "forest")

# Test predictions
predictions_test <- predict(model_train, new_data = split$test)
head(predictions_test)
```

```{r}
# Stratified split (maintains class proportions)
split_strat <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

# Check proportions are maintained
prop.table(table(split_strat$train$Species))
prop.table(table(split_strat$test$Species))
prop.table(table(iris$Species))
```

## Wrapped Packages

tidylearn provides a unified interface to these established R packages:

### Supervised Methods

| Method | Underlying Package | Function Called |
|--------|-------------------|-----------------|
| `"linear"` | stats | `lm()` |
| `"polynomial"` | stats | `lm()` with `poly()` |
| `"logistic"` | stats | `glm(..., family = binomial)` |
| `"ridge"`, `"lasso"`, `"elastic_net"` | glmnet | `glmnet()` |
| `"tree"` | rpart | `rpart()` |
| `"forest"` | randomForest | `randomForest()` |
| `"boost"` | gbm | `gbm()` |
| `"xgboost"` | xgboost | `xgb.train()` |
| `"svm"` | e1071 | `svm()` |
| `"nn"` | nnet | `nnet()` |
| `"deep"` | keras | `keras_model_sequential()` |

### Unsupervised Methods

| Method | Underlying Package | Function Called |
|--------|-------------------|-----------------|
| `"pca"` | stats | `prcomp()` |
| `"mds"` | stats, MASS, smacof | `cmdscale()`, `isoMDS()`, etc. |
| `"kmeans"` | stats | `kmeans()` |
| `"pam"` | cluster | `pam()` |
| `"clara"` | cluster | `clara()` |
| `"hclust"` | stats | `hclust()` |
| `"dbscan"` | dbscan | `dbscan()` |

### Accessing the Underlying Model

The raw model from the underlying package is reachable through `$fit`:

```{r}
# Example: Access the raw randomForest object
model_forest <- tl_model(iris, Species ~ ., method = "forest")
class(model_forest$fit)  # This is the randomForest object

# Use package-specific functions if needed
# randomForest::varImpPlot(model_forest$fit) # nolint
```

An unsupervised method returns tidied components as well, so its `$fit` is
the list holding them and the wrapped object sits at `$fit$model`:

```{r}
model_pca <- tl_model(iris, ~ ., method = "pca")
names(model_pca$fit)
class(model_pca$fit$model)  # This is the prcomp object
```

## The Whole Workflow

Split, fit, predict, score — the four steps this vignette covered, in the
order they run:

```{r}
# Quick example combining everything
data_split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 42)

# Random forests are scale-invariant, so no scaling is needed here. When a
# method does need scaled inputs, the same transformation has to be applied
# to the test set -- see the Supervised Learning vignette.
model_final <- tl_model(data_split$train, Species ~ ., method = "forest")
test_preds <- predict(model_final, new_data = data_split$test)

accuracy <- mean(test_preds$.pred == data_split$test$Species)
cat("Test accuracy:", round(accuracy * 100, 1), "%\n")
```

## Next Steps

- `vignette("data-ingestion")` — reading from files, databases and cloud
  sources
- `vignette("supervised-learning")` — classification and regression in depth,
  and how to replay preprocessing on a test set
- `vignette("unsupervised-learning")` — clustering, ordination, and choosing
  the number of clusters
- `vignette("market-basket")` — association rules
- `vignette("tuning-and-pipelines")` — hyperparameter search, and bundling a
  workflow you can save
- `vignette("automl")` — searching across methods under a time budget
- `vignette("diagnostics")` — assumptions, influence and model comparison
- `vignette("reporting")` — plots and formatted `gt` tables
- `vignette("integration-workflows")` — combining supervised and unsupervised
  steps
- `vignette("compute-backends")` — when a fit is too slow or too large for
  this machine: GPU routing, cost estimates, and the cloud safety model

---
title: "Automated Machine Learning with tidylearn"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Automated Machine Learning with tidylearn}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5,
  # tl_auto_ml() narrates its progress through message(); the narration is
  # useful at the console and noise in a document
  message = FALSE
)
```

```{r setup}
library(tidylearn)
library(dplyr)
```

## Introduction

`tl_auto_ml()` searches across methods rather than within one. It trains
several models, engineers a couple of feature sets, cross-validates what the
budget allows, and ranks the results.

It orchestrates the wrapped packages rather than implementing anything new.
Every model on the leaderboard is an ordinary tidylearn model, so
`predict()`, `plot()` and `$fit` all work on it.

## A First Run

```{r}
result <- tl_auto_ml(
  iris, Species ~ .,
  time_budget = 30,
  cv_folds = 3
)
```

```{r}
result$leaderboard
```

Three columns: which model, what it scored, and how the score was obtained.

```{r}
result$best_model
```

```{r}
result$task
result$metric
round(as.numeric(result$runtime, units = "secs"), 1)
```

Regression is the same call with a numeric response — `task = "auto"` reads
it off the data:

```{r}
result_reg <- tl_auto_ml(
  mtcars, mpg ~ .,
  time_budget = 30,
  cv_folds = 3
)

result_reg$task
```

```{r}
result_reg$leaderboard
```

## How the Search Works

Four phases run in order, each gated on the remaining budget and on the
`use_reduction` and `use_clustering` toggles.

| Phase | What it does | Classification | Regression |
|-------|-------------|----------------|------------|
| 1. Baselines | Standard models on the raw features | tree, logistic\*, forest | tree, linear, forest |
| 2. PCA variants | PCA preprocessing, then the baselines | pca\_tree, pca\_logistic\*, pca\_forest | pca\_tree, pca\_linear, pca\_forest |
| 3. Cluster variants | Cluster assignment added as a feature | clustered\_tree, clustered\_logistic\*, clustered\_forest | clustered\_tree, clustered\_linear, clustered\_forest |
| 4. Advanced | Heavier methods | svm, xgboost | ridge, lasso |

\* Logistic regression is binary only, so it appears only when the response
has exactly two classes. On a multiclass problem such as `iris` the baselines
are tree and forest.

You can see the phases in what actually got trained:

```{r}
names(result$models)
```

```{r, warning = FALSE}
# The same search on a two-class problem picks up the logistic variants
iris_binary <- iris %>%
  filter(Species != "setosa") %>%
  mutate(Species = droplevels(Species))

binary_result <- tl_auto_ml(iris_binary, Species ~ ., time_budget = 30,
                            cv_folds = 3)
names(binary_result$models)
```

That call emits a run of `glm.fit: algorithm did not converge` warnings,
which this chunk hides only because there are a dozen of them. They are worth
understanding rather than ignoring: *versicolor* and *virginica* overlap only
slightly, and a three-fold split of 100 rows will sometimes produce a fold
where the two are perfectly separable. Logistic regression has no finite
maximum likelihood estimate on separable data, so the coefficients diverge
and `glm()` says so.

The fit on the full data is fine:

```{r}
tl_model(iris_binary, Species ~ ., method = "logistic")$spec$method
```

It is the folds, not the dataset. Seeing this on your own data means the
classes are close to separable at that sample size — a reason to prefer a
regularised method, which is what `ridge` and `lasso` are doing in phase 4 of
the regression search.

### The evaluation column

Each model is fitted on the full training data, then cross-validated if the
budget allows. When it does not, the leaderboard falls back to training-set
metrics.

```{r}
table(result$leaderboard$evaluation)
```

Training metrics are optimistically biased, so a leaderboard mixing `"cv"`
and `"train"` is not ranking like with like. If you see both, raise the
budget or lower `cv_folds`.

## The Time Budget

`time_budget` is in seconds and is checked **between** model fits, not during
them. Once a model starts it runs to completion, because the wrapped packages
execute C-level code that R cannot safely interrupt. Wall-clock time can
therefore overshoot by the duration of whichever model started last.

The budget controls how much gets tried. The relationship on this data and
this machine:

```{r}
budgets <- c(2, 5, 10, 30)

sweep <- lapply(budgets, function(b) {
  t0 <- Sys.time()
  r <- tl_auto_ml(iris, Species ~ ., time_budget = b, cv_folds = 3)
  data.frame(
    budget = b,
    elapsed = round(as.numeric(difftime(Sys.time(), t0, units = "secs")), 1),
    models = nrow(r$leaderboard),
    cv_scored = sum(r$leaderboard$evaluation == "cv"),
    best = r$leaderboard$model[1]
  )
})

do.call(rbind, sweep)
```

iris is 150 rows, so everything here is quick and the budget is barely
touched. On data where a single forest fit takes ten seconds the same numbers
look very different, so run the sweep on your own data.

Two things are worth knowing regardless of size:

- The forest baseline and every advanced model are only attempted when
  `time_budget >= 30`.
- Cross-validation is the expensive step. `tl_cv(folds = 5)` fits five models
  where a plain fit costs one, so `cv_folds` is the most effective lever.

## Controlling the Search

```{r}
# Baselines only -- no PCA, no cluster features
baseline_only <- tl_auto_ml(
  iris, Species ~ .,
  use_reduction = FALSE,
  use_clustering = FALSE,
  time_budget = 30,
  cv_folds = 3
)

names(baseline_only$models)
```

```{r}
# Keep PCA, drop clustering
no_clustering <- tl_auto_ml(
  iris, Species ~ .,
  use_clustering = FALSE,
  time_budget = 30,
  cv_folds = 3
)

names(no_clustering$models)
```

Naming a metric changes what the leaderboard optimises. Classification
accepts `accuracy`, `precision`, `recall`, `sensitivity`, `specificity`,
`f1`, `auc` and `pr_auc`; regression accepts `rmse`, `mse`, `mae`, `mape`
and `rsq`.

```{r, warning = FALSE}
by_f1 <- tl_auto_ml(iris_binary, Species ~ ., metric = "f1",
                    time_budget = 30, cv_folds = 3)

by_f1$metric
by_f1$leaderboard
```

## Using the Result

```{r}
split <- tl_split(iris, prop = 0.7, stratify = "Species", seed = 123)

automl <- tl_auto_ml(split$train, Species ~ ., time_budget = 30, cv_folds = 3)

test_preds <- predict(automl$best_model, new_data = split$test)
mean(test_preds$.pred == split$test$Species)
```

Any model on the leaderboard can be pulled out by name:

```{r}
available <- names(automl$models)
available
```

```{r}
scores <- vapply(available, function(nm) {
  preds <- predict(automl$models[[nm]], new_data = split$test)
  mean(preds$.pred == split$test$Species)
}, numeric(1))

data.frame(model = available, test_accuracy = round(scores, 3),
           row.names = NULL) %>%
  arrange(desc(test_accuracy))
```

Comparing the leaderboard against held-out accuracy is worth doing: the
model that won the search is not always the one that generalises, and with a
dataset this small the difference between the top few is noise.

## AutoML Against a Single Choice

```{r}
manual <- tl_model(split$train, Species ~ ., method = "forest")
manual_acc <- mean(
  predict(manual, new_data = split$test)$.pred == split$test$Species
)

automl_acc <- mean(test_preds$.pred == split$test$Species)

data.frame(
  approach = c("forest, chosen by hand", "AutoML best"),
  test_accuracy = round(c(manual_acc, automl_acc), 3)
)
```

On iris a random forest is already the right answer, so the search does not
improve on it. That is the expected result on an easy, well-understood
dataset — AutoML earns its cost when you do not already know which method
suits the data.

## Preprocessing First

AutoML does not impute. Handle missing values before the call, and apply the
same transformation to any data you later predict on:

```{r}
processed <- tl_prepare_data(
  split$train, Species ~ .,
  scale_method = "standardize",
  remove_correlated = TRUE
)

automl_processed <- tl_auto_ml(processed$data, Species ~ .,
                               time_budget = 30, cv_folds = 3)

automl_processed$leaderboard$model[1]
```

`tl_pipeline()` handles that bookkeeping properly — see
`vignette("tuning-and-pipelines")`.

## When a Model Is Missing or Scores NA

A model that fails to fit, or errors during evaluation, is dropped with a
message explaining why and never reaches the leaderboard. Run without
`message = FALSE` to see which and why.

A model that evaluates but produces no value for the chosen metric appears
with an `NA` score. That usually means the metric is not one `tl_evaluate()`
recognises for the task. If every score is `NA`, `tl_auto_ml()` warns and
returns the first model trained rather than pretending to have ranked them.

```{r}
sum(is.na(result$leaderboard$score))
```

## AutoML or Tuning?

They search different axes and compose in one direction:

- `tl_auto_ml()` searches **across methods** at default hyperparameters.
- `tl_tune_grid()` and `tl_tune_random()` search **within one method**.

Run AutoML first to find out which family suits the data, then tune the
winner:

```{r}
winner <- automl$best_model$spec$method
winner
```

```{r}
tuned <- tl_tune_grid(
  split$train, Species ~ .,
  method = winner,
  param_grid = tl_default_param_grid(winner, size = "small"),
  folds = 3,
  verbose = FALSE
)

attr(tuned, "tuning_results")$best_params
```

```{r}
mean(predict(tuned, new_data = split$test)$.pred == split$test$Species)
```

## Guidance

1. **Start at `time_budget = 10`** to confirm the call is well formed, then
   raise it. Nothing about a 10-second run tells you which model is best.
2. **Lower `cv_folds` before lowering `time_budget`.** Going from 5 folds to
   2 removes 60% of the evaluation cost and still gives out-of-sample
   estimates. A 30-second budget at `cv_folds = 2` is more informative than
   the same budget at the default 5, which will skip CV entirely.
3. **Preprocess first.** AutoML does not impute or scale.
4. **Score on held-out data.** The leaderboard ranks; it does not report
   final performance.
5. **Read past the first row.** When the top few scores are within noise of
   each other, pick on interpretability or fit time instead.
6. **Check the `evaluation` column** before comparing scores.

## Where to Go Next

- **Tuning and pipelines** (`vignette("tuning-and-pipelines")`) — search
  within a method, then freeze the recipe.
- **Diagnostics** (`vignette("diagnostics")`) — what to check about the model
  AutoML handed you.
- **Integration workflows** (`vignette("integration-workflows")`) — the PCA
  and clustering steps AutoML applies, driven by hand.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-deep-learning.R
\name{tl_plot_deep_architecture}
\alias{tl_plot_deep_architecture}
\title{Plot deep learning model architecture}
\usage{
tl_plot_deep_architecture(model, ...)
}
\arguments{
\item{model}{A tidylearn deep learning model object}

\item{...}{Additional arguments}
}
\value{
The return value of \code{keras::plot_model()}, an architecture
diagram of the Keras model.
}
\description{
Plot deep learning model architecture
}
\examples{
\dontrun{
if (requireNamespace("keras", quietly = TRUE)) {
  model <- tl_model(iris, Species ~ ., method = "deep", epochs = 5)
  tl_plot_deep_architecture(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_tune_xgboost}
\alias{tl_tune_xgboost}
\title{Tune XGBoost hyperparameters}
\usage{
tl_tune_xgboost(
  data,
  formula,
  is_classification = FALSE,
  param_grid = NULL,
  cv_folds = 5,
  early_stopping_rounds = 10,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{param_grid}{Named list of parameter values to try}

\item{cv_folds}{Number of cross-validation folds (default: 5)}

\item{early_stopping_rounds}{Early stopping rounds (default: 10)}

\item{verbose}{Logical indicating whether to print progress (default: TRUE)}

\item{...}{Additional arguments}
}
\value{
A \code{tidylearn_model} object (the refit on full data using the
best hyperparameters) with an attribute \code{"tuning_results"} containing
a list with elements \code{param_grid}, \code{results} (per-combination CV
output), \code{best_params}, \code{best_iteration}, \code{best_score}, and
\code{minimize}.
}
\description{
Tune XGBoost hyperparameters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{tl_split}
\alias{tl_split}
\title{Split data into train and test sets}
\usage{
tl_split(data, prop = 0.8, stratify = NULL, seed = NULL)
}
\arguments{
\item{data}{A data frame}

\item{prop}{Proportion for training set (default: 0.8)}

\item{stratify}{Column name for stratified splitting}

\item{seed}{Random seed for reproducibility}
}
\value{
A list with two elements:
\describe{
\item{\code{$train}}{A data frame containing the training subset.}
\item{\code{$test}}{A data frame containing the test subset.}
}
}
\description{
Split data into train and test sets
}
\examples{
\donttest{
split_data <- tl_split(iris, prop = 0.7, stratify = "Species")
train <- split_data$train
test <- split_data$test
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-svm.R
\name{tidylearn-svm}
\alias{tidylearn-svm}
\title{Support Vector Machines for tidylearn}
\description{
SVM functionality for classification and
regression
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_plot_xgboost_shap_dependence}
\alias{tl_plot_xgboost_shap_dependence}
\title{Plot SHAP dependence for a specific feature}
\usage{
tl_plot_xgboost_shap_dependence(
  model,
  feature,
  interaction_feature = NULL,
  data = NULL,
  n_samples = 100
)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{feature}{Feature name to plot}

\item{interaction_feature}{Feature to use for coloring (default: NULL)}

\item{data}{Data for SHAP value calculation
(default: NULL, uses training data)}

\item{n_samples}{Number of samples to use
(default: 100, NULL for all)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot SHAP dependence for a specific feature
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{tl_semisupervised}
\alias{tl_semisupervised}
\title{Semi-Supervised Learning via Clustering}
\usage{
tl_semisupervised(
  data,
  formula,
  labeled_indices,
  cluster_method = "kmeans",
  supervised_method = "logistic",
  ...
)
}
\arguments{
\item{data}{A data frame}

\item{formula}{Model formula}

\item{labeled_indices}{Indices of labeled observations}

\item{cluster_method}{Clustering method for label propagation}

\item{supervised_method}{Supervised learning method for final model}

\item{...}{Additional arguments}
}
\value{
A tidylearn model object with additional class
\code{"tidylearn_semisupervised"}, trained on pseudo-labeled data. The
model includes a \code{semisupervised_info} element with
\code{labeled_indices}, \code{cluster_model}, and
\code{label_mapping}.
}
\description{
Train a supervised model with limited labels by first clustering the data
and propagating labels within clusters.
}
\examples{
\donttest{
# Use only 10\% of labels
labeled_idx <- sample(nrow(iris), size = 15)
model <- tl_semisupervised(iris, Species ~ ., labeled_indices = labeled_idx,
  cluster_method = "kmeans",
  supervised_method = "tree"
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-selection.R
\name{tl_step_selection}
\alias{tl_step_selection}
\title{Perform stepwise selection on a linear model}
\usage{
tl_step_selection(
  data,
  formula,
  direction = "backward",
  criterion = "AIC",
  trace = FALSE,
  steps = 1000,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the initial model}

\item{direction}{Direction of stepwise selection:
"forward", "backward", or "both"}

\item{criterion}{Criterion for selection: "AIC" or "BIC"}

\item{trace}{Logical; whether to print progress}

\item{steps}{Maximum number of steps to take}

\item{...}{Additional arguments to pass to step()}
}
\value{
A \code{tidylearn_model} object of class
\code{tidylearn_linear} wrapping the selected \code{\link[stats]{lm}}
model. Access the underlying model via \code{$fit} and the selected
formula via \code{$spec$formula}.
}
\description{
Perform stepwise selection on a linear model
}
\examples{
\donttest{
model <- tl_step_selection(mtcars, mpg ~ ., direction = "backward")
summary(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read}
\alias{tl_read}
\title{Read data from diverse sources}
\usage{
tl_read(source, ..., format = NULL, .quiet = FALSE)
}
\arguments{
\item{source}{A file path, URL, connection string, directory path, or a
character vector of multiple file paths.}

\item{...}{Additional arguments passed to the format-specific reader.}

\item{format}{Optional explicit format override.
One of \code{"csv"}, \code{"tsv"},
\code{"excel"}, \code{"parquet"}, \code{"json"},
\code{"rds"}, \code{"rdata"},
\code{"sqlite"}, \code{"postgres"}, \code{"mysql"}, \code{"bigquery"},
\code{"s3"}, \code{"github"}, \code{"kaggle"}. When \code{NULL} (default),
the format is auto-detected from the file extension
or source pattern. Note: \code{.txt} files default
to CSV; use \code{format = "tsv"} to override.}

\item{.quiet}{Logical. If \code{TRUE}, suppresses
informational messages. Default is \code{FALSE}.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Auto-detects the data format from the file extension or source pattern and
dispatches to the appropriate reader. All readers
return a \code{tidylearn_data} object, which is a
tibble subclass carrying metadata about the data
source.
}
\details{
When \code{source} is a character vector of multiple paths, each file is read
and row-bound into a single result with a \code{source_file} column. When
\code{source} is a directory path, it is equivalent to calling
\code{tl_read_dir()}. When \code{source} is a \code{.zip} file, it is
equivalent to calling \code{tl_read_zip()}.
}
\examples{
\donttest{
# Read a single CSV file
# data <- tl_read("path/to/data.csv")

# Read multiple files and row-bind
# data <- tl_read(c("jan.csv", "feb.csv", "mar.csv"))

# Read all CSVs from a directory
# data <- tl_read("data/")

# Read from a zip archive
# data <- tl_read("data.zip")

# Explicit format override
# data <- tl_read("path/to/data.txt", format = "tsv")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_extract_importance}
\alias{tl_extract_importance}
\title{Extract importance from a tree-based model}
\usage{
tl_extract_importance(model)
}
\arguments{
\item{model}{A tidylearn model object}
}
\value{
A data frame with feature importance values
}
\description{
Extract importance from a tree-based model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{tidy_pca}
\alias{tidy_pca}
\title{Tidy Principal Component Analysis}
\usage{
tidy_pca(data, cols = NULL, scale = TRUE, center = TRUE, method = "prcomp")
}
\arguments{
\item{data}{A data frame or tibble}

\item{cols}{Columns to include in PCA (tidy select
syntax). If NULL, uses all numeric columns.}

\item{scale}{Logical; should variables be scaled to
unit variance? Default TRUE.}

\item{center}{Logical; should variables be centered? Default TRUE.}

\item{method}{Character; "prcomp" (default, recommended) or "princomp"}
}
\value{
A list of class "tidy_pca" containing:
\itemize{
\item scores: tibble of PC scores with observation identifiers
\item loadings: tibble of variable loadings in long format
\item variance: tibble of variance explained by each PC
\item model: the original prcomp/princomp object
\item settings: list of scale, center, method used
}
}
\description{
Performs PCA on a dataset using tidyverse principles. Returns a tidy list
containing scores, loadings, variance explained, and the original model.
}
\examples{
# Basic PCA
pca_result <- tidy_pca(USArrests)


# Access components
pca_result$scores
pca_result$loadings
pca_result$variance

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{summary.tidylearn_pipeline}
\alias{summary.tidylearn_pipeline}
\title{Summarize a tidylearn pipeline}
\usage{
\method{summary}{tidylearn_pipeline}(object, ...)
}
\arguments{
\item{object}{A tidylearn pipeline object}

\item{...}{Additional arguments (not used)}
}
\value{
The input pipeline \code{object}, returned invisibly. Called
for its side effect of printing detailed pipeline and model results.
}
\description{
Summarize a tidylearn pipeline
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .)
summary(pipe)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{visualize_rules}
\alias{visualize_rules}
\title{Visualize Association Rules}
\usage{
visualize_rules(rules_obj, method = "scatter", top_n = 50, ...)
}
\arguments{
\item{rules_obj}{A tidy_apriori object, rules object, or rules tibble}

\item{method}{Visualization method: "scatter"
(default), "graph", "grouped", "paracoord"}

\item{top_n}{Number of top rules to visualize (default: 50)}

\item{...}{Additional arguments passed to plot() for rules visualization}
}
\value{
A \code{\link[ggplot2]{ggplot}} object when
\code{method = "scatter"}.
For other methods, the plot is produced as a side effect via
\pkg{arulesViz}.
}
\description{
Create visualizations of association rules
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  visualize_rules(res, method = "scatter")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{calc_validation_metrics}
\alias{calc_validation_metrics}
\title{Calculate Cluster Validation Metrics}
\usage{
calc_validation_metrics(clusters, data = NULL, dist_mat = NULL)
}
\arguments{
\item{clusters}{Vector of cluster assignments}

\item{data}{Original data frame (for WSS calculation)}

\item{dist_mat}{Distance matrix (for silhouette)}
}
\value{
A single-row tibble with columns \code{k}, \code{min_size},
\code{max_size}, \code{avg_size}, and optionally \code{avg_silhouette},
\code{min_silhouette} (if \code{dist_mat} provided), and \code{total_wss}
(if \code{data} provided).
}
\description{
Comprehensive validation metrics for a clustering result
}
\examples{
\donttest{
km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
d <- dist(iris[, 1:4])
metrics <- calc_validation_metrics(km$cluster, iris[, 1:4], d)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_sqlite}
\alias{tl_read_sqlite}
\title{Read from a SQLite database}
\usage{
tl_read_sqlite(path, query, ...)
}
\arguments{
\item{path}{Path to a SQLite database file (\code{.sqlite} or \code{.db}).}

\item{query}{A SQL query string.}

\item{...}{Additional arguments passed to \code{DBI::dbGetQuery()}.}
}
\value{
A \code{tidylearn_data} object containing the query results.
}
\description{
Opens a SQLite database file, executes a SQL query, and returns the result
as a \code{tidylearn_data} object. The connection is automatically closed
when done. Requires \pkg{DBI} and \pkg{RSQLite}.
}
\examples{
\donttest{
# data <- tl_read_sqlite("my_database.sqlite", "SELECT * FROM my_table")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{tidy_knn_dist}
\alias{tidy_knn_dist}
\title{Compute k-NN Distances}
\usage{
tidy_knn_dist(data, k = 4, cols = NULL)
}
\arguments{
\item{data}{A data frame or matrix}

\item{k}{Number of nearest neighbors (default: 4)}

\item{cols}{Columns to include (tidy select).
If NULL, uses all numeric columns.}
}
\value{
A tibble with columns \code{.obs_id} (observation identifier),
\code{knn_dist} (distance to k-th nearest neighbor), and \code{rank}
(rank of the k-NN distance).
}
\description{
Calculate distances to k-th nearest neighbor for each point
}
\examples{
\donttest{
knn <- tidy_knn_dist(iris[, 1:4], k = 5)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tidylearn-diagnostics}
\alias{tidylearn-diagnostics}
\title{Advanced Diagnostics Functions for tidylearn}
\description{
Functions for advanced model diagnostics,
assumption checking, and outlier detection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-mds.R
\name{plot_mds}
\alias{plot_mds}
\title{Plot MDS Configuration}
\usage{
plot_mds(mds_obj, color_by = NULL, label_points = TRUE, dim_x = 1, dim_y = 2)
}
\arguments{
\item{mds_obj}{A tidy_mds object}

\item{color_by}{Optional variable to color points by}

\item{label_points}{Logical; add point labels? (default: TRUE)}

\item{dim_x}{Which dimension for x-axis (default: 1)}

\item{dim_y}{Which dimension for y-axis (default: 2)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize MDS results
}
\examples{
\donttest{
mds <- tidy_mds(USArrests, method = "classical")
plot_mds(mds)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{inspect_rules}
\alias{inspect_rules}
\title{Inspect Association Rules}
\usage{
inspect_rules(rules_obj, by = "lift", n = 10, decreasing = TRUE)
}
\arguments{
\item{rules_obj}{A tidy_apriori object or rules object}

\item{by}{Sort by: "support", "confidence", "lift" (default), "count"}

\item{n}{Number of rules to display (default: 10)}

\item{decreasing}{Sort in decreasing order? (default: TRUE)}
}
\value{
A tibble of the top \code{n} rules sorted by the specified quality
measure.
}
\description{
View rules sorted by various quality measures
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  inspect_rules(res, by = "lift", n = 5)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tl_tune_nn}
\alias{tl_tune_nn}
\title{Tune a neural network model}
\usage{
tl_tune_nn(
  data,
  formula,
  is_classification = FALSE,
  sizes = c(1, 2, 5, 10),
  decays = c(0, 0.001, 0.01, 0.1),
  folds = 5,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{sizes}{Vector of hidden layer sizes to try}

\item{decays}{Vector of weight decay parameters to try}

\item{folds}{Number of cross-validation folds (default: 5)}

\item{...}{Additional arguments to pass to nnet()}
}
\value{
A list with elements \code{model} (the best fitted \code{nnet}
model), \code{best_size} (optimal hidden-layer size), \code{best_decay}
(optimal weight decay), and \code{tuning_results} (a data frame of all
parameter combinations and their cross-validated errors).
}
\description{
Tune a neural network model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{suggest_eps}
\alias{suggest_eps}
\title{Suggest eps Parameter for DBSCAN}
\usage{
suggest_eps(data, minPts = 5, method = "percentile", percentile = 0.95)
}
\arguments{
\item{data}{A data frame or matrix}

\item{minPts}{Minimum points parameter (used as k for k-NN)}

\item{method}{Method to suggest eps: "knee" (default), "percentile"}

\item{percentile}{If method="percentile", which
percentile to use (default: 0.95)}
}
\value{
A list containing:
\itemize{
\item eps: suggested epsilon value
\item knn_distances: full tibble of k-NN distances
\item method: method used
}
}
\description{
Use k-NN distance plot to suggest eps value
}
\examples{
eps_info <- suggest_eps(iris, minPts = 5)
eps_info$eps

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tl_plot_regularization_path}
\alias{tl_plot_regularization_path}
\title{Plot regularization path for a regularized model}
\usage{
tl_plot_regularization_path(model, label_n = 5, ...)
}
\arguments{
\item{model}{A tidylearn regularized model object}

\item{label_n}{Number of top features to label
(default: 5)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot regularization path for a regularized model
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ ., method = "lasso")
tl_plot_regularization_path(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{tidy_pca_screeplot}
\alias{tidy_pca_screeplot}
\title{Create PCA Scree Plot}
\usage{
tidy_pca_screeplot(pca_obj, type = "proportion", add_line = TRUE)
}
\arguments{
\item{pca_obj}{A tidy_pca object}

\item{type}{Character; "variance" or "proportion" (default)}

\item{add_line}{Logical; add horizontal line at
eigenvalue = 1? (for Kaiser criterion)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize variance explained by each principal component
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
tidy_pca_screeplot(pca)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{tidylearn-core}
\alias{tidylearn-core}
\title{tidylearn: A Unified Tidy Interface to R's Machine Learning Ecosystem}
\description{
Core functionality for tidylearn. This package
provides a unified tidyverse-compatible interface to
established R machine learning packages including glmnet,
randomForest, xgboost, e1071, rpart, gbm, nnet, cluster,
and dbscan. The underlying algorithms are unchanged -
tidylearn wraps them with consistent function signatures,
tidy tibble output, and unified ggplot2-based
visualization. Access raw model objects via model$fit.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{print.tidy_hclust}
\alias{print.tidy_hclust}
\title{Print Method for tidy_hclust}
\usage{
\method{print}{tidy_hclust}(x, ...)
}
\arguments{
\item{x}{A tidy_hclust object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_hclust
}
\examples{
\donttest{
hc <- tidy_hclust(USArrests, method = "ward.D2")
print(hc)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_distance_heatmap}
\alias{plot_distance_heatmap}
\title{Create Distance Heatmap}
\usage{
plot_distance_heatmap(
  dist_mat,
  cluster_order = NULL,
  title = "Distance Heatmap"
)
}
\arguments{
\item{dist_mat}{Distance matrix (dist object)}

\item{cluster_order}{Optional vector to reorder observations by cluster}

\item{title}{Plot title}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize distance matrix as heatmap
}
\examples{
\donttest{
d <- dist(iris[1:20, 1:4])
plot_distance_heatmap(d)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tl_detect_outliers}
\alias{tl_detect_outliers}
\title{Detect outliers in the data}
\usage{
tl_detect_outliers(
  data,
  variables = NULL,
  method = "iqr",
  threshold = NULL,
  plot = TRUE
)
}
\arguments{
\item{data}{A data frame containing the data}

\item{variables}{Character vector of variables to check for outliers}

\item{method}{Method for outlier detection: "boxplot", "z-score", "cook",
"iqr", "mahalanobis"}

\item{threshold}{Threshold for outlier detection}

\item{plot}{Logical; whether to create a plot of outliers}
}
\value{
A list with outlier detection results:
\describe{
\item{method}{The detection method used (character).}
\item{method_name}{Human-readable method name (character).}
\item{threshold}{The threshold value used (numeric).}
\item{threshold_label}{Formatted threshold description (character).}
\item{outlier_flags}{A logical matrix (observations x variables).}
\item{any_outlier}{Logical vector indicating if each observation is an
outlier in any variable.}
\item{outlier_counts}{List with \code{total}, \code{by_variable}, and
\code{by_observation} counts.}
\item{outlier_indices}{Integer vector of outlier row indices.}
\item{plot}{A \code{\link[ggplot2]{ggplot}} object, or \code{NULL} if
\code{plot = FALSE}.}
}
}
\description{
Detect outliers in the data
}
\examples{
\donttest{
tl_detect_outliers(mtcars, variables = c("mpg", "wt"), method = "iqr")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-svm.R
\name{tl_fit_svm}
\alias{tl_fit_svm}
\title{Fit a support vector machine model}
\usage{
tl_fit_svm(
  data,
  formula,
  is_classification = FALSE,
  kernel = "radial",
  cost = 1,
  gamma = NULL,
  degree = 3,
  tune = FALSE,
  tune_folds = 5,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{kernel}{Kernel function
("linear", "polynomial", "radial", "sigmoid")}

\item{cost}{Cost parameter (default: 1)}

\item{gamma}{Gamma parameter for kernels
(default: 1/ncol(data))}

\item{degree}{Degree for polynomial kernel (default: 3)}

\item{tune}{Logical indicating whether to tune
hyperparameters (default: FALSE)}

\item{tune_folds}{Number of folds for cross-validation
during tuning (default: 5)}

\item{...}{Additional arguments to pass to svm()}
}
\value{
A fitted SVM model
}
\description{
Fit a support vector machine model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{print.tidy_apriori}
\alias{print.tidy_apriori}
\title{Print Method for tidy_apriori}
\usage{
\method{print}{tidy_apriori}(x, ...)
}
\arguments{
\item{x}{A tidy_apriori object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_apriori
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  print(res)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{find_related_items}
\alias{find_related_items}
\title{Find Related Items}
\usage{
find_related_items(rules_obj, item, min_lift = 1.5, top_n = 10)
}
\arguments{
\item{rules_obj}{A tidy_apriori object}

\item{item}{Character; item to find associations for}

\item{min_lift}{Minimum lift threshold (default: 1.5)}

\item{top_n}{Number of top associations to return (default: 10)}
}
\value{
A tibble of rules involving the specified \code{item}, filtered by
\code{min_lift} and sorted by lift in descending order.
}
\description{
Find items frequently purchased with a given item
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  find_related_items(res, "whole milk", min_lift = 1.5)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-classification.R
\name{tl_fit_logistic}
\alias{tl_fit_logistic}
\title{Fit a logistic regression model}
\usage{
tl_fit_logistic(data, formula, ...)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{...}{Additional arguments to pass to glm()}
}
\value{
A fitted logistic regression model
}
\description{
Fit a logistic regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tidylearn-read-backends}
\alias{tidylearn-read-backends}
\title{Data Reading Backends for tidylearn}
\description{
Backend readers for databases and cloud/API sources.
All backends are optional dependencies checked at call time via
\code{tl_check_packages()}.
}
\details{
Database backends (via \pkg{DBI}):
\itemize{
\item \strong{SQLite}: via \pkg{RSQLite}
\item \strong{PostgreSQL}: via \pkg{RPostgres}
\item \strong{MySQL/MariaDB}: via \pkg{RMariaDB}
\item \strong{BigQuery}: via \pkg{bigrquery}
}

Cloud/API backends:
\itemize{
\item \strong{S3}: via \pkg{paws.storage}
\item \strong{GitHub}: via base \code{download.file()}
\item \strong{Kaggle}: via Kaggle CLI
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{tidy_pam}
\alias{tidy_pam}
\title{Tidy PAM (Partitioning Around Medoids)}
\usage{
tidy_pam(data, k, metric = "euclidean", cols = NULL)
}
\arguments{
\item{data}{A data frame, tibble, or dist object}

\item{k}{Number of clusters}

\item{metric}{Distance metric (default: "euclidean").
Use "gower" for mixed data types.}

\item{cols}{Columns to include (tidy select). If NULL, uses all columns.}
}
\value{
A list of class "tidy_pam" containing:
\itemize{
\item clusters: tibble with observation IDs and cluster assignments
\item medoids: tibble of medoid indices and values
\item silhouette: average silhouette width
\item model: original pam object
}
}
\description{
Performs PAM clustering with tidy output
}
\examples{
# PAM with Euclidean distance
pam_result <- tidy_pam(iris, k = 3)

# PAM with Gower distance for mixed data
pam_result <- tidy_pam(mtcars, k = 3, metric = "gower")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactions.R
\name{tl_test_interactions}
\alias{tl_test_interactions}
\title{Test for significant interactions between variables}
\usage{
tl_test_interactions(
  data,
  formula,
  var1 = NULL,
  var2 = NULL,
  all_pairs = FALSE,
  categorical_only = FALSE,
  numeric_only = FALSE,
  mixed_only = FALSE,
  alpha = 0.05
)
}
\arguments{
\item{data}{A data frame containing the data}

\item{formula}{A formula specifying the base model without interactions}

\item{var1}{First variable to test for interactions}

\item{var2}{Second variable to test for interactions
(if NULL, tests var1 with all others)}

\item{all_pairs}{Logical; whether to test all variable pairs}

\item{categorical_only}{Logical; whether to only test categorical variables}

\item{numeric_only}{Logical; whether to only test numeric variables}

\item{mixed_only}{Logical; whether to only test numeric-categorical pairs}

\item{alpha}{Significance level for interaction tests}
}
\value{
A data frame with one row per tested interaction pair, containing
columns \code{var1}, \code{var2}, \code{p_value}, \code{significant}
(logical), \code{delta_r2} (change in R-squared), and
\code{f_statistic}, sorted by \code{p_value} ascending.
}
\description{
Test for significant interactions between variables
}
\examples{
\donttest{
results <- tl_test_interactions(mtcars, mpg ~ wt + hp + cyl,
  var1 = "wt", var2 = "hp")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{augment_dbscan}
\alias{augment_dbscan}
\title{Augment Data with DBSCAN Cluster Assignments}
\usage{
augment_dbscan(dbscan_obj, data)
}
\arguments{
\item{dbscan_obj}{A tidy_dbscan object}

\item{data}{Original data frame}
}
\value{
A tibble containing the original \code{data} with additional columns
\code{cluster} (factor), \code{is_noise} (logical), and \code{is_core}
(logical).
}
\description{
Augment Data with DBSCAN Cluster Assignments
}
\examples{
\donttest{
db <- tidy_dbscan(iris[, 1:4], eps = 0.5, minPts = 5)
augmented <- augment_dbscan(db, iris)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-selection.R
\name{tidylearn-model-selection}
\alias{tidylearn-model-selection}
\title{Model Selection Functions for tidylearn}
\description{
Functions for stepwise model selection,
cross-validation, and hyperparameter tuning
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-distance.R
\name{tidy_gower}
\alias{tidy_gower}
\title{Gower Distance Calculation}
\usage{
tidy_gower(data, weights = NULL)
}
\arguments{
\item{data}{A data frame or tibble}

\item{weights}{Optional named vector of variable
weights (default: equal weights)}
}
\value{
A \code{\link[stats]{dist}} object containing Gower distances, with
the \code{method} attribute set to \code{"gower"}.
}
\description{
Computes Gower distance for mixed data types (numeric, factor, ordered)
}
\details{
Gower distance handles mixed data types:
\itemize{
\item Numeric: range-normalized Manhattan distance
\item Factor/Character: 0 if same, 1 if different
\item Ordered: treated as numeric ranks
}

Formula: d_ij = sum(w_k * d_ijk) / sum(w_k)
where d_ijk is the dissimilarity for variable k between obs i and j
}
\examples{
# Create example data with mixed types
car_data <- data.frame(
  horsepower = c(130, 250, 180),
  weight = c(1200, 1650, 1420),
  color = factor(c("red", "black", "blue"))
)

# Compute Gower distance
gower_dist <- tidy_gower(car_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-distance.R
\name{compare_distances}
\alias{compare_distances}
\title{Compare Distance Methods}
\usage{
compare_distances(data, methods = c("euclidean", "manhattan", "maximum"))
}
\arguments{
\item{data}{A data frame or tibble}

\item{methods}{Character vector of methods to compare}
}
\value{
A named list of \code{\link[stats]{dist}} objects, one per method.
}
\description{
Compute distances using multiple methods for comparison
}
\examples{
\donttest{
dists <- compare_distances(
  iris[, 1:4], methods = c("euclidean", "manhattan")
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{get_pca_variance}
\alias{get_pca_variance}
\title{Get Variance Explained Summary}
\usage{
get_pca_variance(pca_obj)
}
\arguments{
\item{pca_obj}{A tidy_pca object}
}
\value{
A tibble with columns \code{component}, \code{sdev},
\code{variance}, \code{prop_variance}, and \code{cum_variance}.
}
\description{
Get Variance Explained Summary
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
get_pca_variance(pca)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{print.tidylearn_model}
\alias{print.tidylearn_model}
\title{Print method for tidylearn models}
\usage{
\method{print}{tidylearn_model}(x, ...)
}
\arguments{
\item{x}{A tidylearn model object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print method for tidylearn models
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
print(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{tidylearn-workflows}
\alias{tidylearn-workflows}
\title{High-Level Workflows for Common Machine Learning Patterns}
\description{
Functions providing end-to-end workflows that showcase
tidylearn's ability to seamlessly combine multiple learning paradigms
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_save_pipeline}
\alias{tl_save_pipeline}
\title{Save a pipeline to disk}
\usage{
tl_save_pipeline(pipeline, file)
}
\arguments{
\item{pipeline}{A tidylearn pipeline object}

\item{file}{Path to save the pipeline}
}
\value{
Called for its side effect of saving to disk; returns
\code{NULL} invisibly.
}
\description{
Save a pipeline to disk
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .)
tl_save_pipeline(pipe, tempfile(fileext = ".rds"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tl_plot_nn_architecture}
\alias{tl_plot_nn_architecture}
\title{Plot neural network architecture}
\usage{
tl_plot_nn_architecture(model, ...)
}
\arguments{
\item{model}{A tidylearn neural network model object}

\item{...}{Additional arguments}
}
\value{
The return value of \code{\link[NeuralNetTools]{plotnet}}, called for
its side effect of drawing the network diagram, or \code{NULL} if the
\pkg{NeuralNetTools} package is not installed.
}
\description{
Plot neural network architecture
}
\examples{
\donttest{
if (requireNamespace("NeuralNetTools", quietly = TRUE)) {
  model <- tl_model(iris, Species ~ ., method = "nn", size = 3)
  tl_plot_nn_architecture(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-mds.R
\name{tidy_mds_kruskal}
\alias{tidy_mds_kruskal}
\title{Kruskal's Non-metric MDS}
\usage{
tidy_mds_kruskal(dist_mat, ndim = 2, ...)
}
\arguments{
\item{dist_mat}{A distance matrix (dist object)}

\item{ndim}{Number of dimensions (default: 2)}

\item{...}{Additional arguments passed to MASS::isoMDS()}
}
\value{
A list of class \code{"tidy_mds"} containing:
\itemize{
\item config: tibble of MDS coordinates
\item stress: Kruskal stress value
\item method: \code{"Kruskal's isoMDS"}
\item model: the \code{\link[MASS]{isoMDS}} result
}
}
\description{
Performs Kruskal's isoMDS
}
\examples{
\donttest{
d <- dist(USArrests)
mds <- tidy_mds_kruskal(d)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{augment_kmeans}
\alias{augment_kmeans}
\title{Augment Data with K-Means Cluster Assignments}
\usage{
augment_kmeans(kmeans_obj, data)
}
\arguments{
\item{kmeans_obj}{A tidy_kmeans object}

\item{data}{Original data frame}
}
\value{
A tibble containing the original \code{data} with an additional
\code{cluster} factor column indicating cluster assignments.
}
\description{
Augment Data with K-Means Cluster Assignments
}
\examples{
\donttest{
km <- tidy_kmeans(iris[, 1:4], k = 3)
augmented <- augment_kmeans(km, iris)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_xgboost_shap}
\alias{tl_xgboost_shap}
\title{Generate SHAP values for XGBoost model interpretation}
\usage{
tl_xgboost_shap(model, data = NULL, n_samples = 100, trees_idx = NULL)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{data}{Data for SHAP value calculation
(default: NULL, uses training data)}

\item{n_samples}{Number of samples to use (default: 100, NULL for all)}

\item{trees_idx}{Trees to include (default: NULL, uses all trees)}
}
\value{
A data frame with one column of SHAP values per feature, a
\code{BIAS} column, a \code{row_id} column, and the original data columns
appended for reference.
}
\description{
Generate SHAP values for XGBoost model interpretation
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE)) {
  model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
  shap <- tl_xgboost_shap(model, n_samples = 20)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{print.tidy_pca}
\alias{print.tidy_pca}
\title{Print Method for tidy_pca}
\usage{
\method{print}{tidy_pca}(x, ...)
}
\arguments{
\item{x}{A tidy_pca object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_pca
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
print(pca)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-svm.R
\name{tl_predict_svm}
\alias{tl_predict_svm}
\title{Predict using a support vector machine model}
\usage{
tl_predict_svm(model, new_data, type = "response", ...)
}
\arguments{
\item{model}{A tidylearn SVM model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response" (default),
"prob" (for classification)}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a support vector machine model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-mds.R
\name{tidy_mds_smacof}
\alias{tidy_mds_smacof}
\title{SMACOF MDS (Metric or Non-metric)}
\usage{
tidy_mds_smacof(dist_mat, ndim = 2, type = "ratio", ...)
}
\arguments{
\item{dist_mat}{A distance matrix (dist object)}

\item{ndim}{Number of dimensions (default: 2)}

\item{type}{Character; "ratio" for metric, "ordinal"
for non-metric (default: "ratio")}

\item{...}{Additional arguments passed to smacof::mds()}
}
\value{
A list of class \code{"tidy_mds"} containing:
\itemize{
\item config: tibble of MDS coordinates
\item stress: stress value from the SMACOF algorithm
\item method: character string describing the MDS type
\item model: the \code{\link[smacof]{mds}} result
}
}
\description{
Performs MDS using SMACOF algorithm from the smacof package
}
\examples{
\donttest{
d <- dist(USArrests)
mds <- tidy_mds_smacof(d, type = "ratio")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_parquet}
\alias{tl_read_parquet}
\title{Read a Parquet file}
\usage{
tl_read_parquet(path, ...)
}
\arguments{
\item{path}{Path to a Parquet file.}

\item{...}{Additional arguments passed to \code{nanoparquet::read_parquet()}.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads a Parquet file into a \code{tidylearn_data} object. Requires the
\pkg{nanoparquet} package.
}
\examples{
\donttest{
# data <- tl_read_parquet("path/to/data.parquet")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{tl_version}
\alias{tl_version}
\title{Get tidylearn version information}
\usage{
tl_version()
}
\value{
A package_version object containing the version number
}
\description{
Get tidylearn version information
}
\examples{
tl_version()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{tidy_pca_biplot}
\alias{tidy_pca_biplot}
\title{Create PCA Biplot}
\usage{
tidy_pca_biplot(
  pca_obj,
  pc_x = 1,
  pc_y = 2,
  color_by = NULL,
  arrow_scale = 1,
  label_obs = FALSE,
  label_vars = TRUE
)
}
\arguments{
\item{pca_obj}{A tidy_pca object}

\item{pc_x}{Principal component for x-axis (default: 1)}

\item{pc_y}{Principal component for y-axis (default: 2)}

\item{color_by}{Optional column name to color points by}

\item{arrow_scale}{Scaling factor for variable arrows (default: 1)}

\item{label_obs}{Logical; label observations? (default: FALSE)}

\item{label_vars}{Logical; label variables? (default: TRUE)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize both observations and variables in PC space
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
tidy_pca_biplot(pca)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-mds.R
\name{tidy_mds}
\alias{tidy_mds}
\title{Tidy Multidimensional Scaling}
\usage{
tidy_mds(data, method = "classical", ndim = 2, distance = "euclidean", ...)
}
\arguments{
\item{data}{A data frame, tibble, or distance matrix}

\item{method}{Character; "classical" (default),
"metric", "nonmetric", "sammon", or "kruskal"}

\item{ndim}{Number of dimensions for output (default: 2)}

\item{distance}{Character; distance metric if data is
not already a dist object (default: "euclidean")}

\item{...}{Additional arguments passed to specific MDS functions}
}
\value{
A list of class "tidy_mds" containing:
\itemize{
\item config: tibble of MDS configuration (coordinates)
\item stress: goodness-of-fit measure (if applicable)
\item method: character string of method used
\item model: original model object
}
}
\description{
Unified interface for MDS methods with tidy output
}
\examples{
# Classical MDS
mds_result <- tidy_mds(eurodist, method = "classical")
print(mds_result)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{tl_add_cluster_features}
\alias{tl_add_cluster_features}
\title{Cluster-Based Features}
\usage{
tl_add_cluster_features(data, response = NULL, method = "kmeans", ...)
}
\arguments{
\item{data}{A data frame}

\item{response}{Response variable name (will be excluded from clustering)}

\item{method}{Clustering method: "kmeans", "pam", "hclust", "dbscan"}

\item{...}{Additional arguments for clustering}
}
\value{
The original data frame with an additional factor column named
\code{cluster_<method>} containing cluster assignments. The fitted
cluster model is stored as an attribute \code{"cluster_model"}.
}
\description{
Add cluster assignments as features for supervised learning.
This semi-supervised approach can capture non-linear patterns.
}
\examples{
\donttest{
# Add cluster features before supervised learning
data_with_clusters <- tl_add_cluster_features(iris, response = "Species",
                                                method = "kmeans", k = 3)
model <- tl_model(data_with_clusters, Species ~ ., method = "forest")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_plot_diagnostics}
\alias{tl_plot_diagnostics}
\title{Plot diagnostics for a regression model}
\usage{
tl_plot_diagnostics(model, which = 1:4, ...)
}
\arguments{
\item{model}{A tidylearn regression model object}

\item{which}{Which plots to create (1:4)}

\item{...}{Additional arguments}
}
\value{
A ggplot object (or list of ggplot objects)
}
\description{
Plot diagnostics for a regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tidylearn-xgboost}
\alias{tidylearn-xgboost}
\title{XGBoost Functions for tidylearn}
\description{
XGBoost-specific implementation for gradient boosting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_plot_importance}
\alias{tl_plot_importance}
\title{Plot variable importance for tree-based models}
\usage{
tl_plot_importance(model, top_n = 20, ...)
}
\arguments{
\item{model}{A tidylearn tree-based model object}

\item{top_n}{Number of top features to display (default: 20)}

\item{...}{Additional arguments}
}
\value{
A ggplot object
}
\description{
Plot variable importance for tree-based models
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{print.tidy_gap}
\alias{print.tidy_gap}
\title{Print Method for tidy_gap}
\usage{
\method{print}{tidy_gap}(x, ...)
}
\arguments{
\item{x}{A tidy_gap object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_gap
}
\examples{
\donttest{
gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
print(gap)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_coefficients}
\alias{tl_table_coefficients}
\title{Formatted model coefficients table}
\usage{
tl_table_coefficients(model, lambda = "1se", digits = 4, ...)
}
\arguments{
\item{model}{A tidylearn model object}

\item{lambda}{For regularised models: "1se" (default) or "min"}

\item{digits}{Number of decimal places (default: 4)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table of model coefficients. Supports linear,
polynomial, logistic, ridge, lasso, and elastic net models.
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_table_coefficients(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tl_plot_nn_tuning}
\alias{tl_plot_nn_tuning}
\title{Plot neural network training history}
\usage{
tl_plot_nn_tuning(model, ...)
}
\arguments{
\item{model}{A tidylearn neural network model object}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot neural network training history
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning.R
\name{tl_plot_tuning_results}
\alias{tl_plot_tuning_results}
\title{Plot hyperparameter tuning results}
\usage{
tl_plot_tuning_results(
  model,
  top_n = 5,
  param1 = NULL,
  param2 = NULL,
  plot_type = "scatter"
)
}
\arguments{
\item{model}{A tidylearn model object with tuning
results}

\item{top_n}{Number of top parameter sets to
highlight}

\item{param1}{First parameter to plot (for 2D grid
or scatter plots)}

\item{param2}{Second parameter to plot (for 2D grid
or scatter plots)}

\item{plot_type}{Type of plot: "scatter", "grid",
"parallel", "importance"}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot hyperparameter tuning results
}
\examples{
\donttest{
model <- tl_tune_grid(iris, Species ~ ., method = "tree",
  param_grid = list(cp = c(0.01, 0.1), minsplit = c(10, 20)),
  folds = 2, verbose = FALSE)
tl_plot_tuning_results(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_comparison}
\alias{tl_table_comparison}
\title{Compare multiple models in a formatted table}
\usage{
tl_table_comparison(..., new_data = NULL, names = NULL, digits = 4)
}
\arguments{
\item{...}{tidylearn model objects to compare}

\item{new_data}{Optional test data for evaluation. If NULL, uses the
training data of the first model.}

\item{names}{Optional character vector of model names}

\item{digits}{Number of decimal places (default: 4)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Evaluates multiple tidylearn models and presents the results side-by-side
in a styled gt table.
}
\examples{
\donttest{
m1 <- tl_model(mtcars, mpg ~ ., method = "linear")
m2 <- tl_model(mtcars, mpg ~ ., method = "lasso")
tl_table_comparison(m1, m2, names = c("Linear", "Lasso"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_db}
\alias{tl_read_db}
\title{Read from a DBI database connection}
\usage{
tl_read_db(conn, query, ...)
}
\arguments{
\item{conn}{A \pkg{DBI} connection object (e.g., from
\code{DBI::dbConnect()}).}

\item{query}{A SQL query string.}

\item{...}{Additional arguments passed to \code{DBI::dbGetQuery()}.}
}
\value{
A \code{tidylearn_data} object containing the query results.
}
\description{
Executes a SQL query against an existing \pkg{DBI} connection and returns
the result as a \code{tidylearn_data} object. The connection is not closed
by this function — the caller is responsible for managing the connection
lifecycle.
}
\examples{
\donttest{
# conn <- DBI::dbConnect(RSQLite::SQLite(), "my_database.sqlite")
# data <- tl_read_db(conn, "SELECT * FROM my_table")
# DBI::dbDisconnect(conn)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_predict_boost}
\alias{tl_predict_boost}
\title{Predict using a gradient boosting model}
\usage{
tl_predict_boost(model, new_data, type = "response", n.trees = NULL, ...)
}
\arguments{
\item{model}{A tidylearn boost model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response"
(default), "prob" (for classification)}

\item{n.trees}{Number of trees to use for prediction
(if NULL, uses optimal number)}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a gradient boosting model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_predict_pipeline}
\alias{tl_predict_pipeline}
\title{Make predictions using a pipeline}
\usage{
tl_predict_pipeline(
  pipeline,
  new_data,
  type = "response",
  model_name = NULL,
  ...
)
}
\arguments{
\item{pipeline}{A tidylearn pipeline object with results}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction (default: "response")}

\item{model_name}{Name of model to use (if NULL, uses the best model)}

\item{...}{Additional arguments passed to predict}
}
\value{
A \link[tibble]{tibble} with a \code{.pred} column containing
predictions from the selected (or best) pipeline model, after
applying the same preprocessing steps used during training.
}
\description{
Make predictions using a pipeline
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_plot_cv_results}
\alias{tl_plot_cv_results}
\title{Plot cross-validation results}
\usage{
tl_plot_cv_results(cv_results, metrics = NULL)
}
\arguments{
\item{cv_results}{Cross-validation results from tl_cv function}

\item{metrics}{Character vector of metrics to plot
(if NULL, plots all metrics)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot cross-validation results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_elbow}
\alias{plot_elbow}
\title{Create Elbow Plot for K-Means}
\usage{
plot_elbow(wss_data, add_line = FALSE, suggested_k = NULL)
}
\arguments{
\item{wss_data}{A tibble with columns k and tot_withinss (from calc_wss)}

\item{add_line}{Add vertical line at suggested optimal k? (default: FALSE)}

\item{suggested_k}{If add_line=TRUE, which k to highlight}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot total within-cluster sum of squares vs number of clusters
}
\examples{
\donttest{
wss <- data.frame(k = 2:6, tot_withinss = c(150, 90, 60, 50, 45))
plot_elbow(wss)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{summarize_rules}
\alias{summarize_rules}
\title{Summarize Association Rules}
\usage{
summarize_rules(rules_obj)
}
\arguments{
\item{rules_obj}{A tidy_apriori object or rules tibble}
}
\value{
A list with \code{n_rules} and summary statistics (\code{min},
\code{max}, \code{mean}, \code{median}) for \code{support},
\code{confidence}, and \code{lift}.
}
\description{
Get summary statistics about rules
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  summarize_rules(res)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_plot_residuals}
\alias{tl_plot_residuals}
\title{Plot residuals for a regression model}
\usage{
tl_plot_residuals(model, type = "fitted", ...)
}
\arguments{
\item{model}{A tidylearn regression model object}

\item{type}{Type of residual plot: "fitted" (default),
"histogram", "predicted"}

\item{...}{Additional arguments}
}
\value{
A ggplot object
}
\description{
Plot residuals for a regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tl_evaluate_thresholds}
\alias{tl_evaluate_thresholds}
\title{Evaluate metrics at different thresholds}
\usage{
tl_evaluate_thresholds(actuals, probs, thresholds, pos_class)
}
\arguments{
\item{actuals}{Actual values (ground truth)}

\item{probs}{Predicted probabilities}

\item{thresholds}{Vector of thresholds to evaluate}

\item{pos_class}{The positive class}
}
\value{
A tibble of metrics at different thresholds
}
\description{
Evaluate metrics at different thresholds
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_loadings}
\alias{tl_table_loadings}
\title{Formatted PCA loadings table}
\usage{
tl_table_loadings(model, n_components = NULL, digits = 3, ...)
}
\arguments{
\item{model}{A tidylearn PCA model object}

\item{n_components}{Number of components to show (default: all)}

\item{digits}{Number of decimal places (default: 3)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table of variable loadings on each principal component,
with a diverging colour scale to highlight strong loadings.
}
\examples{
\donttest{
model <- tl_model(iris[, 1:4], method = "pca")
tl_table_loadings(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning.R
\name{tidylearn-tuning}
\alias{tidylearn-tuning}
\title{Hyperparameter Tuning Functions for tidylearn}
\description{
Functions for automatic hyperparameter tuning and selection
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{tidy_kmeans}
\alias{tidy_kmeans}
\title{Tidy K-Means Clustering}
\usage{
tidy_kmeans(
  data,
  k,
  cols = NULL,
  nstart = 25,
  iter_max = 100,
  algorithm = "Hartigan-Wong"
)
}
\arguments{
\item{data}{A data frame or tibble}

\item{k}{Number of clusters}

\item{cols}{Columns to include (tidy select).
If NULL, uses all numeric columns.}

\item{nstart}{Number of random starts (default: 25)}

\item{iter_max}{Maximum iterations (default: 100)}

\item{algorithm}{K-means algorithm: "Hartigan-Wong"
(default), "Lloyd", "Forgy", "MacQueen"}
}
\value{
A list of class "tidy_kmeans" containing:
\itemize{
\item clusters: tibble with observation IDs and cluster assignments
\item centers: tibble of cluster centers
\item metrics: tibble with clustering quality metrics
\item model: original kmeans object
}
}
\description{
Performs k-means clustering with tidy output
}
\examples{
# Basic k-means
km_result <- tidy_kmeans(iris, k = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tidylearn-pipeline}
\alias{tidylearn-pipeline}
\title{Model Pipeline Functions for tidylearn}
\description{
Functions for creating end-to-end model pipelines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-deep-learning.R
\name{tl_tune_deep}
\alias{tl_tune_deep}
\title{Tune a deep learning model}
\usage{
tl_tune_deep(
  data,
  formula,
  is_classification = FALSE,
  hidden_layers_options = list(c(32), c(64, 32), c(128, 64, 32)),
  learning_rates = c(0.01, 0.001, 1e-04),
  batch_sizes = c(16, 32, 64),
  epochs = 30,
  validation_split = 0.2,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{hidden_layers_options}{List of vectors defining hidden
layer configurations to try}

\item{learning_rates}{Learning rates to try
(default: c(0.01, 0.001, 0.0001))}

\item{batch_sizes}{Batch sizes to try
(default: c(16, 32, 64))}

\item{epochs}{Number of training epochs (default: 30)}

\item{validation_split}{Proportion of data for validation
(default: 0.2)}

\item{...}{Additional arguments}
}
\value{
A list with elements \code{model} (the best fitted deep learning
model), \code{best_hidden_layers} (optimal layer configuration),
\code{best_learning_rate}, \code{best_batch_size}, and
\code{tuning_results} (a data frame of all hyperparameter combinations
and their validation losses).
}
\description{
Tune a deep learning model
}
\examples{
\dontrun{
if (requireNamespace("keras", quietly = TRUE)) {
  result <- tl_tune_deep(iris, Species ~ .,
    is_classification = TRUE,
    hidden_layers_options = list(c(10), c(10, 5)),
    learning_rates = c(0.01, 0.001), batch_sizes = c(32),
    epochs = 5)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tl_calculate_pr_auc}
\alias{tl_calculate_pr_auc}
\title{Calculate the area under the precision-recall curve}
\usage{
tl_calculate_pr_auc(perf)
}
\arguments{
\item{perf}{A ROCR performance object}
}
\value{
The area under the PR curve
}
\description{
Calculate the area under the precision-recall curve
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_dendrogram}
\alias{plot_dendrogram}
\title{Plot Dendrogram with Cluster Highlights}
\usage{
plot_dendrogram(
  hclust_obj,
  k = NULL,
  title = "Hierarchical Clustering Dendrogram"
)
}
\arguments{
\item{hclust_obj}{Hierarchical clustering object (hclust or tidy_hclust)}

\item{k}{Number of clusters to highlight}

\item{title}{Plot title}
}
\value{
Invisibly returns the \code{\link[stats]{hclust}} object. The
dendrogram is drawn as a side effect.
}
\description{
Enhanced dendrogram with colored cluster rectangles
}
\examples{
\donttest{
hc <- hclust(dist(iris[, 1:4]))
plot_dendrogram(hc, k = 3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_plot_tree}
\alias{tl_plot_tree}
\title{Plot a decision tree}
\usage{
tl_plot_tree(model, ...)
}
\arguments{
\item{model}{A tidylearn tree model object}

\item{...}{Additional arguments to pass to rpart.plot()}
}
\value{
The return value of \code{\link[rpart.plot]{rpart.plot}}, called
for its side effect of drawing the tree.
}
\description{
Plot a decision tree
}
\examples{
\donttest{
model <- tl_model(iris, Species ~ ., method = "tree")
tl_plot_tree(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_variance_explained}
\alias{plot_variance_explained}
\title{Plot Variance Explained (PCA)}
\usage{
plot_variance_explained(variance_tbl, threshold = 0.8)
}
\arguments{
\item{variance_tbl}{Variance tibble from tidy_pca}

\item{threshold}{Horizontal line for variance threshold
(default: 0.8 for 80\%)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Create combined scree plot showing individual and cumulative variance
}
\examples{
\donttest{
model <- tl_model(iris[, 1:4], method = "pca")
plot_variance_explained(model$fit$variance_explained)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{compare_clusterings}
\alias{compare_clusterings}
\title{Compare Multiple Clustering Results}
\usage{
compare_clusterings(cluster_list, data, dist_mat = NULL)
}
\arguments{
\item{cluster_list}{Named list of cluster assignment vectors}

\item{data}{Original data}

\item{dist_mat}{Distance matrix}
}
\value{
A tibble with one row per clustering method and columns for each
validation metric (see \code{\link{calc_validation_metrics}}), plus a
\code{method} column identifying the clustering.
}
\description{
Compare Multiple Clustering Results
}
\examples{
\donttest{
km3 <- kmeans(iris[, 1:4], 3, nstart = 25)$cluster
km4 <- kmeans(iris[, 1:4], 4, nstart = 25)$cluster
compare_clusterings(list(k3 = km3, k4 = km4), iris[, 1:4])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{print.tidylearn_pipeline}
\alias{print.tidylearn_pipeline}
\title{Print a tidylearn pipeline}
\usage{
\method{print}{tidylearn_pipeline}(x, ...)
}
\arguments{
\item{x}{A tidylearn pipeline object}

\item{...}{Additional arguments (not used)}
}
\value{
The input pipeline object \code{x}, returned invisibly.
}
\description{
Print a tidylearn pipeline
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .)
print(pipe)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_plot_model_comparison}
\alias{tl_plot_model_comparison}
\title{Plot model comparison}
\usage{
tl_plot_model_comparison(..., new_data = NULL, metrics = NULL, names = NULL)
}
\arguments{
\item{...}{tidylearn model objects to compare}

\item{new_data}{Optional data frame for evaluation
(if NULL, uses training data)}

\item{metrics}{Character vector of metrics to compute}

\item{names}{Optional character vector of model names}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot model comparison
}
\examples{
\donttest{
m1 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "lasso")
tl_plot_model_comparison(m1, m2, names = c("Linear", "Lasso"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{tl_reduce_dimensions}
\alias{tl_reduce_dimensions}
\title{Integration Functions: Combining Supervised and Unsupervised Learning}
\usage{
tl_reduce_dimensions(
  data,
  response = NULL,
  method = "pca",
  n_components = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame}

\item{response}{Response variable name (will be preserved)}

\item{method}{Dimensionality reduction method: "pca", "mds"}

\item{n_components}{Number of components to retain}

\item{...}{Additional arguments for the dimensionality reduction method}
}
\value{
A list with components:
\describe{
\item{data}{The transformed data frame with reduced-dimension columns
and the response variable (if provided).}
\item{reduction_model}{The fitted tidylearn dimensionality reduction
model.}
\item{original_data}{The original input data frame.}
\item{response}{The response variable name, or \code{NULL}.}
}
}
\description{
These functions demonstrate the power of tidylearn's unified approach by
seamlessly integrating supervised and unsupervised learning techniques.
Feature Engineering via Dimensionality Reduction
}
\details{
Use PCA, MDS, or other dimensionality reduction as a preprocessing step
for supervised learning. This can improve model performance
and interpretability.
}
\examples{
\donttest{
# Reduce dimensions before classification
reduced <- tl_reduce_dimensions(
  iris, response = "Species",
  method = "pca", n_components = 3
)
model <- tl_model(reduced$data, Species ~ ., method = "tree")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_rds}
\alias{tl_read_rds}
\title{Read an RDS file}
\usage{
tl_read_rds(path)
}
\arguments{
\item{path}{Path to an RDS file.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads an RDS file into a \code{tidylearn_data} object. Uses base R
\code{readRDS()} — no additional packages required.
}
\examples{
\donttest{
# data <- tl_read_rds("path/to/data.rds")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{tl_plot_unsupervised}
\alias{tl_plot_unsupervised}
\title{Plot an unsupervised tidylearn model}
\usage{
tl_plot_unsupervised(model, type = "auto", ...)
}
\arguments{
\item{model}{A tidylearn unsupervised model object}

\item{type}{Plot type (default: "auto"). Currently unused; reserved for
future sub-type selection.}

\item{...}{Additional arguments passed to the underlying plot function}
}
\value{
A ggplot2 object or invisible result
}
\description{
Dispatches to the appropriate plotting function based on the unsupervised
model method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tidylearn-regression}
\alias{tidylearn-regression}
\title{Regression Functions for tidylearn}
\description{
Linear and polynomial regression functionality
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_metrics}
\alias{tl_table_metrics}
\title{Formatted evaluation metrics table}
\usage{
tl_table_metrics(model, new_data = NULL, digits = 4, ...)
}
\arguments{
\item{model}{A tidylearn supervised model object}

\item{new_data}{Optional test data. If NULL, uses training data.}

\item{digits}{Number of decimal places (default: 4)}

\item{...}{Additional arguments passed to \code{tl_evaluate}}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table of model evaluation metrics from
\code{\link{tl_evaluate}}.
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_table_metrics(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-classification.R
\name{tidylearn-classification}
\alias{tidylearn-classification}
\title{Classification Functions for tidylearn}
\description{
Logistic regression and classification metrics functionality
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{filter_rules_by_item}
\alias{filter_rules_by_item}
\title{Filter Rules by Item}
\usage{
filter_rules_by_item(rules_obj, item, where = "both")
}
\arguments{
\item{rules_obj}{A tidy_apriori object or tibble of rules}

\item{item}{Character; item to filter by}

\item{where}{Character; "lhs", "rhs", or "both" (default: "both")}
}
\value{
A tibble of rules containing the specified \code{item} in the
requested position.
}
\description{
Subset rules containing specific items
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  filter_rules_by_item(res, "whole milk", where = "rhs")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{plot_knn_dist}
\alias{plot_knn_dist}
\title{Plot k-NN Distance Plot}
\usage{
plot_knn_dist(data, k = 4, add_suggestion = TRUE, percentile = 0.95)
}
\arguments{
\item{data}{A data frame or tidy_knn_dist result}

\item{k}{If data is a data frame, k for k-NN (default: 4)}

\item{add_suggestion}{Add suggested eps line? (default: TRUE)}

\item{percentile}{Percentile for suggestion (default: 0.95)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize k-NN distances to help choose eps
}
\examples{
\donttest{
plot_knn_dist(iris[, 1:4], k = 5)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table}
\alias{tl_table}
\title{Create formatted tables for tidylearn models}
\usage{
tl_table(model, type = "auto", ...)
}
\arguments{
\item{model}{A tidylearn model object}

\item{type}{Table type (default: "auto"). For supervised models: "metrics",
"coefficients", "confusion", "importance". For unsupervised models:
"variance", "loadings", "clusters". MDS models are not supported.}

\item{...}{Additional arguments passed to the underlying table function}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Dispatches to the appropriate table function based on model type and
requested table type. Requires the gt package.
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_table(model)
tl_table(model, type = "coefficients")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{plot_silhouette}
\alias{plot_silhouette}
\title{Plot Silhouette Analysis}
\usage{
plot_silhouette(sil_obj)
}
\arguments{
\item{sil_obj}{A tidy_silhouette object or tibble
from tidy_silhouette_analysis}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot Silhouette Analysis
}
\examples{
\donttest{
km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
d <- dist(iris[, 1:4])
sil <- tidy_silhouette(km$cluster, d)
plot_silhouette(sil)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tidylearn-tables}
\alias{tidylearn-tables}
\title{Table Functions for tidylearn}
\description{
Functions for producing formatted gt tables
from tidylearn models. Provides a parallel interface to
the plot functions: \code{tl_table(model, type)}
dispatches to the appropriate table formatter based on model type.
Requires the gt package (suggested dependency).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tl_check_assumptions}
\alias{tl_check_assumptions}
\title{Check model assumptions}
\usage{
tl_check_assumptions(model, test = TRUE, verbose = TRUE)
}
\arguments{
\item{model}{A tidylearn model object}

\item{test}{Logical; whether to perform statistical tests}

\item{verbose}{Logical; whether to print test results and explanations}
}
\value{
A named list with one element per assumption checked
(\code{linearity}, \code{independence}, \code{homoscedasticity},
\code{normality}, \code{multicollinearity}, \code{outliers}), each
containing \code{assumption} (character label), \code{check} (logical
or \code{NULL}), \code{details} (character), and
\code{recommendation} (character). An additional \code{overall} element
summarises the number of assumptions checked, violated, and satisfied.
}
\description{
Check model assumptions
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_check_assumptions(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_json}
\alias{tl_read_json}
\title{Read a JSON file}
\usage{
tl_read_json(path, flatten = TRUE, ...)
}
\arguments{
\item{path}{Path to a JSON file.}

\item{flatten}{Logical. Automatically flatten nested data frames? Default is
\code{TRUE}.}

\item{...}{Additional arguments passed to \code{jsonlite::fromJSON()}.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads a JSON file into a \code{tidylearn_data} object. Expects the JSON to
represent tabular data (array of objects or similar). Requires the
\pkg{jsonlite} package.
}
\examples{
\donttest{
# data <- tl_read_json("path/to/data.json")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tidylearn-visualization}
\alias{tidylearn-visualization}
\title{Visualization Functions for tidylearn}
\description{
General visualization functions for tidylearn models
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_plot_xgboost_shap_summary}
\alias{tl_plot_xgboost_shap_summary}
\title{Plot SHAP summary for XGBoost model}
\usage{
tl_plot_xgboost_shap_summary(model, data = NULL, top_n = 10, n_samples = 100)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{data}{Data for SHAP value calculation
(default: NULL, uses training data)}

\item{top_n}{Number of top features to display (default: 10)}

\item{n_samples}{Number of samples to use
(default: 100, NULL for all)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot SHAP summary for XGBoost model
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE)) {
  model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
  tl_plot_xgboost_shap_summary(model, n_samples = 20)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_pipeline}
\alias{tl_pipeline}
\title{Create a modeling pipeline}
\usage{
tl_pipeline(
  data,
  formula,
  preprocessing = NULL,
  models = NULL,
  evaluation = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame containing the data}

\item{formula}{A formula specifying the model}

\item{preprocessing}{A list of preprocessing steps}

\item{models}{A list of models to train}

\item{evaluation}{A list of evaluation criteria}

\item{...}{Additional arguments}
}
\value{
A \code{tidylearn_pipeline} object (S3 list) with components
\code{$formula}, \code{$data}, \code{$preprocessing},
\code{$models}, \code{$evaluation}, and \code{$results}
(initially \code{NULL}; populated after \code{\link{tl_run_pipeline}}).
}
\description{
Create a modeling pipeline
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .,
  models = list(tree = list(method = "tree")))
print(pipe)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-distance.R
\name{standardize_data}
\alias{standardize_data}
\title{Standardize Data}
\usage{
standardize_data(data, center = TRUE, scale = TRUE)
}
\arguments{
\item{data}{A data frame or tibble}

\item{center}{Logical; center variables? (default: TRUE)}

\item{scale}{Logical; scale variables to unit variance? (default: TRUE)}
}
\value{
A tibble with numeric variables centered and/or scaled as specified;
non-numeric columns are returned unchanged.
}
\description{
Center and/or scale numeric variables
}
\examples{
\donttest{
std <- standardize_data(iris[, 1:4])
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-distance.R
\name{tidy_dist}
\alias{tidy_dist}
\title{Tidy Distance Matrix Computation}
\usage{
tidy_dist(data, method = "euclidean", cols = NULL, ...)
}
\arguments{
\item{data}{A data frame or tibble}

\item{method}{Character; distance method
(default: "euclidean"). Options: "euclidean",
"manhattan", "maximum", "gower"}

\item{cols}{Columns to include (tidy select).
If NULL, uses all numeric columns.}

\item{...}{Additional arguments passed to distance functions}
}
\value{
A \code{\link[stats]{dist}} object containing the computed
distance matrix.
}
\description{
Compute distance matrices with tidy output
}
\examples{
\donttest{
d <- tidy_dist(iris[, 1:4], method = "euclidean")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-selection.R
\name{tl_test_model_difference}
\alias{tl_test_model_difference}
\title{Perform statistical comparison of models using cross-validation}
\usage{
tl_test_model_difference(
  cv_results,
  baseline_model = NULL,
  test = "t.test",
  metric = NULL
)
}
\arguments{
\item{cv_results}{Results from tl_compare_cv function}

\item{baseline_model}{Name of the model to use as baseline for comparison}

\item{test}{Type of statistical test: "t.test" or "wilcox"}

\item{metric}{Name of the metric to compare}
}
\value{
A data frame with columns \code{metric}, \code{model},
\code{baseline}, \code{mean_diff}, \code{p_value}, and
\code{p_adj} (Holm-adjusted p-value) containing pairwise
statistical comparisons against the baseline model.
}
\description{
Perform statistical comparison of models using cross-validation
}
\examples{
\donttest{
m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
tl_test_model_difference(cv, baseline_model = "simple", metric = "rmse")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_tsv}
\alias{tl_read_tsv}
\title{Read a TSV file}
\usage{
tl_read_tsv(path, ...)
}
\arguments{
\item{path}{Path to a TSV file.}

\item{...}{Additional arguments passed to \code{readr::read_tsv()} or
\code{utils::read.delim()}.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads a tab-separated file into a
\code{tidylearn_data} object. Uses \pkg{readr} when
available for faster parsing, with a base R fallback.
}
\examples{
\donttest{
# data <- tl_read_tsv("path/to/data.tsv")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_plot_xgboost_importance}
\alias{tl_plot_xgboost_importance}
\title{Plot feature importance for an XGBoost model}
\usage{
tl_plot_xgboost_importance(model, top_n = 10, importance_type = "gain", ...)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{top_n}{Number of top features to display (default: 10)}

\item{importance_type}{Type of importance: "gain", "cover", "frequency"}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot feature importance for an XGBoost model
}
\examples{
\donttest{
if (requireNamespace("xgboost", quietly = TRUE)) {
  model <- tl_model(mtcars, mpg ~ ., method = "xgboost")
  tl_plot_xgboost_importance(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{recommend_products}
\alias{recommend_products}
\title{Generate Product Recommendations}
\usage{
recommend_products(rules_obj, basket, top_n = 5, min_confidence = 0.5)
}
\arguments{
\item{rules_obj}{A tidy_apriori object}

\item{basket}{Character vector of items in current basket}

\item{top_n}{Number of recommendations to return (default: 5)}

\item{min_confidence}{Minimum confidence threshold (default: 0.5)}
}
\value{
A tibble with columns \code{rhs} (recommended item),
\code{confidence}, \code{lift}, and \code{support}, sorted by lift in
descending order.
}
\description{
Get product recommendations based on basket contents
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  res <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)
  recommend_products(res, basket = c("whole milk", "butter"))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tidylearn-neural-networks}
\alias{tidylearn-neural-networks}
\title{Neural Networks for tidylearn}
\description{
Neural network functionality for classification and regression
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_dir}
\alias{tl_read_dir}
\title{Read all matching files from a directory}
\usage{
tl_read_dir(
  path,
  pattern = NULL,
  format = NULL,
  recursive = FALSE,
  .quiet = FALSE,
  ...
)
}
\arguments{
\item{path}{Path to a directory.}

\item{pattern}{Optional regex pattern to filter file names (e.g.,
\code{"sales_.*\\\\.csv$"}). If \code{NULL}, files are filtered by
\code{format} instead.}

\item{format}{File format to read. If \code{NULL} and \code{pattern} is
\code{NULL}, all recognized data files are read. If specified, only files
with matching extensions are read.}

\item{recursive}{Logical. Should subdirectories be scanned? Default is
\code{FALSE}.}

\item{.quiet}{Suppress messages. Default is \code{FALSE}.}

\item{...}{Additional arguments passed to the format-specific reader.}
}
\value{
A \code{tidylearn_data} object with an additional
\code{source_file} column identifying the origin of each row.
}
\description{
Scans a directory for files matching a pattern or format, reads each one,
and row-binds them into a single \code{tidylearn_data} object with a
\code{source_file} column identifying the origin of each row.
}
\examples{
\donttest{
# Read all CSVs from a directory
# data <- tl_read_dir("data/", format = "csv")

# Read with pattern matching
# data <- tl_read_dir("data/", pattern = "^sales_.*\\.csv$")

# Read all recognized data files recursively
# data <- tl_read_dir("data/", recursive = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{print.tidylearn_data}
\alias{print.tidylearn_data}
\title{Print a tidylearn_data object}
\usage{
\method{print}{tidylearn_data}(x, ...)
}
\arguments{
\item{x}{A \code{tidylearn_data} object.}

\item{...}{Additional arguments passed to the tibble print method.}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print a tidylearn_data object
}
\examples{
\donttest{
f <- tempfile(fileext = ".csv")
write.csv(iris, f, row.names = FALSE)
d <- tl_read(f)
print(d)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_zip}
\alias{tl_read_zip}
\title{Read data from a zip archive}
\usage{
tl_read_zip(path, file = NULL, format = NULL, .quiet = FALSE, ...)
}
\arguments{
\item{path}{Path to a zip file.}

\item{file}{Optional name of a specific file within the archive to read.
Supports partial matching.}

\item{format}{Optional format override for the file(s) inside the archive.}

\item{.quiet}{Suppress messages. Default is \code{FALSE}.}

\item{...}{Additional arguments passed to the format-specific reader.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}. The archive is extracted to a temporary directory
that is cleaned up automatically. If multiple data files are found,
a \code{source_file} column identifies the origin of each row.
}
\description{
Extracts a zip archive to a temporary directory and reads the contents.
If the archive contains a single data file, it is read directly. If
multiple data files are found, they are row-bound with a \code{source_file}
column. Use the \code{file} argument to select a specific file from
the archive.
}
\examples{
\donttest{
# Read from a zip archive
# data <- tl_read_zip("data.zip")

# Read a specific file from the archive
# data <- tl_read_zip("data.zip", file = "train.csv")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{tidy_dendrogram}
\alias{tidy_dendrogram}
\title{Plot Dendrogram}
\usage{
tidy_dendrogram(hclust_obj, k = NULL, hang = 0.01, cex = 0.7)
}
\arguments{
\item{hclust_obj}{A tidy_hclust object or hclust object}

\item{k}{Optional; number of clusters to highlight with rectangles}

\item{hang}{Fraction of plot height to hang labels (default: 0.01)}

\item{cex}{Label size (default: 0.7)}
}
\value{
The \code{\link[stats]{hclust}} object, returned invisibly. The
dendrogram is plotted as a side effect.
}
\description{
Create dendrogram visualization
}
\examples{
\donttest{
hc <- tidy_hclust(USArrests, method = "ward.D2")
tidy_dendrogram(hc, k = 3)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_compare_pipeline_models}
\alias{tl_compare_pipeline_models}
\title{Compare models from a pipeline}
\usage{
tl_compare_pipeline_models(pipeline, metrics = NULL)
}
\arguments{
\item{pipeline}{A tidylearn pipeline object with results}

\item{metrics}{Character vector of metrics to compare
(if NULL, uses all available)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object showing a faceted bar
chart comparing metric values across models, with the best model
highlighted.
}
\description{
Compare models from a pipeline
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-svm.R
\name{tl_plot_svm_boundary}
\alias{tl_plot_svm_boundary}
\title{Plot SVM decision boundary}
\usage{
tl_plot_svm_boundary(model, x_var = NULL, y_var = NULL, grid_size = 100, ...)
}
\arguments{
\item{model}{A tidylearn SVM model object}

\item{x_var}{Name of the x-axis variable}

\item{y_var}{Name of the y-axis variable}

\item{grid_size}{Number of points in each dimension
for the grid (default: 100)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot SVM decision boundary
}
\examples{
\donttest{
if (requireNamespace("e1071", quietly = TRUE)) {
  model <- tl_model(iris, Species ~ ., method = "svm")
  tl_plot_svm_boundary(model,
    x_var = "Sepal.Length", y_var = "Sepal.Width")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_load_pipeline}
\alias{tl_load_pipeline}
\title{Load a pipeline from disk}
\usage{
tl_load_pipeline(file)
}
\arguments{
\item{file}{Path to the pipeline file}
}
\value{
A \code{tidylearn_pipeline} object previously saved with
\code{\link{tl_save_pipeline}}.
}
\description{
Load a pipeline from disk
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .)
f <- tempfile(fileext = ".rds")
tl_save_pipeline(pipe, f)
pipe2 <- tl_load_pipeline(f)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{predict.tidylearn_stratified}
\alias{predict.tidylearn_stratified}
\title{Predict from stratified models}
\usage{
\method{predict}{tidylearn_stratified}(object, new_data = NULL, ...)
}
\arguments{
\item{object}{A tidylearn_stratified model object}

\item{new_data}{New data for predictions}

\item{...}{Additional arguments}
}
\value{
A \link[tibble]{tibble} with a \code{.pred} column containing
predictions and a \code{.cluster} column with cluster assignments.
}
\description{
Predict from stratified models
}
\examples{
\donttest{
models <- tl_stratified_models(mtcars, mpg ~ .,
  cluster_method = "kmeans", k = 2, supervised_method = "linear")
preds <- predict(models)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-classification.R
\name{tl_plot_precision_recall}
\alias{tl_plot_precision_recall}
\title{Plot precision-recall curve for a classification model}
\usage{
tl_plot_precision_recall(model, new_data = NULL, ...)
}
\arguments{
\item{model}{A tidylearn classification model object}

\item{new_data}{Optional data frame for evaluation
(if NULL, uses training data)}

\item{...}{Additional arguments}
}
\value{
A ggplot object with precision-recall curve
}
\description{
Plot precision-recall curve for a classification model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{tl_auto_ml}
\alias{tl_auto_ml}
\title{Auto ML: Automated Machine Learning Workflow}
\usage{
tl_auto_ml(
  data,
  formula,
  task = "auto",
  use_reduction = TRUE,
  use_clustering = TRUE,
  time_budget = 300,
  cv_folds = 5,
  metric = NULL
)
}
\arguments{
\item{data}{A data frame}

\item{formula}{Model formula (for supervised learning)}

\item{task}{Task type: "classification", "regression", or "auto" (default)}

\item{use_reduction}{Whether to try dimensionality reduction (default: TRUE)}

\item{use_clustering}{Whether to add cluster features (default: TRUE)}

\item{time_budget}{Time budget in seconds (default: 300). Controls which
models are attempted and whether cross-validation is used for evaluation.
The budget is checked \strong{between} model fits, not during them -- once a
model starts training it runs to completion because R cannot safely
interrupt C-level code (e.g. randomForest, xgboost, e1071).

How the budget shapes the workflow:
\itemize{
\item \strong{Under 30s}: Only fast models are attempted
(tree, logistic/linear).
Cross-validation is skipped; models are ranked on training-set metrics
only. Expect 2 models in the leaderboard. Use this for quick sanity
checks or interactive exploration.
\item \strong{30--120s}: All baseline models are attempted including random
forest. Cross-validation runs when enough time remains after each
model fit; otherwise training metrics are used. Advanced models
(SVM, XGBoost / ridge, lasso) are attempted if 40\\% of the budget
remains after baselines. Dimensionality reduction and clustering
pipelines run if enabled and 10\\% of the budget remains.
\item \strong{120s+ (recommended)}: The full pipeline runs -- all baselines,
advanced models, PCA-augmented variants, and cluster-augmented
variants, each with cross-validation. Expect 9--11 models in the
leaderboard.
}

Because individual model fits (especially forest, SVM, XGBoost with CV)
can take 5--30s each depending on data size, the actual wall-clock time
may modestly exceed the budget by the duration of the last model that
was started before the budget expired.}

\item{cv_folds}{Number of cross-validation folds (default: 5). Reducing
this (e.g. to 2 or 3) is an effective way to stay closer to the time
budget since CV is typically the most expensive step.}

\item{metric}{Evaluation metric (default: auto-selected based on task).
For classification: "accuracy"; for regression: "rmse".}
}
\value{
A list with class \code{"tidylearn_automl"} containing:
\describe{
\item{best_model}{The best tidylearn model object}
\item{models}{Named list of all successfully trained models}
\item{leaderboard}{Tibble ranking models by the chosen metric}
\item{task}{Detected or specified task type}
\item{metric}{Metric used for ranking}
\item{runtime}{Total elapsed time as a difftime object}
}
}
\description{
Automatically explores multiple modeling approaches including
dimensionality reduction, clustering, and various supervised methods.
Returns the best performing model based on cross-validation.
}
\examples{
\donttest{
# Quick run with fast models only (< 30s budget skips forest/SVM/XGBoost)
result <- tl_auto_ml(iris, Species ~ .,
  time_budget = 10,
  use_reduction = FALSE,
  use_clustering = FALSE,
  cv_folds = 2)
result$leaderboard
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_clusters}
\alias{tl_table_clusters}
\title{Formatted cluster summary table}
\usage{
tl_table_clusters(model, k = 3, digits = 2, ...)
}
\arguments{
\item{model}{A tidylearn clustering model object}

\item{k}{For hclust models, the number of clusters to cut (default: 3)}

\item{digits}{Number of decimal places (default: 2)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table showing cluster sizes and mean feature values.
Supports kmeans, pam, clara, dbscan, and hclust models.
}
\examples{
\donttest{
model <- tl_model(iris[, 1:4], method = "kmeans", k = 3)
tl_table_clusters(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{tidy_dbscan}
\alias{tidy_dbscan}
\title{Tidy DBSCAN Clustering}
\usage{
tidy_dbscan(data, eps, minPts = 5, cols = NULL, distance = "euclidean")
}
\arguments{
\item{data}{A data frame, tibble, or distance matrix}

\item{eps}{Neighborhood radius (epsilon)}

\item{minPts}{Minimum number of points to form a dense region (default: 5)}

\item{cols}{Columns to include (tidy select).
If NULL, uses all numeric columns.}

\item{distance}{Distance metric if data is not a
dist object (default: "euclidean")}
}
\value{
A list of class "tidy_dbscan" containing:
\itemize{
\item clusters: tibble with observation IDs and
cluster assignments (0 = noise)
\item core_points: logical vector indicating core points
\item n_clusters: number of clusters (excluding noise)
\item n_noise: number of noise points
\item model: original dbscan object
}
}
\description{
Performs density-based clustering with tidy output
}
\examples{
# Basic DBSCAN
db_result <- tidy_dbscan(iris, eps = 0.5, minPts = 5)

# With suggested eps from k-NN distance plot
eps_suggestion <- suggest_eps(iris, minPts = 5)
db_result <- tidy_dbscan(iris, eps = eps_suggestion$eps, minPts = 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_fit_tree}
\alias{tl_fit_tree}
\title{Fit a decision tree model}
\usage{
tl_fit_tree(
  data,
  formula,
  is_classification = FALSE,
  cp = 0.01,
  minsplit = 20,
  maxdepth = 30,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{cp}{Complexity parameter (default: 0.01)}

\item{minsplit}{Minimum number of observations in a node for a split}

\item{maxdepth}{Maximum depth of the tree}

\item{...}{Additional arguments to pass to rpart()}
}
\value{
A fitted decision tree model
}
\description{
Fit a decision tree model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-svm.R
\name{tl_plot_svm_tuning}
\alias{tl_plot_svm_tuning}
\title{Plot SVM tuning results}
\usage{
tl_plot_svm_tuning(model, ...)
}
\arguments{
\item{model}{A tidylearn SVM model object}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot SVM tuning results
}
\examples{
\donttest{
if (requireNamespace("e1071", quietly = TRUE)) {
  model <- tl_model(iris, Species ~ ., method = "svm",
    kernel = "linear", tune = TRUE, tune_folds = 2)
  tl_plot_svm_tuning(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_variance}
\alias{tl_table_variance}
\title{Formatted PCA variance explained table}
\usage{
tl_table_variance(model, n_components = NULL, digits = 4, ...)
}
\arguments{
\item{model}{A tidylearn PCA model object}

\item{n_components}{Maximum number of components to show (default: all)}

\item{digits}{Number of decimal places (default: 4)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table of variance explained by each principal component,
with a colour gradient on cumulative variance.
}
\examples{
\donttest{
model <- tl_model(iris[, 1:4], method = "pca")
tl_table_variance(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_predict_polynomial}
\alias{tl_predict_polynomial}
\title{Predict using a polynomial regression model}
\usage{
tl_predict_polynomial(model, new_data, type = "response", level = 0.95, ...)
}
\arguments{
\item{model}{A tidylearn polynomial model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response" (default),
"confidence", "prediction"}

\item{level}{Confidence level for intervals (default: 0.95)}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a polynomial regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_predict_xgboost}
\alias{tl_predict_xgboost}
\title{Predict using an XGBoost model}
\usage{
tl_predict_xgboost(model, new_data, type = "response", ntreelimit = NULL, ...)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response" (default),
"prob" (for classification), "class" (for classification)}

\item{ntreelimit}{Limit number of trees used for prediction
(default: NULL, uses all trees)}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using an XGBoost model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_fit_polynomial}
\alias{tl_fit_polynomial}
\title{Fit a polynomial regression model}
\usage{
tl_fit_polynomial(data, formula, degree = 2, ...)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{degree}{Degree of the polynomial (default: 2)}

\item{...}{Additional arguments to pass to lm()}
}
\value{
A fitted polynomial regression model
}
\description{
Fit a polynomial regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{tl_model}
\alias{tl_model}
\title{Create a tidylearn model}
\usage{
tl_model(data, formula = NULL, method = "linear", ...)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model. For
unsupervised methods, use \code{~ vars} or NULL.}

\item{method}{The modeling method. Supervised: "linear"
(stats::lm), "logistic" (stats::glm), "tree" (rpart),
"forest" (randomForest), "boost" (gbm),
"ridge"/"lasso"/"elastic_net" (glmnet), "svm" (e1071),
"nn" (nnet), "deep" (keras), "xgboost" (xgboost).
Unsupervised: "pca" (stats::prcomp),
"mds" (stats/MASS/smacof), "kmeans" (stats::kmeans),
"pam"/"clara" (cluster), "hclust" (stats::hclust),
"dbscan" (dbscan).}

\item{...}{Additional arguments passed to the underlying model function}
}
\value{
A \code{tidylearn_model} object (S3) containing the fitted model
(\code{$fit}), model specification (\code{$spec}), and training data
(\code{$data}). The object also inherits from a method-specific class
(e.g., \code{tidylearn_linear}) and a paradigm class
(\code{tidylearn_supervised} or \code{tidylearn_unsupervised}).
}
\description{
Unified interface for creating machine learning models
by wrapping established R packages. This function dispatches
to the appropriate underlying package based on the method.
}
\details{
The wrapped packages include: stats (lm, glm, prcomp,
kmeans, hclust), glmnet, randomForest, xgboost, gbm,
e1071, nnet, rpart, cluster, and dbscan. The underlying
algorithms are unchanged - this function provides a
consistent interface and returns tidy output.

Access the raw model object from the underlying package via \code{model$fit}.
}
\examples{
\donttest{
# Classification -> wraps randomForest::randomForest()
model <- tl_model(iris, Species ~ ., method = "forest")
model$fit  # Access the raw randomForest object

# Regression -> wraps stats::lm()
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
model$fit  # Access the raw lm object

# PCA -> wraps stats::prcomp()
model <- tl_model(iris, ~ ., method = "pca")
model$fit  # Access the raw prcomp object

# Clustering -> wraps stats::kmeans()
model <- tl_model(iris, method = "kmeans", k = 3)
model$fit  # Access the raw kmeans object
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{optimal_hclust_k}
\alias{optimal_hclust_k}
\title{Determine Optimal Number of Clusters for Hierarchical Clustering}
\usage{
optimal_hclust_k(hclust_obj, method = "silhouette", max_k = 10)
}
\arguments{
\item{hclust_obj}{A tidy_hclust object}

\item{method}{Character; "silhouette" (default) or "gap"}

\item{max_k}{Maximum number of clusters to test (default: 10)}
}
\value{
A list containing:
\itemize{
\item optimal_k: the recommended number of clusters
\item method: the evaluation method used
\item values: numeric vector of evaluation scores (for silhouette)
\item k_range: integer vector of k values tested (for silhouette)
}
If \code{method = "gap"}, returns a \code{tidy_gap} object instead.
}
\description{
Use silhouette or gap statistic to find optimal k
}
\examples{
\donttest{
hc <- tidy_hclust(USArrests, method = "ward.D2")
opt <- optimal_hclust_k(hc, method = "silhouette", max_k = 6)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{print.tidy_dbscan}
\alias{print.tidy_dbscan}
\title{Print Method for tidy_dbscan}
\usage{
\method{print}{tidy_dbscan}(x, ...)
}
\arguments{
\item{x}{A tidy_dbscan object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_dbscan
}
\examples{
\donttest{
db <- tidy_dbscan(iris[, 1:4], eps = 0.5, minPts = 5)
print(db)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{predict.tidylearn_model}
\alias{predict.tidylearn_model}
\title{Predict using a tidylearn model}
\usage{
\method{predict}{tidylearn_model}(object, new_data = NULL, type = "response", ...)
}
\arguments{
\item{object}{A tidylearn model object}

\item{new_data}{A data frame containing the new data.
If NULL, uses training data.}

\item{type}{Type of prediction. For supervised:
"response" (default), "prob", "class". For
unsupervised: "scores", "clusters", "transform"
depending on method.}

\item{...}{Additional arguments}
}
\value{
A \link[tibble]{tibble} with a \code{.pred} column containing
predictions. For classification with \code{type = "prob"}, returns
columns for each class probability.
}
\description{
Unified prediction interface for both supervised and unsupervised models
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
predict(model)
predict(model, new_data = mtcars[1:5, ])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactions.R
\name{tidylearn-interactions}
\alias{tidylearn-interactions}
\title{Interaction Analysis Functions for tidylearn}
\description{
Functions for testing, visualizing, and analyzing interactions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tidylearn-regularization}
\alias{tidylearn-regularization}
\title{Regularization Functions for tidylearn}
\description{
Ridge, Lasso, and Elastic Net regularization
functionality
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{calc_wss}
\alias{calc_wss}
\title{Calculate Within-Cluster Sum of Squares for Different k}
\usage{
calc_wss(data, max_k = 10, nstart = 25)
}
\arguments{
\item{data}{A data frame or tibble}

\item{max_k}{Maximum number of clusters to test (default: 10)}

\item{nstart}{Number of random starts for each k (default: 25)}
}
\value{
A tibble with columns \code{k} (number of clusters) and
\code{tot_withinss} (total within-cluster sum of squares).
}
\description{
Used for elbow method to determine optimal k
}
\examples{
\donttest{
wss <- calc_wss(iris[, 1:4], max_k = 6)
plot(wss$k, wss$tot_withinss, type = "b")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-deep-learning.R
\name{tidylearn-deep-learning}
\alias{tidylearn-deep-learning}
\title{Deep Learning for tidylearn}
\description{
Deep learning functionality using Keras/TensorFlow
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tl_predict_ridge}
\alias{tl_predict_ridge}
\title{Predict using a Ridge regression model}
\usage{
tl_predict_ridge(model, new_data, type = "response", ...)
}
\arguments{
\item{model}{A tidylearn Ridge model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a Ridge regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_excel}
\alias{tl_read_excel}
\title{Read an Excel file}
\usage{
tl_read_excel(path, sheet = 1, ...)
}
\arguments{
\item{path}{Path to an Excel file.}

\item{sheet}{Sheet to read. Either a string (the name of a sheet) or an
integer (the position of the sheet). Defaults to the first sheet.}

\item{...}{Additional arguments passed to \code{readxl::read_excel()}.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads an Excel file (\code{.xls}, \code{.xlsx}, or \code{.xlsm}) into a
\code{tidylearn_data} object. Requires the \pkg{readxl} package.
}
\examples{
\donttest{
# data <- tl_read_excel("path/to/data.xlsx")
# data <- tl_read_excel("path/to/data.xlsx", sheet = "Sheet2")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{create_cluster_dashboard}
\alias{create_cluster_dashboard}
\title{Create Summary Dashboard}
\usage{
create_cluster_dashboard(
  data,
  cluster_col = "cluster",
  validation_metrics = NULL
)
}
\arguments{
\item{data}{Data frame with cluster assignments}

\item{cluster_col}{Cluster column name}

\item{validation_metrics}{Optional tibble of validation metrics}
}
\value{
Invisibly returns a list of \code{\link[ggplot2]{ggplot}} objects.
The combined plot grid is drawn as a side effect via
\code{\link[gridExtra]{grid.arrange}}.
}
\description{
Generate a multi-panel summary of clustering results
}
\examples{
\donttest{
df <- iris[, 1:4]
df$cluster <- kmeans(df, 3)$cluster
create_cluster_dashboard(df)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{summary.tidylearn_model}
\alias{summary.tidylearn_model}
\title{Summary method for tidylearn models}
\usage{
\method{summary}{tidylearn_model}(object, ...)
}
\arguments{
\item{object}{A tidylearn model object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input \code{object}, returned invisibly. Called for its
side effect of printing model summary and training performance.
}
\description{
Summary method for tidylearn models
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
summary(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{tl_stratified_models}
\alias{tl_stratified_models}
\title{Stratified Features via Clustering}
\usage{
tl_stratified_models(
  data,
  formula,
  cluster_method = "kmeans",
  k = 3,
  supervised_method = "linear",
  ...
)
}
\arguments{
\item{data}{A data frame}

\item{formula}{Model formula}

\item{cluster_method}{Clustering method}

\item{k}{Number of clusters}

\item{supervised_method}{Supervised learning method}

\item{...}{Additional arguments}
}
\value{
A list with class \code{"tidylearn_stratified"} containing:
\describe{
\item{cluster_model}{The fitted clustering model.}
\item{supervised_models}{Named list of tidylearn models, one per
cluster.}
\item{formula}{The model formula.}
\item{data}{The original training data.}
}
}
\description{
Create cluster-specific supervised models for heterogeneous data
}
\examples{
\donttest{
models <- tl_stratified_models(mtcars, mpg ~ ., cluster_method = "kmeans",
                                k = 3, supervised_method = "linear")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_plot_xgboost_tree}
\alias{tl_plot_xgboost_tree}
\title{Plot XGBoost tree visualization}
\usage{
tl_plot_xgboost_tree(model, tree_index = 0, ...)
}
\arguments{
\item{model}{A tidylearn XGBoost model object}

\item{tree_index}{Index of the tree to plot (default: 0, first tree)}

\item{...}{Additional arguments}
}
\value{
The return value of \code{\link[xgboost]{xgb.plot.tree}}, a
tree diagram rendered via the \pkg{DiagrammeR} package.
}
\description{
Plot XGBoost tree visualization
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_mysql}
\alias{tl_read_mysql}
\title{Read from a MySQL/MariaDB database}
\usage{
tl_read_mysql(
  dsn,
  query,
  dbname = NULL,
  user = NULL,
  password = NULL,
  port = 3306,
  ...
)
}
\arguments{
\item{dsn}{A MySQL connection string (e.g.,
\code{"mysql://user:pass@host:port/dbname"}), or the database host if
using named parameters.}

\item{query}{A SQL query string.}

\item{dbname}{Database name (if not in \code{dsn}).}

\item{user}{Username (if not in \code{dsn}).}

\item{password}{Password (if not in \code{dsn}).}

\item{port}{Port number. Default is 3306.}

\item{...}{Additional arguments passed to \code{DBI::dbConnect()}.}
}
\value{
A \code{tidylearn_data} object containing the query results.
}
\description{
Connects to a MySQL or MariaDB database, executes a SQL query, and returns
the result as a \code{tidylearn_data} object. Accepts either a connection
string or individual connection parameters. Requires \pkg{DBI} and
\pkg{RMariaDB}.
}
\examples{
\donttest{
# data <- tl_read_mysql(
#   dsn = "localhost",
#   query = "SELECT * FROM my_table",
#   dbname = "mydb",
#   user = "myuser",
#   password = "mypass"
# )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning.R
\name{tl_default_param_grid}
\alias{tl_default_param_grid}
\title{Create pre-defined parameter grids for common models}
\usage{
tl_default_param_grid(method, size = "medium", is_classification = TRUE)
}
\arguments{
\item{method}{Model method ("tree", "forest",
"boost", "svm", etc.)}

\item{size}{Grid size: "small", "medium", "large"}

\item{is_classification}{Whether the task is
classification or regression}
}
\value{
A named list of parameter values suitable for passing to
\code{\link{tl_tune_grid}} or \code{\link{tl_tune_random}}. Each
element is a numeric or character vector of candidate values for
that hyperparameter.
}
\description{
Create pre-defined parameter grids for common models
}
\examples{
\donttest{
grid <- tl_default_param_grid("tree", size = "small")
grid <- tl_default_param_grid("forest", size = "medium")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_get_importance_regularized}
\alias{tl_get_importance_regularized}
\title{Extract importance from a regularized regression model}
\usage{
tl_get_importance_regularized(model, lambda = "1se")
}
\arguments{
\item{model}{A tidylearn regularized model object}

\item{lambda}{Which lambda to use ("1se" or "min", default: "1se")}
}
\value{
A data frame with feature importance values
}
\description{
Extract importance from a regularized regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{predict.tidylearn_transfer}
\alias{predict.tidylearn_transfer}
\title{Predict with transfer learning model}
\usage{
\method{predict}{tidylearn_transfer}(object, new_data, ...)
}
\arguments{
\item{object}{A tidylearn_transfer model object}

\item{new_data}{New data for predictions}

\item{...}{Additional arguments}
}
\value{
A \link[tibble]{tibble} with a \code{.pred} column containing
predictions.
}
\description{
Predict with transfer learning model
}
\examples{
\donttest{
model <- tl_transfer_learning(iris, Species ~ .,
  pretrain_method = "pca", supervised_method = "logistic")
preds <- predict(model, iris[1:5, ])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{augment_pam}
\alias{augment_pam}
\title{Augment Data with PAM Cluster Assignments}
\usage{
augment_pam(pam_obj, data)
}
\arguments{
\item{pam_obj}{A tidy_pam object}

\item{data}{Original data frame}
}
\value{
A tibble containing the original \code{data} with an additional
\code{cluster} factor column indicating cluster assignments.
}
\description{
Augment Data with PAM Cluster Assignments
}
\examples{
\donttest{
pm <- tidy_pam(iris[, 1:4], k = 3)
augmented <- augment_pam(pm, iris)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_dashboard}
\alias{tl_dashboard}
\title{Create interactive visualization dashboard for a model}
\usage{
tl_dashboard(model, new_data = NULL, ...)
}
\arguments{
\item{model}{A tidylearn model object}

\item{new_data}{Optional data frame for evaluation
(if NULL, uses training data)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[shiny]{shinyApp}} object.
}
\description{
Create interactive visualization dashboard for a model
}
\examples{
\donttest{
if (requireNamespace("shiny")) {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  app <- tl_dashboard(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tidylearn-read}
\alias{tidylearn-read}
\title{Data Reading Functions for tidylearn}
\description{
Functions for reading data from diverse sources into tidy
\code{tidylearn_data} objects. The main dispatcher
\code{tl_read()} auto-detects the format from the file
extension and routes to the appropriate reader.
All readers return a \code{tidylearn_data} object,
which is a tibble subclass carrying metadata about
the data source.
}
\details{
Supported file formats:
\itemize{
\item \strong{CSV}: \code{.csv} files via \pkg{readr}
(with base R fallback)
\item \strong{TSV}: \code{.tsv} files via \pkg{readr}
(with base R fallback)
\item \strong{Excel}: \code{.xls}, \code{.xlsx},
\code{.xlsm} files via \pkg{readxl}
\item \strong{Parquet}: \code{.parquet} files via \pkg{nanoparquet}
\item \strong{JSON}: \code{.json} files via \pkg{jsonlite}
\item \strong{RDS}: \code{.rds} files via base \code{readRDS()}
\item \strong{RData}: \code{.rdata}, \code{.rda}
files via base \code{load()}
}

Supported databases (via \pkg{DBI}):
\itemize{
\item \strong{SQLite}: \code{.sqlite}, \code{.db} files via \pkg{RSQLite}
\item \strong{PostgreSQL}: via \pkg{RPostgres}
\item \strong{MySQL/MariaDB}: via \pkg{RMariaDB}
\item \strong{BigQuery}: via \pkg{bigrquery}
}

Supported cloud/API sources:
\itemize{
\item \strong{S3}: \code{s3://} URIs via \pkg{paws.storage}
\item \strong{GitHub}: raw file download from repositories
\item \strong{Kaggle}: dataset download via Kaggle CLI
}

Multi-file reading:
\itemize{
\item \strong{Multiple paths}: pass a character vector to \code{tl_read()}
\item \strong{Directories}: \code{tl_read_dir()} scans for data files with
optional pattern/format filtering and recursive scanning
\item \strong{Zip archives}: \code{tl_read_zip()} extracts and reads from
\code{.zip} files
}
When combining multiple files, a \code{source_file} column is added to
identify the origin of each row.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_plot_intervals}
\alias{tl_plot_intervals}
\title{Create confidence and prediction interval plots}
\usage{
tl_plot_intervals(model, new_data = NULL, level = 0.95, ...)
}
\arguments{
\item{model}{A tidylearn regression model object}

\item{new_data}{Optional data frame for prediction
(if NULL, uses training data)}

\item{level}{Confidence level (default: 0.95)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Create confidence and prediction interval plots
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt, method = "linear")
tl_plot_intervals(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tl_predict_regularized}
\alias{tl_predict_regularized}
\title{Predict using a regularized regression model}
\usage{
tl_predict_regularized(model, new_data, type = "response", lambda = "1se", ...)
}
\arguments{
\item{model}{A tidylearn regularized model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response" (default),
"class" or "prob" (for classification)}

\item{lambda}{Which lambda to use for prediction
("1se" or "min", default: "1se")}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a regularized regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_plot_importance_comparison}
\alias{tl_plot_importance_comparison}
\title{Plot feature importance across multiple models}
\usage{
tl_plot_importance_comparison(..., top_n = 10, names = NULL)
}
\arguments{
\item{...}{tidylearn model objects to compare}

\item{top_n}{Number of top features to display (default: 10)}

\item{names}{Optional character vector of model names}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot feature importance across multiple models
}
\examples{
\donttest{
m1 <- tl_model(iris, Species ~ ., method = "forest")
m2 <- tl_model(iris, Species ~ ., method = "boost")
tl_plot_importance_comparison(m1, m2, names = c("Forest", "Boost"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tl_fit_regularized}
\alias{tl_fit_regularized}
\title{Fit a regularized regression model}
\usage{
tl_fit_regularized(
  data,
  formula,
  is_classification = FALSE,
  alpha = 0,
  lambda = NULL,
  cv_folds = 5,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{alpha}{Mixing parameter (0 for Ridge, 1 for Lasso,
between 0-1 for Elastic Net)}

\item{lambda}{Regularization parameter (if NULL, uses
cross-validation to select)}

\item{cv_folds}{Number of folds for cross-validation
(default: 5)}

\item{...}{Additional arguments to pass to glmnet()}
}
\value{
A fitted regularized regression model
}
\description{
Fits Ridge, Lasso, or Elastic Net regularization.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_cluster_sizes}
\alias{plot_cluster_sizes}
\title{Plot Cluster Size Distribution}
\usage{
plot_cluster_sizes(clusters, title = "Cluster Size Distribution")
}
\arguments{
\item{clusters}{Vector of cluster assignments}

\item{title}{Plot title (default: "Cluster Size Distribution")}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Create bar plot of cluster sizes
}
\examples{
\donttest{
clusters <- kmeans(iris[, 1:4], 3)$cluster
plot_cluster_sizes(clusters)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tl_diagnostic_dashboard}
\alias{tl_diagnostic_dashboard}
\title{Create a comprehensive diagnostic dashboard}
\usage{
tl_diagnostic_dashboard(
  model,
  include_influence = TRUE,
  include_assumptions = TRUE,
  include_performance = TRUE,
  arrange_plots = "grid"
)
}
\arguments{
\item{model}{A tidylearn model object}

\item{include_influence}{Logical; whether to include influence diagnostics}

\item{include_assumptions}{Logical; whether to include assumption checks}

\item{include_performance}{Logical; whether to include performance metrics}

\item{arrange_plots}{Layout arrangement (e.g., "grid", "row", "column")}
}
\value{
A \code{\link[gridExtra]{grid.arrange}} object (a
\code{\link[grid]{grob}}) containing the arranged diagnostic plots.
}
\description{
Create a comprehensive diagnostic dashboard
}
\examples{
\donttest{
if (requireNamespace("gridExtra")) {
  model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
  tl_diagnostic_dashboard(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{augment_pca}
\alias{augment_pca}
\title{Augment Original Data with PCA Scores}
\usage{
augment_pca(pca_obj, data, n_components = NULL)
}
\arguments{
\item{pca_obj}{A tidy_pca object}

\item{data}{Original data frame}

\item{n_components}{Number of PCs to add (default: all)}
}
\value{
A tibble containing the original \code{data} with additional columns
for each principal component score (named \code{PC1}, \code{PC2}, etc.).
}
\description{
Add PC scores to the original dataset
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
augmented <- augment_pca(pca, USArrests, n_components = 2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactions.R
\name{tl_auto_interactions}
\alias{tl_auto_interactions}
\title{Find important interactions automatically}
\usage{
tl_auto_interactions(
  data,
  formula,
  top_n = 3,
  min_r2_change = 0.01,
  max_p_value = 0.05,
  exclude_vars = NULL
)
}
\arguments{
\item{data}{A data frame containing the data}

\item{formula}{A formula specifying the base model without interactions}

\item{top_n}{Number of top interactions to return}

\item{min_r2_change}{Minimum change in R-squared to consider}

\item{max_p_value}{Maximum p-value for significance}

\item{exclude_vars}{Character vector of variables to exclude
from interaction testing}
}
\value{
A tidylearn model object (class \code{"tidylearn_model"}) fitted
with the top significant interaction terms added to the formula.
The interaction test results and selected interactions are stored as
attributes \code{"interaction_tests"} and
\code{"selected_interactions"}.
}
\description{
Find important interactions automatically
}
\examples{
\donttest{
model <- tl_auto_interactions(mtcars, mpg ~ wt + hp + cyl, top_n = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-mds.R
\name{tidy_mds_classical}
\alias{tidy_mds_classical}
\title{Classical (Metric) MDS}
\usage{
tidy_mds_classical(dist_mat, ndim = 2, add_rownames = TRUE)
}
\arguments{
\item{dist_mat}{A distance matrix (dist object)}

\item{ndim}{Number of dimensions (default: 2)}

\item{add_rownames}{Preserve row names from distance matrix (default: TRUE)}
}
\value{
A list of class \code{"tidy_mds"} containing:
\itemize{
\item config: tibble of MDS coordinates
\item stress: \code{NA} (not applicable for classical MDS)
\item gof: goodness-of-fit (proportion of variance retained)
\item eigenvalues: numeric vector of eigenvalues
\item method: \code{"Classical MDS"}
\item model: the \code{\link[stats]{cmdscale}} result
}
}
\description{
Performs classical multidimensional scaling using cmdscale()
}
\examples{
\donttest{
d <- dist(USArrests)
mds <- tidy_mds_classical(d)
print(mds)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{tidy_hclust}
\alias{tidy_hclust}
\title{Tidy Hierarchical Clustering}
\usage{
tidy_hclust(data, method = "average", distance = "euclidean", cols = NULL)
}
\arguments{
\item{data}{A data frame, tibble, or dist object}

\item{method}{Agglomeration method: "ward.D2",
"single", "complete", "average" (default),
"mcquitty", "median", "centroid"}

\item{distance}{Distance metric if data is not a
dist object (default: "euclidean")}

\item{cols}{Columns to include (tidy select).
If NULL, uses all numeric columns.}
}
\value{
A list of class "tidy_hclust" containing:
\itemize{
\item model: hclust object
\item dist: distance matrix used
\item method: linkage method used
\item data: original data (for plotting)
}
}
\description{
Performs hierarchical clustering with tidy output
}
\examples{
# Basic hierarchical clustering
hc_result <- tidy_hclust(USArrests, method = "average")

# With specific distance
hc_result <- tidy_hclust(mtcars, method = "complete", distance = "manhattan")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tl_cv}
\alias{tl_cv}
\title{Cross-validation for tidylearn models}
\usage{
tl_cv(data, formula, method, folds = 5, ...)
}
\arguments{
\item{data}{Data frame}

\item{formula}{Model formula}

\item{method}{Modeling method}

\item{folds}{Number of cross-validation folds}

\item{...}{Additional arguments}
}
\value{
A list with two elements:
\describe{
\item{\code{$folds}}{A list of per-fold evaluation
\link[tibble]{tibble}s, each with \code{metric} and
\code{value} columns.}
\item{\code{$summary}}{A \link[tibble]{tibble} with columns
\code{metric}, \code{mean}, and \code{sd} summarizing
performance across folds.}
}
}
\description{
Cross-validation for tidylearn models
}
\examples{
\donttest{
cv <- tl_cv(mtcars, mpg ~ wt + hp, method = "linear", folds = 3)
cv$summary
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_fit_boost}
\alias{tl_fit_boost}
\title{Fit a gradient boosting model}
\usage{
tl_fit_boost(
  data,
  formula,
  is_classification = FALSE,
  n.trees = 100,
  interaction.depth = 3,
  shrinkage = 0.1,
  n.minobsinnode = 10,
  cv.folds = 0,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{n.trees}{Number of trees (default: 100)}

\item{interaction.depth}{Depth of interactions
(default: 3)}

\item{shrinkage}{Learning rate (default: 0.1)}

\item{n.minobsinnode}{Minimum number of observations
in terminal nodes (default: 10)}

\item{cv.folds}{Number of cross-validation folds
(default: 0, no CV)}

\item{...}{Additional arguments to pass to gbm()}
}
\value{
A fitted gradient boosting model
}
\description{
Fit a gradient boosting model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_run_pipeline}
\alias{tl_run_pipeline}
\title{Run a tidylearn pipeline}
\usage{
tl_run_pipeline(pipeline, verbose = TRUE)
}
\arguments{
\item{pipeline}{A tidylearn pipeline object}

\item{verbose}{Logical; whether to print progress}
}
\value{
The input \code{tidylearn_pipeline} object with its
\code{$results} component populated. Results include
\code{$processed_data}, \code{$model_results} (a named list of
per-model fits and metrics), \code{$best_model_name},
\code{$best_model} (the winning \code{tidylearn_model}), and
\code{$metric_values}.
}
\description{
Run a tidylearn pipeline
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .,
  models = list(tree = list(method = "tree")),
  evaluation = list(metrics = "accuracy", validation = "cv",
    cv_folds = 2, best_metric = "accuracy"))
pipe <- tl_run_pipeline(pipe, verbose = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tl_calc_classification_metrics}
\alias{tl_calc_classification_metrics}
\title{Calculate classification metrics}
\usage{
tl_calc_classification_metrics(
  actuals,
  predicted,
  predicted_probs = NULL,
  metrics = c("accuracy", "precision", "recall", "f1", "auc"),
  thresholds = NULL,
  ...
)
}
\arguments{
\item{actuals}{Actual values (ground truth)}

\item{predicted}{Predicted class values}

\item{predicted_probs}{Predicted probabilities (for metrics like AUC)}

\item{metrics}{Character vector of metrics to compute}

\item{thresholds}{Optional vector of thresholds to evaluate
for threshold-dependent metrics}

\item{...}{Additional arguments}
}
\value{
A \link[tibble]{tibble} with columns \code{metric} (character)
and \code{value} (numeric) containing the requested classification
metrics. When \code{thresholds} are supplied, additional rows are
appended with threshold-specific metric names.
}
\description{
Calculate classification metrics
}
\examples{
\donttest{
model <- tl_model(iris, Species ~ ., method = "forest")
preds <- predict(model)
tl_calc_classification_metrics(iris$Species, preds$.pred)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{augment_hclust}
\alias{augment_hclust}
\title{Augment Data with Hierarchical Cluster Assignments}
\usage{
augment_hclust(hclust_obj, data, k = NULL, h = NULL)
}
\arguments{
\item{hclust_obj}{A tidy_hclust object}

\item{data}{Original data frame}

\item{k}{Number of clusters (optional)}

\item{h}{Height at which to cut (optional)}
}
\value{
A tibble containing the original \code{data} with an additional
\code{cluster} integer column indicating cluster assignments.
}
\description{
Add cluster assignments to original data
}
\examples{
\donttest{
hc <- tidy_hclust(USArrests, method = "ward.D2")
augmented <- augment_hclust(hc, USArrests, k = 3)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regression.R
\name{tl_plot_actual_predicted}
\alias{tl_plot_actual_predicted}
\title{Plot actual vs predicted values for a regression model}
\usage{
tl_plot_actual_predicted(model, new_data = NULL, ...)
}
\arguments{
\item{model}{A tidylearn regression model object}

\item{new_data}{Optional data frame for evaluation
(if NULL, uses training data)}

\item{...}{Additional arguments}
}
\value{
A ggplot object
}
\description{
Plot actual vs predicted values for a regression model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tl_influence_measures}
\alias{tl_influence_measures}
\title{Calculate influence measures for a linear model}
\usage{
tl_influence_measures(
  model,
  threshold_cook = NULL,
  threshold_leverage = NULL,
  threshold_dffits = NULL
)
}
\arguments{
\item{model}{A tidylearn model object}

\item{threshold_cook}{Cook's distance threshold (default: 4/n)}

\item{threshold_leverage}{Leverage threshold (default: 2*(p+1)/n)}

\item{threshold_dffits}{DFFITS threshold (default: 2*sqrt((p+1)/n))}
}
\value{
A data frame with one row per observation containing influence
measures: \code{cooks_distance}, \code{leverage}, \code{dffits},
\code{std_residual}, \code{stud_residual}, boolean flags for each
threshold (\code{is_cook_influential}, \code{is_leverage_influential},
\code{is_dffits_influential}, \code{is_outlier}), per-coefficient
\code{dfbetas_*} columns, and an overall \code{is_influential} flag.
Threshold values are stored as attributes.
}
\description{
Calculate influence measures for a linear model
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_influence_measures(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preprocessing.R
\name{tl_prepare_data}
\alias{tl_prepare_data}
\title{Data Preprocessing for tidylearn}
\usage{
tl_prepare_data(
  data,
  formula = NULL,
  impute_method = "mean",
  scale_method = "standardize",
  encode_categorical = TRUE,
  remove_zero_variance = TRUE,
  remove_correlated = FALSE,
  correlation_cutoff = 0.95
)
}
\arguments{
\item{data}{A data frame}

\item{formula}{Optional formula (for supervised learning)}

\item{impute_method}{Method for missing value imputation:
"mean", "median", "mode", "knn"}

\item{scale_method}{Scaling method: "standardize",
"normalize", "robust", "none"}

\item{encode_categorical}{Whether to encode categorical
variables (default: TRUE)}

\item{remove_zero_variance}{Remove zero-variance features (default: TRUE)}

\item{remove_correlated}{Remove highly correlated features (default: FALSE)}

\item{correlation_cutoff}{Correlation threshold for removal (default: 0.95)}
}
\value{
A list with components:
\describe{
\item{\code{data}}{The processed data frame.}
\item{\code{original_data}}{The original unprocessed data frame.}
\item{\code{preprocessing_steps}}{A list of metadata for each
preprocessing step applied (imputation values, encoding maps,
scaling parameters, etc.).}
\item{\code{formula}}{The formula passed in (or \code{NULL}).}
}
}
\description{
Unified preprocessing functions that work with both
supervised and unsupervised workflows
Prepare Data for Machine Learning
}
\details{
Comprehensive preprocessing pipeline including imputation, scaling,
encoding, and feature engineering
}
\examples{
\donttest{
processed <- tl_prepare_data(iris, Species ~ ., scale_method = "standardize")
model <- tl_model(processed$data, Species ~ ., method = "tree")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{tidy_clara}
\alias{tidy_clara}
\title{Tidy CLARA (Clustering Large Applications)}
\usage{
tidy_clara(data, k, metric = "euclidean", samples = 50, sampsize = NULL)
}
\arguments{
\item{data}{A data frame or tibble}

\item{k}{Number of clusters}

\item{metric}{Distance metric (default: "euclidean")}

\item{samples}{Number of samples to draw (default: 50)}

\item{sampsize}{Sample size (default: min(n, 40 + 2*k))}
}
\value{
A list of class \code{"tidy_clara"} containing:
\itemize{
\item clusters: tibble with observation IDs and cluster assignments
\item medoids: tibble of medoid values
\item silhouette_avg: average silhouette width
\item model: original \code{\link[cluster]{clara}} object
}
}
\description{
Performs CLARA clustering (scalable version of PAM)
}
\examples{
\donttest{
# CLARA for large datasets
large_data <- iris[rep(1:nrow(iris), 10), 1:4]
clara_result <- tidy_clara(large_data, k = 3, samples = 50)
print(clara_result)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_plot_partial_dependence}
\alias{tl_plot_partial_dependence}
\title{Plot partial dependence for tree-based models}
\usage{
tl_plot_partial_dependence(model, var, n.pts = 20, ...)
}
\arguments{
\item{model}{A tidylearn tree-based model object}

\item{var}{Variable name to plot}

\item{n.pts}{Number of points for continuous
variables (default: 20)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot partial dependence for tree-based models
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ ., method = "forest")
tl_plot_partial_dependence(model, var = "wt")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-selection.R
\name{tl_compare_cv}
\alias{tl_compare_cv}
\title{Compare models using cross-validation}
\usage{
tl_compare_cv(data, models, folds = 5, metrics = NULL, ...)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{models}{A list of tidylearn model objects}

\item{folds}{Number of cross-validation folds}

\item{metrics}{Character vector of metrics to compute}

\item{...}{Additional arguments}
}
\value{
A list with two elements:
\describe{
\item{\code{$fold_metrics}}{A data frame with columns
\code{metric}, \code{value}, \code{fold}, and \code{model}
containing per-fold results for every model.}
\item{\code{$summary}}{A data frame with columns \code{model},
\code{metric}, \code{mean_value}, \code{sd_value},
\code{min_value}, and \code{max_value} summarizing
cross-validation performance.}
}
}
\description{
Compare models using cross-validation
}
\examples{
\donttest{
m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
cv$summary
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tl_fit_forest}
\alias{tl_fit_forest}
\title{Fit a random forest model}
\usage{
tl_fit_forest(
  data,
  formula,
  is_classification = FALSE,
  ntree = 500,
  mtry = NULL,
  importance = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{ntree}{Number of trees to grow (default: 500)}

\item{mtry}{Number of variables randomly sampled at each split}

\item{importance}{Whether to compute variable importance (default: TRUE)}

\item{...}{Additional arguments to pass to randomForest()}
}
\value{
A fitted random forest model
}
\description{
Fit a random forest model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_bigquery}
\alias{tl_read_bigquery}
\title{Read from Google BigQuery}
\usage{
tl_read_bigquery(project, query, dataset = NULL, ...)
}
\arguments{
\item{project}{Google Cloud project ID.}

\item{query}{A SQL query string (Standard SQL).}

\item{dataset}{Optional default dataset for unqualified table names.}

\item{...}{Additional arguments passed to
\code{bigrquery::bq_project_query()}.}
}
\value{
A \code{tidylearn_data} object containing the query results.
}
\description{
Executes a SQL query against Google BigQuery and returns the result as a
\code{tidylearn_data} object. Requires the \pkg{bigrquery} package and
valid Google Cloud authentication.
}
\examples{
\donttest{
# data <- tl_read_bigquery(
#   project = "my-project",
#   query = "SELECT * FROM `my_dataset.my_table` LIMIT 1000"
# )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tl_fit_nn}
\alias{tl_fit_nn}
\title{Fit a neural network model}
\usage{
tl_fit_nn(
  data,
  formula,
  is_classification = FALSE,
  size = 5,
  decay = 0,
  maxit = 100,
  trace = FALSE,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{size}{Number of units in the hidden layer (default: 5)}

\item{decay}{Weight decay parameter (default: 0)}

\item{maxit}{Maximum number of iterations (default: 100)}

\item{trace}{Logical; whether to print progress (default: FALSE)}

\item{...}{Additional arguments to pass to nnet()}
}
\value{
A fitted neural network model
}
\description{
Fit a neural network model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{tl_plot_lift}
\alias{tl_plot_lift}
\title{Plot lift chart for a classification model}
\usage{
tl_plot_lift(model, new_data = NULL, bins = 10, ...)
}
\arguments{
\item{model}{A tidylearn classification model object}

\item{new_data}{Optional data frame for evaluation
(if NULL, uses training data)}

\item{bins}{Number of bins for grouping predictions
(default: 10)}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot lift chart for a classification model
}
\examples{
\donttest{
iris_bin <- iris[iris$Species != "setosa", ]
iris_bin$Species <- factor(iris_bin$Species)
model <- tl_model(iris_bin, Species ~ ., method = "logistic")
tl_plot_lift(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/integration.R
\name{tl_anomaly_aware}
\alias{tl_anomaly_aware}
\title{Anomaly-Aware Supervised Learning}
\usage{
tl_anomaly_aware(
  data,
  formula,
  response,
  anomaly_method = "dbscan",
  action = "flag",
  supervised_method = "logistic",
  ...
)
}
\arguments{
\item{data}{A data frame}

\item{formula}{Model formula}

\item{response}{Response variable name}

\item{anomaly_method}{Method for anomaly detection:
"dbscan", "isolation_forest"}

\item{action}{Action to take: "remove", "flag", "downweight"}

\item{supervised_method}{Supervised learning method}

\item{...}{Additional arguments}
}
\value{
A tidylearn model object with additional class
\code{"tidylearn_anomaly_aware"}. The model includes an
\code{anomaly_info} element with \code{anomaly_model},
\code{is_anomaly} (logical vector), \code{n_anomalies}, and
\code{action}.
}
\description{
Detect outliers using DBSCAN or other methods, then optionally
remove them or down-weight them before supervised learning.
}
\examples{
\donttest{
model <- tl_anomaly_aware(iris, Species ~ ., response = "Species",
                           anomaly_method = "dbscan", action = "flag")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{tidy_silhouette_analysis}
\alias{tidy_silhouette_analysis}
\title{Silhouette Analysis Across Multiple k Values}
\usage{
tidy_silhouette_analysis(
  data,
  max_k = 10,
  method = "kmeans",
  nstart = 25,
  dist_method = "euclidean",
  linkage_method = "average"
)
}
\arguments{
\item{data}{A data frame or tibble}

\item{max_k}{Maximum number of clusters to test (default: 10)}

\item{method}{Clustering method: "kmeans" (default) or "hclust"}

\item{nstart}{If kmeans, number of random starts (default: 25)}

\item{dist_method}{Distance metric (default: "euclidean")}

\item{linkage_method}{If hclust, linkage method (default: "average")}
}
\value{
A tibble with columns \code{k} and \code{avg_sil_width}. The
\code{"optimal_k"} attribute contains the k with the highest average
silhouette width.
}
\description{
Silhouette Analysis Across Multiple k Values
}
\examples{
\donttest{
sil_analysis <- tidy_silhouette_analysis(iris[, 1:4], max_k = 6)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tidylearn-metrics}
\alias{tidylearn-metrics}
\title{Metrics Functionality for tidylearn}
\description{
Functions for calculating model evaluation metrics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{tl_explore}
\alias{tl_explore}
\title{Exploratory Data Analysis Workflow}
\usage{
tl_explore(data, response = NULL, max_components = 5, k_range = 2:6)
}
\arguments{
\item{data}{A data frame}

\item{response}{Optional response variable for colored visualizations}

\item{max_components}{Maximum PCA components to compute (default: 5)}

\item{k_range}{Range of k values for clustering (default: 2:6)}
}
\value{
A list with class \code{"tidylearn_eda"} containing:
\describe{
\item{data}{The original data frame.}
\item{response}{The response variable name, or \code{NULL}.}
\item{pca}{The fitted PCA model.}
\item{optimal_k}{List with optimal cluster count results.}
\item{kmeans}{The fitted k-means model.}
\item{hclust}{The fitted hierarchical clustering model.}
\item{summary}{List with \code{n_obs}, \code{n_vars},
\code{n_components}, and \code{best_k}.}
}
}
\description{
Comprehensive EDA combining unsupervised learning techniques
to understand data structure before modeling
}
\examples{
\donttest{
eda <- tl_explore(iris, response = "Species")
plot(eda)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model-selection.R
\name{tl_plot_cv_comparison}
\alias{tl_plot_cv_comparison}
\title{Plot comparison of cross-validation results}
\usage{
tl_plot_cv_comparison(cv_results, metrics = NULL)
}
\arguments{
\item{cv_results}{Results from tl_compare_cv function}

\item{metrics}{Character vector of metrics to plot
(if NULL, plots all metrics)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object showing boxplots of
cross-validation metric distributions for each model.
}
\description{
Plot comparison of cross-validation results
}
\examples{
\donttest{
m1 <- tl_model(mtcars, mpg ~ wt, method = "linear")
m2 <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
cv <- tl_compare_cv(mtcars, list(simple = m1, full = m2), folds = 3)
tl_plot_cv_comparison(cv)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-pca.R
\name{get_pca_loadings}
\alias{get_pca_loadings}
\title{Get PCA Loadings in Wide Format}
\usage{
get_pca_loadings(pca_obj, n_components = NULL)
}
\arguments{
\item{pca_obj}{A tidy_pca object}

\item{n_components}{Number of components to include (default: all)}
}
\value{
A tibble with one row per variable and one column per principal
component, containing the loading values.
}
\description{
Get PCA Loadings in Wide Format
}
\examples{
\donttest{
pca <- tidy_pca(USArrests)
get_pca_loadings(pca, n_components = 2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactions.R
\name{tl_plot_interaction}
\alias{tl_plot_interaction}
\title{Plot interaction effects}
\usage{
tl_plot_interaction(
  model,
  var1,
  var2,
  n_points = 100,
  fixed_values = NULL,
  confidence = TRUE,
  ...
)
}
\arguments{
\item{model}{A tidylearn model object}

\item{var1}{First variable in the interaction}

\item{var2}{Second variable in the interaction}

\item{n_points}{Number of points to use for continuous variables}

\item{fixed_values}{Named list of values for other variables in the model}

\item{confidence}{Logical; whether to show confidence intervals}

\item{...}{Additional arguments to pass to predict()}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot interaction effects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-deep-learning.R
\name{tl_plot_deep_history}
\alias{tl_plot_deep_history}
\title{Plot deep learning model training history}
\usage{
tl_plot_deep_history(model, metrics = c("loss", "val_loss"), ...)
}
\arguments{
\item{model}{A tidylearn deep learning model object}

\item{metrics}{Which metrics to plot
(default: c("loss", "val_loss"))}

\item{...}{Additional arguments}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot deep learning model training history
}
\examples{
\dontrun{
if (requireNamespace("keras", quietly = TRUE)) {
  model <- tl_model(iris, Species ~ ., method = "deep", epochs = 5)
  tl_plot_deep_history(model)
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{print.tidy_silhouette}
\alias{print.tidy_silhouette}
\title{Print Method for tidy_silhouette}
\usage{
\method{print}{tidy_silhouette}(x, ...)
}
\arguments{
\item{x}{A tidy_silhouette object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print Method for tidy_silhouette
}
\examples{
\donttest{
km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
d <- dist(iris[, 1:4])
sil <- tidy_silhouette(km$cluster, d)
print(sil)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_importance}
\alias{tl_table_importance}
\title{Formatted feature importance table}
\usage{
tl_table_importance(model, top_n = 20, digits = 2, ...)
}
\arguments{
\item{model}{A tidylearn model object}

\item{top_n}{Maximum number of features to display (default: 20)}

\item{digits}{Number of decimal places (default: 2)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt table of feature importance with a colour gradient.
Supports tree-based, regularised, and xgboost models.
}
\examples{
\donttest{
model <- tl_model(iris, Species ~ ., method = "forest")
tl_table_importance(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{plot_gap_stat}
\alias{plot_gap_stat}
\title{Plot Gap Statistic}
\usage{
plot_gap_stat(gap_obj, show_methods = FALSE)
}
\arguments{
\item{gap_obj}{A tidy_gap object}

\item{show_methods}{Logical; show all three k
selection methods? (default: FALSE)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot Gap Statistic
}
\examples{
\donttest{
gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
plot_gap_stat(gap)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read.R
\name{tl_read_rdata}
\alias{tl_read_rdata}
\title{Read an RData file}
\usage{
tl_read_rdata(path, name = NULL, ...)
}
\arguments{
\item{path}{Path to an RData file.}

\item{name}{Optional name of the object to extract from the RData file. If
\code{NULL} (default), the function returns the first data frame found, or
errors if there are multiple data frames.}

\item{...}{Currently unused.}
}
\value{
A \code{tidylearn_data} object (a \link[tibble]{tibble} subclass)
with attributes \code{tl_source}, \code{tl_format}, and
\code{tl_timestamp}.
}
\description{
Reads an RData (\code{.rdata} or \code{.rda}) file
into a \code{tidylearn_data} object. Since RData files
can contain multiple objects, use the \code{name}
argument to specify which object to extract.
If \code{name} is \code{NULL} and
the file contains exactly one data frame, it is returned automatically.
}
\examples{
\donttest{
# data <- tl_read_rdata("path/to/data.rdata")
# data <- tl_read_rdata("path/to/data.rdata", name = "my_data")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_github}
\alias{tl_read_github}
\title{Read from GitHub}
\usage{
tl_read_github(source, path = NULL, ref = "main", ...)
}
\arguments{
\item{source}{A GitHub URL or \code{"owner/repo"} string.}

\item{path}{Path to the file within the repository (required when
\code{source} is \code{"owner/repo"} format).}

\item{ref}{Branch, tag, or commit SHA. Default is \code{"main"}.}

\item{...}{Additional arguments passed to the format-specific reader.}
}
\value{
A \code{tidylearn_data} object containing the downloaded data.
}
\description{
Downloads a raw file from a GitHub repository and reads it into a
\code{tidylearn_data} object. Accepts either a full GitHub URL or a
\code{owner/repo} shorthand with a file path.
}
\examples{
\donttest{
# data <- tl_read_github("user/repo", path = "data/file.csv")
# data <- tl_read_github(
#   "https://github.com/user/repo/blob/main/data/file.csv"
# )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{plot.tidylearn_model}
\alias{plot.tidylearn_model}
\title{Plot method for tidylearn models}
\usage{
\method{plot}{tidylearn_model}(x, type = "auto", ...)
}
\arguments{
\item{x}{A tidylearn model object}

\item{type}{Plot type (default: "auto")}

\item{...}{Additional arguments passed to plotting functions}
}
\value{
A \code{\link[ggplot2]{ggplot}} object. The specific plot depends
on the model paradigm and \code{type} argument.
}
\description{
Plot method for tidylearn models
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
plot(model, type = "actual_predicted")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core.R
\name{tl_plot_model}
\alias{tl_plot_model}
\title{Plot a supervised tidylearn model}
\usage{
tl_plot_model(model, type = "auto", ...)
}
\arguments{
\item{model}{A tidylearn supervised model object}

\item{type}{Plot type. For regression: "auto", "actual_predicted",
"residuals", "diagnostics". For classification: "auto", "confusion",
"roc", "precision_recall", "calibration", "lift", "gain".
"importance" is available for tree-based and regularized models.}

\item{...}{Additional arguments passed to the underlying plot function}
}
\value{
A ggplot2 object (invisibly for base-graphics plots)
}
\description{
Dispatches to the appropriate plotting function based on model type and
requested plot type.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diagnostics.R
\name{tl_plot_influence}
\alias{tl_plot_influence}
\title{Plot influence diagnostics}
\usage{
tl_plot_influence(
  model,
  plot_type = "cook",
  threshold_cook = NULL,
  threshold_leverage = NULL,
  threshold_dffits = NULL,
  n_labels = 3,
  label_size = 3
)
}
\arguments{
\item{model}{A tidylearn model object}

\item{plot_type}{Type of influence plot: "cook", "leverage", "index"}

\item{threshold_cook}{Cook's distance threshold (default: 4/n)}

\item{threshold_leverage}{Leverage threshold (default: 2*(p+1)/n)}

\item{threshold_dffits}{DFFITS threshold (default: 2*sqrt((p+1)/n))}

\item{n_labels}{Number of points to label (default: 3)}

\item{label_size}{Text size for labels (default: 3)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Plot influence diagnostics
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_plot_influence(model, plot_type = "cook")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{plot.tidylearn_eda}
\alias{plot.tidylearn_eda}
\title{Plot EDA results}
\usage{
\method{plot}{tidylearn_eda}(x, ...)
}
\arguments{
\item{x}{A tidylearn_eda object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly. Called for its
side effect of plotting a PCA scatter plot coloured by cluster.
}
\description{
Plot EDA results
}
\examples{
\donttest{
eda <- tl_explore(iris, response = "Species")
plot(eda)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_cluster_comparison}
\alias{plot_cluster_comparison}
\title{Create Cluster Comparison Plot}
\usage{
plot_cluster_comparison(data, cluster_cols, x_col, y_col)
}
\arguments{
\item{data}{Data frame with multiple cluster columns}

\item{cluster_cols}{Vector of cluster column names}

\item{x_col}{X-axis variable}

\item{y_col}{Y-axis variable}
}
\value{
The return value of \code{\link[gridExtra]{grid.arrange}}, a
\code{\link[gtable]{gtable}} drawn as a side effect.
}
\description{
Compare multiple clustering results side-by-side
}
\examples{
\donttest{
df <- iris[, 1:4]
df$km3 <- kmeans(df, 3)$cluster
df$km4 <- kmeans(df, 4)$cluster
plot_cluster_comparison(df, c("km3", "km4"), "Sepal.Length", "Sepal.Width")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{tidy_apriori}
\alias{tidy_apriori}
\title{Tidy Apriori Algorithm}
\usage{
tidy_apriori(
  transactions,
  support = 0.01,
  confidence = 0.5,
  minlen = 2,
  maxlen = 10,
  target = "rules"
)
}
\arguments{
\item{transactions}{A transactions object or data frame}

\item{support}{Minimum support (default: 0.01)}

\item{confidence}{Minimum confidence (default: 0.5)}

\item{minlen}{Minimum rule length (default: 2)}

\item{maxlen}{Maximum rule length (default: 10)}

\item{target}{Type of association mined: "rules"
(default), "frequent itemsets",
"maximally frequent itemsets"}
}
\value{
A list of class "tidy_rules" containing:
\itemize{
\item rules_tbl: tibble of rules with lhs, rhs, and quality measures
\item rules: original rules object
\item parameters: parameters used
}
}
\description{
Mine association rules using the Apriori algorithm with tidy output
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
data("Groceries", package = "arules")

# Basic apriori
rules <- tidy_apriori(Groceries, support = 0.001, confidence = 0.5)

# Access rules
rules$rules_tbl
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-deep-learning.R
\name{tl_fit_deep}
\alias{tl_fit_deep}
\title{Fit a deep learning model}
\usage{
tl_fit_deep(
  data,
  formula,
  is_classification = FALSE,
  hidden_layers = c(32, 16),
  activation = "relu",
  dropout = 0.2,
  epochs = 30,
  batch_size = 32,
  validation_split = 0.2,
  verbose = 0,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{hidden_layers}{Vector of units in each hidden layer
(default: c(32, 16))}

\item{activation}{Activation function for hidden layers
(default: "relu")}

\item{dropout}{Dropout rate for regularization (default: 0.2)}

\item{epochs}{Number of training epochs (default: 30)}

\item{batch_size}{Batch size for training (default: 32)}

\item{validation_split}{Proportion of data for validation
(default: 0.2)}

\item{verbose}{Verbosity mode (0 = silent, 1 = progress bar,
2 = one line per epoch) (default: 0)}

\item{...}{Additional arguments}
}
\value{
A fitted deep learning model
}
\description{
Fit a deep learning model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pipeline.R
\name{tl_get_best_model}
\alias{tl_get_best_model}
\title{Get the best model from a pipeline}
\usage{
tl_get_best_model(pipeline)
}
\arguments{
\item{pipeline}{A tidylearn pipeline object with results}
}
\value{
The best \code{tidylearn_model} object from the pipeline,
selected by the metric specified in \code{evaluation$best_metric}.
}
\description{
Get the best model from a pipeline
}
\examples{
\donttest{
pipe <- tl_pipeline(iris, Species ~ .,
  models = list(tree = list(method = "tree")),
  evaluation = list(metrics = "accuracy", validation = "cv",
    cv_folds = 2, best_metric = "accuracy"))
pipe <- tl_run_pipeline(pipe, verbose = FALSE)
best <- tl_get_best_model(pipe)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interactions.R
\name{tl_interaction_effects}
\alias{tl_interaction_effects}
\title{Calculate partial effects based on a model with interactions}
\usage{
tl_interaction_effects(model, var, by_var, at_values = NULL, intervals = TRUE)
}
\arguments{
\item{model}{A tidylearn model object}

\item{var}{Variable to calculate effects for}

\item{by_var}{Variable to calculate effects by (interaction variable)}

\item{at_values}{Named list of values at which to hold other variables}

\item{intervals}{Logical; whether to include confidence intervals}
}
\value{
For numeric \code{var}: a list with \code{effects} (data frame of
predicted values across the variable range for each level of
\code{by_var}) and \code{slopes} (data frame with estimated slopes and
standard errors per level). For categorical \code{var}: a data frame of
predicted values at each factor level for each level of \code{by_var}.
}
\description{
Calculate partial effects based on a model with interactions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{print.tidylearn_automl}
\alias{print.tidylearn_automl}
\title{Print auto ML results}
\usage{
\method{print}{tidylearn_automl}(x, ...)
}
\arguments{
\item{x}{A tidylearn_automl object}

\item{...}{Additional arguments (ignored)}
}
\value{
The input object \code{x}, returned invisibly.
}
\description{
Print auto ML results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-neural-networks.R
\name{tl_predict_nn}
\alias{tl_predict_nn}
\title{Predict using a neural network model}
\usage{
tl_predict_nn(model, new_data, type = "response", ...)
}
\arguments{
\item{model}{A tidylearn neural network model object}

\item{new_data}{A data frame containing the new data}

\item{type}{Type of prediction: "response" (default),
"prob" (for classification), "class" (for classification)}

\item{...}{Additional arguments}
}
\value{
Predictions
}
\description{
Predict using a neural network model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_kaggle}
\alias{tl_read_kaggle}
\title{Read from Kaggle}
\usage{
tl_read_kaggle(source, file = NULL, dest = tempdir(), type = "dataset", ...)
}
\arguments{
\item{source}{A Kaggle dataset slug (e.g., \code{"user/dataset-name"}) or a
Kaggle URL.}

\item{file}{The specific file to read from the dataset. If \code{NULL} and
the dataset contains exactly one file, it is read automatically.}

\item{dest}{Directory to download files to. Default is a temporary directory.}

\item{type}{Either \code{"dataset"} (default) or \code{"competition"}.}

\item{...}{Additional arguments passed to the format-specific reader.}
}
\value{
A \code{tidylearn_data} object containing the downloaded data.
}
\description{
Downloads a dataset file from Kaggle using the Kaggle CLI and reads it into
a \code{tidylearn_data} object. Requires the Kaggle CLI to be installed and
configured (\code{pip install kaggle}).
}
\examples{
\donttest{
# data <- tl_read_kaggle("zillow/zecon", file = "Zip_time_series.csv")
# data <- tl_read_kaggle("titanic", file = "train.csv", type = "competition")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R
\name{tl_evaluate}
\alias{tl_evaluate}
\title{Evaluate a tidylearn model}
\usage{
tl_evaluate(object, new_data = NULL, ...)
}
\arguments{
\item{object}{A tidylearn model object}

\item{new_data}{Optional new data for evaluation
(if NULL, uses training data)}

\item{...}{Additional arguments}
}
\value{
A \link[tibble]{tibble} with columns \code{metric} (character)
and \code{value} (numeric). For regression models, includes
\code{rmse}, \code{mae}, and \code{rsq}. For classification models,
includes \code{accuracy}.
}
\description{
Evaluate a tidylearn model
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ wt + hp, method = "linear")
tl_evaluate(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-xgboost.R
\name{tl_fit_xgboost}
\alias{tl_fit_xgboost}
\title{Fit an XGBoost model}
\usage{
tl_fit_xgboost(
  data,
  formula,
  is_classification = FALSE,
  nrounds = 100,
  max_depth = 6,
  eta = 0.3,
  subsample = 1,
  colsample_bytree = 1,
  min_child_weight = 1,
  gamma = 0,
  alpha = 0,
  lambda = 1,
  early_stopping_rounds = NULL,
  nthread = NULL,
  verbose = 0,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{is_classification}{Logical indicating if this is a
classification problem}

\item{nrounds}{Number of boosting rounds (default: 100)}

\item{max_depth}{Maximum depth of trees (default: 6)}

\item{eta}{Learning rate (default: 0.3)}

\item{subsample}{Subsample ratio of observations (default: 1)}

\item{colsample_bytree}{Subsample ratio of columns (default: 1)}

\item{min_child_weight}{Minimum sum of instance weight
needed in a child (default: 1)}

\item{gamma}{Minimum loss reduction to make a further partition (default: 0)}

\item{alpha}{L1 regularization term (default: 0)}

\item{lambda}{L2 regularization term (default: 1)}

\item{early_stopping_rounds}{Early stopping rounds (default: NULL)}

\item{nthread}{Number of threads (default: max available)}

\item{verbose}{Verbose output (default: 0)}

\item{...}{Additional arguments to pass to xgb.train()}
}
\value{
A fitted XGBoost model
}
\description{
Fit an XGBoost model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-regularization.R
\name{tl_plot_regularization_cv}
\alias{tl_plot_regularization_cv}
\title{Plot cross-validation results for a regularized model}
\usage{
tl_plot_regularization_cv(model, ...)
}
\arguments{
\item{model}{A tidylearn regularized model object
(ridge, lasso, or elastic_net)}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Shows the cross-validation error as a function of
lambda for ridge, lasso, or elastic net models fitted
with cv.glmnet.
}
\examples{
\donttest{
model <- tl_model(mtcars, mpg ~ ., method = "ridge")
tl_plot_regularization_cv(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tables.R
\name{tl_table_confusion}
\alias{tl_table_confusion}
\title{Formatted confusion matrix table}
\usage{
tl_table_confusion(model, new_data = NULL, ...)
}
\arguments{
\item{model}{A tidylearn classification model}

\item{new_data}{Optional test data. If NULL, uses training data.}

\item{...}{Additional arguments (currently unused)}
}
\value{
A \code{\link[gt]{gt}} table object.
}
\description{
Produces a styled gt confusion matrix with correct predictions highlighted.
Only available for classification models.
}
\examples{
\donttest{
model <- tl_model(iris, Species ~ ., method = "forest")
tl_table_confusion(model)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-hclust.R
\name{tidy_cutree}
\alias{tidy_cutree}
\title{Cut Hierarchical Clustering Tree}
\usage{
tidy_cutree(hclust_obj, k = NULL, h = NULL)
}
\arguments{
\item{hclust_obj}{A tidy_hclust object or hclust object}

\item{k}{Number of clusters (optional)}

\item{h}{Height at which to cut (optional)}
}
\value{
A tibble with columns \code{.obs_id} (observation identifier) and
\code{cluster} (integer cluster assignment).
}
\description{
Cut dendrogram to obtain cluster assignments
}
\examples{
\donttest{
hc <- tidy_hclust(USArrests, method = "ward.D2")
clusters <- tidy_cutree(hc, k = 3)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-clustering.R
\name{optimal_clusters}
\alias{optimal_clusters}
\title{Find Optimal Number of Clusters}
\usage{
optimal_clusters(data, max_k = 10, methods = c("silhouette", "gap", "wss"))
}
\arguments{
\item{data}{A data frame or tibble}

\item{max_k}{Maximum k to test (default: 10)}

\item{methods}{Vector of methods: "silhouette", "gap", "wss" (default: all)}
}
\value{
A list of class \code{"optimal_k_results"} containing one or more of:
\itemize{
\item wss: tibble from \code{\link{calc_wss}} (if "wss" method used)
\item silhouette: tibble from \code{\link{tidy_silhouette_analysis}}
(if "silhouette" method used)
\item gap: a \code{tidy_gap} object from \code{\link{tidy_gap_stat}}
(if "gap" method used)
}
}
\description{
Use multiple methods to suggest optimal k
}
\examples{
\donttest{
opt <- optimal_clusters(iris[, 1:4], max_k = 6, methods = "wss")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-backends.R
\name{tl_read_s3}
\alias{tl_read_s3}
\title{Read from Amazon S3}
\usage{
tl_read_s3(source, format = NULL, region = NULL, ...)
}
\arguments{
\item{source}{An S3 URI (e.g., \code{"s3://bucket/path/to/file.csv"}).}

\item{format}{Optional format override for the downloaded file. If
\code{NULL}, auto-detected from the S3 key extension.}

\item{region}{AWS region. If \code{NULL}, uses the default from your AWS
configuration.}

\item{...}{Additional arguments passed to the format-specific reader.}
}
\value{
A \code{tidylearn_data} object containing the downloaded data.
}
\description{
Downloads a file from an S3 bucket and reads it into a \code{tidylearn_data}
object. The file format is auto-detected from the key's extension, or can be
specified explicitly. Requires the \pkg{paws.storage} package and valid AWS
credentials.
}
\examples{
\donttest{
# data <- tl_read_s3("s3://my-bucket/data/sales.csv")
# data <- tl_read_s3("s3://my-bucket/data/results.parquet")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tuning.R
\name{tl_tune_grid}
\alias{tl_tune_grid}
\title{Tune hyperparameters for a model using grid search}
\usage{
tl_tune_grid(
  data,
  formula,
  method,
  param_grid,
  folds = 5,
  metric = NULL,
  maximize = NULL,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame containing the training data}

\item{formula}{A formula specifying the model}

\item{method}{The modeling method to tune}

\item{param_grid}{A named list of parameter values
to tune}

\item{folds}{Number of cross-validation folds}

\item{metric}{Metric to optimize}

\item{maximize}{Logical; whether to maximize (TRUE)
or minimize (FALSE) the metric}

\item{verbose}{Logical; whether to print progress}

\item{...}{Additional arguments passed to tl_model}
}
\value{
A tidylearn model object fitted with the best hyperparameters.
Tuning results are stored as an attribute \code{"tuning_results"},
a list containing \code{param_grid}, \code{results} (data frame of
all evaluated combinations), \code{best_params}, \code{best_metric},
\code{metric}, and \code{maximize}.
}
\description{
Tune hyperparameters for a model using grid search
}
\examples{
\donttest{
model <- tl_tune_grid(iris, Species ~ ., method = "tree",
  param_grid = list(cp = c(0.01, 0.1), minsplit = c(10, 20)),
  folds = 2, verbose = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supervised-trees.R
\name{tidylearn-trees}
\alias{tidylearn-trees}
\title{Tree-based Methods for tidylearn}
\description{
Decision trees, random forests, and boosting functionality
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visualization.R
\name{plot_clusters}
\alias{plot_clusters}
\title{Plot Clusters in 2D Space}
\usage{
plot_clusters(
  data,
  cluster_col = "cluster",
  x_col = NULL,
  y_col = NULL,
  centers = NULL,
  title = "Cluster Plot",
  color_noise_black = TRUE
)
}
\arguments{
\item{data}{A data frame with cluster assignments}

\item{cluster_col}{Name of cluster column (default: "cluster")}

\item{x_col}{X-axis variable (if NULL, uses first numeric column)}

\item{y_col}{Y-axis variable (if NULL, uses second numeric column)}

\item{centers}{Optional data frame of cluster centers}

\item{title}{Plot title}

\item{color_noise_black}{If TRUE, color noise points (cluster 0) black}
}
\value{
A \code{\link[ggplot2]{ggplot}} object.
}
\description{
Visualize clustering results using first two dimensions
or specified dimensions
}
\examples{
\donttest{
km <- tidy_kmeans(iris[, 1:4], k = 3)
clustered <- augment_kmeans(km, iris[, 1:4])
plot_clusters(clustered)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{tidy_gap_stat}
\alias{tidy_gap_stat}
\title{Tidy Gap Statistic}
\usage{
tidy_gap_stat(data, FUN_cluster = NULL, max_k = 10, B = 50, nstart = 25)
}
\arguments{
\item{data}{A data frame or tibble}

\item{FUN_cluster}{Clustering function (default: uses kmeans internally)}

\item{max_k}{Maximum number of clusters (default: 10)}

\item{B}{Number of bootstrap samples (default: 50)}

\item{nstart}{If using kmeans, number of random starts (default: 25)}
}
\value{
A list of class \code{"tidy_gap"} containing:
\itemize{
\item gap_data: tibble with gap statistics for each k
\item k_firstSEmax: optimal k via firstSEmax method (most conservative)
\item k_globalmax: optimal k via globalmax method
\item k_firstmax: optimal k via firstmax method
\item recommended_k: recommended k (uses firstSEmax)
\item model: the \code{\link[cluster]{clusGap}} result
}
}
\description{
Compute gap statistic for determining optimal number of clusters
}
\examples{
\donttest{
gap <- tidy_gap_stat(iris[, 1:4], max_k = 6, B = 10)
gap$recommended_k
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-market-basket.R
\name{tidy_rules}
\alias{tidy_rules}
\title{Convert Association Rules to Tidy Tibble}
\usage{
tidy_rules(rules)
}
\arguments{
\item{rules}{A rules object from arules}
}
\value{
A tibble with columns \code{rule_id}, \code{lhs}, \code{rhs}, and
quality measures (e.g., \code{support}, \code{confidence}, \code{lift}).
}
\description{
Convert Association Rules to Tidy Tibble
}
\examples{
\donttest{
if (requireNamespace("arules", quietly = TRUE)) {
  data("Groceries", package = "arules")
  rules_obj <- arules::apriori(Groceries,
    parameter = list(supp = 0.001, conf = 0.5))
  rules_tbl <- tidy_rules(rules_obj)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/workflows.R
\name{tl_transfer_learning}
\alias{tl_transfer_learning}
\title{Transfer Learning Workflow}
\usage{
tl_transfer_learning(
  data,
  formula,
  pretrain_method = "pca",
  supervised_method = "logistic",
  ...
)
}
\arguments{
\item{data}{Training data}

\item{formula}{Model formula}

\item{pretrain_method}{Pre-training method: "pca", "autoencoder"}

\item{supervised_method}{Supervised learning method}

\item{...}{Additional arguments}
}
\value{
A list with class \code{"tidylearn_transfer"} containing:
\describe{
\item{pretrain_model}{The fitted dimensionality reduction model.}
\item{supervised_model}{The fitted supervised tidylearn model.}
\item{formula}{The model formula.}
\item{method}{The supervised learning method used.}
}
}
\description{
Use unsupervised pre-training (e.g., autoencoder
features) before supervised learning
}
\examples{
\donttest{
model <- tl_transfer_learning(iris, Species ~ .,
  pretrain_method = "pca", supervised_method = "logistic")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-dbscan.R
\name{explore_dbscan_params}
\alias{explore_dbscan_params}
\title{Explore DBSCAN Parameters}
\usage{
explore_dbscan_params(data, eps_values, minPts_values)
}
\arguments{
\item{data}{A data frame or matrix}

\item{eps_values}{Vector of eps values to test}

\item{minPts_values}{Vector of minPts values to test}
}
\value{
A tibble with columns \code{eps}, \code{minPts}, \code{n_clusters},
\code{n_noise}, and \code{prop_noise} for each parameter combination.
}
\description{
Test multiple eps and minPts combinations
}
\examples{
\donttest{
params <- explore_dbscan_params(iris[, 1:4],
  eps_values = c(0.3, 0.5, 0.8), minPts_values = c(3, 5))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsupervised-validation.R
\name{tidy_silhouette}
\alias{tidy_silhouette}
\title{Tidy Silhouette Analysis}
\usage{
tidy_silhouette(clusters, dist_mat)
}
\arguments{
\item{clusters}{Vector of cluster assignments}

\item{dist_mat}{Distance matrix (dist object)}
}
\value{
A list of class "tidy_silhouette" containing:
\itemize{
\item silhouette_data: tibble with silhouette values for each observation
\item avg_width: average silhouette width
\item cluster_avg: average silhouette width by cluster
}
}
\description{
Compute silhouette statistics for cluster validation
}
\examples{
\donttest{
km <- kmeans(iris[, 1:4], centers = 3, nstart = 25)
d <- dist(iris[, 1:4])
sil <- tidy_silhouette(km$cluster, d)
}

}
